package dk.sosi.sts.integrations.blurring;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.model.BlurredOrganization;
import dk.sosi.sts.model.BlurredOrganization.Type;
import dk.sosi.sts.model.CprNr;
import dk.sundhedsdatastyrelsen.identitetssloering._2023._06._01.GetCurrentSaltRequest;
import dk.sundhedsdatastyrelsen.identitetssloering._2024._01._24.*;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import javax.xml.bind.JAXBException;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.soap.SOAPException;
import java.io.IOException;
import java.util.List;
import java.util.stream.Collectors;

import static dk.sosi.sts.XmlUtil.*;

public class BlurringService {
    private static final Logger logger = Logger.getLogger(BlurringService.class);
    private final BlurringServiceCaller blurringServiceCaller;
    private final SOSIFactory sosiFactory;
    private final IdCardFactory idCardFactory;
    public static final String errorMessage = "Internal error. Call to blurring service failed.";

    public BlurringService(SOSIFactory sosiFactory,
                           IdCardFactory idCardFactory,
                           BlurringServiceCaller blurringServiceCaller) {
        this.blurringServiceCaller = blurringServiceCaller;
        this.sosiFactory = sosiFactory;
        this.idCardFactory = idCardFactory;
    }

    public List<BlurredOrganization> getActiveBlurrings(CprNr patientId, String flowId) {
        String request = createGetActiveBlurringsRequest(patientId, PatientIdClassificationType.CPR, flowId);
        String response = blurringServiceCaller.callGetBlurredOrganisations(request);
        return parseBlurredOrganisationResponse(response);
    }

    public List<BlurredOrganization> getAllActiveOrgBlurrings(String flowId) {
        String request = createListAllActiveOrgBlurrings(flowId);
        String response = blurringServiceCaller.callListActiveOrgBlurrings(request);
        return parseListAllActiveOrgBlurringsResponse(response);
    }

    protected String createGetActiveBlurringsRequest(CprNr patientId,
                                                     PatientIdClassificationType patientIdType,
                                                     String flowId) {
        try {
            Element requestBody = createGetActiveBlurringsRequestBody(patientId, patientIdType).getDocumentElement();
            return createRequest(flowId, requestBody);
        } catch (ParserConfigurationException | JAXBException | IOException | SAXException e) {
            logger.error("Failed to create getActiveBlurringsRequest.", e);
            throw new RuntimeException(errorMessage);
        }
    }

    protected String createListAllActiveOrgBlurrings(String flowId) {
        try {
            Element requestBody = toXml(new ListAllActiveOrgBlurringsRequest()).getDocumentElement();
            return createRequest(flowId, requestBody);
        } catch (ParserConfigurationException | JAXBException | IOException | SAXException e) {
            logger.error("Failed to create listAllActiveOrgBlurrings request.", e);
            throw new RuntimeException(errorMessage);
        }
    }

    protected Document createGetActiveBlurringsRequestBody(CprNr patientId, PatientIdClassificationType patientIdType)
            throws ParserConfigurationException, JAXBException, IOException, SAXException {
        GetBlurredOrganisationsRequest request = new GetBlurredOrganisationsRequest();
        PatientIdType patientIdObject = new PatientIdType();
        patientIdObject.setId(patientId.getValue());
        patientIdObject.setClassification(patientIdType);
        request.setPatientId(patientIdObject);
        return toXml(request);
    }

    private List<BlurredOrganization> parseBlurredOrganisationResponse(String responseXml) {
        try {
            return toModel(toSoapMessage(responseXml).getSOAPBody(), BlurredOrganisationsResponse.class)
                    .getOrganisationId().stream()
                    .map(orgId -> new BlurredOrganization(orgId.getId(),
                                                          toBlurredOrganizationType(orgId.getClassification())))
                    .collect(Collectors.toList());
        } catch (IOException | JAXBException e) {
            logger.error( "Failed to find \"blurred organisations\" in blurring service response. " +
                                  "Response was:\n" + responseXml, e);
            throw new RuntimeException(errorMessage);
        } catch (SOAPException e) {
            logger.error("Error creating MessageFactory.", e);
            throw new RuntimeException(errorMessage);
        }
    }

    private List<BlurredOrganization> parseListAllActiveOrgBlurringsResponse(String responseXml) {
        try {
            return toModel(toSoapMessage(responseXml).getSOAPBody(), ListAllActiveOrgBlurringsResponse.class)
                    .getOrg().stream()
                    .map(orgId -> new BlurredOrganization(orgId.getId(),
                                                          toBlurredOrganizationType(orgId.getClassification())))
                    .collect(Collectors.toList());
        } catch (IOException | JAXBException e) {
            logger.error("Failed to find \"all active org blurrings\" in blurring service response. Response was:\n" + responseXml, e);
            throw new RuntimeException(errorMessage);
        } catch (SOAPException e) {
            logger.error("Error creating MessageFactory.", e);
            throw new RuntimeException(errorMessage);
        }
    }

    private static Type toBlurredOrganizationType(OrganisationIdClassificationType type) {
        switch (type) {
            case CVR:
                return Type.Cvr;
            case SOR:
                return Type.Sor;
            case SHAK:
                return Type.Shak;
            default:
                String error = "Internal error. Could not map value '" + type + "' to type '" + Type.class + "'.";
                logger.error(error);
                throw new RuntimeException(errorMessage);
        }
    }

    public String getSalt(String flowId) {
        String request = createSaltRequest(flowId);
        String response = blurringServiceCaller.callSaltService(request);
        return getSaltFromResponse(response);
    }

    private String createSaltRequest(String flowId)  {
        try {
            Element requestBody = toXml(new GetCurrentSaltRequest()).getDocumentElement();
            return createRequest(flowId, requestBody);
        }
        catch (JAXBException | ParserConfigurationException | SAXException | IOException e) {
            logger.error("Failed to create getSalt request", e);
            throw new RuntimeException(errorMessage);
        }
    }

    private String getSaltFromResponse(String responseXml) {
        try {
            return toModel(toSoapMessage(responseXml).getSOAPBody(), String.class).trim();
        } catch (IOException | JAXBException e) {
            logger.error("Failed to find salt in blurring service response. Response was:\n" + responseXml, e);
            throw new RuntimeException(errorMessage);
        } catch (SOAPException e) {
            logger.error("Error creating MessageFactory.", e);
            throw new RuntimeException(errorMessage);
        }
    }

    private String createRequest(String flowId, Element requestBody)  {
        return idCardFactory.applyIdCard(idCard -> {
            Request request = sosiFactory.createNewRequest(false, flowId);
            request.setIDCard(idCard);
            request.setBody(requestBody);
            // Bemærk, serialize er faktisk her hvor den signerer hele requestet.
            return XmlUtil.node2String(request.serialize2DOMDocument(), false, true);
        });
    }
}