package dk.sosi.sts.integrations.blurring;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.model.BlurredOrganization;
import dk.sosi.sts.model.CprNr;
import dk.sundhedsdatastyrelsen.identitetssloering._2024._01._24.PatientIdClassificationType;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import javax.xml.bind.JAXBException;
import javax.xml.parsers.ParserConfigurationException;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.stream.Collectors;

import static dk.sosi.sts.model.BlurredOrganization.Type.Cvr;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.anyBoolean;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class BlurringServiceTest {
    private String responseXml;

    // Mocks
    SOSIFactory sosiFactory = mock(SOSIFactory.class);
    BlurringServiceCaller blurringServiceCaller = mock(BlurringServiceCaller.class);
    SystemIDCard idCard = mock(SystemIDCard.class);
    Request request = mock(Request.class);
    // Spies
    IdCardFactory idCardFactory = new IdCardFactory(sosiFactory);

    @Before
    public void setUp() throws IOException {
        try (InputStream is = this.getClass().getResourceAsStream("/lookupResponse.xml")) {
            responseXml = new BufferedReader(
                    new InputStreamReader(is, StandardCharsets.UTF_8))
                    .lines()
                    .collect(Collectors.joining("\n"));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        String itSystemName = "itSystemName", careProviderId = "careProviderId", careProviderOrg = "careProviderOrg";
        idCardFactory.setItSystemName(itSystemName);
        idCardFactory.setCareProviderId(careProviderId);
        idCardFactory.setCareProviderOrg(careProviderOrg);
        final CareProvider expectedCareProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, careProviderId, careProviderOrg);

        when(sosiFactory.createNewRequest(anyBoolean(), anyString()))
                .thenReturn(request);
        
        when(sosiFactory.createNewSystemIDCard(eq(itSystemName), eq(expectedCareProvider), any(), any(), any(), any(), any()))
                .thenReturn(idCard);

    }

    @Test
    public void createRequestBody_WhenGivenInput_ReturnsRequestBody() throws JAXBException,
            ParserConfigurationException, IOException, SAXException {
        // Arrange
        BlurringService blurringService = new BlurringService(sosiFactory, idCardFactory, blurringServiceCaller);

        // Act
        Document doc = blurringService.createGetActiveBlurringsRequestBody(new CprNr("1234567890"), PatientIdClassificationType.CPR);

        // Assert
        String s = XmlUtil.node2String(doc, false, true);
        assertTrue(s.contains("<id>1234567890</id>"));
        assertTrue(s.contains("<classification>CPR</classification>"));
    }

    @Test
    public void createRequest_WhenGivenInput_ReturnsRequest() {
        // Arrange
        BlurringService blurringService = new BlurringService(sosiFactory, idCardFactory, blurringServiceCaller);

        // Act
        String res = blurringService.createGetActiveBlurringsRequest(new CprNr("1234567890"),
                                                                     PatientIdClassificationType.CPR,
                                                                     "dummy");
        // Assert
        assertTrue(res.contains("xml"));
    }

    @Test
    public void getActiveBlurrings_WhenCalled_ReturnsListWithSingleCVR() {
        // Arrange
        BlurringService blurringService = new BlurringService(sosiFactory, idCardFactory, blurringServiceCaller);

        // Prepare mocks
        when(blurringServiceCaller.callGetBlurredOrganisations(anyString())).thenReturn(responseXml);

        // Act
        List<BlurredOrganization> res = blurringService.getActiveBlurrings(new CprNr("1234567890"), "dummy");

        // Assert
        assertEquals(new BlurredOrganization("33257872", Cvr), res.get(0));
    }

    String expected = "TkhvHPslclPzA1jyLk+jdg==",
           response = "<soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
            "    <soap:Body>\n" +
            "        <SaltResponse xmlns=\"http://sundhedsdatastyrelsen.dk/identitetssloering/2023/06/01/\">\n" +
            "            " + expected + "</SaltResponse>\n" +
            "    </soap:Body>\n" +
            "</soap:Envelope>";

    @Test
    public void getSalt_WhenCalled_ReturnsSalt() {
        // Prepare mocks
        when(blurringServiceCaller.callSaltService(anyString())).thenReturn(response);

        // Arrange
        BlurringService blurringService = new BlurringService(sosiFactory, idCardFactory, blurringServiceCaller);

        // Act
        String actual = blurringService.getSalt("dummy");

        // Assert
        assertEquals(expected, actual);
    }

    @Test
    public void getSalt_WhenCalled_RequestContainsFlowIdSoTheRequestCanBeTracedBetweenComponents() {
        // Values
        final String flowId = "flowId";

        //Prepare captor
        ArgumentCaptor<String> requestFlowId = ArgumentCaptor.forClass(String.class);

        //Arrange
        BlurringService blurringService = new BlurringService(sosiFactory, idCardFactory, blurringServiceCaller);

        // Act
        try {
            blurringService.getSalt(flowId);
        } catch (Exception ignored) {}

        // Then
        verify(sosiFactory).createNewRequest(anyBoolean(), requestFlowId.capture());
        assertEquals(flowId, requestFlowId.getValue());
    }

    @Test
    public void getBlurrings_WhenCalled_RequestContainsFlowIdSoTheRequestCanBeTracedBetweenComponents() {
        ArgumentCaptor<String> requestFlowId = ArgumentCaptor.forClass(String.class);
        String flowId = "flowId";
        BlurringService blurringService = new BlurringService(sosiFactory, idCardFactory, blurringServiceCaller);

        try { blurringService.getActiveBlurrings(new CprNr("1234567890"), flowId); } catch(Exception ignored) {}

        verify(sosiFactory).createNewRequest(anyBoolean(), requestFlowId.capture());
        assertEquals(flowId, requestFlowId.getValue());
    }
}