package dk.sosi.sts.integrations.blurring;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.model.SystemInfo;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;
import org.w3c.dom.Element;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.time.Instant;
import java.util.Date;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import static org.junit.Assert.*;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.anyString;
import static org.mockito.Mockito.*;
import static org.mockito.Mockito.when;

public class IdCardFactoryTest {

    @Test
    public void getIdCard_WhenCalled_ReturnsTheCreatedIdCard() {
        // Arrange
        SOSIFactory sosiFactory = Mockito.mock(SOSIFactory.class);
        IdCardFactory idCardFactory = new IdCardFactory(sosiFactory);
        idCardFactory.setItSystemName("dummy");
        idCardFactory.setCareProviderId("dummy");
        idCardFactory.setCareProviderOrg("dummy");

        SystemIDCard idCard = Mockito.mock(SystemIDCard.class);
        final CareProvider expectedCareProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "dummy", "dummy");
        Mockito.when(sosiFactory.createNewSystemIDCard(eq("dummy"), eq(expectedCareProvider), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any()))
                .thenReturn(idCard);
        // Act
        idCardFactory.applyIdCard((ic) -> verifySame(idCard, ic));
    }

    @Test
    public void getIdCard_WhenCalledWithItSystemName_UsesItSystemNameAsArgumentToSosiFactory() {
        // Arrange
        SOSIFactory sosiFactory = Mockito.mock(SOSIFactory.class);
        IdCardFactory idCardFactory = new IdCardFactory(sosiFactory);
        String itSystemName = "itSystemName", careProviderId = "id", careProviderOrg = "org";
        SystemIDCard idCard = Mockito.mock(SystemIDCard.class);
        Mockito.when(sosiFactory.createNewSystemIDCard( Mockito.any(),  Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any()))
                .thenReturn(idCard);

        // Act
        idCardFactory.setItSystemName(itSystemName);
        idCardFactory.setCareProviderId(careProviderId);
        idCardFactory.setCareProviderOrg(careProviderOrg);
        idCardFactory.applyIdCard((ic) -> verifySame(idCard, ic));

        // Assert
        ArgumentCaptor<String> argument = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<CareProvider> cpArgument = ArgumentCaptor.forClass(CareProvider.class);
        verify(sosiFactory).createNewSystemIDCard(argument.capture(), cpArgument.capture(), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any());
        assertEquals(itSystemName, argument.getValue());
        assertEquals(careProviderId, cpArgument.getValue().getID());
        assertEquals(careProviderOrg, cpArgument.getValue().getOrgName());
    }

    @Test
    public void getIdCard_WhenCalled_ThenExpiredIdCardShouldBeRenewed() throws ParserConfigurationException {
        SOSIFactory sosiFactory = mock(SOSIFactory.class);
        Instant now = Instant.now();
        SystemIDCard expiredIdCard = createSystemIdCard(Date.from(now.minusSeconds(300)));
        when(sosiFactory.createNewSystemIDCard(anyString(), any(), any(), anyString(), anyString(), any(), anyString()))
                .thenReturn(expiredIdCard);
        IdCardFactory idCardFactory = new IdCardFactory(sosiFactory);
        idCardFactory.setCareProviderId("id");
        idCardFactory.setCareProviderOrg("org");

        idCardFactory.applyIdCard((ic) -> verifySame(expiredIdCard, ic));// Update the cache in IdCardFactory with the expired ID card from the SOSIFactory mock

        SystemIDCard validIdCard = createSystemIdCard(Date.from(now.plusSeconds(300)));
        when(sosiFactory.createNewSystemIDCard(anyString(), any(), any(), anyString(), anyString(), any(), anyString()))
                .thenReturn(validIdCard);

        idCardFactory.applyIdCard((ic) -> verifySame(validIdCard, ic));
    }

    @Test
    public void getIdCard_WhenCalledByTwoThreads_ThenTheIdCardsShouldNotBeTheSame() throws InterruptedException {
        SOSIFactory sosiFactory = mock(SOSIFactory.class);
        IdCardFactory idCardFactory = new IdCardFactory(sosiFactory);
        idCardFactory.setCareProviderId("id");
        idCardFactory.setCareProviderOrg("org");

        SystemIDCard idCard1 = Mockito.mock(SystemIDCard.class);
        SystemIDCard idCard2 = Mockito.mock(SystemIDCard.class);
        Mockito.when(sosiFactory.createNewSystemIDCard(eq("system1"), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any())).thenReturn(idCard1);
        Mockito.when(sosiFactory.createNewSystemIDCard(eq("system2"), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any(), Mockito.any())).thenReturn(idCard2);
        when(idCard1.getExpiryDate()).thenReturn(Date.from(Instant.now().plusSeconds(1000)));
        when(idCard2.getExpiryDate()).thenReturn(Date.from(Instant.now().plusSeconds(1000)));

        AtomicInteger mutex = new AtomicInteger();
        ExecutorService threadPool = Executors.newFixedThreadPool(2);
        Future<Boolean> system1 = threadPool.submit(() -> {
            idCardFactory.setItSystemName("system1");
            return idCardFactory.applyIdCard(idCard -> {
                assertEquals(idCard1, idCard);
                mutex.set(1);
                while (mutex.get() == 1) {
                    try {
                        Thread.sleep(100);
                    } catch (InterruptedException ignored) {
                    }
                }
                return true;
            });
        });

        Future<Boolean> system2 = threadPool.submit(() -> {
            while (mutex.get() != 1) {
                try {
                    Thread.sleep(100);
                } catch (InterruptedException ignored) {
                }
            }
            idCardFactory.setItSystemName("system2");
            return idCardFactory.applyIdCard(idCard -> {
                assertEquals(idCard2, idCard);
                mutex.set(2);
                return true;
            });
        });
        try {
            assertTrue(system1.get());
            assertTrue(system2.get());
        } catch (ExecutionException e) {
            if (e.getCause() instanceof AssertionError) {
                throw (AssertionError) e.getCause();
            } else {
                throw new RuntimeException(e.getCause());
            }
        }

        threadPool.shutdown();
        assertTrue(threadPool.awaitTermination(10, TimeUnit.SECONDS));
    }

    private static SystemIDCard createSystemIdCard(Date expiryDate) throws ParserConfigurationException {
        SystemInfo systemInfo =
                new SystemInfo(new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "cvr", "navn"), "TE-STS");
        Element emptyElement =
                DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument().getDocumentElement();
        return new SystemIDCard("1.0.1",
                                emptyElement,
                                "cardId",
                                AuthenticationLevel.VOCES_TRUSTED_SYSTEM,
                                "certHash",
                                "issuer",
                                systemInfo,
                                null,
                                expiryDate,
                                "alternativeIdentifier",
                                "user",
                                "pass");
    }

    public String verifySame(SystemIDCard expected, SystemIDCard actual) {
        assertSame(expected, actual);
        return null;
    }
}