package dk.sosi.sts.server.facade.bootstrap;

import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.OIOBootstrapToken;
import dk.sosi.seal.model.SignatureConfiguration;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.model.dombuilders.AbstractSAMLBuilder;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import java.util.Date;
import java.util.Properties;

/**
 * Created by Kent on 20-03-2017.
 */
public class BootstrapTokenBuilder extends AbstractSAMLBuilder<BootstrapTokenBuilder, OIOBootstrapToken> {
    public static final int ONE_HOUR = 1000 * 60 * 60;
    private final String nameId;

    private final CredentialVault signingVault;
    private Node subjectNode;

    private final String audienceRestriction = "https://bootstrap.sts.nemlog-in.dk/";

    private final Date created = new Date(System.currentTimeMillis() - ONE_HOUR);
    private Date expires = new Date(System.currentTimeMillis() + ONE_HOUR);

    public BootstrapTokenBuilder(String nameId) {
        setIssuer("https://saml.nemlog-in.dk");
        this.signingVault = getExternalTrustedIdp();
        this.nameId = nameId;
        validate("nameId", nameId);
    }

    private static CredentialVault getExternalTrustedIdp() {
        return new ClasspathCredentialVault(new Properties(), "test-external-trusted-idp.jks", "Test1234");
    }

    public Element buildDom() {
        Document document = createDocument();
        SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[]{assertionID}, assertionID, IDValues.Id);
        signatureConfiguration.setSignatureSiblingNode(subjectNode);
        SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(signingVault), document, signatureConfiguration);
        return document.getDocumentElement();
    }

    @Override
    public OIOBootstrapToken build() throws ModelException {
        return new OIOBootstrapToken(buildDom());
    }

    @Override
    protected final void appendToRoot(Document doc, Element root) {
        super.appendToRoot(doc, root);
        subjectNode = createSubject();
        root.appendChild(subjectNode);
        root.appendChild(createConditions());
        root.appendChild(createAuthnStatement());
        root.appendChild(createAttributeStatement());
    }

    private Node createSubject() {
        Element subjectElm = createElement(SAMLTags.subject);

        Element nameIdElm = createElement(SAMLTags.nameID);
        nameIdElm.setAttribute(SAMLAttributes.FORMAT, SAMLValues.NAMEID_FORMAT_X509_SUBJECT_NAME);
        nameIdElm.setTextContent(nameId);
        subjectElm.appendChild(nameIdElm);

        Element subjectConfirmationElm = createElement(SAMLTags.subjectConfirmation);
        subjectConfirmationElm.setAttribute(SAMLAttributes.METHOD, SAMLValues.CONFIRMATION_METHOD_BEARER);
        subjectElm.appendChild(subjectConfirmationElm);

        Element subjectConfirmationDataElm = createElement(SAMLTags.subjectConfirmationData);
        // TODO should really add certificate here
        subjectConfirmationElm.appendChild(subjectConfirmationDataElm);

        return subjectElm;
    }

    private Node createConditions() {
        Element conditionsElm = createElement(SAMLTags.conditions);
        conditionsElm.setAttribute(SAMLAttributes.NOT_ON_OR_AFTER, XmlUtil.getDateFormat(true).format(expires));

        Element audienceRestrictionElm = createElement(SAMLTags.audienceRestriction);
        conditionsElm.appendChild(audienceRestrictionElm);

        Element audienceElm = createElement(SAMLTags.audience);
        audienceElm.setTextContent(audienceRestriction);
        audienceRestrictionElm.appendChild(audienceElm);

        return conditionsElm;
    }

    private Node createAuthnStatement() {
        Element authnStatementElm = createElement(SAMLTags.authnStatement);
        authnStatementElm.setAttribute(SAMLAttributes.AUTHN_INSTANT, XmlUtil.getDateFormat(true).format(this.created));

        Element authnContextElm = createElement(SAMLTags.authnContext);
        Element authContextClassRefElm = createElement(SAMLTags.authnContextClassRef);
        authContextClassRefElm.setTextContent(SAMLValues.AUTHN_CONTEXT_CLASS_REF);
        authnContextElm.appendChild(authContextClassRefElm);
        authnStatementElm.appendChild(authnContextElm);
        // SessionIndex is mandatory in OIOSAML (to support per session logout), but is meaningless
        // for STS issued assertions
        authnStatementElm.setAttribute(SAMLAttributes.SESSION_INDEX, assertionID);
        return authnStatementElm;
    }

    private Element createAttributeStatement() {
        Element attributeStatementElm = createElement(SAMLTags.attributeStatement);
        final Element spEntityId = createAttributeElement("dk:nemlogin:saml:attribute:SpEntityId", null, "https://saml.test1.fmk.netic.dk/fmk/");
        attributeStatementElm.appendChild(spEntityId);
        return attributeStatementElm;
    }
}
