/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.bootstrap;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.BasicPrivileges;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.OIOBootstrapToIdentityTokenResponse;
import dk.sosi.seal.model.OIOBootstrapToken;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.testtools.testrequest.SoapClient;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Properties;

import static org.junit.Assert.*;

public class BootstrapTokenProcurationIT {
    public static final String STS_HOST = System.getProperty("sts.url","http://test1-cnsp.ekstern-test.nspop.dk:8080");

    protected final Federation federation = new SOSITestFederation(new Properties());

    private SoapClient soapClient;
    private Element bootstrapTokenDom;
    private final OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();

    public static GenericCredentialVault getValidWhitelistedVault() {
        byte[] pkcs12 = CredentialVaultTestUtil.readResource("oces2/PP/FOCES_gyldig.p12");
        return CredentialVaultTestUtil.getCredentialVaultFromPKCS12(new Properties(), pkcs12, "Test1234");
    }

    @Before
    public void initTestClient() throws MalformedURLException {
        soapClient = new SoapClient(new URL(STS_HOST + "/sts/services/Bst2Idws"));
    }

    @Before
    public void prepareSimulatedBootstrapToken() {
        final String currentUserNameId = "C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=PID:9208-2002-2-514358910503";
        this.bootstrapTokenDom = new BootstrapTokenBuilder(currentUserNameId).buildDom();
    }

    @Test
    public void testValidRequestWithProcurations() {
        // bootstrap token dom is normally fetched by the citizen authenticating himself with NemLog-in
        OIOBootstrapToken bootstrapToken = new OIOBootstrapToken(this.bootstrapTokenDom);
        OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();
        OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = oiosamlFactory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
        requestBuilder.setOIOBootstrapToken(bootstrapToken);
        requestBuilder.setSigningVault(getValidWhitelistedVault());
        requestBuilder.setAudience("http://audience/clear");
        requestBuilder.setCPRNumberClaim("0501792275");
        requestBuilder.setProcurationByCprClaim("1111111118");
        Document requestDocument = requestBuilder.build();

        final String requestString = XmlUtil.node2String(requestDocument, false, false);
        final Document responseDoc = soapClient.execute(requestString);

        OIOBootstrapToIdentityTokenResponse response = oiosamlFactory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(responseDoc);
        assertFalse(response.getFaultString(), response.isFault());
        response.validateSignature();
        response.validateSignatureAndTrust(federation);

        final IdentityToken identityToken = response.getIdentityToken();

        // identity token can be used to authenticate against backend service
        assertEquals("0501792275", identityToken.getCpr());
        assertEquals("http://audience/clear", identityToken.getAudienceRestriction());
        assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
        assertEquals("DK-SAML-2.0", identityToken.getAttribute("dk:gov:saml:attribute:SpecVer"));
        assertEquals("3", identityToken.getAssuranceLevel());
        final BasicPrivileges privileges = identityToken.getPrivileges();
        final String privilegeScope = "urn:dk:healthcare:saml:actThroughProcurationBy:cprNumberIdentifier:1111111118";
        assertEquals(1, privileges.getPrivileges(privilegeScope).size());
        assertTrue(privileges.getPrivileges(privilegeScope).contains("urn:dk:nspop:sts:read"));
    }

    @Test
    public void testErrorWhenClaimsNotFulfilled() {
        OIOBootstrapToken bootstrapToken = new OIOBootstrapToken(this.bootstrapTokenDom);
        OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = oiosamlFactory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
        requestBuilder.setOIOBootstrapToken(bootstrapToken);
        requestBuilder.setSigningVault(getValidWhitelistedVault());
        requestBuilder.setAudience("http://audience/clear");
        requestBuilder.setCPRNumberClaim("0501792275");
        requestBuilder.setProcurationByCprClaim("1111111119");
        Document requestDocument = requestBuilder.build();
        final String requestString = XmlUtil.node2String(requestDocument, false, false);
        final Document responseDoc = soapClient.execute(requestString);
        OIOBootstrapToIdentityTokenResponse response = oiosamlFactory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(responseDoc);
        assertTrue("Not a fault", response.isFault());
        assertEquals("dk:sosi:sts:its", response.getFaultActor());
        assertTrue(response.getFaultString(), response.getFaultString().contains("No privileges provided by procuration"));
    }
}
