/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server.facade.jwt;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.BasicPrivileges;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.JWTToIdentityTokenResponse;
import dk.sosi.seal.model.dombuilders.JWTToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.testtools.testrequest.SoapClient;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;
import java.util.Properties;

import static org.junit.Assert.*;

public class JwtExchangeIT {
//    public static final String STS_HOST = System.getProperty("sts.url","http://test1-cnsp.ekstern-test.nspop.dk:8080");
    public static final String STS_HOST = System.getProperty("sts.url","http://localhost:8080");

    private SoapClient soapClient;
    private final OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();
    private String jwtToken;


    @Before
    public void initTestClient() throws MalformedURLException {
        soapClient = new SoapClient(new URL(STS_HOST + "/sts/services/JWT2Idws"));
    }

    // we create a JWT token, signed by our test "idp"
    // fortunately that idp is trusted in test environments
    @Before
    public void prepareSimulatedBootstrapToken() {
        // pid and cpr matches
        // this user also has 'fuldmagter' for 0101603040 and 1111111118
        String cpr = "0501792275";
        final String subjectDN = "C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=PID:9208-2002-2-514358910503";
        JWTTokenBuilder tokenBuilder = new JWTTokenBuilder(new FakeIdp(), subjectDN, cpr);
        jwtToken = tokenBuilder.build();
    }


    @Test
    public void testValidRequestWithCprInJWT() {
        // a test "audience"
        final String audience = "http://audience/clear";
        JWTToIdentityTokenRequestDOMBuilder requestBuilder = new OIOSAMLFactory().createJWTToIdentityTokenRequestDOMBuilder();
        requestBuilder.setJWTToken(jwtToken);
        requestBuilder.setSigningVault(getValidWhitelistedVault());
        requestBuilder.setAudience(audience);
        // request fejler, hvis brugeren ikke har fuldmagt, og dette claimes
        requestBuilder.setProcurationByCprClaim("1111111118");

        Document requestDocument = requestBuilder.build();

        final String requestString = XmlUtil.node2String(requestDocument, false, false);
        final Document responseDoc = soapClient.execute(requestString);

        JWTToIdentityTokenResponse response = oiosamlFactory.createJWTToIdentityTokenResponseModelBuilder().build(responseDoc);
        assertFalse(response.getFaultString(), response.isFault());

        final IdentityToken identityToken = response.getIdentityToken();
        assertEquals("0501792275", identityToken.getCpr());
        final BasicPrivileges privileges = identityToken.getPrivileges();
        if (privileges != null) {
            final List<String> privilegesForUser = privileges.getPrivileges("urn:dk:healthcare:saml:actThroughProcurationBy:cprNumberIdentifier:1111111118");

            System.out.println("Privileges: " + privilegesForUser.size());
            for (String privilege: privilegesForUser) {
                System.out.println("- " + privilege);
            }
        }

        //System.out.println(XmlUtil.node2String(responseDoc, true, true));
    }

    // a webservice consumer which is whitelisted to the test audience
    public static GenericCredentialVault getValidWhitelistedVault() {
        byte[] pkcs12 = CredentialVaultTestUtil.readResource("oces2/PP/FOCES_gyldig.p12");
        return CredentialVaultTestUtil.getCredentialVaultFromPKCS12(new Properties(), pkcs12, "Test1234");
    }
}
