/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ibo;

import dk.sosi.sts.model.IboConfig;
import dk.sosi.sts.service.IboConfigService;
import dk.sosi.sts.util.CertificateUtil;
import org.apache.log4j.Logger;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.RowMapper;

import javax.sql.DataSource;
import java.security.cert.X509Certificate;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

/**
 *
 */
public class DBIboConfigService implements IboConfigService {

	private static final String DEFAULT_SELECT = "SELECT publicKey, recipientURL, includeBST, deliveryNotOnOrAfterOffset, notBeforeOffset, notOnOrAfterOffset, idCardMaxAgeMins FROM iboConfig WHERE audience = ?";
	public static final int IDCARD_MAX_AGE_DEFAULT_MINS = 24 * 60;

	private final Logger log = Logger.getLogger(DBIboConfigService.class);

	private String select;
	private JdbcTemplate jdbcTemplate;

	public void setDataSource(final DataSource dataSource) {
		this.jdbcTemplate = new JdbcTemplate(dataSource);
	}

	public void setSelect(final String select) {
		this.select = select;
	}

	@Override
	public IboConfig getAudienceConfiguration(final String audience) {
		if (audience == null) return null;

		final List<IboConfig> result =
				jdbcTemplate.query(chooseSelect(),
								   new RowMapper<IboConfig>() {
									   @Override
									   public IboConfig mapRow(final ResultSet rs, final int rowNum) throws SQLException {
										   final String publicKey = rs.getString("publicKey");
										   final String recipientURL = rs .getString("recipientURL");
										   final Boolean includeBST = rs.getBoolean("includeBST");
										   final long deliveryNotOnOrAfterOffset = rs.getLong("deliveryNotOnOrAfterOffset");
										   final long notBeforeOffset = rs.getLong("notBeforeOffset");
										   final long notOnOrAfterOffset = rs.getLong("notOnOrAfterOffset");
										   long idCardMaxAgeMins = rs.getLong("idCardMaxAgeMins");
										   if (rs.wasNull()) {
											   idCardMaxAgeMins = IDCARD_MAX_AGE_DEFAULT_MINS;
										   }
										   return new IboConfigImpl(audience, parseCertificate(publicKey),
																	recipientURL,
																	includeBST,
																	deliveryNotOnOrAfterOffset,
																	notBeforeOffset,
																	notOnOrAfterOffset, idCardMaxAgeMins);
									   }
								   },
								   audience);
		return result.isEmpty() ? null : result.get(0);
	}

	private String chooseSelect() {
		return select == null ? DEFAULT_SELECT : select;
	}

	/**
	 * Try to make more sense of the string from the database
	 * @param input the raw database string
	 * @return the PublicKey extracted from the Base64 encoded string
	 */
	private X509Certificate parseCertificate(String input) {
		if (input == null) {
			throw new RuntimeException("No publicKey found!");
		}
		return CertificateUtil.parseCertificateConfiguration(input);
	}


	public static class IboConfigImpl implements IboConfig {
		private final X509Certificate targetCertificate;
		private final String recipientURL;
		private final Boolean includeBootStrapToken;
		private final long deliveryNotOnOrAfterOffset;
		private final long notBeforeOffset;
		private final long notOnOrAfterOffset;
		private final String audience;
		private final long idCardMaxAgeMins;

		public IboConfigImpl(String audience, final X509Certificate targetCertificate, final String url, final Boolean includeBST, final long deliveryNotOnOrAfterOffset, final long notBeforeOffset, final long notOnOrAfterOffset, long idCardMaxAgeMins) {
			this.audience = audience;
			this.targetCertificate = targetCertificate;
			this.recipientURL = url;
			this.includeBootStrapToken = includeBST;
			this.deliveryNotOnOrAfterOffset = deliveryNotOnOrAfterOffset;
			this.notBeforeOffset = notBeforeOffset;
			this.notOnOrAfterOffset = notOnOrAfterOffset;
			this.idCardMaxAgeMins = idCardMaxAgeMins;
		}


		@Override
		public String getAudience() {
			return audience;
		}

		@Override
		public X509Certificate getTargetCertificate() {
			return targetCertificate;
		}

		@Override
		public String getRecipientURL() {
			return recipientURL;
		}

		@Override
		public boolean includeBootStrapToken() {
			return includeBootStrapToken;
		}

		@Override
		public long getDeliveryNotOnOrAfterOffset() {
			return deliveryNotOnOrAfterOffset;
		}

		@Override
		public long getNotBeforeOffset() {
			return notBeforeOffset;
		}

		@Override
		public long getNotOnOrAfterOffset() {
			return notOnOrAfterOffset;
		}

		@Override
		public long getIdCardMaxAgeMinutes() {
			return idCardMaxAgeMins;
		}
	}
}
