/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ibo;

import java.net.URI;
import java.net.URISyntaxException;

public class IboConventions {

	/**
	 * This method will normalize the audience such that comparison in the
	 * database is possible wrt. to this normalization.
	 *
	 * The method has a fixpoint.
	 *
	 * @param audience the (potential) non-normalized form
	 * @return the normalized form under this particular normalization
	 *         or null if audience is null or does not contain valid URI.
	 */
	public static String normalizeAudience(String audience) {
		if (audience == null) return null;

		final URI uri;
		try {
			uri = new URI(audience);
		} catch (URISyntaxException e) {
			return null;
		}
		final String scheme = handleScheme(uri.getScheme());
		final String authority = handleAuthority(uri.getAuthority());
		final String path = handlePath(uri.getPath());
		final String query = orEmpty(uri.getQuery(), "?");
		final String fragment = orEmpty(uri.getFragment(), "#");

		return scheme + authority + path + query + fragment;
	}

	/**
	 * No path is equivalent to null, the empty string or a single slash.
	 * All paths longer than one character should start with slash
	 */
	private static String handlePath(final String path) {
		if (path == null || "".equals(path) || "/".equals(path)) return "";
		if (!path.startsWith("/")) return "/" + path;
		return path;
	}

	/**
	 * The authority is prepended '//' if there is parsed an authority.
	 */
	private static String handleAuthority(final String authority) {
		if (authority == null) return "";
		return "//" + authority.toLowerCase();
	}

	/**
	 * Generic method for prefixing the content if such exists.
	 */
	private static String orEmpty(String str, String prefix) {
		if (str == null) return "";
		else return prefix + str;
	}

	/**
	 * Schemes if exists is postfixed a colon.
	 */
	private static String handleScheme(String scheme) {
		if (scheme == null) return "";
		else return scheme.toLowerCase() + ":";
	}

}

