/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ibo;

import dk.sosi.seal.model.ModelException;
import dk.sosi.sts.model.IboConfig;
import junit.framework.TestCase;
import org.hsqldb.jdbc.JDBCDataSource;
import org.springframework.jdbc.core.JdbcTemplate;

import java.sql.SQLException;

/**
 *
 */
public class DBIboConfigServiceTest extends TestCase {

	private JDBCDataSource dataSource;
	public static final String DDL_CREATE =
			"CREATE TABLE iboConfig (" +
			"    audience VARCHAR(255) NOT NULL PRIMARY KEY," +
			"    publicKey TEXT NOT NULL," +
			"    recipientURL VARCHAR(255) NOT NULL," +
			"    includeBST BOOLEAN NOT NULL," +
			"    deliveryNotOnOrAfterOffset BIGINT," +
			"    notBeforeOffset BIGINT," +
			"    notOnOrAfterOffset BIGINT, idCardMaxAgeMins BIGINT default null)"
	;
	public static final String DDL_DROP = "DROP TABLE iboConfig IF EXISTS" ;

	private DBIboConfigService iboConfig;

	@Override
	protected void setUp() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		execute("set database sql syntax PGS true");
		execute(DDL_CREATE);

		iboConfig = new DBIboConfigService();
		iboConfig.setDataSource(dataSource);
	}

	@Override
	protected void tearDown() throws Exception {
		execute(DDL_DROP);
	}

	public void testEmpty() throws Exception {
		assertNull(iboConfig.getAudienceConfiguration("//sundhed.dk"));
	}

	public void testNull() throws Exception {
		assertNull(iboConfig.getAudienceConfiguration(null));
	}

	public void testEntryWithNoValidPublicKey() throws Exception {
		update("insert into iboConfig values (?,?,?,?,?,?,?,?)",
			   "audience",
			   "publicKey",
			   "recipientURL",
			   true,
			   300000,
			   300000,
			   12000000, null);

		try {
			iboConfig.getAudienceConfiguration("audience");
			fail("Should throw exception");
		} catch(ModelException e) {}
	}

	public void testEntry() throws Exception {
		update("insert into iboConfig values (?,?,?,?,?,?,?,?)",
			   "audience",
			   VicValidVocesPublicKey,
			   "recipientURL",
			   true,
			   300000,
			   300000,
			   12000000, null);

		final IboConfig config = iboConfig.getAudienceConfiguration("audience");
		assertNotNull(config);

		assertEquals("recipientURL", config.getRecipientURL());
		assertTrue(config.includeBootStrapToken());
		assertEquals(300000, config.getNotBeforeOffset());
		assertEquals(300000, config.getDeliveryNotOnOrAfterOffset());
		assertEquals(12000000, config.getNotOnOrAfterOffset());
		assertNotNull(config.getTargetCertificate());
		assertEquals(24*60, config.getIdCardMaxAgeMinutes());
	}

	public void testMultipleEntry() throws Exception {
		update("insert into iboConfig values (?,?,?,?,?,?,?,?)",
			   "audience",
			   VicValidVocesPublicKey,
			   "recipientURL",
			   true,
			   300000,
			   300000,
			   12000000, null);
		update("insert into iboConfig values (?,?,?,?,?,?,?,?)",
			   "another",
			   VicValidVocesPublicKey,
			   "recipientURL",
			   false,
			   300000,
			   300000,
			   12000000, null);

		final IboConfig config = iboConfig.getAudienceConfiguration("another");
		assertNotNull(config);

		assertEquals("recipientURL", config.getRecipientURL());
		assertFalse(config.includeBootStrapToken());
		assertEquals(300000, config.getNotBeforeOffset());
		assertEquals(300000, config.getDeliveryNotOnOrAfterOffset());
		assertEquals(12000000, config.getNotOnOrAfterOffset());
		assertNotNull(config.getTargetCertificate());
	}

	public void testIdCardMaxAge() throws Exception {
		update("insert into iboConfig values (?,?,?,?,?,?,?,?)",
			"audience",
			VicValidVocesPublicKey,
			"recipientURL",
			true,
			300000,
			300000,
			12000000, 30);

		final IboConfig config = iboConfig.getAudienceConfiguration("audience");
		assertNotNull(config);

		assertEquals("recipientURL", config.getRecipientURL());
		assertTrue(config.includeBootStrapToken());
		assertEquals(300000, config.getNotBeforeOffset());
		assertEquals(300000, config.getDeliveryNotOnOrAfterOffset());
		assertEquals(12000000, config.getNotOnOrAfterOffset());
		assertNotNull(config.getTargetCertificate());
		assertEquals(30, config.getIdCardMaxAgeMinutes());
	}


	private void execute(String sql) throws SQLException {
		final JdbcTemplate jdbcTemplate = new JdbcTemplate(dataSource);
		jdbcTemplate.execute(sql);
	}

	private void update(String sql, Object... args) {
		final JdbcTemplate jdbcTemplate = new JdbcTemplate(dataSource);
		jdbcTemplate.update(sql, args);
	}


	private static final String VicValidVocesPublicKey = "MIIFDzCCBHigAwIBAgIEQDhGIzANBgkqhkiG9w0BAQUFADA/MQswCQYDVQQGEwJE\n" +
														 "SzEMMAoGA1UEChMDVERDMSIwIAYDVQQDExlUREMgT0NFUyBTeXN0ZW10ZXN0IENB\n" +
														 "IElJMB4XDTEyMTEwNjA4MjAyM1oXDTE0MTEwNjA4NTAyM1owgZUxCzAJBgNVBAYT\n" +
														 "AkRLMS4wLAYDVQQKEyVKRVJOQUxERVJCWUVOUyBWRU5ORVIgLy8gQ1ZSOjE5MzQz\n" +
														 "NjM0MVYwJQYDVQQFEx5DVlI6MTkzNDM2MzQtVUlEOjExNjU4MTM2MTMxODMwLQYD\n" +
														 "VQQDEyZKRVJOQUxERVJCWUVOUyBWRU5ORVIgLSBWaWMgdmFsaWRWT0NFUzCBnzAN\n" +
														 "BgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEA3Cya1Kbm+mxr6mYBjg6k8jbj2dwubJa6\n" +
														 "jjuz2c/I5yGzX4UQpcx5JxBBPbxDXwaCMefNCP9hXmmrjjhCUV+0dEZxRuNCHZDM\n" +
														 "9Ny9EXdBC1nfjzkowaC5V38LHPPGScwM5g0/ZBf6XBf4W+bSoxb4hxeekSkzs6SX\n" +
														 "4j+kM/E/kYsCAwEAAaOCAr8wggK7MA4GA1UdDwEB/wQEAwIDuDArBgNVHRAEJDAi\n" +
														 "gA8yMDEyMTEwNjA4MjAyM1qBDzIwMTQxMTA2MDg1MDIzWjBGBggrBgEFBQcBAQQ6\n" +
														 "MDgwNgYIKwYBBQUHMAGGKmh0dHA6Ly90ZXN0Lm9jc3AuY2VydGlmaWthdC5kay9v\n" +
														 "Y3NwL3N0YXR1czCCAQMGA1UdIASB+zCB+DCB9QYJKQEBAQEBAQEDMIHnMC8GCCsG\n" +
														 "AQUFBwIBFiNodHRwOi8vd3d3LmNlcnRpZmlrYXQuZGsvcmVwb3NpdG9yeTCBswYI\n" +
														 "KwYBBQUHAgIwgaYwChYDVERDMAMCAQEagZdUREMgVGVzdCBDZXJ0aWZpa2F0ZXIg\n" +
														 "ZnJhIGRlbm5lIENBIHVkc3RlZGVzIHVuZGVyIE9JRCAxLjEuMS4xLjEuMS4xLjEu\n" +
														 "MS4zLiBUREMgVGVzdCBDZXJ0aWZpY2F0ZXMgZnJvbSB0aGlzIENBIGFyZSBpc3N1\n" +
														 "ZWQgdW5kZXIgT0lEIDEuMS4xLjEuMS4xLjEuMS4xLjMuMBwGCWCGSAGG+EIBDQQP\n" +
														 "Fg1vcmdhbldlYk5vRGlyMA8GA1UdEQQIMAaBBG51bGwwgZcGA1UdHwSBjzCBjDBX\n" +
														 "oFWgU6RRME8xCzAJBgNVBAYTAkRLMQwwCgYDVQQKEwNUREMxIjAgBgNVBAMTGVRE\n" +
														 "QyBPQ0VTIFN5c3RlbXRlc3QgQ0EgSUkxDjAMBgNVBAMTBUNSTDM2MDGgL6Athito\n" +
														 "dHRwOi8vdGVzdC5jcmwub2Nlcy5jZXJ0aWZpa2F0LmRrL29jZXMuY3JsMB8GA1Ud\n" +
														 "IwQYMBaAFByYCUcaTDi5EMUEKVvx9E6Aasx+MB0GA1UdDgQWBBSlvlxvAUer8dYn\n" +
														 "o4qlIIgyUBrStzAJBgNVHRMEAjAAMBkGCSqGSIb2fQdBAAQMMAobBFY3LjEDAgOo\n" +
														 "MA0GCSqGSIb3DQEBBQUAA4GBAGPrsDmTKTj7yYEq2mG6IuWqaztKPp8fjGAmdvlJ\n" +
														 "THGnHUf1DKYCXwfnSDbjpIRimYnbVAW1nnld60KoPBl4uE2wAbjXYzaBFUMVA9J0\n" +
														 "qkS5iTnQ9vHm3fChHZdLlDNSYBIBnvZ7BD7i/NydkwkObmsvZ6ax9eo8hVk1kPz6\n" +
														 "T6m3";

}
