/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.idp;

import dk.sosi.sts.model.Configuration;
import dk.sosi.sts.model.IdpConfig;
import dk.sosi.sts.service.IdpConfigService;
import dk.sosi.sts.util.DbUtil;

import org.apache.log4j.Logger;

import javax.sql.DataSource;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.Properties;

/**
 * Service for getting database idp configuration
 */
public class DBIdpConfigService implements IdpConfigService {
	private final Logger log = Logger.getLogger(DBIdpConfigService.class);

    private static final String DEFAULT_SELECT_BY_ISSUER = "SELECT attribute, attribute_value FROM trustedIdpConfiguration WHERE issuer = ?";
    private static final String DEFAULT_SELECT_ALL_ENCRYPTION_KEYS = "SELECT attribute_value FROM trustedIdpConfiguration WHERE attribute like 'encryptionKey%'";
    
    private String sqlSelectByIssuer = DEFAULT_SELECT_BY_ISSUER;
    private String sqlSelectAllEncryptionKeys = DEFAULT_SELECT_ALL_ENCRYPTION_KEYS;

    private final DataSource dataSource;

    public DBIdpConfigService(DataSource dataSource) {
		this.dataSource = dataSource;
	}
    
    public void setSelectByIssuer(String sql) {
    	sqlSelectByIssuer = Optional.ofNullable(sql).orElse(DEFAULT_SELECT_BY_ISSUER);
    }
    
    public void setSelectAllEncryptionKeys(String sql) {
    	sqlSelectAllEncryptionKeys = Optional.ofNullable(sql).orElse(DEFAULT_SELECT_ALL_ENCRYPTION_KEYS);
    }

    @Override
    public IdpConfig getIdpConfiguration(String issuer) {
    	Connection conn = null;
        PreparedStatement pstmt = null;
        Properties result = new Properties();
        try {
            conn = dataSource.getConnection();
			pstmt = conn.prepareStatement(sqlSelectByIssuer);
            pstmt.setString(1, issuer);

            ResultSet rs = pstmt.executeQuery();
            while (rs.next()) {
                final String attribute = rs.getString("attribute");
                final String value = rs.getString("attribute_value");
                result.put(attribute, value);
            }
            rs.close();
        } catch (SQLException e) {
            throw new RuntimeException("Error getting idpConfiguration", e);
        } finally {
            DbUtil.close(pstmt, log);
            DbUtil.close(conn, log);
        }

        if (result.isEmpty())
            return null;
        else {
            return new IdpConfigImpl(issuer, new Configuration(result));
        }
    }
    
    @Override
    public List<String> getAllEncryptionKeys() {
    	Connection conn = null;
        PreparedStatement pstmt = null;
        List<String> result = new ArrayList<>();
        try {
            conn = dataSource.getConnection();
			pstmt = conn.prepareStatement(sqlSelectAllEncryptionKeys);
  
            ResultSet rs = pstmt.executeQuery();
            while (rs.next()) {
                final String value = rs.getString("attribute_value");
                result.add(value);
            }
            rs.close();
        } catch (SQLException e) {
        	throw new RuntimeException("Error getting all encryption keys", e);
        } finally {
            DbUtil.close(pstmt, log);
            DbUtil.close(conn, log);
        }

        return result;
    }

}
