package dk.sosi.sts.idp;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import dk.sosi.sts.model.IdpConfig;
import dk.sosi.sts.service.IdpConfigService;

import org.hsqldb.jdbc.JDBCDataSource;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.sql.Connection;
import java.sql.SQLException;

public class DBIdpConfigServiceTest {
	private JDBCDataSource dataSource;
	public static final String DDL_DROP = "DROP TABLE trustedIdpConfiguration IF EXISTS";
	public static final String DDL_CREATE = "CREATE TABLE trustedIdpConfiguration ( " +
			"   issuer VARCHAR(255) NOT NULL," +
			"   attribute VARCHAR(50) NOT NULL," +
			"   attribute_value VARCHAR(4095)," +
			"   comment VARCHAR(255))";
	public static final String DDL_UNIQUE_INDEX = "CREATE UNIQUE INDEX trustedIdpConfiguration_unique ON "
			+ "trustedIdpConfiguration (issuer, attribute)";
	public static final String DDL_INDEX = "CREATE INDEX trustedIdpConfiguration_index ON "
			+ "trustedIdpConfiguration (attribute)";
	
	public static final String[] DDL_EXECUTE = {
		DDL_DROP, 
		DDL_CREATE,
		DDL_UNIQUE_INDEX,
		DDL_INDEX
	};

	@Before
	public void setUp() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		for (String sql : DDL_EXECUTE) {
			execute(sql);
		}
	}
	@After
	public void tearDown() throws Exception {
		execute(DDL_DROP);
	}
	
	private boolean execute(String sql) throws SQLException {
		Connection conn = dataSource.getConnection();
		try {
			return conn.createStatement().execute(sql);
		} finally {
			conn.close();
		}
	}
	
	@Test
	public void testEmptyIssuerConfiguration() throws Exception {
		IdpConfigService service = new DBIdpConfigService(dataSource);
		IdpConfig config = service.getIdpConfiguration("non-existing-issuer");
		assertNull(config);
	}
	
	@Test
	public void testDefaultIssuerConfiguration() throws Exception {
		String[][] rows = {
				{ "issuer","unknown", "true" }
		};
		for (String[] r : rows) {
			execute("INSERT INTO trustedIdpConfiguration (issuer, attribute, attribute_value) VALUES ('" + r[0] + "', '" + r[1] + "', '" + r[2] +"')");
		}

		DBIdpConfigService service = new DBIdpConfigService(dataSource);
		IdpConfig config = service.getIdpConfiguration("issuer");
		
		assertNotNull(config);
		assertEquals("issuer", config.getIssuer());
	}
	
	@Test
	public void testIssuerConfiguration() throws Exception {
		String[][] rows = {
				{ "issuer","signingKey.1", "xyz"},
				{ "issuer","encryptionKey", "abc"},
				{ "issuer","validateHOK", "true"},
				{ "issuer","tokenProfile", "OIOH3BST"},
 			    { "issuer","tokenProfile.1", "OIOH2BST"},
				{ "issuer","certificate.test1", "certificate1"}
		};
		for (String[] r : rows) {
			execute("INSERT INTO trustedIdpConfiguration (issuer, attribute, attribute_value) VALUES ('" + r[0] + "', '" + r[1] + "', '" + r[2] +"')");
		}
		execute("INSERT INTO trustedIdpConfiguration (issuer, attribute, attribute_value, comment) VALUES ('issuer', 'certificate.test2', 'certificate2', 'comment')");

		DBIdpConfigService service = new DBIdpConfigService(dataSource);
		IdpConfig config = service.getIdpConfiguration("issuer");
		
		assertNotNull(config);
		assertEquals("issuer", config.getIssuer());
		assertTrue(config.getSigningKeys().contains("xyz"));
		assertEquals("[abc]", config.getEncryptionKeys().toString());
		assertEquals("true", config.getValidateHOK());
		assertTrue(config.getTokenProfiles().contains("OIOH3BST"));
		assertTrue(config.getTokenProfiles().contains("OIOH2BST"));
		assertTrue(config.getCertificates().contains("certificate1"));
		assertTrue(config.getCertificates().contains("certificate2"));
	}
}
