/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.its;

import dk.sosi.sts.model.AudienceConfiguration;
import dk.sosi.sts.model.BlurringVersion;
import dk.sosi.sts.model.Configuration;

import java.util.List;
import java.util.Properties;
import java.util.stream.Collectors;

/**
 * User: ht
 * Date: 7/28/11
 * Time: 14:22
 */
public class MappedItsConf implements AudienceConfiguration {

    protected final Configuration conf;
	protected final String audience;

    public MappedItsConf(String audience, final Configuration conf) {
        this.conf = conf;
        this.audience = audience;
    }

    @Override
    public long getLifetime() {
        return conf.getLong("lifetime")*1000L;
    }

    @Override
    public boolean requireCertificateAsReference() {
        return conf.getBoolean("certificateAsReference");
    }

    @Override
    public boolean requireCprNumber() {
        return conf.getBoolean("cprNumber");
    }

    @Override
    public boolean requireCvrNumberIdentifier() {
        return conf.getBoolean("cvrNumberIdentifier");
    }

    @Override
    public boolean requireOrganizationName() {
        return conf.getBoolean("organizationName");
    }

    @Override
	public boolean requireITSystemName() {
		return conf.getBoolean("itSystemName");
	}

    @Override
	public boolean requireAuthorizationCode() {
		return conf.getBoolean("authorizationCode");
	}
    
    @Override
	public boolean requireUserEducationCode() {
		return conf.getBoolean("userEducationCode");
	}

	@Override
    public List<String> getCertificates() {
        final Properties properties = conf.getProperties();
        return properties.keySet().stream().map(String::valueOf).filter(obj -> obj.startsWith("certificate")).map(properties::getProperty).collect(Collectors.toList());
    }

    @Override
    public String getJwtScope() {
        return conf.get("jwtScope");
    }

    @Override
    public String getAudience() {
		return audience;
	}

    @Override
	public Long getIdCardMaxAge() {
		final String key = "idCardMaxAge";
		
		if (conf.containsKey(key))
			return conf.getLong(key)*1000L;
		else 
			return null;
	}

    @Override
    public BlurringVersion requireBlurring() {
        String blurring = conf.get("blurring");
        if (blurring == null) {
            return BlurringVersion.NONE;
        }

        switch (blurring) {
            case "V10":
                return BlurringVersion.V10;
            case "V11":
                return BlurringVersion.V11;
            default:
                throw new IllegalArgumentException(String.format("Unknown blurring version: '%s'. Allowed versions are V10 and V11.", blurring));
        }
    }

    @Override
    public List<String> wardClaimAllowedCertificates() {
        final String key = "srp.ward.certificate";
        final Properties properties = conf.getProperties();
        return properties.keySet().stream()
                .map(String::valueOf).filter(obj -> obj.startsWith(key))
                .map(properties::getProperty)
                .collect(Collectors.toList());
    }

}
