/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.its;

import dk.sosi.sts.model.AudienceConfiguration;
import junit.framework.TestCase;
import org.hsqldb.jdbc.JDBCDataSource;
import org.junit.Test;

import java.sql.Connection;
import java.sql.SQLException;

public class DBItsConfigServiceTest extends TestCase {

	private JDBCDataSource dataSource;
	public static final String DDL_DROP = "DROP TABLE audienceConfiguration IF EXISTS";
	public static final String[] DDL_CREATE = {
		DDL_DROP, 
		"CREATE TABLE audienceConfiguration ( " +
		"   audience VARCHAR(255) NOT NULL," +
		"   attribute VARCHAR(50) NOT NULL," +
		"   attribute_value VARCHAR(4095)," +
		"   comment VARCHAR(255))"
	};

	@Override
	protected void setUp() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		for (String sql : DDL_CREATE) {
			execute(sql);
		}
	}
	
	@Override
	protected void tearDown() throws Exception {
		execute(DDL_DROP);
	}
	
	public void testEmptyAudienceConfiguration() throws Exception {
		DBItsConfigService service = new DBItsConfigService(dataSource, 0);
		AudienceConfiguration config = service.getAudienceConfiguration("audience");
		assertNull(config);
	}

	public void testDefaultAudienceConfiguration() throws Exception {
		String[][] rows = {
				{ "audience","unknown", "true", "comment" }
		};
		for (String[] r : rows) {
			execute("INSERT INTO audienceConfiguration " +	"VALUES ('" + r[0] + "', '" + r[1] + "', '" + r[2] +"', '" + r[3] +"')");
		}

		DBItsConfigService service = new DBItsConfigService(dataSource, 0);
		AudienceConfiguration config = service.getAudienceConfiguration("audience");
		
		assertNotNull(config);
		assertEquals("audience", config.getAudience());
		assertEquals(0, config.getLifetime());
		assertNull(config.getIdCardMaxAge());
		assertFalse(config.requireCertificateAsReference());
		assertFalse(config.requireAuthorizationCode());
		assertFalse(config.requireCprNumber());
		assertFalse(config.requireCvrNumberIdentifier());
		assertFalse(config.requireITSystemName());
		assertFalse(config.requireOrganizationName());
		assertFalse(config.requireUserEducationCode());
	}
	
	public void testAudienceConfiguration() throws Exception {
		String[][] rows = {
				{ "audience", "lifetime", "10" },
				{ "audience", "idCardMaxAge", "100" },
				{ "audience","certificateAsReference", "true" },
				{ "audience","cprNumber", "true" },
				{ "audience","cvrNumberIdentifier", "true" },
				{ "audience","organizationName", "true" },
				{ "audience","itSystemName", "true" },
				{ "audience","authorizationCode", "true" }, 
				{ "audience","userEducationCode", "true" }
		};
		for (String[] r : rows) {
			execute("INSERT INTO audienceConfiguration (audience, attribute, attribute_value) " +
					"VALUES ('" + r[0] + "', '" + r[1] + "', '" + r[2] +"')");
		}

		DBItsConfigService service = new DBItsConfigService(dataSource, 0);
		AudienceConfiguration config = service.getAudienceConfiguration("audience");

		assertNotNull(config);
		assertEquals(10000, config.getLifetime());
		assertEquals(100000, config.getIdCardMaxAge().longValue());
		assertTrue(config.requireCertificateAsReference());
		assertTrue(config.requireAuthorizationCode());
		assertTrue(config.requireCprNumber());
		assertTrue(config.requireCvrNumberIdentifier());
		assertTrue(config.requireITSystemName());
		assertTrue(config.requireOrganizationName());
		assertTrue(config.requireUserEducationCode());
	}
	
	public void testCustomSelect() throws Exception {
		String[][] rows = {
				{ "audience","certificateAsReference", "true" },
				{ "audience","cprNumber", "true" },
				{ "audience","cvrNumberIdentifier", "true" },
				{ "audience","organizationName", "true" },
				{ "audience","itSystemName", "true" },
				{ "audience","authorizationCode", "true" }, 
				{ "audience","userEducationCode", "true" }
		};
		for (String[] r : rows) {
			execute("INSERT INTO audienceConfiguration (audience, attribute, attribute_value) " +
					"VALUES ('" + r[0] + "', '" + r[1] + "', '" + r[2] +"')");
		}

		DBItsConfigService service = new DBItsConfigService(dataSource, 0, 
				"SELECT attribute, 'false' as attribute_value FROM audienceConfiguration WHERE audience = ?");
		AudienceConfiguration config = service.getAudienceConfiguration("audience");

		assertNotNull(config);
		assertFalse(config.requireCertificateAsReference());
		assertFalse(config.requireAuthorizationCode());
		assertFalse(config.requireCprNumber());
		assertFalse(config.requireCvrNumberIdentifier());
		assertFalse(config.requireITSystemName());
		assertFalse(config.requireOrganizationName());
		assertFalse(config.requireUserEducationCode());
	}

	@Test
	public void testAudienceWithCertificate() throws SQLException {
		String[][] rows = {
				{ "audience", "certificate.test1", "CVR:30808460-FID:94731315", "comment1" },
				{ "audience", "certificate.test2", "CVR:30808460-FID:94731316", "comment2" }
		};
		for (String[] r : rows) {
			execute("INSERT INTO audienceConfiguration " +	"VALUES ('" + r[0] + "', '" + r[1] + "', '" + r[2] +"', '" + r[3] +"')");
		}

		DBItsConfigService service = new DBItsConfigService(dataSource, 0);
		AudienceConfiguration config = service.getAudienceConfiguration("audience");

		assertNotNull(config);
		assertEquals(2, config.getCertificates().size());
		assertTrue(config.getCertificates().contains("CVR:30808460-FID:94731315"));
		assertTrue(config.getCertificates().contains("CVR:30808460-FID:94731316"));
	}

	private boolean execute(String sql) throws SQLException {
		Connection conn = dataSource.getConnection();
		try {
			return conn.createStatement().execute(sql);
		} finally {
			conn.close();
		}
	}


	private static final String VicValidVocesPublicKey = "MIIFDzCCBHigAwIBAgIEQDhGIzANBgkqhkiG9w0BAQUFADA/MQswCQYDVQQGEwJE\n" +
			"SzEMMAoGA1UEChMDVERDMSIwIAYDVQQDExlUREMgT0NFUyBTeXN0ZW10ZXN0IENB\n" +
			"IElJMB4XDTEyMTEwNjA4MjAyM1oXDTE0MTEwNjA4NTAyM1owgZUxCzAJBgNVBAYT\n" +
			"AkRLMS4wLAYDVQQKEyVKRVJOQUxERVJCWUVOUyBWRU5ORVIgLy8gQ1ZSOjE5MzQz\n" +
			"NjM0MVYwJQYDVQQFEx5DVlI6MTkzNDM2MzQtVUlEOjExNjU4MTM2MTMxODMwLQYD\n" +
			"VQQDEyZKRVJOQUxERVJCWUVOUyBWRU5ORVIgLSBWaWMgdmFsaWRWT0NFUzCBnzAN\n" +
			"BgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEA3Cya1Kbm+mxr6mYBjg6k8jbj2dwubJa6\n" +
			"jjuz2c/I5yGzX4UQpcx5JxBBPbxDXwaCMefNCP9hXmmrjjhCUV+0dEZxRuNCHZDM\n" +
			"9Ny9EXdBC1nfjzkowaC5V38LHPPGScwM5g0/ZBf6XBf4W+bSoxb4hxeekSkzs6SX\n" +
			"4j+kM/E/kYsCAwEAAaOCAr8wggK7MA4GA1UdDwEB/wQEAwIDuDArBgNVHRAEJDAi\n" +
			"gA8yMDEyMTEwNjA4MjAyM1qBDzIwMTQxMTA2MDg1MDIzWjBGBggrBgEFBQcBAQQ6\n" +
			"MDgwNgYIKwYBBQUHMAGGKmh0dHA6Ly90ZXN0Lm9jc3AuY2VydGlmaWthdC5kay9v\n" +
			"Y3NwL3N0YXR1czCCAQMGA1UdIASB+zCB+DCB9QYJKQEBAQEBAQEDMIHnMC8GCCsG\n" +
			"AQUFBwIBFiNodHRwOi8vd3d3LmNlcnRpZmlrYXQuZGsvcmVwb3NpdG9yeTCBswYI\n" +
			"KwYBBQUHAgIwgaYwChYDVERDMAMCAQEagZdUREMgVGVzdCBDZXJ0aWZpa2F0ZXIg\n" +
			"ZnJhIGRlbm5lIENBIHVkc3RlZGVzIHVuZGVyIE9JRCAxLjEuMS4xLjEuMS4xLjEu\n" +
			"MS4zLiBUREMgVGVzdCBDZXJ0aWZpY2F0ZXMgZnJvbSB0aGlzIENBIGFyZSBpc3N1\n" +
			"ZWQgdW5kZXIgT0lEIDEuMS4xLjEuMS4xLjEuMS4xLjMuMBwGCWCGSAGG+EIBDQQP\n" +
			"Fg1vcmdhbldlYk5vRGlyMA8GA1UdEQQIMAaBBG51bGwwgZcGA1UdHwSBjzCBjDBX\n" +
			"oFWgU6RRME8xCzAJBgNVBAYTAkRLMQwwCgYDVQQKEwNUREMxIjAgBgNVBAMTGVRE\n" +
			"QyBPQ0VTIFN5c3RlbXRlc3QgQ0EgSUkxDjAMBgNVBAMTBUNSTDM2MDGgL6Athito\n" +
			"dHRwOi8vdGVzdC5jcmwub2Nlcy5jZXJ0aWZpa2F0LmRrL29jZXMuY3JsMB8GA1Ud\n" +
			"IwQYMBaAFByYCUcaTDi5EMUEKVvx9E6Aasx+MB0GA1UdDgQWBBSlvlxvAUer8dYn\n" +
			"o4qlIIgyUBrStzAJBgNVHRMEAjAAMBkGCSqGSIb2fQdBAAQMMAobBFY3LjEDAgOo\n" +
			"MA0GCSqGSIb3DQEBBQUAA4GBAGPrsDmTKTj7yYEq2mG6IuWqaztKPp8fjGAmdvlJ\n" +
			"THGnHUf1DKYCXwfnSDbjpIRimYnbVAW1nnld60KoPBl4uE2wAbjXYzaBFUMVA9J0\n" +
			"qkS5iTnQ9vHm3fChHZdLlDNSYBIBnvZ7BD7i/NydkwkObmsvZ6ax9eo8hVk1kPz6\n" +
			"T6m3";



}
