/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ocescpr.cvrrid.service.impl;

import dk.sosi.sts.StsException;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;

/**
 * Interface for caching relations between CVR-RID and CPR. The implementation
 * <strong>may</strong> choose to store an opaque representation of the 
 * association, e.g.
 * <pre>
 *   exists(ssn, cpr) == true
 *   find(ssn) == null
 * </pre>
 */
public interface CprCache {
	/**
	 * Find the CPR associated with a CVR-RID
	 * @param ssn the {@link SubjectSerialNumber} to lookup
	 * @return the associated CPR or <code>null</code> if none exist
	 * @throws StsException if an error occurs
	 */
	CprNr find(SubjectSerialNumber ssn) throws StsException;
	
	/**
	 * Check if an association between a CVR-RID and CPR exists 
	 * @param ssn a {@link SubjectSerialNumber} to check 
	 * @param cpr a {@link CprNr} to check
	 * @return <code>true</code> iff the association exists
	 * @throws StsException if an error occurs
	 */
	boolean exists(SubjectSerialNumber ssn, CprNr cpr) throws StsException;

	/**
	 * Create an association between a CVR-RID and a CPR
	 * @param ssn the {@link SubjectSerialNumber} to map
	 * @param cpr the {@link CprNr} to map
	 * @throws StsException if an error occurs
	 */
	void map(SubjectSerialNumber ssn, CprNr cpr) throws StsException;

	/**
	 * Returns the number of cached associations
	 * @return the number of associated pairs in the cache
	 * @throws StsException if an error occurs
	 */
	int size() throws StsException;

	/**
	 * Clear the cache
	 * @throws StsException if an error occurs
	 */
	void clearAll() throws StsException;
}
