/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ocescpr.cvrrid.service.impl;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Timestamp;
import java.util.Calendar;

import javax.sql.DataSource;

import org.apache.log4j.Logger;

import dk.sosi.sts.StsException;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.ocescpr.utilities.HashHelper;
import dk.sosi.sts.util.DbUtil;

public class DbCprHashCache implements CprCache {
	private final Logger log = Logger.getLogger(this.getClass());
	private final DataSource ds;
	
	public DbCprHashCache(DataSource ds) {
		this.ds = ds;
	}
	
	public CprNr find(SubjectSerialNumber ssn) {
		return null;
	}
	
	public boolean exists(SubjectSerialNumber ssn, CprNr cpr) throws StsException {
		PreparedStatement statement = null;
		Connection connection = null;
		ResultSet result = null;
		try {
			connection = ds.getConnection();
			statement = connection.prepareStatement("SELECT createdDate FROM cprhash WHERE hash=?");
			statement.setString(1, HashHelper.sha256Encode(ssn, cpr));
			result = statement.executeQuery();
			if (result.next()) {
				log.debug("Found in cache dated:" + result.getDate(1));
				return true;
			} else {
				return false;
			}
		} catch (Exception e) {
			throw new StsException(e);
		} finally {
			DbUtil.close(connection, statement, result, log);
		}
	}
	
	public synchronized void map(SubjectSerialNumber ssn, CprNr cpr) throws StsException {
		if (exists(ssn,cpr)) return;
		
		Connection connection = null;
		PreparedStatement statement = null;
		try {
			connection = ds.getConnection();
			statement = connection.prepareStatement("INSERT INTO cprhash (hash, createdDate) VALUES (?,?)");
			statement.setString(1, HashHelper.sha256Encode(ssn, cpr));
			statement.setTimestamp(2, new Timestamp(Calendar.getInstance().getTimeInMillis()));
			statement.execute();
		} catch (Exception e) {
			throw new StsException(e);
		} finally {
			DbUtil.close(connection, statement, null, log);
		}
	}

	public int size() throws StsException {
		Connection connection = null;
		PreparedStatement statement = null;
		ResultSet result = null;
		try {
			connection = ds.getConnection();
			statement = connection.prepareStatement("SELECT COUNT(hash) FROM cprhash");
			result = statement.executeQuery();
			if (result.next()) {
				return result.getInt(1);
			} else {
				return 0;
			}
		} catch (Exception e) {
			throw new StsException(e);
		} finally {
			DbUtil.close(connection, statement, result, log);
		}
	}
	
	public void clearAll() throws StsException {
		Connection connection = null;
		PreparedStatement statement = null;
		try {
			connection = ds.getConnection();
			statement = connection.prepareStatement("DELETE FROM cprhash");
			statement.execute();
		} catch (Exception e) {
			throw new StsException(e);
		} finally {
			DbUtil.close(connection, statement, null, log);
		}
	}
	
}
