/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ocescpr.cvrrid.service.impl;

import java.net.URL;
import java.rmi.RemoteException;
import java.util.concurrent.atomic.AtomicLong;

import javax.xml.rpc.ServiceException;

import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.sts.ServiceInfoCollector;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.ocescpr.cvrrid.service.impl.ssl.SSLClientAxisEngineConfig;
import org.apache.axis.EngineConfiguration;
import org.apache.axis.client.Stub;
import org.apache.log4j.Logger;

import localhost.HandleSundhedsportalWSLocator;
import localhost.HandleSundhedsportalWSPort;
import dk.sosi.sts.StsException;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.ocescpr.OcesServiceException;
import dk.sosi.sts.ocescpr.cvrrid.util.ValidateHelper;
import dk.sosi.sts.service.OcesCvrRidService;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import dk.sosi.sts.util.StsUtil;

public class WsOcesCvrRidService implements OcesCvrRidService {

	private static Logger log = Logger.getLogger(WsOcesCvrRidService.class);

	private URL endpoint;
	private HandleSundhedsportalWSLocator serviceLocator;
	private AtomicLong lastSuccess = new AtomicLong(-1);
	private AtomicLong lastFailure = new AtomicLong(-1);

	private String keyStore;
	private String keyStorePassword;
	private String trustStore;
	private String trustStorePassword;

	private int timeout = 0;

	private final ServiceInfoCollector serviceInfoCollector;

    public WsOcesCvrRidService(ServiceInfoCollector serviceInfoCollector) { this.serviceInfoCollector =
			serviceInfoCollector; }

    public EngineConfiguration makeWsConfig() {
		final SSLClientAxisEngineConfig axisConfig = new SSLClientAxisEngineConfig();

		axisConfig.setKeyStore(keyStore);
		axisConfig.setKeyStoreType("JKS");
		axisConfig.setKeyStorePassword(keyStorePassword);
		axisConfig.setTrustStore(trustStore);
		axisConfig.setTrustStoreType("JKS");
		axisConfig.setTrustStorePassword(trustStorePassword);

		return axisConfig;
	}

	public void init() {
		verifyState();
		this.serviceLocator = new HandleSundhedsportalWSLocator(makeWsConfig());
	}

	private void verifyState() {
		if (endpoint == null) {
			throw new RuntimeException("endpoint URL must not be null");
		}
	}

	public boolean isRelated(final OcesInfo ocesInfo, final CprNr cpr) throws StsException {
		try {
			ValidateHelper.validate(ocesInfo.getSubjectSerialNumber(), cpr);
			CprNr cprLookup = findRelatedCpr(ocesInfo);
			boolean equals = cpr.equals(cprLookup);
			serviceInfoCollector.registerSuccess();
			return equals;
		} catch (Exception e) { serviceInfoCollector.registerFailure(); throw e; }
	}

	public CprNr findRelatedCpr(OcesInfo ocesInfo) throws StsException {
		try {
			final SubjectSerialNumber cvrRid = ocesInfo.getSubjectSerialNumber();
			ValidateHelper.validate(cvrRid);

			SLALogItem item = SLALogHelper.createItem(SLALogPoint.CVRRID_FINDRELATED);
			item.addCallParameter("endpoint", endpoint.toString());
			item.addCallParameter("cvr-rid", cvrRid.getValue());
			try {
				CprNr cpr = findRelatedCpr(cvrRid);
				SLALogHelper.storeCallResultOk(SLALogPoint.CVRRID_FINDRELATED);
				serviceInfoCollector.registerSuccess();
				return cpr;
			} catch (Throwable t) {
				SLALogHelper.storeCallResultError(SLALogPoint.CVRRID_FINDRELATED, t.getMessage());
				throw StsUtil.rethrowable(t, StsException.class);
			}
		} catch (Exception e) { serviceInfoCollector.registerFailure(); throw e; }
	}

	private CprNr findRelatedCpr(SubjectSerialNumber cvrRid) throws StsException {
		try {
			log.debug("Performing CPR WS lookup for: " + cvrRid);
			String reply = getCvrRidService().getCPR(cvrRid.getValue());
			lastSuccess.set(System.currentTimeMillis());
			if (reply == null) {
				return null;
			}
			if (reply != null && reply.length() != 10) {
				throw new OcesServiceException("Fejl under cpr opslag: " + cvrRid.getValue() + "=[" + reply + "]");
			}
			return new CprNr(reply);
		} catch (RemoteException | ServiceException e) {
			lastFailure.set(System.currentTimeMillis());
			throw new StsException(e);
		}
	}

	public long lastSuccess() {
		return lastSuccess.get();
	}

	public long lastFailure() {
		return lastFailure.get();
	}

	@Override
	public int getFailedCallsPercentage() { return serviceInfoCollector.getFailedCallsPercentage(); }

	@Override
	public int getHoursBackInTimeToCalculateFailedCallsPercentage() { return serviceInfoCollector.getHoursBackInTime(); }

	public void setEndpoint(final URL endpoint) {
		this.endpoint = endpoint;
	}

	public URL getEndpoint() {
		return endpoint;
	}

	public String getKeyStore() {
		return keyStore;
	}

	public void setKeyStore(String keyStore) {
		this.keyStore = keyStore;
	}

	public String getKeyStorePassword() {
		return keyStorePassword;
	}

	public void setKeyStorePassword(String keyStorePassword) {
		this.keyStorePassword = keyStorePassword;
	}

	public String getTrustStore() {
		return trustStore;
	}

	public void setTrustStore(String trustStore) {
		this.trustStore = trustStore;
	}

	public String getTrustStorePassword() {
		return trustStorePassword;
	}

	public void setTrustStorePassword(String trustStorePassword) {
		this.trustStorePassword = trustStorePassword;
	}

	public int getTimeout() {
		return timeout;
	}

	public void setTimeout(final int timeout) {
		this.timeout = timeout;
	}

	HandleSundhedsportalWSPort getCvrRidService() throws ServiceException {
		final HandleSundhedsportalWSPort handleSundhedsportalWSPort = serviceLocator.getHandleSundhedsportalWSPort(endpoint);
		((Stub) handleSundhedsportalWSPort).setTimeout(timeout);
		return handleSundhedsportalWSPort;
	}

}
