/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ocescpr.uuid.service.impl;

import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.ocescpr.uuid.util.ValidateHelper;
import dk.sosi.sts.service.OcesUUID2CPRService;
import dk.sosi.sts.timing.Timer;
import org.apache.log4j.Logger;

import javax.sql.DataSource;

public class DBCacheOcesUuidService implements OcesUUID2CPRService {

	protected final OcesUUID2CPRService delegate;
	protected final dk.sosi.sts.ocescpr.uuid.service.impl.CprCache cache;
	private static Logger log = Logger.getLogger(DBCacheOcesUuidService.class);

	public DBCacheOcesUuidService(OcesUUID2CPRService delegate, DataSource ds) {
		this(delegate, new DbCprHashCache(ds));
	}

	public DBCacheOcesUuidService(OcesUUID2CPRService delegate, CprCache cache) {
		this.delegate = delegate;
		this.cache = cache;
	}

	public CprNr findRelatedCpr(OcesInfo ocesInfo) throws StsException {
		SubjectSerialNumber ssn = ocesInfo.getSubjectSerialNumber();
		log.info("Executing method [findRelatedCpr] with UUID:" + ssn);
		ValidateHelper.validate(ssn);
		CprNr cpr = cache.find(ssn);
		if (cpr != null) {
			return cpr;
		} else {
			Timer.start(this.getClass(), "delegate:findRelatedCpr");
			try {
				cpr = delegate.findRelatedCpr(ocesInfo);
			} finally {
				Timer.stop(this.getClass(), "delegate:findRelatedCpr");
			}
			if (cpr != null) {
				cache.map(ssn, cpr);
			}
			return cpr;
		}
	}

	public boolean isRelated(OcesInfo ocesInfo, CprNr cpr) throws StsException {
		final SubjectSerialNumber ssn = ocesInfo.getSubjectSerialNumber();
		log.info("Executing method [isRelated] with UUID,cpr:" + ssn + "," + cpr);

		ValidateHelper.validate(ssn, cpr);

		if (cache.exists(ssn, cpr)) {
			return true;
		} else {
			boolean result = false;
			Timer.start(this.getClass(), "delegate:isRelated");
			try {
				result = delegate.isRelated(ocesInfo, cpr);
			} finally {
				Timer.stop(this.getClass(), "delegate:isRelated");
			}
			if (result) {
				cache.map(ssn, cpr);
			}
			return result;
		}
	}

	public long lastSuccess() {
		return delegate.lastSuccess();
	}

	public long lastFailure() {
		return delegate.lastFailure();
	}

	@Override
	public int getFailedCallsPercentage() { return delegate.getFailedCallsPercentage();	}

	@Override
	public int getHoursBackInTimeToCalculateFailedCallsPercentage() {
		return delegate.getHoursBackInTimeToCalculateFailedCallsPercentage();
	}
}
