/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ocescpr.uuid.service.impl;

import dk.sosi.sts.StsException;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import org.apache.log4j.Logger;
import org.springframework.dao.DataAccessException;
import org.springframework.dao.DuplicateKeyException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.RowMapper;

import javax.sql.DataSource;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.List;

// TODO mkn: consider if cache access should throw Exceptions
public class DbCprCache implements CprCache {
	private static final String SQL_CACHE_INSERT = "INSERT INTO uuidcprcache (uuid, cpr, createdDate) VALUES (?, ?, ?)";
	private static final String SQL_CACHE_FIND = "SELECT createdDate, cpr FROM uuidcprcache WHERE uuid=?";
	private static final String SQL_CACHE_SIZE = "SELECT COUNT(1) FROM uuidcprcache";
	private static final String SQL_CACHE_CLEAR = "DELETE FROM uuidcprcache";
	
	private final Logger log = Logger.getLogger(this.getClass());
	private final JdbcTemplate jdbcTemplate;
	
	public DbCprCache(DataSource ds) {
		this.jdbcTemplate = new JdbcTemplate(ds);
	}
	
	public CprNr find(SubjectSerialNumber ssn) throws StsException {
		RowMapper<CprNr> cprMapper = new RowMapper<CprNr>() {
			public CprNr mapRow(ResultSet rs, int rowNum) throws SQLException {
				return new CprNr(rs.getString("cpr"));
			}
		};
		try {
			List<CprNr> result = jdbcTemplate.query(SQL_CACHE_FIND,	cprMapper, ssn.getValue());
			return (result.isEmpty()) ? null : result.get(0);
		} catch (Exception e) {
			throw new StsException(e);
		}
	}
	
	public boolean exists(SubjectSerialNumber ssn, CprNr cpr) throws StsException {
		CprNr cachedCpr = find(ssn);
		return cpr.equals(cachedCpr);
	}
	
	public void map(SubjectSerialNumber ssn, CprNr cpr) throws StsException {
		try {
			jdbcTemplate.update(SQL_CACHE_INSERT, 
					ssn.getValue(),
					cpr.getValue(),
					new Timestamp(Calendar.getInstance().getTimeInMillis()));
		} catch (DuplicateKeyException e) {
			log.info("Cache entry exists (" + ssn + "," + cpr + ")");
		} catch (DataAccessException e) {
			throw new StsException(e);
		}
	}
	
	public int size() throws StsException {
		try {
			return jdbcTemplate.queryForInt(SQL_CACHE_SIZE);
		} catch (DataAccessException e) {
			throw new StsException(e);
		}
	}
	
	public void clearAll() throws StsException {
		try {
			jdbcTemplate.execute(SQL_CACHE_CLEAR);
		} catch (DataAccessException e) {
			throw new StsException(e);
		}
	}
	
}
