package dk.sosi.sts.ocescpr.uuid.uuid2cpr;

import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.sts.ocescpr.uuid.uuid2cpr.model.Cpr;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpResponse;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.util.List;
import java.util.Optional;

public class Uuid2CprProxyClient {

    private static final Logger LOGGER = Logger.getLogger(Uuid2CprProxyClient.class);
    private final List<String> uuid2cprServiceUrls;
    private final CloseableHttpClient httpClient;

    // invoked by spring config
    public Uuid2CprProxyClient(List<String> uuid2cprServiceUrls, int timeout) {
        this(uuid2cprServiceUrls, new HttpClientFactory(timeout, timeout));
    }

    public Uuid2CprProxyClient(List<String> uuid2cprServiceUrls, HttpClientFactory httpClientFactory) {
        this.uuid2cprServiceUrls = uuid2cprServiceUrls;
        this.httpClient = httpClientFactory.buildClient();
    }

    public Optional<String> lookupCpr(String uuid) throws IOException {
        return lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_UUID, uuid, "service/lookup/");
    }

    public Optional<String> lookupCprFromPid(String pid) throws IOException {
        return lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_PID, pid, "service/lookup-pid-cpr/");
    }

    public Optional<String> lookupRidFromUuid(String uuid) throws IOException {
        return lookup(SLALogPoint.UUID2CPRPROXY_FIND_RID_FROM_UUID, uuid, "service/lookup-uuid-rid/");
    }

    protected Optional<String> lookup(SLALogPoint slaLogPoint, String requestArg, String serviceOperation) throws IOException {
        Optional<String> answer = Optional.empty();
        for (String uuid2cprServiceUrl : uuid2cprServiceUrls) {
            HttpGet request = buildRequest(requestArg, uuid2cprServiceUrl, serviceOperation);
            SLALogItem item = SLALogHelper.createItem(slaLogPoint);
            item.addCallParameter("endpoint", request.getURI().toString());
            item.addCallParameter("argument", requestArg);
            try (CloseableHttpResponse response = httpClient.execute(request)) {
                SLALogHelper.storeCallResultOk(slaLogPoint);
                Optional<String> thisResponse = handleResponse(response);
                if (thisResponse.isPresent() && answer.isPresent() && !thisResponse.get().equals(answer.get())) {
                    throw new ClientProtocolException(String.format("Conflicting matches in endpoints for argument: %s", requestArg));
                }
                if (thisResponse.isPresent()) {
                    answer = thisResponse;
                }
            } catch (ClientProtocolException e) {
                LOGGER.info(e);
                throw e;
            } catch (RuntimeException e) {
                LOGGER.error(e);
                SLALogHelper.storeCallResultError(slaLogPoint, e.getMessage());
                throw e;
            } finally {
                request.releaseConnection();
            }
        }
        return answer;
    }

    private HttpGet buildRequest(String uuid, String uuid2cprServiceUrl, String servicePostFix) {
        String url = uuid2cprServiceUrl;
        if (!url.endsWith("/")) {
            url += "/";
        }
        url += servicePostFix;
        if (!url.endsWith("/")) {
            url += "/";
        }
        return new HttpGet(url + uuid);
    }

    private Optional<String> handleResponse(HttpResponse response) throws IOException {
        int status = response.getStatusLine().getStatusCode();
        switch (status) {
            case 200:
                return parseEntity(response);
            case 404:
                return Optional.empty();
            default:
                throw new ClientProtocolException(String.format("Something went wrong invoking Uuid2CprProxy. Statuscode: %d, reason: %s", status, response.getStatusLine().getReasonPhrase()));
        }
    }

    private Optional<String> parseEntity(HttpResponse response) throws IOException {
        // Extract contents.
        String contents = IOUtils.toString(response.getEntity().getContent());

        // Parse contents and extract value.
        Cpr cpr = new ObjectMapper().readValue(contents, Cpr.class);
        return Optional.of(cpr.getValue());
    }
}
