/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ocescpr.cvrrid.service.impl;

import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import junit.framework.TestCase;
import org.hsqldb.jdbc.JDBCDataSource;
import org.junit.Ignore;

import java.sql.Connection;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;

@Ignore("SDS-6660")
public class DbCacheOcesCvrRidServiceTest extends TestCase {

	private JDBCDataSource dataSource;
	private SubjectSerialNumber cvrRid = new SubjectSerialNumber("CVR:19343634-RID:1165813849809");
	private OcesInfo ocesInfo_cvrRid = OcesInfoParser.parseSubjectSerialNumber(cvrRid);
	private CprNr cpr = new CprNr( "1111111118");

	private SubjectSerialNumber cvrRid2 = new SubjectSerialNumber("CVR:19343634-UID:1165814083652");
	private OcesInfo ocesInfo_cvrRid2 = OcesInfoParser.parseSubjectSerialNumber(cvrRid2);
	private CprNr cpr2 = new CprNr( "1111111119");
		
	@Override
	protected void setUp() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		Connection conn = dataSource.getConnection();
		Statement statement = conn.createStatement();
		statement.execute("DROP TABLE cprhash IF EXISTS");
		statement.execute("CREATE TABLE cprhash (hash char(46), createdDate date)");
		conn.close();		
	}
	
	@Override
	protected void tearDown() throws Exception {
		Connection conn = dataSource.getConnection();
		conn.createStatement().execute("DROP TABLE cprhash");
		conn.close();		
	}
	
	public void testIsRelated() throws Exception {
		TestService delegate = createCachedMockService();
		DBCacheOcesCvrRidService service = new DBCacheOcesCvrRidService(delegate, dataSource);

		// must only hit delegate once if related 
		assertTrue(service.isRelated(ocesInfo_cvrRid, cpr));
		assertEquals(1, delegate.isRelatedCount);

		assertTrue(service.isRelated(ocesInfo_cvrRid, cpr));
		assertEquals(1, delegate.isRelatedCount);
		
		// must not cache unrelated 
		assertFalse(service.isRelated(ocesInfo_cvrRid2, cpr));
		assertEquals(2, delegate.isRelatedCount);

		assertFalse(service.isRelated(ocesInfo_cvrRid2, cpr));
		assertEquals(3, delegate.isRelatedCount);
	}
	
	public void testFindRelatedCpr() throws Exception {
		TestService delegate = createCachedMockService();
		DBCacheOcesCvrRidService service = new DBCacheOcesCvrRidService(delegate, dataSource);
		
		assertEquals(cpr, service.findRelatedCpr(ocesInfo_cvrRid));
		assertEquals(1, delegate.findRelatedCount);
		
		// make sure that the CPR is never cached
		assertEquals(cpr, service.findRelatedCpr(ocesInfo_cvrRid));
		assertEquals(2, delegate.findRelatedCount);
	}

	public void testCountRelations() throws Exception {
		TestService delegate = createCachedMockService();
		DbCprHashCache cache = new DbCprHashCache(dataSource);
		DBCacheOcesCvrRidService service = new DBCacheOcesCvrRidService(delegate, cache);
		
		assertTrue(service.isRelated(ocesInfo_cvrRid, cpr));
		assertEquals(1, cache.size());
		assertTrue(service.isRelated(ocesInfo_cvrRid2, cpr2));
		assertEquals(2, cache.size());
		assertTrue(service.isRelated(ocesInfo_cvrRid, cpr));
		assertEquals(2, cache.size());
	}

	public void testClearCache() throws Exception {
		TestService delegate = createCachedMockService();
		DbCprHashCache cache = new DbCprHashCache(dataSource);
		DBCacheOcesCvrRidService service = new DBCacheOcesCvrRidService(delegate, cache);
		
		assertTrue(service.isRelated(ocesInfo_cvrRid, cpr));
		assertTrue(service.isRelated(ocesInfo_cvrRid2, cpr2));
		assertEquals(delegate.isRelatedCount, cache.size());
		
		cache.clearAll();
		assertEquals(0, cache.size());
	}

	private TestService createCachedMockService() {
		Map<SubjectSerialNumber, CprNr> map = new HashMap<SubjectSerialNumber, CprNr>();
		map.put(cvrRid, cpr);
		map.put(cvrRid2, cpr2);
		return new TestService(map);
	}
	
	private static class TestService extends MockOcesCvrRidService {
		public int isRelatedCount = 0;
		public int findRelatedCount = 0;
		
		public TestService(Map<SubjectSerialNumber, CprNr> map) {
			super(map);
		}

		@Override
		public boolean isRelated(final OcesInfo ocesInfo, final CprNr cpr) throws StsException {
			isRelatedCount++;
			return super.isRelated(ocesInfo, cpr);
		}

		@Override
		public CprNr findRelatedCpr(final OcesInfo ocesInfo) throws StsException {
			findRelatedCount++;
			return super.findRelatedCpr(ocesInfo);
		}
	}
}
