/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ocescpr.pid.service;

import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.ocescpr.cvrrid.service.impl.DbCprCache;
import dk.sosi.sts.ocescpr.pid.DBCacheOcesPidService;
import dk.sosi.sts.service.OcesPidService;
import junit.framework.TestCase;
import org.hsqldb.jdbc.JDBCDataSource;

import java.sql.Connection;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;

public class DbCacheOcesPidServiceTest extends TestCase {

	private JDBCDataSource dataSource;
	private SubjectSerialNumber cvrRid = new SubjectSerialNumber("CVR:19343634-RID:1165813849809");
	private CprNr cpr = new CprNr( "1111111118");

	private SubjectSerialNumber cvrRid2 = new SubjectSerialNumber("CVR:19343634-UID:1165814083652");
	private CprNr cpr2 = new CprNr( "1111111119");
		
	@Override
	protected void setUp() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		Connection conn = dataSource.getConnection();
		Statement statement = conn.createStatement();
		statement.execute("DROP TABLE cprcache IF EXISTS");
		statement.execute("CREATE TABLE cprcache (ssn VARCHAR(64), cpr VARCHAR(16), createdDate datetime)");
		conn.close();		
	}
	
	@Override
	protected void tearDown() throws Exception {
		Connection conn = dataSource.getConnection();
		conn.createStatement().execute("DROP TABLE cprcache");
		conn.close();		
	}
	
	public void testIsRelated() throws Exception {
		OcesPidServiceMock delegate = createCachedMockService();
		DbCprCache cache = new DbCprCache(dataSource);
		OcesPidService service = new DBCacheOcesPidService(delegate, cache);

		// must only hit delegate once if related 
		assertTrue(service.isRelated(cvrRid, cpr));
		assertEquals(1, delegate.isRelatedCount);

		assertTrue(service.isRelated(cvrRid, cpr));
		assertEquals(1, delegate.isRelatedCount);
		
		// must not cache unrelated 
		assertFalse(service.isRelated(cvrRid2, cpr));
		assertEquals(2, delegate.isRelatedCount);

		assertFalse(service.isRelated(cvrRid2, cpr));
		assertEquals(3, delegate.isRelatedCount);
	}
	
	public void testCountRelations() throws Exception {
		OcesPidServiceMock delegate = createCachedMockService();
		DbCprCache cache = new DbCprCache(dataSource);
		OcesPidService service = new DBCacheOcesPidService(delegate, cache);
		
		assertTrue(service.isRelated(cvrRid, cpr));
		assertEquals(1, cache.size());
		assertTrue(service.isRelated(cvrRid2, cpr2));
		assertEquals(2, cache.size());
		assertTrue(service.isRelated(cvrRid, cpr));
		assertEquals(2, cache.size());
	}

	public void testClearCache() throws Exception {
		OcesPidServiceMock delegate = createCachedMockService();
		DbCprCache cache = new DbCprCache(dataSource);
		OcesPidService service = new DBCacheOcesPidService(delegate, cache);
		
		assertTrue(service.isRelated(cvrRid, cpr));
		assertTrue(service.isRelated(cvrRid2, cpr2));
		assertEquals(delegate.isRelatedCount, cache.size());
		
		cache.clearAll();
		assertEquals(0, cache.size());
	}

	private OcesPidServiceMock createCachedMockService() {
		Map<SubjectSerialNumber, CprNr> map = new HashMap<SubjectSerialNumber, CprNr>();
		map.put(cvrRid, cpr);
		map.put(cvrRid2, cpr2);
		return new OcesPidServiceMock(map);
	}

}
