package dk.sosi.sts.ocescpr.pid.service;

import dk.sosi.sts.ServiceInfoCollector;
import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.ocescpr.uuid.service.impl.WsOcesUuidService;
import dk.sosi.sts.ocescpr.uuid.uuid2cpr.Uuid2CprProxyClient;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.io.IOException;
import java.util.Optional;

public class WsOcesUuidServiceTest {

    public WsOcesUuidService subject;
    private Uuid2CprProxyClient mockUuid2CprProxyClient;

    @Before
    public void setup() {
        mockUuid2CprProxyClient = Mockito.mock(Uuid2CprProxyClient.class);
        subject = new WsOcesUuidService(mockUuid2CprProxyClient, new ServiceInfoCollector(0));
    }

    @Test
    public void testEmptyAnswerUUid2CprProxyResultsInNullAnswer() throws IOException, StsException {

        // If
        Mockito.when(mockUuid2CprProxyClient.lookupCpr(Mockito.anyString())).thenReturn(Optional.empty());
        OcesInfo ocesInfo = Mockito.mock(OcesInfo.class);
        Mockito.when(ocesInfo.getSubjectSerialNumber()).thenReturn(new SubjectSerialNumber("TestSSN"));

        // When
        CprNr answer = subject.findRelatedCpr(ocesInfo);

        // Then
        Assert.assertNull(answer);
    }

    @Test
    public void testEmptyAnswerUUid2CprProxyResultsInNoMatch() throws IOException, StsException {

        // If
        Mockito.when(mockUuid2CprProxyClient.lookupCpr(Mockito.anyString())).thenReturn(Optional.empty());
        OcesInfo ocesInfo = Mockito.mock(OcesInfo.class);
        Mockito.when(ocesInfo.getSubjectSerialNumber()).thenReturn(new SubjectSerialNumber("TestSSN"));

        // When
        boolean answer = subject.isRelated(ocesInfo, new CprNr("9999999999"));

        // Then
        Assert.assertFalse(answer);
    }

    @Test
    public void testAnswerUUid2CprProxyResultsInCpr() throws IOException, StsException {

        // If
        String correctCpr = "1111111111";
        Mockito.when(mockUuid2CprProxyClient.lookupCpr(Mockito.anyString())).thenReturn(Optional.of(correctCpr));
        OcesInfo ocesInfo = Mockito.mock(OcesInfo.class);
        Mockito.when(ocesInfo.getSubjectSerialNumber()).thenReturn(new SubjectSerialNumber("TestSSN"));

        // When
        CprNr answer = subject.findRelatedCpr(ocesInfo);

        // Then
        Assert.assertNotNull(answer);
        Assert.assertEquals(correctCpr, answer.getValue());
    }
    @Test
    public void testAnswerUUid2CprProxyResultsInMatchIfCorrectCprIsSupplied() throws IOException, StsException {

        // If
        String correctCpr = "1111111111";
        Mockito.when(mockUuid2CprProxyClient.lookupCpr(Mockito.anyString())).thenReturn(Optional.of(correctCpr));
        OcesInfo ocesInfo = Mockito.mock(OcesInfo.class);
        Mockito.when(ocesInfo.getSubjectSerialNumber()).thenReturn(new SubjectSerialNumber("TestSSN"));

        // When
        boolean answer = subject.isRelated(ocesInfo, new CprNr(correctCpr));

        // Then
        Assert.assertTrue(answer);
    }

    @Test
    public void testAnswerUUid2CprProxyResultsInNoMatchIfWrongCprIsSupplied() throws IOException, StsException {

        // If
        String correctCpr = "1111111111";
        Mockito.when(mockUuid2CprProxyClient.lookupCpr(Mockito.anyString())).thenReturn(Optional.of(correctCpr));
        OcesInfo ocesInfo = Mockito.mock(OcesInfo.class);
        Mockito.when(ocesInfo.getSubjectSerialNumber()).thenReturn(new SubjectSerialNumber("TestSSN"));

        // When
        boolean answer = subject.isRelated(ocesInfo, new CprNr("9999999999"));

        // Then
        Assert.assertFalse(answer);
    }

    @Test(expected = StsException.class)
    public void testExceptionFromUUid2CprProxyResultsExceptionWhenFindingRelatedCpr() throws IOException, StsException {

        // If
        Mockito.when(mockUuid2CprProxyClient.lookupCpr(Mockito.anyString())).thenThrow(new IOException());
        OcesInfo ocesInfo = Mockito.mock(OcesInfo.class);
        Mockito.when(ocesInfo.getSubjectSerialNumber()).thenReturn(new SubjectSerialNumber("TestSSN"));

        // When
        try {
            subject.findRelatedCpr(ocesInfo);
        } catch (StsException e) {
            // Then
            Assert.assertEquals("Error looking up cpr for ssn TestSSN", e.getMessage());
            throw e;
        }
    }

    @Test(expected = StsException.class)
    public void testExceptionFromUUid2CprProxyResultsInExceptionWhenMatching() throws IOException, StsException {

        // If
        Mockito.when(mockUuid2CprProxyClient.lookupCpr(Mockito.anyString())).thenThrow(new IOException());
        OcesInfo ocesInfo = Mockito.mock(OcesInfo.class);
        Mockito.when(ocesInfo.getSubjectSerialNumber()).thenReturn(new SubjectSerialNumber("TestSSN"));

        // When
        try {
            subject.isRelated(ocesInfo, new CprNr("9999999999"));
        } catch (StsException e) {
            // Then
            Assert.assertEquals("Error looking up cpr for ssn TestSSN", e.getMessage());
            throw e;
        }
    }

}
