/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.ocescpr.uuid.service.impl;

import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import junit.framework.TestCase;
import org.hsqldb.jdbc.JDBCDataSource;

import java.sql.Connection;
import java.sql.Statement;

public class DbCprCacheTest extends TestCase {
	
	public static final String[] DDL_CREATE = {
			"DROP TABLE uuidcprcache IF EXISTS",
			"CREATE TABLE uuidcprcache (uuid varchar(64) PRIMARY KEY, cpr varchar(16), createdDate date)"
	};
	public static final String DDL_DROP = "DROP TABLE uuidcprcache";
	
	private final SubjectSerialNumber ssn1 = new SubjectSerialNumber("UI:DK-E:C:b2699984-9d7e-41e0-92d1-f502a20bf64c");
	private final CprNr cpr1 = new CprNr( "1111111118");
	private final SubjectSerialNumber ssn2 = new SubjectSerialNumber("UI:DK-E:C:5678");
	private final CprNr cpr2 = new CprNr( "1111111119");

	private JDBCDataSource dataSource;

	@Override
	protected void setUp() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		Connection conn = dataSource.getConnection();
		Statement statement = conn.createStatement();
		for (String sql : DDL_CREATE) {
			statement.execute(sql);
		} 
		conn.close();		
	}
	
	@Override
	protected void tearDown() throws Exception {
		Connection conn = dataSource.getConnection();
		conn.createStatement().execute(DDL_DROP);
		conn.close();		
	}

	public void testFind() throws Exception {
		DbCprCache cache = new DbCprCache(dataSource);
		assertNull(cache.find(ssn1));
		assertNull(cache.find(ssn2));
		cache.map(ssn1, cpr1);
		cache.map(ssn2, cpr2);
		assertEquals(cpr1, cache.find(ssn1));
		assertEquals(cpr2, cache.find(ssn2));
	}

	public void testExists() throws Exception {
		DbCprCache cache = new DbCprCache(dataSource);
		assertFalse(cache.exists(ssn1, cpr1));
		assertFalse(cache.exists(ssn2, cpr2));
		cache.map(ssn1, cpr1);
		assertTrue(cache.exists(ssn1, cpr1));
		assertFalse(cache.exists(ssn2, cpr2));
	}

	public void testMap() throws Exception {
		DbCprCache cache = new DbCprCache(dataSource);

		assertFalse(cache.exists(ssn1, cpr1));
		cache.map(ssn1, cpr1);
		assertTrue(cache.exists(ssn1, cpr1));
		
		assertFalse(cache.exists(ssn2, cpr2));
		cache.map(ssn2, cpr2);
		assertTrue(cache.exists(ssn2, cpr2));

		assertEquals(2, cache.size());
		
		// cache mapping is stable
		cache.map(ssn1, cpr1);
		assertTrue(cache.exists(ssn1, cpr1));
		// should never occur in practice. documents our behavior
		cache.map(ssn1, cpr2);
		assertFalse(cache.exists(ssn1, cpr2));
		assertTrue(cache.exists(ssn1, cpr1));
		
		cache.clearAll();
		assertEquals(0, cache.size());		
	}
}
