package dk.sosi.sts.ocescpr.uuid.uuid2cpr;

import dk.sosi.sts.sla.SLALogPoint;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.impl.client.CloseableHttpClient;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.runners.MockitoJUnitRunner;
import org.mockito.stubbing.Answer;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Optional;

import static org.junit.Assert.*;

@RunWith(MockitoJUnitRunner.class)
public class Uuid2CprProxyClientTest {
    private Uuid2CprProxyClient subject;

    private final String uuid2cprServiceUrl = "http://foo";

    @Mock
    private HttpClientFactory httpClientFactory;

    @Mock
    private CloseableHttpClient httpClient;


    private int responseStatusCode = 200;
    private String responseValue = null;
    private HttpEntity responseEntity;

    @Before
    public void setup() throws IOException {
        Mockito.when(httpClientFactory.buildClient()).thenReturn(httpClient);

        responseEntity = Mockito.mock(HttpEntity.class);
        Mockito.when(httpClient.execute(Mockito.any(HttpUriRequest.class))).thenAnswer(new Answer<HttpResponse>() {
            @Override
            public HttpResponse answer(InvocationOnMock invocation) throws Throwable {
                Object[] args = invocation.getArguments();

                CloseableHttpResponse response = Mockito.mock(CloseableHttpResponse.class);
                StatusLine statusLine = Mockito.mock(StatusLine.class);

                HttpUriRequest request = (HttpUriRequest) invocation.getArguments()[0];
                if (request != null && request.getURI().toString().contains("404")) {
                    Mockito.when(statusLine.getStatusCode()).thenReturn(404);
                    Mockito.when(response.getStatusLine()).thenReturn(statusLine);
                    return response;
                }

                Mockito.when(statusLine.getStatusCode()).thenReturn(responseStatusCode);
                Mockito.when(response.getStatusLine()).thenReturn(statusLine);

                Mockito.when(response.getEntity()).thenReturn(responseEntity);
                if (responseValue != null) {
                    Mockito.when(responseEntity.getContent()).thenReturn(IOUtils.toInputStream(String.format("{\"value\": \"%s\"}", responseValue)));
                }

                return response;
            }
        });

        subject = new Uuid2CprProxyClient(Collections.singletonList(uuid2cprServiceUrl), httpClientFactory);
    }

    @Test
    public void lookup_200_returnsCpr() throws Exception {
        // Arrange
        responseStatusCode = 200;
        responseValue = "1234";

        // Act
        Optional<String> result = subject.lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_UUID, "arg", "operation");

        // Assert
        assertTrue(result.isPresent());
        assertEquals(responseValue, result.get());
    }

    @Test
    public void lookup_200FromTwoEndpointsWithDifferentResponses_throwsError() throws Exception {
        // Arrange
        subject = new Uuid2CprProxyClient(Arrays.asList("http://url1", "http://url2"), httpClientFactory);
        responseStatusCode = 200;
        String response1 = "1234";
        String response2 = "4321";
        Mockito.when(responseEntity.getContent())
                .thenReturn(IOUtils.toInputStream(String.format("{\"value\": \"%s\"}", response1)))
                .thenReturn(IOUtils.toInputStream(String.format("{\"value\": \"%s\"}", response2)));

        // Act
        try {
            subject.lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_UUID, "arg", "operation");
            fail("The lookup should throw an exception");
        } catch (ClientProtocolException e) {
            // Assert
            assertTrue(e.getMessage().contains("Conflicting matches in endpoints for argument: arg"));
        }
    }

    @Test
    public void lookup_200FromTwoEndpointsWithSameResponse_ReturnsResponseValue() throws Exception {
        // Arrange
        subject = new Uuid2CprProxyClient(Arrays.asList("http://url1", "http://url2"), httpClientFactory);
        responseStatusCode = 200;
        responseValue = "test";

        // Act
        Optional<String> result = subject.lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_UUID, "arg", "operation");

        // Assert
        assertTrue(result.isPresent());
        assertEquals(responseValue, result.get());
    }

    @Test
    public void lookup_200and404WithTwoEndpoints_returnsResponseValue() throws Exception {
        // Arrange
        subject = new Uuid2CprProxyClient(Arrays.asList(uuid2cprServiceUrl, "http://notfound.dk/always/returns/404"), httpClientFactory);
        responseStatusCode = 200;
        responseValue = "test";

        // Act
        Optional<String> result = subject.lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_UUID, "arg", "operation");

        // Assert
        assertTrue(result.isPresent());
        assertEquals(responseValue, result.get());
    }


    @Test
    public void lookup_404_returnsEmptyResult() throws Exception {
        // Arrange
        responseStatusCode = 404;

        // Act
        Optional<String> result = subject.lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_UUID, "arg", "operation");

        // Assert
        assertFalse(result.isPresent());
    }

    @Test(expected = ClientProtocolException.class)
    public void lookup_500_throwsException() throws Exception {
        // Arrange
        responseStatusCode = 500;

        // Act
        subject.lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_UUID, "arg", "operation");
    }

    @Test
    public void lookupCpr_callsLookupWithCorrectArguments() throws Exception {
        // Arrange
        Uuid2CprProxyClient subjectSpy = Mockito.spy(this.subject);
        String response = "test";
        Mockito.doReturn(Optional.of(response)).when(subjectSpy).lookup(Mockito.any(), Mockito.any(), Mockito.any());
        String input = "test input";

        // Act
        Optional<String> result = subjectSpy.lookupCpr(input);

        // Assert
        Mockito.verify(subjectSpy).lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_UUID, input, "service/lookup/");
        assertTrue(result.isPresent());
        assertEquals(response, result.get());
    }

    @Test
    public void lookupCprFromPid_callsLookupWithCorrectArguments() throws Exception {
        // Arrange
        Uuid2CprProxyClient subjectSpy = Mockito.spy(this.subject);
        String response = "test";
        Mockito.doReturn(Optional.of(response)).when(subjectSpy).lookup(Mockito.any(), Mockito.any(), Mockito.any());
        String input = "test input";

        // Act
        Optional<String> result = subjectSpy.lookupCprFromPid(input);

        // Assert
        Mockito.verify(subjectSpy).lookup(SLALogPoint.UUID2CPRPROXY_FIND_CPR_FROM_PID, input, "service/lookup-pid-cpr/");
        assertTrue(result.isPresent());
        assertEquals(response, result.get());
    }

    @Test
    public void lookupRidFromUuid_callsLookupWithCorrectArguments() throws Exception {
        // Arrange
        Uuid2CprProxyClient subjectSpy = Mockito.spy(this.subject);
        String response = "test";
        Mockito.doReturn(Optional.of(response)).when(subjectSpy).lookup(Mockito.any(), Mockito.any(), Mockito.any());
        String input = "test input";

        // Act
        Optional<String> result = subjectSpy.lookupRidFromUuid(input);

        // Assert
        Mockito.verify(subjectSpy).lookup(SLALogPoint.UUID2CPRPROXY_FIND_RID_FROM_UUID, input, "service/lookup-uuid-rid/");
        assertTrue(result.isPresent());
        assertEquals(response, result.get());
    }

}