package dk.sosi.sts.integrations.personinformation;

import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sosi.sts.ServiceInfoCollector;
import dk.sosi.sts.service.ExternalService;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;

public class PersonInformationClient implements ExternalService {
    private static final String APPLICATION_JSON = "application/json";
    private static final String CONTENT_TYPE = "Content-Type";
    private static final Logger logger = Logger.getLogger(PersonInformationClient.class);
    private final ObjectMapper objectMapper = new ObjectMapper();
    private final CloseableHttpClient httpClient;
    private String url;
    private final AtomicLong lastSuccess = new AtomicLong(-1);
    private final AtomicLong lastFailure = new AtomicLong(-1);
    private final ServiceInfoCollector serviceInfoCollector;

    // Used by Spring
    public void setServiceEndpoint(String endpoint) {
        this.url = endpoint;
    }

    public PersonInformationClient(CloseableHttpClient httpClient, ServiceInfoCollector serviceInfoCollector) {
        this.httpClient = httpClient;
        this.serviceInfoCollector = serviceInfoCollector;
    }

    public PersonInformationClient(ServiceInfoCollector serviceInfoCollector) {
        this(HttpClients.createDefault(), serviceInfoCollector);
    }

    public Set<Custody> personCustody(String cpr) throws PersonInformationClientException {
        try {
            HttpUriRequest cprRequest = buildGetPersonCustodyRequest(cpr);
            Map<String, Object> response = invokeService(cprRequest);
            Set<Custody> custodies = parseCustodyFromResponse(response);
            serviceInfoCollector.registerSuccess();
            return custodies;
        } catch (PersonInformationClientException e) { serviceInfoCollector.registerFailure(); throw e; }
    }

    public int personAge(String cpr) throws PersonInformationClientException {
        HttpUriRequest cprRequest = buildGetPersonAgeRequest(cpr);
        Map<String, Object> response = invokeService(cprRequest);
        Object age = response.get("age");
        if(age == null){
            throw new PersonInformationClientException(ErrorMessages.personInformationServiceCallNoAge());
        }
        return Integer.parseInt(age.toString());
    }

    private HttpUriRequest buildGetPersonAgeRequest(String cpr) {
        return new HttpGet(String.format("%s/person/%s/age", url, cpr));
    }

    private HttpUriRequest buildGetPersonCustodyRequest(String cpr) {
        return new HttpGet(String.format("%s/person/%s/custody", url, cpr));
    }

    private Map<String, Object> invokeService(HttpUriRequest request) throws PersonInformationClientException {
        try (CloseableHttpResponse response = httpClient.execute(request)) {
            Map<String, Object> result = handleResponse(response);
            registerSucces();
            return result;
        } catch (IOException e) {
            String errorMessage = ErrorMessages.personInformationServiceCallFailed();
            logger.error(errorMessage, e);
            registerError();
            throw new PersonInformationClientException(errorMessage, e);
        } catch (PersonInformationClientException e) {
            registerError();
            throw e;
        }
    }

    private void registerSucces() {
        lastSuccess.set(System.currentTimeMillis());
    }

    private void registerError() {
        lastFailure.set(System.currentTimeMillis());
    }

    @SuppressWarnings("unchecked")
    private Set<Custody> parseCustodyFromResponse(Map<String, Object> response) {
        return response.entrySet().stream()
                .flatMap(entry ->
                        ((ArrayList<Map<String, Object>>) entry.getValue()).stream()
                                .map(custody -> Custody.fromValues(entry.getKey(), custody.get("relationCpr").toString()))
                )
                .collect(Collectors.toSet());
    }

    private Map<String, Object> handleResponse(HttpResponse response) throws PersonInformationClientException {
        int statusCode = response.getStatusLine().getStatusCode();
        switch (statusCode) {
            case HttpStatus.SC_OK: {
                boolean contentTypeOk = contentTypeOk(response);
                Map<String, Object> responseData = parseResponseAsJson(response);
                if (!contentTypeOk) {
                    String error = ErrorMessages.personInformationServiceCallContenTypeWrong();
                    logger.error(error);
                    throw new PersonInformationClientException(error);
                }
                return responseData;
            }
            case HttpStatus.SC_NOT_FOUND: {
                boolean contentTypeOk = contentTypeOk(response);
                if (!contentTypeOk) {
                    String error = ErrorMessages.personInformationServiceCallError404();
                    throw new PersonInformationClientException(error);
                }
                return parseResponseAsJson(response);
            }
            default:
                throw new PersonInformationClientException(ErrorMessages.personInformationServiceCallUnknownStatus(statusCode));
        }
    }

    @SuppressWarnings("unchecked")
    private Map<String, Object> parseResponseAsJson(HttpResponse response) throws PersonInformationClientException {
        try {
            return objectMapper.readValue(response.getEntity().getContent(), Map.class);
        } catch (UnsupportedOperationException | IOException e) {
            logger.error("Kunne ikke parse response body til json", e);
            throw new PersonInformationClientException(ErrorMessages.personInformationServiceCallParseError(), e);
        }
    }

    private boolean contentTypeOk(HttpResponse response) {
        Header headers = response.getFirstHeader(CONTENT_TYPE);
        return headers != null && headers.getValue().equalsIgnoreCase(APPLICATION_JSON);
    }

    @Override
    public long lastSuccess() {
        return lastSuccess.get();
    }

    @Override
    public long lastFailure() {
        return lastFailure.get();
    }

    @Override
    public int getFailedCallsPercentage() { return serviceInfoCollector.getFailedCallsPercentage(); }

    @Override
    public int getHoursBackInTimeToCalculateFailedCallsPercentage() { return serviceInfoCollector.getHoursBackInTime(); }
}