package dk.sosi.sts.integrations.personinformation;

import dk.sosi.sts.ServiceInfoCollector;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.impl.client.CloseableHttpClient;
import org.junit.Before;
import org.junit.Test;

import java.io.IOException;
import java.util.Set;

import static dk.sosi.sts.integrations.personinformation.TestUtil.createHttpResponse;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

public class PersonInformationClientTest {

    private CloseableHttpClient httpClient;
    private PersonInformationClient personInformationClient;

    @Before
    public void setUp() {
        httpClient = mock(CloseableHttpClient.class);
        personInformationClient = new PersonInformationClient(httpClient, new ServiceInfoCollector(0));
        personInformationClient.setServiceEndpoint("test-url");
    }

    @Test
    public void personStatus_WhenClientThrowsIOException_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        // Arrange
        doThrow(IOException.class).when(httpClient).execute(any(HttpUriRequest.class));

        // Act
        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personCustody("1234"));

        // Assert
        assertEquals(ErrorMessages.personInformationServiceCallFailed(), e.getMessage());
    }

    @Test
    public void personExists_WhenServiceReturnsResponse200WithContentTypeText_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(200, "text/plain", "{\"cpr\": \"1234\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        // Act
        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personCustody("1234"));

        // Assert
        assertEquals(ErrorMessages.personInformationServiceCallContenTypeWrong(), e.getMessage());
    }

    @Test
    public void personExists_WhenServiceReturnsResponse200WithInvalidJson_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "invalid body");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        // Act
        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personCustody("1234"));

        // Assert
        assertEquals(ErrorMessages.personInformationServiceCallParseError(), e.getMessage());
    }


    @Test
    public void personExists_WhenServiceReturnsResponse404WithContentTypeText_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(404, "text/plain", "{\"cpr\": \"1234\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        // Act
        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personCustody("1234"));

        // Assert
        assertEquals(ErrorMessages.personInformationServiceCallError404(), e.getMessage());
    }

    @Test
    public void personExists_WhenServiceReturnsResponse404WithInvalidJson_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(404, "application/json", "invalid body");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        // Act
        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personCustody("1234"));

        // Assert
        assertEquals(ErrorMessages.personInformationServiceCallParseError(), e.getMessage());
    }

    @Test
    public void personExists_WhenServiceReturnsResponse400_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(400, "application/json", "{\"test\": \"test\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        // Act
        PersonInformationClientException e = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personCustody("1234"));

        // Assert
        assertEquals(ErrorMessages.personInformationServiceCallUnknownStatus(400), e.getMessage());
    }

    @Test
    public void personCustody_WhenServiceReturnsValidResponse_ReturnsCustodyAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        // Arrange
        String json = "{" +
                "        \"parentalCustody\": [" +
                "            {\"relationCpr\": \"0101011234\"}," +
                "            {\"relationCpr\": \"0101011235\"}" +
                "        ]," +
                "        \"wardCustody\": [" +
                "            {\"relationCpr\": \"0101011236\"}" +
                "        ]" +
                "    }";
        HttpResponse httpResponse = createHttpResponse(200, "application/json", json);

        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        String cpr = "1234";

        // Act
        Set<Custody> custody = personInformationClient.personCustody(cpr);

        // Assert
        assertEquals(3, custody.size());
        assertTrue(custody.contains(Custody.fromValues("PARENTALCUSTODY", "0101011234")));
        assertTrue(custody.contains(Custody.fromValues("PARENTALCUSTODY", "0101011235")));
        assertTrue(custody.contains(Custody.fromValues("WARDCUSTODY", "0101011236")));
        assertFalse(custody.contains(Custody.fromValues("PARENTALCUSTODY", "0101011236")));
    }

    @Test
    public void personCustody_WhenServiceReturnsValidResponseWithEmptyCustody_ReturnsEmptySetAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        // Arrange
        String json = "{" +
                "        \"parentalCustody\": []," +
                "        \"wardCustody\": []" +
                "    }";
        HttpResponse httpResponse = createHttpResponse(200, "application/json", json);

        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        String cpr = "1234";

        // Act
        Set<Custody> custody = personInformationClient.personCustody(cpr);

        // Assert
        assertTrue(custody.isEmpty());
    }

    @Test
    public void personCustody_WhenServiceReturnsValidResponse_RegistersLastSuccess() throws PersonInformationClientException, IOException {
        // Arrange
        String json = "{" +
                "        \"parentalCustody\": []," +
                "        \"wardCustody\": []" +
                "    }";
        HttpResponse httpResponse = createHttpResponse(200, "application/json", json);

        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        String cpr = "1234";

        // Act
        personInformationClient.personCustody(cpr);

        // Assert
        assertTrue(personInformationClient.lastSuccess() > 0);
        assertTrue(personInformationClient.lastFailure() < 0);
    }

    @Test
    public void personCustody_WhenServiceThrowsException_RegistersLastFailure() throws IOException {
        // Arrange
        doThrow(IOException.class).when(httpClient).execute(any(HttpUriRequest.class));

        // Act
        try {
            personInformationClient.personCustody("1234");
        } catch (PersonInformationClientException ignored) {}

        // Assert
        assertTrue(personInformationClient.lastFailure() > 0);
        assertTrue(personInformationClient.lastSuccess() < 0);
    }

    @Test
    public void personAge_WhenServiceReturnsValidResponse_ReturnsSameAge() throws PersonInformationClientException, IOException {
        // Arrange
        int age = 72;
        HttpResponse httpResponse = createHttpResponse(200, "application/json",
                "{\"status\": 1, \"age\": "+age+"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        String cpr = "1234";

        // Act
        int returnAge = personInformationClient.personAge(cpr);

        // Assert
        assertSame(age, returnAge);
    }

    @Test
    public void personAge_WhenServiceReturnsValidResponse_ReturnsNoAge() throws IOException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "{}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        String cpr = "1234";
        // Act
        PersonInformationClientException exception = assertThrows(PersonInformationClientException.class, () -> personInformationClient.personAge(cpr));
        // Assert
        assertEquals(ErrorMessages.personInformationServiceCallNoAge(), exception.getMessage());
    }
}