package dk.sosi.sts;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPMessage;
import javax.xml.ws.soap.SOAPFaultException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import static java.nio.charset.StandardCharsets.UTF_8;
import static java.util.Optional.ofNullable;

public class XmlUtil {
    private XmlUtil() {}

    private final static Logger logger = Logger.getLogger(XmlUtil.class);

    private final static Map<String, JAXBContext> cache = new ConcurrentHashMap<>();

    public static JAXBContext jaxbContext(Class<?> t) {
        return cache.computeIfAbsent(t.getName(), u -> {
            try { return JAXBContext.newInstance(t); }
            catch (JAXBException e) {
                logger.error("Failed to create JAXBContext.", e);
                throw new RuntimeException(e);
            }
        });
    }

    public static <T> Document toXml(T v)
            throws JAXBException, ParserConfigurationException, IOException, SAXException {
        ByteArrayOutputStream s = new ByteArrayOutputStream();
        jaxbContext(v.getClass()).createMarshaller().marshal(v, s);
        return toXml(new String(s.toByteArray(), UTF_8));
    }

    public static Document toXml(JAXBElement<?> v)
            throws JAXBException, ParserConfigurationException, IOException, SAXException {
        ByteArrayOutputStream s = new ByteArrayOutputStream();
        jaxbContext(v.getValue().getClass()).createMarshaller().marshal(v, s);
        return toXml(new String(s.toByteArray(), UTF_8));
    }

    public static Document toXml(String v) throws ParserConfigurationException, IOException, SAXException {
        InputSource is = new InputSource(new StringReader(v));
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        // Disable external entity declarations to prevent XXE attacks
        factory.setFeature("http://xml.org/sax/features/external-general-entities", false);
        factory.setFeature("http://xml.org/sax/features/external-parameter-entities", false);
        // Disable entity expansion to prevent XXE attacks
        factory.setExpandEntityReferences(false);
        return factory.newDocumentBuilder().parse(is);
    }

    public static SOAPMessage toSoapMessage(String v) throws SOAPException, IOException {
        return MessageFactory.newInstance().createMessage(null, new ByteArrayInputStream(v.getBytes()));
    }

    public static <T> T toModel(SOAPBody b, Class<T> modelType)
            throws JAXBException, SOAPException, SOAPFaultException {
        ofNullable(b.getFault()).map(SOAPFaultException::new).ifPresent(ex -> {
            logger.error("SOAP body contains fault.", ex);
            throw ex;
        });
        return jaxbContext(modelType).createUnmarshaller().unmarshal(b.extractContentAsDocument(), modelType)
                .getValue();
    }
}