/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.certificate;

import java.security.cert.X509Certificate;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.pki.OCESUtil;
import dk.sosi.sts.certificate.OcesInfo.CertificateType;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.model.SubjectSerialNumber;

public class Oces2InfoParser extends OcesInfoParser{
    private static final Pattern CVR_RID_PATTERN = Pattern.compile("CVR:([0-9]*)-([URF])ID:([^,+\\s]+)");
    private static final Pattern PID_PATTERN = Pattern.compile("PID:([^},+\\\\s]+)");

    public static SubjectSerialNumber subjectSerialNumberFromCvrAndRid(String cvr, String rid) {
        return new SubjectSerialNumber(String.format("CVR:%s-RID:%s", cvr, rid));
    }

    protected boolean isValid(SubjectSerialNumber subjectSerialNumber) {
        String value = subjectSerialNumber != null ? subjectSerialNumber.getValue() : null;
        return value != null && (CVR_RID_PATTERN.matcher(value).find() || PID_PATTERN.matcher(value).find());
    }

    @Override
    protected boolean isValid(X509Certificate certificate) {
        return OCESUtil.isProbableOCES2Certificate(certificate);
    }

    @Override
    protected OcesInfo parse(X509Certificate certificate) {
        DistinguishedName distinguishedName = new DistinguishedName(certificate.getSubjectDN().getName());
        SubjectSerialNumber subjectSerialNumber = new SubjectSerialNumber(distinguishedName.getSubjectSerialNumber());
        return parse(subjectSerialNumber, certificate);
    }

    @Override
    protected OcesInfo parse(SubjectSerialNumber subjectSerialNumber) {
        return parse(subjectSerialNumber, null);
    }

    private OcesInfo parse(SubjectSerialNumber subjectSerialNumber, X509Certificate certificate) {
        String value = subjectSerialNumber != null ? subjectSerialNumber.getValue() : null;
        if(value != null) {
            Matcher cvrRidMatcher = CVR_RID_PATTERN.matcher(value);
            if(cvrRidMatcher.find()) {
                return getCvrRidOcesInfo(cvrRidMatcher, certificate);
            }

            Matcher pidMatcher = PID_PATTERN.matcher(value);

            if(pidMatcher.find()) {
                return getPidOcesInfo(pidMatcher, certificate);
            }
        }

        /*No matches found.*/
        return getUnknownOcesInfo(certificate);
    }

    private static OcesInfo getCvrRidOcesInfo(final Matcher matcher, final X509Certificate certificate) {
        SubjectSerialNumber ssn = new SubjectSerialNumber(matcher.group(0));
        String cvr = matcher.group(1);
        String idType = matcher.group(2);

        CertificateType type = CertificateType.UNKNOWN;
        if (idType.equals("U")) {
            type = CertificateType.VOCES;
        } else if (idType.equals("R")) {
            type = CertificateType.MOCES;
        } else if (idType.equals("F")) {
            type = CertificateType.FOCES;
        }

        return new OcesInfo(ssn, cvr, type, Version.OCES2, certificate);
    }

    private static OcesInfo getPidOcesInfo(final Matcher matcher, final X509Certificate certificate) {
        SubjectSerialNumber ssn = new SubjectSerialNumber(matcher.group(0));
        return new OcesInfo(ssn, null, CertificateType.POCES, Version.OCES2, certificate);
    }
}
