package dk.sosi.sts.certificate;

import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.pki.OCESUtil;
import dk.sosi.sts.certificate.OcesInfo.CertificateType;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.model.SubjectSerialNumber;

import java.security.cert.X509Certificate;
import java.util.Optional;
import java.util.UUID;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class Oces3InfoParser extends OcesInfoParser {

	// IdentityType is either (P)erson, (O)rganization or (E)mployee.
	// PersistenceLevel is either (G)lobal identifier, long term (C)ertificate or (S)ession specific identifier.
	private static final Pattern UI_PATTERN = Pattern.compile("UI:DK-(?<IdentityType>[POE]):(?<PersistenceLevel>[GCS]):(?<Uuid>[^,+\\s]+)");
	private static final Pattern OI_PATTERN = Pattern.compile("NTRDK-(?<Cvr>[0-9]{1,8})");

	public static SubjectSerialNumber subjectSerialNumberFromGlobalEmployeeUuid(String uuid) {
		return new SubjectSerialNumber("UI:DK-E:G:" + uuid);
	}

	public static Optional<UUID> uuidFromSubjectSerialNumber(SubjectSerialNumber ssn) {
		Matcher matcher = UI_PATTERN.matcher(ssn.getValue());
		if (matcher.find()) {
			String uuid = matcher.group("Uuid");
			return Optional.of(UUID.fromString(uuid));
		}
		return Optional.empty();
	}

	@Override
	protected boolean isValid(X509Certificate certificate) {
		return OCESUtil.isProbableOCES3Certificate(certificate);
	}

	@Override
	protected OcesInfo parse(X509Certificate certificate) {
		DistinguishedName distinguishedName = new DistinguishedName(certificate.getSubjectDN().getName());
		SubjectSerialNumber subjectSerialNumber = new SubjectSerialNumber(distinguishedName.getSubjectSerialNumber());
		return parse(subjectSerialNumber, distinguishedName, certificate);
	}

	@Override
	protected boolean isValid(SubjectSerialNumber subjectSerialNumber) {
		String value = subjectSerialNumber != null ? subjectSerialNumber.getValue() : null;
		return value != null && UI_PATTERN.matcher(value).find();
	}

	@Override
	protected OcesInfo parse(SubjectSerialNumber subjectSerialNumber) {
		return parse(subjectSerialNumber, null, null);
	}

	private OcesInfo parse(SubjectSerialNumber subjectSerialNumber, DistinguishedName distinguishedName, X509Certificate certificate) {
		String ssn = subjectSerialNumber != null ? subjectSerialNumber.getValue() : null;
		if (ssn != null) {
			Matcher uiMatcher = UI_PATTERN.matcher(ssn);
			if (uiMatcher.find()) {
				return getUiOcesInfo(uiMatcher, distinguishedName, certificate);
			}
		}

		/* No matches found. */
		return getUnknownOcesInfo(certificate);
	}

	private static String getCvr(DistinguishedName dn) {
		if(dn == null) {
			return null;
		}

		String organizationIdentifier = dn.getOrganizationIdentifier();
		Matcher oiMatcher = null;
		if(organizationIdentifier != null) {
			oiMatcher = OI_PATTERN.matcher(organizationIdentifier);
		}

		return (oiMatcher != null && oiMatcher.find()) ? oiMatcher.group("Cvr") : null;
	}

	private static CertificateType getCertificateType(String identityType) {
		switch (identityType) {
			case "P": return CertificateType.POCES;
			case "O": return CertificateType.VOCES; // voces and foces is the same in oces3
			case "E": return CertificateType.MOCES;
			default: return CertificateType.UNKNOWN;
		}
	}

	private static OcesInfo.PersistenceLevel getPersistenceLevel(String persistenceLevel) {
		switch (persistenceLevel) {
			case "G": return OcesInfo.PersistenceLevel.GLOBAL;
			case "C": return OcesInfo.PersistenceLevel.LONG_TERM;
			case "S": return OcesInfo.PersistenceLevel.SESSION_SPECIFIC;
			default: return OcesInfo.PersistenceLevel.NONE;
		}
	}

	private static OcesInfo getUiOcesInfo(final Matcher ssnMatcher, final DistinguishedName dn, final X509Certificate certificate) {
		SubjectSerialNumber ssn = new SubjectSerialNumber(ssnMatcher.group());
		String cvr = getCvr(dn);
		CertificateType type = getCertificateType(ssnMatcher.group("IdentityType"));
		OcesInfo.PersistenceLevel persistenceLevel = getPersistenceLevel(ssnMatcher.group("PersistenceLevel"));

		return new OcesInfo(ssn, cvr, type, Version.OCES3, certificate, persistenceLevel);
	}
}
