/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.model;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.stream.Collectors;

public class Configuration {

	protected final Properties data;

	public Configuration() {
		this(new Properties());
	}

	public Configuration(Properties props) {
		data = props;
	}

	/**
	 * Test if the key exists in this config instance
	 *
	 * @param key a key to test for
	 * @return <code>true</code> if and only if this config instance contains this
	 *         key
	 */
	public boolean containsKey(String key) {
		return data.containsKey(key);
	}

	/**
	 * Returns the values starting with <code>keyStartsWith</code>
	 * 
	 * valX=MIIG...= valY=MIIG...= val=MIIG...=
	 * 
	 * @param val
	 * @return
	 */
	public List<String> startsWith(String keyStartsWith) {
		return data.keySet().stream().map(String::valueOf).filter(obj -> obj.startsWith(keyStartsWith)).map(data::getProperty).collect(Collectors.toList());
	}

	/**
	 * Returns the value associated with <code>key</code> if it exists or
	 * <code>null</code> otherwise.
	 *
	 * @param key the name of the key
	 * @return the value for this key
	 */
	public String get(String key) {
		return data.getProperty(key);
	}

	/**
	 * Returns the value associated with <code>key</code> if it exists or
	 * <code>defaultValue</code> otherwise.
	 *
	 * @param key          the name of the key
	 * @param defaultValue a default value
	 * @return the value for this key
	 */
	public String get(String key, String defaultValue) {
		return data.getProperty(key, defaultValue);
	}

	/**
	 * Returns <code>true</code> if and only if a value given by <code>key</code>
	 * exists and is equal to the string <code>"true"</code>.
	 *
	 * @param key the name of the key
	 * @return the boolean value of this key
	 */
	public boolean getBoolean(String key) {
		return Boolean.parseBoolean(data.getProperty(key, "false"));
	}

	/**
	 * Returns <code>true</code> if and only if a value given by <code>key</code>
	 * exists and is equal to the string <code>"true"</code>, or
	 * <code>defaultValue</code> otherwise.
	 *
	 * @param key          the name of the key
	 * @param defaultValue a default value
	 * @return the boolean value of this key
	 */
	public boolean getBoolean(String key, boolean defaultValue) {
		boolean value;
		if (data.containsKey(key)) {
			value = getBoolean(key);
		} else {
			value = defaultValue;
		}
		return value;
	}

	/**
	 * Returns the integer value of <code>key</code> if it exists or <code>0</code>
	 * otherwise.
	 *
	 * @param key the name of the key
	 * @return the integer value of the key
	 * @throws NumberFormatException if the value associated with key is not an
	 *                               integer
	 */
	public int getInteger(String key) {
		return getInteger(key, 0);
	}

	/**
	 * Returns the integer value of <code>key</code> if it exists or
	 * <code>defaultValue</code> otherwise.
	 *
	 * @param key          the name of the key
	 * @param defaultValue a default value
	 * @return the integer value of the key
	 * @throws NumberFormatException if the value associated with key is not an
	 *                               integer
	 */
	public int getInteger(String key, int defaultValue) {
		int value;
		if (data.containsKey(key)) {
			value = Integer.parseInt(data.getProperty(key));
		} else {
			value = defaultValue;
		}
		return value;
	}

	/**
	 * Returns the long value of <code>key</code> if it exists or <code>0</code>
	 * otherwise.
	 *
	 * @param key the name of the key
	 * @return the long value of the key
	 * @throws NumberFormatException if the value associated with key is not a long
	 */
	public long getLong(String key) {
		return getLong(key, 0);
	}

	/**
	 * Returns the long value of <code>key</code> if it exists or
	 * <code>defaultValue</code> otherwise.
	 *
	 * @param key          the name of the key
	 * @param defaultValue a default value
	 * @return the long value of the key
	 * @throws NumberFormatException if the value associated with key is not a long
	 */
	public long getLong(String key, long defaultValue) {
		long value;
		if (data.containsKey(key)) {
			value = Long.parseLong(data.getProperty(key));
		} else {
			value = defaultValue;
		}
		return value;
	}

	/**
	 * Returns a <code>URL</code> if and only if a value given by <code>key</code>
	 * exists and is wellformed, otherwise <code>null</code> is returned.
	 *
	 * @param key the name of the key
	 * @return the URL from the value of this key
	 */
	public URL getUrl(String key) {
		return getUrl(key, null);
	}

	/**
	 * Returns a <code>URL</code> if and only if a value given by <code>key</code>
	 * exists and is wellformed, otherwise <code>defaultValue</code> is returned.
	 *
	 * @param key          the name of the key
	 * @param defaultValue a default value
	 * @return the URL from the value of this key
	 */
	public URL getUrl(String key, URL defaultValue) {
		URL value = null;
		if (data.containsKey(key)) {
			try {
				value = new URL(data.getProperty(key));
			} catch (MalformedURLException e) {
				value = defaultValue;
			}
		} else {
			value = defaultValue;
		}
		return value;
	}

	private void setProperty(String key, String value) {
		data.setProperty(key, value);
	}

	/**
	 * Example: prefix.variable.name = value prefix.variable.name2 = value2
	 * prefix.variable2.name3 = value3 prefix.variable2.name4 = value4
	 *
	 * assumptions: variable can be subdivided, so can prefix, but name cannot.
	 *
	 * @param prefix the prefix of the property key to be used
	 * @return a map from variable to {name, value}-Configurations.
	 */
	public Map<String, Configuration> getPropertySlice(String prefix) {
		final Map<String, Configuration> result = new HashMap<String, Configuration>();
		final Set<String> strings = data.stringPropertyNames();
		for (String string : strings) {
			if (string.startsWith(prefix)) {
				final String variableAndName = string.substring(prefix.length() + 1); // the dot
				final int endIndex = variableAndName.lastIndexOf('.');
				if (endIndex == -1)
					continue;
				final String variable = variableAndName.substring(0, endIndex);
				final String name = variableAndName.substring(endIndex + 1);
				Configuration slice = result.get(variable);
				if (slice == null) {
					slice = new Configuration();
					result.put(variable, slice);
				}
				slice.setProperty(name, data.getProperty(string));
			}
		}
		return result;
	}

	@Override
	public String toString() {
		return "Configuration{" + "data=" + data + '}';
	}

	public Properties getProperties() {
		Properties props = new Properties();
		props.putAll(data);
		return props;
	}
}
