/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.certificate;

import dk.sosi.sts.certificate.OcesInfo.CertificateType;
import dk.sosi.sts.model.SubjectSerialNumber;
import junit.framework.TestCase;

public class Oces2InfoParserTest extends TestCase {

    public void testMoces() {
        SubjectSerialNumber ssn = new SubjectSerialNumber("CVR:19343634-RID:1165813849809");
        OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

        assertEquals(ssn, info.getSubjectSerialNumber());
        assertEquals("19343634", info.getCvr());
        assertEquals(CertificateType.MOCES, info.getCertificateType());
    }

    public void testVoces() {
        SubjectSerialNumber ssn = new SubjectSerialNumber("CVR:19343634-UID:1165813849809");
        OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

        assertEquals(ssn, info.getSubjectSerialNumber());
        assertEquals("19343634", info.getCvr());
        assertEquals(CertificateType.VOCES, info.getCertificateType());
    }

    public void testFoces() {
        SubjectSerialNumber ssn = new SubjectSerialNumber("CVR:30808460-FID:1320135775022");
        OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

        assertEquals(ssn, info.getSubjectSerialNumber());
        assertEquals("30808460", info.getCvr());
        assertEquals(CertificateType.FOCES, info.getCertificateType());
    }

    public void testSpecialCharacters() {
        SubjectSerialNumber ssn = new SubjectSerialNumber("CVR:19343634-RID:1165813849809-1234#=%&/\\5678");
        OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

        assertEquals(ssn, info.getSubjectSerialNumber());
        assertEquals("19343634", info.getCvr());
        assertEquals(CertificateType.MOCES, info.getCertificateType());
    }

    public void testIllegalCommaCharacter() {
        SubjectSerialNumber without = new SubjectSerialNumber("CVR:19343634-RID:1165813849809");
        SubjectSerialNumber ssn = new SubjectSerialNumber("CVR:19343634-RID:1165813849809,1234");
        OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

        assertEquals(without, info.getSubjectSerialNumber());
        assertFalse(ssn.equals(info.getSubjectSerialNumber()));
        assertEquals("19343634", info.getCvr());
        assertEquals(CertificateType.MOCES, info.getCertificateType());
    }

    public void testIllegalPlusCharacter() {
        SubjectSerialNumber without = new SubjectSerialNumber("CVR:19343634-RID:1165813849809");
        SubjectSerialNumber ssn = new SubjectSerialNumber("CVR:19343634-RID:1165813849809+1234");
        OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

        assertEquals(without, info.getSubjectSerialNumber());
        assertFalse(ssn.equals(info.getSubjectSerialNumber()));
        assertEquals("19343634", info.getCvr());
        assertEquals(CertificateType.MOCES, info.getCertificateType());
    }

    public void testIllegalSpaceCharacter() {
        SubjectSerialNumber without = new SubjectSerialNumber("CVR:19343634-RID:1165813849809");
        SubjectSerialNumber ssn = new SubjectSerialNumber("CVR:19343634-RID:1165813849809 1234");
        OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

        assertEquals(without, info.getSubjectSerialNumber());
        assertFalse(ssn.equals(info.getSubjectSerialNumber()));
        assertEquals("19343634", info.getCvr());
        assertEquals(CertificateType.MOCES, info.getCertificateType());
    }

    public void testFullSubjectDN() {
        SubjectSerialNumber ssn = new SubjectSerialNumber("C=DK,O=Praksisudv. i Region Nordjylland // CVR:29541337,SERIALNUMBER=CVR:29541337-RID:SC180913-857721+CN=Peter Sture Winsløw (1081:00PDX)");
        OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

        assertEquals(new SubjectSerialNumber("CVR:29541337-RID:SC180913-857721"), info.getSubjectSerialNumber());
        assertEquals("29541337", info.getCvr());
        assertEquals(CertificateType.MOCES, info.getCertificateType());
    }

    public void testOces() {
        SubjectSerialNumber ssn = new SubjectSerialNumber("PID:1111-1111-1-111111111111");
        OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

        assertEquals(new SubjectSerialNumber("PID:1111-1111-1-111111111111"), info.getSubjectSerialNumber());
        assertNull(info.getCvr());
        assertEquals(CertificateType.POCES, info.getCertificateType());
    }
}