/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.certificate;

import dk.sosi.sts.certificate.OcesInfo.CertificateType;
import dk.sosi.sts.model.SubjectSerialNumber;
import junit.framework.TestCase;

public class Oces3InfoParserTest extends TestCase {

	public void testMocesGlobalUuid() {
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-E:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(ssn, info.getSubjectSerialNumber());
		assertNull(info.getCvr());
		assertEquals(CertificateType.MOCES, info.getCertificateType());
		assertEquals(OcesInfo.PersistenceLevel.GLOBAL, info.getPersistenceLevel());
	}

	public void testMocesLongTermUuid() {
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-E:C:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(ssn, info.getSubjectSerialNumber());
		assertNull(info.getCvr());
		assertEquals(CertificateType.MOCES, info.getCertificateType());
		assertEquals(OcesInfo.PersistenceLevel.LONG_TERM, info.getPersistenceLevel());
	}

	public void testMocesSessionSpecificUuid() {
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-E:S:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(ssn, info.getSubjectSerialNumber());
		assertNull(info.getCvr());
		assertEquals(CertificateType.MOCES, info.getCertificateType());
		assertEquals(OcesInfo.PersistenceLevel.SESSION_SPECIFIC, info.getPersistenceLevel());
	}

	public void testMocesUnknownPersistenceLevel() {
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-E:K:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(CertificateType.UNKNOWN, info.getCertificateType());
	}

	public void testVoces() {
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(ssn, info.getSubjectSerialNumber());
		assertNull(info.getCvr());
		assertEquals(CertificateType.VOCES, info.getCertificateType());
		assertEquals(OcesInfo.PersistenceLevel.GLOBAL, info.getPersistenceLevel());
	}

	public void testPoces() {
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-P:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(ssn, info.getSubjectSerialNumber());
		assertNull(info.getCvr());
		assertEquals(CertificateType.POCES, info.getCertificateType());
	}

	public void testSpecialCharacters() {
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b-1234#=%&/\\5678");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(ssn, info.getSubjectSerialNumber());

		assertEquals(null, info.getCvr());
		assertEquals(CertificateType.VOCES, info.getCertificateType());
	}

	public void testIllegalCommaCharacter() {
		SubjectSerialNumber without = new SubjectSerialNumber("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b,1234");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(without, info.getSubjectSerialNumber());
		assertFalse(ssn.equals(info.getSubjectSerialNumber()));
		assertEquals(null, info.getCvr());
		assertEquals(CertificateType.VOCES, info.getCertificateType());
	}

	public void testIllegalPlusCharacter() {
		SubjectSerialNumber without = new SubjectSerialNumber("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b+1234");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(without, info.getSubjectSerialNumber());

		assertFalse(ssn.equals(info.getSubjectSerialNumber()));

		assertEquals(null, info.getCvr());
		assertEquals(CertificateType.VOCES, info.getCertificateType());
	}

	public void testIllegalSpaceCharacter() {
		SubjectSerialNumber without = new SubjectSerialNumber("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
		SubjectSerialNumber ssn = new SubjectSerialNumber("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b 1234");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(without, info.getSubjectSerialNumber());

		assertFalse(ssn.equals(info.getSubjectSerialNumber()));

		assertEquals(null, info.getCvr());
		assertEquals(CertificateType.VOCES, info.getCertificateType());
	}

	public void testFullSubjectDN() {
		SubjectSerialNumber ssn = new SubjectSerialNumber("C=DK, OID.2.5.4.97=NTRDK-94354969, O=Testorganisation nr. 94354969, SERIALNUMBER=UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b, CN=VOCES_gyldig");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(new SubjectSerialNumber("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b"), info.getSubjectSerialNumber());
		assertEquals(null, info.getCvr());
		assertEquals(CertificateType.VOCES, info.getCertificateType());
	}


	public void testMoces() {
		SubjectSerialNumber ssn = new SubjectSerialNumber("CVR:19343634-RID:1165813849809");
		OcesInfo info = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertEquals(ssn, info.getSubjectSerialNumber());
		assertEquals("19343634", info.getCvr());
		assertEquals(CertificateType.MOCES, info.getCertificateType());
	}
}

