/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.sla;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.concurrent.CountDownLatch;

import javax.servlet.http.HttpServletRequest;

import junit.framework.TestCase;

import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sdsd.nsp.slalog.impl.SLALoggerDummyImpl;

public class SLALogHelperTest extends TestCase {
	
	@Override
	protected void setUp() throws Exception {
		StsContext.clearContext();
		SLALogHelper.clearAll();
	}
	
	@Override
	protected void tearDown() throws Exception {
		StsContext.clearContext();
		SLALogHelper.clearAll();
	}
	
	public void testCreate() throws Exception {
		SLALogItem item1 = SLALogHelper.createItem(SLALogPoint.STS_ISSUEIDCARD);
		assertNotNull(item1);
		assertSame(item1, SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));
		assertEquals(SLALogPoint.STS_ISSUEIDCARD.getId(), item1.getLogPointUniqId());
		assertEquals(SLALogPoint.STS_ISSUEIDCARD.getName(), item1.getLogPoint());

		SLALogItem item2 = SLALogHelper.createItem(SLALogPoint.STS_ISSUEIDCARD);
		assertSame(item2, SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));
		assertNotSame(item1, item2);
	}
	
	public void testGet() throws Exception {
		SLALogHelper.createItem(SLALogPoint.STS_ISSUEIDCARD);
		assertNotNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));
		// check that the SLALogItem is not accessible from another thread
		final SLALogItem[] holder = new SLALogItem[1];
		final CountDownLatch latch = new CountDownLatch(1);
		new Thread() {
			@Override public void run() {
				holder[0] = SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD);
				latch.countDown();
			}
		}.start();
		latch.await();
		assertNull(holder[0]);
	}
	
	public void testClear() throws Exception {
		SLALogHelper.createItem(SLALogPoint.STS_ISSUEIDCARD);
		assertNotNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));

		SLALogHelper.clearItem(SLALogPoint.STS_ISSUEIDCARD);
		assertNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));
	}

	public void testClearAll() throws Exception {
		SLALogHelper.createItem(SLALogPoint.STS_ISSUEIDCARD);
		SLALogHelper.createItem(SLALogPoint.ITS_SERIALIZE);

		SLALogHelper.clearAll();
		
		assertNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));
		assertNull(SLALogHelper.getItem(SLALogPoint.ITS_SERIALIZE));
	}
	
	public void testStoreCallResultOk() throws Exception {
		StsContext.getContext().setCallerIp("127.0.0.1");
		StsContext.getContext().setMessageId("1234567890");
		
		SLALogItem item = SLALogHelper.createItem(SLALogPoint.STS_ISSUEIDCARD);
		assertNotNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));

		SLALogHelper.storeCallResultOk(SLALogPoint.STS_ISSUEIDCARD);
		assertNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));
		assertTrue(item.isCallOk());		
		assertEquals(StsContext.getContext().getCallerIp(), item.getClientIP());
		assertEquals(StsContext.getContext().getMessageId(), item.getMessageId());
	}

	public void testStoreCallResultOk_CallerIPFromRequest() throws Exception {
		HttpServletRequest request = mock(HttpServletRequest.class);
		when(request.getHeader("X-Forwarded-For")).thenReturn("0.0.0.0");
		
		StsContext.getContext().setRequest(request);
		StsContext.getContext().setMessageId("1234567890");
		
		SLALogItem item = SLALogHelper.createItem(SLALogPoint.STS_ISSUEIDCARD);

		SLALogHelper.storeCallResultOk(SLALogPoint.STS_ISSUEIDCARD);

		assertNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));
		assertTrue(item.isCallOk());		
		assertEquals("0.0.0.0", item.getClientIP());
		assertEquals(StsContext.getContext().getMessageId(), item.getMessageId());
	}

	
	public void testStoreCallResultOk_MissingLogItem() throws Exception {
		assertNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));

		// should fail without exception
		SLALogHelper.storeCallResultOk(SLALogPoint.STS_ISSUEIDCARD);		
	}

	public void testStoreCallResultError() throws Exception {
		StsContext.getContext().setCallerIp("127.0.0.1");
		StsContext.getContext().setMessageId("1234567890");

		SLALogItem item = SLALogHelper.createItem(SLALogPoint.STS_ISSUEIDCARD);
		assertNotNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));

		SLALogHelper.storeCallResultError(SLALogPoint.STS_ISSUEIDCARD, "test error");
		assertNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));
		assertFalse(item.isCallOk());
		assertEquals(StsContext.getContext().getCallerIp(), item.getClientIP());
		assertEquals(StsContext.getContext().getMessageId(), item.getMessageId());
	}
	
	public void testStoreCallResultError_MissingLogItem() throws Exception {
		assertNull(SLALogHelper.getItem(SLALogPoint.STS_ISSUEIDCARD));

		// should fail without exception
		SLALogHelper.storeCallResultError(SLALogPoint.STS_ISSUEIDCARD, "test error");
	}
	
	public void testSetLogger() throws Exception {
		// dummy test
		SLALogHelper.setLogger(new SLALoggerDummyImpl());
	}
}
