/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.timing;

import java.util.Arrays;
import java.util.Collections;

import junit.framework.TestCase;

public class TimerTest extends TestCase {
	
	private TestSink sink;
	private Timer timer;

	@Override
	protected void setUp() throws Exception {
		sink = new TestSink();
		timer = new Timer(sink);
	}

	public void testTimer() {
		timer.enter(getClass(), "op1");
		assertEquals(1, 0, 0, sink);
		timer.enter(getClass(), "op2");
		assertEquals(2, 0, 0, sink);

		timer.exit(getClass(), "op2");
		assertEquals(2, 1, 0, sink);
		timer.exit(getClass(), "op1");
		assertEquals(2, 2, 0, sink);
	}

	public void testUnmatchedClass() {
		timer.enter(getClass(), "op1");
		assertEquals(1, 0, 0, sink);

		timer.exit(Object.class, "op1");
		assertEquals(1, 0, 1, sink);
	}

	public void testUnmatchedOperation() {
		timer.enter(getClass(), "op1");
		timer.enter(getClass(), "op2");
		assertEquals(2, 0, 0, sink);

		timer.exit(getClass(), "op1");
		assertEquals(2, 0, 2, sink);
	}
	
	public void testCapacityExceeded() {
		final int MAX_CAPACITY = 8;
		timer = new Timer(sink, Collections.<String>emptyList(), MAX_CAPACITY);
		
		for (int i = 0; i < MAX_CAPACITY; i++) {
			timer.enter(getClass(), "op" + i);
		}
		assertEquals(MAX_CAPACITY, 0, 0, sink);
		
		timer.enter(getClass(), "op");
		assertEquals(MAX_CAPACITY, 0, MAX_CAPACITY, sink);
	}
	
	public void testStartStop() {
		Timer.setTimer(timer);
		try {
			Timer.start(getClass(), "op");
			assertEquals(1, 0, 0, sink);
			Timer.stop(getClass(), "op");
			assertEquals(1, 1, 0, sink);
		} finally {
			// clean-up
			Timer.setTimer(new Timer.NopTimer());
		}
	}
	
	public void testFilter() throws Exception {
		timer = new Timer(sink, 
				Arrays.asList("Object#filter"),
				32);
		
		timer.enter(Object.class, "op1");
		timer.enter(Object.class, "filter1");
		timer.enter(Object.class, "op2");
		timer.exit(Object.class, "op2");
		timer.exit(Object.class, "filter1");
		timer.exit(Object.class, "op1");
		
		// 1 enter, 1 exit, 0, discarded (i.e. 2 filtered)
		assertEquals(1, 1, 0, sink);
	}
	
	private void assertEquals(int entered, int exited, int discarded, TestSink sink) {
		assertEquals(entered, sink.entered);
		assertEquals(exited, sink.exited);
		assertEquals(discarded, sink.discarded);
	}

	private static class TestSink implements TimeSink {
		public int discarded;
		public int entered;
		public int exited;

		public void enter(TimerEvent e) {
			entered++;
		}
		public void exit(TimerEvent e, long duration) {
			exited++;
		}
		public void discard(TimerEvent e, String reason) {
			discarded++;
		}
	}
}
