/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.monitor;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.sts.server.StsStatus;
import dk.sosi.sts.util.LogUtil;
import org.apache.log4j.Logger;

import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.Map;

public class StatusCheck implements Check {

	private final Logger log = Logger.getLogger(StatusCheck.class);

	private final StsStatus stsStatus;

	private final SOSIFactory sosiFactory;

	private final CertificateStatusChecker crlService;

	public StatusCheck(final StsStatus stsStatus, final SOSIFactory sosiFactory, final CertificateStatusChecker crlService) {
		this.stsStatus = stsStatus;
		this.sosiFactory = sosiFactory;
		this.crlService = crlService;
	}


  public CheckResult performCheck() {
    final X509Certificate certificate = sosiFactory
            .getCredentialVault()
            .getSystemCredentialPair()
            .getCertificate();

    final Map<String, String> result = new LinkedHashMap<String, String>();
    final Date now = new Date();


    final boolean validStsCert =
            now.after(certificate.getNotBefore()) && now.before(certificate.getNotAfter());

	  boolean failed = !validStsCert || !revocationCheck(certificate);
		if (!failed) {
			try {
				sosiFactory.getSignatureProvider().sign(certificate.getEncoded());
			} catch (Exception e) {
				failed = true;
				log.trace("Signing test failed", e);
			}
		}

    result.put("Version", stsStatus.getVersion());
    result.put("Local time", now.toString());
    result.put("Started", stsStatus.getStartedDate());
    result.put("Last transaction", MonitorUtil.toDate(stsStatus.getLatestTransaction()));
    result.put("Last issued", MonitorUtil.toDate(stsStatus.getLastIssued()));
    result.put("Sts certificate, valid from", certificate.getNotBefore().toString());
    result.put("Sts certificate, valid until", certificate.getNotAfter().toString());
    result.put("Sts certificate valid", !failed ? "OK" : "not valid");
    result.put("Requests, 24h", ""+stsStatus.getTransactions24h());
	result.put("Requests, total", ""+stsStatus.getTransactionsForever());
	  result.put("Failed requests, 24h", ""+stsStatus.getErrors24h());
	  result.put("Failed requests, total", ""+stsStatus.getErrorsForever());
    result.put("Response time, avg 24h", ""+stsStatus.getAverageTime24h());
    result.put("Response time, max 24h", ""+stsStatus.getMaxTime24h());

    return new CheckResult(failed, result);
  }

	private boolean revocationCheck(final X509Certificate certificate) {
		final CertificateStatus revocationStatus;
		try {
			revocationStatus = crlService.getRevocationStatus(certificate);
		} catch(Throwable t) {
			LogUtil.logOnce(StatusCheck.class.getName(), log, t);
			return false;
		}
		return revocationStatus.isValid();
	}
}
