/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.monitor;

import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.TrustedVaultProvider;
import dk.sosi.sts.util.LogUtil;
import org.apache.log4j.Logger;

import java.security.KeyStoreException;
import java.security.cert.X509Certificate;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class TrustedVaultCheck implements Check {
    private final Logger log = Logger.getLogger(getClass());


    private final TrustedVaultProvider vaultProvider;
    private final CertificateStatusChecker crlService;
    private final String vaultName;
    private final List<String> ignoresAliases;

    public TrustedVaultCheck(TrustedVaultProvider vaultProvider, CertificateStatusChecker crlService, String vaultName, List<String> ignoresAliases) {
        this.vaultProvider = vaultProvider;
        this.crlService = crlService;
        this.vaultName = vaultName;
        this.ignoresAliases = ignoresAliases;
    }

    @Override
    public CheckResult performCheck() {
        final CredentialVault trustedVault = vaultProvider.getTrustedVault();
        try {
            boolean failed = false;
            Map<String, String> messages = new HashMap<>();
            final Enumeration<String> aliases = trustedVault.getKeyStore().aliases();
            while (aliases.hasMoreElements()) {
                String alias = aliases.nextElement();
                X509Certificate cert = (X509Certificate) trustedVault.getKeyStore().getCertificate(alias);
                boolean valid = checkCertificate(cert);
                boolean ignored = ignoresAliases.contains(alias);
                boolean revocationStatusOk = true;
                if (!ignored) {
                	revocationStatusOk = revocationCheck(cert);
                }
                final String certDescription = new DistinguishedName(cert.getSubjectX500Principal()).getSubjectSerialNumber();
                final String key = "Trusted certificate " + vaultName + "/" + alias + " -> " + certDescription;
                String expiryDateText =  String.format("(Expires: %1$td-%1$tm-%1$tY %1$tH:%1$tM:%1$tS)", cert.getNotAfter());
				if (ignored) {
                    messages.put(key, "Ignored "+expiryDateText);
                } else if (!valid) {
                	failed = true;
                    messages.put(key, "Expired "+expiryDateText);
                } else if (!revocationStatusOk) {
                	failed = true;
                    messages.put(key, "Revoked "+expiryDateText);
                } else {
                    messages.put(key, "OK "+expiryDateText);
                }
            }
            return new CheckResult(failed, messages);
        } catch (KeyStoreException e) {
            LogUtil.logOnce(StatusCheck.class.getName(), log, e);
            return new CheckResult(true, Collections.singletonMap("TrustedVaultCheck: " + vaultName, e.getMessage()));
        }
    }

    private boolean checkCertificate(X509Certificate certificate) {
        Date now = new Date();
        return now.after(certificate.getNotBefore()) && now.before(certificate.getNotAfter());
    }

    private boolean revocationCheck(final X509Certificate certificate) {
        final CertificateStatus revocationStatus;
        try {
            revocationStatus = crlService.getRevocationStatus(certificate);
        } catch(Throwable t) {
            LogUtil.logOnce(StatusCheck.class.getName(), log, t);
            return false;
        }
        return revocationStatus.isValid();
    }

}
