/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.roles.nationalroles;

import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.Oces2InfoParser;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.service.NationalRoleService;
import dk.sosi.sts.service.Uuid2RidService;
import org.apache.log4j.Logger;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.PreparedStatementSetter;
import org.springframework.jdbc.core.RowMapper;

import javax.sql.DataSource;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

public class DbNationalRoleService implements NationalRoleService {
    private final Logger log = Logger.getLogger(this.getClass());
    private DataSource dataSource;
    private String sql = "select * from nationalRoles where ssn = ? and ValidFrom <= ? and ValidTo > ?";
    private Uuid2RidService uuid2RidService;

    // invoked by spring config
    public void setDataSource(DataSource dataSource) {
        this.dataSource = dataSource;
    }

    // invoked by spring config
    public void setSql(String sql) {
        this.sql = sql;
    }

    // invoked by spring config
    public void setUuid2RidService(Uuid2RidService uuid2RidService) {
        this.uuid2RidService = uuid2RidService;
    }

    @Override
    public List<String> getRoles(OcesInfo ocesInfo) {
        SubjectSerialNumber subjectSerialNumber = ocesInfo.getSubjectSerialNumber();
        if (subjectSerialNumber == null) {
            return Collections.emptyList();
        }

        List<String> roles = lookupRoles(subjectSerialNumber.getValue());

        if (roles.isEmpty() && ocesInfo.getVersion() == OcesInfo.Version.OCES3) {
            Optional<String> relatedRid = findRelatedRid(subjectSerialNumber);
            if (relatedRid.isPresent()) {
                SubjectSerialNumber cvrRidSsn = Oces2InfoParser.subjectSerialNumberFromCvrAndRid(ocesInfo.getCvr(), relatedRid.get());
                roles = lookupRoles(cvrRidSsn.getValue());
            }
        }

        return roles;
    }

    private Optional<String> findRelatedRid(SubjectSerialNumber oces3Ssn) {
        try {
            return uuid2RidService.findRelatedRid(oces3Ssn.getValue());
        } catch (StsException e) {
            log.error("Unable to lookup rid from uuid. ", e);
        }
        return Optional.empty();
    }

    private List<String> lookupRoles(String inputId) {
        JdbcTemplate template = new JdbcTemplate(dataSource);
        long now = System.currentTimeMillis();
        final Timestamp currentTime = new Timestamp(now);
        PreparedStatementSetter setter = new PreparedStatementSetter() {
            @Override
            public void setValues(PreparedStatement preparedStatement) throws SQLException {
                preparedStatement.setString(1, inputId);
                preparedStatement.setTimestamp(2, currentTime);
                preparedStatement.setTimestamp(3, currentTime);
            }
        };
        RowMapper<String> rowMapper = new RowMapper<String>() {
            @Override
            public String mapRow(ResultSet resultSet, int i) throws SQLException {
                String from = resultSet.getString("ValidFrom");
                String to = resultSet.getString("ValidTo");
                return resultSet.getString("externalName");
            }
        };

        return template.query(sql, setter, rowMapper);
    }
}