/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.roles.roledefinitions;

import dk.sosi.sts.model.RoleDefinition;
import dk.sosi.sts.server.cache.DefinitionServiceCache;
import dk.sosi.sts.service.RoleDefinitionService;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.InitializingBean;

import java.util.List;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicReference;

public class CachingRoleDefinitionsService implements RoleDefinitionService, InitializingBean, DefinitionServiceCache {
	private final Logger log = Logger.getLogger(this.getClass());

	private final AtomicReference<List<RoleDefinition>> cache = new AtomicReference<>();
	private final Object cacheRefreshLock = new Object();
	private final RoleDefinitionService backend;

	public CachingRoleDefinitionsService(RoleDefinitionService backend) {
		this.backend = backend;
	}

	@Override
	public void afterPropertiesSet() {
		refresh();
	}

	@Override
	public List<RoleDefinition> getRoleDefinitions() {
		return ensureCache();
	}

	@Override
	public List<String> getTrustedRoleClaimers(String endpoint, String roleName) {
		return backend.getTrustedRoleClaimers(endpoint, roleName);
	}

	@Override
	public List<String> getTrustedRoleClaimers(String endpoint, Optional<String> roleName) {
		return backend.getTrustedRoleClaimers(endpoint, roleName);
	}

	public void refresh() {
		synchronized (cacheRefreshLock) {
			try {
				refreshCache();
			} catch (RuntimeException e) {
				log.error("An error occured when reading from the Roledefinition tables.", e);
			}
		}
	}

	private List<RoleDefinition> ensureCache() {
		List<RoleDefinition> cached = cache.get();
		if (cached == null) {
			synchronized (cacheRefreshLock) {
				cached = cache.get();
				if (cached == null) {
					return refreshCache();
				} else {
					return cached;
				}
			}
		} else {
			return cached;
		}
	}

	private List<RoleDefinition> refreshCache() {
		log.info("Updating the RoleDefinition Cache");
		final List<RoleDefinition> roleDefinitions = backend.getRoleDefinitions();
		this.cache.set(roleDefinitions);
		log.info("Found " + roleDefinitions.size() + " role definitionss");
		return roleDefinitions;
	}

	@Override
	public String getCacheId() {
		return "RoleDefinition";
	}
}