/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.model.*;
import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.StsException;
import dk.sosi.sts.integrations.blurring.BlurringService;
import dk.sosi.sts.integrations.personinformation.Custody;
import dk.sosi.sts.integrations.personinformation.PersonInformationClient;
import dk.sosi.sts.integrations.personinformation.PersonInformationClientException;
import dk.sosi.sts.its.DocumentResponse;
import dk.sosi.sts.model.AudienceConfiguration;
import dk.sosi.sts.model.BlurringVersion;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.PrivilegeGroup;
import dk.sosi.sts.server.service.Timed;
import dk.sosi.sts.service.ItsConfigService;
import dk.sosi.sts.service.ProcurationService;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import java.io.IOException;
import java.io.OutputStream;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;

import static dk.sosi.seal.model.BlurredOrganisation.OrgType;
import static dk.sosi.seal.model.BlurredOrganisation.Reason;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.*;
import static dk.sosi.sts.OptionalExt.toStream;
import static dk.sosi.sts.model.BlurredOrganization.Type;

/**
 * Created by Kent on 21-02-2017.
 */
public abstract class AbstractIdwsRequestHandler<R extends OIOWSTrustRequest, ConfigType> extends AbstractStsRequestHandler<R, DocumentResponse, ConfigType> {
    protected ItsConfigService itsConfigService;
    protected ProcurationService procurationService;

    private static final Logger logger = Logger.getLogger(AbstractIdwsRequestHandler.class);
    private PersonInformationClient personInformationClient;

    // invoked from spring
    public void setItsConfigService(ItsConfigService itsConfigService) {
        this.itsConfigService = itsConfigService;
    }

    // invoked from spring
    public void setProcurationService(ProcurationService procurationService) {
        this.procurationService = procurationService;
    }
    // invoked by spring
    public void setPersonInformationClient(PersonInformationClient personInformationClient) {
        this.personInformationClient = personInformationClient;
    }

    @Override
    protected final boolean isFault(DocumentResponse stResponse) {
        return stResponse.isFault();
    }

    @Override
    @Timed
    protected final void serializeToOutputStream(DocumentResponse doc, OutputStream outputStream) throws IOException {
        outputStream.write(XmlUtil.serializeXml2ByteArray(serialize(doc), true));
    }

    @Override
    protected final Document serialize(DocumentResponse doc) {
        return doc.getDocument();
    }

    protected final X509Certificate verifyConsumer(OIOWSTrustRequest request, AudienceConfiguration audienceConfiguration) throws VerifyException {
        validateMessageSignature(request);
        X509Certificate signingCertificate = request.getSigningCertificate();
        verifySigningCertificate(signingCertificate);
        verifyConsumerWhitelisted(signingCertificate, audienceConfiguration);
        return signingCertificate;
    }

    private void verifyConsumerWhitelisted(X509Certificate signingCertificate, AudienceConfiguration config) throws VerifyException {
        final String certificateDn = signingCertificate.getSubjectDN().getName();
        if (!config.getCertificates().contains(new DistinguishedName(certificateDn).getSubjectSerialNumber())) {
            final String faultString = "Signing certificate not whitelisted for this audience " +
                    "(cert={"+certificateDn+"}, audience={"+config.getAudience()+"})";
            throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, faultString, StsFaultActorValues.ITS);

        }
    }

    private void verifySigningCertificate(X509Certificate signingCertificate) throws VerifyException {
        new CertificateStatusVerifier(sosiFactory.getFederation())
                .verifyCertificateStatus(signingCertificate, StsFaultActorValues.ITS);
    }

    private void validateMessageSignature(OIOWSTrustRequest req) throws VerifyException {
        new OIOWSTrustValidation(StsFaultActorValues.ITS).validateSignature(req);
    }

    public BlurringInstructionsDomBuilder<Document> createBlurringInstructionsBuilder(BlurringService blurringService,
                                                                                      CprNr patientId,
                                                                                      Optional<CprNr> onBehalfOf,
                                                                                      Optional<SubjectRelations11Impl.VerifiedRelation> verifiedRelation,
                                                                                      String flowId,
                                                                                      BlurringVersion blurringVersion) {
        if (BlurringVersion.NONE.equals(blurringVersion)) {
            return null;
        } else if (BlurringVersion.V10.equals(blurringVersion)) {
            BlurringInstructions10DomBuilder builder = new BlurringInstructions10DomBuilder(blurringService.getSalt(flowId));

            Stream.of(
                getCitizenSpecificBlurrings(blurringService, patientId, SpecificForPerson, flowId),
                toStream(onBehalfOf).flatMap(o -> getCitizenSpecificBlurrings(blurringService, o, FromRelatedPerson, flowId)),
                toStream(verifiedRelation).flatMap(o -> getCitizenSpecificBlurrings(blurringService, new CprNr(o.getRelatedPersonID()), FromRelatedPerson, flowId))
            )
            .flatMap(Function.identity())
            .forEach(builder::addBlurringInstruction);

            return builder;
        } else if (BlurringVersion.V11.equals(blurringVersion)) {
            BlurringInstructions11DomBuilder builder = new BlurringInstructions11DomBuilder(blurringService.getSalt(flowId));

            Stream.of(
                getCitizenSpecificBlurrings(blurringService, patientId, SpecificForPerson, flowId),
                getAllActiveOrgBlurrings(blurringService, flowId),
                toStream(onBehalfOf).flatMap(o -> getCitizenSpecificBlurrings(blurringService, o, FromRelatedPerson, flowId)),
                toStream(verifiedRelation).flatMap(o -> getCitizenSpecificBlurrings(blurringService, new CprNr(o.getRelatedPersonID()), FromRelatedPerson, flowId))
            )
            .flatMap(Function.identity())
            .forEach(builder::addBlurringInstruction);

            return builder;
        } else {
            String error = "Internal error. Could not map value '" + blurringVersion + "' to type '" + BlurringVersion.class + "'.";
            logger.error(error);
            throw new RuntimeException(error);
        }
    }

    private static Stream<BlurredOrganisation> getCitizenSpecificBlurrings(BlurringService blurringService,
                                                                           CprNr cpr,
                                                                           Reason reason,
                                                                           String flowId) {
        return blurringService.getActiveBlurrings(cpr, flowId).stream()
                .map(o -> new BlurredOrganisation(o.value, toOrgTypeForCitizenBlurrings(o.type), reason));
    }

    private static Stream<BlurredOrganisation> getAllActiveOrgBlurrings(BlurringService blurringService,
                                                                        String flowId) {
        return blurringService.getAllActiveOrgBlurrings(flowId).stream()
                .map(o -> new BlurredOrganisation(o.value, toOrgTypeForOrgBlurrings(o.type), SpecificDepartment));
    }

    private static OrgType toOrgTypeForCitizenBlurrings(Type t) {
        if (Objects.requireNonNull(t) == Type.Cvr) {
            return OrgType.Cvr;
        }
        String error = String.format("Internal error. Invalid org type for specific citizen blurring: %s", t);
        logger.error(error);
        throw new RuntimeException(error);
    }

    private static OrgType toOrgTypeForOrgBlurrings(Type t) {
        switch (t) {
            case Sor:
                return OrgType.Sor;
            case Shak:
                return OrgType.Shak;
            default:
                String error = String.format("Internal error. Invalid org type for org blurring: %s", t);
                logger.error(error);
                throw new RuntimeException(error);
        }
    }

    public BasicPrivilegesDOMBuilder createPrivilegeBuilder(CprNr patientId, CprNr onBehalfOf) throws VerifyException {
        if (onBehalfOf == null) { return null; }
        PrivilegeGroup procurationPrivileges = createProcurationPrivileges(patientId, onBehalfOf);
        BasicPrivilegesDOMBuilder privilegesDOMBuilder = new BasicPrivilegesDOMBuilder();
        procurationPrivileges.getPrivileges()
                .forEach(privilege -> privilegesDOMBuilder.addPrivilege(procurationPrivileges.getScope(), privilege));
        return privilegesDOMBuilder;
    }

    public SubjectRelationsDomBuilder createSubjectRelationsBuilder(SubjectRelations11Impl.VerifiedRelation verifiedRelation) {
        if (verifiedRelation == null) { return null; }
        SubjectRelations11Impl subjectRelations = new SubjectRelations11Impl();
        subjectRelations.addVerifiedRelation(verifiedRelation);
        return new SubjectRelations11DomBuilder(subjectRelations);
    }

    protected SubjectRelations11Impl.VerifiedRelation createValidatedSubjectRelations(String wardCustody, String parentCustody, String partlyWardCustody, String cprFromToken) throws VerifyException {
        final SubjectRelations11Impl.PersonIDType relatedPersonIDType = SubjectRelations11Impl.PersonIDType.CPR;

        // Kun en af 'wardCustody' eller 'parentCustody' kan være sat. 'partlyWardCustody' skal altid afvises (SDS-6981).
        if (partlyWardCustody != null) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
                                      "partlyWardCustody is not allowed as a Subject Relation", StsFaultActorValues.ITS);
        } else if (wardCustody != null && parentCustody != null) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
                                      "It's not allowed to have both 'wardCustody' og 'parentCustody'", StsFaultActorValues.ITS);
        } else if (wardCustody != null) {
            SubjectRelations11Impl.VerifiedRelation verifiedRelation = new SubjectRelations11Impl.VerifiedRelation(SubjectRelations11Impl.RelationType.WARD_CUSTODY_HOLDER, wardCustody, relatedPersonIDType);
            validateVerifiedRelation(verifiedRelation, cprFromToken);
            return verifiedRelation;
        } else if (parentCustody != null) {
            try {
                int age = personInformationClient.personAge(parentCustody);
                SubjectRelations11Impl.VerifiedRelation verifiedRelation = new SubjectRelations11Impl.VerifiedRelation(SubjectRelations11Impl.RelationType.PARENTAL_CUSTODY_HOLDER, parentCustody, relatedPersonIDType, age);
                validateVerifiedRelation(verifiedRelation, cprFromToken);
                return verifiedRelation;
            } catch (PersonInformationClientException e) {
                logger.error("Error when calling PersonInformationClient", e);
                throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
                                          "Unable to verify subject relation", StsFaultActorValues.ITS);
            }
        } else {
            return null;
        }
    }

    protected void detectMissingWhitelistingForWardClaim(AudienceConfiguration audienceConfig, X509Certificate signingCertificate, String wardCustody) throws VerifyException {
        List<String> allowedCertsWardClaim = audienceConfig.wardClaimAllowedCertificates();
        String signingCertSSN = new DistinguishedName(signingCertificate.getSubjectDN().getName()).getSubjectSerialNumber();
        boolean isWardClaimAllowed = allowedCertsWardClaim.contains(signingCertSSN);
        if (wardCustody != null && !isWardClaimAllowed) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
                    "Ward claim not whitelisted for this certificate SSN and audience",
                    StsFaultActorValues.ITS);
        }
    }

    private void validateVerifiedRelation(SubjectRelations11Impl.VerifiedRelation verifiedRelation, String cprFromToken) throws VerifyException {
        if (verifiedRelation == null) { return; }
        logger.trace("Validating subject relation");

        Function<Custody.CustodyType, Predicate<Custody>> custodyMatcher = custodyType ->
                c -> c.getCustodyType() == custodyType && c.getRelationCpr().equals(verifiedRelation.getRelatedPersonID());
        try {
            Set<Custody> custody = personInformationClient.personCustody(cprFromToken);
            if (verifiedRelation.getRelationType() == SubjectRelations11Impl.RelationType.WARD_CUSTODY_HOLDER) {
                if (custody.stream().noneMatch(custodyMatcher.apply(Custody.CustodyType.WARDCUSTODY))) {
                    throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
                                              "The person is not in ward custody of the given person", StsFaultActorValues.ITS);
                }
            } else if (verifiedRelation.getRelationType() == SubjectRelations11Impl.RelationType.PARENTAL_CUSTODY_HOLDER) {
                if (custody.stream().noneMatch(custodyMatcher.apply(Custody.CustodyType.PARENTALCUSTODY))) {
                    throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
                                              "The person is not in parental custody of the given person", StsFaultActorValues.ITS);
                }
            }

        } catch (PersonInformationClientException e) {
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
                                      "Unable to verify subject relation", StsFaultActorValues.ITS);
        }
    }

    private PrivilegeGroup createProcurationPrivileges(CprNr representative, CprNr onBehalfOf) throws VerifyException {
        PrivilegeGroup procurationPrivileges;
        try {
            procurationPrivileges = procurationService.getProcurationPrivileges(onBehalfOf, representative);
        } catch (StsException e) {
            log.error("Unable to get procurations", e);
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
                                      "Unable to get procurations", StsFaultActorValues.ITS);
        }

        if (procurationPrivileges.getPrivileges().isEmpty()) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
                                      "No privileges provided by procuration", StsFaultActorValues.ITS);
        }
        return procurationPrivileges;
    }

    protected final CprNr parseCpr(String cpr) throws VerifyException {
        try {
            return new CprNr(cpr);
        } catch (IllegalArgumentException e) {
            log.warn("Unable to parse cpr", e);
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, e.getMessage(), StsFaultActorValues.ITS);
        }
    }
}