/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.modelbuilders.SignatureInvalidModelBuildException;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.seal.xml.XmlUtilException;
import dk.sosi.sts.server.audit.EnvelopeCompressor;
import dk.sosi.sts.server.audit.X509CertificateCompressor;
import dk.sosi.sts.server.service.SelectiveTiming;
import dk.sosi.sts.server.war.LogHelper;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import dk.sosi.sts.timing.Timer;
import org.apache.axis.AxisFault;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import java.io.IOException;
import java.io.OutputStream;
import java.security.KeyStoreException;
import java.security.cert.X509Certificate;

@SelectiveTiming
public abstract class AbstractStsRequestHandler<RequestType, ResponseType, ConfigType> {

    protected static Logger log = Logger.getLogger("dk.sosi.sts.server.RequestHandler");
    protected static Logger auditLog = Logger.getLogger("STS.AUDIT");

	protected SOSIFactory sosiFactory;
	protected StsStatus stsStatus;
	protected CertificateStatusChecker crlService;

	public SOSIFactory getSosiFactory() {
		return sosiFactory;
	}

	public void setSosiFactory(final SOSIFactory sosiFactory) {
		this.sosiFactory = sosiFactory;
	}

	public StsStatus getStsStatus() {
		return stsStatus;
	}

	public void setStsStatus(final StsStatus stsStatus) {
		this.stsStatus = stsStatus;
	}

	public CertificateStatusChecker getCrlService() {
		return crlService;
	}

	public void setCrlService(final CertificateStatusChecker crlService) {
		this.crlService = crlService;
	}

	private EnvelopeCompressor envelopeCompressor;

	public AbstractStsRequestHandler() {
		envelopeCompressor = new EnvelopeCompressor(new X509CertificateCompressor());
	}

	/**
     * Deserialize the string representation to the RequestType.
     * 
     * @param doc the string representation
     * @return the RequestType
     */
    protected abstract RequestType deserializeRequest(String doc);


    /**
	 * Performs the request handler's job. If the job fails a
	 * <code>VerifyException</code> should be thrown.
	 * 
	 * @param req the request
	 * @param configType
     * @return the result
	 * @throws VerifyException 
	 */
    protected abstract ResponseType perform(RequestType req, ConfigType configType) throws VerifyException;
    
    /**
	 * Determines if the response is a fault and subsequently if the resulting
	 * soap message should be a fault.
	 * 
	 * @param stResponse
	 *            the response
	 * @return
	 */
    protected abstract boolean isFault(final ResponseType stResponse);

	protected final void checkCertificate(X509Certificate trustedCertificate, String certificateRole, String faultActor) throws VerifyException {
		long currentTime = System.currentTimeMillis();
		if (currentTime < trustedCertificate.getNotBefore().getTime()) {
			final String message = certificateRole + " not valid yet '" + trustedCertificate.getSubjectX500Principal().getName() + "'";
			log.error(message);
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, message, faultActor);
		} else if (currentTime > trustedCertificate.getNotAfter().getTime()) {
			final String message = certificateRole + " expired '" + trustedCertificate.getSubjectX500Principal().getName() + "'";
			log.error(message);
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, message, faultActor);
		} else if (!crlService.getRevocationStatus(trustedCertificate).isValid()) {
			final String message = certificateRole + " revoked '" + trustedCertificate.getSubjectX500Principal().getName() + "'";
			log.error(message);
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, message, faultActor);
		} else {
			log.debug(certificateRole + " '" + trustedCertificate.getSubjectX500Principal().getName() + "' verified");
		}
	}


	/**
	 * Serialize the response to a document to be part of the soap message.
	 * 
	 * @param doc
	 *            the response
	 * @return the document
	 */
    protected abstract Document serialize(ResponseType doc);

	protected abstract void serializeToOutputStream(ResponseType doc, OutputStream outputStream) throws IOException;

    /**
	 * Constructs an fault response, where no request could be parsed.
	 * 
	 * @param dgwsVersion
	 *            the version of the DGWS used -- optional.
	 * @param msgId
	 *            the message id of the request if found.
	 * @param faultCode
	 *            the necessary fault code for the soap response
	 * @param faultString
	 *            the necessary fault string for the soap response
	 * @param faultActor
	 *            the necessary fault actor the for the soap reponse
	 * @return the fault response constructed
	 */
    protected abstract ResponseType createErrorResponse(String dgwsVersion, String msgId, String faultCode, String faultString, String faultActor);
    
    /**
	 * Constructs an fault response with a parsed request.
	 * 
	 * @param stRequest
	 *            the parsed request
	 * @param faultCode
	 *            the necessary fault code for the soap response
	 * @param faultString
	 *            the necessary fault string for the soap response
	 * @param faultActor
	 *            the necessary fault actor the for the soap reponse
	 * @return the fault response constructed
	 */
    protected abstract ResponseType createErrorResponse(RequestType stRequest, String faultCode, String faultString, String faultActor);

    /**
     * Processes the soap request from Axis.
     * 
     *
	 * @param req
	 * @param resp
	 * @throws AxisFault
     */
    public boolean handleRequest(String req, OutputStream resp) throws AxisFault {
    	log.info("Handling request: "+this.getClass().getSimpleName());
        // Get Request object
        RequestType stsRequest;
        ResponseType stsResponse;
		boolean internalError = false;

        LogHelper.startRequest();
        SLALogItem slaLogItem = SLALogHelper.createItem(SLALogPoint.STSFACADE_REQUEST);
        String slaErrorMsg = null;
        
        long start = System.currentTimeMillis();
        // Build request object implies validation of Signature
        try {
            try {
                slaLogItem.setRequestSize(audit(req));
                stsRequest = timedDeserializeRequest(req);
                try {
                    verifyCertificates();
	                try {
		                final ConfigType configType = verifyConfiguration(stsRequest);
						stsResponse = perform(stsRequest, configType);
                    } catch (VerifyException e) {
                        log.debug("Request not verified: " + e.getFaultCode() + " " + e.getFaultString() + " " + e.getFaultActor());
						stsResponse = timedCreateAndTimeErrorResponse(
								stsRequest,
								e.getFaultCode(),
								e.getFaultString(),
								e.getFaultActor());
                    }
                } catch (RuntimeException t) {
                    slaErrorMsg = t.getMessage();
                    log.error("Request: Unspecified Error in STS", t);
                    stsResponse = timedCreateAndTimeErrorResponse(
                            stsRequest,
                            StsFaultCodeValues.WST_REQUEST_FAILED ,
                            StsFaultStringValues.WST_REQUEST_FAILED + ": " + t.getMessage(),
                            StsFaultActorValues.STS);
                }
            } catch (SignatureInvalidModelBuildException e) {
                slaErrorMsg = e.getMessage();
                log.debug("Request: Validation of signature in request failed", e);
                stsResponse = timedCreateAndTimeErrorResponse(
                        e.getDGWSVersion(),
                        e.getMessageID(),
                        StsFaultCodeValues.WST_AUTHENTICATION_BAD_ELEMENTS,
                        StsFaultStringValues.WST_AUTHENTICATION_BAD_ELEMENTS + ": " + e.getMessage(),
                        StsFaultActorValues.SEAL);
            } catch (XmlUtilException e) {
	            slaErrorMsg = e.getMessage();
	            log.error("Request: Could not be deserialized", e);
	            stsResponse = timedCreateAndTimeErrorResponse(
			            null,
			            StsFaultCodeValues.WST_INVALID_REQUEST ,
			            StsFaultStringValues.WST_INVALID_REQUEST + ": " + e.getMessage(),
			            StsFaultActorValues.STS);
				internalError = true;
            } catch (Exception e) {
                slaErrorMsg = e.getMessage();
                log.error("Request: Seal originated error", e);
                stsResponse = timedCreateAndTimeErrorResponse(
                        null,
                        StsFaultCodeValues.WST_REQUEST_FAILED ,
                        StsFaultStringValues.WST_REQUEST_FAILED + ": " + e.getMessage(),
                        StsFaultActorValues.SEAL);
            }

            timedSerializeToOutputStream(stsResponse, resp);

			boolean isFault = isFault(stsResponse);
			if (!isFault) { stsStatus.issued(); }
			
            slaLogItem.setReplySize(audit(timedSerialize(stsResponse)));
			return internalError;
		} catch (RuntimeException | IOException t) {
        	slaErrorMsg = t.getMessage();
            auditLog.error("ID-Card Response: Unhandled error in STS");
            log.error("ID-Card Response: Unhandled error in STS", t);

            throw new AxisFault(StsFaultCodeValues.WST_REQUEST_FAILED,
                                StsFaultStringValues.WST_REQUEST_FAILED,
                                StsFaultActorValues.STS, null);
        } finally {
            int duration = (int)(System.currentTimeMillis() - start);
	        boolean ok = slaErrorMsg == null;
	        stsStatus.addTransaction(duration, ok);
	        if (ok) {
            	SLALogHelper.storeCallResultOk(SLALogPoint.STSFACADE_REQUEST);
            } else {
            	SLALogHelper.storeCallResultError(SLALogPoint.STSFACADE_REQUEST, slaErrorMsg);
            }
            LogHelper.endRequest();
        }
    }

    /**
     * Check that the STS certificate is valid, not expired, and not revoked.
     * This mathod may be expanded by sub-classes
     * @throws KeyStoreException
     */
    protected final void verifyCertificates() {
        X509Certificate stsCert = getSosiFactory().getCredentialVault()
                                            .getSystemCredentialPair().getCertificate();
        long currentTime = System.currentTimeMillis();
        if (currentTime < stsCert.getNotBefore().getTime()) {
            throw new IllegalStateException("STS certificate not valid yet");
        } else if (currentTime > stsCert.getNotAfter().getTime()) {
            throw new IllegalStateException("STS certificate expired");
        } else if (!crlService.getRevocationStatus(stsCert).isValid()) {
            throw new IllegalStateException("STS certificate revoked");
        } else {
            log.debug("STS certificate verified");
        }
    }

	protected abstract ConfigType verifyConfiguration(RequestType stRequest) throws VerifyException;

    private int audit(String requestString) {
    	if (auditLog.isInfoEnabled()) {
    		String loggedRequest = getEnvelopeForAuditLog(requestString);
    		auditLog.info(auditName() + " Request:\n" + loggedRequest);
    	}

    	return requestString.length();
    }

    private int audit(Document responseDoc) {
    	String responseString = XmlUtil.node2String(responseDoc, false, true);

    	if (auditLog.isInfoEnabled()) {
    		String loggedResponse = getEnvelopeForAuditLog(responseString);
    		auditLog.info(auditName()+" Response:\n" + loggedResponse);
    	}

    	return responseString.length();
    }

    private String getEnvelopeForAuditLog(String envelope) {
    	String loggedEnvelope = envelope;

		try {
			loggedEnvelope = envelopeCompressor.compress(envelope);
		}
		catch(Exception e) {
			log.warn("Failed to compress log message for audit logging", e);
		}

    	return loggedEnvelope;
	}

	protected abstract String auditName();

	// This times the serialize calls, but should be replaced by and non-inheriting "framework"

	private RequestType timedDeserializeRequest(String doc) {
		Timer.start(this.getClass(),
					"deserializeRequest");
		try {
			return deserializeRequest(doc);
		} finally {
			Timer.stop(this.getClass(),
					   "deserializeRequest");
		}
	}

	private Document timedSerialize(ResponseType doc) {
		Timer.start(this.getClass(),
					"serialize");
		try {
			return serialize(doc);
		} finally {
			Timer.stop(this.getClass(),
					   "serialize");
		}
	}

	private void timedSerializeToOutputStream(ResponseType doc, OutputStream outputStream) throws IOException {
		Timer.start(this.getClass(),
					"serializeToOutputStream");
		try {
			serializeToOutputStream(doc, outputStream);
		} finally {
			Timer.stop(this.getClass(),
					   "serializeToOutputStream");
		}
	}

	private ResponseType timedCreateAndTimeErrorResponse(String dgwsVersion, String msgId, String faultCode, String faultString, String faultActor) {
		Timer.start(this.getClass(),
					"createErrorResponse");
		try {
			return createErrorResponse(dgwsVersion, msgId, faultCode, faultString, faultActor);
		} finally {
			Timer.stop(this.getClass(),
					   "createErrorResponse");
		}
	}

	private ResponseType timedCreateAndTimeErrorResponse(RequestType stRequest, String faultCode, String faultString, String faultActor) {
		Timer.start(this.getClass(),
					"createErrorResponse");
		try {
			return createErrorResponse(stRequest, faultCode, faultString, faultActor);
		} finally {
			Timer.stop(this.getClass(),
					   "createErrorResponse");
		}
	}

}
