/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.OIOBSTSAMLAssertion;
import dk.sosi.sts.model.IdpConfig;
import dk.sosi.sts.util.CertificateUtil;

import java.security.cert.X509Certificate;
import java.util.EnumSet;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

public class BST2SOSIConfig {

	private final IdpConfig config;
	private final OIOBSTSAMLAssertion assertion;
	private final List<String> whitelistedCertificates;
	private final List<X509Certificate> signingKeys;

	public BST2SOSIConfig(IdpConfig idpConfiguration, OIOBSTSAMLAssertion oiosamlAssertion, List<String> whitelistedCertificates) {
		this.whitelistedCertificates = whitelistedCertificates;
		if(idpConfiguration == null) {
			throw new ModelException("IdpConfig cannot be null in BST2SOSIConfig");
		} else if(oiosamlAssertion == null) {
			throw new ModelException("OIOBSTSAMLAssertion cannot be null in BST2SOSIConfig");
		}
		this.config = idpConfiguration;
		this.assertion = oiosamlAssertion;
		signingKeys = parseSigningKeys(config);
	}

	private List<X509Certificate> parseSigningKeys(IdpConfig config) {
		try {
			return config.getSigningKeys().stream()
					.filter(Objects::nonNull)
					.map(CertificateUtil::parseCertificateConfiguration)
					.collect(Collectors.toList());
		} catch (Exception e) {
			throw new ModelException("Invalid signingKey in BST2SOSIConfig for issuer " + config.getIssuer(), e);
		}
	}

	public OIOBSTSAMLAssertion getAssertion() {
		return assertion;
	}
	
	protected IdpConfig getConfig() {
		return config;
	}
	
	public String getIssuer() {
		return getConfig().getIssuer();
	}
	
	public EnumSet<OIOBSTSAMLAssertion.Type> getRequiredTypes() {
		List<String> tokenProfiles = getConfig().getTokenProfiles();
		EnumSet<OIOBSTSAMLAssertion.Type> assertionTypes = EnumSet.noneOf(OIOBSTSAMLAssertion.Type.class);
		for (String tokenProfile : tokenProfiles) {
			assertionTypes.add(OIOBSTSAMLAssertion.Type.valueOf(tokenProfile));
		}
		return assertionTypes;
	}
	
	public boolean getValidateHolderOfKey() {
		String validateHOK = getConfig().getValidateHOK();
		return Boolean.parseBoolean(validateHOK);
	}
	
	public List<X509Certificate> getSigningKeys() {
		return signingKeys;
	}

	public List<String> getWhitelistedCertificates() {
		return whitelistedCertificates;
	}
}
