/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.AssuranceLevel.NIST;
import dk.sosi.seal.model.AssuranceLevel.NSIS;
import dk.sosi.seal.model.OIOBSTSAMLAssertion.Type;
import dk.sosi.seal.model.constants.*;
import dk.sosi.seal.model.dombuilders.OIOBSTSAMLAssertionToIDCardResponseDOMBuilder;
import dk.sosi.seal.model.dombuilders.OIOSAMLAssertionToIDCardResponseDOMBuilder;
import dk.sosi.seal.modelbuilders.OIOBSTSAMLAssertionToIDCardRequestModelBuilder;
import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.vault.CredentialPair;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultException;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.Oces2InfoParser;
import dk.sosi.sts.certificate.Oces3InfoParser;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfo.CertificateType;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.its.DocumentResponse;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.IdpConfig;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.model.UserData;
import dk.sosi.sts.server.service.Timed;
import dk.sosi.sts.service.IdpConfigService;
import dk.sosi.sts.service.UserDataService;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import dk.sosi.sts.sla.StsContext;
import org.springframework.util.StringUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.io.IOException;
import java.io.OutputStream;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.*;

/**
 * A request handler for exchanging an OIO Saml bootstrap token with a SOSI id card.
 * This is for employees only.
 * Valid tokens are:
 * <ul>
 *  <li>OIOH2BST</li>
 *  <li>OIOH3BST</li>
 *  <li>OIO3BST</li>
 * </ul>
 * Refer to <a href="https://www.nspop.dk/display/public/web/STS+-+Guide+til+anvendere%3A+Medarbejderomvekslinger">STS Medarbejderomvekslinger</a> for details.
 */
public class BST2SOSIRequestHandler extends AbstractStsRequestHandler<OIOBSTSAMLAssertionToIDCardRequest, DocumentResponse, BST2SOSIConfig> {
	private final String UUID_PREFIX = "urn:uuid:";
	private final List<String> VALID_CLAIM_TYPES = Arrays.asList(MedcomAttributes.IT_SYSTEM_NAME, MedcomAttributes.USER_ROLE, MedcomAttributes.USER_AUTHORIZATION_CODE, SOSIAttributes.SUBJECT_NAME_ID);
	private final OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();
	private String blacklistedITSystemName;
	protected UserValidationService userValidationService;
	protected UserDataService userDataService;
	protected IdpConfigService configService;
	protected long allowedDriftInSeconds;
	protected String allowedAudience;
	protected long idCardDuration;
	protected long fuzzyTime;
	protected boolean whitelistValidation = true;

	@Override
	protected OIOBSTSAMLAssertionToIDCardRequest deserializeRequest(String doc) {
		Document document = XmlUtil.readXml(new Properties(), doc, false);
		OIOBSTSAMLAssertionToIDCardRequestModelBuilder modelBuilder = oiosamlFactory.createOIOBSTSAMLAssertionToIDCardRequestModelBuilder();
		return modelBuilder.build(document);
	}

	public void setBlacklistedITSystemName(String blacklistedITSystemName) {
		this.blacklistedITSystemName = blacklistedITSystemName;
	}

	public void setUserValidationService(UserValidationService userValidationService) {
		this.userValidationService = userValidationService;
	}

	public void setIdpConfigService(IdpConfigService configService) {
		this.configService = configService;
	}

	public void setAllowedDriftInSeconds(String s) {
		allowedDriftInSeconds = Integer.parseInt(s);
		if (allowedDriftInSeconds < 0) {
			throw new IllegalArgumentException("'allowedDriftInSeconds' must not be negative!");
		}
	}

	public long getAllowedDriftInSeconds() {
		return allowedDriftInSeconds;
	}

	public void setAllowedAudience(String allowedAudience) {
		this.allowedAudience = allowedAudience;
	}

	public String getAllowedAudience() {
		return allowedAudience;
	}
	
	public void setUserDataService(UserDataService userDataService) {
		this.userDataService = userDataService;
	}

	public void setWhitelistValidation(boolean whitelistValidation) {
		this.whitelistValidation = whitelistValidation;
	}

	@Override
	protected BST2SOSIConfig verifyConfiguration(OIOBSTSAMLAssertionToIDCardRequest request) throws VerifyException {
		OIOBSTSAMLAssertion oiosamlAssertion = getAssertion(request);
		IdpConfig idpConfiguration = configService.getIdpConfiguration(oiosamlAssertion.getIssuer());
		if (idpConfiguration == null) {
			throw new ModelException("Could not find IdpConfiguration for issuer:" + oiosamlAssertion.getIssuer());
		}
		List<String> whitelistedCertificates = getWhitelistedCertificates(idpConfiguration);

		return new BST2SOSIConfig(idpConfiguration, oiosamlAssertion, whitelistedCertificates);
	}

	private List<String> getWhitelistedCertificates(IdpConfig idpConfiguration) {
		List<String> whitelistedCertificates = idpConfiguration.getCertificates();
		// Whitelisted certificates might be configured with '*' as issuer, meaning that the certificate is whitelisted for all issuers.
		IdpConfig allIssuerConfiguration = configService.getIdpConfiguration("*");
		if (allIssuerConfiguration != null) {
			whitelistedCertificates = new ArrayList<>(whitelistedCertificates);
			whitelistedCertificates.addAll(allIssuerConfiguration.getCertificates());
		}
		return whitelistedCertificates;
	}

	private OIOBSTSAMLAssertion getAssertion(OIOBSTSAMLAssertionToIDCardRequest request) {
		OIOBSTSAMLAssertion oiosamlAssertion;
		if (request.isAssertionEncrypted()) {
			oiosamlAssertion = decryptOioSamlAssertion(request);
		} else {
			oiosamlAssertion = request.getOIOBSTSAMLAssertion();
		}
		if (oiosamlAssertion == null) {
			throw new ModelException("No assertion element found on request.");
		}
		return oiosamlAssertion;
	}

	private OIOBSTSAMLAssertion decryptOioSamlAssertion(OIOBSTSAMLAssertionToIDCardRequest request) {
		OIOBSTSAMLAssertion oiosamlAssertion = null;
		List<String> allEncryptionKeys = configService.getAllEncryptionKeys();
		for (String encryptionKey : allEncryptionKeys) {
			PrivateKey privateKey = EncryptionUtil.generatePrivateKeyFromString(encryptionKey, "RSA");
			if (privateKey != null) {
				try {
					oiosamlAssertion = request.getOIOBSTSAMLAssertion(privateKey);
				} catch (ModelException e) {
					log.debug("Could not get OIOSAMLAssertion with private key", e);
				}
			} else {
				log.warn("Unable to generate Private Key from IdpConfigService");
			}
		}
		if (oiosamlAssertion == null) {
			throw new ModelException("Could not decrypt Assertion with any of the encryption keys in IdpConfigService.");
		}
		return oiosamlAssertion;
	}

	@Override
	protected DocumentResponse perform(OIOBSTSAMLAssertionToIDCardRequest req, BST2SOSIConfig config) throws VerifyException {
		SLALogHelper.createItem(SLALogPoint.BST2SOSI_SERIALIZE);
		StsContext.getContext().setMessageId(req.getMessageID());
		try {
			try {
				validateEnvelope(req, config);
				// Assertion has been extracted and decrypted (if encrypted) from request in verifyConfiguration
				OIOBSTSAMLAssertion assertion = config.getAssertion();
				validateAssertion(assertion, config);
				if (config.getValidateHolderOfKey()) {
					req.validateHolderOfKeyRelation();
				}
				validateClaims(req);

				UserInfo userInfo = buildUserInfo(assertion, req);
				SystemInfo systemInfo = buildSystemInfo(assertion, req);
				IDCard signedIdCard = buildSignedIdCard(req, userInfo, systemInfo);

				OIOBSTSAMLAssertionToIDCardResponseDOMBuilder response = oiosamlFactory.createOIOBSTSAMLAssertionToIDCardResponseDOMBuilder();
				response.setAudienceRestriction(req.getAppliesTo());
				response.setIDCard(signedIdCard);
				response.setContext(req.getContext());
				response.setRelatesTo(req.getMessageID());
				response.setSigningVault(sosiFactory.getCredentialVault());

				Document documentResponse = response.build();
				SLALogHelper.storeCallResultOk(SLALogPoint.BST2SOSI_SERIALIZE);
				return new DocumentResponse(false, documentResponse);
			} catch (ModelException me) {
				throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, StsFaultStringValues.WST_INVALID_REQUEST + ": " + me.getMessage(), StsFaultActorValues.BST2SOSI);
			}
		} catch (VerifyException v) {
			SLALogHelper.storeCallResultError(SLALogPoint.BST2SOSI_SERIALIZE, v.getFaultString());
			throw v;
		}
	}

	protected void validateEnvelope(final OIOBSTSAMLAssertionToIDCardRequest request, BST2SOSIConfig config) throws VerifyException {
		validateITSystemName(request.getITSystemName());
		// Validate the signature and check certificate
		new OIOWSTrustValidation(StsFaultActorValues.BST2SOSI).validateSignature(request);
		// check for issued by right OCES CA
		new CertificateStatusVerifier(sosiFactory.getFederation()).verifyCertificateStatus(request.getSigningCertificate(), StsFaultActorValues.BST2SOSI);

		if (whitelistValidation) {
			verifyConsumerWhitelisted(request, config);
		}
	}

	private void validateITSystemName(String itSystemName) throws VerifyException {
		if (blacklistedITSystemName.equals(itSystemName)) {
			throw new VerifyException(StsFaultCodeValues.WST_BAD_REQUEST,
					StsFaultStringValues.WST_BAD_REQUEST + ": Invalid ITSystemName",
					StsFaultActorValues.BST2SOSI);
		}
	}

	private void verifyConsumerWhitelisted(final OIOBSTSAMLAssertionToIDCardRequest request, BST2SOSIConfig config) throws VerifyException {
		final String certificateDn = request.getSigningCertificate().getSubjectDN().getName();
		if (!config.getWhitelistedCertificates().contains(new DistinguishedName(certificateDn).getSubjectSerialNumber())) {
			final String faultString = "Signing certificate not whitelisted for this issuer (cert={"+certificateDn+"}, issuer={"+config.getIssuer()+"})";
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, faultString, StsFaultActorValues.BST2SOSI);
		}
	}

	private IDCard buildSignedIdCard(OIOBSTSAMLAssertionToIDCardRequest request, UserInfo userInfo, SystemInfo systemInfo) {
		Date createdDate = new Date(System.currentTimeMillis() - getFuzzyTime());
		Date expiryDate = new Date(createdDate.getTime() + getIdCardDuration());
		Element document = null;// document == null => needs signature
		String cardId = XmlUtil.createGUID();
		// SDS-6483: If the SubjectNameId is set, use that as alternativeIdentifier, otherwise set it to null.
		String alternativeIdentifier = request.getSubjectNameId() != null ? request.getSubjectNameId() : null;
		return sign(new UserIDCard(DGWSConstants.VERSION_1_0_1, document, cardId, AuthenticationLevel.MOCES_TRUSTED_USER, null, sosiFactory.getIssuer(), systemInfo, userInfo, createdDate, expiryDate, alternativeIdentifier, null, null));

	}
	
	/**
	 * Sign User Id Card
	 */
	private UserIDCard sign(final UserIDCard idCard) {
		Request tmp = new Request(DGWSConstants.VERSION_1_0_1, false, "", sosiFactory);
		tmp.setIDCard(idCard);
		Document doc = tmp.serialize2DOMDocument();
		idCard.sign(doc, SignatureProviderFactory.fromCredentialVault(sosiFactory.getCredentialVault()));
		return (UserIDCard) tmp.getIDCard();
	}

	public void setIdCardDuration(long idCardDuration) {
		this.idCardDuration = idCardDuration;
	}

	private long getIdCardDuration() {
		return idCardDuration;
	}
	
	
	public void setFuzzyTime(long fuzzyTime) {
		this.fuzzyTime = fuzzyTime;
	}

	private long getFuzzyTime() {
		return fuzzyTime;
	}

	private SystemInfo buildSystemInfo(OIOBSTSAMLAssertion assertion, OIOBSTSAMLAssertionToIDCardRequest req) throws VerifyException {
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, assertion.getCvrNumberIdentifier(), assertion.getOrganizationName());
		return new SystemInfo(careProvider, req.getITSystemName());
	}

	/**
	 * Følgende skal valideres for et bootstraptoken:<br>
	 * <ul>
	 * <li>Token signatur</li>
	 * <li>Om token er tidsmæssigt gyldigt</li>
	 * <li>Om tokenet er i overensstemmelse med den konfigurerede tokenprofil
	 * (OIOH3BST, OIOH2BST eller OIO3BST)</li>
	 * <li>Om standard til angivelse af sikringsniveau i token er høj nok. </li>
	 * <li>Validering af om audience er sat til SOSI-STS’en.</li>
	 * <li>Om basic privilege rolle strukturen er gyldig, hvis til stede.</li>
	 * </ul>
	 */
	private void validateAssertion(OIOBSTSAMLAssertion assertion, BST2SOSIConfig config) throws VerifyException {
		CredentialVault credentialVault = createCredentialVaultFromConfig(config);
		assertion.validateSignatureAndTrust(credentialVault);

		checkCertificate(assertion.getSigningCertificate(), "BST2SOSI Certificate", StsFaultActorValues.BST2SOSI);

		assertion.validateTimestamp(getAllowedDriftInSeconds());

		if (!config.getRequiredTypes().contains(assertion.getType())) {
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, "BST2SOSI assertion is not allowed type. Actual type:" + assertion.getType() + ", Allowed types:" + config.getRequiredTypes(), StsFaultActorValues.BST2SOSI);
		}

		validateAssuranceLevel(assertion.getType(), assertion.getAssuranceLevel());

		String audienceRestriction = assertion.getAudienceRestriction();
		if (!getAllowedAudience().equals(audienceRestriction)) {
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, "BST2SOSI assertion does not have allowed audience. Actual type:" + audienceRestriction + ", Allowed type:" + getAllowedAudience(), StsFaultActorValues.BST2SOSI);
		}

		try {
			assertion.getBasicPrivileges(); // Fails if mandatory and cannot be decoded
		} catch (Exception e) {
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, "BST2SOSI assertion contains an invalid basic privilege profile: " + e.getMessage(), StsFaultActorValues.BST2SOSI);
		}

	}

	/**
	 * For OIOH2BST bootstraptokens, a NIST assurance level of 3 or higher is required.
	 * For OIOH3BST bootstraptokens, a NSIS assurance level of ’Substantial’ or ’High’ is required.
	 * For OIO3BST bootstraptokens, any of the above is accepted.
	 */
	private void validateAssuranceLevel(final Type assertionType, final String assuranceLevel) throws VerifyException {
		AssuranceLevel assuranceLevelObject = AssuranceLevel.ofValue(assuranceLevel);
		switch (assertionType) {
			case OIOH2BST:
				validateAssuranceLevelType(assuranceLevelObject, AssuranceLevel.Type.NIST);
				validateAssuranceLevel((NIST) assuranceLevelObject);
				break;
			case OIOH3BST:
				validateAssuranceLevelType(assuranceLevelObject, AssuranceLevel.Type.NSIS);
				validateAssuranceLevel((NSIS) assuranceLevelObject);
				break;
			case OIO3BST:
				if (assuranceLevelObject.getType() == AssuranceLevel.Type.NIST) {
					validateAssuranceLevel((NIST) assuranceLevelObject);
				} else {
					validateAssuranceLevel((NSIS) assuranceLevelObject);
				}
				break;
		}
	}

	private void validateAssuranceLevelType(AssuranceLevel assuranceLevel, AssuranceLevel.Type expectedType) throws VerifyException {
		if (assuranceLevel.getType() != expectedType) {
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, "BST2SOSI assertion do not have a valid type of AssuranceLevel. Type:" + assuranceLevel.getType(), StsFaultActorValues.BST2SOSI);
		}
	}

	private void validateAssuranceLevel(final NIST nist) throws VerifyException {
		if (nist != NIST.NIVEAU_3 && nist != NIST.NIVEAU_4) {
			String faultString = String.format("BST2SOSI assertion do not have a valid AssuranceLevel %s. AssuranceLevel %s or %s required.", nist, NIST.NIVEAU_3, NIST.NIVEAU_4);
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, faultString, StsFaultActorValues.BST2SOSI);
		}
	}

	private void validateAssuranceLevel(final NSIS nsis) throws VerifyException {
		if (nsis != NSIS.Substantial && nsis != NSIS.High) {
			String faultString = String.format("BST2SOSI assertion do not have a valid AssuranceLevel %s. AssuranceLevel %s or %s required.", nsis, NSIS.Substantial, NSIS.High);
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, faultString, StsFaultActorValues.BST2SOSI);
		}
	}

	private UserInfo buildUserInfo(final OIOBSTSAMLAssertion assertion, final OIOBSTSAMLAssertionToIDCardRequest request) throws VerifyException {
		OcesInfo ocesInfo = buildOcesInfo(assertion);
		final CprNr inputCpr = getInputCpr(assertion.getCpr());
		final String inputRole = request.getUserRole();
		final String inputAuthorizationCode = request.getUserAuthorizationCode();
		final boolean cprTrustedForSigner = true; // we trust the cpr-number - at this point it has been checked that the BST was signed by a trusted party, so that is not checked again.
		VerifyInput verifyInput = new VerifyInput(ocesInfo, inputCpr, inputRole, inputAuthorizationCode);
		final String endpoint = "BST2SOSI"; // WARNING: this name may be used to check if we trust roles - beware, if changing
		VerifyResult verifyResult = userValidationService.verifyCprAndRole(verifyInput, cprTrustedForSigner, endpoint, assertion.getBasicPrivileges(), true);

		List<UserData> userDataList = userDataService.getUserInfo(verifyResult.cpr);
		if(userDataList == null || userDataList.isEmpty()) {
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, "Could not find userData for input CPR.", StsFaultActorValues.BST2SOSI);			
		} else if(userDataList.size() > 1){
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, "More than one set of UserData for input CPR. Unable to decide", StsFaultActorValues.BST2SOSI);						
		}
		
		UserData userData = userDataList.get(0);
		
		return new UserInfo(verifyResult.cpr.getValue(), userData.getGivenName(), userData.getSurName(), null, null, verifyResult.role, verifyResult.authorizationCode);
	}

	private OcesInfo buildOcesInfo(OIOBSTSAMLAssertion assertion) throws VerifyException {
		X509Certificate certificate = assertion.getSigningCertificate(); // Used to choose which cpr service to use

		// OIOH2BST may contain either RID or UUID as identification.
		// OIOH3BST and OIO3BST must contain UUID as identification.
		if (assertion.getType() == Type.OIOH2BST && assertion.getRidNumberIdentifier() != null) {
			return buildOcesInfoFromRid(assertion, certificate);
		} else if (assertion.getUuid() != null) {
			return buildOcesInfoFromUuid(assertion, certificate);
		} else if (assertion.getType() == Type.OIOH2BST) {
			/* SDS-6660 kan denne fjernes? */
			throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
					String.format("BST2SOSI assertion must contain one of the attributes: '%s' or '%s'", OIOSAMLAttributes.RID_NUMBER, OIOSAML3Attributes.PROFESSIONAL_UUID),
					StsFaultActorValues.BST2SOSI);
		} else {
			throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
					String.format("BST2SOSI assertion must contain the attribute '%s'", OIOSAML3Attributes.PROFESSIONAL_UUID), StsFaultActorValues.BST2SOSI);
		}
	}

	private OcesInfo buildOcesInfoFromUuid(OIOBSTSAMLAssertion assertion, X509Certificate certificate) throws VerifyException {
        String uuid = assertion.getUuid();
        if(uuid.startsWith(UUID_PREFIX)) {
            uuid = uuid.substring(UUID_PREFIX.length());
        }
        if(!isValidUuid(uuid)) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "BST2SOSI assertion do not have a valid uuid: " + assertion.getUuid(), StsFaultActorValues.BST2SOSI);
        }

        String cvr = assertion.getCvrNumberIdentifier();
		SubjectSerialNumber subjectSerialNumber = Oces3InfoParser.subjectSerialNumberFromGlobalEmployeeUuid(uuid);
		return new OcesInfo(subjectSerialNumber, cvr, CertificateType.MOCES, Version.OCES3, certificate);
	}

    private boolean isValidUuid(String uuid) {
        try {
            UUID.fromString(uuid);
        }
        catch(IllegalArgumentException e) {
            return false;
        }
        return true;
    }

	private CprNr getInputCpr(String assertionCpr) {
		final CprNr inputCpr;
		if (assertionCpr == null || "".equals(assertionCpr)) {
			inputCpr = null;
		} else {
			inputCpr = new CprNr(assertionCpr);
		}
		return inputCpr;
	}

	private OcesInfo buildOcesInfoFromRid(OIOBSTSAMLAssertion assertion, X509Certificate certificate) {
		String rid = assertion.getRidNumberIdentifier();
		String cvr = assertion.getCvrNumberIdentifier();
		SubjectSerialNumber subjectSerialNumber = Oces2InfoParser.subjectSerialNumberFromCvrAndRid(cvr, rid);
		return new OcesInfo(subjectSerialNumber, cvr, CertificateType.MOCES, Version.OCES2, certificate);
	}

	/**
	 * Validate Claims from Request
	 */
	private void validateClaims(OIOBSTSAMLAssertionToIDCardRequest req) {
		if (!VALID_CLAIM_TYPES.containsAll(req.getClaimMap().keySet())) {
			throw new ModelException("Request contains invalid claims. Valid claims are: " + VALID_CLAIM_TYPES);
		}

		/* Claims */
		String itSystemName = req.getITSystemName();
		if (StringUtils.isEmpty(itSystemName)) {
			throw new ModelException("ITSystemName is required in request.");
		}
		String userAuthorizationCode = req.getUserAuthorizationCode();
		String userRole = req.getUserRole();
		if (userAuthorizationCode != null && userRole != null) {
			throw new ModelException("Only one of UserAuthorizationCode or UserRole is allowed in request.");
		}
	}

	@Override
	protected boolean isFault(final DocumentResponse stResponse) {
		return stResponse.isFault();
	}

	@Override
	protected Document serialize(final DocumentResponse response) {
		return response.getDocument();
	}

	@Override
	@Timed
	protected void serializeToOutputStream(final DocumentResponse doc, final OutputStream outputStream) throws IOException {
		outputStream.write(XmlUtil.serializeXml2ByteArray(serialize(doc), true));
	}

	@Override
	protected DocumentResponse createErrorResponse(final String dgwsVersion, final String msgId, final String faultCode, final String faultString, final String faultActor) {
		return createErrorResponse(null, faultCode, faultString, faultActor);
	}

	@Timed
	@Override
	protected DocumentResponse createErrorResponse(final OIOBSTSAMLAssertionToIDCardRequest stRequest, final String faultCode, final String faultString, final String faultActor) {
		OIOSAMLAssertionToIDCardResponseDOMBuilder response = oiosamlFactory.createOIOSAMLAssertionToIDCardResponseDOMBuilder();

		response.setSigningVault(sosiFactory.getCredentialVault());
		response.setFaultCode(faultCode);
		response.setFaultActor(faultActor);
		response.setFaultString(faultString);
		response.setRelatesTo(stRequest == null ? "<unknown>" : stRequest.getMessageID());

		Document output = response.build();
		return new DocumentResponse(true, output);
	}

	@Override
	protected String auditName() {
		return "BST2SOSI";
	}

	/**
	 * Create a credential vault for validating a certificate against the trusted signing keys in the given configuration
	 */
	private CredentialVault createCredentialVaultFromConfig(BST2SOSIConfig config) {
		return new CredentialVault() {
			@Override
			public boolean isTrustedCertificate(X509Certificate certificate) throws CredentialVaultException {
				return config.getSigningKeys().stream().anyMatch(certificate::equals);
			}

			@Override
			public CredentialPair getSystemCredentialPair() throws CredentialVaultException {
				return null;
			}

			@Override
			public void setSystemCredentialPair(CredentialPair credentialPair) throws CredentialVaultException {
			}

			@Override
			public KeyStore getKeyStore() {
				return null;
			}
		};
	}

}
