/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.HealthcareSAMLAttributes;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;
import dk.sosi.seal.model.dombuilders.AbstractOIOToIdentityTokenResponseDOMBuilder;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenResponseDOMBuilder;
import dk.sosi.seal.model.dombuilders.OIOSAMLAssertionToIDCardResponseDOMBuilder;
import dk.sosi.seal.vault.CredentialPair;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultException;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.StsException;
import dk.sosi.sts.ibo.IboConventions;
import dk.sosi.sts.integrations.blurring.BlurringService;
import dk.sosi.sts.its.DocumentResponse;
import dk.sosi.sts.model.AudienceConfiguration;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.IdpConfig;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.service.IdpConfigService;
import dk.sosi.sts.service.OcesPidService;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import dk.sosi.sts.sla.StsContext;
import org.w3c.dom.Document;

import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.*;

import static java.util.Optional.*;

/**
 * A request handler for exchanging an OIO Saml bootstrap token with an OIO IDWS identity token.
 * This is for citizens only.
 * Valid tokens are:
 * <ul>
 *  <li>OIO2BST_LEGACY</li>
 *  <li>OIO2BST_CITIZEN</li>
 *  <li>OIO3BST_CITIZEN</li>
 * </ul>
 * Refer to <a href="https://www.nspop.dk/display/web/STS+-+Guide+til+anvendere%3A+Borgeromvekslinger">STS Borgeromvekslinger</a> for details.
 */
public class Bootstrap2IdwsRequestHandler extends AbstractIdwsRequestHandler<OIOBootstrapToIdentityTokenRequest, Bootstrap2IdwsConfig> {
    private final List<String> VALID_CLAIM_TYPES = Arrays.asList(OIOSAMLAttributes.CPR_NUMBER, HealthcareSAMLAttributes.ON_BEHALF_OF);
    protected NboConfig nboConfig;
    protected IdpConfigService idpConfigService;
    private OcesPidService ocesPidService;
    private BlurringService blurringService;

    // invoked by spring
    public void setOcesPidService(OcesPidService service) {
        this.ocesPidService = service;
    }
    // invoked by spring
    public void setBlurringService(BlurringService blurringService) {
        this.blurringService = blurringService;
    }

    public BlurringService getBlurringService() { return blurringService; }

    // the name used to identify the service in auditlog
    @Override
    protected final String auditName() {
        return "NBT2Idws";
    }

    @Override
    protected OIOBootstrapToIdentityTokenRequest deserializeRequest(String doc) {
        final Document requestDom = XmlUtil.readXml(new Properties(), doc, false);
        return new OIOSAMLFactory().createOIOBootstrapToIdentityTokenRequestModelBuilder().build(requestDom);
    }

    @Override
    protected Bootstrap2IdwsConfig verifyConfiguration(OIOBootstrapToIdentityTokenRequest stRequest) throws VerifyException {
        final String reqAudience = stRequest.getAppliesTo();
        if (reqAudience == null || reqAudience.isEmpty()) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Audience is missing", StsFaultActorValues.ITS);
        }
        // Clients using standard .net classes (URI) will erroneously append an extra slash in the end so this is removed
        String normalizedAudience = IboConventions.normalizeAudience(reqAudience);
        final AudienceConfiguration audienceConfiguration = itsConfigService.getAudienceConfiguration(normalizedAudience);
        if (audienceConfiguration == null) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Audience not configured: " + reqAudience, StsFaultActorValues.ITS);
        }

        OIOBSTSAMLAssertion assertion = getAssertion(stRequest);
        IdpConfig idpConfiguration = idpConfigService.getIdpConfiguration(assertion.getIssuer());

        return new Bootstrap2IdwsConfig(idpConfiguration, audienceConfiguration);
    }

    @Override
    protected DocumentResponse perform(OIOBootstrapToIdentityTokenRequest req, Bootstrap2IdwsConfig config) throws VerifyException {
        SLALogHelper.createItem(SLALogPoint.NBO_IDWS);
        StsContext.getContext().setMessageId(req.getMessageID());
        try {
            final DocumentResponse documentResponse = doPerform(req, config);
            SLALogHelper.storeCallResultOk(SLALogPoint.NBO_IDWS);
            return documentResponse;
        } catch (ModelException me) {
            final String faultString = StsFaultStringValues.WST_INVALID_REQUEST + ": " + me.getMessage();
            SLALogHelper.storeCallResultError(SLALogPoint.NBO_IDWS, faultString);
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, faultString, StsFaultActorValues.ITS);
        } catch (VerifyException v) {
            SLALogHelper.storeCallResultError(SLALogPoint.NBO_IDWS, v.getFaultString());
            throw v;
        }
    }

    @Override
    protected final DocumentResponse createErrorResponse(OIOBootstrapToIdentityTokenRequest stRequest, String faultCode, String faultString, String faultActor) {
        final String messageId = stRequest == null ? "<unknown>" : stRequest.getMessageID();
        return createErrorResponse(null, messageId, faultCode, faultString, faultActor);
    }

    @Override
    protected final DocumentResponse createErrorResponse(String dgwsVersion, String msgId, String faultCode, String faultString, String faultActor) {
        OIOSAMLAssertionToIDCardResponseDOMBuilder response = new OIOSAMLFactory().createOIOSAMLAssertionToIDCardResponseDOMBuilder();

        response.setSigningVault(sosiFactory.getCredentialVault());
        response.setFaultCode(faultCode);
        response.setFaultActor(faultActor);
        response.setFaultString(faultString);
        response.setRelatesTo(msgId == null ? "<unknown>" : msgId);

        Document output = response.build();
        return new DocumentResponse(true, output);
    }

    private DocumentResponse doPerform(OIOBootstrapToIdentityTokenRequest request, Bootstrap2IdwsConfig config) throws VerifyException {
        AudienceConfiguration audienceConfig = config.getAudienceConfig();
        final X509Certificate signingCertificate = verifyConsumer(request, audienceConfig);

        OIOBSTSAMLAssertion assertion = request.getOIOBSTSAMLAssertion(); // Assertion has been extracted and decrypted (if encrypted) in verifyConfiguration
        validateAssertion(assertion, config);
        // Legacy bootstrap token doesn't use holder-of-key validation
        if (assertion.getType() != OIOBSTSAMLAssertion.Type.OIO2BST_LEGACY) {
            request.validateHolderOfKeyRelation();
        }
        validateClaims(request);
        detectMissingWhitelistingForWardClaim(audienceConfig, signingCertificate, request.getWardCustodyClaim());

        final String cprNumberClaim = request.getCPRNumberClaim();
        validateCpr(assertion, cprNumberClaim);

        if(Objects.equals(request.getProcurationForCprClaim(), "")) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
                                      "OnBehalfOf claim must not be empty",
                                      StsFaultActorValues.ITS);
        }

        BasicPrivilegesDOMBuilder privilegesDOMBuilder = createPrivilegeBuilder(
                parseCpr(cprNumberClaim),
                request.getProcurationForCprClaim() != null
                        ? parseCpr(request.getProcurationForCprClaim())
                        : null);

        SubjectRelations11Impl.VerifiedRelation verifiedRelation = createValidatedSubjectRelations(
                request.getWardCustodyClaim(), request.getParentalCustodyClaim(), request.getPartlyWardCustodyClaim(), cprNumberClaim);
        SubjectRelationsDomBuilder subjectRelationsDomBuilder = createSubjectRelationsBuilder(verifiedRelation);

        BlurringInstructionsDomBuilder blurringInstructionBuilder =
                createBlurringInstructionsBuilder(blurringService,
                        parseCpr(cprNumberClaim),
                        request.getProcurationForCprClaim() != null
                                ? of(parseCpr(request.getProcurationForCprClaim()))
                                : empty(),
                        Optional.ofNullable(verifiedRelation),
                        request.getMessageID(),
                        audienceConfig.requireBlurring());

        IdentityToken identityToken = createIdentityToken(audienceConfig,
                                                          assertion,
                                                          signingCertificate,
                                                          cprNumberClaim,
                                                          privilegesDOMBuilder,
                                                          blurringInstructionBuilder,
                                                          subjectRelationsDomBuilder);

        AbstractOIOToIdentityTokenResponseDOMBuilder<?> responseBuilder =
                new OIOBootstrapToIdentityTokenResponseDOMBuilder();
        responseBuilder.setContext(request.getContext());
        responseBuilder.setIdentityToken(identityToken);
        responseBuilder.setRelatesTo(request.getMessageID());
        responseBuilder.setSigningVault(getSosiFactory().getCredentialVault(),
                                        request.isUsingSecurityTokenReference()
                                                ? SignatureConfiguration.KeyInfoType.WSS_REFERENCE
                                                : null);

        return new DocumentResponse(false, responseBuilder.build());
    }

    private OIOBSTSAMLAssertion getAssertion(OIOBootstrapToIdentityTokenRequest request) {
        OIOBSTSAMLAssertion oiosamlAssertion = request.isAssertionEncrypted() ? decryptAssertion(request) : request.getOIOBSTSAMLAssertion();
        if (oiosamlAssertion == null) {
            throw new ModelException("No assertion element found on request.");
        }
        return oiosamlAssertion;
    }

    private OIOBSTSAMLAssertion decryptAssertion(OIOBootstrapToIdentityTokenRequest request) {
        OIOBSTSAMLAssertion oiosamlAssertion = null;
        List<String> allEncryptionKeys = idpConfigService.getAllEncryptionKeys();
        for (String encryptionKey : allEncryptionKeys) {
            PrivateKey privateKey = EncryptionUtil.generatePrivateKeyFromString(encryptionKey, "RSA");
            if (privateKey != null) {
                try {
                    oiosamlAssertion = request.getOIOBSTSAMLAssertion(privateKey);
                } catch (ModelException e) {
                    log.debug("Could not get OIOSAMLAssertion with private key", e);
                }
            } else {
                log.warn("Unable to generate Private Key from IdpConfigService");
            }
        }
        if (oiosamlAssertion == null) {
            throw new ModelException("Could not decrypt Assertion with any of the encryption keys in IdpConfigService.");
        }
        return oiosamlAssertion;
    }

    /**
     * Check that the given bootstraptoken is trusted and valid
     */
    private void validateAssertion(OIOBSTSAMLAssertion assertion, Bootstrap2IdwsConfig config) throws VerifyException {
        // Assertion check trust
        CredentialVault credentialVault = createCredentialVaultFromConfig(config);
        assertion.validateSignatureAndTrust(credentialVault);
        checkCertificate(assertion.getSigningCertificate(), "Bootstrap token signer", StsFaultActorValues.ITS);

        if (!config.getRequiredTypes().contains(assertion.getType())) {
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, "BST2IDWS assertion is not allowed type. Actual type:" + assertion.getType() + ", Allowed types:" + config.getRequiredTypes(), StsFaultActorValues.ITS);
        }
        if (assertion.getAudienceRestriction() == null || !assertion.getAudienceRestriction().equals(config.getTokenAudience())) {
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, "BST2IDWS assertion has an invalid audience. Actual audience:" + assertion.getAudienceRestriction() + ", Allowed audience:" + config.getTokenAudience(), StsFaultActorValues.ITS);
        }
        validateAssuranceLevel(assertion.getType(), assertion.getAssuranceLevel());

        // Check time for assertion ... notbefore not present here
        final long allowedDriftInSeconds = nboConfig.getAllowedDriftInSeconds();
        if(allowedDriftInSeconds < 0L) {
            throw new IllegalArgumentException("\'allowedDriftInSeconds\' must not be negative!");
        } else {
            Date now = new Date();
            if(!(new Date(now.getTime() - allowedDriftInSeconds * 1000L)).before(assertion.getNotOnOrAfter())) {
                throw new ModelException("Bootstrap token no longer valid");
            }
        }
    }

    private void validateClaims(OIOBootstrapToIdentityTokenRequest request) {
        Map<String, String> claimMap = request.getClaimMap();
        if (!VALID_CLAIM_TYPES.containsAll(claimMap.keySet())) {
            throw new ModelException("Request contains invalid claims. Valid claims are: " + VALID_CLAIM_TYPES);
        }
    }

    /**
     * For OIO2BST bootstraptokens, a NIST assurance level of 3 or higher is required.
     * For OIO3BST bootstraptokens, a NSIS assurance level of ’Substantial’ or ’High’ is required.
     */
    private void validateAssuranceLevel(final OIOBSTSAMLAssertion.Type assertionType, final String assuranceLevel) throws VerifyException {
        if (assertionType == OIOBSTSAMLAssertion.Type.OIO2BST_LEGACY && assuranceLevel == null) {
            // Assurance level is optional for OIO2BST_LEGACY
            return;
        }
        AssuranceLevel assuranceLevelObject = AssuranceLevel.ofValue(assuranceLevel);
        switch (assertionType) {
            case OIO2BST_CITIZEN:
            case OIO2BST_LEGACY:
                validateAssuranceLevelType(assuranceLevelObject, AssuranceLevel.Type.NIST);
                validateAssuranceLevel((AssuranceLevel.NIST) assuranceLevelObject);
                break;
            case OIO3BST_CITIZEN:
                validateAssuranceLevelType(assuranceLevelObject, AssuranceLevel.Type.NSIS);
                validateAssuranceLevel((AssuranceLevel.NSIS) assuranceLevelObject);
                break;
        }
    }

    private void validateAssuranceLevelType(AssuranceLevel assuranceLevel, AssuranceLevel.Type expectedType) throws VerifyException {
        if (assuranceLevel.getType() != expectedType) {
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, "BST2IDWS assertion do not have a valid type of AssuranceLevel. Type:" + assuranceLevel.getType(), StsFaultActorValues.ITS);
        }
    }

    private void validateAssuranceLevel(final AssuranceLevel.NIST nist) throws VerifyException {
        if (nist == AssuranceLevel.NIST.NIVEAU_1 || nist == AssuranceLevel.NIST.NIVEAU_2) {
            String faultString = String.format("BST2IDWS assertion do not have a valid AssuranceLevel %s. AssuranceLevel %s or higher required.", nist, AssuranceLevel.NIST.NIVEAU_3);
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, faultString, StsFaultActorValues.ITS);
        }
    }

    private void validateAssuranceLevel(final AssuranceLevel.NSIS nsis) throws VerifyException {
        if (nsis != AssuranceLevel.NSIS.Substantial && nsis != AssuranceLevel.NSIS.High) {
            String faultString = String.format("BST2IDWS assertion do not have a valid AssuranceLevel %s. AssuranceLevel %s or %s required.", nsis, AssuranceLevel.NSIS.Substantial, AssuranceLevel.NSIS.High);
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED, faultString, StsFaultActorValues.ITS);
        }
    }

    /**
     * Validate the cpr claim from the request against the included bootstraptoken/assertion.
     * Legacy bootstraptokens contain PID and no cpr, while other tokens contain cpr and no PID.
     * The cpr claim in a request with a legacy bootstraptoken is therefore validated against the PID,
     * while the cpr claim is validated against the cpr from the given token in the other case.
     */
    private void validateCpr(OIOBSTSAMLAssertion assertion, String cprClaim) throws VerifyException {
        if (cprClaim == null || cprClaim.isEmpty()) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "CPR claim missing", StsFaultActorValues.ITS);
        }

        String subjectNameID = assertion.getSubjectNameID();
        if (assertion.getType() == OIOBSTSAMLAssertion.Type.OIO2BST_LEGACY) {
            SubjectSerialNumber pidSsn = VerifyUtil.verifyCitizenToken(subjectNameID);
            validateCprAgainstPid(pidSsn, cprClaim);
        } else if (subjectNameID.startsWith(OIOSAMLAttributes.CPR_NUMBER + ":")) {
            validateCprAgainstAssertionCprAndSubjectNameId(assertion.getCpr(), subjectNameID, cprClaim);
        } else {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Invalid Subject NameID: " + subjectNameID, StsFaultActorValues.ITS);
        }
    }

    private void validateCprAgainstAssertionCprAndSubjectNameId(String assertionCpr, String subjectNameId, String cprClaim) throws VerifyException {
        String subjectCpr = subjectNameId.substring(OIOSAMLAttributes.CPR_NUMBER.length()+1);
        if (!cprClaim.equals(assertionCpr) || !cprClaim.equals(subjectCpr)) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Cpr claim does not match cpr in assertion or Subject NameID.", StsFaultActorValues.ITS);
        }
    }

    private void validateCprAgainstPid(SubjectSerialNumber pidSsn, String cprClaim) throws VerifyException {
        final CprNr cprNr = parseCpr(cprClaim);
        boolean result;
        try {
            result = ocesPidService.isRelated(pidSsn, cprNr);
        } catch (StsException e) {
            log.error("Unable to validateScope cpr", e);
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validateScope cpr", StsFaultActorValues.ITS);
        }
        if (!result) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "CPR/PID mismatch", StsFaultActorValues.ITS);
        }
    }

    protected final IdentityToken createIdentityToken(final AudienceConfiguration config,
                                                      final OIOBSTSAMLAssertion assertion,
                                                      X509Certificate holderOfKeyCertificate,
                                                      String currentUserCpr,
                                                      BasicPrivilegesDOMBuilder privilegesDOMBuilder,
                                                      BlurringInstructionsDomBuilder blurringInstructionBuilder,
                                                      SubjectRelationsDomBuilder subjectRelationsDomBuilder) {
        final CitizenIdentityTokenBuilder identityTokenBuilder = new OIOSAMLFactory().createCitizenIdentityTokenBuilder();

        identityTokenBuilder.setSubjectNameID(assertion.getSubjectNameID());
        identityTokenBuilder.setSubjectNameIDFormat(assertion.getSubjectNameIDFormat());
        identityTokenBuilder.setCprNumberAttribute(currentUserCpr);
        identityTokenBuilder.setRecipientURL(config.getAudience());
        identityTokenBuilder.setAudienceRestriction(config.getAudience());
        identityTokenBuilder.setHolderOfKeyCertificate(holderOfKeyCertificate);
        identityTokenBuilder.setSigningVault(getSosiFactory().getCredentialVault());
        identityTokenBuilder.setPrivileges(privilegesDOMBuilder);
        identityTokenBuilder.setBlurringInstructions(blurringInstructionBuilder);
        identityTokenBuilder.setSubjectRelations(subjectRelationsDomBuilder);

        final long now = System.currentTimeMillis();
        // not before not present in input assertion
        final Date notBefore = new Date(now - nboConfig.getFuzzyTime());
        final Date notOnOrAfter = new Date(Math.min(now + config.getLifetime(), assertion.getNotOnOrAfter().getTime()));
        return identityTokenBuilder.setNotBefore(notBefore)
                .setNotOnOrAfter(notOnOrAfter)
                .setDeliveryNotOnOrAfter(notOnOrAfter)
                .setIssuer(getSosiFactory().getIssuer())
                .build();
    }

    /**
     * Create a credential vault for validating a certificate against the trusted signing keys in the given configuration
     */
    private CredentialVault createCredentialVaultFromConfig(Bootstrap2IdwsConfig config) {
        return new CredentialVault() {
            @Override
            public boolean isTrustedCertificate(X509Certificate certificate) throws CredentialVaultException {
                return config.getSigningKeys().stream().anyMatch(certificate::equals);
            }

            @Override
            public CredentialPair getSystemCredentialPair() throws CredentialVaultException {
                return null;
            }

            @Override
            public void setSystemCredentialPair(CredentialPair credentialPair) throws CredentialVaultException {
            }

            @Override
            public KeyStore getKeyStore() {
                return null;
            }
        };
    }


    // invoked from spring
    public void setNboConfig(NboConfig nboConfig) {
        this.nboConfig = nboConfig;
    }

    // invoked from spring
    public void setIdpConfigService(IdpConfigService idpConfigService) {
        this.idpConfigService = idpConfigService;
    }
}