/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.IDCardToOIOSAMLAssertionRequest;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.OIOSAMLAssertion;
import dk.sosi.seal.model.OIOSAMLAssertionBuilder;
import dk.sosi.seal.model.dombuilders.IDCardToOIOSAMLAssertionResponseDOMBuilder;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.ibo.IboConventions;
import dk.sosi.sts.its.DocumentResponse;
import dk.sosi.sts.model.IboConfig;
import dk.sosi.sts.service.IboConfigService;
import org.w3c.dom.Document;

import java.io.IOException;
import java.io.OutputStream;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.Properties;
import java.util.UUID;

/**
 */
public class IboRequestHandler extends AbstractStsRequestHandler<IDCardToOIOSAMLAssertionRequest, DocumentResponse, IboConfig>{

	public static final String URN_STS_PREFIX = "urn:sts:generated:";

	private final OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();
	private SOSIFactory sosiFactory;
	private IboConfigService iboConfigService;
	private SecurityTokenService sts;
	private boolean validateEnvelopeSignature = true;
	private boolean checkTargetCertificate = true;
	private String emptyAttributeValue = "NONE";

	@Override
	protected IDCardToOIOSAMLAssertionRequest deserializeRequest(final String doc) {
		return oiosamlFactory.createIDCardToOIOSAMLAssertionRequestModelBuilder()
							 .build(XmlUtil.readXml(new Properties(),
													doc,
													false));
	}

	@Override
	protected IboConfig verifyConfiguration(IDCardToOIOSAMLAssertionRequest req) throws VerifyException {
		// Clients using standard .net classes (URI) will erroneously append an extra slash in the end so this is removed
		final String audience = IboConventions.normalizeAudience(req.getAppliesTo());

		final IboConfig conf = iboConfigService.getAudienceConfiguration(audience);
		if (conf == null) {
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
				StsFaultStringValues.WST_REQUEST_FAILED
					+ ": No configuration found for audience, '" + req.getAppliesTo() + "'"
					+ (audience != null ? " (normalized to '" +audience+ "')" : ""),
				StsFaultActorValues.IBO);
		}
		if (checkTargetCertificate) {
			final X509Certificate targetCertificate = conf.getTargetCertificate();
			try {
				CertificateStatus status = sosiFactory.getFederation().getCertificateStatus(targetCertificate);
				if (!status.isValid() && status.getTimestamp() != null) {
					throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
											  StsFaultStringValues.WST_REQUEST_FAILED + ": Certificate revoked (" + status.getTimestamp() + ") for audience " + audience,
											  StsFaultActorValues.STS);
				} else if (!status.isValid()){
					throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
											  StsFaultStringValues.WST_REQUEST_FAILED + ": Certificate expired or not yet valid for audience " + audience,
											  StsFaultActorValues.STS);
				}
			} catch (PKIException e) {
				throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
										  StsFaultStringValues.WST_REQUEST_FAILED + ": Certificate invalid for audience " + audience + " (" + e.getMessage() +")",
										  StsFaultActorValues.STS);
			}
		}
		return conf;
	}


	@Override
	protected DocumentResponse perform(final IDCardToOIOSAMLAssertionRequest req, IboConfig conf) throws VerifyException {
		// To support SOSI-GW work flow, where the request cannot be signed after the ID-card
		// is supplied to the request.
		if (validateEnvelopeSignature) {
			validateEnvelope(req);
		}
		validateIDCard(req, conf);

		final OIOSAMLAssertionBuilder builder = oiosamlFactory.createOIOSAMLAssertionBuilder()
				.setDeliveryNotOnOrAfter(getDeliveryNotOnOrAfter(conf))
				.setNotBefore(getNotBefore(conf))
				.setNotOnOrAfter(getNotOnOrAfter(conf))
				.setRecipientURL(conf.getRecipientURL())
				.setAudienceRestriction(conf.getAudience())
				.setIssuer(getIssuer())
				.setSigningVault(sosiFactory.getCredentialVault())
				.setUserIdCard(req.getUserIDCard())
				.setEmptyAttributeValue(emptyAttributeValue);

		if (conf.includeBootStrapToken()) {
			builder.includeIDCardAsBootstrapToken();
		}
		final OIOSAMLAssertion assertion = builder.build();

		final IDCardToOIOSAMLAssertionResponseDOMBuilder responseBuilder = oiosamlFactory
				.createIDCardToOIOSAMLAssertionResponseDOMBuilder();

		responseBuilder.setSigningVault(sosiFactory.getCredentialVault());
		responseBuilder.setEncryptionKey(conf.getTargetCertificate().getPublicKey());
		responseBuilder.setRelatesTo(getMessageId(req));
		responseBuilder.setOIOSAMLAssertion(assertion);
		responseBuilder.setContext(req.getContext());

		return new DocumentResponse(false,
									responseBuilder.build());
	}

	private void validateIDCard(final IDCardToOIOSAMLAssertionRequest req, IboConfig config) throws VerifyException {
		// Verify the id-card wrt validity in time
		sts.verifyIdCard(req.getUserIDCard());

		float ageMillis = (float) System.currentTimeMillis() - req.getUserIDCard().getCreatedDate().getTime();
		float ageMins = (ageMillis / (60*1000) );

		long maxAgeMins = config.getIdCardMaxAgeMinutes();
		if (ageMins > maxAgeMins) {
			throw new VerifyException(StsFaultCodeValues.WST_INVALID_TIME_RANGE,
				StsFaultStringValues.WST_INVALID_TIME_RANGE +  ": IDCard is too old for this audience: " + config.getAudience() + " [" + ageMins +  ">" + maxAgeMins + "]",
				StsFaultActorValues.STS);
		}


		// Check STS signed ID-card
		try {
			req.getUserIDCard().validateSignatureAndTrust(sosiFactory.getFederation());
		} catch (ModelException me) {
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
									  StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": " + me.getMessage(),
									  StsFaultActorValues.IBO);
		}
	}

	private void validateEnvelope(final IDCardToOIOSAMLAssertionRequest req) throws VerifyException {
		try {
			req.validateSignature();
		} catch(ModelException me) {
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
									  StsFaultStringValues.WST_REQUEST_FAILED + ": " + me.getMessage(),
									  StsFaultActorValues.IBO);
		} catch(ModelBuildException me) {
			throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
									  StsFaultStringValues.WST_REQUEST_FAILED + ": " + me.getMessage(),
									  StsFaultActorValues.IBO);
		}
		// Exceptions/faults from these will be interchangeable with the one generated from the signing
		// service. The difference should however be clear from the service being called.
		sts.checkIssuer(req.getSigningCertificate(), StsFaultActorValues.STS); // check for issued by right OCES CA
	}

	private Date getDeliveryNotOnOrAfter(final IboConfig conf) {
		return new Date(System.currentTimeMillis() + conf.getDeliveryNotOnOrAfterOffset());
	}

	private Date getNotOnOrAfter(final IboConfig conf) {
		return new Date(System.currentTimeMillis() + conf.getNotOnOrAfterOffset());
	}

	private Date getNotBefore(final IboConfig conf) {
		return new Date(System.currentTimeMillis() + conf.getNotBeforeOffset());
	}

	private String getIssuer() {
		return sosiFactory.getIssuer();
	}

	@Override
	protected boolean isFault(final DocumentResponse response) {
		return response.isFault();
	}

	@Override
	protected Document serialize(final DocumentResponse response) {
		return response.getDocument();
	}

	@Override
	protected void serializeToOutputStream(final DocumentResponse doc, final OutputStream outputStream) throws IOException {
		outputStream.write(XmlUtil.serializeXml2ByteArray(serialize(doc),
														  true));
	}

	@Override
	protected DocumentResponse createErrorResponse(final String dgwsVersion, final String msgId, final String faultCode, final String faultString, final String faultActor) {
		final IDCardToOIOSAMLAssertionResponseDOMBuilder builder = oiosamlFactory
				.createIDCardToOIOSAMLAssertionResponseDOMBuilder();

		builder.setSigningVault(sosiFactory.getCredentialVault());
		builder.setFaultActor(faultActor);
		builder.setFaultString(faultString);
		builder.setFaultCode(faultCode);
		builder.setRelatesTo(msgId);

		return new DocumentResponse(true,
									builder.build());
	}

	@Override
	protected DocumentResponse createErrorResponse(final IDCardToOIOSAMLAssertionRequest stRequest, final String faultCode, final String faultString, final String faultActor) {
		return createErrorResponse("Sosi2OIOSaml Version", // ignored
										  getMessageId(stRequest),
										  faultCode,
										  faultString,
										  faultActor);
	}

	/**
	 * STS-20:
	 * This is a fix for the case where requests go threw SOSI-GW, which will erase all wsa-headers.
	 *
	 * IDCardToOIOSAMLAssertionRequest#getMessageID will throw a NPE in case the wsa:MessageID is
	 * not present in the request. This method will catch it and substitute a meaningful value.
	 *
	 * @param request the exchange request.
	 * @return the MessageID from the exchange request or a meaningful value.
	 */
	private String getMessageId(IDCardToOIOSAMLAssertionRequest request) {
        String messageID = request.getMessageID();
        return messageID != null ? messageID : URN_STS_PREFIX + UUID.randomUUID();
	}


	@Override
	protected String auditName() {
		return "IBO";
	}

	public void setIboConfigService(final IboConfigService iboConfigService) {
		this.iboConfigService = iboConfigService;
	}

	public SOSIFactory getSosiFactory() {
		return sosiFactory;
	}

	public void setSosiFactory(final SOSIFactory sosiFactory) {
		this.sosiFactory = sosiFactory;
	}

	public void setSts(final SecurityTokenService sts) {
		this.sts = sts;
	}

	public void setValidateEnvelopeSignature(final boolean validateEnvelopeSignature) {
		this.validateEnvelopeSignature = validateEnvelopeSignature;
	}

	public void setCheckTargetCertificate(final boolean checkTargetCertificate) {
		this.checkTargetCertificate = checkTargetCertificate;
	}

	public void setEmptyAttributeValue(final String emptyAttributeValue) {
		this.emptyAttributeValue = emptyAttributeValue;
	}

}
