/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.IDWSHFactory;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.IdentityTokenBuilder;
import dk.sosi.seal.model.IdentityTokenRequest;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.dombuilders.IdentityTokenResponseDOMBuilder;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.its.DocumentResponse;
import dk.sosi.sts.model.AudienceConfiguration;
import dk.sosi.sts.server.service.Timed;
import dk.sosi.sts.service.ItsConfigService;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import dk.sosi.sts.sla.StsContext;
import dk.sosi.sts.timing.Timer;
import dk.sosi.sts.util.StsUtil;
import org.apache.axis.Message;
import org.apache.axis.message.SOAPEnvelope;
import org.w3c.dom.Document;

import javax.xml.soap.SOAPException;
import java.io.IOException;
import java.io.OutputStream;
import java.security.KeyStoreException;
import java.util.Date;
import java.util.Properties;
import java.util.UUID;

public class IdentityTokenRequestHandler extends AbstractStsRequestHandler<IdentityTokenRequest,DocumentResponse,AudienceConfiguration> {

    private IDWSHFactory idwshFactory;
    private ItsConfigService configService;

	public IdentityTokenRequestHandler() {
	}

	public IdentityTokenRequestHandler(final SOSIFactory sosiFactory,
									   final StsStatus stsStatus,
									   final CertificateStatusChecker crlService,
									   final IDWSHFactory idwshFactory,
									   final ItsConfigService itsConfigService) {
		setSosiFactory(sosiFactory);
		setStsStatus(stsStatus);
		setCrlService(crlService);
		setIdwshFactory(idwshFactory);
		setItsConfigService(itsConfigService);
	}


	public IDWSHFactory getIdwshFactory() {
		return idwshFactory;
	}

	public void setIdwshFactory(final IDWSHFactory idwshFactory) {
		this.idwshFactory = idwshFactory;
	}

	public ItsConfigService getItsConfigService() {
		return configService;
	}

	public void setItsConfigService(final ItsConfigService configService) {
		this.configService = configService;
	}

	@Override
	protected String auditName() {
		return "ITS";
	}

	@Override
	@Timed
    protected IdentityTokenRequest deserializeRequest(final String doc) {
		return idwshFactory.createIdentityTokenRequestModelBuilder()
						   .buildModel(XmlUtil.readXml(new Properties(), doc, false));
	}

    @Override
	@Timed
    protected DocumentResponse perform(final IdentityTokenRequest identityTokenRequest, AudienceConfiguration config) throws VerifyException {
		SLALogHelper.createItem(SLALogPoint.ITS_SERIALIZE);
		
        final IDCard reqIDCard = identityTokenRequest.getIDCard();
        final String context = identityTokenRequest.getContext();
        final String reqMessageID = identityTokenRequest.getMessageID();
        

        StsContext.getContext().setMessageId(reqMessageID);

        if (!(reqIDCard instanceof UserIDCard)) {
            String faultStr = StsFaultStringValues.WST_INVALID_REQUEST + ": identity token service only support user card.";
			SLALogHelper.storeCallResultError(SLALogPoint.ITS_SERIALIZE, faultStr);
			throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, faultStr, StsFaultActorValues.ITS);
        }
        
    	// check time-range in idCard 
        if (!reqIDCard.isValidInTime()) {
            String faultStr = StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": ID card is not valid in time.";
			SLALogHelper.storeCallResultError(SLALogPoint.ITS_SERIALIZE, faultStr);
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, faultStr, StsFaultActorValues.ITS);
        }
        
        if (config.getIdCardMaxAge() != null
        	&& (System.currentTimeMillis() - reqIDCard.getCreatedDate().getTime() >= config.getIdCardMaxAge())) {
            String faultStr = StsFaultStringValues.WST_INVALID_TIME_RANGE + ": ID card is not valid in time due to maximal age restriction for this audience.";
			SLALogHelper.storeCallResultError(SLALogPoint.ITS_SERIALIZE, faultStr);
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, faultStr, StsFaultActorValues.ITS);
        }
        	
		Timer.start(this.getClass(), "seal:createIdentityTokenResponseDOMBuilder");
		final DocumentResponse itsResponse;
		try {
			final Document document = idwshFactory.createIdentityTokenResponseDOMBuilder()
					.setIdentityToken(createIdentityToken(config,(UserIDCard) reqIDCard))
					.setContext(context) // what is this?
					.setRelatesTo(reqMessageID)
					.build();
			itsResponse = new DocumentResponse(false,
					document);
			SLALogHelper.storeCallResultOk(SLALogPoint.ITS_SERIALIZE);
		} catch (Throwable t) {
			SLALogHelper.storeCallResultError(SLALogPoint.ITS_SERIALIZE, t.getMessage());
			throw StsUtil.rethrowable(t, VerifyException.class);
		} finally {
			Timer.stop(this.getClass(), "seal:createIdentityTokenResponseDOMBuilder");
		}

		return itsResponse;
    }

    private IdentityToken createIdentityToken(final AudienceConfiguration config, final UserIDCard signedIdCard) throws VerifyException {
        final IdentityTokenBuilder identityTokenBuilder = idwshFactory.createIdentityTokenBuilder();

        identityTokenBuilder.setAudienceRestriction(config.getAudience());

        if (config.requireCertificateAsReference()) {
            identityTokenBuilder.requireCertificateAsReference();
        }
        if (config.requireCprNumber()) {
            identityTokenBuilder.requireCprNumber();
        }
        if (config.requireCvrNumberIdentifier()) {
            identityTokenBuilder.requireCvrNumberIdentifier();
        }
        if (config.requireOrganizationName()) {
            identityTokenBuilder.requireOrganizationName();
        }
        if (config.requireITSystemName()) {
        	identityTokenBuilder.requireITSystemName();
        }
        if (config.requireAuthorizationCode()) {
        	identityTokenBuilder.requireUserAuthorizationCode();
        }
        if (config.requireUserEducationCode()) {
        	identityTokenBuilder.requireUserEducationCode();
        }

		final long now = System.currentTimeMillis();

		return identityTokenBuilder.setNotBefore(new Date(now - configService.getFuzzyTime()))
                                   .setNotOnOrAfter(new Date(Math.min(now + config.getLifetime(),
                                                                      signedIdCard.getExpiryDate().getTime())))
                                   .setUserIdCard(signedIdCard)
                                   .setIssuer(signedIdCard.getIssuer())
                                   .build();
    }

	@Override
    protected boolean isFault(final DocumentResponse stResponse) {
        return stResponse.isFault();
    }

    @Override
    protected Document serialize(final DocumentResponse resp) {
        return resp.getDocument();
    }

	@Timed
	@Override
	protected void serializeToOutputStream(final DocumentResponse doc, final OutputStream outputStream) throws IOException {
		// This might not be the best method for ITS, but it is the one, that was deployed in 1.3.x series
		final Document responseDoc = doc.getDocument();
		final SOAPEnvelope resp = new SOAPEnvelope();
		try {
			AxisUtil.addHeadersToSOAPEnvelope(resp, responseDoc, isFault(doc),
											  false); // do not include id
			AxisUtil.addBodyToSOAPEnvelope(resp, responseDoc);

			new Message(resp, false).writeTo(outputStream);
		} catch (SOAPException se) {
			throw new RuntimeException(se);
		}
	}

	@Timed
	@Override
    protected DocumentResponse createErrorResponse(final String dgwsVersion, final String msgId, final String faultCode, final String faultString, final String faultActor) {
        return createErrorResponse(null, faultCode, faultString, faultActor);
    }

	@Timed
    @Override
    protected DocumentResponse createErrorResponse(final IdentityTokenRequest stRequest, final String faultCode, final String faultString, final String faultActor) {
        final IdentityTokenResponseDOMBuilder identityTokenResponseDOMBuilder = idwshFactory
                .createIdentityTokenResponseDOMBuilder();

        identityTokenResponseDOMBuilder.setFaultCode(faultCode);
        identityTokenResponseDOMBuilder.setFaultString(faultString);
        identityTokenResponseDOMBuilder.setFaultFactor(faultActor);
        identityTokenResponseDOMBuilder.setRelatesTo(stRequest == null ? "<unknown>" : getMessageId(stRequest));

        final Document output = identityTokenResponseDOMBuilder.build();
        return new DocumentResponse(true, output);
    }

	@Override
	protected AudienceConfiguration verifyConfiguration(IdentityTokenRequest stRequest) throws VerifyException {
		final String reqAudience = stRequest.getAudience();
		final AudienceConfiguration config = configService.getAudienceConfiguration(reqAudience);
		if (config == null) {
		    throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST,
		                              StsFaultStringValues.WST_INVALID_REQUEST + ": audience not known.",
		                              StsFaultActorValues.ITS);
		}
		return config;
	}

	private String getMessageId(IdentityTokenRequest request) {
        String messageID = request.getMessageID();
        return messageID != null ? messageID : IboRequestHandler.URN_STS_PREFIX + UUID.randomUUID();
    }

}
