/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.model.dombuilders.AbstractOIOToIdentityTokenResponseDOMBuilder;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenResponseDOMBuilder;
import dk.sosi.seal.model.dombuilders.OIOSAMLAssertionToIDCardResponseDOMBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.StsException;
import dk.sosi.sts.TrustedVaultProvider;
import dk.sosi.sts.ibo.IboConventions;
import dk.sosi.sts.integrations.blurring.BlurringService;
import dk.sosi.sts.its.DocumentResponse;
import dk.sosi.sts.model.AudienceConfiguration;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.server.service.jwt.JwtIdpStrategy;
import dk.sosi.sts.service.OcesPidService;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import dk.sosi.sts.sla.StsContext;
import io.jsonwebtoken.*;
import org.springframework.beans.factory.InitializingBean;
import org.w3c.dom.Document;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.util.*;

import static java.util.Optional.*;

/**
 * Created by Kent on 21-02-2017.
 */
public class JWT2IdwsRequestHandler extends AbstractIdwsRequestHandler<JWTToIdentityTokenRequest, AudienceConfiguration> implements InitializingBean, TrustedVaultProvider {
    private OcesPidService ocesPidService;
    private long fuzzyTime;
    private long allowedDriftInSeconds;
    private File trustStoreFile;
    private String trustStorePassword;
    private KeyStore trustedVault;
    private Map<String, JwtIdpStrategy> issuerStrategies = new HashMap<>();
    private BlurringService blurringService;


    // invoked by spring
    public void setOcesPidService(OcesPidService service) {
        this.ocesPidService = service;
    }

    // invoked by spring
    public void setFuzzyTime(final long fuzzyTime) {
        this.fuzzyTime = fuzzyTime;
    }

    // invoked by spring
    public void setTrustStoreFile(File file) {
        this.trustStoreFile = file;
    }

    // invoked by spring
    public void setTrustStorePassword(String trustStorePassword) {
        this.trustStorePassword = trustStorePassword;
    }

    // invoked by spring
    public void setAllowedDriftInSeconds(final long allowedDriftInSeconds) {
        this.allowedDriftInSeconds = allowedDriftInSeconds;
    }
    // invoked by spring
    public void setIssuerStrategies(Map<String, JwtIdpStrategy> issuerStrategies) {
        this.issuerStrategies = issuerStrategies;
    }
    // invoked by spring
    public void setBlurringService(BlurringService blurringService) {
        this.blurringService = blurringService;
    }

    public BlurringService getBlurringService() { return blurringService; }

    @Override
    public void afterPropertiesSet() throws Exception {
        if (!trustStoreFile.exists()) {
            throw new FileNotFoundException(trustStoreFile.getAbsolutePath());
        }
        trustedVault = KeyStore.getInstance("JKS");
        try (FileInputStream is = new FileInputStream(trustStoreFile)) {
            trustedVault.load(is, trustStorePassword.toCharArray());
        }
    }

    // the name used to identify the service in auditlog
    @Override
    protected final String auditName() {
        return "JWT2Idws";
    }



    @Override
    protected JWTToIdentityTokenRequest deserializeRequest(String doc) {
        final Document requestDom = XmlUtil.readXml(new Properties(), doc, false);
        return new OIOSAMLFactory().createJWTToIdentityTokenRequestModelBuilder().build(requestDom);
    }

    @Override
    protected AudienceConfiguration verifyConfiguration(JWTToIdentityTokenRequest stRequest) throws VerifyException {
        final String reqAudience = stRequest.getAppliesTo();
        if (reqAudience == null || reqAudience.isEmpty()) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Audience is missing", StsFaultActorValues.ITS);
        }
        // Clients using standard .net classes (URI) will erroneously append an extra slash in the end so this is removed
        String normalizedAudience = IboConventions.normalizeAudience(reqAudience);

        final AudienceConfiguration config = itsConfigService.getAudienceConfiguration(normalizedAudience);
        if (config == null) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Audience not configured: " + reqAudience, StsFaultActorValues.ITS);
        }
        return config;
    }


    @Override
    protected DocumentResponse perform(JWTToIdentityTokenRequest req, AudienceConfiguration config) throws VerifyException {
        SLALogHelper.createItem(SLALogPoint.NBO_IDWS);
        StsContext.getContext().setMessageId(req.getMessageID());
        try {
            final DocumentResponse documentResponse = doPerform(req, config);
            SLALogHelper.storeCallResultOk(SLALogPoint.NBO_IDWS);
            return documentResponse;
        } catch (ModelException me) {
            final String faultString = StsFaultStringValues.WST_INVALID_REQUEST + ": " + me.getMessage();
            SLALogHelper.storeCallResultError(SLALogPoint.NBO_IDWS, faultString);
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, faultString, StsFaultActorValues.ITS);
        } catch (VerifyException v) {
            SLALogHelper.storeCallResultError(SLALogPoint.NBO_IDWS, v.getFaultString());
            throw v;
        }
    }

    @Override
    protected final DocumentResponse createErrorResponse(JWTToIdentityTokenRequest stRequest, String faultCode, String faultString, String faultActor) {
        final String messageId = stRequest == null ? "<unknown>" : stRequest.getMessageID();
        return createErrorResponse(null, messageId, faultCode, faultString, faultActor);
    }

    @Override
    protected final DocumentResponse createErrorResponse(String dgwsVersion, String msgId, String faultCode, String faultString, String faultActor) {
        OIOSAMLAssertionToIDCardResponseDOMBuilder response = new OIOSAMLFactory().createOIOSAMLAssertionToIDCardResponseDOMBuilder();

        response.setSigningVault(sosiFactory.getCredentialVault());
        response.setFaultCode(faultCode);
        response.setFaultActor(faultActor);
        response.setFaultString(faultString);
        response.setRelatesTo(msgId == null ? "<unknown>" : msgId);

        Document output = response.build();
        return new DocumentResponse(true, output);
    }

    private DocumentResponse doPerform(JWTToIdentityTokenRequest req, AudienceConfiguration config) throws VerifyException {
        final X509Certificate signingCertificate = verifyConsumer(req, config);

        Jws<Claims> jwtToken = getActAsToken(req);
        String issuer = jwtToken.getBody().getIssuer();
        JwtIdpStrategy jwtStrategy = getStrategy(issuer);
        SubjectSerialNumber pidSubjectSerial = jwtStrategy.getPidSubjectSerial(jwtToken);

        jwtStrategy.validateScope(jwtToken, config);
        detectMissingWhitelistingForWardClaim(config, signingCertificate, req.getWardCustodyClaim());

        final String cprFromToken = jwtStrategy.getTrustedCpr(jwtToken);
        final String cprNumber;
        
        final String subjectDN;
        final String subjectDNFormat;

        if (cprFromToken != null) {
            cprNumber = parseCpr(cprFromToken).getValue();
            
        	subjectDN = cprNumber;
        	subjectDNFormat = SAMLValues.NAMEID_FORMAT_PERSISTENT;
        } else {
        	subjectDN = jwtStrategy.getSubjectDistinguishedName(jwtToken);
        	subjectDNFormat = SAMLValues.NAMEID_FORMAT_X509_SUBJECT_NAME;
            String cprNumberClaim = req.getCPRNumberClaim();
            if (cprNumberClaim == null || cprNumberClaim.isEmpty()) {
                throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "CPR claim missing", StsFaultActorValues.ITS);
            }
            validateCpr(pidSubjectSerial, cprNumberClaim);
            cprNumber = cprNumberClaim;
        }

        AuditHelper.addNonPersonal("JWT2IdwsRequestHandler", "jwtIssuer", issuer);
        AuditHelper.addNonPersonal("JWT2IdwsRequestHandler", "jwtScope", jwtStrategy.getScope(jwtToken));
        AuditHelper.addPersonal("JWT2IdwsRequestHandler", "jwtUserSubjectDN", subjectDN);
        AuditHelper.addSensitive("JWT2IdwsRequestHandler", "jwtUserCpr", cprNumber);

        final BasicPrivilegesDOMBuilder privilegesDOMBuilder = createPrivilegeBuilder(
                parseCpr(cprNumber),
                req.getProcurationForCprClaim() != null
                        ? parseCpr(req.getProcurationForCprClaim())
                        : null);

        SubjectRelations11Impl.VerifiedRelation verifiedRelation = createValidatedSubjectRelations(
                req.getWardCustodyClaim(), req.getParentalCustodyClaim(), req.getPartlyWardCustodyClaim(), cprFromToken);
        SubjectRelationsDomBuilder subjectRelationsDomBuilder = createSubjectRelationsBuilder(verifiedRelation);

        BlurringInstructionsDomBuilder blurringInstructionsBuilder =
                createBlurringInstructionsBuilder(blurringService,
                                                  parseCpr(cprNumber),
                                                  req.getProcurationForCprClaim() != null
                                                          ? of(parseCpr(req.getProcurationForCprClaim()))
                                                          : empty(),
                                                  ofNullable(verifiedRelation),
                                                  req.getMessageID(),
                                                  config.requireBlurring());

        final IdentityToken identityToken = createIdentityToken(config,
                                                                jwtToken,
                                                                subjectDN,
                                                                subjectDNFormat,
                                                                signingCertificate,
                                                                cprNumber,
                                                                privilegesDOMBuilder,
                                                                blurringInstructionsBuilder,
                                                                subjectRelationsDomBuilder);
        final AbstractOIOToIdentityTokenResponseDOMBuilder<OIOBootstrapToIdentityTokenResponseDOMBuilder> responseBuilder = new OIOBootstrapToIdentityTokenResponseDOMBuilder();
        responseBuilder.setContext(req.getContext());
        responseBuilder.setIdentityToken(identityToken);
        responseBuilder.setRelatesTo(req.getMessageID());
        responseBuilder.setSigningVault(getSosiFactory().getCredentialVault());
        final Document responseDom = responseBuilder.build();
        return new DocumentResponse(false, responseDom);
    }

    private Jws<Claims> getActAsToken(JWTToIdentityTokenRequest req) throws VerifyException {
        String jwt = req.getJsonWebToken();

        // Assertion check trust
        final JwtParser jwtParser = Jwts.parser();
        jwtParser.setSigningKeyResolver(new SigningKeyResolver() {
            @Override
            public Key resolveSigningKey(JwsHeader jwsHeader, Claims claims) {
                final String kid = jwsHeader.getKeyId();
                return lookupKey(kid).getPublicKey(); // throws error if not found
            }

            @Override
            public Key resolveSigningKey(JwsHeader jwsHeader, String s) {
                throw new UnsupportedOperationException("Method not implemented");
            }
        });
        jwtParser.setAllowedClockSkewSeconds(allowedDriftInSeconds);
        final Jws<Claims> claimsJws = jwtParser.parseClaimsJws(jwt);

        // have already been found during parsing, so never expected to fail
        final String kid = claimsJws.getHeader().getKeyId();
        X509Certificate idpCertificate = lookupKey(kid);

        checkCertificate(idpCertificate, "JWT signer", StsFaultActorValues.ITS);
        return claimsJws;
    }

    private void validateCpr(SubjectSerialNumber ssn, String cprClaim) throws VerifyException {
        final CprNr cprNr = parseCpr(cprClaim);
        boolean result;
        try {
            result = ocesPidService.isRelated(ssn, cprNr);
        } catch (StsException e) {
            log.error("Unable to validateScope cpr", e);
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validateScope cpr", StsFaultActorValues.ITS);
        }
        if (!result) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "CPR/PID mismatch", StsFaultActorValues.ITS);
        }
    }

    private IdentityToken createIdentityToken(final AudienceConfiguration config,
                                              final Jws<Claims> jwtToken,
                                              String subjectNameId,
                                              String subjectNameIdFormat,
                                              X509Certificate holderOfKeyCertificate,
                                              String currentUserCpr,
                                              BasicPrivilegesDOMBuilder privilegesDOMBuilder,
                                              BlurringInstructionsDomBuilder blurringInstructionsBuilder,
                                              SubjectRelationsDomBuilder subjectRelationsDomBuilder) {
        final CitizenIdentityTokenBuilder identityTokenBuilder = new OIOSAMLFactory().createCitizenIdentityTokenBuilder();

        identityTokenBuilder.setSubjectNameID(subjectNameId);
        identityTokenBuilder.setSubjectNameIDFormat(subjectNameIdFormat);

        identityTokenBuilder.setCprNumberAttribute(currentUserCpr);

        identityTokenBuilder.setRecipientURL(config.getAudience());

        identityTokenBuilder.setAudienceRestriction(config.getAudience());
        identityTokenBuilder.setHolderOfKeyCertificate(holderOfKeyCertificate);
        identityTokenBuilder.setSigningVault(getSosiFactory().getCredentialVault());
        identityTokenBuilder.setBlurringInstructions(blurringInstructionsBuilder);
        identityTokenBuilder.setPrivileges(privilegesDOMBuilder);
        identityTokenBuilder.setSubjectRelations(subjectRelationsDomBuilder);

        final long now = System.currentTimeMillis();
        // not before not present in input assertion
        final Date notBefore = new Date(now - fuzzyTime);
        final Date notOnOrAfter = new Date(Math.min(now + config.getLifetime(), jwtToken.getBody().getExpiration().getTime()));
        return identityTokenBuilder.setNotBefore(notBefore)
                .setNotOnOrAfter(notOnOrAfter)
                .setDeliveryNotOnOrAfter(notOnOrAfter)
                .setIssuer(getSosiFactory().getIssuer())
                .build();
    }

    private X509Certificate lookupKey(String keyIdentifier) {
        try {
            final Certificate certificate = trustedVault.getCertificate(keyIdentifier);
            if (certificate != null) {
                return (X509Certificate) certificate;
            } else {
                throw new SignatureException("Unable to get signing certificate for kid " + keyIdentifier);
            }
        } catch (KeyStoreException e) {
            throw new SignatureException("Unable to get signing certificate for kid " + keyIdentifier, e);
        }
    }

    private JwtIdpStrategy getStrategy(String issuer) throws VerifyException {
        final JwtIdpStrategy jwtIdpStrategy = issuerStrategies.get(issuer);
        if (jwtIdpStrategy == null) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Unknown issuer: " + issuer, StsFaultActorValues.ITS);
        }
        return jwtIdpStrategy;
    }

    @Override
    public CredentialVault getTrustedVault() {
        return new GenericCredentialVault(new Properties(), trustedVault, trustStorePassword);
    }
}