/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.DGWSConstants;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.model.dombuilders.OIOSAMLAssertionToIDCardResponseDOMBuilder;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.its.DocumentResponse;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.server.service.Timed;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import dk.sosi.sts.sla.StsContext;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.io.IOException;
import java.io.OutputStream;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.Properties;

/**
 *
 */
public class NboRequestHandler extends AbstractStsRequestHandler<OIOSAMLAssertionToIDCardRequest, DocumentResponse, Void> {

	private final OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();

	private UserValidationService userValidationService;
	private NboConfig nboConfig;
	private String blacklistedITSystemName;

	public void setUserValidationService(UserValidationService userValidationService) {
		this.userValidationService = userValidationService;
	}

	public void setNboConfig(final NboConfig nboConfig) {
		this.nboConfig = nboConfig;
	}

	public void setBlacklistedITSystemName(String blacklistedITSystemName) {
		this.blacklistedITSystemName = blacklistedITSystemName;
	}

	@Timed
	@Override
	protected OIOSAMLAssertionToIDCardRequest deserializeRequest(final String doc) {
		return oiosamlFactory.createOIOSAMLAssertionToIDCardRequestModelBuilder().build(XmlUtil.readXml(new Properties(), doc, false));
	}

	@Override
	protected Void verifyConfiguration(OIOSAMLAssertionToIDCardRequest stRequest) {
		return null;
	}

	@Override
	protected DocumentResponse perform(final OIOSAMLAssertionToIDCardRequest req, Void configType) throws VerifyException {
		SLALogHelper.createItem(SLALogPoint.NBO_SERIALIZE);

		StsContext.getContext().setMessageId(req.getMessageID());

		try {
			try {
				validateEnvelope(req);
				OIOSAMLAssertion assertion = req.getOIOSAMLAssertion();
				// Assertion check trust
								
				CredentialVault trustedVault = nboConfig.getTrustedVault();
				assertion.validateSignatureAndTrust(trustedVault);
				checkCertificate(assertion.getSigningCertificate(), "NBO certificate", StsFaultActorValues.NBO);

				validateAssertion(assertion);

				UserInfo userInfo = buildUserInfo(req);
				SystemInfo systemInfo = buildSystemInfo(assertion, req);
				IDCard signedIdCard = buildSignedIdCard(assertion, userInfo, systemInfo);

				OIOSAMLAssertionToIDCardResponseDOMBuilder response = oiosamlFactory.createOIOSAMLAssertionToIDCardResponseDOMBuilder();
				response.setIDCard(signedIdCard);
				response.setContext(req.getContext());
				response.setRelatesTo(req.getMessageID());
				response.setSigningVault(sosiFactory.getCredentialVault(), req.isUsingSecurityTokenReference());

				Document built = response.build();
				SLALogHelper.storeCallResultOk(SLALogPoint.NBO_SERIALIZE);
				return new DocumentResponse(false, built);
			} catch (ModelException me) {
				throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, StsFaultStringValues.WST_INVALID_REQUEST + ": " + me.getMessage(), StsFaultActorValues.NBO);
			}
		} catch (VerifyException v) {
			SLALogHelper.storeCallResultError(SLALogPoint.NBO_SERIALIZE, v.getFaultString());
			throw v;
		}

	}

	protected void validateAssertion(final OIOSAMLAssertion assertion) throws VerifyException {
		// Check time for assertion
		assertion.validateTimestamp(nboConfig.getAllowedDriftInSeconds());

		if (!acceptAssuranceLevel(assertion.getAssuranceLevel())) {
			throw new VerifyException(StsFaultCodeValues.WST_ASSURANCE_LEVEL, StsFaultStringValues.WST_ASSURANCE_LEVEL, StsFaultActorValues.NBO);
		}
	}

	/**
	 * OIOSAML definerer AssuranceLevel som string med følgende tilladte værdier: “1”, “2”, “3”, “4” og “test”, så lad
	 * os holde det som string i Seal - men i STS synes jeg vi skal vedtage at det kun er "3" og "4" der understøttes og
	 * lad kald med et andet AssuranceLevel fejle.
	 */
	private boolean acceptAssuranceLevel(final String assuranceLevel) {
		if (assuranceLevel == null || "test".equals(assuranceLevel)) {
			return false;
		}
		try {
			int assuranceLevelInt = Integer.parseInt(assuranceLevel);
			switch (assuranceLevelInt) {
			case 3:
			case 4:
				return true;
			default:
				return false;
			}
		} catch (NumberFormatException n) {
			return false;
		}
	}

	protected void validateEnvelope(final OIOSAMLAssertionToIDCardRequest req) throws VerifyException {
		validateITSystemName(req.getITSystemName());
		// Validate the signature and check certificate
		new OIOWSTrustValidation(StsFaultActorValues.NBO).validateSignature(req);
		// check for issued by right OCES CA
		new CertificateStatusVerifier(sosiFactory.getFederation()).verifyCertificateStatus(req.getSigningCertificate(), StsFaultActorValues.NBO);

	}

	private void validateITSystemName(String itSystemName) throws VerifyException {
		if (blacklistedITSystemName.equals(itSystemName)) {
			throw new VerifyException(StsFaultCodeValues.WST_BAD_REQUEST,
					StsFaultStringValues.WST_BAD_REQUEST + ": Invalid ITSystemName",
					StsFaultActorValues.NBO);
		}
	}

	private IDCard buildSignedIdCard(final OIOSAMLAssertion assertion, final UserInfo userInfo, final SystemInfo systemInfo) {
		String certHash;
		X509Certificate userCertificate = assertion.getUserCertificate();
		// User certificate is optional according to OIOSaml (OCES sub profile)
		if (userCertificate != null) {
			certHash = SignatureUtil.getDigestOfCertificate(userCertificate);
		} else {
			certHash = null;
		}

		Date createdDate = new Date(System.currentTimeMillis() - nboConfig.getFuzzyTime());
		Date expiryDate = new Date(createdDate.getTime() + nboConfig.getIdCardDuration());
		Element document = null;// document == null => needs signature
		String cardId = XmlUtil.createGUID();

		return sign(new UserIDCard(DGWSConstants.VERSION_1_0_1, document, cardId, AuthenticationLevel.MOCES_TRUSTED_USER, certHash, sosiFactory.getIssuer(), systemInfo, userInfo, createdDate, expiryDate, null, null, null));
	}

	private UserIDCard sign(final UserIDCard idCard) {
		// This will not work since, the time range is explicitly set disregarding the set values
		// return (UserIDCard) sosiFactory.copyToVOCESSignedIDCard(idCard);

		// A "hack" (I guess) from SOSIFactory#getSignedIdCard
		Request tmp = new Request(DGWSConstants.VERSION_1_0_1, false, "", sosiFactory);
		tmp.setIDCard(idCard);
		Document doc = tmp.serialize2DOMDocument();
		idCard.sign(doc, SignatureProviderFactory.fromCredentialVault(sosiFactory.getCredentialVault()));
		return (UserIDCard) tmp.getIDCard();

		// This will result in a NPE
		// final Document doc = XmlUtil.createEmptyDocument();
		// idCard.serialize2DOMDocument(sosiFactory, doc);
		// idCard.sign(doc, SignatureProviderFactory.fromCredentialVault(sosiFactory.getCredentialVault()));
		// return idCard;
	}

	private SystemInfo buildSystemInfo(final OIOSAMLAssertion assertion, final OIOSAMLAssertionToIDCardRequest req) throws VerifyException {
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, assertion.getCvrNumberIdentifier(), assertion.getOrganizationName());

		if (req.getITSystemName() == null) {
			throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, StsFaultStringValues.WST_INVALID_REQUEST + ": IT System must be specified.", StsFaultActorValues.NBO);
		}

		return new SystemInfo(careProvider, req.getITSystemName());
	}

	private OcesInfo buildOcesInfo(OIOSAMLAssertion assertion) {
		// We trust that the CvrNumberIdentifier and the RidNumberIdentifier are correct in the token
		X509Certificate userCertificate = assertion.getUserCertificate();
		// User certificate is optional according to OIOSaml (OCES sub profile)
		if (userCertificate != null) {
			return OcesInfoParser.parseCertificate(userCertificate);
		} else {
			// according to OCES sub profile of OIO-SAML, this should be the certificates SSN.
			String uidAttribute = assertion.getAttribute(OIOSAMLAttributes.UID);
			SubjectSerialNumber ssn = new SubjectSerialNumber(uidAttribute);
			return OcesInfoParser.parseSubjectSerialNumber(ssn);
		}
	}

	private UserInfo buildUserInfo(final OIOSAMLAssertionToIDCardRequest request) throws VerifyException {
		OIOSAMLAssertion assertion = request.getOIOSAMLAssertion();
		OcesInfo ocesInfo = buildOcesInfo(assertion);
		final CprNr inputCpr = getInputCpr(assertion.getAttribute(OIOSAMLAttributes.CPR_NUMBER));
		final String inputRole = request.getUserEducationCode();
		final String inputAuthorizationCode = request.getUserAuthorizationCode();

		final boolean cprTrustedForSigner = isCprTrustedForSigner(assertion.getSigningCertificate());

		VerifyInput verifyInput = new VerifyInput(ocesInfo, inputCpr, inputRole, inputAuthorizationCode);

		final String endpoint = "NBO"; // WARNING: this name may be used to check if we trust roles - beware, if changing
		VerifyResult verifyResult = userValidationService.verifyCprAndRole(verifyInput, cprTrustedForSigner, endpoint);
		if (verifyResult.role == null) {
			throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, StsFaultStringValues.WST_INVALID_REQUEST + ": EducationCode must be specified or unique authorization must exist.", StsFaultActorValues.NBO);
		}

		final String givenName;
		final String surName;
		if (request.getUserGivenName() != null && request.getUserSurName() != null) {
			givenName = request.getUserGivenName();
			surName = request.getUserSurName();
		} else {
			// The IdP cannot split CommonName and neither should we (assertion.getSurName() returns null)
			givenName = assertion.getCommonName();
			surName = "-";
		}
		String email = assertion.getEmail();
		String occupation = null;
		return new UserInfo(verifyResult.cpr.getValue(), givenName, surName, email, occupation, verifyResult.role, verifyResult.authorizationCode);
	}

	private CprNr getInputCpr(String assertionCpr) {
		final CprNr inputCpr;
		if (assertionCpr == null || "".equals(assertionCpr)) {
			inputCpr = null;
		} else {
			inputCpr = new CprNr(assertionCpr);
		}
		return inputCpr;
	}

	private boolean isCprTrustedForSigner(X509Certificate signingCertificate) {
		String ssn = OcesInfoParser.parseCertificate(signingCertificate).getSubjectSerialNumber().getValue();
		return nboConfig.isCprTrustedFrom(ssn);
	}

	@Override
	protected boolean isFault(final DocumentResponse stResponse) {
		return stResponse.isFault();
	}

	@Override
	protected Document serialize(final DocumentResponse response) {
		return response.getDocument();
	}

	@Override
	@Timed
	protected void serializeToOutputStream(final DocumentResponse doc, final OutputStream outputStream) throws IOException {
		outputStream.write(XmlUtil.serializeXml2ByteArray(serialize(doc), true));
	}

	@Override
	protected DocumentResponse createErrorResponse(final String dgwsVersion, final String msgId, final String faultCode, final String faultString, final String faultActor) {
		return createErrorResponse(null, faultCode, faultString, faultActor);
	}

	@Timed
	@Override
	protected DocumentResponse createErrorResponse(final OIOSAMLAssertionToIDCardRequest stRequest, final String faultCode, final String faultString, final String faultActor) {
		OIOSAMLAssertionToIDCardResponseDOMBuilder response = oiosamlFactory.createOIOSAMLAssertionToIDCardResponseDOMBuilder();

		response.setSigningVault(sosiFactory.getCredentialVault());
		response.setFaultCode(faultCode);
		response.setFaultActor(faultActor);
		response.setFaultString(faultString);
		response.setRelatesTo(stRequest == null ? "<unknown>" : stRequest.getMessageID());

		Document output = response.build();
		return new DocumentResponse(true, output);
	}

	@Override
	protected String auditName() {
		return "NBO";
	}
}
