/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStream;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.util.*;

import org.springframework.beans.factory.InitializingBean;
import org.w3c.dom.Document;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.BootstrapOIOSAMLAssertion;
import dk.sosi.seal.model.BootstrapOIOSAMLAssertionBuilder;
import dk.sosi.seal.model.JWTToOIOSAMLAssertionRequest;
import dk.sosi.seal.model.OIOBootstrapToken;
import dk.sosi.seal.model.OIOBootstrapTokenBuilder;
import dk.sosi.seal.model.OIOWSTrustMessage;
import dk.sosi.seal.model.OIOWSTrustRequest;
import dk.sosi.seal.model.dombuilders.IDCardToOIOSAMLAssertionResponseDOMBuilder;
import dk.sosi.seal.model.dombuilders.JWTToOIOSAMLAssertionResponseDOMBuilder;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.StsException;
import dk.sosi.sts.ibo.IboConventions;
import dk.sosi.sts.its.DocumentResponse;
import dk.sosi.sts.its.SboAudienceConfigurationImpl;
import dk.sosi.sts.model.AudienceConfigurationBuilder;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SboAudienceConfiguration;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.server.service.jwt.JwtIdpStrategy;
import dk.sosi.sts.service.ItsConfigService;
import dk.sosi.sts.service.OcesPidService;
import io.jsonwebtoken.Claims;
import io.jsonwebtoken.Jws;
import io.jsonwebtoken.JwsHeader;
import io.jsonwebtoken.JwtParser;
import io.jsonwebtoken.Jwts;
import io.jsonwebtoken.SignatureException;
import io.jsonwebtoken.SigningKeyResolver;

/**
 */
public class SboRequestHandler extends AbstractStsRequestHandler<JWTToOIOSAMLAssertionRequest, DocumentResponse, SboAudienceConfiguration> implements InitializingBean{

	public static final String URN_STS_PREFIX = "urn:sts:generated:";

	private final OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();

	private SOSIFactory sosiFactory;
	private boolean checkTargetCertificate = true;

	// AbstractIDWSRequest
	protected ItsConfigService itsConfigService;
	// FROM JWT2IDWS
	private long fuzzyTime;
	private OcesPidService ocesPidService;
	private long allowedDriftInSeconds;
	private KeyStore trustedVault;
	private Map<String, JwtIdpStrategy> issuerStrategies = new HashMap<>();
    private File trustStoreFile;
    private String trustStorePassword;

	@Override
	protected JWTToOIOSAMLAssertionRequest deserializeRequest(final String doc) {
		Document readXml = XmlUtil.readXml(new Properties(), doc, false);
		return oiosamlFactory.createJWTToOIOSAMLAssertionRequestModelBuilder().build(readXml);
	}

	@Override
	protected SboAudienceConfiguration verifyConfiguration(JWTToOIOSAMLAssertionRequest req) throws VerifyException {
		// Requested audience er konfigureret i STS
        final String audience = req.getAppliesTo();
        if (audience == null || audience.isEmpty()) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Audience is missing", StsFaultActorValues.ITS);
        }
        // Clients using standard .net classes (URI) will erroneously append an extra slash in the end so this is removed
        String normalizedAudience = IboConventions.normalizeAudience(audience);

        AudienceConfigurationBuilder<SboAudienceConfiguration> audienceConfigurationBuilder = new AudienceConfigurationBuilder<SboAudienceConfiguration>(SboAudienceConfigurationImpl.class);
        final SboAudienceConfiguration audienceConfiguration = itsConfigService.getAudienceConfiguration(normalizedAudience, audienceConfigurationBuilder);
        
        if (audienceConfiguration == null) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Audience not configured: " + audience, StsFaultActorValues.ITS);
        }
        
        if (checkTargetCertificate) {
			final X509Certificate targetCertificate = audienceConfiguration.getTargetCertificate();
			try {
				CertificateStatus status = sosiFactory.getFederation().getCertificateStatus(targetCertificate);
				if (!status.isValid() && status.getTimestamp() != null) {
					throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
											  StsFaultStringValues.WST_REQUEST_FAILED + ": Certificate revoked (" + status.getTimestamp() + ") for audience " + audience,
											  StsFaultActorValues.STS);
				} else if (!status.isValid()){
					throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
											  StsFaultStringValues.WST_REQUEST_FAILED + ": Certificate expired or not yet valid for audience " + audience,
											  StsFaultActorValues.STS);
				}
			} catch (PKIException e) {
				throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
										  StsFaultStringValues.WST_REQUEST_FAILED + ": Certificate invalid for audience " + audience + " (" + e.getMessage() +")",
										  StsFaultActorValues.STS);
			}
		}
        
        
        return audienceConfiguration;
	}
	
	
	@Override
	protected DocumentResponse perform(JWTToOIOSAMLAssertionRequest req, SboAudienceConfiguration config) throws VerifyException {
		/*REQUEST*/
        
        // signeret af en whitelistet part
		final X509Certificate signingCertificate = verifyConsumer(req, config);
		//CertificateInfo certInfo = new CertificateInfo(signingCertificate);
        
        // Valid request hvor JWT:
        //  - er gyldig i tid
        //  - indeholder et ’certsubdn’ formateret som et X509SubjectName for et POCES certifikat og som dermed indeholder en PID
        //  - er signeret af en trusted part (fx en OIDC AS), som STS’en er konfigureret med (STS’en bør benytte ’iss’ til at finde rette udsteders certifikat)
        Jws<Claims> jwtToken = getActAsToken(req);
        
        String issuer = jwtToken.getBody().getIssuer();
        JwtIdpStrategy jwtStrategy = getStrategy(issuer);
        
        // indeholder et ’certsubdn’ formateret som et X509SubjectName for et POCES certifikat og som dermed indeholder en PID
        String subjectDN = jwtStrategy.getSubjectDistinguishedName(jwtToken);
        SubjectSerialNumber pidSubjectSerial = jwtStrategy.getPidSubjectSerial(jwtToken);
        

        // indeholder valid værdi i dens ’scope’ liste
        jwtStrategy.validateScope(jwtToken, config);

		//- er gyldig i tid		- 
		//- er signeret af en trusted part (fx en OIDC AS), som STS’en er konfigureret med (STS’en bør benytte ’iss’ til at finde rette udsteders certifikat)

        final String cprFromToken = jwtStrategy.getTrustedCpr(jwtToken);
        final String cprNumber;
        if (cprFromToken != null) {
        	// JWT indeholder CPR
            cprNumber = parseCpr(cprFromToken).getValue();
        } else {
        	// WS-Trust indeholder CPR
            String cprNumberClaim = req.getCPRNumberClaim();
            if (cprNumberClaim == null || cprNumberClaim.isEmpty()) {
                throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "CPR claim missing", StsFaultActorValues.ITS);
            }
            //CPR claimet matcher PID’en fra NameID
            validateCpr(pidSubjectSerial, cprNumberClaim);
            cprNumber = cprNumberClaim;
        }
        
        /*AUDIT*/
        AuditHelper.addNonPersonal("JWT2IdwsRequestHandler", "jwtIssuer", issuer);
        AuditHelper.addNonPersonal("JWT2IdwsRequestHandler", "jwtScope", jwtStrategy.getScope(jwtToken));
        AuditHelper.addPersonal("JWT2IdwsRequestHandler", "jwtUserSubjectDN", subjectDN);
        AuditHelper.addSensitive("JWT2IdwsRequestHandler", "jwtUserCpr", cprNumber);
        
		
		/*RESPONSE*/
        OIOBootstrapTokenBuilder oioBootstrapTokenBuilder = new OIOBootstrapTokenBuilder(sosiFactory.getCredentialVault(), sosiFactory.getIssuer()); 
		final BootstrapOIOSAMLAssertion assertion = createBootstrapOIOSAMLAssertion(config, jwtToken, subjectDN, signingCertificate, cprNumber, oioBootstrapTokenBuilder);//builder.build();

		final JWTToOIOSAMLAssertionResponseDOMBuilder responseBuilder = oiosamlFactory.createJWTToOIOSAMLAssertionResponseDOMBuilder();
		responseBuilder.setSigningVault(sosiFactory.getCredentialVault());
		responseBuilder.setEncryptionKey(config.getTargetCertificate().getPublicKey());
		responseBuilder.setRelatesTo(getMessageId(req));
		responseBuilder.setOIOSAMLAssertion(assertion);
		responseBuilder.setContext(req.getContext());
		
		return new DocumentResponse(false,
									responseBuilder.build());
	}
	
	protected final BootstrapOIOSAMLAssertion createBootstrapOIOSAMLAssertion(
			final SboAudienceConfiguration config, final Jws<Claims> jwtToken, String subjectDN, X509Certificate holderOfKeyCertificate, String cprNumber, OIOBootstrapTokenBuilder oioBootstrapTokenBuilder) {
				
		final long now = System.currentTimeMillis();
	       
	      // not before not present in input assertion
	      final Date notBefore = new Date(now - fuzzyTime);
	      final Date notOnOrAfter = new Date(Math.min(now + config.getLifetime(), jwtToken.getBody().getExpiration().getTime()));
	  
	     
		 BootstrapOIOSAMLAssertionBuilder bootstrapOIOSAMLAssertionBuilder = oiosamlFactory.createBootstrapOIOSAMLAssertionBuilder()
				.setDeliveryNotOnOrAfter(notOnOrAfter)
				.setNotBefore(notBefore)
				.setNotOnOrAfter(notOnOrAfter)
				.setRecipientURL(config.getRecipientURL())
				.setAudienceRestriction(config.getAudience())
				.setIssuer(sosiFactory.getIssuer())
				.setSigningVault(sosiFactory.getCredentialVault())
				.setCreated(jwtToken.getBody().getIssuedAt())
				.setCpr(cprNumber)
				.setSubjectName(subjectDN);


		if (config.includeBootStrapToken()) {
			OIOBootstrapToken oioBootstrapToken = oioBootstrapTokenBuilder
					.setCpr(cprNumber)
					.setSubjectName(subjectDN)
					.setAudienceRestriction("http://sosi.dk")
					.validFrom(notBefore)
					.expires(notOnOrAfter)
					.setHolderOfKeyCertificate(holderOfKeyCertificate)
					.build();

			bootstrapOIOSAMLAssertionBuilder.setBootstrapToken(oioBootstrapToken).includeBootstrapToken();
		}

		return bootstrapOIOSAMLAssertionBuilder.build();
	}

	@Override
	protected boolean isFault(final DocumentResponse response) {
		return response.isFault();
	}

	@Override
	protected Document serialize(final DocumentResponse response) {
		return response.getDocument();
	}

	@Override
	protected void serializeToOutputStream(final DocumentResponse doc, final OutputStream outputStream) throws IOException {
		outputStream.write(XmlUtil.serializeXml2ByteArray(serialize(doc),
														  true));
	}

	@Override
	protected DocumentResponse createErrorResponse(final String dgwsVersion, final String msgId, final String faultCode, final String faultString, final String faultActor) {
		final IDCardToOIOSAMLAssertionResponseDOMBuilder builder = oiosamlFactory
				.createIDCardToOIOSAMLAssertionResponseDOMBuilder();

		builder.setSigningVault(sosiFactory.getCredentialVault());
		builder.setFaultActor(faultActor);
		builder.setFaultString(faultString);
		builder.setFaultCode(faultCode);
		builder.setRelatesTo(msgId);

		return new DocumentResponse(true,
									builder.build());
	}

	@Override
	protected DocumentResponse createErrorResponse(final JWTToOIOSAMLAssertionRequest stRequest, final String faultCode, final String faultString, final String faultActor) {
		return createErrorResponse("JWT2OIOSaml Version", // ignored
										  getMessageId(stRequest),
										  faultCode,
										  faultString,
										  faultActor);
	}

	/**
	 * STS-20:
	 * This is a fix for the case where requests go threw SOSI-GW, which will erase all wsa-headers.
	 *
	 * IDCardToOIOSAMLAssertionRequest#getMessageID will throw a NPE in case the wsa:MessageID is
	 * not present in the request. This method will catch it and substitute a meaningful value.
	 *
	 * @param request the exchange request.
	 * @return the MessageID from the exchange request or a meaningful value.
	 */
	private String getMessageId(OIOWSTrustMessage request) {
        String messageID = request.getMessageID();
        return messageID != null ? messageID : URN_STS_PREFIX + UUID.randomUUID();
	}


	@Override
	protected String auditName() {
		return "JWT2OIO";
	}

	public SOSIFactory getSosiFactory() {
		return sosiFactory;
	}

	public void setSosiFactory(final SOSIFactory sosiFactory) {
		this.sosiFactory = sosiFactory;
	}

	// From IboRequest
	public void setCheckTargetCertificate(final boolean checkTargetCertificate) {
		this.checkTargetCertificate = checkTargetCertificate;
	}
	// From JWT2IDWS
	private Jws<Claims> getActAsToken(JWTToOIOSAMLAssertionRequest req) throws VerifyException {
        String jwt = req.getJsonWebToken();

        // Assertion check trust
        final JwtParser jwtParser = Jwts.parser();
        jwtParser.setSigningKeyResolver(new SigningKeyResolver() {
            @Override
            public Key resolveSigningKey(@SuppressWarnings("rawtypes") JwsHeader jwsHeader, Claims claims) {
            	// STS’en bør benytte ’iss’ til at finde rette udsteders certifikat
                final String kid = jwsHeader.getKeyId();//claims.getIssuer();
                return lookupKey(kid).getPublicKey(); // throws error if not found
            }

            @Override
            public Key resolveSigningKey(@SuppressWarnings("rawtypes") JwsHeader jwsHeader, String s) {
                throw new UnsupportedOperationException("Method not implemented");
            }
        });
        // gyldig i tid
        jwtParser.setAllowedClockSkewSeconds(allowedDriftInSeconds);
        final Jws<Claims> claimsJws = jwtParser.parseClaimsJws(jwt);

        // have already been found during parsing, so never expected to fail
        final String kid = claimsJws.getHeader().getKeyId();
        
        // signeret af en trusted part
        X509Certificate idpCertificate = lookupKey(kid);
        checkCertificate(idpCertificate, "JWT signer", StsFaultActorValues.ITS);
        return claimsJws;
    }
	
	private X509Certificate lookupKey(String keyIdentifier) {
        try {
            final Certificate certificate = trustedVault.getCertificate(keyIdentifier);
            if (certificate != null) {
                return (X509Certificate) certificate;
            } else {
                throw new SignatureException("Unable to get signing certificate for kid " + keyIdentifier);
            }
        } catch (KeyStoreException e) {
            throw new SignatureException("Unable to get signing certificate for kid " + keyIdentifier, e);
        }
    }
	
	private JwtIdpStrategy getStrategy(String issuer) throws VerifyException {
        final JwtIdpStrategy jwtIdpStrategy = issuerStrategies.get(issuer);
        if (jwtIdpStrategy == null) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Unknown issuer: " + issuer, StsFaultActorValues.ITS);
        }
        return jwtIdpStrategy;
    }
	
	private void validateCpr(SubjectSerialNumber ssn, String cprClaim) throws VerifyException {
        final CprNr cprNr = parseCpr(cprClaim);
        boolean result;
        try {
            result = ocesPidService.isRelated(ssn, cprNr);
        } catch (StsException e) {
            log.error("Unable to validateScope cpr", e);
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validateScope cpr", StsFaultActorValues.ITS);
        }
        if (!result) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "CPR/PID mismatch", StsFaultActorValues.ITS);
        }
    }
	
	@Override
    public void afterPropertiesSet() throws Exception {
        if (!trustStoreFile.exists()) {
            throw new FileNotFoundException(trustStoreFile.getAbsolutePath());
        }
        trustedVault = KeyStore.getInstance("JKS");
        try (FileInputStream is = new FileInputStream(trustStoreFile)) {
            trustedVault.load(is, trustStorePassword.toCharArray());
        }
    }
	
	// From AbstractIdwsRequest
    protected final X509Certificate verifyConsumer(OIOWSTrustRequest request, SboAudienceConfiguration audienceConfiguration) throws VerifyException {
        validateMessageSignature(request);
        X509Certificate signingCertificate = request.getSigningCertificate();
        verifySigningCertificate(signingCertificate);
        verifyConsumerWhitelisted(signingCertificate, audienceConfiguration);
        return signingCertificate;
    }
    
    private void verifyConsumerWhitelisted(X509Certificate signingCertificate, SboAudienceConfiguration config) throws VerifyException {
        final String certificateDn = signingCertificate.getSubjectDN().getName();
        if (!config.getCertificates().contains(new DistinguishedName(certificateDn).getSubjectSerialNumber())) {
            final String faultString = "Signing certificate not whitelisted for this audience (cert={"+certificateDn+"}, audience={"+config.getAudience()+"})";
            throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, faultString, StsFaultActorValues.ITS);
        }
    }

    private void verifySigningCertificate(X509Certificate signingCertificate) throws VerifyException {
        new CertificateStatusVerifier(sosiFactory.getFederation()).verifyCertificateStatus(signingCertificate, StsFaultActorValues.ITS);
    }

    private void validateMessageSignature(OIOWSTrustRequest req) throws VerifyException {
        new OIOWSTrustValidation(StsFaultActorValues.ITS).validateSignature(req);
    }
    
    protected final CprNr parseCpr(String cpr) throws VerifyException {
        try {
            return new CprNr(cpr);
        } catch (IllegalArgumentException e) {
            log.warn("Unable to parse cpr", e);
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, e.getMessage(), StsFaultActorValues.ITS);
        }
    }
    
    // invoked by spring
    public void setTrustStoreFile(File file) {
        this.trustStoreFile = file;
    }

    // invoked by spring
    public void setTrustStorePassword(String trustStorePassword) {
        this.trustStorePassword = trustStorePassword;
    }
    
    // invoked from spring
    public void setItsConfigService(ItsConfigService itsConfigService) {
        this.itsConfigService = itsConfigService;
    }
    
	
    // invoked by spring
    public void setOcesPidService(OcesPidService service) {
        this.ocesPidService = service;
    }


    // invoked by spring
    public void setFuzzyTime(final long fuzzyTime) {
        this.fuzzyTime = fuzzyTime;
    }


    // invoked by spring
    public void setAllowedDriftInSeconds(final long allowedDriftInSeconds) {
        this.allowedDriftInSeconds = allowedDriftInSeconds;
    }

    // invoked by spring
    public void setIssuerStrategies(Map<String, JwtIdpStrategy> issuerStrategies) {
        this.issuerStrategies = issuerStrategies;
    }
}
