/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.SecurityTokenResponse;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.model.SystemInfo;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfo.CertificateType;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.service.AclService;
import org.apache.log4j.Logger;

import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;

public class SecurityTokenService {

	private final Logger log = Logger.getLogger(SecurityTokenService.class);

	private final SOSIFactory sosiFactory;
	private final AclService aclService;
    private final UserValidationService userValidationService;
	private final String blacklistedITSystemName;

    /**
	 * By default, the nameId is not replaced. This behavior will change in the future, since
	 * the non-replacing service will be deprecated.
	 */
	private boolean replaceNameId = false;

	/**
	 * Som default er denne feature slået fra. Det er muligt at slå den til vha. konfigurationsfilerne.
	 */
	private boolean allowUserIDCardWithVoces = false;

	public SecurityTokenService(SOSIFactory sosiFactory,
								AclService aclService,
								UserValidationService userValidationService,
								String blacklistedITSystemName) {
		this.sosiFactory = sosiFactory;
		this.aclService = aclService;
        this.userValidationService = userValidationService;
		this.blacklistedITSystemName = blacklistedITSystemName;
	}

	public void setReplaceNameId(final boolean replaceNameId) {
		this.replaceNameId = replaceNameId;
	}

	public boolean isAllowUserIDCardWithVoces() {
		return allowUserIDCardWithVoces;
	}

	public void setAllowUserIDCardWithVoces(boolean allowUserIDCardWithVoces) {
		this.allowUserIDCardWithVoces = allowUserIDCardWithVoces;
	}

	public SecurityTokenResponse issueIdCard(SecurityTokenRequest request) {

		// change value of replaceNameId in case of OCES3
		boolean replaceNameId = shouldReplaceNameId(request.getIDCard());

		// Verify the IDCard of the requestor
		VerifyResult result;
		try {
			result = verify(request.getIDCard());
		} catch (VerifyException e) {
			log.debug("Request not verified: " + e.getFaultCode() + " " + e.getFaultString() + " " + e.getFaultActor());
			return sosiFactory.createNewSecurityTokenErrorResponse(
					request,
					e.getFaultCode(),
					e.getFaultString(),
					e.getFaultActor());
		}
		final IDCard idCard = request.getIDCard();
		// Create the WS-Trust response
		SecurityTokenResponse response = sosiFactory.createNewSecurityTokenResponse(request);

		// Create a new STS signed IDCard from the IDCard of the requestor
		IDCard signedIdCard;
		if (idCard instanceof UserIDCard) {
			UserIDCard userCard = ((UserIDCard)idCard);
			signedIdCard = sosiFactory.copyToVOCESSignedIdCard(userCard, createUserInfo(userCard.getUserInfo(), result), replaceNameId);
		} else {
			signedIdCard = sosiFactory.copyToVOCESSignedIDCard(idCard, replaceNameId);
		}

		// Attach the signed IDCard to WS-Trust response
		response.setIDCard(signedIdCard);
		return response;
	}

	/**
	 * The requested IDCard is first verified according to AuthenticationLevel,
	 * and second against the type of IDCard.
	 *
	 * Refer to requirements: K1
	 *
	 * AuthenticationLevel.MOCES_TRUSTED_USER
	 * <ol>
	 * <li>check revocation of MOCES</li>
	 * <li>verify signature</li>
	 * <li>check user black list</li>
	 * <li>verify/attach cpr from cvr-rid</li>
	 * <li>check authorizationId</li>
	 * </ol>
	 *
	 * AuthenticationLevel.VOCES_TRUSTED_SYSTEM:
	 * <ol>
	 * <li>check revocation of VOCES certificate</li>
	 * <li>verify signature</li>
	 * <li>check system white list</li>
	 * </ol>
	 *
	 * @param idCard the {@link IDCard} to verify
	 * @return a {@link VerifyResult} that never will be <code>null</code>
	 * @throws VerifyException if verification error
	 */
	public VerifyResult verify(IDCard idCard) throws VerifyException {
		verifyITSystemName(idCard);
		verifyNameId(idCard);
		if (AuthenticationLevel.VOCES_TRUSTED_SYSTEM.equals(idCard.getAuthenticationLevel())) {
			return verifyVocesTrustedSystem(idCard);
		} else if (AuthenticationLevel.MOCES_TRUSTED_USER.equals(idCard.getAuthenticationLevel())) {
			return verifyMocesTrustedUser(idCard);
		} else { // unsupported AuthenticationLevel
			throw new VerifyException(StsFaultCodeValues.WST_BAD_REQUEST,
					StsFaultStringValues.WST_BAD_REQUEST + ": Authentication level not supported (" + idCard.getAuthenticationLevel().getLevel() + ")",
					StsFaultActorValues.STS);
		}
	}

	private void verifyITSystemName(IDCard idCard) throws VerifyException {
		if (!(idCard instanceof SystemIDCard)) {
			return;
		}
		SystemInfo systemInfo = ((SystemIDCard) idCard).getSystemInfo();
		if (systemInfo != null && blacklistedITSystemName.equals(systemInfo.getITSystemName())) {
			throw new VerifyException(StsFaultCodeValues.WST_BAD_REQUEST,
					StsFaultStringValues.WST_BAD_REQUEST + ": Invalid ITSystemName",
					StsFaultActorValues.STS);
		}
	}

	private void verifyNameId(final IDCard idCard) throws VerifyException {
		// if it gets replaced anyway, we do not care what it contains
		if (shouldReplaceNameId(idCard)) {
			return;
		}

		if ( ! CertificateInfo.isProbableCertificateInfoString(idCard.getAlternativeIdentifier()) ) {
			return; // leave non-certificateInfoString intact
		}

		// verify that certificate and NameId match
		final CertificateInfo fromNameId = CertificateInfo.fromString(idCard.getAlternativeIdentifier());
		final CertificateInfo fromCert = new CertificateInfo(idCard.getSignedByCertificate());
		if (! fromNameId.toString().equals(fromCert.toString())) {
			throw new VerifyException(StsFaultCodeValues.WST_BAD_REQUEST,
					StsFaultStringValues.WST_BAD_REQUEST + ": IDCard nameId must match information from certificate.",
					StsFaultActorValues.STS);
		}
	}

	// verify IDCard with AuthenticationLevel.VOCES_TRUSTED_SYSTEM, i.e. including FOCES signed
	private VerifyResult verifyVocesTrustedSystem(IDCard idCard) throws VerifyException {
		if (!(idCard instanceof SystemIDCard)) {
			throw new VerifyException(StsFaultCodeValues.WST_BAD_REQUEST,
					StsFaultStringValues.WST_BAD_REQUEST + ": ID-Card is not a System ID-Card",
					StsFaultActorValues.STS);
		}
		verifyIdCard(idCard);

		X509Certificate voces = idCard.getSignedByCertificate();
		OcesInfo ocesInfo = OcesInfoParser.parseCertificate(voces);
		verifyCertificate(voces, true, ocesInfo);
		SystemInfo systemInfo = ((SystemIDCard)idCard).getSystemInfo();
		String systemName = systemInfo.getITSystemName();
		if (!aclService.isWhitelisted(ocesInfo, systemName)) {
			String whiteKey = "CVR:" + ocesInfo.getCvr() + "-SYSTEM:" + systemName;
			log.info("Not whitelisted: " + whiteKey );
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
					StsFaultStringValues.WST_FAILED_AUTHENTICATION +  ":" + " VOCES not in whitelist [" + whiteKey + "]",
					StsFaultActorValues.STS);
		}

		verifySystemInfo(systemInfo, ocesInfo);
		if (idCard instanceof UserIDCard) {
			log.warn("UserIDCard is being signed with VOCES/FOCES from system: " + systemName);
			if(isAllowUserIDCardWithVoces()) {
				/*
				 * Det giver ingen mening at verificere UserInfo fra et Voces Trusted System (AuthenticationLevel 3),
				 * da metoden verifyUserInfo forventer at ID kortet er signeret med MOCES.
				 * Fremadrettet er default for isAllowUserIDCardWithVoces sat til false, så denne kode vil aldrig blive ramt.
				 * isAllowUserIDCardWithVoces kan dog aktiveres ved at sætte propertyen til true i konfigurationesfilen.
				 *
				 * Hvis isAllowUserIDCardWithVoces sættes til true, så kopieres cpr, role og authorizationCode fra UserInfo/UserIDCard direkte over i VerifyResult.
				 */
				UserIDCard userIdCard = (UserIDCard)idCard;
				VerifyResult result = new VerifyResult();
				result.cpr = getCprNr(userIdCard.getUserInfo());
				result.role = userIdCard.getUserInfo().getRole();
				result.authorizationCode = userIdCard.getUserInfo().getAuthorizationCode();
				return result;
			} else {
				throw new VerifyException(StsFaultCodeValues.WST_BAD_REQUEST,
						StsFaultStringValues.WST_BAD_REQUEST +  ": ID-Card is a User ID-Card signed with VOCES/FOCES",
						StsFaultActorValues.STS);
			}
		} else {
			return new VerifyResult();
		}

	}

	void verifyIdCard(IDCard idCard) throws VerifyException {
		if (!idCard.isValidInTime()) {
			throw new VerifyException(StsFaultCodeValues.WST_INVALID_TIME_RANGE,
					StsFaultStringValues.WST_INVALID_TIME_RANGE +  ": IDCard is created after or expires before current system time",
					StsFaultActorValues.STS);
		}
		long seconds = (idCard.getExpiryDate().getTime() - idCard.getCreatedDate().getTime()) / 1000;
		float hours = (((float)seconds) / (60*60));
		if (hours > 24) {
			throw new VerifyException(StsFaultCodeValues.WST_INVALID_TIME_RANGE,
					StsFaultStringValues.WST_INVALID_TIME_RANGE +  ": IDCard have maximum validity of 24 hours + [" + hours +"]",
					StsFaultActorValues.STS);
		}
	}

	private VerifyResult verifyMocesTrustedUser(IDCard idCard) throws VerifyException {
		if (!(idCard instanceof UserIDCard)) {
			throw new VerifyException(StsFaultCodeValues.WST_BAD_REQUEST,
					StsFaultStringValues.WST_BAD_REQUEST + ": ID-Card is not a User ID-Card",
					StsFaultActorValues.STS);
		}
		verifyIdCard(idCard);

		X509Certificate moces = idCard.getSignedByCertificate();
		OcesInfo ocesInfo = OcesInfoParser.parseCertificate(moces);
		verifyCertificate(moces, false, ocesInfo);
		String systemName = ((UserIDCard)(idCard)).getSystemInfo().getITSystemName();

		if (!aclService.isWhitelisted(ocesInfo, systemName)) {
			String whiteKey = "CVR:" + ocesInfo.getCvr() + "-SYSTEM:" + systemName;
			log.info("Not whitelisted: " + whiteKey );
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
					StsFaultStringValues.WST_FAILED_AUTHENTICATION +  ": MOCES not in whitelist [" + whiteKey + "]",
					StsFaultActorValues.STS);
		}

		if (aclService.isBlacklisted(ocesInfo, systemName)) {
			log.info("STS blacklisted: " + ocesInfo.getSubjectSerialNumber().getValue());
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
					StsFaultStringValues.WST_FAILED_AUTHENTICATION +  ": MOCES is blacklisted [" + ocesInfo.getSubjectSerialNumber().getValue() + "]",
					StsFaultActorValues.STS);
		}

		UserIDCard userIdCard = (UserIDCard)idCard;
		VerifyResult result  = verifyUserInfo(userIdCard.getUserInfo(), ocesInfo);
		verifySystemInfo(userIdCard.getSystemInfo(), ocesInfo);
		return result;
	}

	private void verifyCertificate(X509Certificate oces, boolean isSystem, OcesInfo ocesInfo) throws VerifyException {
		if (ocesInfo.getCertificateType() != CertificateType.UNKNOWN) {
			checkValidCertificate(oces);
			checkCertificateType(isSystem, ocesInfo);
			checkIssuer(oces, StsFaultActorValues.STS);
		} else {
			SubjectSerialNumber subjectSerialNumber = ocesInfo.getSubjectSerialNumber();
			log.info("Unsupported certificate type: " + subjectSerialNumber);

			String faultString;
			if(subjectSerialNumber == null || subjectSerialNumber.getValue() == null) {
				faultString = StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": signed with unsupported certificate type. Subject SerialNumber is missing.";
			} else {
				faultString = StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": signed with unsupported certificate type";
			}

			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
					faultString,
					StsFaultActorValues.STS);
		}
	}

	public void checkIssuer(X509Certificate oces, String faultActor) throws VerifyException {
		new CertificateStatusVerifier(sosiFactory.getFederation()).verifyCertificateStatus(oces, faultActor);
	}

	void checkCertificateType(boolean isSystem, OcesInfo ocesInfo) throws VerifyException {
		if ((isSystem && ocesInfo.getCertificateType() == CertificateType.MOCES)
				|| (!isSystem && ocesInfo.getCertificateType() != CertificateType.MOCES)) {
			log.info("Certificate of wrong type");
			String certType = (isSystem ? "VOCES/FOCES" : "MOCES");
			String fault = StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": must be signed with " + certType;
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, fault, StsFaultActorValues.STS);
		}
		if (ocesInfo.getVersion() == Version.OCES3 && ocesInfo.getCertificateType() == CertificateType.MOCES && ocesInfo.getPersistenceLevel() != OcesInfo.PersistenceLevel.GLOBAL) {
			String fault = StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": signed with unsupported certificate - persistence level of moces3 certificates must be global";
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, fault, StsFaultActorValues.STS);
		}
	}

	public void checkValidCertificate(X509Certificate oces) throws VerifyException {
		try {
			oces.checkValidity();
		} catch (CertificateExpiredException e) {
			log.info("Certificate expired: " + oces.getSerialNumber());
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
					StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": certificate expired " + oces.getSerialNumber(),
					StsFaultActorValues.STS);
		} catch (CertificateNotYetValidException e) {
			log.info("Certificate not yet valid: " + oces.getSerialNumber());
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
					StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": certificate not yet valid " + oces.getSerialNumber(),
					StsFaultActorValues.STS);
		}
	}

	void verifySystemInfo(SystemInfo info, OcesInfo ocesInfo) throws VerifyException {
		if (!info.getCareProvider().getID().equals(ocesInfo.getCvr())) {
			log.info("cvr mismatch: " + info.getCareProvider().getID() + " != " + ocesInfo.getCvr());
			throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
					StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": cvr mismatch",
					StsFaultActorValues.STS);
		}
	}

	private VerifyResult verifyUserInfo(UserInfo input, OcesInfo ocesInfo) throws VerifyException {
		// Check cpr and certificate and save for later
		if (ocesInfo.getCertificateType() == CertificateType.MOCES) {
			CprNr inputCpr = getCprNr(input);
			final String inputRole = input.getRole();
			final String inputAuthorizationCode = input.getAuthorizationCode();
			VerifyInput verifyInput = new VerifyInput(ocesInfo, inputCpr, inputRole, inputAuthorizationCode);
			final boolean cprTrustedForSigner = false; // we never trust cpr number
			final String endpoint = "STS"; // WARNING: this name may be used to check if we trust roles - beware, if changing
			return userValidationService.verifyCprAndRole(verifyInput, cprTrustedForSigner, endpoint);
		} else {
			return new VerifyResult();
		}
	}

	static UserInfo createUserInfo(UserInfo template, VerifyResult result) {
        return new UserInfo(
				(result.cpr == null ? template.getCPR() : result.cpr.getValue()),
				template.getGivenName(),
				template.getSurName(),
				template.getEmail(),
				template.getOccupation(),
                result.getRole().orElse(null),
                result.getAuthorizationCode().orElse(null));
	}

	private CprNr getCprNr(UserInfo info) {
		if (info.getCPR() == null || info.getCPR().trim().length() == 0) {
			return null;
		} else {
			return new CprNr(info.getCPR());
		}
	}

	private boolean shouldReplaceNameId(IDCard idCard) {
		X509Certificate moces = idCard.getSignedByCertificate();
		OcesInfo ocesInfo = OcesInfoParser.parseCertificate(moces);

		// For moces3 idcards, the nameID should never be replaced
		if (ocesInfo.getCertificateType() == CertificateType.MOCES && ocesInfo.getVersion() == Version.OCES3) {
			return false;
		} else {
			return replaceNameId;
		}
	}
}
