/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.sts.sla.StsContext;
import org.apache.commons.io.ByteOrderMark;
import org.apache.commons.io.input.BOMInputStream;
import org.apache.log4j.Logger;
import org.springframework.web.servlet.ModelAndView;
import org.springframework.web.servlet.mvc.Controller;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.nio.CharBuffer;

/**
 *
 */
public class ServiceServlet extends HttpServlet implements Controller {

	private final Logger log = Logger.getLogger(ServiceServlet.class);

	private final AbstractStsRequestHandler<?, ?, ?> handler;
	private final String wsdlUrl;

	public ServiceServlet(final AbstractStsRequestHandler<?, ?, ?> handler,
						  final String wsdlUrl) {
		this.handler = handler;
		this.wsdlUrl = wsdlUrl;
	}

	@Override
	protected void doPost(final HttpServletRequest req, final HttpServletResponse resp) throws ServletException, IOException {
		resp.setContentType("text/xml;charset=utf-8");
		try {
			handle(req, resp);
		} catch (Throwable e) {
			log.error("General error:", e);
			resp.setStatus(500);
		}
	}

	public void handle(HttpServletRequest req, HttpServletResponse resp) throws IOException {
		String characterEncoding = req.getCharacterEncoding();
		InputStream inputStream = req.getInputStream();
		Readable readable = getReader(characterEncoding, inputStream);
		final String request = asString(readable);
		final boolean internalError = handler.handleRequest(request, resp.getOutputStream());
		if (internalError) {
			resp.setStatus(500);
		}
	}

	private Readable getReader(String characterEncoding, InputStream inputStream) throws IOException {
		boolean includeBoms = false;
		BOMInputStream bomInputStream = new BOMInputStream(inputStream, includeBoms, ByteOrderMark.UTF_8, ByteOrderMark.UTF_16BE, ByteOrderMark.UTF_16LE);

		String bomCharsetName = bomInputStream.getBOMCharsetName();

		final String effectiveCharset;
		if (characterEncoding != null) {
			effectiveCharset = characterEncoding;
			if (bomCharsetName != null && !bomCharsetName.equalsIgnoreCase(characterEncoding)) {
				log.warn("Character set indicated by HTTP header was " + characterEncoding + " but HTTP body contained a " + bomCharsetName + " byte order marker.");
			}
		} else if (bomCharsetName != null) {
			effectiveCharset = bomCharsetName;
		} else {
			effectiveCharset = ByteOrderMark.UTF_8.getCharsetName();
		}

		return new InputStreamReader(bomInputStream, effectiveCharset);
	}

	private String asString(Readable readable) throws IOException {
		StringBuilder sb = new StringBuilder();
		copy(readable, sb);
		return sb.toString();
	}

	private static final int BUF_SIZE = 0x800; // 2K chars (4K bytes)

	public static long copy(Readable from, Appendable to) throws IOException {
		CharBuffer buf = CharBuffer.allocate(BUF_SIZE);
		long total = 0;
		while (from.read(buf) != -1) {
			buf.flip();
			to.append(buf);
			total += buf.remaining();
			buf.clear();
		}
		return total;
	}

	@Override
	protected void doGet(final HttpServletRequest req, final HttpServletResponse res) throws ServletException, IOException {
		/* The SOSI-GW relies on the URL, /sts/services/SecurityTokenService, returning 200. This is used to
		 * mark that the specific STS is indeed working. Instead of just returning the wsdl, which would then
		 * show up on the SOSI-GW status check, the STS will simply return 'OK'.
		 *
		 * One could consider forwarding to checkstatus, this would although increase the load on that url.
		 */
		if ("wsdl".equalsIgnoreCase(req.getQueryString())) {
			req.getRequestDispatcher(wsdlUrl).forward(req, res);
			return;
		}

		res.setContentType("text/plain;charset=utf-8");
		final PrintWriter writer = res.getWriter();
		writer.println("OK");
		writer.flush();
		writer.close();
	}

	@Override
	public ModelAndView handleRequest(final HttpServletRequest req, final HttpServletResponse resp) throws Exception {
		StsContext.getContext().setRequest(req);
		try {
			super.service(req, resp);
		} finally {
			StsContext.clearContext();
		}
		return null;
	}
}
