/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.SecurityTokenResponse;
import dk.sosi.seal.model.constants.DGWSConstants;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.sts.server.service.Timed;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import dk.sosi.sts.sla.StsContext;
import dk.sosi.sts.util.StsUtil;
import org.apache.axis.Message;
import org.apache.axis.message.SOAPEnvelope;
import org.w3c.dom.Document;

import javax.xml.soap.SOAPException;
import java.io.IOException;
import java.io.OutputStream;
import java.security.KeyStoreException;
import java.security.cert.X509Certificate;

public class StsRequestHandler extends AbstractStsRequestHandler<SecurityTokenRequest, SecurityTokenResponse, Void> {

	private SecurityTokenService sts;

	public StsRequestHandler() {
	}

	public StsRequestHandler(final SOSIFactory sosiFactory,
							 final SecurityTokenService sts,
							 final StsStatus stsStatus,
							 final CertificateStatusChecker service) {
		setSosiFactory(sosiFactory);
		setSts(sts);
		setStsStatus(stsStatus);
		setCrlService(service);
	}

	public SecurityTokenService getSts() {
		return sts;
	}

	public void setSts(final SecurityTokenService sts) {
		this.sts = sts;
	}

	@Override
    protected SecurityTokenRequest deserializeRequest(final String req) {
        SecurityTokenRequest stRequest;
        // Rebuild SecurityTokenRequest from the string representation of the request
		stRequest = sosiFactory.deserializeSecurityTokenRequest(req);
        if (AuthenticationLevel.NO_AUTHENTICATION.equals(stRequest.getIDCard().getAuthenticationLevel())) {
            log.warn("Not signed by certificate. Unsupported authentication level:" + AuthenticationLevel.NO_AUTHENTICATION);
        } else {
            X509Certificate signedByCertificate = stRequest.getIDCard().getSignedByCertificate();
            if (signedByCertificate != null) {
                log.debug("Signed by certificate [" + signedByCertificate.getSerialNumber() + "]:[" + signedByCertificate.getSubjectDN().getName() + "]");
            }
        }
        return stRequest;
    }

	@Override
	protected Void verifyConfiguration(SecurityTokenRequest stRequest) {
		return null;
	}

	@Override
    protected SecurityTokenResponse perform(final SecurityTokenRequest stRequest, Void configType) throws VerifyException {
        StsContext.getContext().setMessageId(stRequest.getMessageID());
		SLALogHelper.createItem(SLALogPoint.STS_ISSUEIDCARD);
		
		SecurityTokenResponse response;
		try {
			response = sts.issueIdCard(stRequest);
			if (!response.isFault()) {
				SLALogHelper.storeCallResultOk(SLALogPoint.STS_ISSUEIDCARD);
			} else {
				SLALogHelper.storeCallResultError(SLALogPoint.STS_ISSUEIDCARD, response.getFaultString());
			}
			return response;
		} catch (Throwable t) {
			SLALogHelper.storeCallResultError(SLALogPoint.STS_ISSUEIDCARD, t.getMessage());
			throw StsUtil.rethrowable(t, VerifyException.class);
		}
    }

    @Override
    protected boolean isFault(final SecurityTokenResponse stResponse) {
        return stResponse.isFault();
    }

    @Override
	@Timed
    protected Document serialize(final SecurityTokenResponse stResponse) {
        return stResponse.serialize2DOMDocument();
    }

	@Override
	@Timed
	protected void serializeToOutputStream(final SecurityTokenResponse stResponse, final OutputStream outputStream) throws IOException {
		final Document responseDoc = stResponse.serialize2DOMDocument();
		final SOAPEnvelope resp = new SOAPEnvelope();
		try {
			AxisUtil.addHeadersToSOAPEnvelope(resp, responseDoc, isFault(stResponse),
											  true);
			AxisUtil.addBodyToSOAPEnvelope(resp, responseDoc);

			new Message(resp, false).writeTo(outputStream);

		} catch (SOAPException se) {
			throw new RuntimeException(se);
		}
	}


	@Override
	@Timed
    protected SecurityTokenResponse createErrorResponse(SecurityTokenRequest stRequest,
	                                                    String faultCode, String faultString, String faultActor) {
        String dgwsVersion = (stRequest != null) ? stRequest.getDGWSVersion() : null;
        String msgId = (stRequest != null) ? stRequest.getMessageID() : null;

        return createErrorResponse(dgwsVersion, msgId, faultCode, faultString, faultActor);
    }

	@Override
	protected String auditName() {
		return "ID-Card";
	}

	@Override
	@Timed
    protected SecurityTokenResponse createErrorResponse(String dgwsVersion, String msgId,
	                                                    String faultCode, String faultString, String faultActor) {
		if (dgwsVersion == null) dgwsVersion = DGWSConstants.VERSION_1_0;
		if (msgId == null) msgId = "";
		SecurityTokenResponse stResponse = sosiFactory.createNewSecurityTokenErrorResponse(
				dgwsVersion, msgId, faultCode, faultString, faultActor);
		return stResponse;
    }
}
