/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import org.springframework.web.context.ServletContextAware;

import javax.servlet.ServletContext;
import java.util.Date;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Expose transaction information since last start and per day, along with
 * version information and when the STS was last started.
 */
public class StsStatus implements ServletContextAware {

	private final static int HOUR_MS = 60 * 60 * 1000;
	private final static int DAY_MS = 24 * HOUR_MS;

	private long transactions;
	private long errors;
	private int maxDuration;
	private long totalDuration;
	private long transactions24h;
	private long errors24h;
	private int maxDuration24h;
	private long totalDuration24h;
	private String startDate;
	private String version;
	private String sealVersion;
	private long latestTransaction;

	private long nextReset;

	private final AtomicLong issued = new AtomicLong(-1);

	public String getAverageTime24h() {
		checkReset24h();
		long avg = 0;
		if (transactions24h != 0) {
			avg = totalDuration24h / transactions24h;
		}
		return avg + " ms";
	}

	public String getAverageTimeForever() {
		long avg = 0;
		if (transactions != 0) {
			avg = totalDuration / transactions;
		}
		return avg + " ms";
	}

	public String getMaxTime24h() {
		checkReset24h();
		return maxDuration24h + " ms";
	}

	public String getMaxTimeForever() {
		return maxDuration + " ms";
	}

	public String getStartedDate() {
		return startDate;
	}

	public int getTransactions24h() {
		checkReset24h();
		return (int) transactions24h;
	}

	public long getTransactionsForever() {
		return transactions;
	}

	public int getErrors24h() {
		checkReset24h();
		return (int) errors24h;
	}

	public long getErrorsForever() {
		return errors;
	}

	public String getVersion() {
		return version;
	}

	public String getSealVersion() {
		return sealVersion;
	}

	public long getLatestTransaction() {
		return latestTransaction;
	}

	/**
	 * Collect duration information for a single transaction
	 *
	 * @param duration in milliseconds
	 * @param success
	 */
	public synchronized void addTransaction(int duration, boolean success) {
		checkReset24h();
		latestTransaction = getTimeInMillis();
		transactions++;
		transactions24h++;
		if (!success) {
			errors++;
			errors24h++;
		}
		totalDuration += duration;
		if (duration > maxDuration) {
			maxDuration = duration;
		}
		totalDuration24h += duration;
		if (duration > maxDuration24h) {
			maxDuration24h = duration;
		}
	}

	protected synchronized void checkReset24h() {
		if (nextReset < getTimeInMillis()) {
			transactions24h = 0;
			errors24h = 0;
			maxDuration24h = 0;
			totalDuration24h = 0;
			nextReset = (getTimeInMillis() / DAY_MS + 1) * DAY_MS;
		}
	}

	protected long getTimeInMillis() {
		return System.currentTimeMillis();
	}

	public StsStatus(Date start, String version, String sealVersion) {
		this.transactions = 0;
		this.totalDuration = 0;
		this.maxDuration = 0;
		this.transactions24h = 0;
		this.totalDuration24h = 0;
		this.maxDuration24h = 0;
		this.startDate = start.toString();
		this.version = version;
		this.sealVersion = sealVersion;
		this.latestTransaction = -1;
		checkReset24h();
	}

	public StsStatus() {
		this(new Date(), "-", "-");
	}

	public void issued() {
		this.issued.set(System.currentTimeMillis());
	}

	public long getLastIssued() {
		return this.issued.get();
	}

	@Override
	public void setServletContext(final ServletContext servletContext) {
		version = servletContext.getInitParameter("sts.version");
		sealVersion = servletContext.getInitParameter("seal.version");
	}
}
