/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sds.nsp.audit.AuditBuilder;
import dk.sosi.seal.model.BasicPrivileges;
import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.model.Authorization;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.RoleDefinition;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.service.*;
import org.apache.log4j.Logger;

import java.text.MessageFormat;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Service for validating cpr and role of a user.
 * Refer to <a href="https://www.nspop.dk/display/public/web/STS+-+Guide+til+anvendere%3A+Medarbejderomvekslinger">STS Medarbejderomvekslinger</a>
 * for a detailed description of the algorithm.
 */
public class UserValidationService {
    private final Logger log = Logger.getLogger(getClass());
    
    private static final String NO_USER_ROLE = "urn:dk:healthcare:no-role";
    private static final String BASIC_PRIVILEGES_CVR_SCOPE_PREFIX = "urn:dk:gov:saml:cvrNumberIdentifier:";
    private static final String UNKNOWN_USER_ROLE = "urn:dk:healthcare:unknown";
    
    private final OcesCvrRidService cvrRidService;
    private final OcesUUID2CPRService uuid2CprService;
    private final AuthorizationService authorizationService;
    private final RoleDefinitionService roleDefinitionService;
    private final NationalRoleService nationalRoleService;
    private final AuthorizationDefinitionService authorizationDefinitionService;

    public UserValidationService(OcesCvrRidService cvrRidService, AuthorizationService authorizationService, RoleDefinitionService roleDefinitionService, NationalRoleService nationalRoleService, AuthorizationDefinitionService authorizationDefinitionService, OcesUUID2CPRService uuid2CprService) {
        this.cvrRidService = cvrRidService;
        this.authorizationService = authorizationService;
        this.roleDefinitionService = roleDefinitionService;
        this.nationalRoleService = nationalRoleService;
        this.authorizationDefinitionService = authorizationDefinitionService;
        this.uuid2CprService = uuid2CprService;
    }

    public VerifyResult verifyCprAndRole(VerifyInput verifyInput, boolean cprTrustedForSigner, String endpoint) throws VerifyException {
        return verifyCprAndRole(verifyInput, cprTrustedForSigner, endpoint, null, false);
    }

    public VerifyResult verifyCprAndRole(VerifyInput verifyInput, boolean cprTrustedForSigner, String endpoint, BasicPrivileges basicPrivileges, boolean replaceEmptyRole) throws VerifyException {
        final CprNr foundCpr;
        if (verifyInput.getInputCpr() != null && cprTrustedForSigner) {
            foundCpr = verifyInput.getInputCpr();
        } else {
            foundCpr = verifyCpr(verifyInput.getInputCpr(), verifyInput.getOcesInfo());
        }
        VerifyResult verifyResult = new VerifyResult();
        verifyResult.cpr = foundCpr;
        verifyRole(foundCpr, verifyInput.getOcesInfo(), verifyInput.getInputRole(), verifyInput.getInputAuthorizationCode(), verifyResult, endpoint, basicPrivileges, replaceEmptyRole);
        
        return verifyResult;
    }


    protected CprNr verifyCpr(final CprNr inputCpr, OcesInfo ocesInfo) throws VerifyException {
        if (inputCpr != null) {
            verifyCprAndOcesInfoRelation(inputCpr, ocesInfo);
            return inputCpr;
        } else {
            return lookupCprFromOcesInfo(ocesInfo);
        }
    }

    private CprNr lookupCprFromOcesInfo(OcesInfo ocesInfo) throws VerifyException {
        if (ocesInfo.getVersion() == OcesInfo.Version.OCES3) {
            return lookupCprFromUuid(ocesInfo);
        } else {
            return lookupCprFromCvrRid(ocesInfo);
        }
    }

    private CprNr lookupCprFromCvrRid(OcesInfo ocesInfo) throws VerifyException {
        SubjectSerialNumber ssn = ocesInfo.getSubjectSerialNumber();
        final CprNr foundCpr;
        try {
            foundCpr = cvrRidService.findRelatedCpr(ocesInfo);
        } catch (StsException e) {
            String detail = "cvrrid-cpr lookup failed [" + ssn.getValue() + "]";
            log.error(detail, e);
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
                    StsFaultStringValues.WST_REQUEST_FAILED + ": " + detail,
                    StsFaultActorValues.CVRRIDCPR);
        }
        if (foundCpr == null) {
            throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                    StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": illegal cvrrid [" + ssn.getValue() + "]",
                    StsFaultActorValues.CVRRIDCPR);
        }
        return foundCpr;
    }

    private CprNr lookupCprFromUuid(OcesInfo ocesInfo) throws VerifyException {
        SubjectSerialNumber ssn = ocesInfo.getSubjectSerialNumber();
        final CprNr foundCpr;
        try {
            foundCpr = uuid2CprService.findRelatedCpr(ocesInfo);
        } catch (StsException e) {
            String detail = "uuid-cpr lookup failed [" + ssn.getValue() + "]";
            log.error(detail, e);
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
                    StsFaultStringValues.WST_REQUEST_FAILED + ": " + detail,
                    StsFaultActorValues.UUIDCPR);
        }
        if (foundCpr == null) {
            throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                    StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": illegal uuid [" + ssn.getValue() + "].",
                    StsFaultActorValues.UUIDCPR);
        }
        return foundCpr;
    }

    private void verifyCprAndOcesInfoRelation(CprNr inputCpr, OcesInfo ocesInfo) throws VerifyException {
        if (ocesInfo.getVersion() == OcesInfo.Version.OCES3) {
            verifyCprAndUuidRelation(inputCpr, ocesInfo);
        } else {
            verifyCprAndCvrRidRelation(inputCpr, ocesInfo);
        }
    }

    private void verifyCprAndUuidRelation(CprNr inputCpr, OcesInfo ocesInfo) throws VerifyException {
        SubjectSerialNumber ssn = ocesInfo.getSubjectSerialNumber();
        final boolean isValid;
        try {
            isValid = uuid2CprService.isRelated(ocesInfo, inputCpr);
        } catch (StsException e) {
            String detail = "uuid-cpr lookup failed [" + ssn.getValue() + "]";
            log.error(detail, e);
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
                    StsFaultStringValues.WST_REQUEST_FAILED + ": " + detail,
                    StsFaultActorValues.UUIDCPR);
        }
        if (!isValid) {
            throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                    StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": uuid-cpr mismatch [" + ssn.getValue() + "," + inputCpr.getValue() + "].",
                    StsFaultActorValues.UUIDCPR);
        }
    }

    private void verifyCprAndCvrRidRelation(CprNr inputCpr, OcesInfo ocesInfo) throws VerifyException {
        SubjectSerialNumber ssn = ocesInfo.getSubjectSerialNumber();
        final boolean isValid;
        try {
            isValid = cvrRidService.isRelated(ocesInfo, inputCpr);
        } catch (StsException e) {
            String detail = "cvrrid-cpr lookup failed [" + ssn.getValue() + "]";
            log.error(detail, e);
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
                    StsFaultStringValues.WST_REQUEST_FAILED + ": " + detail,
                    StsFaultActorValues.CVRRIDCPR);
        }
        if (!isValid) {
            throw new VerifyException(StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                    StsFaultStringValues.WST_FAILED_AUTHENTICATION + ": cvrrid-cpr mismatch [" + ssn.getValue() + "," + inputCpr.getValue() + "]",
                    StsFaultActorValues.CVRRIDCPR);
        }
    }

    protected void verifyRole(CprNr validCpr, OcesInfo ocesInfo, String inputRole, String inputAuthorizationCode, VerifyResult result, String endpoint) throws VerifyException {
        verifyRole(validCpr, ocesInfo, inputRole, inputAuthorizationCode, result, endpoint, null, false);
    }

    private void verifyRole(CprNr validCpr, OcesInfo ocesInfo, String inputRole, String inputAuthorizationCode, VerifyResult result, String endpoint, BasicPrivileges basicPrivileges, boolean replaceEmptyRole) throws VerifyException {

        if (RoleDefinition.isNationalRole(inputRole)) {
            List<String> cvrsWithTrustedRoles = roleDefinitionService.getTrustedRoleClaimers(endpoint, RoleDefinition.extractRoleName(inputRole));
            final String nationalRoleForUser = getMatchingNationalRole(ocesInfo, inputRole, cvrsWithTrustedRoles, basicPrivileges);
            result.role = nationalRoleForUser;
            result.authorizationCode = null;
            AuditHelper.get().addAuditInformation("STS", endpoint, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "nationalRole", nationalRoleForUser);
            return;
        }

        final Authorization inputAuthorization = getInputAuthorization(inputRole, inputAuthorizationCode);
        if (inputAuthorization != null) {
            Authorization authorization = findMatchingAuthorization(validCpr, inputAuthorization);
            result.role = authorization.getRole();
            result.authorizationCode = authorization.getCode();
            AuditHelper.get().addAuditInformation("STS", endpoint, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "authorizationCode", authorization.getCode());
            AuditHelper.get().addAuditInformation("STS", endpoint, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "educationCode", authorization.getRole());
            return;
        }
        
        final Authorization authorization = noRoleOrAuthorizationWanted(inputRole) ? null : findSingleAuthorization(validCpr);
        if (authorization != null) {
            result.role = authorization.getRole();
            result.authorizationCode = authorization.getCode();
            AuditHelper.get().addAuditInformation("STS", endpoint, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "authorizationCode", authorization.getCode());
            AuditHelper.get().addAuditInformation("STS", endpoint, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "educationCode", authorization.getRole());
            return;
        }

        final String nationalRoleForUser = noRoleOrAuthorizationWanted(inputRole) ? null : getSingleNationalRole(ocesInfo, basicPrivileges);
        if (nationalRoleForUser != null) {
            result.role = nationalRoleForUser;
            result.authorizationCode = null;
            AuditHelper.get().addAuditInformation("STS", endpoint, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "nationalRole", nationalRoleForUser);
            return;
        }

        if (inputRole != null && !"".equals(inputRole.trim())) {
            AuditHelper.get().addAuditInformation("STS", endpoint, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "claimedRole", inputRole);
        } else if (replaceEmptyRole){
            inputRole = UNKNOWN_USER_ROLE;
            AuditHelper.get().addAuditInformation("STS", endpoint, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "replacedRole", UNKNOWN_USER_ROLE);
        }

        result.role = inputRole;
        result.authorizationCode = null;
    }
    
    private static boolean noRoleOrAuthorizationWanted(String inputRole) {
       return NO_USER_ROLE.equals(inputRole);
    }

    private Authorization getInputAuthorization(String role, String authorizationCode) {
        // find authorizations matching ID-card attributes
        // See https://jira.nspop.dk/browse/SDS-2633 for use of authorizationDefinitionService
        String matchEducationCode = authorizationDefinitionService.isValidRole(role) ? role : null;
        // Seal.NET (<2.0.0.7) special case: treat white space(s) like null
        String matchAuthCode = authorizationCode != null && !authorizationCode.trim().isEmpty() ? authorizationCode : null;
        return matchEducationCode != null || matchAuthCode != null ? new Authorization(matchAuthCode, matchEducationCode) : null;
    } 

    private Authorization findMatchingAuthorization(CprNr cpr, Authorization inputAuthorization) throws VerifyException {
        // get authorizations for cpr
        List<Authorization> auts = queryAuthorizations(cpr);
        Set<Authorization> matches = getMatchingAuthorizations(auts, inputAuthorization);
        // handle matching authorizations
        if (matches.isEmpty()) {
            throw auts.isEmpty() ? VerifyException.createMissingAuthorization() : VerifyException.createAuthorizationNotMatched(auts);
        } else if (matches.size() == 1) {
            return matches.iterator().next();
        } else {
            throw VerifyException.createAuthorizationMultipleFound(matches);
        }
    }

    private Set<Authorization> getMatchingAuthorizations(List<Authorization> auts, Authorization inputAuthorization) {
        Stream<Authorization> authsStream = auts.stream();
        if (inputAuthorization.getCode() !=null) {
            authsStream = authsStream.filter(a -> inputAuthorization.getCode().equals(a.getCode()));
        }
        if (inputAuthorization.getRole() != null) {
            authsStream = authsStream.filter(a -> inputAuthorization.getRole().equals(a.getRole()));
        }
        return authsStream.collect(Collectors.toSet());
    }

    private Authorization findSingleAuthorization(CprNr cpr) throws VerifyException {
        // get authorizations for cpr
        List<Authorization> auts = queryAuthorizations(cpr);
        // handle matching authorizations
        if (!auts.isEmpty()) {
            if (auts.size() == 1) {
                return auts.iterator().next();
            } else {
                throw VerifyException.createAuthorizationMultipleFound(auts);
            }
        }
        // ignore invalid aut attributes
        return null;
    }

    private List<Authorization> queryAuthorizations(CprNr cpr) throws VerifyException {
        List<Authorization> auts;
        try {
            auts = authorizationService.getAuthorizations(cpr);
        } catch (StsException e) {
            String detail = "authorization lookup failed [" + cpr.getValue() + "]";
            log.error(detail, e);
            throw new VerifyException(StsFaultCodeValues.WST_REQUEST_FAILED,
                    StsFaultStringValues.WST_REQUEST_FAILED + ": " + detail,
                    StsFaultActorValues.AUTHORIZATION);
        }
        return auts;
    }

    private String getMatchingNationalRole(OcesInfo ocesInfo, String inputRole, List<String> cvrWithTrustedRoles, BasicPrivileges basicPrivileges) throws VerifyException {
        RoleDefinition matchedRoleDefinition = getMatchingRoleDefinition(inputRole);
        if (cvrWithTrustedRoles.contains(ocesInfo.getCvr())) {
            return matchedRoleDefinition.getPublicName();
        }
        List<String> privilegeRoles = getBasicPrivilegesRoles(ocesInfo, basicPrivileges);
        if (privilegeRoles.contains(matchedRoleDefinition.getPublicName()) || privilegeRoles.contains(matchedRoleDefinition.getPublicNameNoCode())) {
            return matchedRoleDefinition.getPublicName();
        }
        final List<String> externalRoles = nationalRoleService.getRoles(ocesInfo);
        if (externalRoles.contains(matchedRoleDefinition.getExternalName())) {
            return matchedRoleDefinition.getPublicName();
        }
        throw VerifyException.failedAuth(MessageFormat.format(VerifyException.ROLE_MISSING, inputRole));
    }

    private List<String> getBasicPrivilegesRoles(OcesInfo ocesInfo, BasicPrivileges basicPrivileges) {
        return basicPrivileges != null && ocesInfo != null ? basicPrivileges.getPrivileges(BASIC_PRIVILEGES_CVR_SCOPE_PREFIX + ocesInfo.getCvr()) : Collections.emptyList();
    }

    private RoleDefinition getMatchingRoleDefinition(String inputRole) throws VerifyException {
        final List<RoleDefinition> configuredRoleDefinitions = roleDefinitionService.getRoleDefinitions();
        RoleDefinition matchedRoleDefinition = null;
        for (RoleDefinition roleDefinition : configuredRoleDefinitions) {
            if (roleDefinition.getPublicName().equals(inputRole) || roleDefinition.getPublicNameNoCode().equals(inputRole)) {
                matchedRoleDefinition = roleDefinition;
                break;
            }
        }
        if (matchedRoleDefinition == null) {
            throw VerifyException.failedAuth(MessageFormat.format(VerifyException.ROLE_INVALID, inputRole));
        }
        return matchedRoleDefinition;
    }

    private String getSingleNationalRole(OcesInfo ocesInfo, BasicPrivileges basicPrivileges) throws VerifyException {
        final String nationalRoleForUser;
        final List<String> matchedRoles = getNationalRolesForUser(ocesInfo, basicPrivileges);
        if (!matchedRoles.isEmpty()) {
            if (matchedRoles.size() == 1) {
                nationalRoleForUser = matchedRoles.get(0);
            } else {
                throw VerifyException.failedAuth(MessageFormat.format(VerifyException.ROLE_MULTIPLE_FOUND, matchedRoles));
            }
        } else {
            nationalRoleForUser = null;
        }
        return nationalRoleForUser;
    }

    private List<String> getNationalRolesForUser(OcesInfo ocesInfo, BasicPrivileges basicPrivileges) {
        final List<RoleDefinition> configuredRoleDefinitions = roleDefinitionService.getRoleDefinitions();
        final List<String> privilegeRoles = getBasicPrivilegesRoles(ocesInfo, basicPrivileges);
        final List<String> externalRoles = nationalRoleService.getRoles(ocesInfo);
        return configuredRoleDefinitions.stream()
                .filter(role -> privilegeRoles.contains(role.getPublicName()) || externalRoles.contains(role.getExternalName()))
                .map(RoleDefinition::getPublicName)
                .collect(Collectors.toList());
    }
}