/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.aop;

import net.sf.cglib.proxy.Callback;
import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.MethodProxy;
import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.springframework.beans.factory.config.AbstractFactoryBean;

import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

/**
 *
 */
public class StsProxyFactory<T> extends AbstractFactoryBean<T> {

	private Class<?> proxyClass;
	private Object[] arguments;

	public static final ConcurrentHashMap<Class<?>, MethodInterceptor[]> globalInterceptors = new ConcurrentHashMap<Class<?>, MethodInterceptor[]>();

	@Override
	public Class<?> getObjectType() {
		return proxyClass;
	}

	@Override
	@SuppressWarnings("unchecked")
	protected T createInstance() throws Exception {
		final Enhancer enhancer = new Enhancer();
		enhancer.setClassLoader(proxyClass.getClassLoader());
		enhancer.setSuperclass(proxyClass);
		enhancer.setCallbacks(new Callback[]{new MultiMethodInterceptorAdapter()});
		enhancer.setInterfaces(new Class[] { StsProxy.class });
		if (arguments != null && arguments.length > 0) {
			final Class<?>[] argumentTypes = findArgumentTypes();
			return (T) enhancer.create(argumentTypes, arguments);
		} else
			return (T) enhancer.create();
	}

	private Class<?>[] findArgumentTypes() {
		final Constructor<?>[] constructors = proxyClass.getConstructors();
		for (Constructor<?> constructor : constructors) {
			final Class<?>[] parameterTypes = constructor.getParameterTypes();
			if (parameterTypes.length == arguments.length) {
				boolean allMatch = true;
				for (int i = 0; i < parameterTypes.length; i++) {
					final Class<?> parameterType = parameterTypes[i];
					if (arguments[i] != null && !parameterType.isInstance(arguments[i])) {
						allMatch = false;
						break;
					}
				}
				if (allMatch) {
					return parameterTypes;
				}
			}
		}

		final List<Class<?>> actuals = new ArrayList<Class<?>>();
		for (Object argument : arguments) {
			actuals.add(argument.getClass());
		}

		throw new UnsupportedOperationException("No matching constructor found for " + proxyClass + ": " + actuals);
	}

	public Class<?> getProxyClass() {
		return proxyClass;
	}

	public void setProxyClass(final Class<?> proxyClass) {
		this.proxyClass = proxyClass;
	}

	public Object[] getArguments() {
		return arguments;
	}

	public void setArguments(final Object[] arguments) {
		this.arguments = arguments;
	}

	public static class MultiMethodInterceptorAdapter implements net.sf.cglib.proxy.MethodInterceptor {

		private MethodInterceptor[] getInterceptors(final Class<?> clazz) {
			MethodInterceptor[] methodInterceptors = globalInterceptors.get(clazz);
			if (methodInterceptors == null) {
				return new MethodInterceptor[] {};
			} else {
				return methodInterceptors;
			}
		}

		@Override
		public Object intercept(final Object o, final Method method, final Object[] objects, final MethodProxy methodProxy) throws Throwable {
			final MethodInterceptor[] interceptors = getInterceptors(o.getClass().getSuperclass());

			final MethodInvocation chain = new MethodInvocation() {
				int index = 0;

				MethodInterceptor nextInterceptor() {
					return interceptors[index++];
				}

				boolean ifLast() {
					return index >= interceptors.length;
				}

				@Override
				public Method getMethod() {
					return method;
				}

				@Override
				public Object[] getArguments() {
					return objects;
				}

				@Override
				public Object proceed() throws Throwable {
					if (ifLast())
						return methodProxy.invokeSuper(o, objects);
					else
						return nextInterceptor().invoke(this);
				}

				@Override
				public Object getThis() {
					return o;
				}

				@Override
				public AccessibleObject getStaticPart() {
					return null;
				}
			};
			return chain.proceed();
		}
	}
}


