/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server.cache;

import java.util.concurrent.atomic.AtomicReference;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.InitializingBean;

public abstract class CachingService<SERVICE, CACHED_OBJ> implements DefinitionServiceCache, InitializingBean {
	protected final Logger log = Logger.getLogger(this.getClass());

	private final SERVICE backend;
	private final Object cacheRefreshLock = new Object();
	private final AtomicReference<CACHED_OBJ> cache = new AtomicReference<>();

	public CachingService(SERVICE backend) {
		this.backend = backend;
	}

	@Override
	public final void afterPropertiesSet() throws Exception {
		refresh();
	}

	@Override
	public final void refresh() {
		synchronized (cacheRefreshLock) {
			try {
				log.info("Updating the "+getCacheId()+" Cache");
				CACHED_OBJ obj = getUncachedValue(backend);
				setCachedValue(obj);
			} catch (RuntimeException e) {
				log.error("An error occured when reading from the "+getCacheId()+" tables.", e);
			}
		}
	}

	protected final CACHED_OBJ ensureCache() {
		CACHED_OBJ cached = getCachedValue();
		if (cached == null) {
			synchronized (cacheRefreshLock) {
				cached = getCachedValue();
				if (cached == null) {
					refresh();
					cached = getCachedValue();
				}
			}
		}
		return cached;
	}

	protected abstract CACHED_OBJ getUncachedValue(SERVICE backend);

	private void setCachedValue(CACHED_OBJ obj) {
		this.cache.set(obj);
	}

	private CACHED_OBJ getCachedValue() {
		return this.cache.get();
	}
}
