package dk.sosi.sts.server.cache;

import dk.sosi.seal.pki.CertificateCache;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.log4j.Logger;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;

import javax.sql.DataSource;
import java.security.cert.X509Certificate;
import java.util.Optional;

public class CraDbCertificateCache implements CertificateCache {
    private static final Logger LOGGER = Logger.getLogger(CraDbCertificateCache.class);

    private final NamedParameterJdbcTemplate template;
    private String selectQuery = "SELECT certificate FROM caintermediate WHERE url=:url";
    private Optional<CertificateCache> internalCache = Optional.empty();
    public CraDbCertificateCache(DataSource cradb) {
        this.template = new NamedParameterJdbcTemplate(cradb);
    }

    /**
     * Angiv en intern cache, hvor værdier skal gemmes. Værdien er optional.
     * @param cache
     */
    public void setInternalCache(CertificateCache cache) {
        internalCache = Optional.ofNullable(cache);
        LOGGER.debug("Setting cache handler to "+internalCache+" for internal cache.");

    }

    /**
     * Angiv en SQL query, som skal bruges til at hente certifikater fra CRADB.
     * @param selectQuery
     */
    public void setSelectQuery(String selectQuery) {
            this.selectQuery = selectQuery;
    }

    @Override
    /**
     * Indsæt X509Certificate i en intern cache
     */
    public void putCertificate(final Category category, final String s, final X509Certificate x509Certificate) throws PKIException {
        LOGGER.debug("Putting certificate "+s+" into HashMapCertificateCache with Category "+category);
        internalCache.ifPresent(c -> c.putCertificate(category, s, x509Certificate));
    }

    @Override
    /**
     * Hent X509Certificate fra HashMapCertificateCache, hvis det ikke findes der, så hentes det fra CRADB.
     * @return X509Certificate
     */
    public X509Certificate getCertificate(Category category, String s) throws PKIException {
        LOGGER.debug("Getting certificate "+s+" from HashMapCertificateCache with Category "+category);
        X509Certificate certificate = internalCache.map(c -> c.getCertificate(category, s)).orElse(null);

        if(certificate == null && category == Category.IntermediateCert){
            LOGGER.debug("Could not find IntermediateCert "+s+" - need to lookup in CRADB!");
            certificate = getIntermediateCertificateFromCra(s);
            if(certificate != null){
                putCertificate(category, s, certificate);
            } else {
                LOGGER.error("IntermediateCert "+s+" not found in CRADB!");
            }
        }

        return certificate;
    }

    /**
     * Hent Intermediate X509Certificate fra CRADB
     * @param url
     * @return
     */
     X509Certificate getIntermediateCertificateFromCra(String url) {
        LOGGER.debug("Getting IntermediateCert "+url+" from CRADB using "+selectQuery);
        final SqlParameterSource params = new MapSqlParameterSource("url", url);

        return template.query(
                selectQuery,
                params,
                (rs, rowNum) -> decode(rs.getString("certificate"))
        ).stream().findFirst().orElse(null);
    }

    /**
     * Konverter et base64 encodet tekst til et X509Certificate objekt
     * @param encodedCertificate
     * @return
     */
    public static X509Certificate decode(String encodedCertificate){
        if(encodedCertificate == null){
            return null;
        }
        byte[] certificateAsBytes = XmlUtil.fromBase64(encodedCertificate);
        return CertificateParser.asCertificate(certificateAsBytes);
    }
}
