/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.cert.X509Certificate;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicReference;

import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.DERIA5String;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.x509.CRLDistPoint;
import org.bouncycastle.asn1.x509.DistributionPoint;
import org.bouncycastle.asn1.x509.DistributionPointName;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;
import org.springframework.beans.factory.InitializingBean;

import dk.sosi.seal.pki.CRLCache.CRLInfo;
import dk.sosi.seal.pki.CRLCertificateStatusChecker;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.pki.PKIException;

/**
 * 
 * @author Jacob Qvortrup
 */
public class CraOcesCrlServiceCacheImpl implements CertificateStatusChecker, StsOcesCrlService, InitializingBean {

	private final Logger log = Logger.getLogger(CraOcesCrlServiceCacheImpl.class);

	private final AtomicReference<Cache> cache = new AtomicReference<>();

	private final Object cacheRefreshLock = new Object();

	/**
	 * The datasource containing the CRA tables.
	 */
	private DataSource dataSource;

	/**
	 * The SQL used to fetch the known CRL URLs.
	 */
	private String crlsSQL = "SELECT url, lastmodified, nextupdate FROM crl";

	/**
	 * The SQL used to fetch the revoked certificates.
	 */
	private String revokedSQL = "SELECT c.url as url, r.serialnumber as serialnumber FROM revoked r INNER JOIN crl c ON (r.crlid=c.id)";

	/**
	 * The strictness of the check.
	 */
	private boolean strict = true;

	/**
	 * Time to live for a CRL. A CRL must satisfy: CRL.getNextUpdate() + ttl > now before it can be used.
	 * 
	 * ttl == {@link CRLCertificateStatusChecker#NEVER} means that the check is disregarded.
	 */
	private long ttl;

	private PreparedStatement crlsStatement;

	private PreparedStatement revokedCertificatesStatement;

	public CraOcesCrlServiceCacheImpl() {
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public CertificateStatus getRevocationStatus(X509Certificate certificate) throws PKIException {
		ensureCache();

		if (certificate == null) {
			throw new PKIException("Certificate must not be null");
		}

		String url = getCRLUrlFromCertificate(certificate);
		if (url == null) {
			if (strict) {
				throw new PKIException("Certificate must have a CRL url");
			} else {
				return new CertificateStatus(true, CRLCertificateStatusChecker.INVALID_TIMESTAMP);
			}
		}

		BigInteger serialNumber = certificate.getSerialNumber();
		if (serialNumber == null) {
			throw new PKIException("Certificate must have a serial number");
		}

		Long lastModified = getLastModified(url);
		Long nextUpdate = getNextUpdate(url);
		if (lastModified == null) {
			if (strict) {
				/*
				 * The CRA have no data for the CRL URL so we throw an exception.
				 */
				log.error("Certificate has an unknown CRL url: " + url);
				throw new PKIException("Certificate has an unknown CRL url: " + url);
			} else {
				/*
				 * In a non-strict mode we validate all unknown certificates.
				 */
				return new CertificateStatus(true, CRLCertificateStatusChecker.INVALID_TIMESTAMP);
			}
		} else if (isRevoked(url, serialNumber)) {
			/*
			 * Either the specifik certificate is revoked or all certifictes with this CRL URL are revoked.
			 */
			return new CertificateStatus(false, new Date(lastModified));
		} else if (ttl != CRLCertificateStatusChecker.NEVER && nextUpdate + TimeUnit.MINUTES.toMillis(ttl) < System.currentTimeMillis()) {
			if (strict) {
				/*
				 * The CRA tables have not been updated in a timely fashion so we throw an exception.
				 */
				throw new PKIException("Certificate Revocation Authority not up to date");
			} else {
				/*
				 * In a non-strict mode we ignore that CRA has old data.
				 */
				return new CertificateStatus(true, new Date(lastModified));
			}
		} else {
			/*
			 * The certificate is valid.
			 */
			return new CertificateStatus(true, new Date(lastModified));
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public long getLastSuccessful() {
		long lastSuccessful = -1;
		for (long lastModified : cache.get().lastModified.values()) {
			lastSuccessful = Math.max(lastSuccessful, lastModified);
		}
		return lastSuccessful;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public long getLastFailed() {
		return -1;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public CRLInfo loadFromURL(String url, CRLInfo crlInfo) throws IOException {
		ensureCache();
		if (cache.get().lastModified.get(url) == null) {
			log.warn("The CRA data source does not contain the CRL URL " + url);
		}
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void refresh() {
		synchronized (cacheRefreshLock) {
			try {
				refreshCertificateRevocationList();
			} catch (SQLException e) {
				log.error("An SQL error occured when reading from the CRA tables.", e);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void afterPropertiesSet() throws Exception {
		crlsStatement = dataSource.getConnection().prepareStatement(crlsSQL);
		revokedCertificatesStatement = dataSource.getConnection().prepareStatement(revokedSQL);
	}

	private void ensureCache() {
		if (cache.get() == null) {
			synchronized (cacheRefreshLock) {
				if (cache.get() == null) {
					try {
						refreshCertificateRevocationList();
					} catch (SQLException e) {
						log.error("An SQL error occured when reading from the CRA tables.", e);
					}
				}
			}
		}
	}

	private void refreshCertificateRevocationList() throws SQLException {
		log.info("Updating the Certificate Revocation Cache");
		int certificateRevocationsLists = 0;
		Cache cache = new Cache();
		for (ResultSet resultSet = crlsStatement.executeQuery(); resultSet.next();) {
			certificateRevocationsLists++;
			String url = resultSet.getString("url");
			cache.lastModified.put(url, resultSet.getLong("lastmodified"));
			cache.nextUpdate.put(url, resultSet.getLong("nextupdate"));
		}
		log.info("Found " + certificateRevocationsLists + " certificate revocation lists");
		int revokedCertificates = 0;
		for (ResultSet resultSet = revokedCertificatesStatement.executeQuery(); resultSet.next();) {
			revokedCertificates++;
			String url = resultSet.getString("url");
			String encodedSerialNumber = resultSet.getString("serialnumber");
			BigInteger serialnumber;
			if (encodedSerialNumber != null) {
				serialnumber = new BigInteger(encodedSerialNumber);
			} else {
				serialnumber = null; // meaning all certificates with this url is revoked.
			}
			Set<BigInteger> serialnumbers = cache.revoked.get(url);
			if (serialnumbers == null) {
				serialnumbers = new HashSet<BigInteger>();
				cache.revoked.put(url, serialnumbers);
			}
			serialnumbers.add(serialnumber);
		}


		this.cache.set(cache);
		log.info("Found " + revokedCertificates + " revoked certificates");
	}

	private Long getLastModified(String crlUrl) {
		return this.cache.get().lastModified.get(crlUrl);
	}

	private Long getNextUpdate(String crlUrl) {
		return this.cache.get().nextUpdate.get(crlUrl);
	}

	private boolean isRevoked(String crlUrl, BigInteger serialNumber) {
		Set<BigInteger> serialNumbers = cache.get().revoked.get(crlUrl);
		return serialNumbers != null && (serialNumbers.contains(serialNumber) || serialNumbers.contains(null));
	}

	public void setDataSource(DataSource dataSource) {
		this.dataSource = dataSource;
	}

	public void setCrlsSQL(String crlsSQL) {
		this.crlsSQL = crlsSQL;
	}

	public void setRevokedSQL(String revokedSQL) {
		this.revokedSQL = revokedSQL;
	}

	public void setStrict(boolean strict) {
		this.strict = strict;
	}

	public void setTtl(long ttl) {
		this.ttl = ttl;
	}

	protected static String getCRLUrlFromCertificate(final X509Certificate certificate) {
		final byte[] extensionValue = certificate.getExtensionValue(Extension.cRLDistributionPoints.getId());

		// In case no such extension exists
		if (extensionValue == null) {
			return null;
		}

		final DEROctetString oct;
		final ASN1Sequence seq;
		try {
			oct = (DEROctetString) new ASN1InputStream(new ByteArrayInputStream(extensionValue)).readObject();
			seq = (ASN1Sequence) new ASN1InputStream(oct.getOctets()).readObject();
		} catch (IOException e) {
			return null;
		}

		// Special error handling if no full_name was found or if no URI was
		// found
		final CRLDistPoint distPoint = CRLDistPoint.getInstance(seq);
		for (final DistributionPoint distributionPoint : distPoint.getDistributionPoints()) {

			if (distributionPoint.getDistributionPoint().getType() == DistributionPointName.FULL_NAME) {

				for (final GeneralName name : ((GeneralNames) distributionPoint.getDistributionPoint().getName()).getNames()) {
					if (name.getTagNo() == GeneralName.uniformResourceIdentifier) {
						return DERIA5String.getInstance(name.getName().toASN1Primitive()).getString();
					}
				}

			}
		}

		return null;
	}

	private static class Cache {

		private final Map<String, Long> lastModified;

		private final Map<String, Long> nextUpdate;

		private final Map<String, Set<BigInteger>> revoked;

		public Cache() {
			this.lastModified = new HashMap<String, Long>();
			this.nextUpdate = new HashMap<String, Long>();
			this.revoked = new HashMap<String, Set<BigInteger>>();
		}
	}
}
