/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.cert.X509Certificate;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.concurrent.TimeUnit;

import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.DERIA5String;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.x509.CRLDistPoint;
import org.bouncycastle.asn1.x509.DistributionPoint;
import org.bouncycastle.asn1.x509.DistributionPointName;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;

import dk.sosi.seal.pki.CRLCache.CRLInfo;
import dk.sosi.seal.pki.CRLCertificateStatusChecker;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.pki.PKIException;

/**
 * 
 * @author Jacob Qvortrup
 */
public class CraOcesCrlServiceDbImpl implements CertificateStatusChecker, StsOcesCrlService {

	private final Logger log = Logger.getLogger(CraOcesCrlServiceDbImpl.class);

	private final ThreadLocal<PreparedStatement> revokedCertificatesStatements = new ThreadLocal<>();

	/**
	 * The datasource containing the CRA tables.
	 */
	private DataSource dataSource;

	/**
	 * The SQL used to check for revoked certificates.
	 */
	private String revokedSQL = "SELECT c.lastmodified, c.nextupdate, r.added FROM crl c LEFT JOIN revoked r ON (r.crlid = c.id and (r.serialnumber = ? or r.serialnumber is null)) WHERE c.url = ?";

	/**
	 * The strictness of the check.
	 */
	private boolean strict = true;

	/**
	 * Time to live for a CRL. A CRL must satisfy: CRL.getNextUpdate() + ttl > now before it can be used.
	 * 
	 * ttl == {@link CRLCertificateStatusChecker#NEVER} means that the check is disregarded.
	 */
	private long ttl;

	private Long lastSuccessful = -1L;

	public CraOcesCrlServiceDbImpl() {
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public CertificateStatus getRevocationStatus(X509Certificate certificate) throws PKIException {
		if (certificate == null) {
			throw new PKIException("Certificate must not be null");
		}

		String url = getCRLUrlFromCertificate(certificate);
		if (url == null) {
			if (strict) {
				throw new PKIException("Certificate must have a CRL url");
			} else {
				return new CertificateStatus(true, CRLCertificateStatusChecker.INVALID_TIMESTAMP);
			}
		}

		BigInteger serialNumber = certificate.getSerialNumber();
		if (serialNumber == null) {
			throw new PKIException("Certificate must have a serial number");
		}

		long lastModified = 0;
		long nextUpdate = 0;
		long added = 0;
		try {
			PreparedStatement revokedCertificatesStatement = getRevokedCertificatesStatement();
			revokedCertificatesStatement.setString(1, serialNumber.toString());
			revokedCertificatesStatement.setString(2, url);
			for (ResultSet resultSet = revokedCertificatesStatement.executeQuery(); resultSet.next();) {
				if (lastModified != 0) {
					throw new PKIException("More than one revocation of the serial number " + serialNumber + " was detected");
				}
				lastModified = resultSet.getLong(1);
				nextUpdate = resultSet.getLong(2);
				added = resultSet.getLong(3);
			}
		} catch (SQLException e) {
			log.error("An error occured reading from the CRA data source");
			throw new PKIException("Could not verify certificate against the Certificate Revocation Authority");
		}

		if (lastModified != 0) {
			synchronized (lastSuccessful) {
				lastSuccessful = Math.max(lastSuccessful, lastModified);
			}
		}
		
		if (lastModified == 0) {
			if (strict) {
				/*
				 * The CRA have no data for the CRL URL so we throw an exception.
				 */
				log.error("Certificate has an unknown CRL url: " + url);
				throw new PKIException("Certificate has an unknown CRL url: " + url);
			} else {
				/*
				 * In a non-strict mode we validate all unknown certificates.
				 */
				return new CertificateStatus(true, CRLCertificateStatusChecker.INVALID_TIMESTAMP);
			}
		} else if (added != 0) {
			/*
			 * Either the specifik certificate is revoked or all certifictes with this CRL URL are revoked.
			 */
			return new CertificateStatus(false, new Date(lastModified));
		} else if (ttl != CRLCertificateStatusChecker.NEVER && nextUpdate + TimeUnit.MINUTES.toMillis(ttl) < System.currentTimeMillis()) {
			if (strict) {
				/*
				 * The CRA tables have not been updated in a timely fashion so we throw an exception.
				 */
				throw new PKIException("Certificate Revocation Authority not up to date");
			} else {
				/*
				 * In a non-strict mode we ignore that CRA has old data.
				 */
				return new CertificateStatus(true, new Date(lastModified));
			}
		} else {
			/*
			 * The certificate is valid.
			 */
			return new CertificateStatus(true, new Date(lastModified));
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public long getLastSuccessful() {
		return lastSuccessful;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public long getLastFailed() {
		return -1;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public CRLInfo loadFromURL(String url, CRLInfo crlInfo) throws IOException {
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void refresh() {
	}

	public void setDataSource(DataSource dataSource) {
		this.dataSource = dataSource;
	}

	public void setRevokedSQL(String revokedSQL) {
		this.revokedSQL = revokedSQL;
	}

	public void setStrict(boolean strict) {
		this.strict = strict;
	}

	public void setTtl(long ttl) {
		this.ttl = ttl;
	}

	private PreparedStatement getRevokedCertificatesStatement() throws SQLException {
		PreparedStatement revokedCertificatesStatement = revokedCertificatesStatements.get();
		if (revokedCertificatesStatement == null) {
			revokedCertificatesStatement = dataSource.getConnection().prepareStatement(revokedSQL);
			revokedCertificatesStatements.set(revokedCertificatesStatement);
		}
		return revokedCertificatesStatement;
	}

	protected static String getCRLUrlFromCertificate(final X509Certificate certificate) {
		final byte[] extensionValue = certificate.getExtensionValue(Extension.cRLDistributionPoints.getId());

		// In case no such extension exists
		if (extensionValue == null) {
			return null;
		}

		final DEROctetString oct;
		final ASN1Sequence seq;
		try {
			oct = (DEROctetString) new ASN1InputStream(new ByteArrayInputStream(extensionValue)).readObject();
			seq = (ASN1Sequence) new ASN1InputStream(oct.getOctets()).readObject();
		} catch (IOException e) {
			return null;
		}

		// Special error handling if no full_name was found or if no URI was
		// found
		final CRLDistPoint distPoint = CRLDistPoint.getInstance(seq);
		for (final DistributionPoint distributionPoint : distPoint.getDistributionPoints()) {

			if (distributionPoint.getDistributionPoint().getType() == DistributionPointName.FULL_NAME) {

				for (final GeneralName name : ((GeneralNames) distributionPoint.getDistributionPoint().getName()).getNames()) {
					if (name.getTagNo() == GeneralName.uniformResourceIdentifier) {
						return DERIA5String.getInstance(name.getName().toASN1Primitive()).getString();
					}
				}

			}
		}

		return null;
	}
}
