/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service;

import java.io.IOException;
import java.security.Provider;
import java.security.Security;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicLong;

import org.apache.log4j.Logger;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.springframework.beans.factory.InitializingBean;

import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.seal.pki.CRLCache;
import dk.sosi.seal.pki.CertificateResolver;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.pki.InMemoryCRLCache;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;
import dk.sosi.sts.timing.Timer;
import dk.sosi.sts.util.StsUtil;


public class StsOcesCrlServiceImpl implements CertificateStatusChecker, StsOcesCrlService, InitializingBean {

	private Logger log = Logger.getLogger(StsOcesCrlServiceImpl.class);

	private final AtomicLong lastSuccessful = new AtomicLong(-1);
	private final AtomicLong lastFailed = new AtomicLong(-1);

	private CRLCache crlCache;
	private StsCertificateStatusChecker certificateStatusChecker;
	private CertificateResolver certificateResolver;
	private Properties checkerProperties;

	private int interval;
	private boolean strict = true;
	private int crl_ttl = StsCertificateStatusChecker.NEVER;
	private int readTimeout;
	private int connectTimeout;

	class StsCertificateStatusChecker extends CRLCertificateStatusCheckerCopy {

		/**
		 * Creates a new <code>CRLCertificateStatusChecker</code> with the supplied
		 * cache, interval and strictness.
		 *
		 * @param cache               the used cache.
		 * @param interval            the interval in seconds, see NEVER and ALWAYS for special values.
		 * @param strict              see class javadoc for semantic.
		 * @param ttl                 specifies a CRL's extra time to live
		 * @param certificateResolver the resolver for finding the issuer certificate used for CRL verification
		 */
		public StsCertificateStatusChecker(final CRLCache cache, final int interval, final boolean strict, final int ttl, final CertificateResolver certificateResolver) {
			super(cache, interval, strict, ttl, certificateResolver);
		}
		
		@Override
		protected Provider getCertificateFactoryProvider() {
			// use the BouncyCastle provider if installed
			Provider result = Security.getProvider(BouncyCastleProvider.PROVIDER_NAME);
			if (result != null) {
				log.debug("Using Bouncy Castle provider for CRL verification");
			} else {
				log.warn("Not present: No Bouncy Castle provider for CRL verification");
			}
			return result;
		}

		@Override
		protected CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
			SLALogItem item = SLALogHelper.createItem(SLALogPoint.CRL_LOAD);
			item.addCallParameter("url", url);
			Timer.start(this.getClass(), "CRLCertificateStatusChecker:load");
			try {
				CRLCache.CRLInfo result = fetchURL(url, crlInfo);
				lastSuccessful.set(System.currentTimeMillis());
				SLALogHelper.storeCallResultOk(SLALogPoint.CRL_LOAD);
				return result;
			} catch (Throwable t) {
				SLALogHelper.storeCallResultError(SLALogPoint.CRL_LOAD, t.getMessage());
				lastFailed.set(System.currentTimeMillis());
				throw StsUtil.rethrowable(t, IOException.class);
			} finally {
				Timer.stop(this.getClass(), "CRLCertificateStatusChecker:load");
			}
		}

		public CRLCache.CRLInfo fetchURL(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
			return super.load(url, crlInfo);
		}


	}

	public StsOcesCrlServiceImpl() {
	}

	@Override
	public void refresh() {
		for (Map.Entry<String, CRLCache.CRLInfo> entry : crlCache.entries()) {
			log.debug("Refresh CRL: "+entry.getKey());
			try {
				final CRLCache.CRLInfo oldEntry = entry.getValue();
				final CRLCache.CRLInfo newEntry = load(entry.getKey(), oldEntry);
				log.debug((oldEntry.getLastModified() == newEntry.getLastModified() ? "Not updated" : "Updated")
									+ ": lastModified = " + new Date(newEntry.getLastModified()));
				crlCache.update(entry.getKey(), newEntry);
			} catch (Throwable t) {
				log.error("Problem with refresh: ", t);
			}
		}
	}

	public CRLCache.CRLInfo loadFromURL(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
		return load(url, crlInfo);
	}

	public CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
		return certificateStatusChecker.load(url, crlInfo);
	}

	public long getLastSuccessful() {
		return lastSuccessful.get();
	}

	public long getLastFailed() {
		return lastFailed.get();
	}

	@Override
	public CertificateStatus getRevocationStatus(final X509Certificate certificate) throws PKIException {
		return certificateStatusChecker.getRevocationStatus(certificate);
	}

	@Override
	public void afterPropertiesSet() throws Exception {
		if (crlCache == null)
			setCrlCache(new InMemoryCRLCache());

		setCertificateStatusChecker(new StsCertificateStatusChecker(crlCache,
																	interval,
																	strict,
																	crl_ttl * 1000, /* The TTL in Seal is in ms */
																	certificateResolver));
		// refresh the value on the instance
		setReadTimeout(readTimeout);
		setConnectTimeout(connectTimeout);
	}

	public void setCertificateStatusChecker(final StsCertificateStatusChecker certificateStatusChecker) {
		this.certificateStatusChecker = certificateStatusChecker;
	}

	public CRLCache getCrlCache() {
		return crlCache;
	}

	public void setCrlCache(final CRLCache crlCache) {
		this.crlCache = crlCache;
	}

	public CertificateStatusChecker getCertificateStatusChecker() {
		return certificateStatusChecker;
	}

	public CertificateResolver getCertificateResolver() {
		return certificateResolver;
	}

	public void setCertificateResolver(final CertificateResolver certificateResolver) {
		this.certificateResolver = certificateResolver;
	}

	public Properties getCheckerProperties() {
		return checkerProperties;
	}

	public void setCheckerProperties(final Properties checkerProperties) {
		this.checkerProperties = checkerProperties;
	}

	public void setConnectTimeout(int connectTimeout) {
		if (connectTimeout <= 0) return;

		this.connectTimeout = connectTimeout;
		if (certificateStatusChecker != null)
			certificateStatusChecker.setConnectTimeout(connectTimeout);
	}

	public void setReadTimeout(int readTimeout) {
		if (readTimeout <= 0) return;

		this.readTimeout = readTimeout;
		if (certificateStatusChecker != null)
			certificateStatusChecker.setReadTimeout(readTimeout);
	}

	public int getInterval() {
		return interval;
	}

	public void setInterval(final int interval) {
		this.interval = interval;
	}

	public boolean isStrict() {
		return strict;
	}

	public void setStrict(final boolean strict) {
		this.strict = strict;
	}

	public int getCrl_ttl() {
		return crl_ttl;
	}

	public void setCrl_ttl(final int crl_ttl) {
		this.crl_ttl = crl_ttl;
	}
}
