/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service;

import java.security.cert.X509Certificate;
import java.util.Map;

import dk.sosi.seal.pki.CertificateResolver;
import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.service.OcesCvrRidService;

/**
 * This class constitutes a mapping from SDN to OcesCvrRidServices enabling
 * one service for multiple CVR-RID endpoints.
 */
public class StsOcesCvrRidService implements OcesCvrRidService {

	private Map<String, OcesCvrRidService> wsEndpoints;
	private CertificateResolver certificateResolver;
	private int certificateChainMaxDepth = 10;

    public Map<String, OcesCvrRidService> getWsEndpoints() {
		return wsEndpoints;
	}

	public void setWsEndpoints(final Map<String, OcesCvrRidService> wsEndpoints) {
		this.wsEndpoints = wsEndpoints;
	}

	public CertificateResolver getCertificateResolver() {
		return certificateResolver;
	}

	public void setCertificateResolver(final CertificateResolver certificateResolver) {
		this.certificateResolver = certificateResolver;
	}

	private OcesCvrRidService findService(X509Certificate cert) throws StsException {
		final X509Certificate rootCertificate = getRootIssuingCertificate(cert);
		final String sdn = getKey(rootCertificate);
		final OcesCvrRidService ocesCvrRidService = wsEndpoints.get(sdn);
		if (ocesCvrRidService == null) {
			throw new StsException("no web service endpoint found for subjectDN: " + sdn);
		}
		return ocesCvrRidService;
	}

	public int getCertificateChainMaxDepth() {
		return certificateChainMaxDepth;
	}

	public void setCertificateChainMaxDepth(final int certificateChainMaxDepth) {
		this.certificateChainMaxDepth = certificateChainMaxDepth;
	}

	public boolean isRelated(final OcesInfo ocesInfo, final CprNr cpr) throws StsException {
		return findService(ocesInfo.getCertificate()).isRelated(ocesInfo, cpr);
	}

	public CprNr findRelatedCpr(final OcesInfo ocesInfo) throws StsException {
		return findService(ocesInfo.getCertificate()).findRelatedCpr(ocesInfo);
	}

	public long lastSuccess() {
		long latest = -1;
		for (Map.Entry<String, OcesCvrRidService> entry : wsEndpoints.entrySet()) {
			latest = Math.max(latest, entry.getValue().lastSuccess());
		}
		return latest;
	}

	public long lastFailure() {
		long latest = -1;
		for (Map.Entry<String, OcesCvrRidService> entry : wsEndpoints.entrySet()) {
			latest = Math.max(latest, entry.getValue().lastFailure());
		}
		return latest;
	}

	@Override
	public int getFailedCallsPercentage() { return 0; }

	@Override
	public int getHoursBackInTimeToCalculateFailedCallsPercentage() { return 0; }

	String getKey(X509Certificate cert) {
		return cert.getSubjectDN().getName();
	}

	X509Certificate getIssuingCertificate(X509Certificate cert) {
		return certificateResolver.getIssuingCertificate(cert);
	}

	/**
	 * @param cert the given certificate
	 * @return
	 *     Fetches the root certificate for the given certificate
	 * @throws StsException is thrown if maximum depth is reached for searching the certificate chain
	 */
	X509Certificate getRootIssuingCertificate(X509Certificate cert) throws StsException {
		boolean fixPointFound = false;
		int depth = certificateChainMaxDepth;

		while(!fixPointFound) {
			final X509Certificate issuingCertificate = getIssuingCertificate(cert);
			fixPointFound = issuingCertificate == cert;
			cert = issuingCertificate;
			if (--depth < 0) {
				throw new StsException("no root certificate found for "+getKey(cert));
			}
		}

		return cert;
	}

	String getRootIssuerKey(X509Certificate cert) throws StsException {
		return getKey(getRootIssuingCertificate(cert));
	}
}
