/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service.jwt;

import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.sts.model.AudienceConfiguration;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.VerifyException;
import dk.sosi.sts.server.VerifyUtil;

import io.jsonwebtoken.Claims;
import io.jsonwebtoken.Jws;

import java.util.Arrays;
import java.util.List;

public class OcesKeyCloakJwtStrategy implements JwtIdpStrategy {
    public static final String CPR_CLAIM = "cpr";
    public static final String CERTSUB_DN = "certsubdn";

    private String requiredStsScope;


    // spring
    public void setRequiredStsScope(String requiredStsScope) {
        this.requiredStsScope = requiredStsScope;
    }

    @Override
    public void validateScope(Jws<Claims> jwtToken, AudienceConfiguration config) throws VerifyException {
        final String scope = getScope(jwtToken);
        final List<String> scopes = Arrays.asList(scope.split(" "));
        if (!scopes.contains(requiredStsScope)) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Incoming JSON webtoken not scoped for this sts: " + scope, StsFaultActorValues.ITS);
        }
        String requiredScopeForAudience = config.getJwtScope();
        if (requiredScopeForAudience == null) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "JWT scope not configured for audience " + config.getAudience(), StsFaultActorValues.ITS);
        }
        if (!scopes.contains(requiredScopeForAudience)) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "Incoming JSON webtoken not scoped for audience ("+ config.getAudience()+ "): " + scope, StsFaultActorValues.ITS);
        }
    }

    @Override
    public String getScope(Jws<Claims> jwtToken) {
        return jwtToken.getBody().get("scope", String.class);
    }

    @Override
    public SubjectSerialNumber getPidSubjectSerial(Jws<Claims> jwtToken) throws VerifyException {
        String subjectDN = getSubjectDistinguishedName(jwtToken);
        if (subjectDN == null) {
            throw new VerifyException(StsFaultCodeValues.WST_INVALID_REQUEST, "JWT claim missing: " + CERTSUB_DN, StsFaultActorValues.ITS);
        }
        SubjectSerialNumber pidSubjectSerial = getSubjectSerialNumber(subjectDN);
        return pidSubjectSerial;
    }

    private SubjectSerialNumber getSubjectSerialNumber(String subjectNameID) throws VerifyException {
        return VerifyUtil.verifyCitizenToken(subjectNameID);
    }

    @Override
    public String getTrustedCpr(Jws<Claims> jwtToken) {
        return jwtToken.getBody().get(CPR_CLAIM, String.class);
    }

    @Override
    public String getSubjectDistinguishedName(Jws<Claims> jwtToken) {
        return jwtToken.getBody().get(CERTSUB_DN, String.class);
    }
}
