package dk.sosi.sts.server.service.procuration;

import dk.sosi.seal.model.constants.HealthcareSAMLAttributes;
import dk.sosi.sts.ServiceInfoCollector;
import dk.sosi.sts.StsException;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.PrivilegeGroup;
import dk.sosi.sts.service.ProcurationService;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;

public class AggregatedProcurationService implements ProcurationService {
    private final Logger logger = Logger.getLogger(AggregatedProcurationService.class);

    private final ProcurationService[] services;

    private final ServiceInfoCollector serviceInfoCollector;
    private final AtomicLong lastSuccess = new AtomicLong(-1);
    private final AtomicLong lastFailure = new AtomicLong(-1);


    public AggregatedProcurationService(ServiceInfoCollector serviceInfoCollector, ProcurationService... services) {
        this.serviceInfoCollector = serviceInfoCollector;
        this.services = services;
    }

    @Override
    public PrivilegeGroup getProcurationPrivileges(CprNr providedBy, CprNr representative) throws StsException {
        PrivilegeGroup aggregatedPrivilegeGroup = new PrivilegeGroup(
                HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + providedBy.getValue());

        List<Exception> exceptions = new ArrayList<>();
        for (ProcurationService service : services) {
            try {
                PrivilegeGroup privilegeGroup = service.getProcurationPrivileges(providedBy, representative);
                aggregatedPrivilegeGroup.getPrivileges().addAll(privilegeGroup.getPrivileges());
                aggregatedPrivilegeGroup.getConstraints().addAll(privilegeGroup.getConstraints());
            } catch (StsException e) {
                logger.error("Failed to get procuration privileges from service: " + service.getClass().getName(), e);
                exceptions.add(e);
            }
        }

        if (exceptions.size() == services.length) {
            serviceInfoCollector.registerFailure();
            lastFailure.set(System.currentTimeMillis());
            throw new StsException("Error from aggregated procuration service. All services failed.");
        }

        serviceInfoCollector.registerSuccess();
        lastSuccess.set(System.currentTimeMillis());

        return aggregatedPrivilegeGroup;
    }

    @Override
    public long lastSuccess() {
        return lastSuccess.get();
    }

    @Override
    public long lastFailure() {
        return lastFailure.get();
    }

    @Override
    public int getFailedCallsPercentage() {
        return serviceInfoCollector.getFailedCallsPercentage();
    }

    @Override
    public int getHoursBackInTimeToCalculateFailedCallsPercentage() {
        return serviceInfoCollector.getHoursBackInTime();
    }
}
