/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service.procuration;

import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.seal.model.constants.HealthcareSAMLAttributes;
import dk.sosi.sts.ServiceInfoCollector;
import dk.sosi.sts.StsException;
import dk.sosi.sts.fuldmagt.*;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.PrivilegeGroup;
import dk.sosi.sts.service.ProcurationService;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.sla.SLALogPoint;

import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.WebServiceException;
import java.net.URL;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;

public class ProcurationWebservice implements ProcurationService {
    public static final String NAMESPACE = "http://schemas.datacontract.org/2004/07/DK.OES.KFOBS.Delegation.Frontend.DelegationWebService";
    private URL endpoint;
    private IQueryWebService port;
    private String entityId;

    private AtomicLong lastSuccess = new AtomicLong(-1);
    private AtomicLong lastFailure = new AtomicLong(-1);

    private final ServiceInfoCollector serviceInfoCollector;

    public ProcurationWebservice(ServiceInfoCollector serviceInfoCollector) {
        this.serviceInfoCollector = serviceInfoCollector;
    }

    // injected by spring
    public void setEndpoint(final URL endpoint) {
        this.endpoint = endpoint;
    }

    // injected by spring
    public void setEntityId(String entityId) {
        this.entityId = entityId;
    }

    // invoked by spring
    public void init() {
        final URL wsdlResource = getClass().getClassLoader().getResource("dqws.wsdl");
        final IQueryWebService port = new QueryWebService(wsdlResource).getBasicHttpBindingIQueryWebService();
        BindingProvider bindingProvider = (BindingProvider) port;
        bindingProvider.getRequestContext().put(BindingProvider.SOAPACTION_USE_PROPERTY, Boolean.TRUE);
        bindingProvider.getRequestContext().put(BindingProvider.SOAPACTION_URI_PROPERTY, "https://DelegationQuery.Nemlog-in.dk/IQueryWebService/GetDelegations");
        bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpoint.toExternalForm());
        this.port = port;
    }

    @Override
    public PrivilegeGroup getProcurationPrivileges(CprNr providedBy, CprNr representative) throws StsException {
        try {
            PrivilegeGroup privilegeGroup = new PrivilegeGroup(
                    HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + providedBy.getValue());
            RepresentativeId representativeId = new RepresentativeId();
            QName representativeQname = new QName(NAMESPACE, "CPR");
            representativeId.setCPR(new JAXBElement<>(representativeQname, String.class, representative.getValue()));
            final DelegationResponse response = getDelegations(representativeId);
            final List<Delegation> delegations = response.getDelegations().getValue().getDelegation();
            for (Delegation delegation : delegations) {
                if (delegation.getCitizenCpr().getValue().equals(providedBy.getValue())) {
                    delegation.getPrivileges().getValue().getString().forEach(privilegeGroup::addPrivilege);
                    return privilegeGroup;
                }
            }
            serviceInfoCollector.registerSuccess();
            return privilegeGroup;
        } catch (Exception e) { serviceInfoCollector.registerFailure(); throw e; }
    }

    @Override
    public long lastSuccess() {
        return lastSuccess.get();
    }

    @Override
    public long lastFailure() {
        return lastFailure.get();
    }

    @Override
    public int getFailedCallsPercentage() { return serviceInfoCollector.getFailedCallsPercentage(); }

    @Override
    public int getHoursBackInTimeToCalculateFailedCallsPercentage() { return serviceInfoCollector.getHoursBackInTime(); }


    private DelegationResponse getDelegations(RepresentativeId representativeId) throws StsException {
        SLALogItem item = SLALogHelper.createItem(SLALogPoint.FULDMAGT_GETDELEGATIONS);
        item.addCallParameter("endpoint", endpoint.toString());
        item.addCallParameter("cpr", representativeId.getCPR().getValue());

        DelegationResponse response;
        try {
            response = port.getDelegations(entityId, representativeId);
            SLALogHelper.storeCallResultOk(SLALogPoint.FULDMAGT_GETDELEGATIONS);
            lastSuccess.set(System.currentTimeMillis());
            return response;
        } catch (WebServiceException e) {
            SLALogHelper.storeCallResultError(SLALogPoint.FULDMAGT_GETDELEGATIONS, e.getMessage());
            lastFailure.set(System.currentTimeMillis());
            throw new StsException("Error from procurationservice: " + e.getMessage(), e);
        } catch (IQueryWebServiceGetDelegationsQueryServiceFaultFaultFaultMessage e) {
            final QueryServiceFault faultInfo = e.getFaultInfo();
            final String errorDetails = getErrorDetails(faultInfo);
            SLALogHelper.storeCallResultError(SLALogPoint.FULDMAGT_GETDELEGATIONS, errorDetails);
            lastFailure.set(System.currentTimeMillis());
            throw new StsException("Error from procurationservice: " + errorDetails, e);
        }

    }

    private String getErrorDetails(QueryServiceFault faultInfo) {
        if (faultInfo == null) {
            return "(no faultinfo)";
        }
        String errorCode = faultInfo.getErrorCode() != null ? faultInfo.getErrorCode().value() : "(null)";
        String errorMessage = faultInfo.getMessage() != null ? faultInfo.getMessage().getValue() : "(null)";
        return "code=" + errorCode + " message=" + errorMessage;
    }
}