/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.war;

import dk.sdsd.nsp.slalog.api.SLALogConfig;
import dk.sdsd.nsp.slalog.impl.BackendInstantiator;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.vault.CredentialPair;
import dk.sosi.sts.sla.SLALogHelper;
import dk.sosi.sts.timing.CommonsLoggingTimeSink;
import dk.sosi.sts.timing.Timer;
import org.apache.log4j.Logger;

import java.util.Arrays;
import java.util.Map;
import java.util.Properties;

/**
 *
 */
public class Startup {

	private static final String SLA_LOG_APP_NAME = "NSP STS";
	private static final String SLA_LOG_APP_SHORT_NAME = "sosists";

	public static void configureSlaLogging(boolean enabled) {
		Logger logger = Logger.getLogger(Startup.class);

		SLALogConfig slaLogConfig;
		if (enabled) {
			logger.info("Enabling SLA logging for '" + SLA_LOG_APP_NAME + "' (" + SLA_LOG_APP_SHORT_NAME + ")");
			slaLogConfig = new SLALogConfig(SLA_LOG_APP_NAME, SLA_LOG_APP_SHORT_NAME);
		} else {
			// disable SLA logging
			Properties properties = new Properties();
			properties.put(BackendInstantiator.NSPSLALOG_LOG4J_CONFIGURE_KEY, false);
			slaLogConfig = new SLALogConfig(properties);
		}
		SLALogHelper.setLogger(slaLogConfig.getSLALogger());
	}

	public static void setupTimer(String category,
								  final String filterList,
								  final int timingMax) {
		Logger timingLog = Logger.getLogger(category); // stsConfig.get("sts.timing.category", "STS.TIMING"));
		CommonsLoggingTimeSink sink = new CommonsLoggingTimeSink(timingLog);
		Timer timer = new Timer(sink,
								Arrays.asList(parseList(filterList)), // stsConfig.get("sts.timing.filter", ""))),
								timingMax); // stsConfig.getInteger("sts.timing.max", 32));
		Timer.setTimer(timer);
	}

	private static String[] parseList(final String filterString) {
		if (filterString.isEmpty()) return new String[] {};

		return filterString.split(" ");
	}

	public static Properties mergeWithCrypto(Properties props) {
		final Properties properties = SignatureUtil.setupCryptoProviderForJVM();
		for (Map.Entry<Object, Object> entry : props.entrySet()) {
			properties.put(entry.getKey(), entry.getValue());
		}

		return properties;
	}

	public static void verifySts(SOSIFactory sosiFactory) {
		Logger log = Logger.getLogger(Startup.class);

		CredentialPair pair = sosiFactory.getCredentialVault().getSystemCredentialPair();
		log.info("Verifying STS SOSI configuration");

		RuntimeException e = null;
		if (pair == null) {
			e = new IllegalStateException("Missing STS credential pair");
		} else if (pair.getPrivateKey() == null) {
			e = new IllegalStateException("Missing STS private key");
		} else if (pair.getCertificate() == null) {
			e = new IllegalStateException("Missing STS certificate");
		} else if (!sosiFactory.getFederation().isValidSTSCertificate(pair.getCertificate())) {
			e = new IllegalStateException("STS certificate not valid for: "
										  + sosiFactory.getFederation().getSTSSubjectSerialNumber());
		}

		if (e == null) {
			log.info("STS SOSI configuration verified");
		} else {
			log.error("STS SOSI configuration verify error: " + e.getMessage());
			throw e;
		}
	}
}
