package dk.sosi.seal.model;

import dk.sosi.seal.model.OIOBSTSAMLAssertion.SAMLAttribute;
import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import java.util.Date;
import java.util.Properties;

/**
 * Created by Kent on 20-03-2017.
 */
public class BootstrapTokenBuilder extends AbstractOIOBSTSAMLAssertionBuilder<BootstrapTokenBuilder, OIOBootstrapToken> {
    public static final int ONE_HOUR = 1000 * 60 * 60;

    private final Date created = new Date(System.currentTimeMillis() - ONE_HOUR);
    private Date expires = new Date(System.currentTimeMillis() + 1000 * 60 * 60);

    /**
     *
     * @param signingVault a credential Vault containing the private key for an identity provider trusted by STS (normally NemLog-In IdP)
     */
    public BootstrapTokenBuilder(CredentialVault signingVault, CertificateInfo certInfo) {
        this(signingVault, formatDN(certInfo.getSubjectDN()));
        setNotOnOrAfter(expires);
        setCpr("1234567890"); // At least one attribute is required
    }

    public BootstrapTokenBuilder(CredentialVault signingVault, String nameId) {
        setIssuer("https://oio2bst-legacy-issuer.dk");
        setAudience("http://audience/clear"); // configured in EmbeddedBootstrapTokenToIdentityTokenBuilder
        setSigningVault(signingVault);
        setHolderOfKeyCertificate(signingVault.getSystemCredentialPair().getCertificate());
        validate("signingVault", signingVault);
        setNameId(nameId);
        validate("nameId", nameId);
        setNotOnOrAfter(expires);
        setCpr("1234567890"); // At least one attribute is required
    }

    public BootstrapTokenBuilder expires(Date expires) {
        setNotOnOrAfter(expires);
        return this;
    }

    @Override
    protected OIOBootstrapToken createAssertion(Element elm) {
        return new OIOBootstrapToken(elm);
    }

    public OIOBootstrapToken buildInvalidSignature() throws ModelException {
        Document document = createDocument();
        // This makes it easier to pass the document around without messing up the signature.
        String documentString = XmlUtil.node2String(document.getDocumentElement());
        document = XmlUtil.readXml(new Properties(), documentString, false);

        signAssertion(document);

        Element subjectNameId = (Element) document.getElementsByTagNameNS(SAMLTags.nameID.getNS(), SAMLTags.nameID.getTagName()).item(0);
        subjectNameId.setTextContent(subjectNameId.getTextContent() + "breaksignature");

        return new OIOBootstrapToken(document.getDocumentElement());
    }

    @Override
    protected SAMLAttribute[] getAttributes() {
        return OIOBootstrapToken.Attribute.values();
    }

    @Override
    protected final void appendToRoot(Document doc, Element root) {
        super.appendToRoot(doc, root);
        root.appendChild(createAuthnStatement());
    }

    @Override
    protected String getNameIDFormat() {
        return SAMLValues.NAMEID_FORMAT_X509_SUBJECT_NAME;
    }

    protected Node createAuthnStatement() {
        Element authnStatementElm = createElement(SAMLTags.authnStatement);
        authnStatementElm.setAttribute(SAMLAttributes.AUTHN_INSTANT, XmlUtil.getDateFormat(true).format(this.created));

        Element authnContextElm = createElement(SAMLTags.authnContext);
        Element authContextClassRefElm = createElement(SAMLTags.authnContextClassRef);
        authContextClassRefElm.setTextContent(SAMLValues.AUTHN_CONTEXT_CLASS_REF);
        authnContextElm.appendChild(authContextClassRefElm);
        authnStatementElm.appendChild(authnContextElm);
        // SessionIndex is mandatory in OIOSAML (to support per session logout), but is meaningless
        // for STS issued assertions
        authnStatementElm.setAttribute(SAMLAttributes.SESSION_INDEX, assertionID);
        return authnStatementElm;
    }


    private static String formatDN(DistinguishedName name) {
        StringBuilder builder = new StringBuilder("C=");
        builder.append(name.getCountry());
        builder.append(",O=");
        builder.append(name.getOrganization());
        builder.append(",CN=");
        builder.append(name.getCommonName());
        builder.append(",Serial=");
        builder.append(name.getSubjectSerialNumber());
        return builder.toString();
    }

    public void setAssuranceLevel(AssuranceLevel.NIST nist) {
        addAttribute(OIOBootstrapToken.Attribute.ASSURANCE_LEVEL, nist);
    }

    public void setCpr(String cpr) {
        addAttribute(OIOBootstrapToken.Attribute.CPR, cpr);
    }

    public static void removeAttribute(AbstractOIOBSTSAMLAssertionBuilder bootstrapTokenBuilder, OIOBSTSAMLAssertion.SAMLAttribute attribute) {
        bootstrapTokenBuilder.removeAttribute(attribute);
    }
}
