/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.acl;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.server.StsConfig;
import dk.sosi.sts.server.test.StsCredentialVaults;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;

import java.security.cert.X509Certificate;
import java.util.Collections;
import java.util.HashSet;
import java.util.Properties;

@RunWith(Parameterized.class)
public class MockAclServiceTest {

	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public X509Certificate testCert;

	private String SYSTEM_NAME = "SOSITEST";
	private static OcesInfo ocesInfo;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate());

		return builder.build();
	}

	@Before
	public void setup() {
		ocesInfo = OcesInfoParser.parseCertificate(testCert);
	}

	@Test
	public void testNullLists() {
		MockAclService acl = new MockAclService(null, null);

		assertFalse(acl.isBlacklisted(ocesInfo, SYSTEM_NAME));
		assertTrue(acl.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testEmptyLists() {
		MockAclService acl = new MockAclService(Collections.<String>emptySet(), Collections.<String>emptySet());

		assertFalse(acl.isBlacklisted(ocesInfo, SYSTEM_NAME));
		assertFalse(acl.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testNonEmptyLists() {
		HashSet<String> blackSet = new HashSet<String>();
		blackSet.add(ocesInfo.getSubjectSerialNumber().getValue());
		HashSet<String> whiteSet = new HashSet<String>();
		whiteSet.add(MockAclService.calculateWhitelistKey(ocesInfo, SYSTEM_NAME));
		MockAclService acl = new MockAclService(blackSet, whiteSet);

		assertTrue(acl.isBlacklisted(ocesInfo, SYSTEM_NAME));
		assertTrue(acl.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testDifferentLists() {
		HashSet<String> blackSet = new HashSet<String>();
		blackSet.add(ocesInfo.getSubjectSerialNumber().getValue());
		MockAclService acl = new MockAclService(blackSet, Collections.<String>emptySet());

		assertTrue(acl.isBlacklisted(ocesInfo, SYSTEM_NAME));
		assertFalse(acl.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testStsConfigConstructorNoKeys() {
		Properties props = new Properties();
		StsConfig config = new StsConfig(props);
		MockAclService acl = new MockAclService(config);

		assertFalse(acl.isBlacklisted(ocesInfo, SYSTEM_NAME));
		assertTrue(acl.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testStsConfigConstructorEmptyKeys() {
		Properties props = new Properties();
		props.put(MockAclService.class.getName() + ".blacklist", "");
		props.put(MockAclService.class.getName() + ".whitelist", "");
		StsConfig config = new StsConfig(props);
		MockAclService acl = new MockAclService(config);

		assertFalse(acl.isBlacklisted(ocesInfo, SYSTEM_NAME));
		assertFalse(acl.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testStsConfigConstructor() {
		Properties props = new Properties();
		String whiteList = "1234  " + MockAclService.calculateWhitelistKey(ocesInfo, SYSTEM_NAME) + "  5678";
		String blackList = "1234  " + ocesInfo.getSubjectSerialNumber().getValue() + "  5678";
		props.put(MockAclService.class.getName() + ".blacklist", blackList);
		props.put(MockAclService.class.getName() + ".whitelist", whiteList);
		StsConfig config = new StsConfig(props);
		MockAclService acl = new MockAclService(config);

		assertTrue(acl.isBlacklisted(ocesInfo, SYSTEM_NAME));
		assertTrue(acl.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}
}
