/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.monitor;

import static org.mockito.Matchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.security.cert.X509Certificate;
import java.util.Date;

import org.mockito.Mockito;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.pki.SignatureProvider;
import dk.sosi.seal.vault.CredentialPair;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.server.StsStatus;
import junit.framework.TestCase;

/**
 *
 */
public class StatusCheckTest extends TestCase {

	CredentialVault mockCredVault;
	CredentialPair mockCredPair;
	X509Certificate mockCert;
	SOSIFactory mockSosiFactory;
	private SignatureProvider mockSigner;

	private void setup(Date from,
										Date to) {
		mockSosiFactory = mock(SOSIFactory.class);

		mockCredVault = mock(CredentialVault.class);
		mockCredPair = mock(CredentialPair.class);
		mockCert = mock(X509Certificate.class);
		when(mockCredPair.getCertificate()).thenReturn(mockCert);
		when(mockCredVault.getSystemCredentialPair()).thenReturn(mockCredPair);
		when(mockSosiFactory.getCredentialVault()).thenReturn(mockCredVault);
		mockSigner = mock(SignatureProvider.class);
		when(mockSigner.sign(any(byte[].class))).thenReturn(null);
		when(mockSosiFactory.getSignatureProvider()).thenReturn(mockSigner);

		when(mockCert.getNotAfter()).thenReturn(to);
		when(mockCert.getNotBefore()).thenReturn(from);
	}

	public void testPerform() throws Exception {
		final Date from = new Date(0);
		final Date to = new Date(System.currentTimeMillis()+10000);

		final CertificateStatusChecker mockCSC = mock(CertificateStatusChecker.class);

		final CertificateStatus status = new CertificateStatus(true, new Date());

		when(mockCSC.getRevocationStatus(Mockito.<X509Certificate>any())).thenReturn(status);

		setup(from, to);

		final StsStatus stsStatus = new StsStatus();
		final StatusCheck statusCheck = new StatusCheck(stsStatus,
														mockSosiFactory,
														mockCSC);

		final Check.CheckResult checkResult = statusCheck.performCheck();

		assertEquals(14, checkResult.result.size());

		assertEquals(stsStatus.getVersion(), checkResult.result.get("Version"));
		assertEquals(stsStatus.getStartedDate(), checkResult.result.get("Started"));
		assertEquals(MonitorUtil.toDate(stsStatus.getLatestTransaction()), checkResult.result.get("Last transaction"));
		assertEquals(MonitorUtil.toDate(stsStatus.getLastIssued()), checkResult.result.get("Last issued"));
		assertEquals(from.toString(), checkResult.result.get("Sts certificate, valid from"));
		assertEquals(to.toString(), checkResult.result.get("Sts certificate, valid until"));
		assertEquals("OK", checkResult.result.get("Sts certificate valid"));
		assertEquals("" + stsStatus.getTransactions24h(), checkResult.result.get("Requests, 24h"));
		assertEquals("" + stsStatus.getTransactionsForever(), checkResult.result.get("Requests, total"));
		assertEquals("" + stsStatus.getErrors24h(), checkResult.result.get("Failed requests, 24h"));
		assertEquals("" + stsStatus.getErrorsForever(), checkResult.result.get("Failed requests, total"));
		assertEquals("" + stsStatus.getAverageTime24h(), checkResult.result.get("Response time, avg 24h"));
		assertEquals("" + stsStatus.getMaxTime24h(), checkResult.result.get("Response time, max 24h"));
		assertFalse(checkResult.failed);
	}

	public void testPerformInvalidCertificate() throws Exception {
		final Date from = new Date(0);
		final Date to = new Date(System.currentTimeMillis()-10000);

		final CertificateStatusChecker mockCSC = mock(CertificateStatusChecker.class);

		final CertificateStatus status = new CertificateStatus(true, new Date());

		when(mockCSC.getRevocationStatus(Mockito.<X509Certificate>any())).thenReturn(status);

		setup(from, to);

		final StsStatus stsStatus = new StsStatus();
		final StatusCheck statusCheck = new StatusCheck(stsStatus,
														mockSosiFactory,
														mockCSC);

		final Check.CheckResult checkResult = statusCheck.performCheck();

		assertEquals(14, checkResult.result.size());

		assertEquals("not valid", checkResult.result.get("Sts certificate valid"));
		assertTrue(checkResult.failed);
	}

	public void testPerformRevokedCertificate() throws Exception {
		final Date from = new Date(0);
		final Date to = new Date(System.currentTimeMillis()+10000);

		final CertificateStatusChecker mockCSC = mock(CertificateStatusChecker.class);

		final CertificateStatus status = new CertificateStatus(false, new Date());

		when(mockCSC.getRevocationStatus(Mockito.<X509Certificate>any())).thenReturn(status);

		setup(from, to);

		final StsStatus stsStatus = new StsStatus();
		final StatusCheck statusCheck = new StatusCheck(stsStatus,
														mockSosiFactory,
														mockCSC);

		final Check.CheckResult checkResult = statusCheck.performCheck();

		assertEquals(14, checkResult.result.size());

		assertEquals("not valid", checkResult.result.get("Sts certificate valid"));
		assertTrue(checkResult.failed);
	}
}
