package dk.sosi.sts.monitor;

import java.security.KeyStore;
import java.security.KeyStoreSpi;
import java.security.cert.X509Certificate;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.security.auth.x500.X500Principal;

import org.junit.Assert;
import org.mockito.Mockito;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.stubbing.Answer;

import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.TrustedVaultProvider;
import dk.sosi.sts.monitor.Check.CheckResult;
import junit.framework.TestCase;

public class TrustedVaultCheckTest extends TestCase {

	
  	public TrustedVaultCheck subject;
  	
  	private CertificateStatusChecker mockedCrlService;
	
	private List<String> ignoresAliases = new LinkedList<String>();

	private Map<String, X509Certificate> testKeystore = new HashMap<>();

	@Override
	protected void setUp() throws Exception {
		super.setUp();

		mockedCrlService = Mockito.mock(CertificateStatusChecker.class);
		Mockito.when(mockedCrlService.getRevocationStatus(Mockito.any())).thenReturn(new CertificateStatus(true, new Date()));

		KeyStoreSpi keyStoreSpiMock = Mockito.mock(KeyStoreSpi.class);
		KeyStore mockKeystore = new KeyStore(keyStoreSpiMock, null, "test"){ };
		mockKeystore.load(null);  // this is important to put the internal state "initialized" to true

		Mockito.when(mockKeystore.aliases()).thenAnswer(new Answer<Enumeration<String>>() {
			@Override
			public Enumeration<String> answer(InvocationOnMock invocation) throws Throwable {
				return Collections.enumeration(testKeystore.keySet());
			}
		});
		Mockito.when(mockKeystore.getCertificate(Mockito.anyString())).thenAnswer(new Answer<X509Certificate>() {
			@Override
			public X509Certificate answer(InvocationOnMock invocation) throws Throwable {
				String alias = (String) invocation.getArguments()[0];
				return testKeystore.get(alias);
			}
		});

		CredentialVault testCredentialVault = Mockito.mock(CredentialVault.class);
		Mockito.when(testCredentialVault.getKeyStore()).thenReturn(mockKeystore);

		TrustedVaultProvider vaultProvider = Mockito.mock(TrustedVaultProvider.class);
		Mockito.when(vaultProvider.getTrustedVault()).thenReturn(testCredentialVault);
		
		subject = new TrustedVaultCheck(vaultProvider, mockedCrlService, "Test", ignoresAliases);
	}
	
	
	public void testEmptyKeystore() {
		
		// Given
		
		
		// When
		CheckResult checkResult = subject.performCheck();

		
		// Then
		Assert.assertNotNull(checkResult);
		Assert.assertFalse(checkResult.failed);
		Assert.assertEquals(0, checkResult.result.size());
	}
	
	public void testKeystoreWithValidCertificate() {
		
		// Given
		X509Certificate certificate = Mockito.mock(X509Certificate.class);
		testKeystore.put("valid", certificate);
		Instant now = Instant.now();
		Date yesterday = Date.from(now.minus(1, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotBefore()).thenReturn(yesterday);
		Date tomorrow = Date.from(now.plus(1, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotAfter()).thenReturn(tomorrow);
		X500Principal x500prinicipal = new X500Principal("CN=Duke, SERIALNUMBER=12345, OU=JavaSoft, O=Sun Microsystems, C=US");
		Mockito.when(certificate.getSubjectX500Principal()).thenReturn(x500prinicipal);

		
		// When
		CheckResult checkResult = subject.performCheck();

		
		// Then
		String key = "Trusted certificate Test/valid -> 12345";
		Assert.assertNotNull(checkResult);
		Assert.assertFalse(checkResult.failed);
		Assert.assertEquals(1, checkResult.result.size());
		Assert.assertTrue(checkResult.result.containsKey(key));
		String result = checkResult.result.get(key);
		Assert.assertTrue(result.startsWith("OK"));
		Assert.assertTrue(result.contains(String.format("(Expires: %1$td-%1$tm-%1$tY %1$tH:%1$tM:%1$tS)", tomorrow)));
	}

	public void testKeystoreWithExpiredCertificate() {
		
		// Given
		X509Certificate certificate = Mockito.mock(X509Certificate.class);
		testKeystore.put("expired", certificate);
		Instant now = Instant.now();
		Date dayBeforeYesterday = Date.from(now.minus(2, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotBefore()).thenReturn(dayBeforeYesterday);
		Date yesterday = Date.from(now.minus(1, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotAfter()).thenReturn(yesterday);
		X500Principal x500prinicipal = new X500Principal("CN=Duke, SERIALNUMBER=12345, OU=JavaSoft, O=Sun Microsystems, C=US");
		Mockito.when(certificate.getSubjectX500Principal()).thenReturn(x500prinicipal);

		
		// When
		CheckResult checkResult = subject.performCheck();

		
		// Then
		String key = "Trusted certificate Test/expired -> 12345";
		Assert.assertNotNull(checkResult);
		Assert.assertTrue(checkResult.failed);
		Assert.assertEquals(1, checkResult.result.size());
		Assert.assertTrue(checkResult.result.containsKey(key));
		String result = checkResult.result.get(key);
		Assert.assertTrue(result.startsWith("Expired"));
		Assert.assertTrue(result.contains(String.format("(Expires: %1$td-%1$tm-%1$tY %1$tH:%1$tM:%1$tS)", yesterday)));
	}

	public void testKeystoreWithExpiredCertificateIgnored() {
		
		// Given
		X509Certificate certificate = Mockito.mock(X509Certificate.class);
		testKeystore.put("expiredignored", certificate);
		Instant now = Instant.now();
		Date dayBeforeYesterday = Date.from(now.minus(2, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotBefore()).thenReturn(dayBeforeYesterday);
		Date yesterday = Date.from(now.minus(1, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotAfter()).thenReturn(yesterday);
		X500Principal x500prinicipal = new X500Principal("CN=Duke, SERIALNUMBER=12345, OU=JavaSoft, O=Sun Microsystems, C=US");
		Mockito.when(certificate.getSubjectX500Principal()).thenReturn(x500prinicipal);
		ignoresAliases.add("expiredignored");

		
		// When
		CheckResult checkResult = subject.performCheck();

		// Then
		String key = "Trusted certificate Test/expiredignored -> 12345";
		Assert.assertNotNull(checkResult);
		Assert.assertFalse(checkResult.failed);
		Assert.assertEquals(1, checkResult.result.size());
		Assert.assertTrue(checkResult.result.containsKey(key));
		String result = checkResult.result.get(key);
		Assert.assertTrue(result.startsWith("Ignored"));
		Assert.assertTrue(result.contains(String.format("(Expires: %1$td-%1$tm-%1$tY %1$tH:%1$tM:%1$tS)", yesterday)));
	}

	public void testKeystoreWithRevokedCertificate() {
		
		// Given
		X509Certificate certificate = Mockito.mock(X509Certificate.class);
		testKeystore.put("revoked", certificate);
		Instant now = Instant.now();
		Date yesterday = Date.from(now.minus(1, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotBefore()).thenReturn(yesterday);
		Date tomorrow = Date.from(now.plus(1, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotAfter()).thenReturn(tomorrow);
		X500Principal x500prinicipal = new X500Principal("CN=Duke, SERIALNUMBER=12345, OU=JavaSoft, O=Sun Microsystems, C=US");
		Mockito.when(certificate.getSubjectX500Principal()).thenReturn(x500prinicipal);
		Mockito.when(mockedCrlService.getRevocationStatus(Mockito.any())).thenReturn(new CertificateStatus(false, yesterday));
		
		
		// When
		CheckResult checkResult = subject.performCheck();

		
		// Then
		String key = "Trusted certificate Test/revoked -> 12345";
		Assert.assertNotNull(checkResult);
		Assert.assertTrue(checkResult.failed);
		Assert.assertEquals(1, checkResult.result.size());
		Assert.assertTrue(checkResult.result.containsKey(key));
		String result = checkResult.result.get(key);
		Assert.assertTrue(result.startsWith("Revoked"));
		Assert.assertTrue(result.contains(String.format("(Expires: %1$td-%1$tm-%1$tY %1$tH:%1$tM:%1$tS)", tomorrow)));
	}

	public void testKeystoreWithRevokedCertificateIgnored() {
		
		// Given
		X509Certificate certificate = Mockito.mock(X509Certificate.class);
		testKeystore.put("revokedignored", certificate);
		Instant now = Instant.now();
		Date yesterday = Date.from(now.minus(1, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotBefore()).thenReturn(yesterday);
		Date tomorrow = Date.from(now.plus(1, ChronoUnit.DAYS));
		Mockito.when(certificate.getNotAfter()).thenReturn(tomorrow);
		X500Principal x500prinicipal = new X500Principal("CN=Duke, SERIALNUMBER=12345, OU=JavaSoft, O=Sun Microsystems, C=US");
		Mockito.when(certificate.getSubjectX500Principal()).thenReturn(x500prinicipal);
		Mockito.when(mockedCrlService.getRevocationStatus(Mockito.any())).thenReturn(new CertificateStatus(false, yesterday));
		ignoresAliases.add("revokedignored");
		
		// When
		CheckResult checkResult = subject.performCheck();

		
		// Then
		String key = "Trusted certificate Test/revokedignored -> 12345";
		Assert.assertNotNull(checkResult);
		Assert.assertFalse(checkResult.failed);
		Assert.assertEquals(1, checkResult.result.size());
		Assert.assertTrue(checkResult.result.containsKey(key));
		String result = checkResult.result.get(key);
		Assert.assertTrue(result.startsWith("Ignored"));
		Assert.assertTrue(result.contains(String.format("(Expires: %1$td-%1$tm-%1$tY %1$tH:%1$tM:%1$tS)", tomorrow)));
	}
}
