/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.roles.authorization;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.hsqldb.jdbc.JDBCDataSource;

import dk.sosi.sts.model.Authorization;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.util.DbUtil;
import junit.framework.TestCase;

public class DbAuthorizationServiceTest extends TestCase {

	public static final String[] DDL_CREATE = {
		"DROP TABLE cprcache IF EXISTS",
		// simplified schema for testing only
		"CREATE TABLE autreg (" +
		"  cpr CHAR(10), aut_id CHAR(5) NOT NULL, edu_id CHAR(4) NOT NULL," +
		"  valid_from date, valid_to date)"
	};
	public static final String DDL_DROP = "DROP TABLE autreg";

	private static final CprNr CPR1 = new CprNr( "1111111118");
	private static final CprNr CPR2 = new CprNr( "1111111119");
	
	private JDBCDataSource dataSource;

	@Override
	protected void setUp() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		Connection conn = dataSource.getConnection();
		Statement statement = conn.createStatement();
		for (String sql : DDL_CREATE) {
			statement.execute(sql);
		} 
		conn.close();		
	}
	
	@Override
	protected void tearDown() throws Exception {
		Connection conn = dataSource.getConnection();
		conn.createStatement().execute(DDL_DROP);
		conn.close();		
	}

	public void testDbAuthorizationService() throws Exception {
		new DbAuthorizationService(dataSource);
		new DbAuthorizationService(dataSource, "SELECT 1");

		new DbAuthorizationService(dataSource);
		new DbAuthorizationService(dataSource, "SELECT 1");
	}

	public void testEmpty() throws Exception {
		DbAuthorizationService service = new DbAuthorizationService(dataSource);
		List<Authorization> list = service.getAuthorizations(CPR1);
		
		assertTrue(list.isEmpty());
	}

	public void testSetSelect() throws Exception {
		String[] r = { CPR1.getValue(), "0013F", "5153", "2000-01-01", "2100-01-01" };
		insertRecord(r);

		DbAuthorizationService service1 = new DbAuthorizationService(dataSource);
		assertFalse(service1.getAuthorizations(CPR1).isEmpty());

		DbAuthorizationService service2 = new DbAuthorizationService(dataSource,
				"SELECT * FROM autreg WHERE false AND cpr=?");
		assertTrue(service2.getAuthorizations(CPR1).isEmpty());
	}

	public void testGetAuthorizations() throws Exception {
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
		Date now = new Date();
		Date earlier = new Date(now.getTime() - TimeUnit.MILLISECONDS.convert(30, TimeUnit.DAYS));
		Date yesterday = new Date(now.getTime() - TimeUnit.MILLISECONDS.convert(1, TimeUnit.DAYS));
		Date tomorrow = new Date(now.getTime() + TimeUnit.MILLISECONDS.convert(1, TimeUnit.DAYS));
		Date later = new Date(now.getTime() + TimeUnit.MILLISECONDS.convert(30, TimeUnit.DAYS));
		String[][] rows = {
				{ CPR1.getValue(), "0013F", "5153", sdf.format(earlier), sdf.format(later) },
				{ CPR1.getValue(), "0013H", "5155", sdf.format(earlier), sdf.format(later) },
				{ CPR1.getValue(), "1DEAD", "DEAD", sdf.format(earlier), sdf.format(yesterday) }, // expired
				{ CPR2.getValue(), "0015J", "5151", sdf.format(earlier), sdf.format(later) },
				{ CPR2.getValue(), "2DEAD", "DEAD", sdf.format(tomorrow), sdf.format(later) } // not yet valid
		};
		for (String[] r : rows) {
			insertRecord(r);
		}
		
		DbAuthorizationService service = new DbAuthorizationService(dataSource, "SELECT aut_id, edu_id " +
			"FROM autreg WHERE cpr=? AND CURRENT_DATE BETWEEN valid_from AND valid_to");

		List<Authorization> list = service.getAuthorizations(CPR1);
		assertEquals(2, list.size());
		for (Authorization a : list) {
			assertNotNull(a.getCode());
			assertNotNull(a.getRole());
		}

		List<Authorization> list2 = service.getAuthorizations(CPR2);
		assertEquals(1, list2.size());
		assertEquals("0015J", list2.get(0).getCode());
		assertEquals("5151", list2.get(0).getRole());
	}

	private void insertRecord(String[] r) throws SQLException {
		Connection conn = dataSource.getConnection();
		Statement s = null;
		try {
			s = conn.createStatement();
			String sql = "INSERT INTO autreg " +
					"VALUES ('" + r[0] + "', '" + r[1] + "', '" + r[2] +
					"', '" + r[3] + "', '" + r[4] + "')";
			s.execute(sql);
		} finally {
			DbUtil.close(conn, s, null, null);
		}
	}
}
