/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.roles.authorizationdefinitions;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import dk.sosi.sts.server.facade.HttpTestClient;
import org.junit.Assume;
import org.junit.Before;
import org.junit.Test;
import org.springframework.jdbc.CannotGetJdbcConnectionException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.RowCallbackHandler;

import dk.sosi.sts.roles.DBTestUtil;

// marked as IT since it uses mysql on docker
public class DbAuthorizationDefinitionsIT {

	private DataSource dataSource;
	private DbAuthorizationDefinitionsService service;
	
	@Before
	public void setUp() throws Exception {
		Assume.assumeTrue(HttpTestClient.testEnvironmentHost.contains("localhost"));

		dataSource = DBTestUtil.getTestDataSource("sts");
		service = new DbAuthorizationDefinitionsService();
		service.setDataSource(dataSource);
		assumeMysqlAvailable();
	}

	private void assumeMysqlAvailable() {
		try {
			new JdbcTemplate(dataSource).execute("SELECT 1");
		} catch (CannotGetJdbcConnectionException e) {
			Assume.assumeFalse(Boolean.getBoolean("unittests.skip"));
		}
	}
	
	@Test
	public void testNotEmpty() throws Exception {
		List<String> list = service.getAuthorizationDefinitions();
		assertFalse(list.isEmpty());
	}
		
	@Test
	public void testRoleIsValid() throws Exception {
		assertTrue(service.isValidRole("7170"));
	}

	@Test
	public void testRoleIsInValid() throws Exception {
		assertFalse(service.isValidRole("8600"));
	}
	
	@Test
	public void testRoleWithCharactersIsValid() throws Exception {
		assertTrue(service.isValidRole("A511"));
	}
	
	@Test
	public void testSqlExtractFromAuth() {
		final List<String> expected_codes_from_authreg = Arrays.asList("1234","7170", "5159", "5175", "5166", "5151", "5158", "5155");
		final Map<String, Integer> expected_visits_to_codes_from_autreg = new HashMap<>();
		
		new JdbcTemplate(dataSource).query("select EDU_ID from stamdata.autreg",new RowCallbackHandler() {
			@Override
			public void processRow(ResultSet rs) throws SQLException {
				final String actual_code = rs.getString("EDU_ID");
				assertTrue("Forventede koden fra autreg tabellen var en af følgende:"+expected_codes_from_authreg,expected_codes_from_authreg.contains(actual_code));
				
				Integer visits = expected_visits_to_codes_from_autreg.get(actual_code);
				if(visits == null) {
					visits = 0;
				}
				expected_visits_to_codes_from_autreg.put(actual_code, ++visits);
			}
		});	
		
		Integer visits_1234 = expected_visits_to_codes_from_autreg.remove("1234");
		Integer visits_7170 = expected_visits_to_codes_from_autreg.remove("7170");
		Integer visits_5159 = expected_visits_to_codes_from_autreg.remove("5159");
		Integer visits_5175 = expected_visits_to_codes_from_autreg.remove("5175");
		Integer visits_5166 = expected_visits_to_codes_from_autreg.remove("5166");
		Integer visits_5151 = expected_visits_to_codes_from_autreg.remove("5151");
		Integer visits_5158 = expected_visits_to_codes_from_autreg.remove("5158");
		Integer visits_5155 = expected_visits_to_codes_from_autreg.remove("5155");

		assertEquals("Forventede ikke flere koder fra autreg. Koder:"+expected_visits_to_codes_from_autreg.keySet(), 0, expected_visits_to_codes_from_autreg.size());

		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 3 gange. Code:1234", Integer.valueOf(3), visits_1234);
		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 16 gange. Code:7170", Integer.valueOf(16), visits_7170);
		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 1 gang. Code:5159", Integer.valueOf(1), visits_5159);
		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 3 gange. Code:5175", Integer.valueOf(3), visits_5175);
		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 2 gange. Code:5166", Integer.valueOf(2), visits_5166);
		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 2 gange. Code:5151", Integer.valueOf(2), visits_5151);
		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 2 gange. Code:5158", Integer.valueOf(2), visits_5158);
		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 2 gange. Code:5155", Integer.valueOf(2), visits_5155);
	}
	
	
	@Test
	public void testSqlExtractFromAuthRegMedDistinct() {
		final List<String> expected_codes_from_authreg = Arrays.asList("1234","7170", "5159", "5175", "5166", "5151", "5158", "5175", "5155");
		final Map<String, Integer> expected_visits_to_codes_from_autreg = new HashMap<>();
		
		new JdbcTemplate(dataSource).query("select distinct EDU_ID from stamdata.autreg",new RowCallbackHandler() {	
			@Override
			public void processRow(ResultSet rs) throws SQLException {
				final String actual_code = rs.getString("EDU_ID");
				assertTrue("Forventede koden fra autreg tabellen var en af følgende:"+expected_codes_from_authreg,expected_codes_from_authreg.contains(actual_code));
				
				Integer visits = expected_visits_to_codes_from_autreg.get(actual_code);
				if(visits == null) {
					visits = 0;
				}
				expected_visits_to_codes_from_autreg.put(actual_code, ++visits);
			}
		});	
		
		for (String code : expected_codes_from_authreg) {
			assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 1 gang. Code:"+code, Integer.valueOf(1), expected_visits_to_codes_from_autreg.get(code)); 
		}
	}
	
	@Test
	public void testSqlExtractFromAuthRegAndInsertIntoAuthorizationDefinitions() {
		new JdbcTemplate(dataSource).execute("insert into sts_audconf.authorizationDefinitions (role) select distinct ar.EDU_ID from stamdata.autreg ar where not exists (select 1 from sts_audconf.authorizationDefinitions ad where ad.role=ar.EDU_ID)");
		assertTrue(service.isValidRole("7170"));
		assertTrue(service.isValidRole("A511"));
		assertTrue(service.isValidRole("1234"));
	}
}
