/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.roles.authorizationdefinitions;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.naming.NamingException;
import javax.sql.DataSource;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.springframework.context.ApplicationContext;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.RowCallbackHandler;
import org.springframework.mock.jndi.SimpleNamingContextBuilder;

import dk.sosi.sts.server.war.SpringTester;

public class DbAuthorizationDefinitionsServiceTest {

	private DbAuthorizationDefinitionsService service;
	private JdbcTemplate template;
	private static SimpleNamingContextBuilder namingContextBuilder;
	
	@BeforeClass
	public static void init() throws IllegalStateException, NamingException, IOException {
		namingContextBuilder = SpringTester.getNamingContextBuilderWithDataSources();
	}

	@Before
	public void setUp() throws Exception {
		namingContextBuilder.activate();
		ApplicationContext ctx = SpringTester.getApplicationContext();
		template = new JdbcTemplate(ctx.getBean("sts.db", DataSource.class));
		service = ctx.getBean(DbAuthorizationDefinitionsService.class);	
	}
	
	@Test
	public void testNotEmpty() throws Exception {
		List<String> list = service.getAuthorizationDefinitions();
		assertFalse(list.isEmpty());
	}
		
	@Test
	public void testRoleIsValid() throws Exception {
		assertTrue(service.isValidRole("7170"));
	}

	@Test
	public void testRoleIsInValid() throws Exception {
		assertFalse(service.isValidRole("8600"));
	}
	
	@Test
	public void testRoleWithCharactersIsValid() throws Exception {
		assertTrue(service.isValidRole("A511"));
	}
	
	@Test
	public void testSqlExtractFromAuth() {
		final List<String> expected_codes_from_authreg = Arrays.asList("1234","7170");
		final Map<String, Integer> expected_visits_to_codes_from_autreg = new HashMap<>();
		
		template.query("select EDU_ID from stamdata.autreg",new RowCallbackHandler() {	
			@Override
			public void processRow(ResultSet rs) throws SQLException {
				final String actual_code = rs.getString("EDU_ID");
				assertTrue("Forventede koden fra autreg tabellen var en af følgende:"+expected_codes_from_authreg,expected_codes_from_authreg.contains(actual_code));
				
				Integer visits = expected_visits_to_codes_from_autreg.get(actual_code);
				if(visits == null) {
					visits = 0;
				}
				expected_visits_to_codes_from_autreg.put(actual_code, ++visits);
			}
		});	
		
		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 3 gang. Code:1234", Integer.valueOf(3), expected_visits_to_codes_from_autreg.get("1234")); 
		assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 1 gang. Code:7170", Integer.valueOf(1), expected_visits_to_codes_from_autreg.get("7170")); 

		expected_visits_to_codes_from_autreg.remove("1234");
		expected_visits_to_codes_from_autreg.remove("7170");
		
		assertEquals("Forventede ikke flere koder fra autreg. Koder:"+expected_visits_to_codes_from_autreg.keySet(), 0, expected_visits_to_codes_from_autreg.size());
	}
	
	@Test
	public void testSqlExtractFromAuthRegMedDistinct() {
		final List<String> expected_codes_from_authreg = Arrays.asList("1234","7170");
		final Map<String, Integer> expected_visits_to_codes_from_autreg = new HashMap<>();
		
		template.query("select distinct EDU_ID from stamdata.autreg",new RowCallbackHandler() {	
			@Override
			public void processRow(ResultSet rs) throws SQLException {
				final String actual_code = rs.getString("EDU_ID");
				assertTrue("Forventede koden fra autreg tabellen var en af følgende:"+expected_codes_from_authreg,expected_codes_from_authreg.contains(actual_code));
				
				Integer visits = expected_visits_to_codes_from_autreg.get(actual_code);
				if(visits == null) {
					visits = 0;
				}
				expected_visits_to_codes_from_autreg.put(actual_code, ++visits);
			}
		});	
		
		for (String code : expected_codes_from_authreg) {
			assertEquals("Forventede hver edu id fra autreg tabellen ville blive returneret 1 gang. Code:"+code, Integer.valueOf(1), expected_visits_to_codes_from_autreg.get(code)); 
		}
	}
	
	@Test
	public void testSqlExtractFromAuthRegAndInsertIntoAuthorizationDefinitions() {
		template.execute("insert into sts_audconf.authorizationDefinitions (role) select distinct ar.EDU_ID from stamdata.autreg ar where not exists (select 1 from sts_audconf.authorizationDefinitions ad where ad.role=ar.EDU_ID)");
		assertTrue(service.isValidRole("7170"));
		assertTrue(service.isValidRole("A511"));
		assertTrue(service.isValidRole("1234"));
	}
//
//	@Test
//	public void testSingleSelect() throws Exception {
//		insertRecord(new String[] {"nspLaegesekretaer", "40001", "Laegesekretaer"});
//
//		final List<RoleDefinition> roleDefinitions = service.getRoleDefinitions();
//		assertEquals(1, roleDefinitions.size());
//		assertEquals("nspLaegesekretaer", roleDefinitions.get(0).getExternalName());
//		assertEquals("urn:dk:healthcare:national-federation-role:code:40001:value:Laegesekretaer", roleDefinitions.get(0).getPublicName());
//	}
//
//    @Test
//	public void testGetMultiple() throws Exception {
//		insertRecord(new String[] {"nspLaegesekretaer", "40001", "Laegesekretaer"});
//		insertRecord(new String[] {"nspPortoer",       "40003", "Portoer"});
//
//		final List<RoleDefinition> roleDefinitions = service.getRoleDefinitions();
//		assertEquals(2, roleDefinitions.size());
//		roleDefinitions.sort(Comparator.comparing(RoleDefinition::getRoleId));
//
//		assertEquals("nspLaegesekretaer", roleDefinitions.get(0).getExternalName());
//		assertEquals("nspPortoer", roleDefinitions.get(1).getExternalName());
//	}
/*
	private void insertRecord(String[] r) throws SQLException {
		String sql = "INSERT INTO roleDefinitions VALUES ('NationalFederation', '" + r[0] + "', '" + r[1] + "', '" + r[2]+"')";
		try (Connection conn = dataSource.getConnection()) {
			conn.createStatement().execute(sql);
		}
	}
*/
}
