/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.roles.nationalroles;

import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.service.Uuid2RidService;
import org.hsqldb.jdbc.JDBCDataSource;
import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.PreparedStatementSetter;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.*;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class DbNationalRolesServiceTest {
	public static final String CVR = "30808460";
	public static final String RID = "42634739";
	public static final String SSN = "CVR:" + CVR + "-RID:" + RID;

	public static final String DDL_CREATE =
			"CREATE TABLE IF NOT EXISTS nationalRoles (\n" +
					"    RolePID BIGINT IDENTITY PRIMARY KEY,\n" +
					"    ssn VARCHAR(64) NOT NULL,\n" +
					"    externalName VARCHAR(255) NOT NULL," +
					"    CreatedDate DATETIME NOT NULL,\n" +
					"    ModifiedDate DATETIME NOT NULL,\n" +
					"    ValidFrom DATETIME NOT NULL,\n" +
					"    ValidTo DATETIME NOT NULL,\n\n" +
					")";

	public static final String DDL_DROP = "DROP TABLE nationalRoles if exists";

	private JDBCDataSource dataSource;
	private DbNationalRoleService service;
	private Uuid2RidService uuid2RidService;

	@Before
	public void setUp() {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		new JdbcTemplate(dataSource).update(DDL_CREATE);
		uuid2RidService = Mockito.mock(Uuid2RidService.class);
		service = new DbNationalRoleService();
		service.setDataSource(dataSource);
		service.setUuid2RidService(uuid2RidService);
	}

	@After
	public void tearDown() {
		new JdbcTemplate(dataSource).update(DDL_DROP);
	}

	@Test
	public void testEmpty() {
		OcesInfo ocesInfo = new OcesInfo(new SubjectSerialNumber("xx"), null, null, null, null);
		List<String> list = service.getRoles(ocesInfo);
		assertTrue(list.isEmpty());
	}

	@Test
	public void testNoMatchWithWrongSSN() throws Exception {
		insertRecord(SSN, "nspLaegesekretaer", new Date(), year2100());

		OcesInfo ocesInfo = new OcesInfo(new SubjectSerialNumber("xx"), null, null, null, null);
		List<String> list = service.getRoles(ocesInfo);
		assertTrue(list.isEmpty());
	}

	@Test
	public void testSinglematch() throws Exception {
		insertRecord(SSN, "nspLaegesekretaer", new Date(), year2100());

		OcesInfo ocesInfo = new OcesInfo(new SubjectSerialNumber(SSN), null, null, null, null);
		List<String> list = service.getRoles(ocesInfo);
		assertEquals(1, list.size());
		assertEquals("nspLaegesekretaer", list.get(0));
	}

	@Test
	public void testSinglematchUUID() throws Exception {
		String ssnUuid = "UI:DK-E:G:61eb1911-9f54-4a78-9c44-4efab71cdc1c";
		insertRecord(ssnUuid, "nspLaegesekretaer", new Date(), year2100());

		OcesInfo ocesInfo = new OcesInfo(new SubjectSerialNumber(ssnUuid), null, null, null, null);
		List<String> list = service.getRoles(ocesInfo);
		assertEquals(1, list.size());
		assertEquals("nspLaegesekretaer", list.get(0));
	}


	@Test
	public void testGetMultiple() throws Exception {
		insertRecord(SSN, "nspPortoer", new Date(), year2100());
		insertRecord(SSN, "nspLaegesekretaer", new Date(), year2100());

		OcesInfo ocesInfo = new OcesInfo(new SubjectSerialNumber(SSN), null, null, null, null);
		List<String> list = service.getRoles(ocesInfo);
		assertEquals(2, list.size());
		assertTrue(list.contains("nspLaegesekretaer"));
		assertTrue(list.contains("nspPortoer"));
	}

	@Test
	public void testExpiredAreIgnored() throws SQLException {
		long now = System.currentTimeMillis();
		final Date yesterYear = new Date(now - 365 * 24 * 60 * 60 * 1000);
		final Date anHourAgo = new Date(now - 1 * 60 * 60 * 1000);
		insertRecord(SSN, "nspPortoer", yesterYear, anHourAgo);

		OcesInfo ocesInfo = new OcesInfo(new SubjectSerialNumber(SSN), null, null, null, null);
		List<String> list = service.getRoles(ocesInfo);
		assertEquals(0, list.size());
	}

	@Test
	public void getRoles_WithUuidNotInDbNoRelatedRid_ReturnsEmptyList() throws Exception {
		String uuid = "UI:DK-E:G:f07daf8f-4f07-447e-a2b9-32cd15f6403c";
		insertRecord(SSN, "nspLaegesekretaer", new Date(), year2100());
		Mockito.when(uuid2RidService.findRelatedRid(uuid)).thenReturn(Optional.empty());

		OcesInfo ocesInfo = new OcesInfo(new SubjectSerialNumber(uuid), CVR, null, OcesInfo.Version.OCES3, null);
		List<String> list = service.getRoles(ocesInfo);

		assertTrue(list.isEmpty());
	}

	@Ignore("SDS-6660 - skal der testes for related CVR-RID?")
	@Test
	public void getRoles_WithNeitherUuidOrRelatedCvrRidInDb_ReturnsEmptyList() throws Exception {
		String uuid = "UI:DK-E:G:f07daf8f-4f07-447e-a2b9-32cd15f6403c";
		insertRecord(SSN, "nspLaegesekretaer", new Date(), year2100());
		Mockito.when(uuid2RidService.findRelatedRid(uuid)).thenReturn(Optional.of("xx"));

		OcesInfo ocesInfo = new OcesInfo(new SubjectSerialNumber(uuid), CVR, null, OcesInfo.Version.OCES3, null);
		List<String> list = service.getRoles(ocesInfo);

		assertTrue(list.isEmpty());
	}

	@Ignore("SDS-6660 - skal der testes for related CVR-RID?")
	@Test
	public void getRoles_WithUuidNotInDbButRelatedCvrRidInDb_ReturnsRoles() throws Exception {
		String uuid = "UI:DK-E:G:f07daf8f-4f07-447e-a2b9-32cd15f6403c";
		insertRecord(SSN, "nspPortoer", new Date(), year2100());
		insertRecord(SSN, "nspLaegesekretaer", new Date(), year2100());
		Mockito.when(uuid2RidService.findRelatedRid(uuid)).thenReturn(Optional.of(RID));

		OcesInfo ocesInfo = new OcesInfo(new SubjectSerialNumber(uuid), CVR, null, OcesInfo.Version.OCES3, null);
		List<String> list = service.getRoles(ocesInfo);

		assertEquals(2, list.size());
		assertTrue(list.contains("nspLaegesekretaer"));
		assertTrue(list.contains("nspPortoer"));
	}

	private void insertRecord(String ssn, String externalName, Date from, Date to) {
		String sql = "INSERT INTO nationalRoles (ssn, externalName, CreatedDate, ModifiedDate, ValidFrom, ValidTo) VALUES (?,?,?,?,?,?)";
		PreparedStatementSetter setter = preparedStatement -> {
            preparedStatement.setString(1, ssn);
            preparedStatement.setString(2, externalName);
            preparedStatement.setTimestamp(3, new Timestamp(System.currentTimeMillis()));
            preparedStatement.setTimestamp(4, new Timestamp(System.currentTimeMillis()));
            preparedStatement.setTimestamp(5, new Timestamp(from.getTime()));
            preparedStatement.setTimestamp(6, new Timestamp(to.getTime()));
        };
		new JdbcTemplate(dataSource).update(sql, setter);
	}

	private static Date year2100() {
		Calendar cal = GregorianCalendar.getInstance();
		cal.set(Calendar.YEAR, 2100);
		cal.set(Calendar.MONTH, 0);
		cal.set(Calendar.DAY_OF_MONTH, 1);
		return cal.getTime();
	}
}
