/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.roles.roledefinitions;

import static java.util.Optional.*;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.sql.SQLException;
import java.util.List;

import javax.sql.DataSource;

import dk.sosi.sts.server.facade.HttpTestClient;
import org.junit.*;
import org.springframework.dao.DuplicateKeyException;
import org.springframework.jdbc.CannotGetJdbcConnectionException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.PreparedStatementSetter;

import dk.sosi.sts.model.RoleDefinition;
import dk.sosi.sts.roles.DBTestUtil;
import dk.sosi.sts.server.facade.RoleDefinitionServiceStub;

// marked as IT since it uses mysql on docker
public class DbRoleDefinitionsIT {

	private DataSource dataSource;
	private DbRoleDefinitionsService service;

	private boolean available;

	@Before
	public void setUp() throws Exception {
		Assume.assumeTrue(HttpTestClient.testEnvironmentHost.contains("localhost"));

		dataSource = DBTestUtil.getTestDataSource("sts");
		service = new DbRoleDefinitionsService();
		service.setDataSource(dataSource);
		assumeMysqlAvailable();
	}

	private void assumeMysqlAvailable() {
		try {
			new JdbcTemplate(dataSource).execute("SELECT 1");
			available = true;
		} catch (CannotGetJdbcConnectionException e) {
			Assume.assumeFalse(Boolean.getBoolean("unittests.skip"));
		}
	}

	@After
	public void tearDown() throws SQLException {
		if (!available) 
			return;
		
		cleardata();
	}

	private void cleardata() throws SQLException {
		deleteRoleDefinition(RoleDefinitionServiceStub.UKENDT);
		deleteTrustedCvr("46837428", "XTX");
		deleteTrustedCvr("00000000", "STS");
	}

	@Test
	public void testSelectRoleDefinitions() throws Exception {
		List<RoleDefinition> roleDefinitions = service.getRoleDefinitions();
		
		assertEquals(5, roleDefinitions.size());
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.PORTOER)));
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.LAEGE_SEKRETAER)));
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.PLEJE)));
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.TESTROLLE1)));
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.TESTROLLE2)));
	}
	
	@Test
	public void testInsertAndSelectRoleDefinitions() throws Exception {
		insertRoleDefinition(RoleDefinitionServiceStub.UKENDT);

		List<RoleDefinition> roleDefinitions = service.getRoleDefinitions();
			
		assertEquals(6, roleDefinitions.size());

		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.PORTOER)));
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.LAEGE_SEKRETAER)));
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.PLEJE)));
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.TESTROLLE1)));
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.TESTROLLE2)));
		assertTrue(roleDefinitions.stream().anyMatch(roleDefinition -> roleDefinitionsEquals(roleDefinition, RoleDefinitionServiceStub.UKENDT)));
	}

	@Test(expected = DuplicateKeyException.class)
	public void testInsertDuplicateRoleDefinition() throws Exception {
		insertRoleDefinition(RoleDefinitionServiceStub.PORTOER);
	}
	
	@Test
	public void selectTrustedCvr() {
		List<String> trustedRoleClaimers = service.getTrustedRoleClaimers("STS", empty());

		assertEquals(3, trustedRoleClaimers.size());
		assertTrue(trustedRoleClaimers.contains("33257872"));
		assertTrue(trustedRoleClaimers.contains("46837428"));
		assertTrue(trustedRoleClaimers.contains("91026150"));
	}

	@Test
	public void selectTrustedCvrWithRole() {
		List<String> trustedRoleClaimers = service.getTrustedRoleClaimers("TestEndpoint", "TestRoleName");
		assertEquals(2, trustedRoleClaimers.size());
		assertTrue(trustedRoleClaimers.contains("70811176"));
		assertTrue(trustedRoleClaimers.contains("70811177"));
	}

	@Test
	public void selectTrustedCvrWithNoMatchingRole() {
		List<String> trustedRoleClaimers = service.getTrustedRoleClaimers("TestEndpoint", "NoMatch");
		assertEquals(1, trustedRoleClaimers.size());
		assertTrue(trustedRoleClaimers.contains("70811177"));
	}

	@Test
	public void selectTrustedCvrWithEmptyRole() {
		List<String> trustedRoleClaimers = service.getTrustedRoleClaimers("TestEndpoint", empty());
		assertEquals(2, trustedRoleClaimers.size());
		assertTrue(trustedRoleClaimers.contains("70811176"));
		assertTrue(trustedRoleClaimers.contains("70811177"));
	}
	
	@Test
	public void testInsertAndSelectTrustedCvr() {
		insertTrustedCvr("46837428", "XTX");
		
		List<String> trustedRoleClaimers = service.getTrustedRoleClaimers("XTX", empty());
			
		assertEquals(1, trustedRoleClaimers.size());
		assertEquals("46837428", trustedRoleClaimers.get(0));
		
		trustedRoleClaimers = service.getTrustedRoleClaimers("STS", empty());
		
		assertEquals(3, trustedRoleClaimers.size());
		assertTrue(trustedRoleClaimers.contains("33257872"));
		assertTrue(trustedRoleClaimers.contains("46837428"));

		insertTrustedCvr("00000000", "STS");
		
		trustedRoleClaimers = service.getTrustedRoleClaimers("XTX", empty());
		
		assertEquals(1, trustedRoleClaimers.size());
		assertEquals("46837428", trustedRoleClaimers.get(0));
		
		trustedRoleClaimers = service.getTrustedRoleClaimers("STS", empty());
		
		assertEquals(4, trustedRoleClaimers.size());
		assertTrue(trustedRoleClaimers.contains("00000000"));
		assertTrue(trustedRoleClaimers.contains("33257872"));
		assertTrue(trustedRoleClaimers.contains("46837428"));
	}
	
	private void insertRoleDefinition(RoleDefinition roleDefinition) throws SQLException {
		String sql = "INSERT INTO sts_audconf.roleDefinitions VALUES ('NationalFederation', ?, ?, ?)";
		executeInsertRoleDefinitionSql(sql, roleDefinition.getExternalName(), roleDefinition.getRoleId(), roleDefinition.getRoleName());
	}
	
	
	private void deleteRoleDefinition(RoleDefinition roleDefinition) throws SQLException {
		String sql = "DELETE FROM sts_audconf.roleDefinitions WHERE source='NationalFederation' and externalName=? and code=? and value=?";
		executeInsertRoleDefinitionSql(sql, roleDefinition.getExternalName(), roleDefinition.getRoleId(), roleDefinition.getRoleName());
	}
	
	private void executeInsertRoleDefinitionSql(String sql, String externalName, String roleId, String roleName) {
		JdbcTemplate template = new JdbcTemplate(dataSource);
		PreparedStatementSetter setter = preparedStatement -> {
			preparedStatement.setString(1, externalName);
			preparedStatement.setString(2, roleId);
			preparedStatement.setString(3, roleName);
		};
		template.update(sql, setter);
	}

	private void insertTrustedCvr(String cvr, String endpoint) {
		String sql = "INSERT INTO sts_audconf.trustedCvr (cvr, endpoint) VALUES (?, ?)";
		executeTrustedCvrSql(sql, cvr, endpoint);

	}
	
	private void deleteTrustedCvr(String cvr, String endpoint) {
		String sql = "DELETE FROM sts_audconf.trustedCvr WHERE cvr=? and endpoint=?";
		executeTrustedCvrSql(sql, cvr, endpoint);
	}
	
	private void executeTrustedCvrSql(String sql, String cvr, String endpoint) {
		JdbcTemplate template = new JdbcTemplate(dataSource);
		PreparedStatementSetter setter = preparedStatement -> {
			preparedStatement.setString(1, cvr);
			preparedStatement.setString(2, endpoint);
		};
		template.update(sql, setter);
	}

	private boolean roleDefinitionsEquals(RoleDefinition rd1, RoleDefinition rd2) {
		return rd1.getPublicName().equals(rd2.getPublicName()) && rd1.getExternalName().equals(rd2.getExternalName());
	}
}
