/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.roles.roledefinitions;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Comparator;
import java.util.List;

import org.hsqldb.jdbc.JDBCDataSource;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.springframework.jdbc.core.JdbcTemplate;

import dk.sosi.sts.model.RoleDefinition;

public class DbRoleDefinitionsServiceTest {

	public static final String DDL_CREATE =
		"CREATE TABLE roleDefinitions (\n" +
				" source VARCHAR(255) NOT NULL,\n" +
				" externalName VARCHAR(255) NOT NULL,\n" +
				" code VARCHAR(255) NOT NULL,\n" +
				" value VARCHAR(255))";

	public static final String DDL_DROP = "DROP TABLE roleDefinitions";

	private JDBCDataSource dataSource;
	private DbRoleDefinitionsService service;

	@Before
	public void setUp() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		Connection conn = dataSource.getConnection();
		Statement statement = conn.createStatement();
		statement.execute(DDL_CREATE);
		conn.close();
		service = new DbRoleDefinitionsService();
		service.setDataSource(dataSource);
		service.setSql("select * FROM roleDefinitions where source = ?");
	}
	
	@After
	public void tearDown() throws Exception {
		new JdbcTemplate(dataSource).execute("delete from roleDefinitions");
		Connection conn = dataSource.getConnection();
		conn.createStatement().execute(DDL_DROP);
		conn.close();		
	}

	@Test
	public void testEmpty() throws Exception {
		List<RoleDefinition> list = service.getRoleDefinitions();
		assertTrue(list.isEmpty());
	}

	@Test
	public void testSingleSelect() throws Exception {
		insertRecord(new String[] {"nspLaegesekretaer", "40001", "Laegesekretaer"});

		final List<RoleDefinition> roleDefinitions = service.getRoleDefinitions();
		assertEquals(1, roleDefinitions.size());
		assertEquals("nspLaegesekretaer", roleDefinitions.get(0).getExternalName());
		assertEquals("urn:dk:healthcare:national-federation-role:code:40001:value:Laegesekretaer", roleDefinitions.get(0).getPublicName());
	}

    @Test
	public void testGetMultiple() throws Exception {
		insertRecord(new String[] {"nspLaegesekretaer", "40001", "Laegesekretaer"});
		insertRecord(new String[] {"nspPortoer",       "40003", "Portoer"});

		final List<RoleDefinition> roleDefinitions = service.getRoleDefinitions();
		assertEquals(2, roleDefinitions.size());
		roleDefinitions.sort(Comparator.comparing(RoleDefinition::getRoleId));

		assertEquals("nspLaegesekretaer", roleDefinitions.get(0).getExternalName());
		assertEquals("nspPortoer", roleDefinitions.get(1).getExternalName());
	}

	private void insertRecord(String[] r) throws SQLException {
		String sql = "INSERT INTO roleDefinitions VALUES ('NationalFederation', '" + r[0] + "', '" + r[1] + "', '" + r[2]+"')";
		try (Connection conn = dataSource.getConnection()) {
			conn.createStatement().execute(sql);
		}
	}
}
