/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.*;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.StsException;
import dk.sosi.sts.acl.MockAclService;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.model.Authorization;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.ocescpr.cvrrid.service.impl.MockOcesCvrRidService;
import dk.sosi.sts.ocescpr.uuid.service.impl.MockOcesUuidService;
import dk.sosi.sts.server.facade.MockServicesImpl;
import dk.sosi.sts.server.facade.STSMessageTest;
import dk.sosi.sts.server.test.*;
import dk.sosi.sts.service.AclService;
import dk.sosi.sts.service.AuthorizationService;
import org.junit.After;
import org.junit.Assume;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;

import java.lang.reflect.Field;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

import static org.junit.Assert.*;
import static org.mockito.Mockito.mock;

/**
 * See also {@link STSMessageTest} and Seal testtools.
 */
@RunWith(Parameterized.class)
public class SecurityTokenServiceTest {
	private static final String BLACKLISTED_IT_SYSTEM_NAME = "blacklisted";

	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	private SystemInfo systemInfo;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3);

		return builder.build();
	}

	@Before
	public void setUp() throws Exception {
		TestUtil.disableLogging();
		// create test data
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "cvr", "navn");
		systemInfo = new SystemInfo(careProvider, "TE-STS");
		AuditHelper.init();
	}

	@After
	public void tearDown() {
		TestUtil.revertLogging();
		AuditHelper.destroy();

	}

	@Test
	public void testVerifyNoAuth() {
		try {
			IDCard idCard = createSystemIdCard(systemInfo, AuthenticationLevel.NO_AUTHENTICATION);
			new SecurityTokenService(TestUtil.createStsSosiFactory(version), mock(AclService.class), mock(UserValidationService.class), "dummy blacklisted ITSystemName").verify(idCard);
			fail("must fail not allowed");
		} catch (VerifyException e) {
			assertVerifyException(e, "not supported", StsFaultCodeValues.WST_BAD_REQUEST, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testVerify_UsernamePassword() {
		try {
			IDCard idCard = TestUtil.createStsSosiFactory(version).createNewSystemIDCard(systemInfo.getITSystemName(), systemInfo.getCareProvider(), AuthenticationLevel.USERNAME_PASSWORD_AUTHENTICATION, "username", "password", null, null);
			new SecurityTokenService(TestUtil.createStsSosiFactory(version), mock(AclService.class), mock(UserValidationService.class), "dummy blacklisted ITSystemName").verify(idCard);
			fail("Authentication level username/password not allowed");
		} catch (VerifyException e) {
			assertVerifyException(e, "not supported", StsFaultCodeValues.WST_BAD_REQUEST, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testVerifyIdCard_CreatedInFuture() {
		IDCard idCard = createSystemIdCard(systemInfo, AuthenticationLevel.NO_AUTHENTICATION);
		setPrivateField(IDCard.class, idCard, "createdDate", new Date(idCard.getCreatedDate().getTime() + TimeUnit.DAYS.toMillis(1)));
		SecurityTokenService sts = createMockedSts();

		try {
			sts.verifyIdCard(idCard);
		} catch (VerifyException e) {
			assertVerifyException(e, "system time", StsFaultCodeValues.WST_INVALID_TIME_RANGE, StsFaultActorValues.STS);
		}
	}

	@Test
	public void verify_WithBlacklistedITSystemName_ThrowsException() {
		SystemIDCard idCard = createSystemIdCard(new SystemInfo(new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "cvr", "navn"), BLACKLISTED_IT_SYSTEM_NAME), AuthenticationLevel.VOCES_TRUSTED_SYSTEM);
		SecurityTokenService sts = createMockedSts();

		VerifyException e = assertThrows(VerifyException.class, () -> sts.verify(idCard));
		assertEquals(StsFaultStringValues.WST_BAD_REQUEST + ": Invalid ITSystemName", e.getFaultString());
	}

	@Test
	public void testVerifyIdCard_ExpiresInPast() {
		IDCard idCard = createSystemIdCard(systemInfo, AuthenticationLevel.NO_AUTHENTICATION);
		setPrivateField(IDCard.class, idCard, "createdDate", new Date(idCard.getCreatedDate().getTime() - TimeUnit.DAYS.toMillis(1)));
		setPrivateField(IDCard.class, idCard, "expiryDate", new Date(idCard.getExpiryDate().getTime() - TimeUnit.DAYS.toMillis(1)));
		SecurityTokenService sts = createMockedSts();

		try {
			sts.verifyIdCard(idCard);
		} catch (VerifyException e) {
			assertVerifyException(e, "system time", StsFaultCodeValues.WST_INVALID_TIME_RANGE, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testVerifyIdCard_ValidityExceeded() {
		IDCard idCard = createSystemIdCard(systemInfo, AuthenticationLevel.NO_AUTHENTICATION);
		setPrivateField(IDCard.class, idCard, "expiryDate", new Date(idCard.getExpiryDate().getTime() + TimeUnit.DAYS.toMillis(1)));
		SecurityTokenService sts = createMockedSts();

		try {
			sts.verifyIdCard(idCard);
		} catch (VerifyException e) {
			assertVerifyException(e, "maximum validity", StsFaultCodeValues.WST_INVALID_TIME_RANGE, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testValidVoces() {
		SOSIFactory stsSosiFactory = TestUtil.createStsSosiFactory(version);
		GenericCredentialVault clientCredentialVault = CertificateStore.getValidVocesVault(version);
		SOSIFactory clientSosiFactory = TestUtil.createSOSIFactory(clientCredentialVault, false, null);
		OcesInfo clientOcesInfo = OcesInfoParser.parseCertificate(clientCredentialVault.getSystemCredentialPair().getCertificate());

		SecurityTokenRequest req = clientSosiFactory.createNewSecurityTokenRequest();
		IDCard idcard = clientSosiFactory.createNewSystemIDCard("SOSITEST", new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, clientOcesInfo.getCvr(), "orgName"), AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, null);
		req.setIDCard(idcard);

		req.serialize2DOMDocument(); // OBS: signing is done here for voces...

		SecurityTokenService sts = new SecurityTokenService(stsSosiFactory, new MockAclService(null, null), mock(UserValidationService.class), "dummy blacklisted ITSystemName");
		SecurityTokenResponse response = sts.issueIdCard(req);
		assertFalse(response.isFault());
	}

	@Test
	public void testUserIdCardSignedWithVocesIsAllowed() {
		SOSIFactory stsSosiFactory = TestUtil.createStsSosiFactory(version);
		GenericCredentialVault clientCredentialVault = CertificateStore.getValidVocesVault(version);
		SOSIFactory clientSosiFactory = TestUtil.createSOSIFactory(clientCredentialVault, false, null);
		OcesInfo clientOcesInfo = OcesInfoParser.parseCertificate(clientCredentialVault.getSystemCredentialPair().getCertificate());

		SecurityTokenRequest req = clientSosiFactory.createNewSecurityTokenRequest();

		IDCard userIdCard = clientSosiFactory.createNewUserIDCard("SOSITEST", new UserInfo("1111111118", "Anders", "And", "aand@andeby.dk", "coolDuck", "DOCTOR", "19901"), new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, clientOcesInfo.getCvr(), "orgName"),
				AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, null);
		req.setIDCard(userIdCard);

		req.serialize2DOMDocument(); // OBS: signing is done here for voces...

		SecurityTokenService sts = new SecurityTokenService(stsSosiFactory, new MockAclService(null, null), mock(UserValidationService.class), "dummy blacklisted ITSystemName");

		sts.setAllowUserIDCardWithVoces(true);

		SecurityTokenResponse response = sts.issueIdCard(req);

		if (response.isFault()) {
			fail(response.getFaultString());
		}
	}

	@Test
	public void testUserIdCardSignedWithVocesIsNotAllowed() {
		SOSIFactory stsSosiFactory = TestUtil.createStsSosiFactory(version);
		GenericCredentialVault clientCredentialVault = CertificateStore.getValidVocesVault(version);
		SOSIFactory clientSosiFactory = TestUtil.createSOSIFactory(clientCredentialVault, false, null);
		OcesInfo clientOcesInfo = OcesInfoParser.parseCertificate(clientCredentialVault.getSystemCredentialPair().getCertificate());

		SecurityTokenRequest req = clientSosiFactory.createNewSecurityTokenRequest();

		IDCard userIdCard = clientSosiFactory.createNewUserIDCard("SOSITEST", new UserInfo("1111111118", "Anders", "And", "aand@andeby.dk", "coolDuck", "DOCTOR", "19901"), new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, clientOcesInfo.getCvr(), "orgName"),
				AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, null);
		req.setIDCard(userIdCard);

		req.serialize2DOMDocument(); // OBS: signing is done here for voces...

		SecurityTokenService sts = new SecurityTokenService(stsSosiFactory, new MockAclService(null, null), mock(UserValidationService.class), "dummy blacklisted ITSystemName");

		sts.setAllowUserIDCardWithVoces(false);

		SecurityTokenResponse response = sts.issueIdCard(req);

		if (!response.isFault()) {
			fail("Forventede fejl.");
		}

		assertEquals("wst:BadRequest", response.getFaultCode());
		assertEquals("The specified RequestSecurityToken is not understood: ID-Card is a User ID-Card signed with VOCES/FOCES", response.getFaultString());
		assertEquals("dk:sosi:sts", response.getFaultActor());
	}

	@Test
	public void testValidFoces() {
		CredentialVault vault = CertificateStore.getValidFocesVault(version);
		String certificateCvr = OcesInfoParser.parseCertificate(vault.getSystemCredentialPair().getCertificate()).getCvr();
		SOSIFactory sosiFactory = TestUtil.createSOSIFactory(vault, false, null);

		SecurityTokenRequest req = sosiFactory.createNewSecurityTokenRequest();
		IDCard idcard = sosiFactory.createNewSystemIDCard("SOSITEST", new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, certificateCvr, "orgName"), AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, null);
		req.setIDCard(idcard);

		req.serialize2DOMDocument(); // OBS: signing is done here

		SecurityTokenService sts = new SecurityTokenService(sosiFactory, new MockAclService(null, null), mock(UserValidationService.class), "dummy blacklisted ITSystemName");
		SecurityTokenResponse r = sts.issueIdCard(req);

		if (r.isFault()) {
			fail(r.getFaultString());
		}
	}

	@Test
	public void testValidMoces() {
		CredentialVault vault = SealPPCredentialVaults.getValidMocesWithCpr(version);
		OcesInfo ocesInfo = OcesInfoParser.parseCertificate(vault.getSystemCredentialPair().getCertificate());
		SOSIFactory clientFactory = TestUtil.createSOSIFactory(vault, false, null);

		SecurityTokenRequest req = clientFactory.createNewSecurityTokenRequest();
		IDCard idcard = clientFactory.createNewUserIDCard("SOSITEST", new UserInfo("1111111118", "Anders", "And", "aand@andeby.dk", "coolDuck", "DOCTOR", "19901"), new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, ocesInfo.getCvr(), "orgName"),
				AuthenticationLevel.MOCES_TRUSTED_USER, null, null, null, null);

		req.setIDCard(idcard);

		Document doc = req.serialize2DOMDocument();
		SignatureUtil.sign(new String[] { IDValues.IDCARD }, clientFactory.getCredentialVault().getSystemCredentialPair(), doc, IDValues.IDCARD);

		MockOcesCvrRidService cvrRidMockService = new MockOcesCvrRidService();
		cvrRidMockService.add(ocesInfo.getSubjectSerialNumber().getValue(), "1111111118");

		MockOcesUuidService uuidMockService = new MockOcesUuidService();
		uuidMockService.add(ocesInfo.getSubjectSerialNumber().getValue(), "1111111118");

		String xml = XmlUtil.node2String(doc, false, true);
		final SOSIFactory sosiFactory = TestUtil.createStsSosiFactory(version);
		SecurityTokenRequest req2 = sosiFactory.deserializeSecurityTokenRequest(xml);
		final AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) throws StsException {
				return Arrays.asList(new Authorization("19901", "7170"));
			}
		};
		new MockServicesImpl(version);
		SecurityTokenService sts = new SecurityTokenService(sosiFactory, new MockAclService(), UserValidationServiceTestUtil.createUserValidationService(version, cvrRidMockService, authorizationService, uuidMockService), "dummy blacklisted ITSystemName"
		// new UserValidationService(cvrRidMockService, authorizationService,
		// roleDefinitionService, new MockNationalRolesService())
		);
		SecurityTokenResponse r = sts.issueIdCard(req2);

		assertFalse(toString(r), r.isFault());
	}

	@Test
	public void testCheckIssuer_MocesValid() throws Exception {
		X509Certificate validMoces = CertificateStore.getValidMoces(version);

		final Date now = new Date();
		SecurityTokenService sts = createMockedSts(new CertificateStatusChecker() {
			@Override
			public CertificateStatus getRevocationStatus(X509Certificate certificate) throws PKIException {
				return new CertificateStatus(true, now);
			}
		});
		sts.checkIssuer(validMoces, StsFaultActorValues.STS);
	}

	@Test
	public void testCheckIssuer_MocesInvalidIssuer() throws Exception {
		X509Certificate moces = SealTestHelper.issueCertificate();

		SecurityTokenService sts = createMockedSts(new CertificateStatusChecker() {
			@Override
			public CertificateStatus getRevocationStatus(X509Certificate certificate) throws PKIException {
				return new CertificateStatus(true, new Date());
			}
		});
		try {
			sts.checkIssuer(moces, StsFaultActorValues.STS);
			fail("Expected expiration");
		} catch (VerifyException e) {
			assertVerifyException(e, "invalid", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testCheckIssuer_MocesExpired() throws Exception {
		X509Certificate revokedMoces = CertificateStore.getExpiredMoces(version);

		SecurityTokenService sts = createMockedSts(new CertificateStatusChecker() {
			@Override
			public CertificateStatus getRevocationStatus(X509Certificate certificate) throws PKIException {
				return new CertificateStatus(false, new Date());
			}
		});
		try {
			sts.checkIssuer(revokedMoces, StsFaultActorValues.STS);
			fail("Expected expiration");
		} catch (VerifyException e) {
			assertVerifyException(e, "expire", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testCheckIssuer_MocesRevoked() throws Exception {
		X509Certificate revokedMoces = CertificateStore.getRevokedMoces(version);

		SecurityTokenService sts = createMockedSts(new CertificateStatusChecker() {
			@Override
			public CertificateStatus getRevocationStatus(X509Certificate certificate) throws PKIException {
				return new CertificateStatus(false, new Date());
			}
		});
		try {
			sts.checkIssuer(revokedMoces, StsFaultActorValues.STS);
			fail("Expected revokation");
		} catch (VerifyException e) {
			assertVerifyException(e, "revoked", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testCheckCertificateType_MOCES() throws Exception {
		SecurityTokenService sts = createMockedSts();
		OcesInfo info = OcesInfoParser.parseCertificate(CertificateStore.getValidMoces(version));

		sts.checkCertificateType(false, info);

		try {
			sts.checkCertificateType(true, info);
			fail("Expected exception");
		} catch (VerifyException e) {
			assertVerifyException(e, "VOCES", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testCheckCertificateType_NonGlobalMOCES3() throws Exception {
		Assume.assumeTrue(version == Version.OCES3);
		SecurityTokenService sts = createMockedSts();
		X509Certificate certificate = TestIdentityProvider.getNonGlobalMoces3Identity().getCredentialVault().getSystemCredentialPair().getCertificate();
		OcesInfo info = OcesInfoParser.parseCertificate(certificate);

		try {
			sts.checkCertificateType(false, info);
			fail("Expected exception");
		} catch (VerifyException e) {
			assertVerifyException(e, "signed with unsupported certificate - persistence level of moces3 certificates must be global", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testCheckCertificateType_VOCES() throws Exception {
		SecurityTokenService sts = createMockedSts();
		OcesInfo info = OcesInfoParser.parseCertificate(CertificateStore.getValidVoces(version));

		sts.checkCertificateType(true, info);

		try {
			sts.checkCertificateType(false, info);
			fail("Expected exception");
		} catch (VerifyException e) {
			assertVerifyException(e, "MOCES", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testCheckCertificateType_FOCES() throws Exception {
		SecurityTokenService sts = createMockedSts();
		OcesInfo info = OcesInfoParser.parseCertificate(CertificateStore.getValidFoces(version));

		sts.checkCertificateType(true, info);

		try {
			sts.checkCertificateType(false, info);
			fail("Expected exception");
		} catch (VerifyException e) {
			assertVerifyException(e, "MOCES", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testCheckValidCertificate() throws Exception {
		SecurityTokenService sts = createMockedSts();

		X509Certificate moces = CertificateStore.getValidMoces(version);
		sts.checkValidCertificate(moces);
	}

	@Test
	public void testCkeckValidCertificate_Expired() throws Exception {
		SecurityTokenService sts = createMockedSts();

		X509Certificate moces = CertificateStore.getExpiredMoces(version);
		try {
			sts.checkValidCertificate(moces);
			fail("Expected excption");
		} catch (VerifyException e) {
			assertVerifyException(e, "expired", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testVerifySystemInfo_ValidCvr() throws Exception {
		SecurityTokenService sts = createMockedSts();

		OcesInfo oces = OcesInfoParser.parseCertificate(CertificateStore.getValidMoces(version));
		SystemInfo system = new SystemInfo(new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, oces.getCvr(), "JERNALDERBYENS VENNER"), "STS-TEST");
		sts.verifySystemInfo(system, oces);
	}

	@Test
	public void testVerifySystemInfo_InvalidCvr() throws Exception {
		SecurityTokenService sts = createMockedSts();

		OcesInfo oces = OcesInfoParser.parseCertificate(CertificateStore.getValidMoces(version));
		SystemInfo system = new SystemInfo(new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "invalid-cvr", "JERNALDERBYENS VENNER"), "STS-TEST");

		try {
			sts.verifySystemInfo(system, oces);
			fail("Expected excption");
		} catch (VerifyException e) {
			assertVerifyException(e, "cvr", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.STS);
		}
	}

	@Test
	public void testIssueIdCard_WithMOCES3AndReplaceNameIdSet_DoesNotReplaceNameId() throws Exception {
		Assume.assumeTrue(version == Version.OCES3);
		final String setNameId = "NameId";

		CredentialVault vault = SealPPCredentialVaults.getValidMocesWithCpr(version);
		OcesInfo ocesInfo = OcesInfoParser.parseCertificate(vault.getSystemCredentialPair().getCertificate());
		SOSIFactory clientFactory = TestUtil.createSOSIFactory(vault, false, null);

		SecurityTokenRequest req = clientFactory.createNewSecurityTokenRequest();
		IDCard idcard = clientFactory.createNewUserIDCard("SOSITEST", new UserInfo("1111111118", "Anders", "And", "aand@andeby.dk", "coolDuck", "DOCTOR", "19901"), new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, ocesInfo.getCvr(), "orgName"),
				AuthenticationLevel.MOCES_TRUSTED_USER, null, null, null, setNameId);

		req.setIDCard(idcard);

		Document doc = req.serialize2DOMDocument();
		SignatureUtil.sign(new String[] { IDValues.IDCARD }, clientFactory.getCredentialVault().getSystemCredentialPair(), doc, IDValues.IDCARD);

		MockOcesCvrRidService cvrRidMockService = new MockOcesCvrRidService();
		cvrRidMockService.add(ocesInfo.getSubjectSerialNumber().getValue(), "1111111118");

		MockOcesUuidService uuidMockService = new MockOcesUuidService();
		uuidMockService.add(ocesInfo.getSubjectSerialNumber().getValue(), "1111111118");

		String xml = XmlUtil.node2String(doc, false, true);
		final SOSIFactory sosiFactory = TestUtil.createStsSosiFactory(version);
		SecurityTokenRequest req2 = sosiFactory.deserializeSecurityTokenRequest(xml);
		final AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) throws StsException {
				return Arrays.asList(new Authorization("19901", "7170"));
			}
		};
		new MockServicesImpl(version);
		SecurityTokenService sts = new SecurityTokenService(sosiFactory, new MockAclService(), UserValidationServiceTestUtil.createUserValidationService(version, cvrRidMockService, authorizationService, uuidMockService), "dummy blacklisted ITSystemName"
		// new UserValidationService(cvrRidMockService, authorizationService,
		// roleDefinitionService, new MockNationalRolesService())
		);
		sts.setReplaceNameId(true);
		SecurityTokenResponse response = sts.issueIdCard(req2);

		// Check that the nameId was not changed
		assertEquals(setNameId, response.getIDCard().getAlternativeIdentifier());
	}

	@Test
	public void testCreateUserInfo_EmptyVerifyResult() {
		UserInfo template = new UserInfo("1111111118", "Hans", "Hansen", "hans.hansen@example.com", "occupation", "role", "code");
		VerifyResult result = new VerifyResult();
		ModelException modelException = assertThrows("Forventede ModelException når SecurityTokenService.createUserInfo kaldes med VerifyResult uden data.", ModelException.class, () -> SecurityTokenService.createUserInfo(template, result));
		assertEquals("UserRole must be specified", modelException.getMessage());
	}

	@Test
	public void testCreateUserInfo_Changed() {
		UserInfo template = new UserInfo("1111111118", "Hans", "Hansen", "hans.hansen@example.com", "occupation", "role", "code");
		VerifyResult result = new VerifyResult();
		result.cpr = new CprNr("1111111119");
		result.role = "role2";
		result.authorizationCode = "code2";

		UserInfo info = SecurityTokenService.createUserInfo(template, result);

		assertFalse(template.equals(info));
		assertEquals(result.getAuthorizationCode().get(), info.getAuthorizationCode());
		assertEquals(result.getRole().get(), info.getRole());
		assertEquals(result.cpr.getValue(), info.getCPR());
		assertEquals(template.getEmail(), info.getEmail());
		assertEquals(template.getGivenName(), info.getGivenName());
		assertEquals(template.getOccupation(), info.getOccupation());
		assertEquals(template.getSurName(), info.getSurName());
	}




	/**
	 * SYSID-1
	 * Input: (F/V)OCES3-signeret systemidkort-request med NameID, som ikke er i 'certifikat-info' format
	 * Endpoint: SecurityTokenService
	 * Output: Response med systemidkort med samme NameID som i input
	 */
	@Test
	public void testReplaceNameIdNotSetWithNonCertificateNameId() {
		final String setNameId = "NameId";
		final SecurityTokenService sts = createMockedSts();
		sts.setReplaceNameId(false);

		GenericCredentialVault clientVault = CertificateStore.getValidVocesVault(version);
		OcesInfo clientOces = OcesInfoParser.parseCertificate(clientVault.getSystemCredentialPair().getCertificate());
		SOSIFactory sosiFactory = TestUtil.createSOSIFactory(clientVault, false, null);

		SecurityTokenRequest req = sosiFactory.createNewSecurityTokenRequest();
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, clientOces.getCvr(), "orgName");
		IDCard idcard = sosiFactory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, setNameId);
		req.setIDCard(idcard);

		Document doc = req.serialize2DOMDocument(); // OBS: signing is done here for voces...

		SecurityTokenResponse response = sts.issueIdCard(req);
		if (response.isFault()) {
			fail(response.getFaultString());
		}

		// Check that the nameId was not changed
		final String nameId = response.getIDCard().getAlternativeIdentifier();
		assertEquals(setNameId, nameId);
	}

	/**
	 * SYSID-2
	 * Input: (F/V)OCES3-signeret systemidkort-request med NameID i 'certifikat-info' format, som matcher (F/V)OCES3 certifikatet
	 * Endpoint: SecurityTokenService
	 * Output: Response med systemidkort med samme NameID som i input (dvs. i 'certifikat-info' format)
	 */
	@Test
	public void testReplaceNameIdNotSetWithCertificateNameIdMatchingCertificate() {
		final SecurityTokenService sts = createMockedSts();
		sts.setReplaceNameId(false);

		GenericCredentialVault clientVault = CertificateStore.getValidVocesVault(version);
		OcesInfo clientOces = OcesInfoParser.parseCertificate(clientVault.getSystemCredentialPair().getCertificate());
		String setNameId = new CertificateInfo(clientOces.getCertificate()).toString();
		SOSIFactory sosiFactory = TestUtil.createSOSIFactory(clientVault, false, null);

		SecurityTokenRequest req = sosiFactory.createNewSecurityTokenRequest();
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, clientOces.getCvr(), "orgName");
		IDCard idcard = sosiFactory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, setNameId);
		req.setIDCard(idcard);

		Document doc = req.serialize2DOMDocument(); // OBS: signing is done here for voces...

		SecurityTokenResponse response = sts.issueIdCard(req);
		if (response.isFault()) {
			fail(response.getFaultString());
		}

		// Check that the nameId was not changed
		final String nameId = response.getIDCard().getAlternativeIdentifier();
		assertEquals(setNameId, nameId);
	}

	/**
	 * SYSID-3
	 * Input: (F/V)OCES3-signeret systemidkort-request med NameID i 'certifikat-info' format, som ikke matcher (F/V)OCES3 certifikatet (fx forkert UUID eller cert. serial)
	 * Endpoint: SecurityTokenService
	 * Output: Fejlsvar ("The specified RequestSecurityToken is not understood: IDCard nameId must match information from certificate.")
	 */
	@Test
	public void testReplaceNameIdNotSetWithCertificateNameIdNotMatchingCertificate() {
		final SecurityTokenService sts = createMockedSts();
		sts.setReplaceNameId(false);

		GenericCredentialVault clientVault = CertificateStore.getValidVocesVault(version);
		OcesInfo clientOces = OcesInfoParser.parseCertificate(clientVault.getSystemCredentialPair().getCertificate());
		String setNameId = new CertificateInfo(clientOces.getCertificate()).toString().replaceAll("CertSerial=\\{[0-9]*\\}", "CertSerial={000000}");
		SOSIFactory sosiFactory = TestUtil.createSOSIFactory(clientVault, false, null);

		SecurityTokenRequest req = sosiFactory.createNewSecurityTokenRequest();
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, clientOces.getCvr(), "orgName");
		IDCard idcard = sosiFactory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, setNameId);
		req.setIDCard(idcard);

		Document doc = req.serialize2DOMDocument(); // OBS: signing is done here for voces...

		SecurityTokenResponse response = sts.issueIdCard(req);
		assertTrue(response.isFault());
		assertTrue(response.getFaultString().contains("IDCard nameId must match information from certificate"));
	}

	/**
	 * SYSID-4
	 * Input: (F/V)OCES3-signeret systemidkort-request med NameID, som ikke er i 'certifikat-info' format
	 * Endpoint: NewSecurityTokenService
	 * Output: Response med systemidkort med NameID i certifikat-info format, som matcher certifikatet
	 */
	@Test
	public void testReplaceNameIdSetWithNonCertificateNameId() {
		final String setNameId = "NameId";
		final SecurityTokenService sts = createMockedSts();
		sts.setReplaceNameId(true);

		GenericCredentialVault clientVault = CertificateStore.getValidVocesVault(version);
		OcesInfo clientOces = OcesInfoParser.parseCertificate(clientVault.getSystemCredentialPair().getCertificate());
		SOSIFactory sosiFactory = TestUtil.createSOSIFactory(clientVault, false, null);

		SecurityTokenRequest req = sosiFactory.createNewSecurityTokenRequest();
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, clientOces.getCvr(), "orgName");
		IDCard idcard = sosiFactory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, setNameId);
		req.setIDCard(idcard);

		Document doc = req.serialize2DOMDocument(); // OBS: signing is done here for voces...

		SecurityTokenResponse response = sts.issueIdCard(req);
		if (response.isFault()) {
			fail(response.getFaultString());
		}

		// Check that the nameId was not changed
		final String nameId = response.getIDCard().getAlternativeIdentifier();
		assertEquals(new CertificateInfo(clientOces.getCertificate()).toString() , nameId);
	}

	/**
	 * SYSID-5
	 * Input: (F/V)OCES3-signeret systemidkort-request med NameID i 'certifikat-info' format, som matcher (F/V)OCES3 certifikatet
	 * Endpoint: NewSecurityTokenService
	 * Output: Response med systemidkort med NameID i certifikat-info format, som matcher certifikatet
	 */
	@Test
	public void testReplaceNameIdSetWithCertificateNameIdMatchingCertificate() {
		final SecurityTokenService sts = createMockedSts();
		sts.setReplaceNameId(true);

		GenericCredentialVault clientVault = CertificateStore.getValidVocesVault(version);
		OcesInfo clientOces = OcesInfoParser.parseCertificate(clientVault.getSystemCredentialPair().getCertificate());
		String setNameId = new CertificateInfo(clientOces.getCertificate()).toString();
		SOSIFactory sosiFactory = TestUtil.createSOSIFactory(clientVault, false, null);

		SecurityTokenRequest req = sosiFactory.createNewSecurityTokenRequest();
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, clientOces.getCvr(), "orgName");
		IDCard idcard = sosiFactory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, setNameId);
		req.setIDCard(idcard);

		Document doc = req.serialize2DOMDocument(); // OBS: signing is done here for voces...

		SecurityTokenResponse response = sts.issueIdCard(req);
		if (response.isFault()) {
			fail(response.getFaultString());
		}

		// Check that the nameId was not changed
		final String nameId = response.getIDCard().getAlternativeIdentifier();
		assertEquals(setNameId, nameId);
	}

	/**
	 * SYSID-6
	 * Input: (F/V)OCES3-signeret systemidkort-request med NameID i 'certifikat-info' format, som ikke matcher (F/V)OCES3 certifikatet (fx forkert UUID eller cert. serial)
	 * Endpoint: NewSecurityTokenService
	 * Output: Response med systemidkort med NameID i 'certifikat-info' format, som matcher certifikatet
	 */
	@Test
	public void testReplaceNameIdSetWithCertificateNameIdNotMatchingCertificate() {
		final SecurityTokenService sts = createMockedSts();
		sts.setReplaceNameId(true);

		GenericCredentialVault clientVault = CertificateStore.getValidVocesVault(version);
		OcesInfo clientOces = OcesInfoParser.parseCertificate(clientVault.getSystemCredentialPair().getCertificate());
		String setNameId = new CertificateInfo(clientOces.getCertificate()).toString().replaceAll("CertSerial=\\{[0-9]*\\}", "CertSerial={000000}");
		SOSIFactory sosiFactory = TestUtil.createSOSIFactory(clientVault, false, null);

		SecurityTokenRequest req = sosiFactory.createNewSecurityTokenRequest();
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, clientOces.getCvr(), "orgName");
		IDCard idcard = sosiFactory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, setNameId);
		req.setIDCard(idcard);

		Document doc = req.serialize2DOMDocument(); // OBS: signing is done here for voces...

		SecurityTokenResponse response = sts.issueIdCard(req);
		if (response.isFault()) {
			fail(response.getFaultString());
		}

		// Check that the nameId was not changed
		final String nameId = response.getIDCard().getAlternativeIdentifier();
		assertEquals(new CertificateInfo(clientOces.getCertificate()).toString(), nameId);
	}

	private void assertVerifyException(VerifyException e, String faultStringPart, String faultCode, String faultActor) {
		assertTrue(e.getFaultString(), e.getFaultString().toLowerCase().contains(faultStringPart.toLowerCase()));
		assertEquals(faultCode, e.getFaultCode());
		assertEquals(faultActor, e.getFaultActor());
	}

	private SecurityTokenService createMockedSts() {
		MockServicesImpl services = new MockServicesImpl(version);
		SOSIFactory sosiFactory = services.buildSosiFactory(false);
		return new SecurityTokenService(sosiFactory, services, UserValidationServiceTestUtil.createUserValidationService(version, services.getCvrRidService(), new AuthorizationServiceStub(), services.getUUID2CPRService()), BLACKLISTED_IT_SYSTEM_NAME);
	}

	private SecurityTokenService createMockedSts(final CertificateStatusChecker services) {
		SOSITestFederation federation = new SOSITestFederation(new Properties(), StsHashMapCertificateCacheFactory.create(), services);
		SOSIFactory sosiFactory = new SOSIFactory(federation, StsCredentialVaults.getStsCredentialVault(version), new Properties());
		MockServicesImpl mockServices = new MockServicesImpl(version);
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, mockServices.getCvrRidService(), new AuthorizationServiceStub(), mockServices.getUUID2CPRService());
		return new SecurityTokenService(sosiFactory, new MockAclService(null, null), userValidationService, "dummy blacklisted ITSystemName");
	}

	private String toString(SecurityTokenResponse r) {
		String s = "";
		if (r.isFault()) {
			s = "[code:" + r.getFaultCode() + ", fault:" + r.getFaultString() + ", actor: " + r.getFaultActor() + "]";
		}
		return s;
	}

	private SystemIDCard createSystemIdCard(SystemInfo si, AuthenticationLevel level) {
		return createSystemIdCard(si, level, null);
	}

	private SystemIDCard createSystemIdCard(SystemInfo si, AuthenticationLevel level, String alternativeIdentifier) {
		return TestUtil.createStsSosiFactory(version).createNewSystemIDCard(si.getITSystemName(), si.getCareProvider(), level, null, null, null, alternativeIdentifier);
	}

	public static void setPrivateField(Class<?> clz, Object obj, String nameOfField, Object value) {
		try {
			Field f = clz.getDeclaredField(nameOfField);
			f.setAccessible(true);
			f.set(obj, value);
			f.setAccessible(false);
		} catch (Throwable t) {
			throw new RuntimeException(t);
		}
	}
}
