/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import java.net.URL;
import java.util.HashMap;
import java.util.Properties;

import junit.framework.TestCase;

public class StsConfigTest extends TestCase {

	public void testContainsKey() {
		Properties props = new Properties();
		props.put("key", "value");
		StsConfig stsConfig = new StsConfig(props);
		
		assertTrue(stsConfig.containsKey("key"));
		assertFalse(stsConfig.containsKey("no_such_key"));
	}
	
	public void testGet() {
		Properties props = new Properties();
		props.put("key", "value");
		StsConfig stsConfig = new StsConfig(props);
		
		// check existing key
		assertEquals("value", stsConfig.get("key"));
		assertFalse("default".equals(stsConfig.get("key", "default")));
		// check missing key
		assertNull(stsConfig.get("missing"));
		assertEquals("default", stsConfig.get("missing", "default"));
	}
	
	public void testGetBoolean() {
		Properties props = new Properties();
		props.put("key1", "true");
		props.put("key2", "false");
		props.put("key3", "invalid");
		StsConfig config = new StsConfig(props);
		
		assertEquals(true, config.getBoolean("key1"));
		assertEquals(false, config.getBoolean("key2"));
		assertEquals(false, config.getBoolean("key3"));
		assertEquals(false, config.getBoolean("no_such_key"));
		
		assertEquals(true, config.getBoolean("no_such_key", true));
		assertEquals(false, config.getBoolean("no_such_key", false));
	}
	
	public void testGetInteger() {
		Properties props = new Properties();
		props.put("key1", "42");
		props.put("key2", "nan");
		StsConfig config = new StsConfig(props);
		
		assertEquals(42, config.getInteger("key1"));
		try {
			config.getInteger("key2");
			fail("value is not an integer: " + config.get("key2"));
		} catch (NumberFormatException e) { /* expected */ }
		assertEquals(0, config.getInteger("no_such_key"));
		
		assertEquals(42, config.getInteger("no_such_key", 42));
		assertEquals(-1, config.getInteger("no_such_key", -1));
	}

	public void testGetLong() {
		Properties props = new Properties();
		props.put("key1", "42");
		props.put("key2", "nan");
		props.put("key3", "" + Long.MAX_VALUE);
		StsConfig config = new StsConfig(props);
		
		assertEquals(42, config.getLong("key1"));
		try {
			config.getLong("key2");
			fail("value is not a long: " + config.get("key2"));
		} catch (NumberFormatException e) { /* expected */ }
		assertEquals(Long.MAX_VALUE, config.getLong("key3"));
		assertEquals(0, config.getLong("no_such_key"));
		
		assertEquals(42, config.getLong("no_such_key", 42));
		assertEquals(-1, config.getLong("no_such_key", -1));
	}
	
	public void testGetUrl() {
		Properties props = new Properties();
		props.put("key1", "http://www.ecample.com");
		props.put("key2", "malformed_url");
		StsConfig config = new StsConfig(props);

		URL url = config.getUrl("key1");
		assertEquals("http://www.ecample.com", url.toExternalForm());
		assertNull(config.getUrl("key2"));
		assertNull(config.getUrl("no_such_key"));

		assertEquals(url, config.getUrl("key2", url));
		assertEquals(url, config.getUrl("no_such_key", url));
	}

	public void testGetResource() {
		Properties props = new Properties();
		HashMap<String, Object> map = new HashMap<String, Object>();
		Integer res1 = new Integer(42);
		map.put("res1", res1);
		StsConfig config = new StsConfig(props, map);
		
		assertEquals(res1, config.getResource("res1"));
		assertEquals(null, config.getResource("no_such_resource"));
	}
	
}
