package dk.sosi.sts.server;

import dk.sosi.seal.pki.PKIException;
import dk.sosi.seal.pki.impl.HashMapCertificateCache;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;

import java.security.cert.X509Certificate;

public class StsHashMapCertificateCache extends HashMapCertificateCache {

    private static final String OCES3_TEST_INTERMEDIATE_CERTIFICATE_BASE_64 =
            "MIIGYTCCBBWgAwIBAgIUc01wpi5SHNF8ZEyZoJAOHVngYa4wQQYJKoZIhvcNAQEK"+
                    "MDSgDzANBglghkgBZQMEAgMFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgMF"+
                    "AKIDAgFAMGIxJDAiBgNVBAMMG0RlbiBEYW5za2UgU3RhdCBPQ0VTIHJvZC1DQTET"+
                    "MBEGA1UECwwKVGVzdCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNrZSBTdGF0MQsw"+
                    "CQYDVQQGEwJESzAeFw0yMTAzMTIwNzQyNTlaFw0zMTAzMTAwNzQyNThaMGsxLTAr"+
                    "BgNVBAMMJERlbiBEYW5za2UgU3RhdCBPQ0VTIHVkc3RlZGVuZGUtQ0EgMTETMBEG"+
                    "A1UECwwKVGVzdCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNrZSBTdGF0MQswCQYD"+
                    "VQQGEwJESzCCAaIwDQYJKoZIhvcNAQEBBQADggGPADCCAYoCggGBANOLofW5cEL4"+
                    "GKGtDfVb75BI+UXI5nl/gnoi1CJcsirR+07jFP4x7mSZe/Olr0z1M3boTVaLCtdZ"+
                    "urqR6AysVI9z9ymzE8zCLxE5U1x3ay/x/0J0iTxpqHqDhSF0OH+Ymb1vgp+9ta6k"+
                    "PweSdG+WjJL+5xkeynGP2Zd80O2RusLAjx7s7b8sR73ImevflENJXNAJwL6MQrx4"+
                    "wiTZYdPOVOSyWKTMHnBJ0MTinPyFxIv8zfbpJ2w8I+/YNTzI9EsXDDdBpQ4+asKx"+
                    "A4qZY/1+OQbVTGgMTNQKl7GireR2R+cTobGiD4zzd/7mt12/gQDkjOFtSVCuOfCU"+
                    "82EKilEysrnKhqFgtyGzUgRQ0PFePdI8DS50popRVtNlKKo17eE3P2GlFr9SIm9U"+
                    "48GN10Eb1q+X+K1OjmNJBUVYVHF3fxC2iPySLbwkS2DSj/A6fBznj/FkIGL1BzcG"+
                    "+DQPimsqKDZoIBm5SBOmi4ujoTxi+4w7h6uLyUt3D3vsPL1yUkl+qQIDAQABo4IB"+
                    "HDCCARgwEgYDVR0TAQH/BAgwBgEB/wIBADAfBgNVHSMEGDAWgBQ53N7e0JAmR6Dg"+
                    "xm5JfybyqS/2GzBzBggrBgEFBQcBAQRnMGUwOwYIKwYBBQUHMAKGL2h0dHA6Ly9j"+
                    "YTEuY3RpLWdvdi5kay9vY2VzL3Jvb3QvY2FjZXJ0L3Jvb3QuY2VyMCYGCCsGAQUF"+
                    "BzABhhpodHRwOi8vY2ExLmN0aS1nb3YuZGsvb2NzcDA9BgNVHR8ENjA0MDKgMKAu"+
                    "hixodHRwOi8vY2ExLmN0aS1nb3YuZGsvb2Nlcy9yb290L2NybC9yb290LmNybDAd"+
                    "BgNVHQ4EFgQUfyif2XGZQuJ159c1di5NCCVtdl4wDgYDVR0PAQH/BAQDAgEGMEEG"+
                    "CSqGSIb3DQEBCjA0oA8wDQYJYIZIAWUDBAIDBQChHDAaBgkqhkiG9w0BAQgwDQYJ"+
                    "YIZIAWUDBAIDBQCiAwIBQAOCAgEAuJcPSd3E6SqABGEckwvdF0W0TrScGglvOyyX"+
                    "5Sx7jChPy1FFJS8vxgoqd8CYoXhzzRzuaTLLWE0P7sz51sZJKGMMA4p+f7/h4ftz"+
                    "fEON1359fckE/SK5RA9IDdzVmiKxDoi8bAO/CyO/mXnZm0WFdtsU7OGgHCwAItIj"+
                    "NLanSnRijazcwsNfxNJoh99S6dAV/GGzo9NMPIRrLNG3IZboPbQe07MT/DFeHP7/"+
                    "xA5KxB2SdnPTfFevulJOARq4Fv+rThaSp+j67AhmzI6qrIj8r4GlCyr3+EI474ie"+
                    "ZeiiClMtT+DbHpPPfDMvNk1DYOqWNRi4warsDXhVkhXdGZrAP6OE3hnVbS94q5B7"+
                    "FhwugFl5XzSu06F3s3BLHrEbMCff2QZuGtf2xZIMRlWO8tivguxM2KLJYewGbOuM"+
                    "uk1SQcB22fF0NQH2yzqnW6DFTGORDgYfy8IJItv82IQxKiS7/FMaoMp3mLyqb9dN"+
                    "rgRyvnxwMuzLiVBDSJX0Hf78Nx/MwFvdbLawzdi9Fro9EWJ1hzjfJMZTiCP7my/E"+
                    "5UJMj/Sjce3jLep25c4pu8qrY7PC6o4OqICOkLJAVNL1TU+OVpKnUd7flWBOA4rH"+
                    "aR0dZY+OC/ZQhymMgT6Zg+Yj6d3cNwh8tiJTlCUYSlX27NP2HGhjZ5OlyHcOXtvQ"+
                    "3VlJhI4=";

    public static final X509Certificate OCES_3_TEST_INTERMEDIATE_CERTIFICATE = asCertificate(OCES3_TEST_INTERMEDIATE_CERTIFICATE_BASE_64);

    public StsHashMapCertificateCache() {
    }

    public X509Certificate getCertificate(Category category, String cacheKey) throws PKIException {
        if(category == Category.IntermediateCert && "http://ca1.cti-gov.dk/oces/issuing/1/cacert/issuing.cer".equals(cacheKey)) {
            return OCES_3_TEST_INTERMEDIATE_CERTIFICATE;
        }
        return super.getCertificate(category, cacheKey);
    }

    private static X509Certificate asCertificate(String base64data){
        return CertificateParser.asCertificate(XmlUtil.fromBase64(base64data));
    }

}
