/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server;

import java.util.Date;

import junit.framework.TestCase;

public class StsStatusTest extends TestCase {
	

	public void testGetStartedDate() {
		Date date = new Date();
		StsStatus status = new StsStatus(date, "unknown", "unknown");
		
		assertEquals(date.toString(), status.getStartedDate());
	}

	public void testVersion() {
		StsStatus status = new StsStatus(new Date(), "version", "unknown");

		assertEquals("version", status.getVersion());
	}
	
	public void testSealVersion() {
		StsStatus status = new StsStatus(new Date(), "unknown", "seal_version");

		assertEquals("seal_version", status.getSealVersion());
	}


	public void testNoAddDuration() {
		StsStatus status = new StsStatus(new Date(), "unknown", "unknown");
		assertStats(0, 0, 0, status);
		assertStats24h(0, 0, 0, status);
	}
	
	public void testSingleAddDuration() {
		StsStatus status = new StsStatus(new Date(), "unknown", "unknown");
		
		status.addTransaction(10, true);
		assertStats(1, 10, 10, status);
		assertStats24h(1, 10, 10, status);
	}
	
	public void testMultipleAddDuration() {
		StsStatus status = new StsStatus(new Date(), "unknown", "unknown");
		
		for (int i = 0; i < 5; i++) {
			status.addTransaction(i * 10, true);
		}
		assertStats(5, 40, 20, status);
		assertStats24h(5, 40, 20, status);
	}
	
	public void testCheckReset24h() {
		final int DAY_MS = 24 * 60 * 60 * 1000; 
		MockStsStatus status = new MockStsStatus(new Date(), System.currentTimeMillis());
		
		for (int i = 0; i < 5; i++) status.addTransaction(i * 10, true);
		assertStats(5, 40, 20, status);
		assertStats24h(5, 40, 20, status);

		// check that reset of 24h values is performed
		status.setTimeInMillis(System.currentTimeMillis() + DAY_MS);
		for (int i = 0; i < 5; i++)	status.addTransaction(i * 10, true);
		assertStats(10, 40, 20, status);
		assertStats24h(5, 40, 20, status);
		
		// check that subsequent resets are properly detected
		status.setTimeInMillis(System.currentTimeMillis() + 2*DAY_MS);		
		for (int i = 0; i < 5; i++) status.addTransaction(i * 10, true);
		assertStats(15, 40, 20, status);
		assertStats24h(5, 40, 20, status);
	}

	private void assertStats(long count, int max, int avg, StsStatus status) {
		assertEquals(count, status.getTransactionsForever());
		assertEquals(max + " ms", status.getMaxTimeForever());
		assertEquals(avg + " ms", status.getAverageTimeForever());
	}

	private void assertStats24h(int count, int max, int avg, StsStatus status) {
		assertEquals(count, status.getTransactions24h());
		assertEquals(max + " ms", status.getMaxTime24h());
		assertEquals(avg + " ms", status.getAverageTime24h());
	}
	
	private static class MockStsStatus extends StsStatus {
		private long time;
		
		private MockStsStatus(Date start, long time) {
			super(start, "unknown", "unknown");
			this.time = time;
		}

		public void setTimeInMillis(long time) {
			this.time = time;
		}
		
		@Override
		protected long getTimeInMillis() {
			return time;
		}
	}
}
