/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server;

import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfo.CertificateType;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.model.Authorization;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.RoleDefinition;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.roles.nationalroles.MockNationalRolesService;
import dk.sosi.sts.server.facade.RoleDefinitionServiceStub;
import dk.sosi.sts.server.test.StsCredentialVaults;
import dk.sosi.sts.server.test.TestIdentityProvider;
import dk.sosi.sts.server.test.TestUtil;
import dk.sosi.sts.server.test.UserValidationServiceTestUtil;
import dk.sosi.sts.service.*;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.LogManager;
import org.apache.log4j.SimpleLayout;
import org.junit.After;
import org.junit.Assume;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.mockito.Mockito;

import java.text.MessageFormat;
import java.text.ParseException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import static org.junit.Assert.*;
import static org.mockito.Matchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static dk.sosi.sts.server.test.UserValidationServiceTestUtil.*;

@RunWith(Parameterized.class)
public class UserValidationServiceTest {

	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	private static final CprNr CPR1 = new CprNr("1111111118");
	private static final String VALID_CODE1 = "0DEAD";
	private static final String VALID_CODE2 = "0CAFE";
	private static final String EDU_CODE_DOCTOR = "7170";
	private static final String EDU_CODE_NURSE = "5166";
	private static final String IGNORED_ROLE = "doctor";

	private static final SubjectSerialNumber CVR_RID_X = new SubjectSerialNumber("CVR:12345678-RID:12345678");
	private static final OcesInfo OCES_INFO_UNTRUSTED = OcesInfoParser.parseSubjectSerialNumber(CVR_RID_X);
	private OcesInfo OCES_INFO_TRUSTED;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		// SDS-6660: OCES2 test genindført. Se evt. SDS-6785
		builder.newTest("OCES2").add(Version.OCES2);
		//builder.newTest("OCES3").add(Version.OCES3);

		return builder.build();
	}

	@BeforeClass
	public static void init() {
		LogManager.getRootLogger().addAppender(new ConsoleAppender(new SimpleLayout()));
	}

	@Before
	public void setUp() throws Exception {
		// TestUtil.disableLogging();
		// System.out.println(LogManager.getRootLogger().getLevel());
		// create test data
		System.setProperty("dk.sds.nsp.audit.DevelopmentAuditProvider.enabled", "true");
		AuditHelper.init();

		OCES_INFO_TRUSTED = OcesInfoParser.parseCertificate(StsCredentialVaults.getRandomMocesWithCvrTrustedNationalRole(version).getSystemCredentialPair().getCertificate());
	}

	@After
	public void tearDown() {
		TestUtil.revertLogging();
		AuditHelper.destroy();
	}

	@Test
	public void testInputCprMatched() throws Exception {
		// SDS-6660: OCES2 test genindført. Se evt. SDS-6785
		Assume.assumeTrue(version == Version.OCES3);

		// OCES2
		OcesCvrRidService cvrRidService = mock(OcesCvrRidService.class);
		when(cvrRidService.isRelated(any(OcesInfo.class), any(CprNr.class))).thenReturn(true);

		// OCES3
		OcesUUID2CPRService uuid2CprService = mock(OcesUUID2CPRService.class);
		when(uuid2CprService.isRelated(any(OcesInfo.class), any(CprNr.class))).thenReturn(true);

		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, cvrRidService, uuid2CprService);

		final OcesInfo testInfo = OcesInfoParser.parseCertificate(CertificateStore.getValidMoces(version));
		CprNr cpr = userValidationService.verifyCpr(CPR1, testInfo);
		assertEquals(CPR1, cpr);
	}

	@Test
	public void testInputCprNotMatched() throws Exception {
		// SDS-6660: OCES2 test genindført. Se evt. SDS-6785
		Assume.assumeTrue(version == Version.OCES3);

		// OCES2
		OcesCvrRidService cvrRidService = mock(OcesCvrRidService.class);
		when(cvrRidService.isRelated(any(OcesInfo.class), any(CprNr.class))).thenReturn(false);

		// OCES3
		OcesUUID2CPRService uuid2CprService = mock(OcesUUID2CPRService.class);
		when(uuid2CprService.isRelated(any(OcesInfo.class), any(CprNr.class))).thenReturn(false);

		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, cvrRidService, uuid2CprService);

		final OcesInfo testInfo = OcesInfoParser.parseCertificate(CertificateStore.getValidMoces(version));
		try {
			userValidationService.verifyCpr(new CprNr("1111111119"), testInfo);
			fail("Exception expected");
		} catch (VerifyException e) {
			assertVerifyException(e, "uuid-cpr mismatch", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.UUIDCPR);
		}
	}

	@Test
	public void testCprLookupFound() throws Exception {
		// SDS-6660: OCES2 test genindført. Se evt. SDS-6785
		Assume.assumeTrue(version == Version.OCES3);

		final OcesInfo testInfo = OcesInfoParser.parseCertificate(CertificateStore.getValidMoces(version));

		// OCES2
		OcesCvrRidService cvrRidService = mock(OcesCvrRidService.class);
		when(cvrRidService.findRelatedCpr(any(OcesInfo.class))).thenReturn(CPR1);

		// OCES 3
		OcesUUID2CPRService uuid2CprService = mock(OcesUUID2CPRService.class);
		when(uuid2CprService.findRelatedCpr(any(OcesInfo.class))).thenReturn(CPR1);

		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, cvrRidService, uuid2CprService);
		CprNr cpr = userValidationService.verifyCpr(null, testInfo);
		assertEquals(CPR1, cpr);
	}

	@Test
	public void testCprLookupNotFound() throws Exception {
		// SDS-6660: OCES2 test genindført. Se evt. SDS-6785
		Assume.assumeTrue(version == Version.OCES3);

		// OCES2
		OcesCvrRidService cvrRidService = mock(OcesCvrRidService.class);
		when(cvrRidService.findRelatedCpr(any(OcesInfo.class))).thenReturn(null);

		// OCES3
		OcesUUID2CPRService uuid2CprService = mock(OcesUUID2CPRService.class);
		when(uuid2CprService.findRelatedCpr(any(OcesInfo.class))).thenReturn(null);

		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, cvrRidService, uuid2CprService);

		OcesInfo testInfo2 = OcesInfoParser.parseCertificate(CertificateStore.getRevokedMoces(version)); // HACK: revocation status not checked
		try {
			userValidationService.verifyCpr(null, testInfo2);
			fail("Exception expected");
		} catch (VerifyException e) {
			assertVerifyException(e, "illegal uuid", StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.UUIDCPR);
		}
	}

	@Test
	public void testCprLookupFromFailure() throws Exception {
		// SDS-6660: OCES2 test genindført. Se evt. SDS-6785
		Assume.assumeTrue(version == Version.OCES3);

		// OCES2
		OcesCvrRidService cvrRidService = mock(OcesCvrRidService.class);
		when(cvrRidService.findRelatedCpr(any(OcesInfo.class))).thenThrow(new StsException("findRelated failed"));

		// OCES3
		OcesUUID2CPRService uuid2CprService = mock(OcesUUID2CPRService.class);
		when(uuid2CprService.findRelatedCpr(any(OcesInfo.class))).thenThrow(new StsException("findRelated failed"));

		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, cvrRidService, uuid2CprService);

		OcesInfo testInfo = OcesInfoParser.parseCertificate(CertificateStore.getValidMoces(version));
		try {
			userValidationService.verifyCpr(null, testInfo);
			fail("Exception expected");
		} catch (VerifyException e) {
			assertVerifyException(e, "failed", StsFaultCodeValues.WST_REQUEST_FAILED, StsFaultActorValues.UUIDCPR);
		}
	}

	@Test
	public void testUnconfiguredNationalRoleFail() {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.UKENDT);
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, RoleDefinitionServiceStub.UKENDT.getPublicName(), null, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.ROLE_INVALID);
		}
	}

	@Test
	public void testNationalRoleSuppliedAndFound() throws Exception {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.PORTOER);
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName(), null, result, "test");
		assertNull(result.authorizationCode);
		assertEquals(result.role, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName());
	}

	@Test
	public void testNationalRoleSuppliedAndFoundAuthorizationsIgnored() throws Exception {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.PORTOER);
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR), new Authorization(VALID_CODE2, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName(), null, result, "test");
		assertNull(result.authorizationCode);
		assertEquals(result.role, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName());
	}

	@Test
	public void testNationalRoleSuppliedButNotFound() throws Exception {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.PORTOER);
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.UKENDT);
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR), new Authorization(VALID_CODE2, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		String role = RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName();
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, role, null, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, "Authentication failed: user does not have the requested role (" + role + ")");
		}
	}

	@Test
	public void testNoNationalRoleButTrustForCorrectEndpoint() throws VerifyException {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_TRUSTED, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName(), null, result, "STS");
		assertNull(result.authorizationCode);
		assertEquals(result.role, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName());
	}

	@Test
	public void givenRoleDefinitionThenRoleDefinitionServiceShouldReceiveRoleNameOfRoleDefinition()
			throws VerifyException {
		RoleDefinition rd = RoleDefinitionServiceStub.LAEGE_SEKRETAER;
		RoleDefinitionService assertingRds = new RoleDefinitionService() {
			@Override
			public List<RoleDefinition> getRoleDefinitions() { return Collections.singletonList(rd); }

			@Override
			public List<String> getTrustedRoleClaimers(String endpoint, String roleName) {
				fail();
				return Collections.emptyList();
			}

			public List<String> getTrustedRoleClaimers(String endpoint, Optional<String> roleName) {
				assertEquals(rd.getRoleName(), roleName.get());
				String ssiCvr = OcesInfoParser.parseCertificate(OCES_INFO_TRUSTED.getCertificate()).getCvr();
				String trustedOces3Cvr = TestIdentityProvider
						.getIdentityWithTrustedCvrAndThreeAuthorizations(Version.OCES3).getCvr();
				return Collections.unmodifiableList(Arrays.asList(ssiCvr, trustedOces3Cvr));
			}
		};
		createUserValidationService(version, noAuthorizations(), new MockNationalRolesService(), assertingRds)
				.verifyRole(CPR1, OCES_INFO_TRUSTED, rd.getPublicName(), null,
						new VerifyResult(), "STS");
	}

	@Test
	public void testVerifyRole_WithInputRoleWithoutCodeNotNationalRoleButTrustedCvr_SetsResultWithCodeInRole() throws VerifyException {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();

		userValidationService.verifyRole(CPR1, OCES_INFO_TRUSTED, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicNameNoCode(), null, result, "STS");

		assertEquals(result.authorizationCode, null);
		assertEquals(result.role, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName());
	}

	@Test
	public void testNationalRoleOverwrittenByTrustedCvr() throws VerifyException {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.PORTOER);

		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService);

		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_TRUSTED, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName(), null, result, "STS");
		assertEquals(result.authorizationCode, null);
		assertEquals(result.role, RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName());
	}

	@Test
	public void testNoNationalRoleButTrustForWrongEndpoint() throws VerifyException {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		String role = RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName();
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, role, null, result, "NBO");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, "Authentication failed: user does not have the requested role (" + role + ")");
		}
	}

	// Test cases for aut check

	// 1.2.1: (code = null, role = valid, match = none) -> Failure: Report
	// Autorisation missing/not matched
	// 1.2.2: (code = null, role = valid, match = single) -> Success: Attach code
	// 1.2.3: (code = null, role = valid, match = multiple) -> Failure: Report
	// multiple Autorisation

	@Test
	public void testEducationCodeSuppliedWithNoMatches_a() throws Exception {
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService);
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, EDU_CODE_DOCTOR, null, result, "STS");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_MISSING);
		}
	}

	@Test
	public void testEducationCodeSuppliedWithNoMatches_b() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_NURSE), new Authorization(VALID_CODE2, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, EDU_CODE_DOCTOR, null, result, "STS");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_NOT_MATCHED);
		}
	}

	@Test
	public void testEducationCodeSuppliedWithNoMatches_c() throws Exception {
		AuthorizationService authorizationService = noAuthorizations();
		final MockNationalRolesService nationalRoleService = new MockNationalRolesService();
		nationalRoleService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, nationalRoleService);
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, EDU_CODE_DOCTOR, null, result, "STS");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_MISSING);
		}
	}

	@Test
	public void testEducationCodeSuppliedWithSingleMatch() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR), new Authorization(VALID_CODE1, EDU_CODE_NURSE), new Authorization(VALID_CODE2, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, EDU_CODE_DOCTOR, null, result, "STS");
		assertEquals(result.authorizationCode, VALID_CODE1);
		assertEquals(result.role, EDU_CODE_DOCTOR);
	}

	@Test
	public void testEducationCodeSuppliedWithMultipleMatches() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR), new Authorization(VALID_CODE2, EDU_CODE_DOCTOR));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, EDU_CODE_DOCTOR, null, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_MULTIPLE_FOUND);
		}
	}

	// 2.1.1: (code = valid, role = ignored, match = none) -> Failure: Report
	// Autorisation missing/not matched
	// 2.1.2: (code = valid, role = ignored, match = single) -> Success: Attach role
	// 2.1.3: (code = valid, role = ignored, match = multiple) -> Failure: Report
	// multiple Autorisation

	@Test
	public void testAuthorizationCodeSuppliedWithNoMatches_a() throws Exception {
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService);
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, EDU_CODE_DOCTOR, VALID_CODE1, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_MISSING);
		}
	}

	@Test
	public void testAuthorizationCodeSuppliedWithNoMatches_b() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE2, EDU_CODE_DOCTOR), new Authorization(VALID_CODE2, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService);
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, VALID_CODE1, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_NOT_MATCHED);
		}
	}

	@Test
	public void testAuthorizationCodeSuppliedWithSingleoMatch() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR), new Authorization(VALID_CODE2, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, VALID_CODE1, result, "test");
		assertEquals(result.authorizationCode, VALID_CODE1);
		assertEquals(result.role, EDU_CODE_DOCTOR);
	}

	@Test
	public void testAuthorizationCodeSuppliedWithMultipleMatches_() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR), new Authorization(VALID_CODE1, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, VALID_CODE1, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_MULTIPLE_FOUND);
		}
	}

	// 2.2.1: (code = valid, role = valid, match = none) -> Failure: Report
	// Autorisation missing/not matched
	// 2.2.2: (code = valid, role = valid, match = single) -> Success: Attributes
	// verified
	// 2.2.3: (code = valid, role = valid, match = multiple) -> n/a

	@Test
	public void testFullAuthorizationSuppliedWithNoMatches_a() throws Exception {
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, EDU_CODE_DOCTOR, VALID_CODE1, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_MISSING);
		}
	}

	@Test
	public void testFullAuthorizationSuppliedWithNoMatches_b() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_NURSE), new Authorization(VALID_CODE2, EDU_CODE_DOCTOR), new Authorization(VALID_CODE2, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, EDU_CODE_DOCTOR, VALID_CODE1, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_NOT_MATCHED);
		}
	}

	@Test
	public void testFullAuthorizationSuppliedWithSingleMatch() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR), new Authorization(VALID_CODE1, EDU_CODE_NURSE), new Authorization(VALID_CODE2, EDU_CODE_DOCTOR), new Authorization(VALID_CODE2, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());

		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, EDU_CODE_DOCTOR, VALID_CODE1, result, "test");
		assertEquals(result.authorizationCode, VALID_CODE1);
		assertEquals(result.role, EDU_CODE_DOCTOR);
	}

	// shouldn't happen. consider robustness check only
	@Test
	public void testFullAuthorizationSuppliedWithMultipleMatches() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR), new Authorization(VALID_CODE1, EDU_CODE_DOCTOR));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, VALID_CODE1, result, "test");
		assertEquals(result.authorizationCode, VALID_CODE1);
		assertEquals(result.role, EDU_CODE_DOCTOR);
	}

	// 1.1.1: (code = null, role = ignored, match = none) -> Success: Ignore code
	// and role
	// 1.1.2: (code = null, role = ignored, match = single) -> Success: Attach code
	// and role
	// 1.1.3: (code = null, role = ignored, match = multiple) -> Failure: Report
	// multiple Autorisation

	@Ignore("SDS-6660")
	@Test
	public void testVerifyAuthorization_TestCase_1_1_1() throws Exception {
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService);
		VerifyResult result = new VerifyResult();
		/* OCES_INFO_UNTRUSTED ... noget med CVR-RID.. */
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, null, result, "STS");
		assertEquals(IGNORED_ROLE, result.role);
		assertNull(result.authorizationCode);
	}

	@Test
	public void testVerifyAuthorization_TestCase_1_1_2() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR));
			}
		};
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, null, result, "STS");
		assertEquals(VALID_CODE1, result.authorizationCode);
		assertEquals(EDU_CODE_DOCTOR, result.role);
	}

	@Test
	public void testVerifyAuthorization_NoAttributesMultipleAuthorizationsXXX() throws Exception {
		AuthorizationService authorizationService = new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Arrays.asList(new Authorization(VALID_CODE1, EDU_CODE_DOCTOR), new Authorization(VALID_CODE2, EDU_CODE_NURSE));
			}
		};
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, getNspLaegesekretaer());
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, null, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.AUT_MULTIPLE_FOUND);
		}
	}

	@Test
	public void testNoAuthNoRoles() throws Exception {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, null, result, "test");
		assertEquals(IGNORED_ROLE, result.role);
		assertEquals(null, result.authorizationCode);
	}

	@Test
	public void testNoAUthUnconfiguredRoleNotMatched() throws Exception {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.UKENDT);
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, null, result, "test");
		assertEquals(IGNORED_ROLE, result.role);
		assertEquals(null, result.authorizationCode);
	}

	@Test
	public void testNoAuthFoundNationalRole() throws Exception {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, null, result, "test");
		assertEquals(RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName(), result.role);
		assertEquals(null, result.authorizationCode);
	}

	@Test
	public void testNoAuthFoundNationalRoleUnconfiguredRoleIgnored() throws Exception {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.UKENDT);
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, null, result, "test");
		assertEquals(RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName(), result.role);
		assertEquals(null, result.authorizationCode);
	}

	@Test
	public void testNoAuthMultiFoundNationalRoleUnconfiguredRoleIgnored() throws Exception {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.PORTOER);
		AuthorizationService authorizationService = noAuthorizations();
		final UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);
		VerifyResult result = new VerifyResult();
		try {
			userValidationService.verifyRole(CPR1, OCES_INFO_UNTRUSTED, IGNORED_ROLE, null, result, "test");
			fail("Exception expected");
		} catch (VerifyException e) {
			assertAuthorizationVerifyExceptionFormat(e, VerifyException.ROLE_MULTIPLE_FOUND);
		}
	}

	@Test
	public void testValidRIDInvalidSubjectSerialNumberNotAllowed() {
		DistinguishedName dn = new DistinguishedName("c=DK,o=Lægehuset i Nørregade // CVR:10203040,cn=Sebastian Larsen,Serial=CVR:10203040-RID:9208-2002-2-194967071622");
		SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
		OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertNotNull(ocesInfo.getSubjectSerialNumber());
		assertNotNull(ocesInfo.getCvr());
		assertEquals(CertificateType.MOCES, ocesInfo.getCertificateType());

		UserValidationService userValidation = getUserValidationWithNationalRolesServiceThrowingNPE();

		try {
			VerifyInput input = new VerifyInput(ocesInfo, CPR1, null, null);
			VerifyResult result = userValidation.verifyCprAndRole(input, true, "test");
			// userValidation.verifyRole(CPR1, ocesInfo, null, null, result, "test");
			assertEquals(CPR1, result.cpr);
			assertEquals(null, result.authorizationCode);
			assertEquals(null, result.role);
		} catch (RuntimeException | VerifyException e) {
			fail("Forventede ikke en exception. Message:" + e.getMessage());
		}
	}

	@Test
	public void testValidPIDInvalidSubjectSerialNumberNotAllowed() {
		DistinguishedName dn = new DistinguishedName("C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=PID:9208-2002-2-514358910503");
		SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
		OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertNotNull(ocesInfo.getSubjectSerialNumber());
		assertNull(ocesInfo.getCvr());
		assertEquals(CertificateType.POCES, ocesInfo.getCertificateType());

		UserValidationService userValidation = getUserValidationWithNationalRolesServiceThrowingNPE();

		try {
			VerifyInput input = new VerifyInput(ocesInfo, CPR1, null, null);
			VerifyResult result = userValidation.verifyCprAndRole(input, true, "test");
			// userValidation.verifyRole(CPR1, ocesInfo, null, null, result, "test");
			assertEquals(CPR1, result.cpr);
			assertEquals(null, result.authorizationCode);
			assertEquals(null, result.role);
		} catch (RuntimeException | VerifyException e) {
			fail("Forventede ikke en exception. Message:" + e.getMessage());
		}
	}

	@Test
	public void testInvalidPIDInvalidSubjectSerialNumberNotAllowed() {
		DistinguishedName dn = new DistinguishedName("C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=XPIDX:9208-2002-2-514358910503");
		SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
		OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);

		// If SubjectSerialNumber is invalid then OcesInfoParser will set null as
		// SubjectSerialNumber on OcesInfo
		assertNull(ocesInfo.getSubjectSerialNumber());
		assertNull(ocesInfo.getCvr());
		// If SubjectSerialNumber is invalid then OcesInfoParser will set
		// CertificateType UNKOWN on OcesInfo
		assertEquals(CertificateType.UNKNOWN, ocesInfo.getCertificateType());

		UserValidationService userValidation = getUserValidationWithNationalRolesServiceThrowingNPE();

		try {
			VerifyInput input = new VerifyInput(ocesInfo, CPR1, null, null);
			@SuppressWarnings("unused")
			VerifyResult result = userValidation.verifyCprAndRole(input, true, "test");
			fail("Forventede nullpointer exception. Da vi bruger UserValidationService der ikke fejler, hvis SubjectSerialNumber er null på OcesInfo.");
		} catch (RuntimeException | VerifyException e) {
			assertEquals(NullPointerException.class, e.getClass());
			assertEquals("SubjectSerialNumber is null in NationalRolesService.getRoles(OcesInfo)", e.getMessage());
		}
	}

	private UserValidationService getUserValidationWithNationalRolesServiceThrowingNPE() {
		AuthorizationService authorizationService = noAuthorizations();
		MockNationalRolesService rolesService = new MockNationalRolesService() {
			@Override
			public List<String> getRoles(OcesInfo ocesInfo) {
				if (ocesInfo.getSubjectSerialNumber() == null) {
					throw new NullPointerException("SubjectSerialNumber is null in NationalRolesService.getRoles(OcesInfo)");
				}
				return super.getRoles(ocesInfo);
			}
		};
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.PORTOER);

		UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);

		return userValidationService;
	}

	@Test
	public void testValidRIDInvalidSubjectSerialNumberAllowed() {
		DistinguishedName dn = new DistinguishedName("c=DK,o=Lægehuset i Nørregade // CVR:10203040,cn=Sebastian Larsen,Serial=CVR:10203040-RID:9208-2002-2-194967071622");
		SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
		OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertNotNull(ocesInfo.getSubjectSerialNumber());
		assertNotNull(ocesInfo.getCvr());
		assertEquals(CertificateType.MOCES, ocesInfo.getCertificateType());

		UserValidationService userValidation = getUserValidationWithNationalRolesServiceReturningEmptyListWhenSubjectSerialNumberIsNull();

		try {
			VerifyInput input = new VerifyInput(ocesInfo, CPR1, null, null);
			VerifyResult result = userValidation.verifyCprAndRole(input, true, "test");
			// userValidation.verifyRole(CPR1, ocesInfo, null, null, result, "test");
			assertEquals(CPR1, result.cpr);
			assertEquals(null, result.authorizationCode);
			assertEquals(null, result.role);
		} catch (RuntimeException | VerifyException e) {
			fail("Forventede ikke en exception. Message:" + e.getMessage());
		}
	}

	@Test
	public void testValidPIDInvalidSubjectSerialNumberAllowed() {
		DistinguishedName dn = new DistinguishedName("C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=PID:9208-2002-2-514358910503");
		SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
		OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);

		assertNotNull(ocesInfo.getSubjectSerialNumber());
		assertNull(ocesInfo.getCvr());
		assertEquals(CertificateType.POCES, ocesInfo.getCertificateType());

		UserValidationService userValidation = getUserValidationWithNationalRolesServiceReturningEmptyListWhenSubjectSerialNumberIsNull();

		try {
			VerifyInput input = new VerifyInput(ocesInfo, CPR1, null, null);
			VerifyResult result = userValidation.verifyCprAndRole(input, true, "test");
			// userValidation.verifyRole(CPR1, ocesInfo, null, null, result, "test");
			assertEquals(CPR1, result.cpr);
			assertEquals(null, result.authorizationCode);
			assertEquals(null, result.role);
		} catch (RuntimeException | VerifyException e) {
			fail("Forventede ikke en exception. Message:" + e.getMessage());
		}
	}

	@Test
	public void testInvalidPIDInvalidSubjectSerialNumberAllowed() {
		DistinguishedName dn = new DistinguishedName("C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=XPIDX:9208-2002-2-514358910503");
		SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
		OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);

		// If SubjectSerialNumber is invalid then OcesInfoParser will set null as
		// SubjectSerialNumber on OcesInfo
		assertNull(ocesInfo.getSubjectSerialNumber());
		assertNull(ocesInfo.getCvr());
		// If SubjectSerialNumber is invalid then OcesInfoParser will set
		// CertificateType UNKOWN on OcesInfo
		assertEquals(CertificateType.UNKNOWN, ocesInfo.getCertificateType());

		UserValidationService userValidation = getUserValidationWithNationalRolesServiceReturningEmptyListWhenSubjectSerialNumberIsNull();

		try {
			VerifyInput input = new VerifyInput(ocesInfo, CPR1, null, null);
			VerifyResult result = userValidation.verifyCprAndRole(input, true, "test");
			// userValidation.verifyRole(CPR1, ocesInfo, null, null, result, "test");
			assertEquals(CPR1, result.cpr);
			assertEquals(null, result.authorizationCode);
			assertEquals(null, result.role);
		} catch (RuntimeException | VerifyException e) {
			fail("Forventede ikke en exception. Message:" + e.getMessage());
		}
	}

    @Test
    public void verifyCprAndRole_cprIsPresentAndTrusted_lookupIsSkipped() throws Exception {
        // Arrange
        OcesUUID2CPRService uuid2CPRService = Mockito.mock(OcesUUID2CPRService.class);
        UserValidationService subject = UserValidationServiceTestUtil.createUserValidationService(version, Mockito.mock(NationalRoleService.class), uuid2CPRService);

        DistinguishedName dn = new DistinguishedName("C=DK,OID.2.5.4.97=NTRDK-90177580,O=Testorganisation nr. 90177580,SERIALNUMBER=UI:DK-E:G:2634ccc3-225a-44ee-94bc-565904f46ead,SURNAME=Friis,GIVENNAME=Thorgot,CN=Thorgot Friis");
        SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
        OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);

        VerifyInput verifyInput = new VerifyInput(ocesInfo, CPR1, null, null);
        boolean cprTrustedForSigner = true;
        String endpoint = "http://foo";

        // Act
        subject.verifyCprAndRole(verifyInput, cprTrustedForSigner, endpoint);

        // Assert
        Mockito.verify(uuid2CPRService, Mockito.never()).isRelated(Mockito.any(), Mockito.any());
        Mockito.verify(uuid2CPRService, Mockito.never()).findRelatedCpr(Mockito.any());
    }

    @Test
    public void verifyCprAndRole_cprIsNotTrusted_verificationIsPerformed() throws Exception {
        // Arrange
        OcesUUID2CPRService uuid2CPRService = Mockito.mock(OcesUUID2CPRService.class);
        UserValidationService subject = UserValidationServiceTestUtil.createUserValidationService(version, Mockito.mock(NationalRoleService.class), uuid2CPRService);

        DistinguishedName dn = new DistinguishedName("C=DK,OID.2.5.4.97=NTRDK-90177580,O=Testorganisation nr. 90177580,SERIALNUMBER=UI:DK-E:G:2634ccc3-225a-44ee-94bc-565904f46ead,SURNAME=Friis,GIVENNAME=Thorgot,CN=Thorgot Friis");
        SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
        OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);

        VerifyInput verifyInput = new VerifyInput(ocesInfo, CPR1, null, null);
        boolean cprTrustedForSigner = false;
        String endpoint = "http://foo";

        Mockito.when(uuid2CPRService.isRelated(ocesInfo, CPR1)).thenReturn(true);

        // Act
        subject.verifyCprAndRole(verifyInput, cprTrustedForSigner, endpoint);

        // Assert
        Mockito.verify(uuid2CPRService).isRelated(ocesInfo, CPR1);
    }

    @Test
    public void verifyCprAndRole_cprIsMissing_lookupIsPerformed() throws Exception {
        // Arrange
        OcesUUID2CPRService uuid2CPRService = Mockito.mock(OcesUUID2CPRService.class);
        UserValidationService subject = UserValidationServiceTestUtil.createUserValidationService(version, Mockito.mock(NationalRoleService.class), uuid2CPRService);

        DistinguishedName dn = new DistinguishedName("C=DK,OID.2.5.4.97=NTRDK-90177580,O=Testorganisation nr. 90177580,SERIALNUMBER=UI:DK-E:G:2634ccc3-225a-44ee-94bc-565904f46ead,SURNAME=Friis,GIVENNAME=Thorgot,CN=Thorgot Friis");
        SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
        OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);

        VerifyInput verifyInput = new VerifyInput(ocesInfo, null, null, null);
        boolean cprTrustedForSigner = false;
        String endpoint = "http://foo";

        Mockito.when(uuid2CPRService.findRelatedCpr(ocesInfo)).thenReturn(CPR1);

        // Act
        subject.verifyCprAndRole(verifyInput, cprTrustedForSigner, endpoint);

        // Assert
        Mockito.verify(uuid2CPRService).findRelatedCpr(ocesInfo);
    }

	private UserValidationService getUserValidationWithNationalRolesServiceReturningEmptyListWhenSubjectSerialNumberIsNull() {
		AuthorizationService authorizationService = noAuthorizations();
		MockNationalRolesService rolesService = new MockNationalRolesService() {
			@Override
			public List<String> getRoles(OcesInfo ocesInfo) {
				if (ocesInfo.getSubjectSerialNumber() == null) {
					return Collections.emptyList();
				}

				return super.getRoles(ocesInfo);
			}
		};
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.PORTOER);

		UserValidationService userValidationService = UserValidationServiceTestUtil.createUserValidationService(version, authorizationService, rolesService);

		return userValidationService;
	}

//    @Test
//    public void testX() throws VerifyException {
//    	String x = "c=DK,o=Lægehuset i Nørregade // CVR:10203040,cn=Sebastian Larsen,Serial=CVR:10203040-RID:9208-2002-2-194967071622";
//    	x = "C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=PID:9208-2002-2-514358910503";
//    	DistinguishedName dn = new DistinguishedName(x);
//    	System.out.println(dn);
//    	SubjectSerialNumber ssn = new SubjectSerialNumber(dn.getSubjectSerialNumber());
//    	OcesInfo ocesInfo = OcesInfoParser.parseSubjectSerialNumber(ssn);
//    	assertNotNull(ocesInfo);
//    	
//
//    	
//    	VerifyResult result = new VerifyResult();
//    	
//    	userValidationService.verifyRole(CPR1, ocesInfo, null, null, result, "test");
//    	
//    	System.out.println("cpr="+result.cpr+", role="+result.role+", auth="+result.authorizationCode);
//    	
//    }

	private AuthorizationService noAuthorizations() {
		return new AuthorizationService() {
			@Override
			public List<Authorization> getAuthorizations(CprNr cpr) {
				return Collections.emptyList();
			}
		};
	}

	private NationalRoleService getNspLaegesekretaer() {
		MockNationalRolesService rolesService = new MockNationalRolesService();
		rolesService.put(CVR_RID_X, RoleDefinitionServiceStub.LAEGE_SEKRETAER);

		return rolesService;
	}

	private void assertAuthorizationVerifyExceptionFormat(VerifyException e, String format) {
		assertVerifyExceptionFormat(e, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, StsFaultActorValues.AUTHORIZATION, format);
	}

	private void assertVerifyException(VerifyException e, String faultStringPart, String faultCode, String faultActor) {
		assertTrue(e.getFaultString(), e.getFaultString().toLowerCase().contains(faultStringPart.toLowerCase()));
		assertEquals(faultCode, e.getFaultCode());
		assertEquals(faultActor, e.getFaultActor());
	}

	private void assertVerifyExceptionFormat(VerifyException ve, String code, String actor, String format, Object... args) {
		assertEquals(code, ve.getFaultCode());
		assertEquals(actor, ve.getFaultActor());

		MessageFormat mf = new MessageFormat(format);
		try {
			mf.parse(ve.getFaultString());
		} catch (ParseException e) {
			fail("Expected '" + format + "', actual '" + ve.getFaultString());
		}
	}
}
