/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.aop;

import java.util.concurrent.atomic.AtomicBoolean;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;

import junit.framework.TestCase;

/**
 *
 */
public class StsProxyFactoryTest extends TestCase {

	public static class TestingClass {

		private final String prefix;

		public TestingClass(String prefix) {
			this.prefix = prefix;
		}

		public String doMuch(String withThis) {
			return prefix + "-" + withThis;
		}
	}

	public void testCreateInstance() throws Exception {
		final StsProxyFactory<TestingClass> factory = new StsProxyFactory<TestingClass>();
		factory.setProxyClass(TestingClass.class);
		assertEquals(TestingClass.class, factory.getProxyClass());

		factory.setArguments(new Object[] {"blah"});
		factory.setSingleton(true);

		final TestingClass instance = factory.createInstance();

		assertEquals("blah-Hello", instance.doMuch("Hello"));
		assertTrue(instance instanceof StsProxy);

		final AtomicBoolean beenThere = new AtomicBoolean(false);

		StsProxyFactory.globalInterceptors.put(TestingClass.class,
											   new MethodInterceptor[]{
													   new MethodInterceptor() {
														   @Override
														   public Object invoke(final MethodInvocation invocation) throws Throwable {
															   beenThere.set(true);
															   assertTrue(invocation.getThis() == instance);
															   assertEquals("Hello", invocation.getArguments()[0]);
															   assertEquals("doMuch", invocation.getMethod().getName());
															   return "first"+invocation.proceed();
														   }
													   },
													   new MethodInterceptor() {
														   @Override
														   public Object invoke(final MethodInvocation invocation) throws Throwable {
															   beenThere.set(true);
															   return "more"+invocation.proceed();
														   }
													   }
											   });

		assertEquals("firstmoreblah-Hello", instance.doMuch("Hello"));
		assertTrue(beenThere.get());

		final StsAopProxySetup stsAopProxySetup = new StsAopProxySetup();

		stsAopProxySetup.setInterceptor(new MethodInterceptor() {
			@Override
			public Object invoke(final MethodInvocation invocation) throws Throwable {
				return "last"+invocation.proceed();
			}
		});

		stsAopProxySetup.setTargets(new Object[] {instance});
		assertTrue(instance == stsAopProxySetup.getTargets()[0]);
		assertNotNull(stsAopProxySetup.getInterceptor());
		stsAopProxySetup.afterPropertiesSet();

		beenThere.set(false);
		assertEquals("firstmorelastblah-Hello", instance.doMuch("Hello"));
		assertTrue(beenThere.get());
	}

	public void testNoMatchingConstructor() throws Exception {
		final StsProxyFactory<TestingClass> factory = new StsProxyFactory<TestingClass>();
		factory.setProxyClass(TestingClass.class);
		factory.setArguments(new Object[] {42});

		factory.setSingleton(true);

		try {
			final TestingClass instance = factory.createInstance();
			fail("No matching constructor should be found");
		} catch (UnsupportedOperationException e ) {

		}
	}
}
