/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.Message;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.server.test.TestUtil;
import dk.sosi.testtools.testrequest.STSMessageGenerator;
import dk.sosi.testtools.testrequest.WSTError;
import org.junit.After;
import org.junit.Before;
import org.w3c.dom.Document;

import java.net.MalformedURLException;

import static org.junit.Assert.*;

public abstract class AbstractTestStsMessage {

	SOSIFactory factory;
	protected TestClient testClient;

	@Before
	public void setUp() throws MalformedURLException {
		TestUtil.disableLogging();
		testClient = makeTestClient();
		factory = TestUtil.createStsSosiFactory(Version.OCES3);
	}

	protected abstract TestClient makeTestClient() throws MalformedURLException;

	@After
	public void tearDown() {
		TestUtil.revertLogging();
	}

	protected Message checkDeserializeResponse(Document response, boolean checkMessageID) {
		Message msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		if (checkMessageID && (msg.getMessageID() == null || msg.getMessageID().trim().isEmpty())) {
			fail("Warning: deserialized response has empty message ID!");
		}
		return msg;
	}

	protected void assertSuccess(Document response) {
		WSTError error = getError(response);
		if (error != null) {
			fail(error.getFaultString());
		}
	}

	public void assertFaultActor(Document response, String faultActor) {
		WSTError error = getError(response);
		if (error == null) {
			fail(XmlUtil.node2String(response, true, false));
		}
		assertEquals("wrong fault actor " + describe(error), faultActor, error.getFaultActor());
	}

	public void assertFaultCode(Document response, String faultCode) {
		WSTError error = getError(response);
		if (error == null) {
			fail(faultCode);
		}
		assertEquals("wrong fault code " + describe(error), faultCode, error.getFaultCode());
	}

	public void assertFaultStringContains(Document response, String... faultStringParts) {
		WSTError error = getError(response);
		if (error == null) {
			fail("Expected a fault");
		}
		for (String faultStringPart : faultStringParts) {
			boolean partFound = error.getFaultString().contains(faultStringPart);
			assertTrue("faultstring: " + error.getFaultString() + " is missing " + faultStringPart + " " + describe(error), partFound);
		}
	}

	public WSTError getError(Document response) {
		/* SDS-6660 - Måske skal VERSION.OCES3 fjernes igen? */
		return new STSMessageGenerator(SignatureUtil.setupCryptoProviderForJVM(), STSMessageGenerator.VERSION.OCES3).getError(response);
	}

	private String describe(WSTError error) {
		return "faultactor=" + error.getFaultActor() + ", faultCode=" + error.getFaultCode() + ", faultString=" + error.getFaultString();
	}
}
