/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import dk.sosi.testtools.testrequest.MessageGeneratorException;
import org.apache.log4j.Logger;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;

import static java.nio.charset.StandardCharsets.UTF_8;
import static java.util.Optional.ofNullable;

public class HttpTestClient implements TestClient {
	//public static final String testEnvironmentHost = "http://localhost:9090";
	//public static final String testEnvironmentHost = System.getProperty("sts.url","http://test1-cnsp.ekstern-test.nspop.dk:8080");

	public static final String
			local = "http://localhost:9090",
			testEnvironmentHost = System.getProperty("sts.url", local);

	public static final boolean isLocalTest = testEnvironmentHost.equals(local);

	private final URL url;
	
	public HttpTestClient(String path) throws MalformedURLException {
		this.url = new URL(testEnvironmentHost + path);
	}
	
	/**
	 * This will send the body as content of the HTTP request to the
	 * <code>sts.url</code>
	 *
	 * @param soapEnvolope
	 * @return the response parsed as a <code>Document</code>
	 */
	@Override
	public Response execute(String soapEnvolope) { return new SoapClient(url).execute(soapEnvolope); }

	/**
	 * This is the SoapClient from Seal's TestTools, but the encoding is forced to utf-8.
	 * The header and body would otherwise disagree.
	 */
	private static class SoapClient {
		private final URL url;

		private final static Logger logger = Logger.getLogger(SoapClient.class);

		public SoapClient(URL SOAPUrl) { url = SOAPUrl; }

		public Response execute(String soapEnvelope) {
			logger.debug("SOAPClient.execute:" + soapEnvelope);
			try {
				HttpURLConnection connection = (HttpURLConnection) url.openConnection();
				final byte[] bytes = soapEnvelope.getBytes(UTF_8);
				connection.setRequestProperty("Content-Length", String.valueOf(bytes.length));
				connection.setRequestProperty("Content-Type", "text/xml; charset=utf-8");
				connection.setRequestProperty("SOAPAction", "\"\"");
				connection.setRequestMethod("POST");
				connection.setDoOutput(true);
				connection.setDoInput(true);
				OutputStream out = connection.getOutputStream();
				out.write(bytes);
				out.close();
				final String contentType = connection.getContentType();
				if (!(contentType.contains("text/xml") && contentType.contains("utf-8"))) {
					throw new IllegalStateException("STS should set content type include 'text/xml' and 'charset=utf-8'. " +
															"This is not right: '" + contentType+"'");
				}
				InputSource response = new InputSource(connection.getResponseCode() >= 500
															   ? connection.getErrorStream()
															   : connection.getInputStream());
				DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
				factory.setValidating(false);
				factory.setNamespaceAware(true);
				try { return new Response(factory.newDocumentBuilder().parse(response), connection.getResponseCode()); }
				catch (SAXException sxe) {
					Exception x = ofNullable(sxe.getException()).orElse(sxe);
					throw new MessageGeneratorException(x.getMessage(),x);
				} catch (ParserConfigurationException pce) {
					throw new MessageGeneratorException(pce.getMessage(),pce);
				}
			} catch (IOException e) {
				throw new MessageGeneratorException("Error during soap invocation at " + url + " : " +
															e.getMessage(), e);
			}
		}
	}

}
