/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.roles.nationalroles.MockNationalRolesService;
import dk.sosi.sts.server.AuthorizationServiceStub;
import dk.sosi.sts.server.UserValidationService;
import dk.sosi.sts.server.test.*;
import dk.sosi.sts.service.*;

import java.lang.reflect.Field;
import java.security.cert.X509Certificate;
import java.util.*;

/**
 * User: ht
 * Date: 1/21/11
 * Time: 12:08
 */
public class MockServicesImpl implements AclService, CertificateStatusChecker {

    private final CredentialVault credentialVault;
    private MockNationalRolesService nationalRolesService = null;

    public static void setExpiredDate(UserIDCard userIDCard, Date expiredDate) {
        try {
            Field f = IDCard.class.getDeclaredField("expiryDate");
            f.setAccessible(true);
            f.set(userIDCard, expiredDate);
            f.setAccessible(false);
        } catch (Throwable t) {
            throw new RuntimeException(t);
        }
    }

    public static void setCreatedDate(UserIDCard userIDCard, Date createdDate) {
        try {
            Field f = IDCard.class.getDeclaredField("createdDate");
            f.setAccessible(true);
            f.set(userIDCard, createdDate);
            f.setAccessible(false);
        } catch (Throwable t) {
            throw new RuntimeException(t);
        }
    }


    private void trace(String method, String... args) {
        if (Boolean.getBoolean("sts.test.debug")) {
            String start = "MockServicesImpl: " + method + " > ";
            for (String arg : args) {
                start = start + ", " + arg;
            }
            start = start.replace("> ,", "> ");
            System.out.println(start);
        }
    }

//  private void trace(String msg) {
//    if (Boolean.getBoolean("sts.test.debug")) {
//      System.out.println(msg);
//    }
//  }

    private final Version version;

    public MockServicesImpl(Version version) {
        this(StsCredentialVaults.getStsCredentialVault(version), version);
    }

    public MockServicesImpl(CredentialVault stsCredentialVault, Version version) {
        this.version = version;

        loadDefaultRelations();
        this.credentialVault = stsCredentialVault;
        revokedCertificates.add(SealPPCredentialVaults.getRevokedVocesVault(version).getSystemCredentialPair().getCertificate());
        revokedCertificates.add(SealPPCredentialVaults.getRevokedFocesVault(version).getSystemCredentialPair().getCertificate());
        revokedCertificates.add(SealPPCredentialVaults.getRevokedMocesVault(version).getSystemCredentialPair().getCertificate());
    }

    public SOSIFactory buildSosiFactory(boolean isEmbeddedTest) {
        Federation federation = TestUtil.createTestFederation(this, isEmbeddedTest);
        return buildSosiFactory(federation);
    }

    public SOSIFactory buildSosiFactory(boolean isEmbeddedTest, CredentialVault vault) {
        Federation federation = TestUtil.createTestFederation(this, isEmbeddedTest);
        return buildSosiFactory(federation, vault);
    }

    public SOSIFactory buildSosiFactory(Federation federation) {
        return buildSosiFactory(federation, credentialVault);
    }

    public SOSIFactory buildSosiFactory(Federation federation, CredentialVault vault) {
        Properties props = new Properties(System.getProperties());
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_ISSUER, "TESTSTS");
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(false));
        return new SOSIFactory(federation, vault, props);
    }

    public boolean isWhitelisted(final OcesInfo ocesInfo, final String systemName) {
        return true;
    }

    public boolean isBlacklisted(final OcesInfo ocesInfo, final String systemName) {
        return false;
    }

    private List<X509Certificate> revokedCertificates = new ArrayList<>();

    public CertificateStatus getRevocationStatus(final X509Certificate certificate) throws PKIException {
        trace("revoked", certificate.getSubjectDN().toString());
        final boolean revoked = revokedCertificates.contains(certificate);
        return new CertificateStatus(!revoked, new Date());
    }

    // SDS-6660 - bør fjernes;
    private Map<String, String> cvrRid2Cpr = new HashMap<String, String>();
    private Map<String, String> uuid2Cpr = new HashMap<String, String>();

    private void loadDefaultRelations() {
        // SDS-6660 - bør fjernes;
        cvrRid2Cpr.put("CVR:46837428-RID:83701009", "0309691444"); // SSI/SonjaBach

        uuid2Cpr.put("UI:DK-E:G:f98cb2b9-30e3-4ef7-a429-fce908a3703b", "1010903333"); // Ole Olsen

        for (TestIdentity moces3Identity : TestIdentityProvider.MOCES3_IDENTITIES) {
            uuid2Cpr.put(moces3Identity.getSubjectSerialNumber(), moces3Identity.getCpr());
        }
    }


    public OcesCvrRidService getCvrRidService() {
        return new OcesCvrRidService() {
            public boolean isRelated(final OcesInfo ocesInfo, final CprNr cpr) throws StsException {
                String ssnValue = ocesInfo.getSubjectSerialNumber().getValue();
                String foundCpr = cvrRid2Cpr.get(ssnValue);
                return cpr.getValue().equals(foundCpr);
            }

            public CprNr findRelatedCpr(final OcesInfo ocesInfo) throws StsException {
                trace("findRelatedCpr (cvr)", ocesInfo.getSubjectSerialNumber().toString());
                String ssnValue = ocesInfo.getSubjectSerialNumber().getValue();
                String foundCpr = cvrRid2Cpr.get(ssnValue);
                return foundCpr != null ? new CprNr(foundCpr) : null;
            }

            public long lastSuccess() {
                return 0;
            }

            public long lastFailure() {
                return 0;
            }

            @Override
            public int getFailedCallsPercentage() { return 0; }

            @Override
            public int getHoursBackInTimeToCalculateFailedCallsPercentage() { return 0; }
        };
    }

    public OcesUUID2CPRService getUUID2CPRService() {
        return new OcesUUID2CPRService() {
            public boolean isRelated(final OcesInfo ocesInfo, final CprNr cpr) {
                String ssnUuid = ocesInfo.getSubjectSerialNumber().getValue();
                String foundCpr = uuid2Cpr.get(ssnUuid);
                return cpr.getValue().equals(foundCpr);
            }

            public CprNr findRelatedCpr(final OcesInfo ocesInfo) throws StsException {
                trace("findRelatedCpr (cvr)", ocesInfo.getSubjectSerialNumber().toString());
                String ssnUuid = ocesInfo.getSubjectSerialNumber().getValue();
                String foundCpr = uuid2Cpr.get(ssnUuid);
                return foundCpr != null ? new CprNr(foundCpr) : null;
            }

            public long lastSuccess() {
                return 0;
            }

            public long lastFailure() {
                return 0;
            }

            @Override
            public int getFailedCallsPercentage() { return 0; }

            @Override
            public int getHoursBackInTimeToCalculateFailedCallsPercentage() { return 0; }
        };
    }

    public MockNationalRolesService getExistingNationalRoleService() {
        return nationalRolesService;
    }

    public MockNationalRolesService getNationalRoleService() {
        nationalRolesService = new MockNationalRolesService();
        nationalRolesService.put(new SubjectSerialNumber("CVR:25450442-RID:48863622"), RoleDefinitionServiceStub.LAEGE_SEKRETAER);
        nationalRolesService.put(
                new SubjectSerialNumber(TestIdentityProvider.getMoces3IdentityWithNoAuthorizationAndOneNationalRoleForMappedSsn().getSubjectSerialNumber()),
                RoleDefinitionServiceStub.PORTOER);
        return nationalRolesService;
    }

    public UserValidationService getUserValidationService() {
        final OcesCvrRidService cvrRidService = getCvrRidService();
        final AuthorizationService authorizationService = new AuthorizationServiceStub();
        final NationalRoleService nationalRoleService = getNationalRoleService();
        final OcesUUID2CPRService uuidService = getUUID2CPRService();

        return UserValidationServiceTestUtil.createUserValidationService(version, cvrRidService, authorizationService, nationalRoleService, uuidService);
    }
}
