/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import java.security.PublicKey;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.Date;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import dk.sosi.seal.model.EncryptionUtil;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.OIOSAMLAssertion;
import dk.sosi.seal.model.SignatureConfiguration;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;
import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.model.constants.Tag;
import dk.sosi.seal.model.dombuilders.AbstractSAMLBuilder;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;

/**
 * User: ksr
 * Date: 23-10-13
 * Time: 13:27
 */
public class NemLoginTokenBuilder extends AbstractSAMLBuilder<NemLoginTokenBuilder, OIOSAMLAssertion> {
  public static final int ONE_HOUR = 1000 * 60 * 60;

  private CredentialVault signingVault;
  private Node subjectNode;

  private String recipientURL;
  private String audienceRestriction;

  private final Date created = new Date(System.currentTimeMillis() - ONE_HOUR);
  private Date expires = new Date(System.currentTimeMillis() + 1000 * 60 * 60);

  private String cpr;
  private String givenName;
  private String surName;
  private String email;

  private X509Certificate certificate;
  private CertificateInfo certInfo;
  private String cvr;
  private String orgName;
  
  private boolean isSimulatedMoces = false;

  public NemLoginTokenBuilder setSimulatedMoces(boolean b) {
	  isSimulatedMoces = b;
	  return this;
  }

  public NemLoginTokenBuilder setSigningVault(CredentialVault signingVault) {
    this.signingVault = signingVault;
    return this;
  }

  public NemLoginTokenBuilder setAudienceRestriction(String audienceRestriction) {
    this.audienceRestriction = audienceRestriction;
    return this;
  }

  public NemLoginTokenBuilder setRecipientURL(String recipientURL) {
    this.recipientURL = recipientURL;
    return this;
  }

  public NemLoginTokenBuilder setUserCpr(String cpr) {
    this.cpr = cpr;
    return this;
  }

  public NemLoginTokenBuilder setUserName(String givenName, String surName) {
    this.givenName = givenName;
    this.surName = surName;
    return this;
  }

  public NemLoginTokenBuilder expires(Date expires) {
    this.expires = expires;
    return this;
  }

  public NemLoginTokenBuilder setUserEmail(String email) {
    this.email = email;
    return this;
  }

  public NemLoginTokenBuilder setUserCertificate(X509Certificate certificate) {
    this.certificate = certificate;
    this.certInfo = new CertificateInfo(certificate);
    return this;
  }

  public NemLoginTokenBuilder setCertificateInfo(CertificateInfo certInfo) {
    this.certificate = null;
    this.certInfo = certInfo;
    return this;
  }

  public NemLoginTokenBuilder setCvr(String cvr) {
    this.cvr = cvr;
    return this;
  }

  public NemLoginTokenBuilder setOrgName(String orgName) {
    this.orgName = orgName;
    return this;
  }


  @Override
  protected void validateBeforeBuild() {
    super.validateBeforeBuild();
    validate("audienceRestriction", audienceRestriction);
    validate("signingVault", signingVault);

    validate("certificateInfo", certInfo);
    validate("recipientURL", recipientURL);
  }

  @Override
  public OIOSAMLAssertion build() throws ModelException {
    Document document = createDocument();
    signAssertion(document);
    return new OIOSAMLAssertion(document.getDocumentElement());
  }

  public Element buildEncrypted(PublicKey publicKey) throws ModelException {
    Document document = createDocument();
    signAssertion(document);
    return buildEncrypted(document, publicKey);
  }

  public static Element buildEncrypted(Document document, PublicKey publicKey) {
    EncryptionUtil.encrypt(document.getDocumentElement(), publicKey);

    Document doc2 = XmlUtil.createEmptyDocument();
    Tag tag = SAMLTags.encryptedAssertion;
    final Element encryptedRoot = doc2.createElementNS(tag.getNS(), tag.getPrefix() + ":" + tag.getTagName());

    doc2.appendChild(encryptedRoot);
    final Node importedNode = doc2.importNode(document.getDocumentElement(), true);
    encryptedRoot.appendChild(importedNode);

    return doc2.getDocumentElement();
  }

  public void signAssertion(Document document) {
    SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[]{assertionID}, assertionID, IDValues.Id);
    signatureConfiguration.setSignatureSiblingNode(subjectNode);
    SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(signingVault), document, signatureConfiguration);
  }

  public Document createDocument() {
    return super.createDocument();
  }


  @Override
  protected final void appendToRoot(Document doc, Element root) {
    super.appendToRoot(doc, root);
    subjectNode = createSubject();
    root.appendChild(subjectNode);
    root.appendChild(createConditions());
    root.appendChild(createAuthnStatement());
    root.appendChild(createAttributeStatement());
  }

  private Node createSubject() {
    Element subjectElm = createElement(SAMLTags.subject);

    Element nameIdElm = createElement(SAMLTags.nameID);
    nameIdElm.setAttribute(SAMLAttributes.FORMAT, SAMLValues.NAMEID_FORMAT_X509_SUBJECT_NAME);
    
    if(isSimulatedMoces) {
    	assert this.cvr != null;
    	assert this.orgName != null;
    	DistinguishedName certInfoSdn = certInfo.getSubjectDN(); 
    	assert certInfoSdn != null;
    	assert certInfoSdn.getCountry() != null;
    	assert certInfoSdn.getCommonName() != null;
    	assert certInfoSdn.getSubjectSerialNumber() != null;
    	assert certInfoSdn.getSubjectSerialNumber().startsWith("PID:");
    	DistinguishedName dn = new DistinguishedName("c="+certInfoSdn.getCountry()+",o="+this.orgName+" // CVR:"+this.cvr+",cn="+certInfoSdn.getCommonName()+",Serial=CVR:"+this.cvr+"-RID:"+certInfoSdn.getSubjectSerialNumber().substring(3));
    	nameIdElm.setTextContent(formatDN(dn, true));
    } else {
    	nameIdElm.setTextContent(formatDN(certInfo.getSubjectDN(), true));
    }
    
    subjectElm.appendChild(nameIdElm);

    Element subjectConfirmationElm = createElement(SAMLTags.subjectConfirmation);
    subjectConfirmationElm.setAttribute(SAMLAttributes.METHOD, SAMLValues.CONFIRMATION_METHOD_BEARER);
    subjectElm.appendChild(subjectConfirmationElm);

    Element subjectConfirmationDataElm = createElement(SAMLTags.subjectConfirmationData);
    subjectConfirmationDataElm.setAttribute(SAMLAttributes.RECIPIENT, recipientURL);
    subjectConfirmationDataElm.setAttribute(SAMLAttributes.NOT_ON_OR_AFTER, XmlUtil.getDateFormat(true).format(expires));
    subjectConfirmationElm.appendChild(subjectConfirmationDataElm);

    return subjectElm;
  }

  private Node createConditions() {
    Element conditionsElm = createElement(SAMLTags.conditions);
    conditionsElm.setAttribute(SAMLAttributes.NOT_BEFORE, XmlUtil.getDateFormat(true).format(created));
    conditionsElm.setAttribute(SAMLAttributes.NOT_ON_OR_AFTER, XmlUtil.getDateFormat(true).format(expires));

    Element audienceRestrictionElm = createElement(SAMLTags.audienceRestriction);
    conditionsElm.appendChild(audienceRestrictionElm);

    Element audienceElm = createElement(SAMLTags.audience);
    audienceElm.setTextContent(audienceRestriction);
    audienceRestrictionElm.appendChild(audienceElm);

    return conditionsElm;
  }

  private Node createAuthnStatement() {
    Element authnStatementElm = createElement(SAMLTags.authnStatement);
    authnStatementElm.setAttribute(SAMLAttributes.AUTHN_INSTANT, XmlUtil.getDateFormat(true).format(this.created));

    Element authnContextElm = createElement(SAMLTags.authnContext);
    Element authContextClassRefElm = createElement(SAMLTags.authnContextClassRef);
    authContextClassRefElm.setTextContent(SAMLValues.AUTHN_CONTEXT_CLASS_REF);
    authnContextElm.appendChild(authContextClassRefElm);
    authnStatementElm.appendChild(authnContextElm);
    // SessionIndex is mandatory in OIOSAML (to support per session logout), but is meaningless
    // for STS issued assertions
    authnStatementElm.setAttribute(SAMLAttributes.SESSION_INDEX, assertionID);
    return authnStatementElm;
  }

  private Element createAttributeStatement() {
    Element attributeStatementElm = createElement(SAMLTags.attributeStatement);
    attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.SURNAME, OIOSAMLAttributes.SURNAME_FRIENDLY, surName));
    attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.COMMON_NAME, OIOSAMLAttributes.COMMON_NAME_FRIENDLY, givenName + " " + surName));
    attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.EMAIL, OIOSAMLAttributes.EMAIL_FRIENDLY, email));
    attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.SPEC_VERSION, null, SAMLValues.DK_SPEC_VERSION));
    attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.ASSURANCE_LEVEL, null, SAMLValues.ASSURANCELEVEL_3));
    if (cvr != null) {
      attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.CVR_NUMBER, null, cvr));
    }
    if (orgName != null) {
      attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.ORGANIZATION_NAME, OIOSAMLAttributes.ORGANIZATION_NAME_FRIENDLY, orgName));
    }
    if (cpr != null) {
      attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.CPR_NUMBER, null, cpr));
    }
    attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.CERTIFICATE_ISSUER, null, formatDN(certInfo.getIssuerDN(), false)));
    String subjectSerialNumber = certInfo.getSubjectDN().getSubjectSerialNumber();
    attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.UID, OIOSAMLAttributes.UID_FRIENDLY, subjectSerialNumber));
    Element ccesIdAttribute = createOcesIdAttribute(subjectSerialNumber);
    if(ccesIdAttribute != null) {
    	attributeStatementElm.appendChild(ccesIdAttribute);
    }
    attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.CERTIFICATE_SERIAL, OIOSAMLAttributes.CERTIFICATE_SERIAL_FRIENDLY, certInfo.getCertificateSerialNumber().toString(16)));
    attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.IS_YOUTH_CERT, null, Boolean.FALSE.toString()));
    if (certificate != null) {
      try {
        final String certificateBase64 = XmlUtil.toBase64(certificate.getEncoded());
        attributeStatementElm.appendChild(createAttributeElement(OIOSAMLAttributes.USER_CERTIFICATE, null, certificateBase64));
      } catch (CertificateEncodingException e) {
        throw new RuntimeException(e);
      }
    }
    return attributeStatementElm;
  }

  private Element createOcesIdAttribute(String subjectSerialNumber) {
    if (subjectSerialNumber.contains("RID:")) {
      int ridIndex = subjectSerialNumber.indexOf("RID:");
      final String result = subjectSerialNumber.substring(ridIndex + 4);
      return createAttributeElement(OIOSAMLAttributes.RID_NUMBER, null, result);
    } else if (subjectSerialNumber.contains("PID:")) {
      int pidIndex = subjectSerialNumber.indexOf("PID:");
      final String result = subjectSerialNumber.substring(pidIndex + 4);
      return createAttributeElement(OIOSAMLAttributes.PID_NUMBER, null, result);
    } else if (subjectSerialNumber.contains("UI:")) {
        return null;
    } else {
      throw new ModelException("Certificate should be either a CVR-RID or a PID: '" + subjectSerialNumber + "'");
    }
  }


  private static String formatDN(DistinguishedName name, boolean requireSerial) {
    StringBuilder builder = new StringBuilder("C=");
    builder.append(name.getCountry());
    builder.append(",O=");
    builder.append(name.getOrganization());
    builder.append(",CN=");
    builder.append(name.getCommonName());
    if (requireSerial) {
      builder.append(",Serial=");
      builder.append(name.getSubjectSerialNumber());
    }
    return builder.toString();
  }
}
