package dk.sosi.sts.server.facade;

import dk.sosi.seal.model.Message;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.testtools.testrequest.MessageGeneratorHelper;
import org.junit.Test;
import org.w3c.dom.Document;

import java.net.MalformedURLException;
import java.security.cert.X509Certificate;

import static org.junit.Assert.assertEquals;

public class NewStsIT extends STSMessageTest {

    @Override
    protected TestClient makeTestClient() throws MalformedURLException {
        return new HttpTestClient("/sts/services/NewSecurityTokenService");
    }

    /**
     * M2-5.1: Test at for MOCES2 signerede SOSI IdKort gælder at: NewSecurityTokenService erstatter Subject.NameID i output idkortet med diverse certifikat-informationer.
     * M2-5.2: Test at for MOCES3 signerede SOSI IdKort gælder at: NewSecurityTokenService bibeholder Subject.NameID fra input til output idkort
     */
    @Test
    public void testMocesIdCardWithNameId() throws Exception {
        // Given
        String inputNameId = "1234";
        Document request = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", actualAuthorizationCode, inputNameId);
        Document signedRequest = MessageGeneratorHelper.sign(request, testIdentity.getCredentialVault());

        // When
        Document response = testClient.execute(signedRequest);

        // Then
        Message responseMessage = checkDeserializeResponse(response, true);
        assertSuccess(response);
        String certificateNameId = new CertificateInfo(CertificateStore.getValidMoces(version)).toString();
        String expectedOutputNameId = inputNameId;
        assertEquals(expectedOutputNameId, responseMessage.getIDCard().getAlternativeIdentifier());
    }

    /**
     * SYSID-4
     * Input: (F/V)OCES3-signeret systemidkort-request med NameID, som ikke er i 'certifikat-info' format
     * Endpoint: NewSecurityTokenService
     * Output: Response med systemidkort med NameID i certifikat-info format, som matcher certifikatet
     */
    @Test
    public void testVocesIdCardWithNonCertificateNameId() throws Exception {
        // Given
        String inputNameId = "1234";
        String cvr = OcesInfoParser.parseCertificate(CertificateStore.getValidVoces(version)).getCvr();
        Document request = stsMessageFactory.createSystemSecurityTokenRequest(cvr, inputNameId);
        Document signedRequest = MessageGeneratorHelper.sign(request, CertificateStore.getValidVocesVault(version));

        // When
        Document response = testClient.execute(signedRequest);

        // Then
        Message responseMessage = checkDeserializeResponse(response, true);
        assertSuccess(response);
        String certificateNameId = new CertificateInfo(CertificateStore.getValidVoces(version)).toString();
        assertEquals(certificateNameId, responseMessage.getIDCard().getAlternativeIdentifier());
    }

    /**
     * SYSID-5
     * Input: (F/V)OCES3-signeret systemidkort-request med NameID i 'certifikat-info' format, som matcher (F/V)OCES3 certifikatet
     * Endpoint: NewSecurityTokenService
     * Output: Response med systemidkort med NameID i certifikat-info format, som matcher certifikatet
     */
    @Test
    public void testVocesIdCardWithCertificateNameId() throws Exception {
        // Given
        GenericCredentialVault vocesVault = CertificateStore.getValidVocesVault(version);
        X509Certificate vocesCertificate = vocesVault.getSystemCredentialPair().getCertificate();
        String inputNameId = new CertificateInfo(vocesCertificate).toString();
        String cvr = OcesInfoParser.parseCertificate(vocesCertificate).getCvr();
        Document request = stsMessageFactory.createSystemSecurityTokenRequest(cvr, inputNameId);
        Document signedRequest = MessageGeneratorHelper.sign(request, vocesVault);

        // When
        Document response = testClient.execute(signedRequest);

        // Then
        Message responseMessage = checkDeserializeResponse(response, true);
        assertSuccess(response);
        assertEquals(inputNameId, responseMessage.getIDCard().getAlternativeIdentifier());
    }

    /**
     * SYSID-6
     * Input: (F/V)OCES3-signeret systemidkort-request med NameID i 'certifikat-info' format, som ikke matcher (F/V)OCES3 certifikatet (fx forkert UUID eller cert. serial)
     * Endpoint: NewSecurityTokenService
     * Output: Response med systemidkort med NameID i 'certifikat-info' format, som matcher certifikatet
     */
    @Test
    public void testVocesIdCardWithNonMatchingCertificateNameId() throws Exception {
        // Given
        GenericCredentialVault vocesVault = CertificateStore.getValidVocesVault(version);
        X509Certificate vocesCertificate = vocesVault.getSystemCredentialPair().getCertificate();
        String inputNameId = new CertificateInfo(CertificateStore.getValidFoces(version)).toString(); // Different certificate
        String cvr = OcesInfoParser.parseCertificate(vocesCertificate).getCvr();
        Document request = stsMessageFactory.createSystemSecurityTokenRequest(cvr, inputNameId);
        Document signedRequest = MessageGeneratorHelper.sign(request, vocesVault);

        // When
        Document response = testClient.execute(signedRequest);

        // Then
        Message responseMessage = checkDeserializeResponse(response, true);
        assertSuccess(response);
        String matchingCertificateNameId = new CertificateInfo(vocesCertificate).toString();
        assertEquals(matchingCertificateNameId, responseMessage.getIDCard().getAlternativeIdentifier());
    }
}
