/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server.facade;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import dk.sosi.seal.model.OIOBootstrapToken;
import dk.sosi.seal.model.OIOWSTrustResponse;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.Tag;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.xml.XmlUtil;

public class OIOWSTrustResponseValidation {
	
	public static void assertHolderOfKey(OIOBootstrapToken bootstrapToken, X509Certificate certificate) throws CertificateEncodingException {
		final Document bootstrapTokenDom = bootstrapToken.getDOM();
		final Element subject = getFirstChild(bootstrapTokenDom.getDocumentElement(), SAMLTags.subject);
		final Element subjectConfirmation = getFirstChild(subject, SAMLTags.subjectConfirmation);
		assertEquals("urn:oasis:names:tc:SAML:2.0:cm:holder-of-key", subjectConfirmation.getAttributes().getNamedItem("Method").getNodeValue());

		final Element subjectConfirmationData = getFirstChild(subjectConfirmation, SAMLTags.subjectConfirmationData);
		final Element keyInfo = getFirstChild(subjectConfirmationData, DSTags.keyInfo);
		final Element keyData = getFirstChild(keyInfo, DSTags.x509Data);
		final Element x509Cert = getFirstChild(keyData, DSTags.x509Certificate);

		final String signingCertBase64 = XmlUtil.toBase64(certificate.getEncoded());
		assertEquals(signingCertBase64, XmlUtil.getTextNodeValue(x509Cert));
	}
	
	public static Element getFirstChild(Element ancestor, Tag tag) {
		return XmlUtil.getFirstChildElementNS(ancestor, tag.getNS(), tag.getTagName());
	}
	
	public static void assertFault(OIOWSTrustResponse response, String requiredActor, String requiredErrorStringPart) {
		assertTrue("Not a fault", response.isFault());
		assertEquals("Unexpected fault actor: " + response.getFaultString(), requiredActor, response.getFaultActor());
		assertTrue("Unexpected fault string: " + response.getFaultString(), response.getFaultString().contains(requiredErrorStringPart));
	}

	public static void assertFault(OIOWSTrustResponse response, String requiredActor, String requiredFaultCode, String... requiredErrorStringParts) {
		assertTrue("Not a fault", response.isFault());
		assertEquals("Unexpected fault actor: " + response.getFaultString(), requiredActor, response.getFaultActor());
		for (String requiredErrorStringPart : requiredErrorStringParts) {
			assertTrue("Unexpected fault string: " + response.getFaultString(), response.getFaultString().contains(requiredErrorStringPart));
		}
		assertEquals(requiredFaultCode, response.getFaultCode());
	}

	public static void assertSuccess(OIOWSTrustResponse response, Federation federation) {
		assertFalse(response.getFaultString(), response.isFault());
		response.validateSignature();
		response.validateSignatureAndTrust(federation);
		assertNotNull(response.getCreated());
		assertNotNull(response.getExpires());
	}
}
