/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.Message;
import dk.sosi.seal.model.SecurityTokenResponse;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialPair;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.model.Authorization;
import dk.sosi.sts.model.RoleDefinition;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.StsCredentialVaults;
import dk.sosi.sts.server.test.TestIdentity;
import dk.sosi.sts.server.test.TestIdentityProvider;
import dk.sosi.sts.server.test.TestUtil;
import dk.sosi.testtools.testrequest.MessageGeneratorHelper;
import dk.sosi.testtools.testrequest.STSMessageGenerator;
import org.apache.commons.lang3.time.DateUtils;
import org.bouncycastle.x509.X509V3CertificateGenerator;
import org.junit.Assume;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.mockito.Mockito;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import javax.security.auth.x500.X500Principal;
import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.Date;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

/**
 * These are the test based on the new batch of certificates regarding the OCES2
 * shipment.
 */
@SuppressWarnings("deprecation")
@RunWith(Parameterized.class)
public abstract class STSMessage2Test extends AbstractTestStsMessage {
	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	@Parameter(2)
	public TestIdentity testIdentity;

	private String actualAuthorizationCode;

	private final String voces3Cvr = "94354969"; // cvr i MOCES3 og VOCES3 test certifikaterne er forskellige

	protected StsMessageFactory stsMessageFactory;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3, TestIdentityProvider.getIdentityWithTrustedCvrAndOneAuthorization(Version.OCES3));

		return builder.build();
	}

	@Before
	public void setUpSTSMessageTest() throws Exception {
		STSMessageGenerator.VERSION version = STSMessageGenerator.VERSION.valueOf(this.version.name());
		stsMessageFactory = new StsMessageFactory(version);
		actualAuthorizationCode = testIdentity.getAuthorizations().get(0).getCode();
	}

	/**
	 * tc02 moces certificate with another CPR than tc01 (valid?)
	 * <p/>
	 * The cpr is not registered in the mock test service, hence this request will
	 * fail.
	 */
	@Test
	public void testMocesWithCprValidCprSupplied() {
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "7170", actualAuthorizationCode );
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertSuccess(response);
	}

	/**
	 * tc02 moces certificate with another CPR than tc01 (valid?)
	 * <p/>
	 * The cpr is not registered in the mock test service, hence this request will
	 * fail.
	 */
	@Test
	public void testMocesWithCprNoCprSupplied() {
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "7170", actualAuthorizationCode);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertSuccess(response);
	}

	/**
	 * tc02 moces certificate with another CPR than tc01 (valid?)
	 * <p/>
	 * The cpr is not registered in the mock test service, hence this request will
	 * fail.
	 */
	@Test
	public void testMocesWithCprWrongCprSupplied() {
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), "1111111118", "doctor", "19901");
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.UUIDCPR);
		assertFaultStringContains(response, "-cpr mismatch"); // Either "cvrrid-cpr mismatch" or "uuid-cpr mismatch"
	}

	/**
	 * tc03 moces certificate with no CPR (valid?)
	 */
	@Test
	public void testMocesWithoutCprNoCprSupplied() {
		final CredentialVault vault = SealPPCredentialVaults.getValidMocesNoCpr(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "doctor", "19901");
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.UUIDCPR);
		assertFaultStringContains(response, "illegal uuid");
	}

	/**
	 * tc03 moces certificate with no CPR (valid?)
	 */
	@Test
	public void testMocesWithoutCprWrongCprSupplied() {
		final CredentialVault vault = SealPPCredentialVaults.getValidMocesNoCpr(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), "1111111118", "doctor", "19901");
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.UUIDCPR);
		assertFaultStringContains(response, "-cpr mismatch");
	}

	@Test
	public void testMocesRevoked() {
		Assume.assumeFalse(version == Version.OCES3); // TODO: Remove when CRA can handle oces3 certificates on test1 environment
		final CredentialVault vault = SealPPCredentialVaults.getRevokedMocesVault(version);

		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "doctor", "19901");
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "revoked");
	}

	@Test
	public void testMocesExpired() {
		final CredentialVault vault = SealPPCredentialVaults.getExpiredMocesVault(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "doctor", "19901");
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "certificate expired");
	}

	@Test
	public void testVocesValid() {
		final CredentialVault vault = SealPPCredentialVaults.getValidVocesVault(version);
		String cvr = version == Version.OCES3 ? voces3Cvr : testIdentity.getCvr();
		Document domClone = stsMessageFactory.createSystemSecurityTokenRequest(cvr);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
		Document response = testClient.execute(signedRequest);

		Message message = checkDeserializeResponse(response, true);
		assertSuccess(response);
		validateSystemIdCard(message.getIDCard(), cvr);
	}

	private void validateSystemIdCard(IDCard idCard, String cvr) {
		assertNotNull(idCard);
		assertTrue(idCard instanceof SystemIDCard);
		SystemIDCard systemIDCard = (SystemIDCard) idCard;
		systemIDCard.validateSignatureAndTrust(TestUtil.createTestFederation());
		assertTrue(systemIDCard.getIssuer().contains("STS")); // Afhænger af test miljø, men indeholder altid STS.
		assertEquals(cvr, systemIDCard.getSystemInfo().getCareProvider().getID());
	}

	@Test
	public void testVocesValidWrongCvr() {
		final CredentialVault vault = SealPPCredentialVaults.getValidVocesVault(version);
		Document domClone = stsMessageFactory.createSystemSecurityTokenRequest("12345678");
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
		Document response = testClient.execute(signedRequest);

		checkDeserializeResponse(response, true);

		assertFaultActor(response, "dk:sosi:sts");
		assertFaultCode(response, "wst:FailedAuthentication");
		assertFaultStringContains(response, "Authentication failed: cvr mismatch");
	}

	@Test
	public void testVocesTampered() {
		final CredentialVault vault = SealPPCredentialVaults.getValidVocesVault(version);
		Document domClone = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);

		insertCpr(signedRequest);

		Document response = testClient.execute(signedRequest);
		// System.out.println(XmlUtil.node2String(response, true, false));

		checkDeserializeResponse(response, true);

		assertFaultActor(response, "dk:sosi:sts:seal");
		assertFaultCode(response, "wst:AuthenticationBadElements");
		assertFaultStringContains(response, "Insufficient Digest Elements: Signature could not be validated");
	}

	protected void insertCpr(Document doc) {
		Element envelope = doc.getDocumentElement();
		Node body = envelope.getChildNodes().item(1); // Body
		Node token = body.getFirstChild(); // requestSecurityToken
		Node claims = token.getChildNodes().item(2); // Claims
		Node assertion = claims.getFirstChild(); // Assertion
		Node subject = assertion.getChildNodes().item(1); // Subject
		Node nameID = subject.getFirstChild();
		nameID.setTextContent("1234567890");
	}

	@Test
	public void testVocesRevoked() {
		final CredentialVault vault = SealPPCredentialVaults.getRevokedVocesVault(version);

		Document domClone = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
		Document response = testClient.execute(signedRequest);
		// System.out.println(XmlUtil.node2String(response, true, false));
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "revoked");
	}

	@Test
	public void testVocesExpired() {
		final CredentialVault credentialVaultFromPKCS12 = SealPPCredentialVaults.getExpiredVocesVault(version);
		Document domClone = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document signedRequest = MessageGeneratorHelper.sign(domClone, credentialVaultFromPKCS12);
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "certificate expired");
	}

	@Test
	public void testFocesValid() {
		CredentialVault focesVault = CertificateStore.getValidFocesVault(version);
		String cvr = OcesInfoParser.parseCertificate(focesVault.getSystemCredentialPair().getCertificate()).getCvr();
		Document dom = stsMessageFactory.createSystemSecurityTokenRequest(cvr);
		Document signedRequest = MessageGeneratorHelper.sign(dom, focesVault);
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertSuccess(response);
	}

	@Test
	public void testFocesExpired() {
		CredentialVault focesVault = SealPPCredentialVaults.getExpiredFocesVault(version);

		String cvr = OcesInfoParser.parseCertificate(focesVault.getSystemCredentialPair().getCertificate()).getCvr();
		Document dom = stsMessageFactory.createSystemSecurityTokenRequest(cvr);
		Document signedRequest = MessageGeneratorHelper.sign(dom, focesVault);
		Document response = testClient.execute(signedRequest);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "certificate expired");
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		checkDeserializeResponse(response, true);
	}

	@Test
	public void testFocesRevoked() {
		CredentialVault focesVault = SealPPCredentialVaults.getRevokedFocesVault(version);
		X509Certificate certificate = focesVault.getSystemCredentialPair().getCertificate();
		String cvr = OcesInfoParser.parseCertificate(certificate).getCvr();

		Document dom = stsMessageFactory.createSystemSecurityTokenRequest(cvr);
		Document signedRequest = MessageGeneratorHelper.sign(dom, focesVault);
		Document response = testClient.execute(signedRequest);
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testUnauthorizedWithRoleSupplied() {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithOneNationalRole(version);
		String role = testIdentity.getNationalRoles().get(0).getPublicName();
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, role, null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
		Document response = testClient.execute(signedRequest);
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(role, userIDCard.getUserInfo().getRole());
	}

	@Test
	public void testUnauthorizedNoRoleSupplied() {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithOneNationalRole(version);
		String role = testIdentity.getNationalRoles().get(0).getPublicName();
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "Klamphugger", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
		Document response = testClient.execute(signedRequest);
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(role, userIDCard.getUserInfo().getRole());
	}

	@Test
	public void testUnauthorizedNoNationalRoleWithUntrustedRoleSupplied() {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);

		String role = RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName();
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, role, null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
		Document response = testClient.execute(signedRequest);
		this.assertFaultStringContains(response, "Authentication failed: user does not have the requested role (" + role + ")");
	}

	@Test
	public void testUnauthorizedNoNationalRoleWithTrustedRoleSupplied() {
		CredentialVault vault = StsCredentialVaults.getRandomMocesWithCvrTrustedNationalRole(version);
		String role = RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName();
		String cvr = OcesInfoParser.parseCertificate(vault.getSystemCredentialPair().getCertificate()).getCvr();
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(cvr, null, role, null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
		Document response = testClient.execute(signedRequest);

		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName(), userIDCard.getUserInfo().getRole());
	}

	@Test
	public void testUnauthorizedNoNationalRoleNoRoleSupplied() {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "Piccoline", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
		Document response = testClient.execute(signedRequest);
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals("Piccoline", userIDCard.getUserInfo().getRole());
	}

	@Test
	@SuppressWarnings("deprecation")
	public void testSelfSignedCertificateWithoutSerialNumber() throws NoSuchAlgorithmException, CertificateEncodingException, InvalidKeyException, IllegalStateException, NoSuchProviderException, SignatureException {
		KeyPairGenerator generator = KeyPairGenerator.getInstance("RSA", "BC");
		/* Generate keypair */
		KeyPair keyPair = generator.generateKeyPair();
		X509V3CertificateGenerator certGen = new X509V3CertificateGenerator();
		/* Set certificate values */
		certGen.setSerialNumber(BigInteger.valueOf(System.currentTimeMillis()));
		certGen.setSignatureAlgorithm("SHA256WithRSAEncryption");
		certGen.setIssuerDN(new X500Principal("CN=TRUST2408 Systemtest XXXIV CA, O=TRUST2408, C=DK"));
		certGen.setNotBefore(DateUtils.addDays(new Date(), -1));
		certGen.setNotAfter(DateUtils.addDays(new Date(), 1));
		certGen.setSubjectDN(new X500Principal("CN=www.regionh.dk, O=CIMT, ST=Some-State, C=DK"));
		certGen.setPublicKey(keyPair.getPublic());
		/* Generate certificate */
		X509Certificate cert = certGen.generate(keyPair.getPrivate(), "BC");

		CredentialVault vault = Mockito.mock(CredentialVault.class);
		CredentialPair pair = Mockito.mock(CredentialPair.class);

		Mockito.when(vault.getSystemCredentialPair()).thenReturn(pair);
		Mockito.when(pair.getPrivateKey()).thenReturn(keyPair.getPrivate());
		Mockito.when(pair.getCertificate()).thenReturn(cert);

		Document domClone = stsMessageFactory.createUserSecurityTokenRequest("25450442", null, "Piccoline", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
		Document response = testClient.execute(signedRequest);

		assertFaultActor(response, "dk:sosi:sts");
		assertFaultCode(response, "wst:FailedAuthentication");
		assertFaultStringContains(response, "Authentication failed: signed with unsupported certificate type. Subject SerialNumber is missing");
	}

    /** *************************************************************************************************************
     *																											    *
     *												MitID Tests													    *
     *   			  																								*
     **************************************************************************************************************** */


	/**
	 * Testcase NSTS1
	 * <p/>
	 * Tjekker at der returneres en fejl, hvis der opsættes en ugyldig national
	 * rolle i userRole (dvs. national rolle er ikke på NSP liste over gyldige
	 * roller)
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor en ugyldig national rolle (den
	 * nationale roller er ugyldig i henhold til NSP listen over gyldige nationale
	 * roller) er opsat i userRole. SOSI STS kaldes (SecurityTokenService) for at få
	 * udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS1InvalidNationalRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), RoleDefinition.PREFIX + "code:11111:value:NonExistingRole", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.AUTHORIZATION);
		assertFaultStringContains(response, "Authentication failed: invalid role supplied (urn:dk:healthcare:national-federation-role:code:11111:value:NonExistingRole)");
	}

	/**
	 * Testcase NSTS2
	 * <p/>
	 * Tjekker at nationale rolle kan administreres lokalt og uden SEB, når CVR er
	 * på trust whitelist og natinoal rolle opsættes i IdKort
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort med ønsket national rolle opsat og
	 * kalder SOSI STS (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS2NationalRoleLocalAdmin() {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithTrustedCvrAndThreeAuthorizations(version);

		// Given
		final CredentialVault vault = StsCredentialVaults.getMitIDAuthorizations3WithNoNationalRoleTrustedCvr(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), RoleDefinition.PREFIX + "code:41002:value:SundAssistR2", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals("urn:dk:healthcare:national-federation-role:code:41002:value:SundAssistR2", userIDCard.getUserInfo().getRole());
	}

	/**
	 * Testcase NSTS3
	 * <p/>
	 * Tjekker at der returneres en fejl, hvis ønsket er forkert i henhold til
	 * udfaldsrum. Her er ønsket en national rolle
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor ønsket national rolle er opsat.
	 * Sundhedsfaglig person har ikke ønsket national rolle ifølge SEB. SOSI STS
	 * kaldes (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS3ErrorWrongNationalRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		String role = RoleDefinition.PREFIX + "code:99992:value:TestRolle2";
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), role, null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		checkDeserializeResponse(response, true);
		assertFaultActor(response, StsFaultActorValues.AUTHORIZATION);
		assertFaultStringContains(response, "Authentication failed: user does not have the requested role (" + role + ")");

	}

	/**
	 * Testcase NSTS4
	 * <p/>
	 * Tjekker at der udstedes IdKort med ønsket rolle, hvis ønske er rigtig i
	 * henhold til udfaldsrum. Her er ønsket en national rolle
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort med ønsket national rolle opsat og
	 * kalder SOSI STS (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS4NationalRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		String role = testIdentity.getNationalRoles().get(0).getPublicName();
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), role, null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(role, userIDCard.getUserInfo().getRole());
	}

	/**
	 * Tjekker at der udstedes IdKort med ønsket rolle, hvis ønske er rigtig i
	 * henhold til udfaldsrum. Her er ønsket en national rolle hvor rollen kun er angivet ved navn.
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort med ønsket national rolle opsat og
	 * kalder SOSI STS (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS4NationalRoleNoCode() {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		String role = testIdentity.getNationalRoles().get(0).getPublicName();
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), role, null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(role, userIDCard.getUserInfo().getRole());
	}

	/**
	 * Testcase NSTS5
	 * <p/>
	 * Tjekker at national rolle ønskes og vælges, hvis der både opsættes
	 * autorisationskode og national rolle i det selvsignerede SOSI IdKort
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor ønsket autorisationsnummer samt
	 * ønsket national rolle er opsat. SOSI STS kaldes (SecurityTokenService) for at
	 * få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS5NationalRoleWithAuthorization() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		String role = testIdentity.getNationalRoles().get(0).getPublicName();
		String authorization = testIdentity.getAuthorizations().get(0).getCode();
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), role, authorization);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(role, userIDCard.getUserInfo().getRole());
		assertNull(userIDCard.getUserInfo().getAuthorizationCode());
	}

	/**
	 * Testcase NSTS6
	 * <p/>
	 * Tjekker at autorisationvælges, når der er opsat et autorisationsId og
	 * uddannelseskode som matcher autorisaiontID i det selvsignerede SOSI IdKort
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor ønsket autorisationsID er opsat,
	 * samt uddanelseskode i userRole, som matcher autorisationsID. SOSI STS kaldes
	 * (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS6AuthorizationUsingAuthAndEduCode() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		Authorization authorization = testIdentity.getAuthorizations().get(0);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), authorization.getRole(), authorization.getCode());
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(authorization.getRole(), userIDCard.getUserInfo().getRole());
		assertEquals(authorization.getCode(), userIDCard.getUserInfo().getAuthorizationCode());
	}

	/**
	 * Testcase NSTS7
	 * <p/>
	 * Tjekker at der udstedes IdKort med ønsket rolle, hvis ønske er rigtigt i
	 * henhold til udfaldsrum. Her er ønsket en autorisation
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort med ønsket autorisationsnummer opsat og
	 * kalder SOSI STS (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS7AuthorizationUsingRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		Authorization authorization = testIdentity.getAuthorizations().get(0);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "test", authorization.getCode());
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(authorization.getRole(), userIDCard.getUserInfo().getRole());
		assertEquals(authorization.getCode(), userIDCard.getUserInfo().getAuthorizationCode());

	}

	/**
	 * Testcase NSTS8
	 * <p/>
	 * Tjekker at der returneres en fejl, hvis der opsættes et autorisationsId i det
	 * selvsignerede SOSI IdKort. Men medarbejderen ingen autorisationer har
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor ønsket autorisationsID er opsat.
	 * SOSI STS kaldes (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS8ErrorEmployeeWithoutAuthorization() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "doctor", "XYZ12");
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		checkDeserializeResponse(response, true);
		assertFaultStringContains(response, "Authentication failed: missing authorization");
	}

	/**
	 * Testcase NSTS9
	 * <p/>
	 * Tjekker at der returneres en fejl, hvis ønsket er forkert i henhold til
	 * udfaldsrum. Her er ønsket en autorisation
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor ønsket autorisationsnummer er
	 * opsat. Sundhedsfaglig person ikke har ønsket autorisation ifølge
	 * autorisationsregister. SOSI STS kaldes (SecurityTokenService) for at få
	 * udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS9ErrorWrongAuthorization() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "doctor", "003NX");
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		checkDeserializeResponse(response, true);
		assertFaultStringContains(response, "Authentication failed: authorization not found",
				testIdentity.getAuthorizations().get(0).getCode(), testIdentity.getAuthorizations().get(1).getCode(), testIdentity.getAuthorizations().get(2).getCode());
	}

	/**
	 * Testcase NSTS10
	 * <p/>
	 * Tjekker at autorisation kan ønskes via opsætning af uddannelseskode i
	 * userRole i det selvsignerede SOSI IdKort
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor ønsket sundhedsfaglig
	 * uddanelseskode er opsat i userRole. SOSI STS kaldes (SecurityTokenService)
	 * for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS10AuthorizationUsingEduCodeAndUserRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		Authorization authorization = testIdentity.getAuthorizations().get(0);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, authorization.getRole(), null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(authorization.getRole(), userIDCard.getUserInfo().getRole());
		assertEquals(authorization.getCode(), userIDCard.getUserInfo().getAuthorizationCode());
	}

	/**
	 * Testcase NSTS11
	 * <p/>
	 * Tjekker at der returneres fejl, hvis der opsættes en uddannelseskode i
	 * userRole, som matcher med flere autorisationer i autorisationsregisteret
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor ønsket sundhedsfaglig
	 * uddanelseskode er opsat i userRole (den kode der matcher de to
	 * autorisationer). SOSI STS kaldes (SecurityTokenService) for at få udsted
	 * adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS11ErrorSeveralAuthorizationsUsingEduCodeAndUserRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsWithTwoMatchingUddCodesAndThreeNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "7170", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		checkDeserializeResponse(response, true);
		assertFaultStringContains(response, "Authentication failed: multiple authorizations found",
				testIdentity.getAuthorizations().get(0).getCode(), testIdentity.getAuthorizations().get(1).getCode());
	}

	/**
	 * Testcase NSTS12
	 * <p/>
	 * Tjekker at der returneres fejl, hvis der opsættes en uddannelseskode i
	 * userRole, men medarbejder har igen autorisationer
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor ønsket uddannelseskode er opsat.
	 * SOSI STS kaldes (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS12ErrorEmployeeWithoutAuthorizationUsingEduCodeAndUserRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "7170", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		checkDeserializeResponse(response, true);
		assertFaultStringContains(response, "Authentication failed: missing authorization");
	}

	/**
	 * Testcase NSTS13
	 * <p/>
	 * Tjekker at der returneres fejl, hvis der opsættes en uddannelseskode i
	 * userRole, som ikke matcher en autorisation i autorisationsregisteret
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor forkert (en som sundhedsperson
	 * ikke har) sundhedsfaglig uddanelseskode er opsat i userRole. SOSI STS kaldes
	 * (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS13ErrorNoAuthorizationUsingEduCodeAndUserRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndThreeNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "5157", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		checkDeserializeResponse(response, true);
		assertFaultStringContains(response, "Authentication failed: multiple authorizations found",
				testIdentity.getAuthorizations().get(0).getCode(), testIdentity.getAuthorizations().get(1).getCode(), testIdentity.getAuthorizations().get(2).getCode());
	}

	/**
	 * Testcase NSTS14
	 * <p/>
	 * Tjekker at autorisation prioriteres over nationale roller. Dette gælder også
	 * når ansat har mange autorisationer og én natinoal rolle. Der skal her
	 * udstedes en fejlbesked med mulige autorisationer
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort uden ønske til autorisation eller
	 * national rolle og kalder SOSI STS (SecurityTokenService) for at få udsted
	 * adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS14ErrorSeveralAuthorizationsUsingAuthAndNationalRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndOneNationalRole(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "test", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		checkDeserializeResponse(response, true);
		assertFaultStringContains(response, "Authentication failed: multiple authorizations found",
				testIdentity.getAuthorizations().get(0).getCode(), testIdentity.getAuthorizations().get(1).getCode(), testIdentity.getAuthorizations().get(2).getCode());
	}

	/**
	 * Testcase NSTS15
	 * <p/>
	 * Tjekker der udstedes et IdKort med autorisation, hvis der ikke ønskes noget
	 * og bruger kun har én autorisation i henhold til autorisationsregister
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort uden ønske til autorisation eller
	 * national rolle og kalder SOSI STS (SecurityTokenService) for at få udsted
	 * adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS15AuthorizationUsingOneAuthorization() {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithOneAuthorization(version);
		Authorization authorization = testIdentity.getAuthorizations().get(0);

		// Given
		CredentialVault vault = StsCredentialVaults.getMitIDAuthorizedWithNoNationalRole(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "test", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(authorization.getRole(), userIDCard.getUserInfo().getRole());
		assertEquals(authorization.getCode(), userIDCard.getUserInfo().getAuthorizationCode());
	}

	/**
	 * Testcase NSTS16
	 * <p/>
	 * Tjekker at der udstedes en struktureret fejlkode med mulige roller, hvis der
	 * ikke ønskes noget og medarbejder har flere nationale roller i SEB
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort uden ønske til autorisation eller
	 * national rolle og kalder SOSI STS (SecurityTokenService) for at få udsted
	 * adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS16ErrorNoNationalRoleUsingSeveralNationalRoles() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "test", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		checkDeserializeResponse(response, true);
		assertFaultStringContains(response, "Authentication failed: multiple national roles found",
				testIdentity.getNationalRoles().get(0).getRoleName(), testIdentity.getNationalRoles().get(1).getRoleName(), testIdentity.getNationalRoles().get(2).getRoleName());
	}

	/**
	 * Testcase NSTS17
	 * <p/>
	 * Tjekker at der udstedes et IdKort med national rolle, hvis der ikke ønskes
	 * noget, samt medarbejder har tomt autorisationregister og en rolle i SEB
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort uden ønske til autorisation eller
	 * national rolle og kalder SOSI STS (SecurityTokenService) for at få udsted
	 * adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS17NationalRoleUsingOneNationalRoleAndZeroAuthorizations() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithOneNationalRole(version);
		String role = testIdentity.getNationalRoles().get(0).getPublicName();
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "test", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals(role, userIDCard.getUserInfo().getRole());
		assertNull(userIDCard.getUserInfo().getAuthorizationCode());
	}

	/**
	 * <p/>
	 * Tjekker at der udstedes et IdKort med rollen "urn:dk:healthcare:no-role", hvis det angives, at
	 * der ikke ønskes en national rolle eller autorisation ("urn:dk:healthcare:no-role" i userRole) og
	 * bruger af flere autorisationer i autorisationsreg. og en national rolle
	 * ifølge SEB
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort uden ønske til autorisation og med
	 * "urn:dk:healthcare:no-role" i userRole. SOSI STS kaldes
	 * (SecurityTokenService) for at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS18NationalRoleUsingOneNationalRoleAndSeveralAuthorizations() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndOneNationalRole(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "urn:dk:healthcare:no-role", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertNull(userIDCard.getUserInfo().getAuthorizationCode());
		assertEquals("urn:dk:healthcare:no-role", userIDCard.getUserInfo().getRole());
	}

	/**
	 * Testcase NSTS19
	 * <p/>
	 * Tjekker at der udstedes et IdKort uden indhold i autorisation og userRole,
	 * hvis der ikke ønskes noget og ikke skrives fritekst i userRole, samt
	 * medarbejder har tomt autorisationregister og SEB
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort uden ønske til autorisation eller
	 * national rolle og kalder SOSI STS (SecurityTokenService) for at få udsted
	 * adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS19EmptyAuthorizationUsingEmptyParameters() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "test", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals("test", userIDCard.getUserInfo().getRole());
	}

	/**
	 * Testcase NSTS20
	 * <p/>
	 * Tjekker at fritekst i selvsigneret IdKort feltet userRole overlever, hvis
	 * bruger ikke har autorisation eller national rolle
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort uden ønske til autorisation, men med
	 * fritekst i userRole attributten. SOSI STS kaldes (SecurityTokenService) for
	 * at få udsted adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS20UserTextAndRoleUsingNoAuthorizationAndNoNationalRole() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "userRole fritekst", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals("userRole fritekst", userIDCard.getUserInfo().getRole());
	}

	/**
	 * Testcase NSTS21
	 * <p/>
	 * Tjekker at der ikke ønskes noget, hvis der opsættes en uddannelseskode i
	 * userRole, som ikke matcher en gyldig uddannelseskode (dvs. er på listen over
	 * gyldige uddannelseskoder)
	 * <p/>
	 * Fagsystem danner selvsigneret IdKort, hvor ugyldig (en kode der ikke er på
	 * NSP lisen over gyldige uddannelseskoder) sundhedsfaglige uddannelseskode er
	 * opsat i userRole. SOS STS (SecurityTokenService) kaldes for at få udsted
	 * adgangbillet
	 * <p/>
	 */
	@Test
	public void testNSTS21EmptyAuthorizationUsingInvalidEduCode() {
		// Given
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "7171", null);
		Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		SecurityTokenResponse msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		assertSuccess(response);
		UserIDCard userIDCard = (UserIDCard) msg.getIDCard();
		assertEquals("7171", userIDCard.getUserInfo().getRole());
	}
}
