/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.Message;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.model.SystemInfo;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.model.constants.MedcomAttributes;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.TestIdentity;
import dk.sosi.sts.server.test.TestIdentityProvider;
import dk.sosi.sts.server.test.TestUtil;
import dk.sosi.testtools.testrequest.MessageGeneratorHelper;
import dk.sosi.testtools.testrequest.STSMessageGenerator;
import org.junit.Assume;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;

// This is mostly from TestSTSMessageGenerator from testtools
@RunWith(Parameterized.class)
public abstract class STSMessageTest extends AbstractTestStsMessage {
	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	@Parameter(2)
	public TestIdentity testIdentity;

	public String actualAuthorizationCode;

	protected final String voces3Cvr = "94354969"; // cvr i MOCES3 og VOCES3 test certifikaterne er forskellige

	private STSMessageGenerator stsMessageGenerator;
	protected StsMessageFactory stsMessageFactory;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3, TestIdentityProvider.getIdentityWithTrustedCvrAndOneAuthorization(Version.OCES3));

		return builder.build();
	}

	@Before
	public void setUpSTSMessageTest() throws Exception {
		STSMessageGenerator.VERSION version = STSMessageGenerator.VERSION.valueOf(this.version.name());
		stsMessageGenerator = new STSMessageGenerator(SignatureUtil.setupCryptoProviderForJVM(), version);
		stsMessageFactory = new StsMessageFactory(version);
		actualAuthorizationCode = testIdentity.getAuthorizations().get(0).getCode();
	}

	protected static SystemIDCard getSystemIDCard(Message msg) {
		return (SystemIDCard) msg.getIDCard();
	}

	protected static UserIDCard getUserIDCard(Message msg) {
		return (UserIDCard) msg.getIDCard();
	}

	protected static UserInfo getUserInfo(Message msg) {
		return getUserIDCard(msg).getUserInfo();
	}

	protected static SystemInfo getSystemInfo(Message msg) {
		switch (msg.getIDCard().getAuthenticationLevel().getLevel()) {
			case 3 /*AuthenticationLevel.VOCES_TRUSTED_SYSTEM*/:
				return getSystemIDCard(msg).getSystemInfo();
			case 4 /*AuthenticationLevel.MOCES_TRUSTED_USER*/:
				return getUserIDCard(msg).getSystemInfo();
			default:
				return null;
		}
	}

	// T1
	@Test
	public void testInvalidSecurityTokenRequest() {
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document invalid = stsMessageGenerator.generateInvalidXmlRequest(dom3system);
		Document response = testClient.execute(invalid);
		checkDeserializeResponse(response, false);
		assertFaultCode(response, StsFaultCodeValues.WST_REQUEST_FAILED);
		assertFaultActor(response, StsFaultActorValues.SEAL);
		assertFaultStringContains(response, "IDCard version 'null' not supported");
	}

	// T2
	@Test
	public void testInvalidVersion() {
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		stsMessageFactory.setSamlAttribute(dom3system, "sosi:IDCardVersion", "0.8");
		Document generateInvalidVersion = MessageGeneratorHelper.sign(dom3system, SealPPCredentialVaults.getValidVocesVault(version));
		Document response = testClient.execute(generateInvalidVersion);
		checkDeserializeResponse(response, false);
		assertFaultCode(response, StsFaultCodeValues.WST_REQUEST_FAILED);
		assertFaultActor(response, StsFaultActorValues.SEAL);
	}

	@Test
	public void testInvalidIdCardTypeLevelCombination() {
		Document dom4system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr(), 4);
		Document generateInvalidIdCardTypeLevelCombination = MessageGeneratorHelper.sign(dom4system, testIdentity.getCredentialVault());
		Document response = testClient.execute(generateInvalidIdCardTypeLevelCombination);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_BAD_REQUEST);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testInvalidGyldighed() {
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document generateInvalidGyldighed = stsMessageGenerator.generateInvalidGyldighed(dom3system);
		Document response = testClient.execute(generateInvalidGyldighed);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_INVALID_TIME_RANGE);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testInvalidAutentifikationLevel() {
		Document dom1system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr(), 1);
		Document generateInvalidAutentifikationLevel = MessageGeneratorHelper.sign(dom1system, SealPPCredentialVaults.getValidVocesVault(version));
		Document response = testClient.execute(generateInvalidAutentifikationLevel);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_BAD_REQUEST);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testInvalidIdCardType() {
		Document dom3unknown = stsMessageFactory.createUnknownSecurityTokenRequest(testIdentity.getCvr());
		Document generateInvalidIdCardType = MessageGeneratorHelper.sign(dom3unknown, SealPPCredentialVaults.getValidVocesVault(version));
		Document response = testClient.execute(generateInvalidIdCardType);
		checkDeserializeResponse(response, false);
		assertFaultCode(response, StsFaultCodeValues.WST_REQUEST_FAILED);
		assertFaultActor(response, StsFaultActorValues.SEAL);
	}

// T3	

	@Test
	public void testUnsignedLevel3() {
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document response = testClient.execute(dom3system);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_AUTHENTICATION_BAD_ELEMENTS);
		assertFaultActor(response, StsFaultActorValues.SEAL);
	}

	@Test
	public void testVocesLevel4() {
		Document dom4user = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", "12345");
		Document generateVocesLevel4 = MessageGeneratorHelper.sign(dom4user, SealPPCredentialVaults.getValidVocesVault(version));
		Document response = testClient.execute(generateVocesLevel4);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testVocesRevoked() {
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document generateVocesRevoked = MessageGeneratorHelper.sign(dom3system, SealPPCredentialVaults.getRevokedVocesVault(version));
		Document response = testClient.execute(generateVocesRevoked);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testVocesInvalidIssuer() {
		Assume.assumeFalse(version == Version.OCES3); // We have no oces3 certificates with invalid issuer yet
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document generateVocesInvalidIssuer = MessageGeneratorHelper.sign(dom3system, stsMessageGenerator.vaultInvalidIssuerVoces);
		Document response = testClient.execute(generateVocesInvalidIssuer);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testVocesExpired() {
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document generateVocesExpired = MessageGeneratorHelper.sign(dom3system, SealPPCredentialVaults.getExpiredVocesVault(version));
		Document response = testClient.execute(generateVocesExpired);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testVocesInvalidSignature() {
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document generateVocesInvalidSignature = MessageGeneratorHelper.sign(dom3system, SealPPCredentialVaults.getValidVocesVault(version));
		stsMessageFactory.setSamlAttribute(generateVocesInvalidSignature, "medcom:ITSystemName", "malice");

		Document response = testClient.execute(generateVocesInvalidSignature);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_AUTHENTICATION_BAD_ELEMENTS);
		assertFaultActor(response, StsFaultActorValues.SEAL);
	}

// T4
	@Test
	public void testUnsignedLevel4() {
		Document dom4user = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", "12345");
		Document response = testClient.execute(dom4user);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_AUTHENTICATION_BAD_ELEMENTS);
		assertFaultActor(response, StsFaultActorValues.SEAL);
	}

	@Test
	public void testMocesLevel3() {
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		Document generateMocesLevel3 = MessageGeneratorHelper.sign(dom3system, testIdentity.getCredentialVault());
		Document response = testClient.execute(generateMocesLevel3);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testMocesRevoked() {
		Document dom4user = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", "12345");
		Document generateMocesRevoked = MessageGeneratorHelper.sign(dom4user, SealPPCredentialVaults.getRevokedMocesVault(version));
		Document response = testClient.execute(generateMocesRevoked);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testMocesInvalidIssuer() {
		Assume.assumeFalse(version == Version.OCES3); // We have no expired certificates with invalid issuer yet
		Document dom4user = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", "12345");
		Document generateMocesInvalidIssuer = MessageGeneratorHelper.sign(dom4user, stsMessageGenerator.vaultInvalidIssuerMoces);
		Document response = testClient.execute(generateMocesInvalidIssuer);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testMocesExpired() {
		Document dom4user = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", "12345");
		Document generateMocesExpired = MessageGeneratorHelper.sign(dom4user, SealPPCredentialVaults.getExpiredMocesVault(version));
		Document response = testClient.execute(generateMocesExpired);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

	@Test
	public void testMocesInvalidSignature() {
		Document dom4user = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", "12345");
		Document generateMocesInvalidSignature = MessageGeneratorHelper.sign(dom4user, testIdentity.getCredentialVault());
		stsMessageFactory.setSamlAttribute(generateMocesInvalidSignature, "medcom:UserEmailAddress", "alice@malice.dk");

		Document response = testClient.execute(generateMocesInvalidSignature);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_AUTHENTICATION_BAD_ELEMENTS);
		assertFaultActor(response, StsFaultActorValues.SEAL);
	}

//	 T5
	@Test
	public void testInvalidCareProviderID() {
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(testIdentity.getCvr());
		stsMessageFactory.setSamlAttribute(dom3system, "medcom:CareProviderID", "00000000");
		Document generateInvalidCareProviderID = MessageGeneratorHelper.sign(dom3system, SealPPCredentialVaults.getValidVocesVault(version));
		Document response = testClient.execute(generateInvalidCareProviderID);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
	}

// T6
	@Test
	public void testEmptyUserAuthorizationCode() {
		Document dom4user = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", "12345");
		// create
		stsMessageFactory.setSamlAttribute(dom4user, MedcomAttributes.USER_AUTHORIZATION_CODE, " \t\n ");
		Document request = MessageGeneratorHelper.sign(dom4user, testIdentity.getCredentialVault());
		Document response = testClient.execute(request);

		Message msg = checkDeserializeResponse(response, true);
		assertSuccess(response);
		assertEquals(testIdentity.getCpr(), getUserInfo(msg).getCPR());
		assertEquals(testIdentity.getCvr(), getSystemInfo(msg).getCareProvider().getID());
		/*UserValidationService.getSingleNationalRole*/
		assertEquals("7170", getUserInfo(msg).getRole());
		assertEquals("KT2Z4", getUserInfo(msg).getAuthorizationCode());
	}

	@Test
	public void testInvalidUserCivilRegistrationNumber() {
		Document dom4user = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", "12345");
		stsMessageFactory.setSamlAttribute(dom4user, "medcom:UserCivilRegistrationNumber", "0000000000");
		Document generateInvalidUserCivilRegistrationNumber = MessageGeneratorHelper.sign(dom4user, testIdentity.getCredentialVault());
		Document response = testClient.execute(generateInvalidUserCivilRegistrationNumber);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.UUIDCPR);
	}

//	 T7
	@Test
	public void testMocesValidLevel4User() {
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "7170", actualAuthorizationCode);

		Document generateMocesValidLevel4User = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
		Document response = testClient.execute(generateMocesValidLevel4User);
		Message msg = checkDeserializeResponse(response, true);
		assertSuccess(response);
		assertEquals(testIdentity.getCpr(), getUserInfo(msg).getCPR());
		assertEquals(testIdentity.getCvr(), getSystemInfo(msg).getCareProvider().getID());
		assertEquals("7170", getUserInfo(msg).getRole());
		assertEquals(actualAuthorizationCode, getUserInfo(msg).getAuthorizationCode());
	}

	@Test
	public void testMocesValidLevel4UserWithValidNationalRolleAndRandomAuthorizationCode() {
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "urn:dk:healthcare:national-federation-role:code:99991:value:TestRolle1", "INVALID");
		Document generateMocesValidLevel4User = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());

		Document response = testClient.execute(generateMocesValidLevel4User);
		Message msg = checkDeserializeResponse(response, true);
		assertSuccess(response);
		assertEquals(testIdentity.getCpr(), getUserInfo(msg).getCPR());
		assertEquals(testIdentity.getCvr(), getSystemInfo(msg).getCareProvider().getID());
		assertEquals("urn:dk:healthcare:national-federation-role:code:99991:value:TestRolle1", getUserInfo(msg).getRole());
		assertNull(getUserInfo(msg).getAuthorizationCode());
	}

	@Test
	public void testVocesValidLevel3System() {
		String cvr = version == Version.OCES3 ? voces3Cvr : testIdentity.getCvr();
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest(cvr);
		Document generateVocesValidLevel3System = MessageGeneratorHelper.sign(dom3system, SealPPCredentialVaults.getValidVocesVault(version));
		Document response = testClient.execute(generateVocesValidLevel3System);
		Message msg = checkDeserializeResponse(response, true);
		assertSuccess(response);
		assertEquals(cvr, getSystemInfo(msg).getCareProvider().getID());
	}

	@Test
	public void testVocesValidLevel3User() {
		String cvr = version == Version.OCES3 ? voces3Cvr : testIdentity.getCvr();
		final int authenticationLevel = AuthenticationLevel.VOCES_TRUSTED_SYSTEM.getLevel();
		Document dom3user = stsMessageFactory.createUserSecurityTokenRequest(cvr, testIdentity.getCpr(), "doctor", "12345", authenticationLevel);
		Document generateVocesValidLevel3User = MessageGeneratorHelper.sign(dom3user, SealPPCredentialVaults.getValidVocesVault(version));
		Document response = testClient.execute(generateVocesValidLevel3User);
		Message msg = checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_BAD_REQUEST);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "ID-Card is a User ID-Card signed with VOCES/FOCES");
	}

	// TODO: refactor when Seal Testtools gets FOCES support
	@Test
	public void testFocesValidLevel3System() throws Exception {
		String cvr = OcesInfoParser.parseCertificate(CertificateStore.getValidFoces(version)).getCvr();
		Document dom = stsMessageFactory.createSystemSecurityTokenRequest(cvr);
		Document signedDom = MessageGeneratorHelper.sign(dom, CertificateStore.getValidFocesVault(version));
		Document response = testClient.execute(signedDom);
		Message msg = checkDeserializeResponse(response, true);
		assertSuccess(response);
		assertEquals(cvr, getSystemInfo(msg).getCareProvider().getID());
	}

	@Test
	public void testFocesValidLevel3User() throws Exception {
		final int authenticationLevel = AuthenticationLevel.VOCES_TRUSTED_SYSTEM.getLevel();
		final String cvr = OcesInfoParser.parseCertificate(CertificateStore.getValidFoces(version)).getCvr();
		Document dom = stsMessageFactory.createUserSecurityTokenRequest(cvr, "1111111118", "doctor", "19901", authenticationLevel);
		Document signedDom = MessageGeneratorHelper.sign(dom, CertificateStore.getValidFocesVault(version));
		Document response = testClient.execute(signedDom);
		Message msg = checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_BAD_REQUEST);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "ID-Card is a User ID-Card signed with VOCES/FOCES");
	}

	@Test
	public void testNonGlobalMoces() {
		Assume.assumeTrue(version == Version.OCES3);
		TestIdentity nonGlobalMoces3Identity = TestIdentityProvider.getNonGlobalMoces3Identity();
		Document domClone = stsMessageFactory.createUserSecurityTokenRequest(nonGlobalMoces3Identity.getCvr(), nonGlobalMoces3Identity.getCpr(), "anything", "anything");

		Document generateMocesValidLevel4User = MessageGeneratorHelper.sign(domClone, nonGlobalMoces3Identity.getCredentialVault());
		Document response = testClient.execute(generateMocesValidLevel4User);
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "signed with unsupported certificate - persistence level of moces3 certificates must be global");
	}

	@Test
	public void testVocesValidLevel3System_WithBlacklistedItSystemName_ReturnsFault() {
		String cvr = version == Version.OCES3 ? voces3Cvr : testIdentity.getCvr();
		Document dom3system = stsMessageFactory.createSystemSecurityTokenRequestWithItSystemName(cvr, TestUtil.STS_IT_SYSTEM_NAME);
		Document generateVocesValidLevel3System = MessageGeneratorHelper.sign(dom3system, SealPPCredentialVaults.getValidVocesVault(version));

		Document response = testClient.execute(generateVocesValidLevel3System);

		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_BAD_REQUEST);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "Invalid ITSystemName");
	}
}
