/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package dk.sosi.sts.server.facade;

import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.*;

import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.model.*;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.NboConfig;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.service.ItsConfigService;
import dk.sosi.sts.service.OcesPidService;

/**
 * Created by Kent on 21-03-2017.
 */
public class SamlTokenEmbeddedTestHelper {
	public static NboConfig buildNboConfig(CredentialVault trustedVault) {
		NboConfig nboConfig = new NboConfig();

		nboConfig.setTrustedVault(trustedVault);
		nboConfig.setIdCardDuration(86400000);
		nboConfig.setFuzzyTime(300000);
		return nboConfig;
	}

	public static ItsConfigService buildItsConfigService(SamlTokenTestModel testModel) {
		ItsConfigService itsConfig = mock(ItsConfigService.class);
		when(itsConfig.getFuzzyTime()).thenReturn(10000L);

		final String ssnValid = getSubjectSerialNumber(testModel.getValidWhitelistedVault());
		final String ssnValid2 = getSubjectSerialNumber(SealPPCredentialVaults.getValidVocesVault(testModel.getVersion()));
		final String ssnRevoked = getSubjectSerialNumber(SealPPCredentialVaults.getRevokedFocesVault(testModel.getVersion()));
		
		List<String> certificates = Arrays.asList(ssnValid, ssnValid2, ssnRevoked);

		AudienceConfiguration clearAudienceConfiguration = createClearAudienceConfiguration(mock(AudienceConfiguration.class), testModel, certificates);
		when(itsConfig.getAudienceConfiguration(eq(clearAudienceConfiguration.getAudience()))).thenReturn(clearAudienceConfiguration);

		AudienceConfiguration clearAudienceNoWardConfiguration = createClearAudienceNoWardConfiguration(mock(AudienceConfiguration.class), testModel, certificates);
		when(itsConfig.getAudienceConfiguration(eq(clearAudienceNoWardConfiguration.getAudience()))).thenReturn(clearAudienceNoWardConfiguration);
		
		SboAudienceConfiguration clearAudienceConfigurationJwt2Oio = createClearAudienceConfiguration(mock(SboAudienceConfiguration.class), testModel, certificates);
		when(itsConfig.getAudienceConfiguration(eq(clearAudienceConfigurationJwt2Oio.getAudience()), any())).thenReturn(clearAudienceConfigurationJwt2Oio);

		AudienceConfiguration cryptAudienceConfiguration = createCryptAudienceConfiguration(mock(AudienceConfiguration.class), testModel, Collections.singletonList(ssnValid));
		when(itsConfig.getAudienceConfiguration(eq(cryptAudienceConfiguration.getAudience()))).thenReturn(cryptAudienceConfiguration);
		
		SboAudienceConfiguration cryptAudienceConfigurationJwt2Oio = createCryptAudienceConfiguration(mock(SboAudienceConfiguration.class), testModel, certificates);
		when(itsConfig.getAudienceConfiguration(eq(cryptAudienceConfigurationJwt2Oio.getAudience()), any())).thenReturn(cryptAudienceConfigurationJwt2Oio);
		
		SboAudienceConfiguration sbo2AudienceConfigurationJwt2OioWithBst = createSBO2AudienceConfiguration(mock(SboAudienceConfiguration.class), testModel, certificates, true);
		when(itsConfig.getAudienceConfiguration(eq(sbo2AudienceConfigurationJwt2OioWithBst.getAudience()), any())).thenReturn(sbo2AudienceConfigurationJwt2OioWithBst);

		SboAudienceConfiguration jwt2oiosaml2AudienceConfiguration = createSBO2AudienceConfiguration(mock(SboAudienceConfiguration.class), testModel, certificates, false);
		when(itsConfig.getAudienceConfiguration(eq(jwt2oiosaml2AudienceConfiguration.getAudience()), any())).thenReturn(jwt2oiosaml2AudienceConfiguration);
		
		SboAudienceConfiguration sbo3AudienceConfigurationJwt2OioWithBst = createSBO3AudienceConfiguration(mock(SboAudienceConfiguration.class), testModel, certificates, true);
		when(itsConfig.getAudienceConfiguration(eq(sbo3AudienceConfigurationJwt2OioWithBst.getAudience()), any())).thenReturn(sbo3AudienceConfigurationJwt2OioWithBst);

		SboAudienceConfiguration jwt2oiosaml3AudienceConfiguration = createSBO3AudienceConfiguration(mock(SboAudienceConfiguration.class), testModel, certificates, false);
		when(itsConfig.getAudienceConfiguration(eq(jwt2oiosaml3AudienceConfiguration.getAudience()), any())).thenReturn(jwt2oiosaml3AudienceConfiguration);

		AudienceConfiguration blurringAudienceConfiguration = createBlurringAudienceConfiguration(mock(AudienceConfiguration.class), testModel, certificates);
		when(itsConfig.getAudienceConfiguration(eq(blurringAudienceConfiguration.getAudience()))).thenReturn(blurringAudienceConfiguration);

		return itsConfig;
	}
	
	private static <E extends AudienceConfiguration> E createClearAudienceConfiguration(E clearAudienceConfiguration, SamlTokenTestModel testModel, List<String> certificates) {
		when(clearAudienceConfiguration.getAudience()).thenReturn(testModel.getClearTextAudience());
		when(clearAudienceConfiguration.getCertificates()).thenReturn(certificates);
		when(clearAudienceConfiguration.getLifetime()).thenReturn(300000L);
		when(clearAudienceConfiguration.getJwtScope()).thenReturn("clear");
		when(clearAudienceConfiguration.requireBlurring()).thenReturn(BlurringVersion.NONE);
		if(clearAudienceConfiguration instanceof SboAudienceConfiguration) {
			SboAudienceConfiguration clearAudienceConfigurationJwt2Oio = (SboAudienceConfiguration) clearAudienceConfiguration;
			when(clearAudienceConfigurationJwt2Oio.getTargetCertificate()).thenReturn(CertificateStore.getValidVocesVault(testModel.getVersion()).getSystemCredentialPair().getCertificate());
			when(clearAudienceConfigurationJwt2Oio.getRecipientURL()).thenReturn("/dev/null");
			when(clearAudienceConfigurationJwt2Oio.includeBootStrapToken()).thenReturn(false);					
		}
		when(clearAudienceConfiguration.wardClaimAllowedCertificates()).thenReturn(new ArrayList<String>(){{add("UI:DK-O:G:23550132-5e1f-4e43-a5f9-048acf49e0b8");}});
		return clearAudienceConfiguration;
	}

	private static <E extends AudienceConfiguration> E createClearAudienceNoWardConfiguration(E clearAudienceConfiguration, SamlTokenTestModel testModel, List<String> certificates) {
		when(clearAudienceConfiguration.getAudience()).thenReturn(testModel.getClearTextAudienceNoWardWhitelisting());
		when(clearAudienceConfiguration.getCertificates()).thenReturn(certificates);
		when(clearAudienceConfiguration.getLifetime()).thenReturn(300000L);
		when(clearAudienceConfiguration.getJwtScope()).thenReturn("clear");
		when(clearAudienceConfiguration.requireBlurring()).thenReturn(BlurringVersion.NONE);
		if(clearAudienceConfiguration instanceof SboAudienceConfiguration) {
			SboAudienceConfiguration clearAudienceConfigurationJwt2Oio = (SboAudienceConfiguration) clearAudienceConfiguration;
			when(clearAudienceConfigurationJwt2Oio.getTargetCertificate()).thenReturn(CertificateStore.getValidVocesVault(testModel.getVersion()).getSystemCredentialPair().getCertificate());
			when(clearAudienceConfigurationJwt2Oio.getRecipientURL()).thenReturn("/dev/null");
			when(clearAudienceConfigurationJwt2Oio.includeBootStrapToken()).thenReturn(false);
		}
		return clearAudienceConfiguration;
	}
	
	private static <E extends AudienceConfiguration> E createCryptAudienceConfiguration(E cryptAudienceConfiguration, SamlTokenTestModel testModel, List<String> certificates) {
		when(cryptAudienceConfiguration.getAudience()).thenReturn(testModel.getCryptAudience());
		when(cryptAudienceConfiguration.getCertificates()).thenReturn(certificates);
		when(cryptAudienceConfiguration.getLifetime()).thenReturn(300000L);
		when(cryptAudienceConfiguration.requireBlurring()).thenReturn(BlurringVersion.NONE);
		if(cryptAudienceConfiguration instanceof SboAudienceConfiguration) {
			SboAudienceConfiguration cryptAudienceConfigurationJwt2Oio = (SboAudienceConfiguration) cryptAudienceConfiguration;
			when(cryptAudienceConfigurationJwt2Oio.getTargetCertificate()).thenReturn(CertificateStore.getValidVocesVault(testModel.getVersion()).getSystemCredentialPair().getCertificate());
			when(cryptAudienceConfigurationJwt2Oio.getRecipientURL()).thenReturn("/dev/null");
			when(cryptAudienceConfigurationJwt2Oio.includeBootStrapToken()).thenReturn(false);					
		}
		return cryptAudienceConfiguration;
	}

	private static <E extends AudienceConfiguration> E createSBO2AudienceConfiguration(E sboAudienceConfiguration, SamlTokenTestModel testModel, List<String> certificates, boolean includeBst) {
		final String sboAudience = DefaultSamlTokenTestModel.getSBOAudience(Version.OCES3, includeBst);
		when(sboAudienceConfiguration.getAudience()).thenReturn(sboAudience);
		when(sboAudienceConfiguration.getCertificates()).thenReturn(certificates);
		when(sboAudienceConfiguration.getLifetime()).thenReturn(300000L);
		when(sboAudienceConfiguration.getJwtScope()).thenReturn("clear");
		when(sboAudienceConfiguration.requireBlurring()).thenReturn(BlurringVersion.NONE);
		if(sboAudienceConfiguration instanceof SboAudienceConfiguration) {
			SboAudienceConfiguration sboAudienceConfigurationJwt2Oio = (SboAudienceConfiguration) sboAudienceConfiguration;
			when(sboAudienceConfigurationJwt2Oio.getTargetCertificate()).thenReturn(CertificateStore.getValidVocesVault(testModel.getVersion()).getSystemCredentialPair().getCertificate());
			when(sboAudienceConfigurationJwt2Oio.getRecipientURL()).thenReturn("/ststest/login");
			when(sboAudienceConfigurationJwt2Oio.includeBootStrapToken()).thenReturn(includeBst);					
		}

		return sboAudienceConfiguration;
	}
	
	private static <E extends AudienceConfiguration> E createSBO3AudienceConfiguration(E sboAudienceConfiguration, SamlTokenTestModel testModel, List<String> certificates, boolean includeBst) {
		final String sboAudience = DefaultSamlTokenTestModel.getSBOAudience(Version.OCES3,  includeBst);
		when(sboAudienceConfiguration.getAudience()).thenReturn(sboAudience);
		when(sboAudienceConfiguration.getCertificates()).thenReturn(certificates);
		when(sboAudienceConfiguration.getLifetime()).thenReturn(300000L);
		when(sboAudienceConfiguration.getJwtScope()).thenReturn("clear");
		when(sboAudienceConfiguration.requireBlurring()).thenReturn(BlurringVersion.NONE);
		if(sboAudienceConfiguration instanceof SboAudienceConfiguration) {
			SboAudienceConfiguration sboAudienceConfigurationJwt2Oio = (SboAudienceConfiguration) sboAudienceConfiguration;
			when(sboAudienceConfigurationJwt2Oio.getTargetCertificate()).thenReturn(CertificateStore.getValidVocesVault(testModel.getVersion()).getSystemCredentialPair().getCertificate());
			when(sboAudienceConfigurationJwt2Oio.getRecipientURL()).thenReturn("/ststest/login");
			when(sboAudienceConfigurationJwt2Oio.includeBootStrapToken()).thenReturn(includeBst);					
		}

		return sboAudienceConfiguration;
	}

	private static <E extends AudienceConfiguration> E createBlurringAudienceConfiguration(E blurringAudienceConfiguration, SamlTokenTestModel testModel, List<String> certificates) {
		when(blurringAudienceConfiguration.getAudience()).thenReturn(testModel.getBlurringAudience());
		when(blurringAudienceConfiguration.getCertificates()).thenReturn(certificates);
		when(blurringAudienceConfiguration.getLifetime()).thenReturn(300000L);
		when(blurringAudienceConfiguration.getJwtScope()).thenReturn("clear");
		when(blurringAudienceConfiguration.requireBlurring()).thenReturn(BlurringVersion.V11);
		if (blurringAudienceConfiguration instanceof SboAudienceConfiguration) {
			SboAudienceConfiguration clearAudienceConfigurationJwt2Oio = (SboAudienceConfiguration) blurringAudienceConfiguration;
			when(clearAudienceConfigurationJwt2Oio.getTargetCertificate()).thenReturn(CertificateStore.getValidVocesVault(testModel.getVersion()).getSystemCredentialPair().getCertificate());
			when(clearAudienceConfigurationJwt2Oio.getRecipientURL()).thenReturn("/dev/null");
			when(clearAudienceConfigurationJwt2Oio.includeBootStrapToken()).thenReturn(false);
			when(blurringAudienceConfiguration.requireBlurring()).thenReturn(BlurringVersion.NONE);
		}
		when(blurringAudienceConfiguration.wardClaimAllowedCertificates()).thenReturn(new ArrayList<String>(){{add("UI:DK-O:G:23550132-5e1f-4e43-a5f9-048acf49e0b8");}});
		return blurringAudienceConfiguration;
	}

	private static String getSubjectSerialNumber(CredentialVault vault) {
		return new DistinguishedName(vault.getSystemCredentialPair().getCertificate().getSubjectDN().getName()).getSubjectSerialNumber();
	}

	public static OcesPidService initMockOcesPidService() {
		final HashMap<SubjectSerialNumber, CprNr> relations = new HashMap<>();
		relations.put(new SubjectSerialNumber("PID:9208-2002-2-514358910503"), new CprNr("0501792275"));
		return new OcesPidMockService(relations);
	}
}
