/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.Message;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.SecurityTokenService;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.StsRequestHandler;
import dk.sosi.sts.server.StsStatus;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.TestUtil;
import dk.sosi.testtools.testrequest.MessageGeneratorHelper;
import dk.sosi.testtools.testrequest.STSMessageGenerator;
import org.junit.Test;
import org.w3c.dom.Document;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.StringWriter;

import static org.junit.Assert.assertEquals;

public class StsEmbeddedTest extends STSMessageTest {
    private MockServicesImpl mockServices;

    private SecurityTokenService sts;
    @Override
    protected TestClient makeTestClient() {
        mockServices = new MockServicesImpl(version);
        SOSIFactory sosiFactory = mockServices.buildSosiFactory(true);
        sts = new SecurityTokenService(sosiFactory, mockServices, mockServices.getUserValidationService(), TestUtil.STS_IT_SYSTEM_NAME);
        StsRequestHandler stsService = new StsRequestHandler(sosiFactory, sts, new StsStatus(), mockServices);
        return new TestClientHelper(stsService);
    }

    @Test
    public void testInvalidXml() throws  Exception {
        Document dom3system = stsMessageFactory.createSystemSecurityTokenRequest("30808460");
        Document generateInvalidXmlRequest = new STSMessageGenerator(SignatureUtil.setupCryptoProviderForJVM(), STSMessageGenerator.VERSION.OCES3).generateInvalidXmlRequest(dom3system);

        Source source = new DOMSource(generateInvalidXmlRequest);
        StringWriter sw = new StringWriter();
        Result result = new StreamResult(sw);
        Transformer xformer = TransformerFactory.newInstance().newTransformer();
        xformer.transform(source, result);
        String str = sw.toString().replaceFirst("saml:Attribute" ,"smal:Attribute");

        Document response = testClient.execute(str).document;
        checkDeserializeResponse(response, false); // If the XML is totally screwed up, the STS cannot return a messageID

        assertFaultActor(response, StsFaultActorValues.STS);
        assertFaultCode(response, StsFaultCodeValues.WST_INVALID_REQUEST);
    }

    @Test
    public void testFocesValidLevel3UserWithIsAllowUserIDCardWithVoces() throws Exception {
        sts.setAllowUserIDCardWithVoces(true);

        final int authenticationLevel = AuthenticationLevel.VOCES_TRUSTED_SYSTEM.getLevel();
        final String cvr = OcesInfoParser.parseCertificate(CertificateStore.getValidFoces(version)).getCvr();
        Document dom = stsMessageFactory.createUserSecurityTokenRequest(cvr, "1111111118", "doctor", "19901", authenticationLevel);
        Document signedDom = MessageGeneratorHelper.sign(dom, CertificateStore.getValidFocesVault(version));
        Document response = testClient.execute(signedDom);
        Message msg = checkDeserializeResponse(response, true);
        assertSuccess(response);
        assertEquals("1111111118", getUserInfo(msg).getCPR());
        assertEquals(cvr, getSystemInfo(msg).getCareProvider().getID());
        assertEquals("doctor", getUserInfo(msg).getRole());
        assertEquals("19901", getUserInfo(msg).getAuthorizationCode());
    }

    @Test
    public void testVocesValidLevel3UserWithIsAllowUserIDCardWithVoces() {
        sts.setAllowUserIDCardWithVoces(true);

        String cvr = version == OcesInfo.Version.OCES3 ? voces3Cvr : testIdentity.getCvr();
        final int authenticationLevel = AuthenticationLevel.VOCES_TRUSTED_SYSTEM.getLevel();
        Document dom3user = stsMessageFactory.createUserSecurityTokenRequest(cvr, testIdentity.getCpr(), "doctor", "12345", authenticationLevel);
        Document generateVocesValidLevel3User = MessageGeneratorHelper.sign(dom3user, SealPPCredentialVaults.getValidVocesVault(version));
        Document response = testClient.execute(generateVocesValidLevel3User);
        Message msg = checkDeserializeResponse(response, true);
        assertSuccess(response);
        assertEquals(testIdentity.getCpr(), getUserInfo(msg).getCPR());
        assertEquals(cvr, getSystemInfo(msg).getCareProvider().getID());
        assertEquals("doctor", getUserInfo(msg).getRole());
        assertEquals("12345", getUserInfo(msg).getAuthorizationCode());
    }


}
