/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import dk.sosi.seal.model.Message;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.testtools.testrequest.MessageGeneratorHelper;
import org.junit.Test;
import org.w3c.dom.Document;

import java.net.MalformedURLException;
import java.security.cert.X509Certificate;

import static org.junit.Assert.assertEquals;

public class StsIT extends STSMessageTest {

	@Override
	protected TestClient makeTestClient() throws MalformedURLException {
		return new HttpTestClient("/sts/services/SecurityTokenService");
	}

	/**
	 * M2-5.1: Test at for MOCES2 signerede SOSI IdKort gælder at: SecurityTokenService bibeholder Subject.NameID fra input til output.
	 * M2-5.2: Test at for MOCES3 signerede SOSI IdKort gælder at: SecurityTokenService bibeholder Subject.NameID fra input til output.
	 */
	@Test
	public void testMocesIdCardWithNameId() throws Exception {
		// Given
		String inputNameId = "1234";
		Document request = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), testIdentity.getCpr(), "doctor", actualAuthorizationCode, inputNameId);
		Document signedRequest = MessageGeneratorHelper.sign(request, testIdentity.getCredentialVault());

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		Message responseMessage = checkDeserializeResponse(response, true);
		assertSuccess(response);
		assertEquals(inputNameId, responseMessage.getIDCard().getAlternativeIdentifier());
	}

	/**
	 * SYSID-1
	 * Input: (F/V)OCES3-signeret systemidkort-request med NameID, som ikke er i 'certifikat-info' format
	 * Endpoint: SecurityTokenService
	 * Output: Response med systemidkort med samme NameID som i input
	 */
	@Test
	public void testVocesIdCardWithNonCertificateNameId() throws Exception {
		// Given
		String inputNameId = "1234";
		String cvr = OcesInfoParser.parseCertificate(CertificateStore.getValidVoces(version)).getCvr();
		Document request = stsMessageFactory.createSystemSecurityTokenRequest(cvr, inputNameId);
		Document signedRequest = MessageGeneratorHelper.sign(request, CertificateStore.getValidVocesVault(version));

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		Message responseMessage = checkDeserializeResponse(response, true);
		assertSuccess(response);
		assertEquals(inputNameId, responseMessage.getIDCard().getAlternativeIdentifier());
	}

	/**
	 * SYSID-2
	 * Input: (F/V)OCES3-signeret systemidkort-request med NameID i 'certifikat-info' format, som matcher (F/V)OCES3 certifikatet
	 * Endpoint: SecurityTokenService
	 * Output: Response med systemidkort med samme NameID som i input (dvs. i 'certifikat-info' format)
	 */
	@Test
	public void testVocesIdCardWithCertificateNameId() throws Exception {
		// Given
		GenericCredentialVault vocesVault = CertificateStore.getValidVocesVault(version);
		X509Certificate vocesCertificate = vocesVault.getSystemCredentialPair().getCertificate();
		String inputNameId = new CertificateInfo(vocesCertificate).toString();
		String cvr = OcesInfoParser.parseCertificate(vocesCertificate).getCvr();
		Document request = stsMessageFactory.createSystemSecurityTokenRequest(cvr, inputNameId);
		Document signedRequest = MessageGeneratorHelper.sign(request, vocesVault);

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		Message responseMessage = checkDeserializeResponse(response, true);
		assertSuccess(response);
		assertEquals(inputNameId, responseMessage.getIDCard().getAlternativeIdentifier());
	}

	/**
	 * SYSID-3
	 * Input: (F/V)OCES3-signeret systemidkort-request med NameID i 'certifikat-info' format, som ikke matcher (F/V)OCES3 certifikatet (fx forkert UUID eller cert. serial)
	 * Endpoint: SecurityTokenService
	 * Output: Fejlsvar ("The specified RequestSecurityToken is not understood: IDCard nameId must match information from certificate.")
	 */
	@Test
	public void testVocesIdCardWithNonMatchingCertificateNameId() throws Exception {
		// Given
		GenericCredentialVault vocesVault = CertificateStore.getValidVocesVault(version);
		X509Certificate vocesCertificate = vocesVault.getSystemCredentialPair().getCertificate();
		String inputNameId = new CertificateInfo(CertificateStore.getValidFoces(version)).toString(); // Different certificate
		String cvr = OcesInfoParser.parseCertificate(vocesCertificate).getCvr();
		Document request = stsMessageFactory.createSystemSecurityTokenRequest(cvr, inputNameId);
		Document signedRequest = MessageGeneratorHelper.sign(request, vocesVault);

		// When
		Document response = testClient.execute(signedRequest);

		// Then
		checkDeserializeResponse(response, true);
		assertFaultCode(response, StsFaultCodeValues.WST_BAD_REQUEST);
		assertFaultActor(response, StsFaultActorValues.STS);
		assertFaultStringContains(response, "IDCard nameId must match information from certificate");
	}
}
