/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.DGWSConstants;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.vault.EmptyCredentialVault;
import dk.sosi.testtools.testrequest.STSMessageGenerator;
import dk.sosi.testtools.testrequest.STSMessageGenerator.VERSION;
import org.w3c.dom.Document;

import java.util.Properties;

import static dk.sosi.testtools.testrequest.STSMessageGenerator.VERSION.*;

public class StsMessageFactory {

    private final STSMessageGenerator messageGenerator;

    public StsMessageFactory(VERSION version) {
        messageGenerator = new STSMessageGenerator(SignatureUtil.setupCryptoProviderForJVM(), version);
    }

    public Document createSystemSecurityTokenRequest(String cvr) {
        return createSystemSecurityTokenRequest(cvr, null);
    }

    public Document createSystemSecurityTokenRequestWithItSystemName(String cvr, String itSystemName) {
        Document tmp = createSecurityTokenRequest(cvr, "0000000000", "doctor", "19901", null, itSystemName);
        return messageGenerator.setLevelType(tmp, AuthenticationLevel.VOCES_TRUSTED_SYSTEM.getLevel(), "system");
    }

    public Document createSystemSecurityTokenRequest(String cvr, String alternativeIdentifier) {
        return createSystemSecurityTokenRequest(cvr, AuthenticationLevel.VOCES_TRUSTED_SYSTEM.getLevel(), alternativeIdentifier);
    }

    protected Document createSystemSecurityTokenRequest(String cvr, int authenticationLevel) {
        return createSystemSecurityTokenRequest(cvr, authenticationLevel, null);
    }

    private Document createSystemSecurityTokenRequest(String cvr, int authenticationLevel, String alternativeIdentifier) {
        Document tmp = createSecurityTokenRequest(cvr, "0000000000", "doctor", "19901", alternativeIdentifier, "SOSITEST");
        return messageGenerator.setLevelType(tmp, authenticationLevel, "system");
    }

    public Document createUserSecurityTokenRequest(String cvr, String cpr, String role, String code, String alternativeIdentifier) {
        return createUserSecurityTokenRequest(cvr, cpr, role, code, AuthenticationLevel.MOCES_TRUSTED_USER.getLevel(), alternativeIdentifier);
    }

    public Document createUserSecurityTokenRequest(String cvr, String cpr, String role, String code) {
        return createUserSecurityTokenRequest(cvr, cpr, role, code, AuthenticationLevel.MOCES_TRUSTED_USER.getLevel());
    }

    public Document createUserSecurityTokenRequest(String cvr, String cpr, String role, String code, int authenticationLevel) {
        return createUserSecurityTokenRequest(cvr, cpr, role, code, authenticationLevel, null);
    }

    public Document createUserSecurityTokenRequest(String cvr, String cpr, String role, String code, int authenticationLevel, String alternativeIdentifier) {
        Document tmp = createSecurityTokenRequest(cvr, cpr, role, code, alternativeIdentifier, "SOSITEST");
        return messageGenerator.setLevelType(tmp, authenticationLevel, "user");
    }

    public Document createUnknownSecurityTokenRequest(String cvr) {
    		Document document = createSecurityTokenRequest(cvr, "0000000000", "doctor", "19901", null, "SOSITEST");
    		return messageGenerator.setLevelType(document, 3, "ukendt");
    }

    private Document createSecurityTokenRequest(String cvr, String cpr, String role, String authorizationCode, String alternativeIdentifier, String itSystemName) {
        Properties props = System.getProperties();
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_DGWS_VERSION, DGWSConstants.VERSION_1_0_1);
        SOSIFactory factory = new SOSIFactory(new EmptyCredentialVault(), props);
        SecurityTokenRequest req = factory.createNewSecurityTokenRequest();
        UserInfo userInfo = new UserInfo(cpr, "Anders", "And", "aand@andeby.dk", "coolDuck", role!=null?role:"dummy", authorizationCode){
            @Override
            public String getRole() {
                return role;
            }
        };
        IDCard idcard = factory.createNewUserIDCard(
                itSystemName, userInfo, new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, cvr, "orgName"),
                AuthenticationLevel.MOCES_TRUSTED_USER, null, null, null, alternativeIdentifier);
        req.setIDCard(idcard);
        return req.serialize2DOMDocument();
    }

    public void setSamlAttribute(Document doc, String id, String value) {
        new STSMessageGenerator(SignatureUtil.setupCryptoProviderForJVM(), OCES3) {
            protected void setSamlAttribute(Document doc, String id, String value) {
                super.setSamlAttribute(doc, id, value);
            }
        }.setSamlAttribute(doc, id, value);
    }
}
