package dk.sosi.sts.server.facade;

import dk.sosi.seal.model.Message;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.model.RoleDefinition;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.StsCredentialVaults;
import dk.sosi.sts.server.test.TestIdentity;
import dk.sosi.sts.server.test.TestIdentityProvider;
import dk.sosi.testtools.testrequest.MessageGeneratorHelper;
import dk.sosi.testtools.testrequest.STSMessageGenerator;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;

import java.net.MalformedURLException;

import static org.junit.Assert.assertEquals;

public class StsMoces3IT extends AbstractTestStsMessage {
    private OcesInfo.Version version = OcesInfo.Version.OCES3;

    private String cvr = "94354969";

    protected StsMessageFactory stsMessageFactory;

    @Override
	protected TestClient makeTestClient() throws MalformedURLException {
        return new HttpTestClient("/sts/services/SecurityTokenService");
    }

    @Before
    public void setUpSTSMessageTest() throws Exception {
        STSMessageGenerator.VERSION version = STSMessageGenerator.VERSION.valueOf(this.version.name());
        stsMessageFactory = new StsMessageFactory(version);
    }

    /**
      * SSNCPR-MOCES3IT1
      */
    @Test
    public void globalUuid_noCprInRequest_responseIncludesCpr() {
        final CredentialVault vault = SealPPCredentialVaults.getValidMocesGlobalUuid();
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(cvr, null, "doctor", "108NX");
        Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
        Document response = testClient.execute(signedRequest);
        Message message = checkDeserializeResponse(response, true);
        assertSuccess(response);

        assertEquals("1010903333", ((UserIDCard) message.getIDCard()).getUserInfo().getCPR());
    }

    /**
     * SSNCPR-MOCES3IT2
     */
    @Test
    public void globalUuid_correctCprInRequest_responseIncludesCpr() {
        final CredentialVault vault = SealPPCredentialVaults.getValidMocesGlobalUuid();
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(cvr, "1010903333", "doctor", "108NX");
        Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
        Document response = testClient.execute(signedRequest);
        Message message = checkDeserializeResponse(response, true);
        assertSuccess(response);

        assertEquals("1010903333", ((UserIDCard) message.getIDCard()).getUserInfo().getCPR());
    }

    /**
     * SSNCPR-MOCES3IT3
     */
    @Test
    public void globalUuid_wrongCprInRequest_failure() {
        final CredentialVault vault = SealPPCredentialVaults.getValidMocesGlobalUuid();
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(cvr, "1011129876", "doctor", "008NX");
        Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
        Document response = testClient.execute(signedRequest);

        assertFaultActor(response, StsFaultActorValues.UUIDCPR);
        assertFaultStringContains(response, "uuid-cpr mismatch");
    }

    /**
     * SSNCPR-MOCES3IT4
     */
    @Test
    public void certificateSpecificUuid_noCprInRequest_failure() {
        final CredentialVault vault = SealPPCredentialVaults.getValidMocesCertificateSpecificUuid();
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(cvr, null, "doctor", "006NX");
        Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
        Document response = testClient.execute(signedRequest);

        assertFaultActor(response, StsFaultActorValues.STS);
        assertFaultStringContains(response, "signed with unsupported certificate");
    }

    /**
     * SSNCPR-MOCES3IT5
     */
    @Test
    public void certificateSpecificUuid_correctCprInRequest_responseIncludesCpr() {
        final CredentialVault vault = SealPPCredentialVaults.getValidMocesCertificateSpecificUuid();
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(cvr, "0909744321", "doctor", "006NX");
        Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
        Document response = testClient.execute(signedRequest);

        assertFaultActor(response, StsFaultActorValues.STS);
        assertFaultStringContains(response, "signed with unsupported certificate");
    }

    /**
     * SSNCPR-MOCES3IT6
     */
    @Test
    public void certificateSpecificUuid_wrongCprInRequest_failure() {
        final CredentialVault vault = SealPPCredentialVaults.getValidMocesCertificateSpecificUuid();
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(cvr, "1011129876", "doctor", "006NX");
        Document signedRequest = MessageGeneratorHelper.sign(domClone, vault);
        Document response = testClient.execute(signedRequest);
        Message message = checkDeserializeResponse(response, true);

        assertFaultActor(response, StsFaultActorValues.STS);
        assertFaultStringContains(response, "signed with unsupported certificate");
    }

    @Test
    public void globalUuid_noCprInRequest_responseIncludesExpectedRole() {
        TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndOneNationalRole(OcesInfo.Version.OCES3);
        String role = testIdentity.getNationalRoles().get(0).getPublicName();
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, role, null);
        Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
        Document response = testClient.execute(signedRequest);
        Message message = checkDeserializeResponse(response, true);
        assertSuccess(response);

        assertEquals(role, ((UserIDCard) message.getIDCard()).getUserInfo().getRole());
    }
}
