/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.server.SecurityTokenService;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsRequestHandler;
import dk.sosi.sts.server.StsStatus;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.TestUtil;
import dk.sosi.testtools.testrequest.MessageGeneratorHelper;
import org.junit.Ignore;
import org.junit.Test;
import org.mockito.Mockito;
import org.w3c.dom.Document;

@Ignore("SDS-6660")
public class StsOces2EmbeddedTest extends STSMessage2Test {
    private Federation federationSpy;

    @Override
    protected TestClient makeTestClient() {
        MockServicesImpl mockServices = new MockServicesImpl(version);
        federationSpy = Mockito.spy(TestUtil.createTestFederation(mockServices));
        SOSIFactory sosiFactory = mockServices.buildSosiFactory(federationSpy);
        SecurityTokenService sts = new SecurityTokenService(sosiFactory, mockServices, mockServices.getUserValidationService(), "dummy blacklisted ITSystemName");
        StsRequestHandler stsService = new StsRequestHandler(sosiFactory, sts, new StsStatus(), mockServices);
        return new TestClientHelper(stsService);
    }

    /**
     * tc06 moces certificate without cross certificate (invalid)
     * <p/>
     * TODO: in future versions it would be nice to change this, but seal handles these generally
     */
    @Test
    public void testMoces_tc06() {
        Mockito.doThrow(new PKIException("Invalid certificate - CA Issuers (1.3.6.1.5.5.7.48.2) not found under Authority Information Access.")).when(federationSpy).getCertificateStatus(Mockito.any());
        final CredentialVault credentialVaultFromPKCS12 = SealPPCredentialVaults.getValidMocesWithCpr(version);
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "doctor", "19901");
        Document signedRequest = MessageGeneratorHelper.sign(domClone, credentialVaultFromPKCS12);
        Document response = testClient.execute(signedRequest);
        checkDeserializeResponse(response, true);
        assertFaultActor(response, StsFaultActorValues.STS);
        assertFaultStringContains(response, "Invalid certificate - CA Issuers");
    }

    /**
     * tc07 moces certificate with reference to non-existing cross certificate (invalid)
     * <p/>
     * TODO: in future versions it would be nice to change this, but seal handles these generally
     */
    @Test
    public void testMoces_tc07() {
        Mockito.doThrow(new PKIException("Intermediate certificate could not be found. Statuscode = 404. Responsebody: ")).when(federationSpy).getCertificateStatus(Mockito.any());
        final CredentialVault credentialVaultFromPKCS12 = SealPPCredentialVaults.getValidMocesWithCpr(version);
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "doctor", "19901");
        Document signedRequest = MessageGeneratorHelper.sign(domClone, credentialVaultFromPKCS12);
        Document response = testClient.execute(signedRequest);
        checkDeserializeResponse(response, true);
        assertFaultActor(response, StsFaultActorValues.STS);
        assertFaultStringContains(response, "Intermediate certificate could not be found");
    }

    /**
     * tc08 moces certificate with reference to wrong but valid cross certificate (?)
     * <p/>
     * FIXME:
     * Change behavior of STS!?
     * <p/>
     * (there was a problem with this one; see chg mail)
     */
    @Test
    public void testMoces_tc08() {
        Mockito.doThrow(new PKIException("Intermediate certificate not valid in time")).when(federationSpy).getCertificateStatus(Mockito.any());
        Document domClone = stsMessageFactory.createUserSecurityTokenRequest(testIdentity.getCvr(), null, "doctor", "19901");
        Document signedRequest = MessageGeneratorHelper.sign(domClone, testIdentity.getCredentialVault());
        Document response = testClient.execute(signedRequest);
        checkDeserializeResponse(response, true);
        assertFaultActor(response, StsFaultActorValues.STS);
        assertFaultStringContains(response, "Certificate invalid");
    }

}
