/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.bootstrap;

import dk.sosi.seal.model.BasicPrivileges;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.OIOBootstrapToIdentityTokenResponse;
import dk.sosi.seal.model.constants.HealthcareSAMLAttributes;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.facade.DefaultSamlTokenTestModel;
import dk.sosi.sts.server.facade.SamlTokenTestModel;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;

import static dk.sosi.seal.model.OnBehalfOfClaim.OnBehalfOfClaimType.Procuration;
import static dk.sosi.sts.fuldmagt.stub.ProcurationTestModel.*;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.*;
import static org.junit.Assert.*;

public abstract class BootstrapTokenProcurationTest extends BootstrapTokenToIdentityTokenParameterizedTest {
	protected SamlTokenTestModel testModel;

	public BootstrapTokenProcurationTest(boolean isEmbeddedTest) {
		super(isEmbeddedTest);
	}

	@Before
	public void initTestModel() {
		this.testModel = new DefaultSamlTokenTestModel(version);
	}

	@Test
	public void testValidRequestWithProcurations1() throws Exception {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(),
				testModel.getClearTextAudience(), testModel.getPatientCpr().getValue(), false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(CPR_1, Procuration);
		Document requestDocument = buildRequestDocument(requestBuilder);
		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();

		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());

		assertTrue(FMK_READ + " missing", identityToken.getPrivileges().getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + CPR_1).contains(FMK_READ));
		assertNull(identityToken.getPrivileges().getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + CPR_2));
		assertEquals(testModel.getClearTextAudience(), identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute(OIOSAMLAttributes.SPEC_VERSION));
		assertEquals("3", identityToken.getAssuranceLevel());

		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
	}

	@Test
	public void testNoPrivilegesIfClaimsAreNotSupplied() throws Exception {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(),
				testModel.getClearTextAudience(), testModel.getPatientCpr().getValue(), false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();

		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		assertNull(identityToken.getPrivileges());
		assertEquals(testModel.getClearTextAudience(), identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute(OIOSAMLAttributes.SPEC_VERSION));
		assertEquals("3", identityToken.getAssuranceLevel());

		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
	}

	@Test
	public void testValidRequestWithProcurations2() throws Exception {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(),
				testModel.getClearTextAudience(), testModel.getPatientCpr().getValue(), false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(CPR_2, Procuration);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();

		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		final BasicPrivileges privileges = identityToken.getPrivileges();
		assertNull(privileges.getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + CPR_1));
		assertTrue(FMK_WRITE + " missing", identityToken.getPrivileges().getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + CPR_2).contains(FMK_WRITE));
		assertEquals(testModel.getClearTextAudience(), identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute(OIOSAMLAttributes.SPEC_VERSION));
		assertEquals("3", identityToken.getAssuranceLevel());

		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
	}

	@Test
	public void testErrorWhenClaimsNotFulfilled() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(),
				testModel.getClearTextAudience(), testModel.getPatientCpr().getValue(), false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim("1111111119", Procuration);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, "No privileges provided by procuration");
	}

}
