/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.bootstrap;

import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.OIOBSTSAMLAssertion;
import dk.sosi.seal.model.OIOBootstrapToIdentityTokenResponse;
import dk.sosi.seal.model.SubjectRelations11Impl;
import dk.sosi.seal.model.constants.WSSETags;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.StsFaultStringValues;
import dk.sosi.sts.server.facade.DefaultSamlTokenTestModel;
import dk.sosi.sts.server.facade.IdentityTokenResponseValidation;
import dk.sosi.sts.server.facade.SamlTokenTestModel;
import dk.sosi.sts.server.facade.TestClient;
import dk.sosi.sts.server.facade.TestClientHelper;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

import java.io.IOException;

import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.*;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

/**
 * Created by Kent on 22-02-2017.
 */
public class BootstrapTokenToIdentityTokenAdditionalEmbeddedTest extends BootstrapTokenToIdentityTokenParameterizedTest {

	protected TestClientHelper testClientHelper;

	protected SamlTokenTestModel testModel;

	public BootstrapTokenToIdentityTokenAdditionalEmbeddedTest() {
		super(true);
	}

	@Before
	public void initTestModel() {
		this.testModel = new DefaultSamlTokenTestModel(version);
	}

	@Before
	public void setUp() throws Exception {
		EmbeddedBootstrapTokenToIdentityTokenBuilder builder = new EmbeddedBootstrapTokenToIdentityTokenBuilder(testModel, version);
		testClientHelper = builder.build();
	}

	@Override
	protected TestClient getTestClient() {
		return testClientHelper;
	}

	/**
	 * An "extra" testcase, not specified
	 */
	@Test
	public void testOIOSAML_RequestInvalidXML() {

		final Document responseDoc = getResponseDocument(XmlUtil.createEmptyDocument());
		final OIOBootstrapToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertTrue(assertionToIDCardResponse.isFault());
		assertEquals(StsFaultActorValues.STS, assertionToIDCardResponse.getFaultActor());
		assertEquals(StsFaultCodeValues.WST_INVALID_REQUEST, assertionToIDCardResponse.getFaultCode());
		assertTrue(assertionToIDCardResponse.getFaultString(), assertionToIDCardResponse.getFaultString().contains("Unable to parse"));
		assertTrue(assertionToIDCardResponse.getFaultString(), assertionToIDCardResponse.getFaultString().contains(StsFaultStringValues.WST_INVALID_REQUEST));
	}

	@Test
	public void testOIOSAML_RequestBootstrapTokenSignerUntrusted() throws IOException {
		assertionBuilder.setSigningVault(SealPPCredentialVaults.getValidMocesNoCpr(version)); // Vault not in idp configuration
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(),
				testModel.getClearTextAudience(), testModel.getPatientCpr().getValue(), false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);
		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		IdentityTokenResponseValidation.assertFault(response, StsFaultActorValues.ITS, "The certificate that signed the security token is not trusted");
	}

	@Test
	public void testOIOSAML_RequestValidXml() throws Exception {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(),
				testModel.getClearTextAudience(), testModel.getPatientCpr().getValue(), false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		NodeList binarySecurityTokenTagsIRequest = TagUtil.getElementsByTagNameNS(requestDocument.getDocumentElement(), WSSETags.binarySecurityToken);
		assertEquals("Forventede ikke BinarySecurityToken  i request", 0, binarySecurityTokenTagsIRequest.getLength());

		final Document responseDoc = getResponseDocument(requestDocument);

		NodeList binarySecurityTokenTagsIResponse = TagUtil.getElementsByTagNameNS(responseDoc.getDocumentElement(), WSSETags.binarySecurityToken);
		assertEquals("Forventede ikke BinarySecurityToken i response", 0, binarySecurityTokenTagsIResponse.getLength());

		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFalse(response.getFaultString(), response.isFault());
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();

		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		assertEquals(testModel.getClearTextAudience(), identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute("dk:gov:saml:attribute:SpecVer"));
		assertEquals("3", identityToken.getAssuranceLevel());

		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
	}

	@Test
	public void testOIOSAML_RequestValidXmlUsingSecurityTokenReference() throws Exception {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(),
				testModel.getClearTextAudience(), testModel.getPatientCpr().getValue(), true, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		NodeList binarySecurityTokenTagsIRequest = TagUtil.getElementsByTagNameNS(requestDocument.getDocumentElement(), WSSETags.binarySecurityToken);
		assertEquals("Forventede at BinarySecurityToken var angivet i request", 1, binarySecurityTokenTagsIRequest.getLength());

		final Document responseDoc = getResponseDocument(requestDocument);

		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFalse(response.getFaultString(), response.isFault());
		assertSuccess(response, federation);

		NodeList binarySecurityTokenTagsIResponse = TagUtil.getElementsByTagNameNS(responseDoc.getDocumentElement(), WSSETags.binarySecurityToken);
		assertEquals("Forventede at BinarySecurityToken var angivet i response", 1, binarySecurityTokenTagsIResponse.getLength());

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();

		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		assertEquals(testModel.getClearTextAudience(), identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute("dk:gov:saml:attribute:SpecVer"));
		assertEquals("3", identityToken.getAssuranceLevel());

		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
	}
}
