package dk.sosi.sts.server.facade.bootstrap;

import dk.sosi.seal.model.BootstrapTokenBuilder;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.OIOBootstrapToIdentityTokenResponse;
import dk.sosi.seal.model.OIOBootstrapToken;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.Tag;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.facade.DefaultSamlTokenTestModel;
import dk.sosi.sts.server.facade.SamlTokenTestModel;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.io.IOException;
import java.security.cert.X509Certificate;
import java.util.Date;

import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.assertFault;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.assertHolderOfKey;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.assertSuccess;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.assertValidInTime;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

/**
 * Created by Kent on 20-03-2017.
 */
public abstract class BootstrapTokenToIdentityTokenLegacySpecificationTest extends BootstrapTokenToIdentityTokenTest {

    protected SamlTokenTestModel testModel;

	// Udkommenteret - SDS-6660:
	Version version = Version.OCES3; // Legacy token is OCES2

	@Before
	public void initTestModel() {
		this.testModel = new DefaultSamlTokenTestModel(version);
	}

	/**
	 * This is Testcase #T1 in the specification document
	 */
	@Ignore("SDS-6660")
	@Test
	public void testValidRequestWithNoSignatureConfiguration() throws Exception {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();
		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFalse(response.getFaultString(), response.isFault());
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();

		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		assertEquals(audience, identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute("dk:gov:saml:attribute:SpecVer"));
		assertEquals("3", identityToken.getAssuranceLevel());

		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
	}

	/**
	 * This is Testcase #T3 in the specification document
	 */
	@Test
	public void testInvalidEnvelopeSignature() throws Exception {
		BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();

		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());
		Element header = XmlUtil.getFirstChildElementNS(requestDocument.getDocumentElement(), "http://schemas.xmlsoap.org/soap/envelope/", "Header");
		Element security = XmlUtil.getFirstChildElementNS(header, "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd", "Security");
		Element signature = getFirstChild(security, DSTags.signature);
		Element signatureValue = XmlUtil.getFirstChildElementNS(signature, "http://www.w3.org/2000/09/xmldsig#", "SignatureValue");

		signatureValue.getFirstChild().setNodeValue(signatureValue.getFirstChild().getNodeValue() + "-break-");

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * This is Testcase #T4 in the specification document
	 */
	@Test
	public void testSigningCertificateNotWhitelisted() throws Exception {
		BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();

		Document requestDocument = buildRequestDocument(bootstrapToken, SealPPCredentialVaults.getValidMocesNoCpr(version), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Signing certificate not whitelisted for this audience");
	}

	/**
	 * This is Testcase #T5 in the specification document
	 */
	@Test
	public void testSigningCertificateExpired() throws Exception {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();
		Document requestDocument = buildRequestDocument(bootstrapToken, SealPPCredentialVaults.getExpiredVocesVault(version), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Certificate expired or not yet valid");
	}

	/*
	 * Testcase #T6 only in embedded mode
	 */

	/**
	 * This is Testcase #T7 in the specification document
	 */
	@Test
	public void testMissingCprClaim() throws Exception {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();

		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = new OIOBootstrapToIdentityTokenRequestDOMBuilder() {
			protected void validateBeforeBuild() throws ModelException {
			}
		};
		requestBuilder.setCPRNumberClaim(null);

		requestBuilder.setOIOBootstrapToken(bootstrapToken);
		requestBuilder.setSigningVault(testModel.getValidWhitelistedVault());
		requestBuilder.setAudience(testModel.getClearTextAudience());

		final Document requestDocument = requestBuilder.build();
		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR claim missing");
	}

	/**
	 * This is Testcase #T8 in the specification document
	 */
	@Ignore("SDS-6660")
	@Test
	public void testMalformedCprClaim() throws Exception {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();

		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "123456789");
		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR-number must be 10 digits");
	}

	/**
	 * This is Testcase #T9 in the specification document
	 */
	@Ignore("SDS-6660")
	@Test
	public void testWrongCprClaimInRequest() throws Exception {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();

		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "1234567890");
		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR/PID mismatch");
	}

	/**
	 * This is Testcase #T10 in the specification document
	 */
	@Test
	public void testMissingAudience() throws Exception {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();

		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = new OIOBootstrapToIdentityTokenRequestDOMBuilder() {
			protected void validateBeforeBuild() throws ModelException {
			}
		};
		requestBuilder.setCPRNumberClaim(testModel.getPatientCpr().getValue());

		requestBuilder.setOIOBootstrapToken(bootstrapToken);
		requestBuilder.setSigningVault(testModel.getValidWhitelistedVault());
		requestBuilder.setAudience(null);

		final Document requestDocument = requestBuilder.build();
		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Audience is missing");
	}

	/**
	 * This is Testcase #T11 in the specification document
	 */
	@Test
	public void testUnknownAudience() throws Exception {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();
		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), "unknown audience", testModel.getPatientCpr().getValue());

		final Document responseDoc = getResponseDocument(requestDocument);
		final OIOBootstrapToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Audience not configured");
	}

	/**
	 * This is Testcase #T12 in the specification document
	 */
	@Test
	public void testInvalidSignatureBootstrapToken() throws IOException {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.buildInvalidSignature();
		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());

		final Document responseDoc = getResponseDocument(requestDocument);
		final OIOBootstrapToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Signature on OIOSAMLAssertion is invalid");
	}

	/**
	 * This is Testcase #T13 in the specification document
	 */
	@Ignore("SDS-6660")
	@Test
	public void testUntrustedSignerBootstrapToken() throws IOException {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getValidWhitelistedVault(), testModel.getPatientCertificateInfo());
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();
		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());

		final Document responseDoc = getResponseDocument(requestDocument);
		final OIOBootstrapToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "The certificate that signed the security token is not trusted");
	}

	/*
	 * Testcase #T14 and #T15 only in embedded mode
	 */

	/**
	 * This is Testcase #T16 in the specification document
	 */
	@Test
	public void testMocesBootstrapToken() throws IOException {
		final X509Certificate mocesCert = SealPPCredentialVaults.getValidMocesWithCpr(version).getSystemCredentialPair().getCertificate();
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), new CertificateInfo(mocesCert));
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();
		final String cprClaim = "1802602810"; // matches the moces
		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), cprClaim);

		final Document responseDoc = getResponseDocument(requestDocument);
		final OIOBootstrapToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Token not issued to a citizen");
	}

	/**
	 * This is Testcase #T17 in the specification document
	 */
	@Test
	public void testMalformedNameIdBootstrapToken() throws IOException {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), "nameid");
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();
		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());

		final Document responseDoc = getResponseDocument(requestDocument);
		final OIOBootstrapToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Invalid name: nameid");
	}

	/**
	 * This is Testcase #T18 in the specification document
	 */
	@Test
	public void testExpiredBootstrapToken() throws Exception {
		final BootstrapTokenBuilder assertionBuilder = new BootstrapTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo()).expires(new Date(0));
		OIOBootstrapToken bootstrapToken = assertionBuilder.build();

		Document requestDocument = buildRequestDocument(bootstrapToken, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());

		final Document responseDoc = getResponseDocument(requestDocument);
		final OIOBootstrapToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);

		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Bootstrap token no longer valid");
	}

	private static Element getFirstChild(Element ancestor, Tag tag) {
		return (Element) ancestor.getElementsByTagNameNS(tag.getNS(), tag.getTagName()).item(0);
	}
}