/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server.facade.bootstrap;

import dk.sosi.seal.model.*;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.fuldmagt.stub.ProcurationTestModel;
import dk.sosi.sts.integrations.blurring.BlurringService;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.server.Bootstrap2IdwsRequestHandler;
import dk.sosi.sts.server.facade.TestClient;
import dk.sosi.sts.server.facade.TestClientHelper;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.w3c.dom.Document;

import java.util.List;
import java.util.stream.Collectors;

import static dk.sosi.seal.model.BlurredOrganisation.OrgType.*;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.*;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.assertSuccess;
import static dk.sosi.sts.server.facade.TestClientHelper.getMessageId;
import static java.util.Collections.emptyList;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;
import static org.junit.Assume.assumeTrue;
import static org.mockito.Mockito.*;

public class BootstrapTokenToIdentityTokenSpecificationEmbeddedTest extends BootstrapTokenToIdentityTokenSpecificationTest {
	protected TestClientHelper testClientHelper;

	public BootstrapTokenToIdentityTokenSpecificationEmbeddedTest() {
		super(true);
	}

	@Before
	public void setUp() throws Exception {
		EmbeddedBootstrapTokenToIdentityTokenBuilder builder = new EmbeddedBootstrapTokenToIdentityTokenBuilder(testModel, version);
		testClientHelper = builder.build();
	}

	@Override
	protected TestClient getTestClient() {
		return testClientHelper;
	}

	@Test
	public void testWithAudienceThatDoesNotRequiringBlurring_ResponseDoesNotContainBlurring() {
		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getClearTextAudience(),
														   testModel.getPatientCpr().getValue(),
														   false,
														   encryptionKey).build();
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(getResponseDocument(requestDocument));
		assertSuccess(response, federation);
        assertNull(response.getIdentityToken().getBlurringInstructions());
	}

	@Test
	public void testWithAudienceThatRequiresBlurringButUserDoesNot_WithoutProcurationClaim_ResponseDoesNotContainBlurring() {
		assumeTrue(version == OcesInfo.Version.OCES3);
		final CprNr userCpr = new CprNr("1234567890");
		if (assertionBuilder instanceof OIO3BSTCitizenSAMLAssertionBuilder) {
			OIO3BSTCitizenSAMLAssertionBuilder builder = (OIO3BSTCitizenSAMLAssertionBuilder) assertionBuilder;
			builder.setCpr(userCpr.getValue());
			builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + userCpr.getValue());
		} else {
			OIO2BSTCitizenSAMLAssertionBuilder builder = (OIO2BSTCitizenSAMLAssertionBuilder) assertionBuilder;
			builder.setCpr(userCpr.getValue());
			builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + userCpr.getValue());
		}

		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getBlurringAudience(),
														   userCpr.getValue(),
														   false,
														   encryptionKey).build();
		BlurringInstructions11Impl blurringInstructions = assertResponseAndGetBlurringInstructions(requestDocument);
		assertEquals(emptyList(), blurringInstructions.getBlurredOrganisations().stream().filter(
				b -> b.getReason() == FromRelatedPerson).collect(Collectors.toList()));
		assertEquals("test salt", blurringInstructions.getSalt());
	}

	@Test
	public void testWithAudienceThatRequiresBlurringWithoutProcurationClaim_ResponseContainsBlurringForPatient() {
		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getBlurringAudience(),
														   testModel.getPatientCpr().getValue(),
														   false,
														   encryptionKey).build();
		BlurringInstructions11Impl blurringInstructions = assertResponseAndGetBlurringInstructions(requestDocument);
		List<BlurredOrganisation> blurring = blurringInstructions.getBlurredOrganisations().stream().filter(
				b -> b.getReason() == SpecificForPerson).collect(Collectors.toList());
		assertEquals(1, blurring.size());
		assertEquals("12345678", blurring.get(0).getValue());
		assertEquals(Cvr, blurring.get(0).getOrgType());
		assertEquals("test salt", blurringInstructions.getSalt());
	}

	@Test
	public void testWithAudienceThatRequiresBlurring_ResponseContainsBlurringForDepartment() {
		// Arrange
		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getBlurringAudience(),
														   testModel.getPatientCpr().getValue(),
														   false,
														   encryptionKey).build();

		// Act
		BlurringInstructions11Impl blurringInstructions = assertResponseAndGetBlurringInstructions(requestDocument);
		List<BlurredOrganisation> blurring = blurringInstructions.getBlurredOrganisations().stream().filter(
				b -> b.getReason() == SpecificDepartment).collect(Collectors.toList());

		// Assert
		assertEquals(2, blurring.size());
		assertEquals("650901000016000", blurring.get(0).getValue());
		assertEquals(Sor, blurring.get(0).getOrgType());
		assertEquals("7042010", blurring.get(1).getValue());
		assertEquals(Shak, blurring.get(1).getOrgType());
	}

	@Test
	public void testWithAudienceThatDoesNotRequireBlurringWithProcurationClaim_ResponseDoesNotContainBlurring() {
		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getClearTextAudience(),
														   testModel.getPatientCpr().getValue(),
														   new CprNr(ProcurationTestModel.CPR_1),
														   OnBehalfOfClaim.OnBehalfOfClaimType.Procuration,
														   false,
														   encryptionKey).build();
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(getResponseDocument(requestDocument));
		assertSuccess(response, federation);
		assertNull(response.getIdentityToken().getBlurringInstructions());
	}

	@Test
	public void testWithAudienceThatRequiresBlurringWithProcurationClaim_ResponseContainsBlurringForPatients() {
		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getBlurringAudience(),
														   testModel.getPatientCpr().getValue(),
														   new CprNr(ProcurationTestModel.CPR_1),
														   OnBehalfOfClaim.OnBehalfOfClaimType.Procuration,
														   false,
														   encryptionKey).build();
		BlurringInstructions11Impl blurringInstructions = assertResponseAndGetBlurringInstructions(requestDocument);
		List<BlurredOrganisation> blurring = blurringInstructions.getBlurredOrganisations().stream().filter(
				b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList());
		assertEquals(2, blurring.size());
		assertEquals(Cvr, blurring.get(0).getOrgType());
		assertEquals("12345678", blurring.get(0).getValue());
		assertEquals(SpecificForPerson, blurring.get(0).getReason());

		assertEquals(Cvr, blurring.get(1).getOrgType());
		assertEquals("87654321", blurring.get(1).getValue());
		assertEquals(FromRelatedPerson, blurring.get(1).getReason());
		assertEquals("test salt", blurringInstructions.getSalt());
	}

	@Test
	public void testBlurringWithParentalClaim_ResponseContainsBlurring() {
		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
				testModel.getValidWhitelistedVault(),
				testModel.getBlurringAudience(),
				testModel.getPatientCpr().getValue(),
				new CprNr(ProcurationTestModel.CPR_1),
				OnBehalfOfClaim.OnBehalfOfClaimType.ParentalCustody,
				false,
				encryptionKey).build();

		BlurringInstructions11Impl blurringInstructions = assertResponseAndGetBlurringInstructions(requestDocument);
		List<BlurredOrganisation> blurring = blurringInstructions.getBlurredOrganisations().stream().filter(
				b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList());
		assertEquals(2, blurring.size());
		assertEquals(Cvr, blurring.get(0).getOrgType());
		assertEquals("12345678", blurring.get(0).getValue());
		assertEquals(SpecificForPerson, blurring.get(0).getReason());

		assertEquals(Cvr, blurring.get(1).getOrgType());
		assertEquals("87654321", blurring.get(1).getValue());
		assertEquals(FromRelatedPerson, blurring.get(1).getReason());
		assertEquals("test salt", blurringInstructions.getSalt());
	}

	@Test
	public void testBlurringWithWardClaim_ResponseContainsBlurring() {
		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
				testModel.getValidWhitelistedVault(),
				testModel.getBlurringAudience(),
				testModel.getPatientCpr().getValue(),
				new CprNr(ProcurationTestModel.CPR_4),
				OnBehalfOfClaim.OnBehalfOfClaimType.WardCustody,
				false,
				encryptionKey).build();

		BlurringInstructions11Impl blurringInstructions = assertResponseAndGetBlurringInstructions(requestDocument);
		List<BlurredOrganisation> blurring = blurringInstructions.getBlurredOrganisations().stream().filter(
				b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList());
		assertEquals(2, blurring.size());
		assertEquals(Cvr, blurring.get(0).getOrgType());
		assertEquals("12345678", blurring.get(0).getValue());
		assertEquals(SpecificForPerson, blurring.get(0).getReason());

		assertEquals(Cvr, blurring.get(1).getOrgType());
		assertEquals("11223344", blurring.get(1).getValue());
		assertEquals(FromRelatedPerson, blurring.get(1).getReason());
		assertEquals("test salt", blurringInstructions.getSalt());
	}

	@Test
	public void testWithAudienceThatRequiresBlurringWithProcurationClaim_OnlyUserRequireBlurring_ResponseContainsBlurringForPatient() {
		final CprNr procurationByCpr = new CprNr("0987654321");
		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getBlurringAudience(),
														   testModel.getPatientCpr().getValue(),
														   procurationByCpr,
				                                           OnBehalfOfClaim.OnBehalfOfClaimType.Procuration,
														   false,
														   encryptionKey).build();
		BlurringInstructions11Impl blurringInstructions = assertResponseAndGetBlurringInstructions(requestDocument);
		List<BlurredOrganisation> blurring = blurringInstructions.getBlurredOrganisations().stream().filter(
				b -> b.getReason() == SpecificForPerson).collect(Collectors.toList());
		assertEquals(1, blurring.size());
		assertEquals("12345678", blurring.get(0).getValue());
		assertEquals(Cvr, blurring.get(0).getOrgType());
		assertEquals(SpecificForPerson, blurring.get(0).getReason());
		assertEquals("test salt", blurringInstructions.getSalt());
	}

	@Test
	public void testWithAudienceThatRequiresBlurringWithProcurationClaim_OnlyOnBehalfOfRequireBlurring_ResponseContainsBlurringForOnBehalfPatient() {
		assumeTrue(version == OcesInfo.Version.OCES3);
		final CprNr cprNumberNotRequiringBlurring = new CprNr("1234567890");

		if (assertionBuilder instanceof OIO3BSTCitizenSAMLAssertionBuilder) {
			OIO3BSTCitizenSAMLAssertionBuilder builder = (OIO3BSTCitizenSAMLAssertionBuilder) assertionBuilder;
			builder.setCpr(cprNumberNotRequiringBlurring.getValue());
			builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + cprNumberNotRequiringBlurring.getValue());
		} else {
			OIO2BSTCitizenSAMLAssertionBuilder builder = (OIO2BSTCitizenSAMLAssertionBuilder) assertionBuilder;
			builder.setCpr(cprNumberNotRequiringBlurring.getValue());
			builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + cprNumberNotRequiringBlurring.getValue());
		}
		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getBlurringAudience(),
														   cprNumberNotRequiringBlurring.getValue(),
														   new CprNr(ProcurationTestModel.CPR_1),
				                                           OnBehalfOfClaim.OnBehalfOfClaimType.Procuration,
														   false,
														   encryptionKey).build();
		BlurringInstructions11Impl blurringInstructions = assertResponseAndGetBlurringInstructions(requestDocument);
		List<BlurredOrganisation> blurring = blurringInstructions.getBlurredOrganisations().stream().filter(
				b -> b.getReason() == FromRelatedPerson).collect(Collectors.toList());
		assertEquals(1, blurring.size());
		assertEquals(Cvr, blurring.get(0).getOrgType());
		assertEquals("87654321", blurring.get(0).getValue());
		assertEquals(FromRelatedPerson, blurring.get(0).getReason());
		assertEquals("test salt", blurringInstructions.getSalt());
	}

	@Test
	public void testWithAudienceThatRequiresBlurringWithProcurationClaim_NoUserRequireBlurring_ResponseDoesNotContainBlurring() {
		assumeTrue(version == OcesInfo.Version.OCES3);
		final CprNr cpr = new CprNr("1234567890"), onBehalfOf = new CprNr("0987654321");
		if (assertionBuilder instanceof OIO3BSTCitizenSAMLAssertionBuilder) {
			OIO3BSTCitizenSAMLAssertionBuilder builder = (OIO3BSTCitizenSAMLAssertionBuilder) assertionBuilder;
			builder.setCpr(cpr.getValue());
			builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + cpr.getValue());
		} else {
			OIO2BSTCitizenSAMLAssertionBuilder builder = (OIO2BSTCitizenSAMLAssertionBuilder) assertionBuilder;
			builder.setCpr(cpr.getValue());
			builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + cpr.getValue());
		}

		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getBlurringAudience(),
														   cpr.getValue(),
														   onBehalfOf,
				                                           OnBehalfOfClaim.OnBehalfOfClaimType.Procuration,
														   false,
														   encryptionKey).build();
		BlurringInstructions11Impl blurring = assertResponseAndGetBlurringInstructions(requestDocument);
        assertEquals(emptyList(), blurring.getBlurredOrganisations().stream().filter(
				b -> b.getReason() == FromRelatedPerson).collect(Collectors.toList()));
		assertEquals("test salt", blurring.getSalt());
	}

	@Test
	public void testWithAudienceThatRequiresBlurringWithProcurationClaim_NoUserRequireBlurring_FlowIdToTraceRequestBetweenComponentsIsCorrect()
			throws Exception {
		assumeTrue(version == OcesInfo.Version.OCES3);
		final CprNr cpr = new CprNr("1234567890"), onBehalfOf = new CprNr("0987654321");
		if (assertionBuilder instanceof OIO3BSTCitizenSAMLAssertionBuilder) {
			OIO3BSTCitizenSAMLAssertionBuilder builder = (OIO3BSTCitizenSAMLAssertionBuilder) assertionBuilder;
			builder.setCpr(cpr.getValue());
			builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + cpr.getValue());
		} else {
			OIO2BSTCitizenSAMLAssertionBuilder builder = (OIO2BSTCitizenSAMLAssertionBuilder) assertionBuilder;
			builder.setCpr(cpr.getValue());
			builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + cpr.getValue());
		}

		Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
														   testModel.getValidWhitelistedVault(),
														   testModel.getBlurringAudience(),
														   cpr.getValue(),
														   onBehalfOf,
				                                           OnBehalfOfClaim.OnBehalfOfClaimType.Procuration,
														   false,
														   encryptionKey).build();
		String expectedFlowId = getMessageId(requestDocument).get();
		ArgumentCaptor<String> getBlurringsFlowId = ArgumentCaptor.forClass(String.class);
		ArgumentCaptor<String> getSaltFlowId = ArgumentCaptor.forClass(String.class);
		BlurringService blurringServiceMock = mock(BlurringService.class);
		when(blurringServiceMock.getActiveBlurrings(any(), getBlurringsFlowId.capture())).thenReturn(emptyList());
		when(blurringServiceMock.getSalt(getSaltFlowId.capture())).thenReturn("dummy");

		TestClientHelper testClientHelper = new EmbeddedBootstrapTokenToIdentityTokenBuilder(testModel, version).build();
		Bootstrap2IdwsRequestHandler stsService = (Bootstrap2IdwsRequestHandler) testClientHelper.stsService;
		stsService.setBlurringService(blurringServiceMock);

		assertResponseAndGetBlurringInstructions(testClientHelper, requestDocument);
		assertEquals(expectedFlowId, getBlurringsFlowId.getValue());
		assertEquals(expectedFlowId, getSaltFlowId.getValue());
	}

	private BlurringInstructions11Impl assertResponseAndGetBlurringInstructions(TestClient client, Document requestDocument) {
		Document responseDocument = getResponseDocument(client, requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDocument);
		assertSuccess(response, federation);
		return (BlurringInstructions11Impl)response.getIdentityToken().getBlurringInstructions();
	}

	private BlurringInstructions11Impl assertResponseAndGetBlurringInstructions(Document requestDocument) {
		Document responseDocument = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDocument);
		assertSuccess(response, federation);
		return (BlurringInstructions11Impl) response.getIdentityToken().getBlurringInstructions();
	}
}