/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.bootstrap;

import dk.nsp.test.idp.model.OrganizationIdentity;
import dk.sosi.seal.model.*;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.server.facade.HttpTestClient;
import dk.sosi.sts.server.facade.TestClient;
import dk.sundhedsdatastyrelsen.identitetssloering._2024._01._24.*;
import org.junit.Test;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import javax.xml.bind.JAXBException;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.stream.Collectors;

import static dk.nsp.test.idp.OrganizationIdentities.sundhedsdatastyrelsen;
import static dk.sosi.seal.model.BlurredOrganisation.OrgType.Cvr;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.SpecificForPerson;
import static dk.sosi.sts.certificate.OcesInfo.Version.OCES3;
import static dk.sosi.sts.server.facade.HttpTestClient.isLocalTest;
import static dk.sosi.sts.server.facade.HttpTestClient.testEnvironmentHost;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.assertSuccess;
import static dk.sosi.sts.server.test.NspSoap.NspSoapResponseData;
import static dk.sosi.sts.server.test.NspSoap.executeNspRequest;
import static dk.sosi.sts.server.util.HttpUtil.httpGet;
import static org.junit.Assert.*;
import static org.junit.Assume.assumeTrue;

public class BootstrapTokenToIdentityTokenSpecificationIT extends BootstrapTokenToIdentityTokenSpecificationTest {
    private final HttpTestClient testClientHelper = new HttpTestClient("/sts/services/Bst2Idws");

    public BootstrapTokenToIdentityTokenSpecificationIT() throws MalformedURLException { super(false); }

    @Override
    protected TestClient getTestClient() { return testClientHelper; }

    public static String createBlurringAndGetCvr(URI uri, CprNr cpr) {
        try {
            String soapAction = "http://sundhedsdatastyrelsen.dk/identitetssloering/2024/01/24/#CreateBlurring";
            PatientIdType patientId = new PatientIdType();
            patientId.setId(cpr.getValue());
            patientId.setClassification(PatientIdClassificationType.CPR);
            CreateBlurringRequest requestBody = new CreateBlurringRequest();
            requestBody.setPatientId(patientId);
            requestBody.setEndDateTime(toXmlGregorian(LocalDateTime.now().plusDays(1)));

            OrganizationIdentity id = sundhedsdatastyrelsen();
            NspSoapResponseData res = executeNspRequest(uri, soapAction, id, requestBody);
            if (res.isFault) { throw new RuntimeException(res.content); }
            return id.getOrganization().getIdentifier();
        } catch (DatatypeConfigurationException | JAXBException | ParserConfigurationException | IOException |
                 SAXException e) { throw new RuntimeException(e); }
    }

    public static XMLGregorianCalendar toXmlGregorian(LocalDateTime datetime) throws DatatypeConfigurationException {
        GregorianCalendar gregorianCalendar = (GregorianCalendar) GregorianCalendar.getInstance();
        gregorianCalendar.setTime(Date.from(datetime.toInstant(ZoneOffset.UTC)));
        return DatatypeFactory.newInstance().newXMLGregorianCalendar(gregorianCalendar);
    }

    @Test
    public void testWithAudienceThatRequiresBlurringWithoutProcurationClaim_ResponseContainsBlurringForPatient()
            throws URISyntaxException {
        assumeTrue("Only OCES3 certificate (serial nr. " +
                           "'UI:DK-O:G:23550132-5e1f-4e43-a5f9-048acf49e0b8') is configured for this test.",
                   version == OCES3);
        CprNr cpr = new CprNr("2406139002");
        String cvr = isLocalTest
                ? "33257872"
                : createBlurringAndGetCvr(new URI(testEnvironmentHost + "/idsas-registration/20240124/registration"),
                                          cpr);
        if (assertionBuilder instanceof OIO3BSTCitizenSAMLAssertionBuilder) {
            OIO3BSTCitizenSAMLAssertionBuilder builder = (OIO3BSTCitizenSAMLAssertionBuilder) assertionBuilder;
            builder.setCpr(cpr.getValue());
            builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + cpr.getValue());
        } else {
            OIO2BSTCitizenSAMLAssertionBuilder builder = (OIO2BSTCitizenSAMLAssertionBuilder) assertionBuilder;
            builder.setCpr(cpr.getValue());
            builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + cpr.getValue());
        }
        Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
                                                           testModel.getValidWhitelistedVault(),
                                                           "https://minlog",
                                                           cpr.getValue(),
                                                           false,
                                                           encryptionKey).build();
        BlurringInstructions blurringInstructions = assertResponseAndGetBlurringInstructions(requestDocument);

        List<BlurredOrganisation> blurring;
        // Dette afhænger af om blurring er sat til V10 eller V11 i databasekonfigurationen
        if (blurringInstructions instanceof BlurringInstructions11Impl) {
            blurring = ((BlurringInstructions11Impl) blurringInstructions).getBlurredOrganisations();
            assertFalse(((BlurringInstructions11Impl) blurringInstructions).getSalt().isEmpty());
        } else if (blurringInstructions instanceof BlurringInstructions10Impl) {
            blurring = ((BlurringInstructions10Impl) blurringInstructions).getBlurredOrganisations();
            assertFalse(((BlurringInstructions10Impl) blurringInstructions).getSalt().isEmpty());
        } else {
            throw new AssertionError("Unexpected BlurringInstructions implementation");
        }

        blurring = blurring.stream().filter(o -> o.getReason() == SpecificForPerson).collect(Collectors.toList());
        assertEquals(1, blurring.size());
        assertEquals(Cvr, blurring.get(0).getOrgType());
        assertEquals(cvr, blurring.get(0).getValue());
        assertEquals(SpecificForPerson, blurring.get(0).getReason());
    }

    @Test
    public void testWithAudienceThatRequiresBlurringWithoutProcurationClaimWithErrorFromBlurringService_ResponseContainsError()
            throws URISyntaxException {
        assumeTrue("Only OCES3 certificate (serial nr. " +
                           "'UI:DK-O:G:23550132-5e1f-4e43-a5f9-048acf49e0b8') is configured for this test.",
                   version == OCES3);
        URI idsasMock = new URI("http://localhost:8081");
        try { httpGet(idsasMock); } catch (IOException ignored) { return; } // To prevent failure when testing against remote STS
        final CprNr idsasMockFailingCpr = new CprNr("1234567890");
        if (assertionBuilder instanceof OIO3BSTCitizenSAMLAssertionBuilder) {
            OIO3BSTCitizenSAMLAssertionBuilder builder = (OIO3BSTCitizenSAMLAssertionBuilder) assertionBuilder;
            builder.setCpr(idsasMockFailingCpr.getValue());
            builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + idsasMockFailingCpr.getValue());
        } else {
            OIO2BSTCitizenSAMLAssertionBuilder builder = (OIO2BSTCitizenSAMLAssertionBuilder) assertionBuilder;
            builder.setCpr(idsasMockFailingCpr.getValue());
            builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:" + idsasMockFailingCpr.getValue());
        }

        Document requestDocument = createRequestDomBuilder(assertionBuilder.build(),
                                                           testModel.getValidWhitelistedVault(),
                                                           "https://minlog",
                                                           idsasMockFailingCpr.getValue(),
                                                           false,
                                                           encryptionKey).build();
        Document responseDocument = getResponseDocument(requestDocument);
        String fault = getResponseModel(responseDocument).getFaultString();
        assertTrue(fault, fault.contains("Internal error. Call to blurring service failed"));
    }

    private BlurringInstructions assertResponseAndGetBlurringInstructions(Document requestDocument) {
        Document responseDocument = getResponseDocument(requestDocument);
        OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDocument);
        assertSuccess(response, federation);
        return response.getIdentityToken().getBlurringInstructions();
    }
}