package dk.sosi.sts.server.facade.bootstrap;

import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.*;
import dk.sosi.seal.model.dombuilders.DOMBuilderHelper;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.fuldmagt.stub.ProcurationTestModel;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.facade.OIOWSTrustResponseValidation;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import org.apache.xml.security.signature.XMLSignature;
import org.junit.Assume;
import org.junit.Ignore;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.xml.crypto.dsig.DigestMethod;
import java.security.cert.CertificateEncodingException;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;
import java.util.List;

import static dk.sosi.seal.model.OnBehalfOfClaim.OnBehalfOfClaimType.*;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.*;
import static org.junit.Assert.*;

public abstract class BootstrapTokenToIdentityTokenSpecificationTest extends BootstrapTokenToIdentityTokenParameterizedTest {

	public BootstrapTokenToIdentityTokenSpecificationTest(boolean isEmbeddedTest) {
		super(isEmbeddedTest);
	}

	/**
	 * Testcase #1
	 * Kald med udløbet XXXBST afvises med passende fejlbesked.
	 */
	@Test
	public void testExpiredToken() {
		assertionBuilder.setNotOnOrAfter(Date.from(Instant.now().minus(2, ChronoUnit.HOURS)));
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Bootstrap token no longer valid");
	}

	/**
	 * Testcase #2
	 * Kald med ’ugyldig’ XXXBST (fx ’volapyk token’ i sted for SAML) afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidToken() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		requestBuilder.setEncryptionKey(null);
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document invalidToken = XmlUtil.createEmptyDocument();
		invalidToken.appendChild(invalidToken.createElement("test"));
		overwriteAssertionInRequest(invalidToken, requestDocument);
		signRequest(requestDocument);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "No assertion element found on request");
	}

	/**
	 * Testcase #3
	 * Kald med XXXBST med ugyldig signatur afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidSignature() throws Exception {
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		Element assertionElement = assertion.getDOM().getDocumentElement();
		Element subjectNameId = (Element) assertionElement.getElementsByTagNameNS(SAMLTags.nameID.getNS(), SAMLTags.nameID.getTagName()).item(0);
		subjectNameId.setTextContent(subjectNameId.getTextContent() + "breaksignature");
		assertion = assertionClass.getDeclaredConstructor(Element.class).newInstance(assertionElement);

		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Signature on OIOSAMLAssertion is invalid");
	}

	/**
	 * Testcase 4
	 * Kald med XXXBST signeret med udløbet certifikat afvises med passende fejlbesked.
	 */
	@Test
	public void testExpiredCertificate() {
		assertionBuilder.setSigningVault(SealPPCredentialVaults.getExpiredVocesVault(version));
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_REQUEST_FAILED, "Bootstrap token signer expired");
	}

	/**
	 * Testcase 5
	 * Kald med XXXBST signeret med spærret certifikat afvises med passende fejlbesked.
	 */
	@Test
	public void testRevokedCertificate() {
		// TODO SDS-6660 Remove assume below - this test should work with OIO2BST
		Assume.assumeFalse(suiteName.equals(OIOBSTSAMLAssertion.Type.OIO2BST_CITIZEN.name()));
		assertionBuilder.setSigningVault(SealPPCredentialVaults.getRevokedVocesVault(version));
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_REQUEST_FAILED, "Bootstrap token signer revoked");
	}

	/**
	 * Testcase 6
	 * Kald med XXXBST signeret med ikke-trustet certifikat afvises med passende fejlbesked.
	 */
	@Test
	public void testUntrustedCertificate() {
		assertionBuilder.setSigningVault(SealPPCredentialVaults.getValidMocesNoCpr(version)); // Vault not in idp configuration
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "The certificate that signed the security token is not trusted!");
	}

	/**
	 * Testcase 7
	 * (Kun for OIO2BST og OIO3BST) Kald med XXXBST som er krypteret til en anden part (dvs. ikke kan dekrypteres) afvises med passende fejlbesked.
	 */
	@Test
	public void testEncryptionWithUnknownKey() {
		Assume.assumeFalse(suiteName.equals(OIOBSTSAMLAssertion.Type.OIO2BST_LEGACY.name()));
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		requestBuilder.setEncryptionKey(SealPPCredentialVaults.getValidMocesNoCpr(version).getSystemCredentialPair().getCertificate().getPublicKey());
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "Could not decrypt Assertion");
	}

	/**
	 * Testcase 8
	 * Kald med XXXBST som ikke overholder den ud fra konfigurationen forventede tokenprofil
	 * (eksempelvis mangel på påkrævede attributter eller forkert tokenprofil) afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidTokenProfile() {
		switch (OIOBSTSAMLAssertion.Type.valueOf(suiteName)) {
			case OIO2BST_LEGACY :
				((BootstrapTokenBuilder) assertionBuilder).setIssuer("https://oio2bst-issuer.dk");
				break;
			case OIO2BST_CITIZEN :
				((OIO2BSTCitizenSAMLAssertionBuilder) assertionBuilder).setIssuer("https://oio2bst-legacy-issuer.dk");
				break;
			case OIO3BST_CITIZEN :
				((OIO3BSTCitizenSAMLAssertionBuilder) assertionBuilder).setIssuer("https://oio2bst-issuer.dk");
				break;
		}

		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_REQUEST_FAILED, "BST2IDWS assertion is not allowed type");
	}

	/**
	 * Testcase 8
	 * Kald med XXXBST som ikke overholder den ud fra konfigurationen forventede tokenprofil
	 * (eksempelvis mangel på påkrævede attributter eller forkert tokenprofil) afvises med passende fejlbesked.
	 */
	@Test
	public void testMissingAttribute() throws Exception {
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		Document assertionDocument = assertion.getDOM();
		Element assertionElement = assertionDocument.getDocumentElement();

		Element attributeStatement = (Element) assertionElement.getElementsByTagNameNS(SAMLTags.attributeStatement.getNS(), SAMLTags.attributeStatement.getTagName()).item(0);
		NodeList attributes = assertionElement.getElementsByTagNameNS(SAMLTags.attribute.getNS(), SAMLTags.attribute.getTagName());
		for (int i = 0; i < attributes.getLength(); i++) {
			Element attribute = (Element) attributes.item(i);
			if (attribute.getAttribute("Name").equals(OIOSAMLAttributes.CPR_NUMBER) || attribute.getAttribute("Name").equals(OIOSAML3Attributes.CPR)) {
				attributeStatement.removeChild(attribute);
			}
		}

		signAssertion(assertionDocument);
		assertion = assertionClass.getDeclaredConstructor(Element.class).newInstance(assertionElement);

		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		if (suiteName.equals(OIOBSTSAMLAssertion.Type.OIO2BST_LEGACY.name())) {
			assertSuccess(response, federation); // Cpr er ikke oblikatorisk for oio2bst-legacy
		} else {
			assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Mandatory 'cpr' SAML attribute", "is missing");
		}
	}

	/**
	 * Testcase 9
	 * Kald med XXXBST med for lav AssuranceLevel afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidAssuranceLevel() {
		String expectedErrorMessage = null;
		switch (OIOBSTSAMLAssertion.Type.valueOf(suiteName)) {
			case OIO2BST_LEGACY :
				((BootstrapTokenBuilder) assertionBuilder).setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_2);
				expectedErrorMessage = "BST2IDWS assertion do not have a valid AssuranceLevel 2. AssuranceLevel 3 or higher required.";
				break;
			case OIO2BST_CITIZEN :
				((OIO2BSTCitizenSAMLAssertionBuilder) assertionBuilder).setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_2);
				expectedErrorMessage = "BST2IDWS assertion do not have a valid AssuranceLevel 2. AssuranceLevel 3 or higher required.";
				break;
			case OIO3BST_CITIZEN :
				((OIO3BSTCitizenSAMLAssertionBuilder) assertionBuilder).setAssuranceLevel(AssuranceLevel.NSIS.Low);
				expectedErrorMessage = "BST2IDWS assertion do not have a valid AssuranceLevel Low. AssuranceLevel Substantial or High required.";
				break;
		}

		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_REQUEST_FAILED, expectedErrorMessage);
	}

	/**
	 * Testcase 10
	 * Kald med XXXBST med anden audience end SOSI-STS afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidAudience() {
		assertionBuilder.setAudience("Invalid audience");
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_REQUEST_FAILED, "BST2IDWS assertion has an invalid audience");
	}

	/**
	 * Testcase 11
	 * Kald med ugyldig holder-of-key relation afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidHolderOfKeyRelation() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		requestBuilder.setSigningVault(SealPPCredentialVaults.getValidVocesVault(version)); // Does not match holder of key in assertion
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		if (suiteName.equals(OIOBSTSAMLAssertion.Type.OIO2BST_LEGACY.name())) {
			assertSuccess(response, federation); // Holder-of-key relation tjekkes ikke for oio2bst-legacy
		} else {
			assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "Signing certificate and holder-of-key certificate relation do not match");
		}
	}

	/**
	 * Testcase 12
	 * Kald med gyldig holder-of-key relation men brudt beskeds-signatur afvises med passende fejlbesked.
	 */
	@Test
	public void testBrokenMessageSignature() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);
		requestDocument.getDocumentElement().getElementsByTagName("auth:ClaimType").item(0).setTextContent(" ");

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase 13
	 * Kald med gyldig holder-of-key relation men manglende signering af påkrævede beskeds-elementer afvises med passende fejlbesked.
	 */
	@Test
	public void testMissingBodySignature() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = DOMBuilderHelper.buildWithoutSigning(requestBuilder);
		new MissingMessageSignatureSigner(testModel.getValidWhitelistedVault(), requestDocument).skipUri("body").enhanceAndSign();

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase 13
	 */
	@Test
	public void testMissingTsSignature() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = DOMBuilderHelper.buildWithoutSigning(requestBuilder);
		new MissingMessageSignatureSigner(testModel.getValidWhitelistedVault(), requestDocument).skipUri("ts").enhanceAndSign();

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase 13
	 */
	@Test
	public void testMissingMessageIDSignature() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = DOMBuilderHelper.buildWithoutSigning(requestBuilder);
		new MissingMessageSignatureSigner(testModel.getValidWhitelistedVault(), requestDocument).skipUri("messageID").enhanceAndSign();

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase 13
	 */
	@Test
	public void testMissingActionSignature() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = DOMBuilderHelper.buildWithoutSigning(requestBuilder);
		new MissingMessageSignatureSigner(testModel.getValidWhitelistedVault(), requestDocument).skipUri("action").enhanceAndSign();

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase 14
	 * Kald med manglende angivelse af ønsket audience afvises med passende fejlbesked.
	 */
	@Test
	public void testMissingAudience() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);
		Element appliesTo = (Element) requestDocument.getDocumentElement().getElementsByTagNameNS(WSPTags.appliesTo.getNS(), WSPTags.appliesTo.getTagName()).item(0);
		appliesTo.getParentNode().removeChild(appliesTo);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Audience is missing");
	}

	/**
	 * Testcase 15
	 * Kald med angivelse af ønsket audience, som ikke er konfigureret i STS, afvises med passende fejlbesked.
	 */
	@Test
	public void testUnknownAudience() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		requestBuilder.setAudience("unknown audience");
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Audience not configured");
	}

	/**
	 * Testcase 16
	 * Kald med andre WS-Trust claims end de som information under Krav 1 angivne afvises med passende fejlbesked.
	 */
	@Test
	public void testUnknownClaims() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim("test", Procuration);
		Document requestDocument = buildRequestDocument(requestBuilder);
		Element claims = (Element) requestDocument.getDocumentElement().getElementsByTagNameNS(WSTTags.claims.getNS(), WSTTags.claims.getTagName()).item(0);

		Element claim = getChildByNameAndAttribute(claims, WSFAuthTags.claimType.getNS(), WSFAuthTags.claimType.getTagName(), "Uri", HealthcareSAMLAttributes.ON_BEHALF_OF);
		assertNotNull("Unexpected test data", claim);
		claim.setAttribute("Uri", HealthcareSAMLAttributes.USER_AUTHORIZATION_CODE); // Replace procuration claim with an invalid claim
		signRequest(requestDocument);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Request contains invalid claims");
	}

	/**
	 * Testcase 17
	 * Kald med manglende påkrævet WS-Trust claim (’CPR-nummer’, se Krav 2) afvises med passende fejlbesked.
	 */
	@Test
	public void testMissingCprClaim() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);
		Element cprClaim = getChildByNameAndAttribute(requestDocument.getDocumentElement(), WSFAuthTags.claimType.getNS(), WSFAuthTags.claimType.getTagName(), "Uri", OIOSAMLAttributes.CPR_NUMBER);
		assertNotNull("Unexpected test data", cprClaim);
		cprClaim.getParentNode().removeChild(cprClaim);
		signRequest(requestDocument);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR claim missing");
	}

	/**
	 * Testcase 18
	 * Kald med ugyldige eller tomme WS-Trust claims afvises med passende fejlbesked.
	 */
	@Test
	public void testEmptyCprClaim() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);
		Element cprClaim = getChildByNameAndAttribute(requestDocument.getDocumentElement(), WSFAuthTags.claimType.getNS(), WSFAuthTags.claimType.getTagName(), "Uri", OIOSAMLAttributes.CPR_NUMBER);
		assertNotNull("Unexpected test data", cprClaim);
		Element cprValue = (Element) cprClaim.getElementsByTagNameNS(WSFAuthTags.value.getNS(), WSFAuthTags.value.getTagName()).item(0);
		cprValue.setTextContent("");
		signRequest(requestDocument);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR claim missing");
	}

	/**
	 * Testcase 18
	 * Kald med ugyldige eller tomme WS-Trust claims afvises med passende fejlbesked.
	 */
	@Test
	public void testEmptyProcurationClaim() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim("", Procuration);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "OnBehalfOf claim must not be empty");
	}

	/**
	 * Testcase 18
	 * Kald med ugyldige eller tomme WS-Trust claims afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidProcurationClaim() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim("1234", Procuration);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR-number must be 10 digits");
	}

	/**
	 * Testcase 19
	 * Kald med gyldig XXXBST og gyldig holder-of-key relation og valide claims,
	 * men hvor klient-certifikatet (holder-of-key certifikatet) ikke er whitelistet
	 * til den i kaldet ønskede audience afvises med passende fejlbesked.
	 */
	@Test
	public void testHolderOfKeyNotWhitelisted() {
		assertionBuilder.setHolderOfKeyCertificate(SealPPCredentialVaults.getValidMocesNoCpr(version).getSystemCredentialPair().getCertificate());
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertionBuilder.build(), testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "0501792275", false, encryptionKey);
		requestBuilder.setSigningVault(SealPPCredentialVaults.getValidMocesNoCpr(version));
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Signing certificate not whitelisted for this audience");
	}

	/**
	 * Testcase 20
	 * Kald med gyldig XXXBST og gyldig holder-of-key relation og valide claims,
	 * og hvor klient-certifikatet (holder-of-key certifikatet) er whitelistet
	 * til den i kaldet ønskede audience går godt.
	 */
	@Test
	public void testValidRequest() throws CertificateEncodingException {
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		final IdentityToken identityToken = response.getIdentityToken();
		assertEquals(cpr, identityToken.getCpr());
		assertEquals(audience, identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getSpecVersion());
		assertEquals("3", identityToken.getAssuranceLevel());
		assertHolderOfKey(identityToken, requestSigningVault.getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
	}

	@Test
	public void testValidRequestWithProcurationClaim() {
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim("1111111118", Procuration); // Setup in ProcurationTestModel
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		final IdentityToken identityToken = response.getIdentityToken();
		assertEquals(cpr, identityToken.getCpr());
		BasicPrivileges privileges = identityToken.getPrivileges();
		assertNotNull(privileges);
		assertEquals(1, privileges.getScopes().size());
		assertTrue(privileges.getScopes().contains(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":1111111118"));
		List<String> privilegesForClaim = privileges.getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":1111111118");
		assertTrue(privilegesForClaim.contains(ProcurationTestModel.FMK_READ));
		assertFalse(privilegesForClaim.contains(ProcurationTestModel.FMK_WRITE));
	}

	@Test
	public void testValidRequestWithInvalidProcurationClaim() {
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim("1111111111", Procuration); // Not in ProcurationTestModel
		Document requestDocument = buildRequestDocument(requestBuilder);

		final Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "No privileges provided by procuration");
	}

	@Test
	public void testValidRequestWithInvalidSubjectRelationsClaim_returnsFault() {
		// Arrange
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		String relatedPersonId = "1111111120";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(relatedPersonId, ParentalCustody);

		// Act
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST,
				"The person is not in parental custody of the given person");
	}

	private void signRequest(Document requestDocument) {
		new LibertyMessageDOMEnhancer(testModel.getValidWhitelistedVault(), SignatureConfiguration.DEFAULT_KEY_INFO_TYPE, requestDocument, false).enhanceAndSign();
	}

	private void signAssertion(Document assertionDocument) {
		Element existingSignature = XmlUtil.getFirstChildElementNS(assertionDocument.getDocumentElement(), NameSpaces.DSIG_SCHEMA, DSTags.SIGNATURE);
		if (existingSignature != null) {
			assertionDocument.getDocumentElement().removeChild(existingSignature);
		}
		String assertionID = assertionDocument.getDocumentElement().getAttribute(SAMLAttributes.ID);
		SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[] { assertionID }, assertionID, IDValues.Id);
		signatureConfiguration.setSignatureMethodAlgorithmUri(XMLSignature.ALGO_ID_SIGNATURE_RSA_SHA256);
		signatureConfiguration.setDigestMethodAlgorithmUri(DigestMethod.SHA256);
		Node subjectNode = assertionDocument.getDocumentElement().getElementsByTagName(SAMLTags.SUBJECT_PREFIXED).item(0);
		signatureConfiguration.setSignatureSiblingNode(subjectNode);
		SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(testModel.getExternalTrustedIdp()), assertionDocument, signatureConfiguration);
	}

	private void overwriteAssertionInRequest(Document assertionDocument, Document requestDocument) {
		Element actAs = (Element)requestDocument.getDocumentElement().getElementsByTagNameNS("http://docs.oasis-open.org/ws-sx/ws-trust/200802", "ActAs").item(0);
		actAs.removeChild(actAs.getFirstChild());
		actAs.appendChild(requestDocument.importNode(assertionDocument.getDocumentElement(), true));
	}

	private static Element getChildByNameAndAttribute(Element ancestor, String tagNS, String tagName, String attributeKey, String attributeValue) {
		NodeList children = ancestor.getElementsByTagNameNS(tagNS, tagName);
		for (int i = 0; i < children.getLength(); i++) {
			Element child = (Element) children.item(i);
			if (attributeValue.equals(child.getAttribute(attributeKey))) {
				return child;
			}
		}
		return null;
	}
}