/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.bootstrap;

import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.OIOBSTSAMLAssertion;
import dk.sosi.seal.model.OIOBootstrapToIdentityTokenResponse;
import dk.sosi.seal.model.SubjectRelations11Impl;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.facade.*;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;

import java.util.Optional;

import static dk.sosi.seal.model.OnBehalfOfClaim.OnBehalfOfClaimType.*;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.*;
import static org.junit.Assert.*;

public class BootstrapTokenToIdentityTokenSubjectRelationsTest extends BootstrapTokenToIdentityTokenParameterizedTest {

	protected TestClientHelper testClientHelper;
	protected TestClientHelper testClientHelperWithFailingPersonInformationClient;
	protected DefaultSamlTokenTestModel testModel;

	public BootstrapTokenToIdentityTokenSubjectRelationsTest() {
		super(true);
	}

	@Before
	public void initTestModel() {
		this.testModel = new DefaultSamlTokenTestModel(version);
	}

	@Before
	public void setUp() throws Exception {
		EmbeddedBootstrapTokenToIdentityTokenBuilder builder = new EmbeddedBootstrapTokenToIdentityTokenBuilder(testModel, version);
		testClientHelper = builder.build();
		testClientHelperWithFailingPersonInformationClient = builder.buildWithFailingPersonInformationClient();
	}

	@Override
	protected TestClient getTestClient() {
		return testClientHelper;
	}

	@Test
	public void testValidRequestWithParentalClaim() {
		// Arrange
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		String relatedPersonId = "1111111118";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(relatedPersonId, ParentalCustody);

		// Act
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		assertSuccess(response, federation);
		assertNotNull(response.getIdentityToken());
		IdentityToken identityToken = response.getIdentityToken();
		assertEquals(cpr, identityToken.getCpr());
		SubjectRelations11Impl subjectRelations = (SubjectRelations11Impl) identityToken.getSubjectRelations();
		assertNotNull(subjectRelations);
		assertEquals(1, subjectRelations.getVerifiedRelations().size());
		assertEquals(relatedPersonId, subjectRelations.getVerifiedRelations().get(0).getRelatedPersonID());
		assertEquals(SubjectRelations11Impl.RelationType.PARENTAL_CUSTODY_HOLDER, subjectRelations.getVerifiedRelations().get(0).getRelationType());
	}

	@Test
	public void testValidRequestWithWardCustodyClaim() {
		// Arrange
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		String wardId = "1111111119";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(wardId, WardCustody);

		// Act
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		assertSuccess(response, federation);
		assertNotNull(response.getIdentityToken());
		IdentityToken identityToken = response.getIdentityToken();
		assertEquals(cpr, identityToken.getCpr());
		SubjectRelations11Impl subjectRelations = (SubjectRelations11Impl) identityToken.getSubjectRelations();
		assertNotNull(subjectRelations);
		assertEquals(1, subjectRelations.getVerifiedRelations().size());
		assertEquals(wardId, subjectRelations.getVerifiedRelations().get(0).getRelatedPersonID());
		assertEquals(SubjectRelations11Impl.RelationType.WARD_CUSTODY_HOLDER, subjectRelations.getVerifiedRelations().get(0).getRelationType());
	}

	@Test
	public void testValidRequestWithWardCustodyClaim_notWhitelisted_returnsFault() {
		// Arrange
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String cpr = "0501792275";
		String wardId = "1111111119";
		String notWhitelistedAudience = testModel.getClearTextAudienceNoWardWhitelisting();
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, notWhitelistedAudience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(wardId, WardCustody);

		// Act
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST,
				"Ward claim not whitelisted for this certificate SSN and audience");
	}

	@Test
	public void testValidRequestWithPartlyWardCustodyClaim_returnsFault() {
		// Arrange
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		String wardId = "1111111118";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(wardId, PartlyWardCustody);

		// Act
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST,
				"partlyWardCustody is not allowed as a Subject Relation");
	}

	@Test
	public void testValidRequestWithInvalidParentalClaim_returnsFault() {
		// Arrange
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		String relatedPersonId = "1111111120";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(relatedPersonId, ParentalCustody);

		// Act
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST,
				"The person is not in parental custody of the given person");
	}

	@Test
	public void testValidRequestWithInvalidWardCustodyClaim_returnsFault() {
		// Arrange
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		String wardId = "1111111120";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(wardId, WardCustody);

		// Act
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST,
				"The person is not in ward custody of the given person");
	}

	@Test
	public void testValidRequestWithParentalClaim_returnsAgeInClaim() {
		// Arrange
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		String relatedPersonId = "1111111118";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(relatedPersonId, ParentalCustody);

		// Act
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document responseDoc = getResponseDocument(requestDocument);
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		IdentityToken identityToken = response.getIdentityToken();
		SubjectRelations11Impl subjectRelations = (SubjectRelations11Impl) identityToken.getSubjectRelations();
		assertEquals(Optional.of(16), subjectRelations.getVerifiedRelations().get(0).getRelatedPersonAge());
	}

	@Test
	public void testValidRequestButPersonInformationCallFails() {
		// Arrange
		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		String audience = testModel.getClearTextAudience();
		String cpr = "0501792275";
		String relatedPersonId = "1111111118";
		GenericCredentialVault requestSigningVault = testModel.getValidWhitelistedVault();
		OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder = createRequestDomBuilder(assertion, requestSigningVault, audience, cpr, false, encryptionKey);
		requestBuilder.setOnBehalfOfClaim(relatedPersonId, ParentalCustody);

		// Act
		Document requestDocument = buildRequestDocument(requestBuilder);
		Document responseDoc = testClientHelperWithFailingPersonInformationClient.execute(
				XmlUtil.node2String(requestDocument, false, false)).document;
		OIOBootstrapToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED,
				"Error calling PersonInformationService");
	}
}
