package dk.sosi.sts.server.facade.bootstrap;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.OIOBSTSAMLAssertion;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.fuldmagt.stub.ProcurationTestModel;
import dk.sosi.sts.integrations.blurring.BlurringService;
import dk.sosi.sts.integrations.blurring.BlurringServiceCaller;
import dk.sosi.sts.integrations.blurring.IdCardFactory;
import dk.sosi.sts.integrations.personinformation.Custody;
import dk.sosi.sts.integrations.personinformation.PersonInformationClient;
import dk.sosi.sts.model.BlurredOrganization;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.IdpConfig;
import dk.sosi.sts.server.Bootstrap2IdwsRequestHandler;
import dk.sosi.sts.server.NboConfig;
import dk.sosi.sts.server.StsStatus;
import dk.sosi.sts.server.facade.*;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.StsCredentialVaults;
import dk.sosi.sts.service.IdpConfigService;
import dk.sosi.sts.service.ItsConfigService;
import org.junit.internal.AssumptionViolatedException;

import java.security.cert.CertificateEncodingException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import static dk.sosi.sts.model.BlurredOrganization.Type.*;
import static java.util.Collections.emptyList;
import static java.util.Collections.singletonList;
import static java.util.stream.Stream.of;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class EmbeddedBootstrapTokenToIdentityTokenBuilder {
    private final SamlTokenTestModel testModel;
    private final OcesInfo.Version version;
    private NboConfig nboConfig;
    private Bootstrap2IdwsRequestHandler requestHandler;

    public EmbeddedBootstrapTokenToIdentityTokenBuilder(SamlTokenTestModel testModel, OcesInfo.Version version) {
        this.testModel = testModel;
        this.version = version;
    }

    public TestClientHelper build() throws Exception {
        requestHandler = createRequestHandler();

        // PersonInformationClient for testing
        PersonInformationClient personInformationClient = mock(PersonInformationClient.class);
        when(personInformationClient.personCustody("0501792275")).thenReturn(of(
                Custody.fromValues("WARDCUSTODY", "1111111119"),
                Custody.fromValues("PARENTALCUSTODY", "1111111118"))
                .collect(Collectors.toSet()));
        when(personInformationClient.personAge("1111111118")).thenReturn(16);
        requestHandler.setPersonInformationClient(personInformationClient);

        return new TestClientHelper(requestHandler);
    }

    public TestClientHelper buildWithFailingPersonInformationClient() throws Exception {
        requestHandler = createRequestHandler();

        PersonInformationClient personInformationClient = mock(PersonInformationClient.class);
        when(personInformationClient.personCustody("0501792275")).thenThrow(new RuntimeException("Error calling PersonInformationService"));
        when(personInformationClient.personAge("1111111118")).thenThrow(new RuntimeException("Error calling PersonInformationService"));
        requestHandler.setPersonInformationClient(personInformationClient);

        return new TestClientHelper(requestHandler);
    }

    private Bootstrap2IdwsRequestHandler createRequestHandler() throws CertificateEncodingException {
        Bootstrap2IdwsRequestHandler requestHandler = new Bootstrap2IdwsRequestHandler();
        nboConfig = SamlTokenEmbeddedTestHelper.buildNboConfig(testModel.getAllTrustedCertificates());
        ItsConfigService itsConfig = SamlTokenEmbeddedTestHelper.buildItsConfigService(testModel);
        final MockServicesImpl mockServices = new MockServicesImpl(version);
        SOSIFactory sosiFactory = mockServices.buildSosiFactory(true);

        requestHandler.setNboConfig(nboConfig);
        requestHandler.setItsConfigService(itsConfig);
        requestHandler.setIdpConfigService(initIdpConfig());
        requestHandler.setCrlService(mockServices);
        requestHandler.setSosiFactory(sosiFactory);
        requestHandler.setStsStatus(new StsStatus());
        requestHandler.setOcesPidService(SamlTokenEmbeddedTestHelper.initMockOcesPidService());
        requestHandler.setProcurationService(new ProcurationMockService());
        requestHandler.setBlurringService(new BlurringService(mock(SOSIFactory.class), mock(IdCardFactory.class), mock(
                BlurringServiceCaller.class)) {

            @Override
            public List<BlurredOrganization> getActiveBlurrings(CprNr patientId, String flowId) {
                return patientId.equals(DefaultSamlTokenTestModel.EXISTING_CPR) ?
                               of(new BlurredOrganization("12345678", Cvr)).collect(Collectors.toList()) :
                       patientId.getValue().equals(ProcurationTestModel.CPR_1) ?
                               of(new BlurredOrganization("87654321", Cvr)).collect(Collectors.toList()) :
                       patientId.getValue().equals(ProcurationTestModel.CPR_4) ?
                               of(new BlurredOrganization("11223344", Cvr)).collect(Collectors.toList()) :
                       emptyList();
            }

            @Override
            public List<BlurredOrganization> getAllActiveOrgBlurrings(String flowId) {
                return of(new BlurredOrganization("650901000016000", Sor),
                        new BlurredOrganization("7042010", Shak)).collect(Collectors.toList());

            }

            @Override
            public String getSalt(String flowId) { return "test salt"; }
        });
        return requestHandler;
    }


    private IdpConfigService initIdpConfig() throws CertificateEncodingException {
        final List<String> signingKeys = getSigningKeys();
        final String encryptionKey = XmlUtil.toBase64(testModel.getValidWhitelistedVault().getSystemCredentialPair().getPrivateKey().getEncoded());
        final List<String> encryptionKeys = new ArrayList<>(singletonList(encryptionKey));

        IdpConfigService configService = mock(IdpConfigService.class);

        when(configService.getAllEncryptionKeys()).thenReturn(encryptionKeys);
        when(configService.getIdpConfiguration("https://oio2bst-issuer.dk")).thenReturn(createIdpConfig(signingKeys, encryptionKeys, "https://oio2bst-issuer.dk", testModel.getClearTextAudience(), OIOBSTSAMLAssertion.Type.OIO2BST_CITIZEN));
        when(configService.getIdpConfiguration("https://oio3bst-issuer.dk")).thenReturn(createIdpConfig(signingKeys, encryptionKeys, "https://oio3bst-issuer.dk", testModel.getClearTextAudience(), OIOBSTSAMLAssertion.Type.OIO3BST_CITIZEN));
        when(configService.getIdpConfiguration("https://oio2bst-legacy-issuer.dk")).thenReturn(createIdpConfig(signingKeys, encryptionKeys, "https://oio2bst-legacy-issuer.dk", testModel.getClearTextAudience(), OIOBSTSAMLAssertion.Type.OIO2BST_LEGACY));
        when(configService.getIdpConfiguration("TESTSTS")).thenReturn(createIdpConfig(signingKeys, encryptionKeys, "TESTSTS", testModel.getClearTextAudience(), OIOBSTSAMLAssertion.Type.OIO2BST_LEGACY));
        return configService;
    }

    private List<String> getSigningKeys() throws CertificateEncodingException {
        final List<String> signingKeys = new ArrayList<>();
        try {
            final String signingKey = XmlUtil.toBase64(testModel.getExternalTrustedIdp().getSystemCredentialPair().getCertificate().getEncoded());
            signingKeys.add(signingKey);
        } catch (AssumptionViolatedException ignored) {}
        try {
            final String signingKey = XmlUtil.toBase64(SealPPCredentialVaults.getExpiredVocesVault(version).getSystemCredentialPair().getCertificate().getEncoded());
            signingKeys.add(signingKey);
        } catch (AssumptionViolatedException ignored) {}
        try {
            final String signingKey = XmlUtil.toBase64(StsCredentialVaults.getStsCredentialVault(version).getSystemCredentialPair().getCertificate().getEncoded());
            signingKeys.add(signingKey);
        } catch (AssumptionViolatedException ignored) {}
        try {
            final String signingKey = XmlUtil.toBase64(SealPPCredentialVaults.getExpiredFocesVault(version).getSystemCredentialPair().getCertificate().getEncoded());
            signingKeys.add(signingKey);
        } catch (AssumptionViolatedException ignored) {}
        try {
            final String signingKey = XmlUtil.toBase64(SealPPCredentialVaults.getRevokedFocesVault(version).getSystemCredentialPair().getCertificate().getEncoded());
            signingKeys.add(signingKey);
        } catch (AssumptionViolatedException ignored) {}
        try {
            final String signingKey = XmlUtil.toBase64(SealPPCredentialVaults.getRevokedVocesVault(version).getSystemCredentialPair().getCertificate().getEncoded());
            signingKeys.add(signingKey);
        } catch (AssumptionViolatedException ignored) {}
        return signingKeys;
    }
    private IdpConfig createIdpConfig(List<String> signingKeys, List<String> encryptionKeys, String issuer, String audience, OIOBSTSAMLAssertion.Type tokenProfile) {
        List<OIOBSTSAMLAssertion.Type> tokenProfiles = new LinkedList<>();
        tokenProfiles.add(tokenProfile);
        return createIdpConfig(signingKeys,encryptionKeys, issuer, audience, tokenProfiles);
    }

    private IdpConfig createIdpConfig(List<String> signingKeys, List<String> encryptionKeys, String issuer, String audience, List<OIOBSTSAMLAssertion.Type> tokenProfiles) {
        return new IdpConfig() {

            @Override
            public String getValidateHOK() {
                return null;
            }

            @Override
            public List<String> getTokenProfiles() {
                List<String> tokenProfileNames = new LinkedList<>();
                for (OIOBSTSAMLAssertion.Type tokenProfile : tokenProfiles) {
                    tokenProfileNames.add(tokenProfile.name());
                }
                return tokenProfileNames;
            }

            @Override
            public List<String> getCertificates() {
                return null;
            }

            @Override
            public List<String> getSigningKeys() {
                return signingKeys;
            }

            @Override
            public String getIssuer() {
                return issuer;
            }

            @Override
            public String getAudience() {
                return audience;
            }

            @Override
            public List<String> getEncryptionKeys() {
                return encryptionKeys;
            }
        };
    }

    public NboConfig getNboConfig() {
        return nboConfig;
    }

    public Bootstrap2IdwsRequestHandler getRequestHandler() {
        return requestHandler;
    }
}