package dk.sosi.sts.server.facade.bst2sosi;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.*;
import dk.sosi.seal.model.dombuilders.DOMBuilderHelper;
import dk.sosi.seal.model.dombuilders.OIOBSTSAMLAssertionToIDCardRequestDOMBuilder;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.*;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.facade.*;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.TestIdentity;
import dk.sosi.sts.server.test.TestIdentityProvider;
import dk.sosi.sts.server.test.TestUtil;
import org.apache.commons.lang3.StringUtils;
import org.apache.xml.security.signature.XMLSignature;
import org.junit.Assume;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.xml.crypto.dsig.DigestMethod;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Base64;
import java.util.Date;

import static dk.sosi.sts.server.facade.TestClient.*;
import static org.junit.Assert.*;


@RunWith(Parameterized.class)
public abstract class Bst2SosiTest {
	private static final OIOSAMLFactory OIOSAML_FACTORY = new OIOSAMLFactory();
	private static final TestIdentity IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndOneNationalRole(OcesInfo.Version.OCES3);
    private static final String OLE_OLSEN_CPR = "1010903333";
	private static Federation FEDERATION;
	private static boolean isEmbeddedTest;
	private static String issuer;

	@Parameterized.Parameter(0)
	public String suiteName;

	@Parameterized.Parameter(1)
	public OcesInfo.Version version;

	@Parameterized.Parameter(2)
	public CredentialVault mocesVault;

	@Parameterized.Parameter(3)
	public CredentialVault vocesVault;

	protected OIOBSTSAMLAssertion.Type tokenType;
	protected OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilder;
	protected AbstractOIOBSTSAMLAssertionBuilder assertionBuilder;
	protected SamlTokenTestModel testModel;

	public Bst2SosiTest(boolean isEmbeddedTest) {
		this.isEmbeddedTest  = isEmbeddedTest;
		FEDERATION = TestUtil.createTestFederation(isEmbeddedTest);
	}

	/* Setting up test parameters */
	@Parameterized.Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		CredentialVault moces3Vault = CredentialVaultTestUtil.getMoces3CredentialVault();
		CredentialVault voces3Vault = CredentialVaultTestUtil.getVoces3CredentialVault();

		TestParameterBuilder builder = new TestParameterBuilder();

		// SDS-6660 - skal OIOH2BST fortsat testes?
		//builder.newTest(OIOBSTSAMLAssertion.Type.OIOH2BST.name()).add(OcesInfo.Version.OCES2, mocesVault, vocesVault);

		builder.newTest(OIOBSTSAMLAssertion.Type.OIO3BST.name()).add(OcesInfo.Version.OCES3, moces3Vault, voces3Vault);
		builder.newTest(OIOBSTSAMLAssertion.Type.OIOH3BST.name()).add(OcesInfo.Version.OCES3, moces3Vault, voces3Vault);

		return builder.build();
	}

	@Before
	public void setup() {
		this.testModel = new DefaultSamlTokenTestModel(version);
		switch (OIOBSTSAMLAssertion.Type.valueOf(suiteName)) {
			case OIO3BST :
				assertionBuilder = createOIO3BSTSAMLAssertionBuilder(mocesVault, vocesVault);
				requestBuilder = requestBuilderFromAssertionBuilder(mocesVault, assertionBuilder);
				requestBuilder.setEncryptionKey(vocesVault.getSystemCredentialPair().getCertificate().getPublicKey());
				break;
			case OIOH3BST :
				assertionBuilder = createOIOH3BSTSAMLAssertionBuilder(mocesVault, vocesVault);
				requestBuilder = requestBuilderFromAssertionBuilder(mocesVault, assertionBuilder);
				break;
			case OIOH2BST :
				/*
				assertionBuilder = createOIOH2BSTSAMLAssertionBuilder(mocesVault, vocesVault); // TODO SDS-6660
				requestBuilder = requestBuilderFromAssertionBuilder(mocesVault, assertionBuilder);
				requestBuilder.setEncryptionKey(vocesVault.getSystemCredentialPair().getCertificate().getPublicKey());
				 */
				break;
			default :
				throw new RuntimeException("Unrecognized token type");

		}
		tokenType = OIOBSTSAMLAssertion.Type.valueOf(suiteName);
	}

	/**
	 * Testcase #1
	 * <p/>
	 * Kald med udløbet XXXBST afvises med passende fejlbesked.
	 */
	@Test
	public void testExpiredToken() throws Exception {
		assertionBuilder.setNotOnOrAfter(Date.from(Instant.now().minus(2, ChronoUnit.HOURS)));
		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "OIOSAML token no longer valid");
	}

	/**
	 * Testcase #2
	 * <p/>
	 * Kald med ’ugyldig’ XXXBST (fx ’volapyk token’ i sted for SAML) afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidToken() throws Exception {
		requestBuilder.setEncryptionKey(null);
		Document requestDocument = requestBuilder.build();
		Document invalidToken = XmlUtil.createEmptyDocument();
		invalidToken.appendChild(invalidToken.createElement("test"));
		overwriteAssertionInRequest(invalidToken, requestDocument, false);
		signDocument(requestDocument);

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "No assertion element found on request");
	}

	/**
	 * Testcase #3
	 * <p/>
	 * Kald med XXXBST med ugyldig signatur afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidSignature() throws Exception {
		requestBuilder.setEncryptionKey(null);
		Document requestDocument = requestBuilder.build();
		requestDocument.getElementsByTagName("saml:NameID").item(0).setTextContent("-break signature-");

		signDocument(requestDocument);

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Signature on OIOSAMLAssertion is invalid");
	}

	/**
	 * Testcase #4
	 * <p/>
	 * Kald med XXXBST signeret med udløbet certifikat afvises med passende fejlbesked.
	 */
	@Test
	public void testExpiredCertificate() throws Exception {
		assertionBuilder.setSigningVault(SealPPCredentialVaults.getExpiredVocesVault(version)); // expired certificate mangler for oces3

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED, "BST2SOSI Certificate expired");
	}

	/**
	 * Testcase #5
	 * <p/>
	 * Kald med XXXBST signeret med spærret certifikat afvises med passende fejlbesked.
	 */
	@Test
	public void testRevokedCertificate() throws Exception {
		assertionBuilder.setSigningVault(SealPPCredentialVaults.getRevokedVocesVault(version));

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED, "BST2SOSI Certificate revoked");
	}

	/**
	 * Testcase #6
	 * <p/>
	 * Kald med XXXBST signeret med ikke-trustet certifikat afvises med passende fejlbesked.
	 */
	@Test
	public void testUntrustedCertificate() throws Exception {
		// Signing key is not listed in configuration
		assertionBuilder.setSigningVault(testModel.getExternalTrustedIdp());

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "The certificate that signed the security token is not trusted!");
	}

	/**
	 * Testcase #7
	 * <p/>
	 * (Kun for OIO3BST) Kald med OIO3BST som er krypteret til en anden part (dvs. ikke kan dekrypteres)
	 * afvises med passende fejlbesked.
	 */
	@Test
	public void testEncryptionWithUnknownKey() throws Exception {
		Assume.assumeTrue(suiteName.equals("OIO3BST"));
		requestBuilder.setEncryptionKey(testModel.getExternalTrustedIdp().getSystemCredentialPair().getCertificate().getPublicKey());

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "Could not decrypt Assertion");
	}

	/**
	 * Testcase #8
	 * <p/>
	 * Kald med XXXBST som ikke overholder den ud fra konfigurationen forventede tokenprofil
	 * (eksempelvis mangel på påkrævede attributter eller forkert tokenprofil) afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidTokenProfile() throws Exception {
		assertionBuilder.setIssuer(issuer + "_wrong-token-profile");
		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED, "BST2SOSI assertion is not allowed type");
	}

	/**
	 * Testcase #9
	 * <p/>
	 * Kald med XXXBST med for lav AssuranceLevel afvises med passende fejlbesked.
	 */
	@Test
	public void testLowAssuranceLevelNISTForOIO3BST() throws Exception {
		Assume.assumeTrue(tokenType == OIOBSTSAMLAssertion.Type.OIO3BST);
		OIO3BSTSAMLAssertionBuilder assertionBuilder = (OIO3BSTSAMLAssertionBuilder) this.assertionBuilder;

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_1);
		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED,
				"BST2SOSI assertion do not have a valid AssuranceLevel 1. AssuranceLevel 3 or 4 required.");

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_2);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED,
				"BST2SOSI assertion do not have a valid AssuranceLevel 2. AssuranceLevel 3 or 4 required.");

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_3);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_4);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_5);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED,
				"BST2SOSI assertion do not have a valid AssuranceLevel 5. AssuranceLevel 3 or 4 required.");
	}

	/**
	 * Testcase #9
	 */
	@Test
	public void testLowAssuranceLevelNSISForOIO3BST() throws Exception {
		Assume.assumeTrue(tokenType == OIOBSTSAMLAssertion.Type.OIO3BST);
		OIO3BSTSAMLAssertionBuilder assertionBuilder = (OIO3BSTSAMLAssertionBuilder) this.assertionBuilder;

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NSIS.Low);
		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED,
				"BST2SOSI assertion do not have a valid AssuranceLevel Low. AssuranceLevel Substantial or High required.");

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NSIS.High);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NSIS.Substantial);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
	}

	/**
	 * Testcase #9
	 */
	@Test
	public void testLowAssuranceLevelOIOH2BST() throws Exception {
		Assume.assumeTrue(tokenType == OIOBSTSAMLAssertion.Type.OIOH2BST);
		OIOH2BSTSAMLAssertionBuilder assertionBuilder = (OIOH2BSTSAMLAssertionBuilder) this.assertionBuilder;

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_1);
		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED,
				"BST2SOSI assertion do not have a valid AssuranceLevel 1. AssuranceLevel 3 or 4 required.");

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_2);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED,
				"BST2SOSI assertion do not have a valid AssuranceLevel 2. AssuranceLevel 3 or 4 required.");

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_3);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_4);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_5);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED,
				"BST2SOSI assertion do not have a valid AssuranceLevel 5. AssuranceLevel 3 or 4 required.");
	}

	/**
	 * Testcase #9
	 */
	@Test
	public void testLowAssuranceLevelOIOH3BST() throws Exception {
		Assume.assumeTrue(tokenType == OIOBSTSAMLAssertion.Type.OIOH3BST);
		OIOH3BSTSAMLAssertionBuilder assertionBuilder = (OIOH3BSTSAMLAssertionBuilder) this.assertionBuilder;

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NSIS.Low);
		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED,
				"BST2SOSI assertion do not have a valid AssuranceLevel Low. AssuranceLevel Substantial or High required.");

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NSIS.High);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

		assertionBuilder.setAssuranceLevel(AssuranceLevel.NSIS.Substantial);
		response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
	}

	/**
	 * Testcase #10
	 * <p/>
	 * Kald med XXXBST med anden audience end SOSI-STS afvises med passende fejlbesked.
	 */
	@Test
	public void testUnkownAudience() throws Exception {
		assertionBuilder.setAudience("Some audience");

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED, "BST2SOSI assertion does not have allowed audience");
	}

	/**
	 * Testcase #11
	 * <p/>
	 * Kald med XXXBST med ugyldig OIOBPP rolle-struktur afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidOIOBPPRoleStructure() throws Exception {
		String nationalRoles =
				"<bpp:PrivilegeList\n" +
						"        xmlns:bpp=\"http://digst.dk/oiosaml/basic_privilege_profile\"\n" +
						"        xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">\n" +
						"    <PrivilegeGroup Scope=\"urn:dk:gov:saml:cvrNumberIdentifier:20301823\">\n" +
						"        <Privilege>urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3\n" +
						"        </Privilege>\n" +
						"        <InvalidTag>test</InvalidTag>\n" +
						"    </PrivilegeGroup>\n" +
						"</bpp:PrivilegeList>";

		String encodedNationalRoles = Base64.getEncoder().encodeToString(nationalRoles.getBytes());

		OIOBSTSAMLAssertion assertion = assertionBuilder.build();
		Document assertionDocument = assertion.getDOM();
		Element attributeStatement = getFirstChild(assertionDocument.getDocumentElement(), "urn:oasis:names:tc:SAML:2.0:assertion", "AttributeStatement");
		Element attribute = assertionDocument.createElementNS("urn:oasis:names:tc:SAML:2.0:assertion", "saml:Attribute");
		String basicPrivilegesAttributeName = "https://data.gov.dk/model/core/eid/privilegesIntermediate";
		attribute.setAttribute("Name", basicPrivilegesAttributeName);
		attribute.setAttribute("NameFormat", "urn:oasis:names:tc:SAML:2.0:attrname-format:uri");
		Element attributeValue = assertionDocument.createElementNS("urn:oasis:names:tc:SAML:2.0:assertion", "saml:AttributeValue");
		attributeValue.setTextContent(encodedNationalRoles);
		attribute.appendChild(attributeValue);
		attributeStatement.appendChild(attribute);

		signAssertion(assertionDocument);

		Document requestDocument = requestBuilder.build();
		boolean encrypt = assertion.getType() == OIOBSTSAMLAssertion.Type.OIO3BST || assertion.getType() == OIOBSTSAMLAssertion.Type.OIOH2BST;
		overwriteAssertionInRequest(assertionDocument, requestDocument, encrypt);
		signDocument(requestDocument);

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_REQUEST_FAILED, "BST2SOSI assertion contains an invalid basic privilege profile");
	}

	private void signAssertion(Document assertionDocument) {
		Element existingSignature = XmlUtil.getFirstChildElementNS(assertionDocument.getDocumentElement(), NameSpaces.DSIG_SCHEMA, DSTags.SIGNATURE);
		if (existingSignature != null) {
			assertionDocument.getDocumentElement().removeChild(existingSignature);
		}
		String assertionID = assertionDocument.getDocumentElement().getAttribute(SAMLAttributes.ID);
		SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[] { assertionID }, assertionID, IDValues.Id);
		signatureConfiguration.setSignatureMethodAlgorithmUri(XMLSignature.ALGO_ID_SIGNATURE_RSA_SHA256);
		signatureConfiguration.setDigestMethodAlgorithmUri(DigestMethod.SHA256);
		Node subjectNode = assertionDocument.getDocumentElement().getElementsByTagName(SAMLTags.SUBJECT_PREFIXED).item(0);
		signatureConfiguration.setSignatureSiblingNode(subjectNode);
		SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(vocesVault), assertionDocument, signatureConfiguration);
	}

	/**
	 * Testcase #12
	 * <p/>
	 * Kald med ugyldig holder-of-key relation afvises med passende fejlbesked (for en konfiguration der kræver holder-of-key).
	 */
	@Test
	public void testInvalidHolderOfKeyWhenRequired() throws Exception {
		assertionBuilder.setHolderOfKeyCertificate(testModel.getExternalTrustedIdp().getSystemCredentialPair().getCertificate());

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "Signing certificate and holder-of-key certificate relation do not match");
	}

	/**
	 * Testcase #13
	 * <p/>
	 * Kald med gyldig holder-of-key relation men brudt beskeds-signatur afvises med passende fejlbesked
	 * (for en konfiguration der kræver holder-of-key).
	 */
	@Test
	public void testBrokenMessageSignature() throws Exception {
		Document requestDocument = requestBuilder.build();
		Element claims = getFirstChild(requestDocument.getDocumentElement(), "http://docs.oasis-open.org/ws-sx/ws-trust/200512", "Claims");
		Element claimValue = getFirstChild(claims, "http://docs.oasis-open.org/wsfed/authorization/200706", "Value");
		claimValue.getFirstChild().setNodeValue("modified");

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase #14
	 * <p/>
	 * Kald med gyldig holder-of-key relation men manglende signering af påkrævede beskeds-elementer
	 * afvises med passende fejlbesked (for en konfiguration der kræver holder-of-key).
	 */
	@Test
	public void testMissingBodySignature() throws Exception {
		Document requestDocument = DOMBuilderHelper.buildWithoutSigning(requestBuilder);
		new MissingMessageSignatureSigner(mocesVault, requestDocument).skipUri("body").enhanceAndSign();

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase #14
	 */
	@Test
	public void testMissingTsSignature() throws Exception {
		Document requestDocument = DOMBuilderHelper.buildWithoutSigning(requestBuilder);
		new MissingMessageSignatureSigner(mocesVault, requestDocument).skipUri("ts").enhanceAndSign();

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase #14
	 */
	@Test
	public void testMissingMessageIDSignature() throws Exception {
		Document requestDocument = DOMBuilderHelper.buildWithoutSigning(requestBuilder);
		new MissingMessageSignatureSigner(mocesVault, requestDocument).skipUri("messageID").enhanceAndSign();

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase #14
	 */
	@Test
	public void testMissingActionSignature() throws Exception {
		Document requestDocument = DOMBuilderHelper.buildWithoutSigning(requestBuilder);
		new MissingMessageSignatureSigner(mocesVault, requestDocument).skipUri("action").enhanceAndSign();

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Testcase #15
	 * <p/>
	 * Kald med andre WS-Trust claims end de i Krav 2 angivne afvises med passende fejlbesked.
	 */
	@Test
	public void testInvalidClaim() throws Exception {
		Document requestDocument = requestBuilder.build();
		Element body = getFirstChild(requestDocument.getDocumentElement(), "http://schemas.xmlsoap.org/soap/envelope/", "Body");
		Element claims = getFirstChild(body, "http://docs.oasis-open.org/ws-sx/ws-trust/200512", "Claims");
		Element claim = getChildByNameAndAttribute(claims, "http://docs.oasis-open.org/wsfed/authorization/200706", "ClaimType", "Uri", "sosi:SubjectNameID");
		assertNotNull("Unexpected test data", claim);
		claim.setAttribute("Uri", "medcom:InvalidClaim");

		signDocument(requestDocument);

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Request contains invalid claims");
	}

	/**
	 * Testcase #15
	 */
	@Test
	public void testDuplicateClaim() throws Exception {
		Document requestDocument = requestBuilder.build();
		Element body = getFirstChild(requestDocument.getDocumentElement(), "http://schemas.xmlsoap.org/soap/envelope/", "Body");
		Element claims = getFirstChild(body, "http://docs.oasis-open.org/ws-sx/ws-trust/200512", "Claims");
		Element claim = getChildByNameAndAttribute(claims, "http://docs.oasis-open.org/wsfed/authorization/200706", "ClaimType", "Uri", "sosi:SubjectNameID");
		assertNotNull("Unexpected test data", claim);
		claim.setAttribute("Uri", "medcom:ITSystemName");
		signDocument(requestDocument);

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Request contains invalid claims");
	}

	/**
	 * Testcase #16
	 * <p/>
	 * Kald med manglende påkrævet WS-Trust claim (’it-system’, se Krav 2) afvises med passende fejlbesked.
	 */
	@Test
	public void testMissingItSystemClaim() throws Exception {
		Document requestDocument = requestBuilder.build();
		Element body = getFirstChild(requestDocument.getDocumentElement(), "http://schemas.xmlsoap.org/soap/envelope/", "Body");
		Element claims = getFirstChild(body, "http://docs.oasis-open.org/ws-sx/ws-trust/200512", "Claims");
		Element itSystemClaim = getChildByNameAndAttribute(claims, "http://docs.oasis-open.org/wsfed/authorization/200706", "ClaimType", "Uri", "medcom:ITSystemName");
		assertNotNull("Unexpected test data", itSystemClaim);
		claims.removeChild(itSystemClaim);
		signDocument(requestDocument);

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "ITSystemName required in OIOBSTSAMLAssertionToIDCardRequest");
	}

	/**
	 * Testcase #17
	 * <p/>
	 * Kald med ugyldige eller tomme WS-Trust claims afvises med passende fejlbesked.
	 */
	@Test
	public void testEmptyItSystemNameClaim() throws Exception {
		Document requestDocument = requestBuilder.build();
		Element body = getFirstChild(requestDocument.getDocumentElement(), "http://schemas.xmlsoap.org/soap/envelope/", "Body");
		Element claims = getFirstChild(body, "http://docs.oasis-open.org/ws-sx/ws-trust/200512", "Claims");
		Element itSystemClaim = getChildByNameAndAttribute(claims, "http://docs.oasis-open.org/wsfed/authorization/200706", "ClaimType", "Uri", "medcom:ITSystemName");
		assertNotNull("Unexpected test data", itSystemClaim);
		Element claimValue = getFirstChild(itSystemClaim, "http://docs.oasis-open.org/wsfed/authorization/200706", "Value");
		claimValue.getFirstChild().setNodeValue("");
		signDocument(requestDocument);

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "ITSystemName is required in request");
	}

	/**
	 * Testcase #17
	 */
	@Test
	public void testInvalidClaims() throws Exception {
		Document requestDocument = requestBuilder.build();
		Element body = getFirstChild(requestDocument.getDocumentElement(), "http://schemas.xmlsoap.org/soap/envelope/", "Body");
		Element claims = getFirstChild(body, "http://docs.oasis-open.org/ws-sx/ws-trust/200512", "Claims");
		Element claim = getChildByNameAndAttribute(claims, "http://docs.oasis-open.org/wsfed/authorization/200706", "ClaimType", "Uri", "sosi:SubjectNameID");
		assertNotNull("Unexpected test data", claim);
		claim.setAttribute("Uri", "medcom:UserRole");
		signDocument(requestDocument);

		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Only one of UserAuthorizationCode or UserRole is allowed in request");
	}

	/**
	 * Testcase #18
	 * <p/>
	 * Gyldige kald med gyldig XXXBST og gyldig holder-of-key relation og valide claims går godt
	 * (for en konfiguration der kræver holder-of-key).
	 */
	@Test
	public void testValidRequest() throws Exception {
		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
	}

	/**
	 * Testcase #19
	 * <p/>
	 * Gyldige kald med gyldig XXXBST uden gyldig holder-of-key relation og valide claims går godt
	 * (for en konfiguration der ikke kræver holder-of-key).
	 */
	@Test
	public void testInvalidHolderOfKeyWhenNotRequired() throws Exception {
		assertionBuilder.setIssuer(issuer + "_no-hok-val");
		assertionBuilder.setHolderOfKeyCertificate(testModel.getExternalTrustedIdp().getSystemCredentialPair().getCertificate());

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
	}

	@Test
	public void testNonWhitelistedCertificate() throws Exception {
		// Certificate not listed in configuration
		assertionBuilder.setHolderOfKeyCertificate(CredentialVaultTestUtil.getVoces3CredentialVault().getSystemCredentialPair().getCertificate());
		requestBuilder.setSigningVault(CredentialVaultTestUtil.getVoces3CredentialVault());

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Signing certificate not whitelisted for this issuer");
	}

	@Test
	public void testCertificateWhitelistedForAllIssuers() throws Exception {
		// Certificate listed in configuration with '*' issuer
		assertionBuilder.setHolderOfKeyCertificate(testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		requestBuilder.setSigningVault(testModel.getValidWhitelistedVault());

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
	}

	@Test
	public void testValidRequestThorough() {
		MockServicesImpl services = new MockServicesImpl(version);
		SOSIFactory sosiFactory = services.buildSosiFactory(isEmbeddedTest);

		Document requestDocument = requestBuilder.build();
		OIOBSTSAMLAssertionToIDCardRequest request = OIOSAML_FACTORY.createOIOBSTSAMLAssertionToIDCardRequestModelBuilder().build(requestDocument);
		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestDocument);

		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
		assertNull("Forventede ikke fejl info", response.getFaultActor());
		assertNull("Forventede ikke fejl info", response.getFaultCode());
		assertNull("Forventede ikke fejl info", response.getFaultString());
		assertEquals("Uventet action", "http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue", response.getAction());
		assertEquals("Uventet Applies To/Audience Restriction", "https://fmk", response.getAppliesTo());
		assertEquals("Uventet Context uuid", request.getContext(), response.getContext());
		assertNotNull("Forventede Created Date i response", response.getCreated());
		assertNotNull("Forventede Expireds Date i response", response.getExpires());
		assertNotNull("Forventede Message uuid i reponse", response.getMessageID());
		assertEquals("Uventet Relates To", request.getMessageID(), response.getRelatesTo());
		assertEquals("Uventet Token Type", "http://docs.oasis-open.org/wss/oasis-wss-saml-token-profile-1.1#SAMLV2.0", response.getTokenType());
		assertTrue("Forventede UserIDCard", response.getIDCard() instanceof UserIDCard);

		UserIDCard idCard = (UserIDCard) response.getIDCard();
		assertEquals("Uventet Alternativ Identifier/Subject Name ID", "Mads_Skjern", idCard.getAlternativeIdentifier());
		assertEquals("Uventet AuthenticationLevel", 4, idCard.getAuthenticationLevel().getLevel());
		assertNull("Forventede ikke cert hash i ID Card", idCard.getCertHash());
		assertNotNull("Forventede Created Date i ID Card", idCard.getCreatedDate());
		assertNotNull("Forventede Expiry Date i ID Card", idCard.getExpiryDate());
		assertNotNull("Forventede ID Card id i ID Card", idCard.getIDCardID());
		assertNotNull("Forventede issuer i ID Card", idCard.getIssuer());
		assertTrue("Uventet issuer: " + idCard.getIssuer(), idCard.getIssuer().contains("STS")); // Afhænger af test miljø, men indeholder altid STS.
		assertNull("Forventede ikke password i ID Card", idCard.getPassword());
		assertNull("Forventede ikke username i ID Card", idCard.getUsername());
		assertEquals("Uventet version", "1.0.1", idCard.getVersion());
		assertTrue("Forventede ID card var valid", idCard.isValidInTime());
		idCard.validateSignature();
		idCard.validateSignatureAndTrust(sosiFactory.getFederation());

		SystemInfo systemInfo = idCard.getSystemInfo();
		assertEquals("Uventet IT System Name", "Korsbæk Kommunes IT systemer", systemInfo.getITSystemName());
		assertEquals("Uventet Care Provider Id", "20301823", systemInfo.getCareProvider().getID());
		assertEquals("Uventet Care Provider Organization Name", "Korsbæk Kommune", systemInfo.getCareProvider().getOrgName());
		assertEquals("Uventet Care Provider Id Type", "medcom:cvrnumber", systemInfo.getCareProvider().getType());

		UserInfo userInfo2 = idCard.getUserInfo();
		assertEquals("Uventet Authorization Code", IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS.getAuthorizations().get(0).getCode(), userInfo2.getAuthorizationCode());
		assertEquals("Uventet CPR", IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS.getCpr(), userInfo2.getCPR());
		assertNull("Forventede ikke Email i ID card", userInfo2.getEmail());
		assertEquals("Uventet Given Name", "NSTSSenAtre", userInfo2.getGivenName());
		assertNull("Forventede ikke Occupation i ID card", userInfo2.getOccupation());
		assertEquals("Uventet Role", "7170", userInfo2.getRole());
		assertEquals("Uventet Sur Name", "Jensen", userInfo2.getSurName());
	}

	@Test
	public void testInvalidUuid() throws Exception {
		switch (tokenType) {
			case OIO3BST:
				((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setUuid("1234");
				break;
			case OIOH2BST:
				((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setUuid("1234");
				break;
			case OIOH3BST:
				((OIOH3BSTSAMLAssertionBuilder) assertionBuilder).setUuid("1234");
				break;
		}

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "BST2SOSI assertion do not have a valid uuid: 1234");
	}

	@Test
	public void testMissingUuid() throws Exception {
		Assume.assumeTrue(tokenType != OIOBSTSAMLAssertion.Type.OIOH2BST);
		BootstrapTokenBuilder.removeAttribute(assertionBuilder, OIO3BSTSAMLAssertion.Attribute.UUID_PROFESSIONAL);
		BootstrapTokenBuilder.removeAttribute(assertionBuilder, OIOH3BSTSAMLAssertion.Attribute.UUID_PROFESSIONAL);

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST,
				String.format("BST2SOSI assertion must contain the attribute '%s'", OIOSAML3Attributes.PROFESSIONAL_UUID));
	}

	@Ignore("SDS-6660")
	@Test
	public void testRidInsteadOfUuid() throws Exception {
		// Udkommenteret SDS-6660:
		//TestIdentity testIdentity = TestIdentityProvider.getIdentityWithThreeAuthorizationsAndOneNationalRole(OcesInfo.Version.OCES2);
		TestIdentity testIdentity = null;
		Assume.assumeTrue(tokenType == OIOBSTSAMLAssertion.Type.OIOH2BST);
		BootstrapTokenBuilder.removeAttribute(assertionBuilder, OIOH2BSTSAMLAssertion.Attribute.UUID);
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setRid("NSTSSenAtre");
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setCpr(testIdentity.getCpr());
		requestBuilder.setUserAuthorizationCode(testIdentity.getAuthorizations().get(0).getCode());

		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
		OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
	}

    /**
     * SSNCPR-BSTIT1
     */
    @Test
    public void testOioH3BstRequestGlobalUuid() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
        String uuid = "f98cb2b9-30e3-4ef7-a429-fce908a3703b"; // Global uuid for Ole Olsen, 1010903333
        assertionBuilder = createOIOH3BSTSAMLAssertionBuilder(mocesVault, vocesVault, uuid);
        requestBuilder = createRequestBuilder(mocesVault, "108NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        assertExpectedCpr(response, OLE_OLSEN_CPR);
    }

    /**
     * SSNCPR-BSTIT2
     */
    @Test
    public void testOioH3BstRequestCertificateSpecificUuid() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
        String uuid = "1885e837-9606-43ac-bc26-137990bdb884"; // Certificate-specific uuid for  Ole Olsen, 1010903333
        assertionBuilder = createOIOH3BSTSAMLAssertionBuilder(mocesVault, vocesVault, uuid);
        requestBuilder = createRequestBuilder(mocesVault, "108NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.UUIDCPR, "wst:FailedAuthentication",  "illegal uuid");
    }

    /**
     * SSNCPR-BSTIT3
     * The side-effect of not invoking the uuid2cpr-service is not checked here - this is unit tested in UserValidationServiceTest instead.
     */
    @Test
    public void testOioH2BstRequestGlobalUuid_correctCpr() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
        String uuid = "f98cb2b9-30e3-4ef7-a429-fce908a3703b"; // Global uuid for Ole Olsen, 1010903333
        String cpr = OLE_OLSEN_CPR;
        assertionBuilder = createOIOH2BSTSAMLAssertionBuilder(mocesVault, vocesVault);
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setUuid(uuid);
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setCpr(cpr);
        requestBuilder = createRequestBuilder(mocesVault, "108NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

        // Check that the cpr is present
        assertExpectedCpr(response, cpr);
    }

    /**
     * SSNCPR-BSTIT4
     */
    @Test
    public void testOioH2BstRequestGlobalUuid_wrongCpr() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
        String uuid = "f98cb2b9-30e3-4ef7-a429-fce908a3703b"; // Global uuid for Ole Olsen, 1010903333
        String cpr = "0707070707"; // The wrong cpr.
        assertionBuilder = createOIOH2BSTSAMLAssertionBuilder(mocesVault, vocesVault);
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setUuid(uuid);
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setCpr(cpr);
        requestBuilder = createRequestBuilder(mocesVault, "009NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

        // Check that the cpr from the request is preferred
        assertExpectedCpr(response, cpr);
    }

    /**
     * SSNCPR-BSTIT5
     */
    @Test
    public void testOioH2BstRequestGlobalUuid_noCpr() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
        String uuid = "f98cb2b9-30e3-4ef7-a429-fce908a3703b"; // Global uuid for Ole Olsen, 1010903333
        assertionBuilder = createNonValidatingOIOH2BSTSAMLAssertionBuilder(mocesVault, vocesVault);
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setUuid(uuid);
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setCpr(null);
        requestBuilder = createRequestBuilder(mocesVault, "108NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        assertExpectedCpr(response, OLE_OLSEN_CPR);
    }

    /**
     * SSNCPR-BSTIT6
     */
    @Test
    public void testOioH2BstRequest_CertificateSpecificUuid() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
        String uuid = "1885e837-9606-43ac-bc26-137990bdb884"; // Certificate-specific uuid for Ole Olsen, 1010903333
        assertionBuilder = createNonValidatingOIOH2BSTSAMLAssertionBuilder(mocesVault, vocesVault);
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setUuid(uuid);
		((OIOH2BSTSAMLAssertionBuilder) assertionBuilder).setCpr(null);
        requestBuilder = createRequestBuilder(mocesVault, "108NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.UUIDCPR, "wst:FailedAuthentication",  "illegal uuid");
    }

    /**
     * SSNCPR-BSTIT7
     * The side-effect of not invoking the uuid2cpr-service is not checked here - this is unit tested in UserValidationServiceTest instead.
     */
    @Test
    public void testOio3BstRequestGlobalUuid_correctCpr() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
		String uuid = "f98cb2b9-30e3-4ef7-a429-fce908a3703b"; // Global uuid for Ole Olsen, 1010903333
        String cpr = OLE_OLSEN_CPR;
        assertionBuilder = createOIO3BSTSAMLAssertionBuilder(mocesVault, vocesVault);
		((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setUuid(uuid);
		((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setCpr(cpr);
        requestBuilder = createRequestBuilder(mocesVault, "108NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

        // Check that the cpr is present
        assertExpectedCpr(response, cpr);
    }

    /**
     * SSNCPR-BSTIT8
     */
    @Test
    public void testOio3BstRequestGlobalUuid_wrongCpr() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
        String uuid = "f98cb2b9-30e3-4ef7-a429-fce908a3703b"; // Global uuid for Ole Olsen, 1010903333
        String cpr = "0707070707"; // The wrong cpr.
        assertionBuilder = createOIO3BSTSAMLAssertionBuilder(mocesVault, vocesVault);
		((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setUuid(uuid);
		((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setCpr(cpr);
        requestBuilder = createRequestBuilder(mocesVault, "009NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

        // Check that the cpr from the request is preferred
        assertExpectedCpr(response, cpr);
    }

    /**
     * SSNCPR-BSTIT9
     */
    @Test
    public void testOio3BstRequestGlobalUuid_noCpr() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
		String uuid = "f98cb2b9-30e3-4ef7-a429-fce908a3703b"; // Global uuid for Ole Olsen, 1010903333
        assertionBuilder = createOIO3BSTSAMLAssertionBuilder(mocesVault, vocesVault);
		((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setUuid(uuid);
		((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setCpr(null);
        requestBuilder = createRequestBuilder(mocesVault, "108NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        assertExpectedCpr(response, OLE_OLSEN_CPR);
    }

    /**
     * SSNCPR-BSTIT10
     */
    @Test
    public void testOio3BstRequest_CertificateSpecificUuid() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
        String uuid = "1885e837-9606-43ac-bc26-137990bdb884"; // Certificate-specific uuid for Ole Olsen, 1010903333
        assertionBuilder = createOIO3BSTSAMLAssertionBuilder(mocesVault, vocesVault);
		((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setUuid(uuid);
		((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setCpr(null);
        requestBuilder = createRequestBuilder(mocesVault, "108NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.UUIDCPR, "wst:FailedAuthentication",  "illegal uuid");
    }

    @Test
    public void testGlobalUuid_withUuidPrefix() throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVaultFromResource(SignatureUtil.setupCryptoProviderForJVM(), "oces3/MOCES_STSTEST_Ole_Olsen_specific_uuid.p12");
        String uuid = "urn:uuid:f98cb2b9-30e3-4ef7-a429-fce908a3703b"; // Global uuid for Ole Olsen, 1010903333
        assertionBuilder = createOIO3BSTSAMLAssertionBuilder(mocesVault, vocesVault);
        ((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setUuid(uuid);
        ((OIO3BSTSAMLAssertionBuilder) assertionBuilder).setCpr(null);
        requestBuilder = createRequestBuilder(mocesVault, "108NX");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
    }

	@Test
	public void testValidRequest_WithBlacklistedITSystemName_ReturnsFault() throws Exception {
		requestBuilder.setITSystemName(TestUtil.STS_IT_SYSTEM_NAME);
		requestBuilder.setOIOBSTSAMLAssertion(assertionBuilder.build());
		OIOBSTSAMLAssertionToIDCardResponse response = sendRequest(requestBuilder.build());

		OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI,
				StsFaultCodeValues.WST_BAD_REQUEST, "Invalid ITSystemName");
	}


	// SDS-6483: If the SubjectNameId is not set, alternativeIdentifier is set to null.
	@Test
	public void testSubjectIdNotSet_alternativeIdShouldBeNull() {
		// Arrange
		requestBuilder.setSubjectNameID(null);

		// Act
		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();

		// Assert
		IDCard card = response.getIDCard();
		assertNull(card.getAlternativeIdentifier());
	}

	// SDS-6483: If the SubjectNameId is set, use that as alternativeIdentifier.
	@Test
	public void testSubjectIdSet_alternativeIdShouldMatch() {
		// Arrange
		String subjectIdName = "Georg Gearløs";
		requestBuilder.setSubjectNameID(subjectIdName);

		// Act
		OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest();

		// Assert
		IDCard card = response.getIDCard();
		assertEquals(subjectIdName, card.getAlternativeIdentifier());
	}

	private OIOBSTSAMLAssertionToIDCardResponse parseResponse(Document responseDoc) {
		return new OIOSAMLFactory().createOIOBSTSAMLAssertionToIDCardResponseModelBuilder().build(responseDoc);
	}

    private static OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilderFromAssertionBuilder(CredentialVault signingVault, AbstractOIOBSTSAMLAssertionBuilder<?,?> assertionBuilder) {
		OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilder = createRequestBuilder(signingVault, IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS.getAuthorizations().get(0).getCode());
		requestBuilder.setOIOBSTSAMLAssertion(assertionBuilder.build());
		return requestBuilder;
    }

	private static OIOBSTSAMLAssertionToIDCardRequestDOMBuilder createRequestBuilder(CredentialVault signingVault, String userAuthorizationCode) {
		OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilder = OIOSAML_FACTORY.createOIOBSTSAMLAssertionToIDCardRequestDOMBuilder();
		requestBuilder.setAudience("https://fmk");
		requestBuilder.setITSystemName("Korsbæk Kommunes IT systemer");
		requestBuilder.setSigningVault(signingVault);
		// Kun en af UserRole eller AuthorizationCode er tilladt
		requestBuilder.setUserAuthorizationCode(userAuthorizationCode);
		requestBuilder.setSubjectNameID("Mads_Skjern");
		return requestBuilder;
	}

	private OIO3BSTSAMLAssertionBuilder createOIO3BSTSAMLAssertionBuilder(CredentialVault mocesVault, CredentialVault signingVault) {
		OIO3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIO3BSTSAMLAssertionBuilder();
		issuer = "https://oio3bst-issuer.dk";
		builder.setIssuer(issuer);
		builder.setNameId("KorsbaekKommune\\MSK");
		builder.setHolderOfKeyCertificate(mocesVault.getSystemCredentialPair().getCertificate());
		builder.setAudience("https://sts.sosi.dk/");
		builder.setNotOnOrAfter(Date.from(Instant.now().plus(2, ChronoUnit.HOURS)));
		builder.setSigningVault(signingVault);
		builder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_4);
		builder.setUuid(uuidFromIdentity(IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS));
		builder.setCvr("20301823");
		builder.setOrganizationName("Korsbæk Kommune");
		builder.setCpr(IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS.getCpr());
		return builder;
	}

    private OIOH3BSTSAMLAssertionBuilder createOIOH3BSTSAMLAssertionBuilder(CredentialVault mocesVault, CredentialVault signingVault) {
        return createOIOH3BSTSAMLAssertionBuilder(mocesVault, signingVault, uuidFromIdentity(IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS));
    }

    private OIOH3BSTSAMLAssertionBuilder createOIOH3BSTSAMLAssertionBuilder(CredentialVault mocesVault, CredentialVault signingVault, String uuid) {
		OIOH3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH3BSTSAMLAssertionBuilder();
		issuer = "https://oioh3bst-issuer.dk";
		builder.setIssuer(issuer);
		builder.setNameId("KorsbaekKommune\\MSK");
		builder.setHolderOfKeyCertificate(mocesVault.getSystemCredentialPair().getCertificate());
		builder.setAudience("https://sts.sosi.dk/");
		builder.setNotOnOrAfter(Date.from(Instant.now().plus(2, ChronoUnit.HOURS)));
		builder.setSigningVault(signingVault);
		builder.setAssuranceLevel(AssuranceLevel.NSIS.High);
		builder.setUuid(uuid);
		builder.setCvr("20301823");
		builder.setOrganizationName("Korsbæk Kommune");
		return builder;
	}

	private OIOH2BSTSAMLAssertionBuilder createOIOH2BSTSAMLAssertionBuilder(CredentialVault mocesVault, CredentialVault signingVault) {
		OIOH2BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH2BSTSAMLAssertionBuilder();
		issuer = "https://oioh2bst-issuer.dk";
		builder.setIssuer(issuer);
		builder.setNameId("KorsbaekKommune\\MSK");
		builder.setHolderOfKeyCertificate(mocesVault.getSystemCredentialPair().getCertificate());
		builder.setAudience("https://sts.sosi.dk/");
		builder.setNotOnOrAfter(Date.from(Instant.now().plus(2, ChronoUnit.HOURS)));
		builder.setSigningVault(signingVault);
		builder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_4);
		builder.setUuid(uuidFromIdentity(IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS));
		builder.setCvr("20301823");
		builder.setOrganizationName("Korsbæk Kommune");
		builder.setCpr(IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS.getCpr());
		return builder;
	}

    private OIOH2BSTSAMLAssertionBuilder createNonValidatingOIOH2BSTSAMLAssertionBuilder(CredentialVault mocesVault, CredentialVault signingVault) {
        OIOH2BSTSAMLAssertionBuilder builder = new OIOH2BSTSAMLAssertionBuilder() {
            @Override
            public void validateBeforeBuild() {
                // no-op - do not validate the BST before sending it. This allows for omitting the cpr-attribute, which is otherwise required.
            }
        };
        issuer = "https://oioh2bst-issuer.dk";
        builder.setIssuer(issuer);
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(mocesVault.getSystemCredentialPair().getCertificate());
        builder.setAudience("https://sts.sosi.dk/");
        builder.setNotOnOrAfter(Date.from(Instant.now().plus(2, ChronoUnit.HOURS)));
        builder.setSigningVault(signingVault);
        builder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_4);
        builder.setUuid(uuidFromIdentity(IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS));
        builder.setCvr("20301823");
        builder.setOrganizationName("Korsbæk Kommune");
        builder.setCpr(IDENTITY_WITH_1_NATIONAL_ROLE_3_AUTHORIZATIONS.getCpr());
        return builder;
    }

	private Response getResponse(Node requestDocument) {
		return getTestClient().execute(XmlUtil.node2String(requestDocument, false, false));
	}

	private OIOBSTSAMLAssertionToIDCardResponse buildAndSendRequest() {
		requestBuilder.setOIOBSTSAMLAssertion(assertionBuilder.build());
		return sendRequest(requestBuilder.build());
	}

	private OIOBSTSAMLAssertionToIDCardResponse sendRequest(Document requestDocument) {
		return parseResponse(getTestClient().execute(XmlUtil.node2String(requestDocument, false, false)).document);
	}

	private static Element getChildren(Element ancestor, Tag tag) {
		return getFirstChild(ancestor, tag.getNS(), tag.getTagName());
	}

	private static Element getFirstChild(Element ancestor, String tagNS, String tagName) {
		return (Element) getChildren(ancestor, tagNS, tagName).item(0);
	}

	private static NodeList getChildren(Element ancestor, String tagNS, String tagName) {
		return tagNS != null ? ancestor.getElementsByTagNameNS(tagNS, tagName) : ancestor.getElementsByTagName(tagName);
	}

	private static Element getChildByNameAndAttribute(Element ancestor, String tagNS, String tagName, String attributeKey, String attributeValue) {
		NodeList children = getChildren(ancestor, tagNS, tagName);
		for (int i = 0; i < children.getLength(); i++) {
			Element child = (Element) children.item(i);
			if (attributeValue.equals(child.getAttribute(attributeKey))) {
				return child;
			}
		}
		return null;
	}

	private void signDocument(Document requestDocument) {
		new LibertyMessageDOMEnhancer(mocesVault, SignatureConfiguration.DEFAULT_KEY_INFO_TYPE, requestDocument, false).enhanceAndSign();
	}

	private void overwriteAssertionInRequest(Document assertionDocument, Document requestDocument, boolean encrypt) {
		Element actAs = getFirstChild(requestDocument.getDocumentElement(), "http://docs.oasis-open.org/ws-sx/ws-trust/200802", "ActAs");
		actAs.removeChild(actAs.getFirstChild());
		if (encrypt) {
			Element assertionElm = assertionDocument.getDocumentElement();
			Element encryptedAssertionElm = assertionDocument.createElementNS(NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.ENCRYPTED_ASSERTION_PREFIXED);
			encryptedAssertionElm.appendChild(assertionElm);
			EncryptionUtil.encrypt(assertionElm, vocesVault.getSystemCredentialPair().getCertificate().getPublicKey());
			actAs.appendChild(requestDocument.importNode(encryptedAssertionElm, true));
		} else {
			actAs.appendChild(requestDocument.importNode(assertionDocument.getDocumentElement(), true));
		}
	}

    private void assertExpectedCpr(OIOBSTSAMLAssertionToIDCardResponse response, String cpr) {
        assertTrue(response.getIDCard() instanceof UserIDCard);
        assertEquals(cpr, ((UserIDCard) response.getIDCard()).getUserInfo().getCPR());
    }

	private String uuidFromIdentity(TestIdentity identity) {
		return StringUtils.replace(identity.getSubjectSerialNumber(), "UI:DK-E:G:", "");
	}

	protected abstract TestClient getTestClient();
}