package dk.sosi.sts.server.facade.bst2sosi;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.BasicPrivileges.Profile;
import dk.sosi.seal.model.dombuilders.OIOBSTSAMLAssertionToIDCardRequestDOMBuilder;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.model.Authorization;
import dk.sosi.sts.model.RoleDefinition;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.facade.OIOWSTrustResponseValidation;
import dk.sosi.sts.server.facade.RoleDefinitionServiceStub;
import dk.sosi.sts.server.facade.TestClient;
import dk.sosi.sts.server.test.TestIdentity;
import dk.sosi.sts.server.test.TestIdentityProvider;
import dk.sosi.sts.server.test.TestUtil;
import org.apache.commons.lang3.StringUtils;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public abstract class Bst2SosiUserRoleTest {
    private static final OIOSAMLFactory OIOSAML_FACTORY = new OIOSAMLFactory();
    private static final TestIdentity IDENTITY_WITH_0_NATIONAL_ROLES_3_AUTHORIZATIONS = TestIdentityProvider.getIdentityWithTrustedCvrAndThreeAuthorizations(OcesInfo.Version.OCES3);
    private static final TestIdentity IDENTITY_WITH_0_NATIONAL_ROLES_1_AUTHORIZATION = TestIdentityProvider.getIdentityWithOneAuthorization(OcesInfo.Version.OCES3);
    private static final TestIdentity IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(OcesInfo.Version.OCES3);
    private static final TestIdentity IDENTITY_WITH_3_NATIONAL_ROLES_3_NONUNIQUE_AUTHORIZATIONS = TestIdentityProvider.getIdentityWithThreeAuthorizationsWithTwoMatchingUddCodesAndThreeNationalRoles(OcesInfo.Version.OCES3);
    private static final TestIdentity IDENTITY_WITH_1_NATIONAL_ROLE_0_AUTHORIZATIONS = TestIdentityProvider.getIdentityWithOneNationalRole(OcesInfo.Version.OCES3);
    private static final TestIdentity IDENTITY_WITH_3_NATIONAL_ROLE_0_AUTHORIZATIONS = TestIdentityProvider.getIdentityWithThreeNationalRoles(OcesInfo.Version.OCES3);
    private static final TestIdentity IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATION_BUT_1_ROLE_FOR_CVRRID = TestIdentityProvider.getMoces3IdentityWithNoAuthorizationAndOneNationalRoleForMappedSsn();

    private static Federation FEDERATION;
    private GenericCredentialVault mocesVault;
    private GenericCredentialVault vocesVault;
    private OIOH2BSTSAMLAssertionBuilder assertionBuilderOIOH2Bst;
    private OIOH3BSTSAMLAssertionBuilder assertionBuilderOIOH3Bst;
    private OIO3BSTSAMLAssertionBuilder assertionBuilderOIO3Bst;
    private OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilder;
    private TestClient testClient;

    public Bst2SosiUserRoleTest(boolean isEmbeddedTest) {
        FEDERATION = TestUtil.createTestFederation(isEmbeddedTest);
    }

    @Before
    public void setup() {
        mocesVault = CredentialVaultTestUtil.getMoces3CredentialVault();
        vocesVault = CredentialVaultTestUtil.getVoces3CredentialVault();
        assertionBuilderOIOH2Bst = null;
        assertionBuilderOIOH3Bst = null;
        assertionBuilderOIO3Bst = null;
        requestBuilder = null;
        testClient = null;
    }

    private void setupOIOH2BstTest() {
        assertionBuilderOIOH2Bst = createOIOH2BSTSAMLAssertionBuilder(mocesVault, vocesVault);
        requestBuilder = requestBuilderFromAssertionBuilder(mocesVault, assertionBuilderOIOH2Bst);
        requestBuilder.setEncryptionKey(vocesVault.getSystemCredentialPair().getCertificate().getPublicKey());
        testClient = getOces3TestClient();
    }

    private void setupOIOH3BstTest() {
        assertionBuilderOIOH3Bst = createOIOH3BSTSAMLAssertionBuilder(mocesVault, vocesVault);
        requestBuilder = requestBuilderFromAssertionBuilder(mocesVault, assertionBuilderOIOH3Bst);
        testClient = getOces3TestClient();
    }

    private void setupOIO3BstTest() {
        assertionBuilderOIO3Bst = createOIO3BSTSAMLAssertionBuilder(mocesVault, vocesVault);
        requestBuilder = requestBuilderFromAssertionBuilder(mocesVault, assertionBuilderOIO3Bst);
        requestBuilder.setEncryptionKey(vocesVault.getSystemCredentialPair().getCertificate().getPublicKey());
        testClient = getOces3TestClient();
    }

    /**
     * Testcase #1
     * <p/>
     * Tjekker at der returneres en fejl, hvis der opsættes (via claim) en ugyldig national rolle i userRole
     * (dvs. national rolle er ikke på NSP liste over gyldige roller)
     *
     * NSP liste -> RoleDefinitionService
     *
     */
    @Test
    public void testInvalidNationalUserRole() throws Exception {
        setupOIOH2BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00003", "role3");
        addNationalRoleToRequest(requestBuilder, "00001", "role1");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Authentication failed: invalid role supplied");
    }

    /**
     * Testcase #2
     * <p/>
     * Tjekker at der kan ønskes (via claim) en national rolle fra de roller der er medsendt i BST
     */
    @Test
    public void testValidNationalUserRoleFromBst() throws Exception {
        setupOIOH2BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "99991", "TestRolle1");
        addNationalRoleToRequest(requestBuilder, "99991", "TestRolle1");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:national-federation-role:code:99991:value:TestRolle1", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #2
     * <p/>
     * Tjekker at der kan ønskes (via claim) en national rolle fra de roller der er medsendt i BST.
     * Nationalrolle med uden kode og claim med kode.
     */
    @Test
    public void testValidNationalUserRoleFromBstNoCodeInBst() {
        setupOIOH2BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "TestRolle1");
        addNationalRoleToRequest(requestBuilder, "99991", "TestRolle1");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:national-federation-role:code:99991:value:TestRolle1", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #2
     * <p/>
     * Tjekker at der kan ønskes (via claim) en national rolle fra de roller der er medsendt i BST
     * Nationalrolle med kode og claim uden kode.
     */
    @Test
    public void testValidNationalUserRoleFromBstNoCodeClaim() {
        setupOIOH2BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "99991", "TestRolle1");
        addNationalRoleToRequest(requestBuilder, "TestRolle1");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:national-federation-role:code:99991:value:TestRolle1", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #2
     * <p/>
     * Tjekker at der kan ønskes (via claim) en national rolle fra de roller der er medsendt i BST
     * Nationalrolle uden kode og claim uden kode.
     */
    @Test
    public void testValidNationalUserRoleFromBstNoCodeBstAndClaim() {
        setupOIOH2BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "TestRolle1");
        addNationalRoleToRequest(requestBuilder, "TestRolle1");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:national-federation-role:code:99991:value:TestRolle1", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #3
     * <p/>
     * Tjekker at der returneres en fejl, hvis ønsket (via claim) er forkert i henhold til udfaldsrum. Her er ønsket en national rolle
     *
     * Udfaldsrum -> NationalRolesService
     */
    @Test
    public void testInvalidNationalUserRole2() throws Exception {
        setupOIOH2BstTest();

        assertionBuilderOIOH2Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00003", "role3");
        String role = addNationalRoleToRequest(requestBuilder, "99991", "TestRolle1");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                "Authentication failed: user does not have the requested role (" + role + ")");
    }

    /**
     * Testcase #4
     * <p/>
     * Tjekker at der udstedes IdKort med ønsket rolle, hvis ønske (via claim)  er rigtig i henhold til udfaldsrum. Her er ønsket en national rolle
     */
    @Test
    public void testValidNationalUserRole() throws Exception {
        setupOIO3BstTest();

        assertionBuilderOIO3Bst.setCpr(IDENTITY_WITH_3_NATIONAL_ROLE_0_AUTHORIZATIONS.getCpr());
        assertionBuilderOIO3Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_3_NATIONAL_ROLE_0_AUTHORIZATIONS));
        RoleDefinition role = IDENTITY_WITH_3_NATIONAL_ROLE_0_AUTHORIZATIONS.getNationalRoles().get(0);
        addNationalRoleToRequest(requestBuilder, role.getRoleId(), role.getRoleName());

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIO3Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", role.getPublicName(), ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #5
     * <p/>
     * Tjekker at der udstedes IdKort med ønsket rolle, hvis ønske er rigtigt i henhold til udfaldsrum. Her er ønsket (via claim) en autorisation
     */
    @Test
    public void testValidUserAuthentication() throws Exception {
        TestIdentity testIdentity = IDENTITY_WITH_0_NATIONAL_ROLES_3_AUTHORIZATIONS;
        String authorizationCode = testIdentity.getAuthorizations().get(0).getCode();
        setupOIOH3BstTest();
        assertionBuilderOIOH3Bst.setUuid(uuidFromIdentity(testIdentity));

        requestBuilder.setUserAuthorizationCode(authorizationCode);

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

        assertTrue("Forventede UserIDCard", response.getIDCard() instanceof UserIDCard);
        UserIDCard idCard = (UserIDCard) response.getIDCard();
        assertEquals("Uventet Alternativ Identifier/Subject Name ID", "Mads_Skjern", idCard.getAlternativeIdentifier());
        assertEquals(authorizationCode, idCard.getUserInfo().getAuthorizationCode());
    }

    /**
     * Testcase #6
     * <p/>
     * Tjekker at der returneres en fejl, hvis der ønskes (via claim) et autorisationsId. Men medarbejderen ingen autorisationer har
     */
    @Test
    public void testUserAuthenticationWithNoAuthentication() throws Exception {
        setupOIOH3BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "00003", "role3");
        requestBuilder.setUserAuthorizationCode("ZXCVB"); // Findes i AuthorizationServiceStub for cpr 1802602810

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "missing authorization");
    }

    /**
     * Testcase #7
     * <p/>
     * Tjekker at der returneres en fejl, hvis ønsket er forkert i henhold til udfaldsrum. Her er ønsket en autorisation
     */
    @Test
    public void testUserAuthenticationWithNoMatchingAuthentication() throws Exception {
        TestIdentity testIdentity = IDENTITY_WITH_0_NATIONAL_ROLES_3_AUTHORIZATIONS;
        setupOIOH2BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00003", "role3");
        assertionBuilderOIOH2Bst.setUuid(uuidFromIdentity(testIdentity));
        assertionBuilderOIOH2Bst.setCpr(testIdentity.getCpr());
        requestBuilder.setUserAuthorizationCode("test");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                "authorization not found (available ", testIdentity.getAuthorizations().get(0).getCode(), testIdentity.getAuthorizations().get(1).getCode(), testIdentity.getAuthorizations().get(2).getCode());
    }

    /**
     * Testcase #8
     * <p/>
     * Tjekker at autorisation kan ønskes via opsætning (via claim) af uddannelseskode
     */
    @Test
    public void testValidUserAuthenticationAsEducationCode() throws Exception {
        TestIdentity testIdentity = IDENTITY_WITH_0_NATIONAL_ROLES_3_AUTHORIZATIONS;
        Authorization authorization = testIdentity.getAuthorizations().get(0);
        setupOIOH2BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00003", "role3");
        assertionBuilderOIOH2Bst.setUuid(uuidFromIdentity(testIdentity));
        assertionBuilderOIOH2Bst.setCpr(testIdentity.getCpr());
        requestBuilder.setUserRole(authorization.getRole());

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);

        assertTrue("Forventede UserIDCard", response.getIDCard() instanceof UserIDCard);
        UserIDCard idCard = (UserIDCard) response.getIDCard();
        assertEquals("Uventet Alternativ Identifier/Subject Name ID", "Mads_Skjern", idCard.getAlternativeIdentifier());
        assertEquals(authorization.getCode(), idCard.getUserInfo().getAuthorizationCode());
    }

    /**
     * Testcase #9
     * <p/>
     * Tjekker at der returneres fejl, hvis der opsættes en uddannelseskode (via claim), som matcher med flere autorisationer i autorisationsregisteret
     */
    @Test
    public void testEducationCodeMatchingMultipleAuthorizations() throws Exception {
        setupOIOH2BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00003", "role3");
        assertionBuilderOIOH2Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_3_NATIONAL_ROLES_3_NONUNIQUE_AUTHORIZATIONS));
        assertionBuilderOIOH2Bst.setCpr(IDENTITY_WITH_3_NATIONAL_ROLES_3_NONUNIQUE_AUTHORIZATIONS.getCpr());
        requestBuilder.setUserRole("7170");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        List<Authorization> authorizations = IDENTITY_WITH_3_NATIONAL_ROLES_3_NONUNIQUE_AUTHORIZATIONS.getAuthorizations();
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                "multiple authorizations found", authorizations.get(0).getCode(), authorizations.get(1).getCode());
    }

    /**
     * Testcase #10
     * <p/>
     * Tjekker at der returneres fejl, hvis der opsættes en uddannelseskode i userRole, men medarbejder har igen autorisationer
     */
    @Test
    public void testEducationCodeWhenNoAuthorizations() throws Exception {
        setupOIOH3BstTest();

        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "00003", "role3");
        requestBuilder.setUserRole("7170");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "missing authorization");
    }

    /**
     * Testcase #11
     * <p/>
     * Tjekker at der returneres fejl, hvis der opsættes en uddannelseskode i userRole, som ikke matcher en autorisation i autorisationsregisteret
     */
    @Test
    public void testEducationCodeNotMatchingAuthorizations() throws Exception {
        TestIdentity testIdentity = IDENTITY_WITH_0_NATIONAL_ROLES_3_AUTHORIZATIONS;
        setupOIOH2BstTest();

        assertionBuilderOIOH2Bst.setUuid(uuidFromIdentity(testIdentity));
        assertionBuilderOIOH2Bst.setCpr(testIdentity.getCpr());
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00001", "role1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00002", "role2");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "00003", "role3");
        requestBuilder.setUserRole("9999");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        List<Authorization> authorizations = testIdentity.getAuthorizations();
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                "multiple authorizations found", authorizations.get(0).getCode(), authorizations.get(1).getCode(), authorizations.get(2).getCode());
    }

    /**
     * Testcase #12
     * <p/>
     * Tjekker at autorisation prioriteres over nationale roller. Dette gælder også når ansat har mange autorisationer og én natinoal rolle. Der skal her udstedes en fejlbesked med mulige autorisationer
     */
    @Test
    public void testWithNeitherUserRoleOrAuthorizationClaimForUserWithMultipleAuthorizations() throws Exception {
        TestIdentity testIdentity = IDENTITY_WITH_0_NATIONAL_ROLES_3_AUTHORIZATIONS;
        setupOIOH2BstTest();

        assertionBuilderOIOH2Bst.setUuid(uuidFromIdentity(testIdentity));
        assertionBuilderOIOH2Bst.setCpr(testIdentity.getCpr());
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "99991", "TestRolle1");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        List<Authorization> authorizations = testIdentity.getAuthorizations();
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                "multiple authorizations found", authorizations.get(0).getCode(), authorizations.get(1).getCode(), authorizations.get(2).getCode());
    }

    /**
     * Testcase #13
     * <p/>
     * Tjekker der udstedes et IdKort med autorisation, hvis der ikke ønskes noget og bruger kun har én autorisation i henhold til autorisationsregister
     */
    @Test
    public void testWithNeitherUserRoleOrAuthorizationClaimForUserWithOneAuthorization() throws Exception {
        TestIdentity testIdentity = IDENTITY_WITH_0_NATIONAL_ROLES_1_AUTHORIZATION;
        setupOIOH3BstTest();
        assertionBuilderOIOH3Bst.setUuid(uuidFromIdentity(testIdentity));

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", testIdentity.getAuthorizations().get(0).getCode(), ((UserIDCard) idCard).getUserInfo().getAuthorizationCode());
    }

    /**
     * Testcase #14
     * <p/>
     * Tjekker at der udstedes en struktureret fejlkode med mulige roller, hvis der ikke ønskes noget og medarbejder har  flere nationale roller som overføres vis BST token
     */
    @Test
    public void testWithNeitherUserRoleOrAuthorizationClaimForUserWithMultipleNationalRolesInBSTToken() throws Exception {
        setupOIOH3BstTest();

        assertionBuilderOIOH3Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "41003", "PlejeAssR3");
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "99991", "TestRolle1");
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "99992", "TestRolle2");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION,
                "multiple national roles found ([" +
                        "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3, " +
                        "urn:dk:healthcare:national-federation-role:code:99991:value:TestRolle1, " +
                        "urn:dk:healthcare:national-federation-role:code:99992:value:TestRolle2])");
    }

    /**
     * Testcase #15
     * <p/>
     * Tjekker at der udstedes et IdKort med national rolle, hvis der ikke ønskes noget,  samt medarbejder har tomt autorisationregister og en  national rolle som overføres vis BST token
     */
    @Test
    public void testWithNeitherUserRoleOrAuthorizationClaimForUserWith1NationalRoleInBSTToken() throws Exception {
        setupOIOH3BstTest();

        assertionBuilderOIOH3Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "99991", "TestRolle1");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:national-federation-role:code:99991:value:TestRolle1", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #16
     * <p/>
     * Tjekker at der udstedes en struktureret fejlkode med mulige roller, hvis der ikke ønskes noget og medarbejder har  flere nationale roller i SEB
     */
    @Test
    public void testWithNeitherUserRoleOrAuthorizationClaimForUserWithMultipleNationalRolesInSEB() throws Exception {
        setupOIO3BstTest();

        assertionBuilderOIO3Bst.setCpr(IDENTITY_WITH_3_NATIONAL_ROLE_0_AUTHORIZATIONS.getCpr());
        assertionBuilderOIO3Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_3_NATIONAL_ROLE_0_AUTHORIZATIONS));

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIO3Bst);
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.AUTHORIZATION, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "multiple national roles found");
        assertTrue("Unexpected fault string: " + response.getFaultString(), response.getFaultString().contains(IDENTITY_WITH_3_NATIONAL_ROLE_0_AUTHORIZATIONS.getNationalRoles().get(0).getPublicName()));
        assertTrue("Unexpected fault string: " + response.getFaultString(), response.getFaultString().contains(IDENTITY_WITH_3_NATIONAL_ROLE_0_AUTHORIZATIONS.getNationalRoles().get(1).getPublicName()));
        assertTrue("Unexpected fault string: " + response.getFaultString(), response.getFaultString().contains(IDENTITY_WITH_3_NATIONAL_ROLE_0_AUTHORIZATIONS.getNationalRoles().get(2).getPublicName()));
    }

    /**
     * Testcase #17
     * <p/>
     * Tjekker at der udstedes et IdKort med national rolle, hvis der ikke ønskes noget,  samt medarbejder har tomt autorisationregister og en rolle i SEB
     */
    @Test
    public void testWithNeitherUserRoleOrAuthorizationClaimForUserWith1NationalRoleInSEB() throws Exception {
        setupOIO3BstTest();

        assertionBuilderOIO3Bst.setCpr(IDENTITY_WITH_1_NATIONAL_ROLE_0_AUTHORIZATIONS.getCpr());
        assertionBuilderOIO3Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_1_NATIONAL_ROLE_0_AUTHORIZATIONS));

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIO3Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #18
     * <p/>
     * Tjekker at der udstedes et IdKort med rollen "urn:dk:healthcare:no-role", hvis det angives, at der ikke ønskes en autorisation eller national rolle
     * ("urn:dk:healthcare:no-role" i userRole) og bruger af flere autorisationer i autorisationsreg. og  1 national rolle som overføres vis BST token
     */
    @Test
    public void testWithNoAuthorizationWantedForUserWithMultipleAuthorizationsAnd1NationalRoleInBst() throws Exception {
        TestIdentity testIdentity = IDENTITY_WITH_0_NATIONAL_ROLES_3_AUTHORIZATIONS;
        setupOIOH3BstTest();

        assertionBuilderOIOH3Bst.setUuid(uuidFromIdentity(testIdentity));
        addNationalRoleToBst(assertionBuilderOIOH3Bst, new BasicPrivilegesDOMBuilder(Profile.V12), "99992", "TestRolle2");
        requestBuilder.setUserRole("urn:dk:healthcare:no-role");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:no-role", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #19
     * <p/>
     * Tjekker at der udstedes et IdKort uden indhold i autorisation og userRole, hvis der ikke ønskes autorisation eller national rolle,
     * samt medarbejder har tomt autorisationregister og SEB
     */
    @Test
    public void testWithNoAuthorizationOrRoleWantedForUserWithNoAuthorizationsOrNationalRoles() throws Exception {
        setupOIOH3BstTest();

        assertionBuilderOIOH3Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:unknown", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #19
     */
    @Test
    public void testWithEmptyRoleForUserWithNoAuthorizationsOrNationalRoles() throws Exception {
        setupOIOH3BstTest();

        assertionBuilderOIOH3Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));
        requestBuilder.setUserRole(" ");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:unknown", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #19
     */
    @Test
    public void testWithUnknownRoleForUserWithNoAuthorizationsOrNationalRoles() throws Exception {
        setupOIOH3BstTest();

        assertionBuilderOIOH3Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));
        requestBuilder.setUserRole("Hest");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "Hest", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Testcase #20
     * <p/>
     * Tjekker at claims af flere autorisationer/roller udløser fejl
     */
    @Test
    public void testAuthorizationClaimAndUserRoleClaim() throws Exception {
        TestIdentity testIdentity = IDENTITY_WITH_0_NATIONAL_ROLES_3_AUTHORIZATIONS;
        setupOIOH2BstTest();

        assertionBuilderOIOH2Bst.setCpr(testIdentity.getCpr());
        assertionBuilderOIOH2Bst.setUuid(uuidFromIdentity(testIdentity));
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "41003", "PlejeAssR3");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "99991", "TestRolle1");
        addNationalRoleToBst(assertionBuilderOIOH2Bst, privilegesBuilder, "99992", "TestRolle2");
        addNationalRoleToRequest(requestBuilder, "41003", "PlejeAssR3");
        requestBuilder.setUserAuthorizationCode(testIdentity.getAuthorizations().get(0).getCode());

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH2Bst);
        OIOWSTrustResponseValidation.assertFault(response, StsFaultActorValues.BST2SOSI, StsFaultCodeValues.WST_INVALID_REQUEST, "Only one of UserAuthorizationCode or UserRole is allowed in request");
    }

    /**
     * Testcase #21
     * <p/>
     * Tjekker at udfaldsrums algoritme fjerner ugyldige nationale rolle overført via BST
     */
    @Test
    public void testInvalidNationalRolesInBst() throws Exception {
        setupOIOH3BstTest();

        assertionBuilderOIOH3Bst.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "41003", "PlejeAssR3");
        addNationalRoleToBst(assertionBuilderOIOH3Bst, privilegesBuilder, "00000", "UgyldigRolle");

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIOH3Bst);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3", ((UserIDCard) idCard).getUserInfo().getRole());
    }

    /**
     * Tjekker at der udstedes IdKort med ønsket rolle, hvis rollen ikke findes med brugerens uuid, men med cvr-rid mappet fra uuid'et (SDS-6026)
     */
    @Test
    public void testOIO3BSTValidNationalUserRoleByCvrRid() throws Exception {
        TestIdentity testIdentity = IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATION_BUT_1_ROLE_FOR_CVRRID;
        setupOIO3BstTest();
        assertionBuilderOIO3Bst.setCpr(testIdentity.getCpr());
        assertionBuilderOIO3Bst.setUuid(uuidFromIdentity(testIdentity));
        RoleDefinition role = RoleDefinitionServiceStub.PORTOER;
        addNationalRoleToRequest(requestBuilder, role.getRoleId(), role.getRoleName());

        OIOBSTSAMLAssertionToIDCardResponse response = buildAndSendRequest(testClient, assertionBuilderOIO3Bst);

        OIOWSTrustResponseValidation.assertSuccess(response, FEDERATION);
        IDCard idCard = response.getIDCard();
        assertTrue("Unexpected id card", idCard instanceof UserIDCard);
        assertEquals("Unexpected user role", role.getPublicName(), ((UserIDCard) idCard).getUserInfo().getRole());
    }

    private OIOBSTSAMLAssertionToIDCardResponse parseResponse(Document responseDoc) {
        return new OIOSAMLFactory().createOIOBSTSAMLAssertionToIDCardResponseModelBuilder().build(responseDoc);
    }

    private static OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilderFromAssertionBuilder(CredentialVault signingVault, AbstractOIOBSTSAMLAssertionBuilder assertionBuilder) {
        OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilder = OIOSAML_FACTORY.createOIOBSTSAMLAssertionToIDCardRequestDOMBuilder();
        requestBuilder.setOIOBSTSAMLAssertion(assertionBuilder.build());
        requestBuilder.setAudience("https://fmk");
        requestBuilder.setITSystemName("Korsbæk Kommunes IT systemer");
        requestBuilder.setSigningVault(signingVault);
        requestBuilder.setSubjectNameID("Mads_Skjern");
        return requestBuilder;
    }

    private static OIO3BSTSAMLAssertionBuilder createOIO3BSTSAMLAssertionBuilder(CredentialVault mocesVault, CredentialVault signingVault) {
        OIO3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIO3BSTSAMLAssertionBuilder();
        builder.setIssuer("https://oio3bst-issuer.dk");
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(mocesVault.getSystemCredentialPair().getCertificate());
        builder.setAudience("https://sts.sosi.dk/");
        builder.setNotOnOrAfter(Date.from(Instant.now().plus(2, ChronoUnit.HOURS)));
        builder.setSigningVault(signingVault);
        builder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_4);
        builder.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));
        builder.setCvr("25450442");
        builder.setOrganizationName("Korsbæk Kommune");
        builder.setCpr(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS.getCpr());
        return builder;
    }

    private static OIOH3BSTSAMLAssertionBuilder createOIOH3BSTSAMLAssertionBuilder(CredentialVault mocesVault, CredentialVault signingVault) {
        OIOH3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH3BSTSAMLAssertionBuilder();
        builder.setIssuer("https://oioh3bst-issuer.dk");
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(mocesVault.getSystemCredentialPair().getCertificate());
        builder.setAudience("https://sts.sosi.dk/");
        builder.setNotOnOrAfter(Date.from(Instant.now().plus(2, ChronoUnit.HOURS)));
        builder.setSigningVault(signingVault);
        builder.setAssuranceLevel(AssuranceLevel.NSIS.High);
        builder.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));
        builder.setCvr("25450442");
        builder.setOrganizationName("Korsbæk Kommune");
        return builder;
    }

    private static OIOH2BSTSAMLAssertionBuilder createOIOH2BSTSAMLAssertionBuilder(CredentialVault mocesVault, CredentialVault signingVault) {
        OIOH2BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH2BSTSAMLAssertionBuilder();
        builder.setIssuer("https://oioh2bst-issuer.dk");
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(mocesVault.getSystemCredentialPair().getCertificate());
        builder.setAudience("https://sts.sosi.dk/");
        builder.setNotOnOrAfter(Date.from(Instant.now().plus(2, ChronoUnit.HOURS)));
        builder.setSigningVault(signingVault);
        builder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_4);
        builder.setUuid(uuidFromIdentity(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS));
        builder.setCvr("25450442");
        builder.setOrganizationName("Korsbæk Kommune");
        builder.setCpr(IDENTITY_WITH_0_NATIONAL_ROLES_0_AUTHORIZATIONS.getCpr());
        return builder;
    }

    private void addNationalRoleToBst(OIOH2BSTSAMLAssertionBuilder assertionBuilder, BasicPrivilegesDOMBuilder privilegesBuilder, String code, String value) {
        privilegesBuilder.addPrivilege("urn:dk:gov:saml:cvrNumberIdentifier:25450442", String.format("urn:dk:healthcare:national-federation-role:code:%s:value:%s", code, value));
        assertionBuilder.setPrivileges(privilegesBuilder);
    }

    private void addNationalRoleToBst(OIOH2BSTSAMLAssertionBuilder assertionBuilder, BasicPrivilegesDOMBuilder privilegesBuilder, String value) {
        privilegesBuilder.addPrivilege("urn:dk:gov:saml:cvrNumberIdentifier:25450442", String.format("urn:dk:healthcare:national-federation-role:%s", value));
        assertionBuilder.setPrivileges(privilegesBuilder);
    }


    private void addNationalRoleToBst(OIOH3BSTSAMLAssertionBuilder assertionBuilder, BasicPrivilegesDOMBuilder privilegesBuilder, String code, String value) {
        privilegesBuilder.addPrivilege("urn:dk:gov:saml:cvrNumberIdentifier:25450442", String.format("urn:dk:healthcare:national-federation-role:code:%s:value:%s", code, value));
        assertionBuilder.setPrivileges(privilegesBuilder);
    }

    private String addNationalRoleToRequest(OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilder, String code, String value) {
        requestBuilder.setUserAuthorizationCode(null);
        String role = String.format("urn:dk:healthcare:national-federation-role:code:%s:value:%s", code, value);
        requestBuilder.setUserRole(role);
        return role;
    }

    private String addNationalRoleToRequest(OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilder, String value) {
        requestBuilder.setUserAuthorizationCode(null);
        String role = String.format("urn:dk:healthcare:national-federation-role:%s", value);
        requestBuilder.setUserRole(role);
        return role;
    }


    private OIOBSTSAMLAssertionToIDCardResponse buildAndSendRequest(TestClient testClient, AbstractOIOBSTSAMLAssertionBuilder assertionBuilder) {
        requestBuilder.setOIOBSTSAMLAssertion(assertionBuilder.build());
        return sendRequest(testClient, requestBuilder.build());
    }
    private OIOBSTSAMLAssertionToIDCardResponse sendRequest(TestClient testClient, Document requestDocument) {
        System.out.println(XmlUtil.node2String(requestDocument, false, false));
        return parseResponse(testClient.execute(XmlUtil.node2String(requestDocument, false, false)).document);
    }

    private static String uuidFromIdentity(TestIdentity identity) {
        return StringUtils.replace(identity.getSubjectSerialNumber(), "UI:DK-E:G:", "");
    }

    protected abstract TestClient getOces3TestClient();
}
