/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.ibo;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.model.dombuilders.IDCardToOIOSAMLAssertionRequestDOMBuilder;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.model.IboConfig;
import dk.sosi.sts.server.*;
import dk.sosi.sts.server.facade.MockServicesImpl;
import dk.sosi.sts.server.facade.TestClientHelper;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.service.IboConfigService;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;

import java.security.cert.X509Certificate;

import static org.junit.Assert.*;

/**
 * User: ksr Date: 19-05-2016 Time: 09:31
 */
@RunWith(Parameterized.class)
public class IboCertificateCheckTest {
	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	OIOSAMLFactory clientFactory = new OIOSAMLFactory();

	static final long UNTIL_OFFSET = 60 * 60 * 1000;

	private final String AUDIENCE = "/oces2.sundhed.dk";
	private MockServicesImpl services;
	private SOSIFactory sosiFactory;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3);

		return builder.build();
	}

	@Before
	public void setUp() {
		services = new MockServicesImpl(version);
		sosiFactory = services.buildSosiFactory(true);
	}

	private static UserIDCard createIDCard(SOSIFactory factory, Version version) {
		final CredentialVault mocesVault = SealPPCredentialVaults.getValidMocesNoCpr(version);
		final SOSIFactory sosiFactory = new SOSIFactory(mocesVault, System.getProperties());
		final CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "12345", "Pelles Pølsefabrik");
		final UserInfo userInfo = new UserInfo("1111111118", "Hans", "Dampf", null, null, "pølsemager", null);
		final String alternativeIdentifier = new CertificateInfo(mocesVault.getSystemCredentialPair().getCertificate()).toString();
		final UserIDCard userIDCard = sosiFactory.createNewUserIDCard("IT-System", userInfo, careProvider, AuthenticationLevel.MOCES_TRUSTED_USER, null, null, null, alternativeIdentifier);

		// cheating: signing with the sts certificate
		final SecurityTokenRequest tokenRequest = sosiFactory.createNewSecurityTokenRequest();
		tokenRequest.setIDCard(userIDCard);
		userIDCard.sign(tokenRequest.serialize2DOMDocument(), factory.getSignatureProvider());

		return userIDCard;
	}

	@Test
	public void testRequestWithNoValidation() throws Exception {
		final X509Certificate certificate = SealPPCredentialVaults.getExpiredVocesVault(version).getSystemCredentialPair().getCertificate();
		TestClientHelper testClientHelper = createTestClientHelper(certificate, false);

		final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();

		domBuilder.setAudience(AUDIENCE);
		UserIDCard idCard = createIDCard(sosiFactory, version);
		domBuilder.setUserIDCard(idCard);
		domBuilder.setSigningVault(SealPPCredentialVaults.getValidVocesVault(version));

		final IDCardToOIOSAMLAssertionResponse response = callIBO(testClientHelper, domBuilder.build());

		assertNotNull(response);
		assertFalse(response.getFaultString(), response.isFault());
		assertNotNull(response.getEncryptedOIOSAMLAssertionElement());

	}

	@Test
	public void testRequestWithValidation() throws Exception {
		final X509Certificate certificate = SealPPCredentialVaults.getExpiredVocesVault(version).getSystemCredentialPair().getCertificate();
		TestClientHelper testClientHelper = createTestClientHelper(certificate, true);

		final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();

		domBuilder.setAudience(AUDIENCE);
		UserIDCard idCard = createIDCard(sosiFactory, version);
		domBuilder.setUserIDCard(idCard);
		domBuilder.setSigningVault(SealPPCredentialVaults.getValidVocesVault(version));

		final IDCardToOIOSAMLAssertionResponse response = callIBO(testClientHelper, domBuilder.build());

		assertNotNull(response);
		assertTrue(response.isFault());
		assertEquals(StsFaultActorValues.STS, response.getFaultActor());
		assertEquals(StsFaultCodeValues.WST_REQUEST_FAILED, response.getFaultCode());
		assertTrue(response.getFaultString(), response.getFaultString().contains("Certificate expired"));
		assertNull(response.getEncryptedOIOSAMLAssertionElement());
	}

	protected IDCardToOIOSAMLAssertionResponse callIBO(TestClientHelper testClientHelper, Document build) {
		Document response = testClientHelper.execute(XmlUtil.node2String(build, false, false)).document;
		return clientFactory.createIDCardToOIOSAMLAssertionResponseModelBuilder().build(response);
	}

	private TestClientHelper createTestClientHelper(final X509Certificate certificate, final boolean checkTargetCertificate) {

		SecurityTokenService sts = new SecurityTokenService(sosiFactory, services, services.getUserValidationService(), "dummy blacklisted ITSystemName");

		final IboRequestHandler iboRequestHandler = new IboRequestHandler();
		iboRequestHandler.setStsStatus(new StsStatus());
		iboRequestHandler.setCrlService(services);
		iboRequestHandler.setSosiFactory(sosiFactory);

		iboRequestHandler.setIboConfigService(getConfigService(certificate));
		iboRequestHandler.setSts(sts);
		iboRequestHandler.setCheckTargetCertificate(checkTargetCertificate);
		return new TestClientHelper(iboRequestHandler);
	}

	private static IboConfigService getConfigService(final X509Certificate targetCertificate) {
		return new IboConfigService() {
			@Override
			public IboConfig getAudienceConfiguration(final String audience) {
				return new IboConfig() {
					@Override
					public String getAudience() {
						return audience;
					}

					@Override
					public X509Certificate getTargetCertificate() {
						return targetCertificate;
					}

					@Override
					public String getRecipientURL() {
						return "http://sundhed.dk/login";
					}

					@Override
					public boolean includeBootStrapToken() {
						return true;
					}

					@Override
					public long getDeliveryNotOnOrAfterOffset() {
						return 0; // NOW
					}

					@Override
					public long getNotBeforeOffset() {
						return 0; // NOW
					}

					@Override
					public long getNotOnOrAfterOffset() {
						return UNTIL_OFFSET;
					}

					@Override
					public long getIdCardMaxAgeMinutes() {
						return 24 * 60;
					}
				};
			}
		};
	}

}
