/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.ibo;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.EncryptionUtil;
import dk.sosi.seal.model.IDCardToOIOSAMLAssertionResponse;
import dk.sosi.seal.model.OIOSAMLAssertion;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.dombuilders.IDCardToOIOSAMLAssertionRequestDOMBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.model.IboConfig;
import dk.sosi.sts.server.IboRequestHandler;
import dk.sosi.sts.server.SecurityTokenService;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.StsStatus;
import dk.sosi.sts.server.facade.MockServicesImpl;
import dk.sosi.sts.server.facade.TestClientHelper;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.service.IboConfigService;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.security.cert.X509Certificate;
import java.util.Date;

import static org.junit.Assert.*;

/**
 * Created by Kent on 24-02-2017.
 */
public class IboEmbeddedTest extends IboTest {
    TestClientHelper testClientHelper;
    private MockServicesImpl mockServices;
    private SOSIFactory stsSosiFactory;
    private IboRequestHandler iboRequestHandler;

    @Before
    public void setUp() throws Exception {
        final IboConfigService configService = new IboConfigService() {
            @Override
            public IboConfig getAudienceConfiguration(final String audience) {
                if (!audience.equals(audienceStsTest)) {
                    return null;
                }

                return new IboConfig() {
                    @Override
                    public String getAudience() {
                        return audience;
                    }

                    @Override
                    public X509Certificate getTargetCertificate() {
                        return getAudienceVault(version).getSystemCredentialPair().getCertificate();
                    }

                    @Override
                    public String getRecipientURL() {
                        return IboTest.getRecipientUrl();
                    }

                    @Override
                    public boolean includeBootStrapToken() {
                        return includeBST;
                    }

                    @Override
                    public long getDeliveryNotOnOrAfterOffset() {
                        return 0; // NOW
                    }

                    @Override
                    public long getNotBeforeOffset() {
                        return 0; // NOW
                    }

                    @Override
                    public long getNotOnOrAfterOffset() {
                        return UNTIL_OFFSET;
                    }

                    @Override
                    public long getIdCardMaxAgeMinutes() {
                        return idCardMaxAgeMins;
                    }
                };
            }
        };

        mockServices = new MockServicesImpl(version);

        stsSosiFactory = mockServices.buildSosiFactory(true);

        SecurityTokenService sts = new SecurityTokenService(stsSosiFactory, mockServices, mockServices.getUserValidationService(), "dummy blacklisted ITSystemName");

        iboRequestHandler = new IboRequestHandler();
        iboRequestHandler.setStsStatus(new StsStatus());
        iboRequestHandler.setCrlService(mockServices);
        iboRequestHandler.setSosiFactory(stsSosiFactory);
        iboRequestHandler.setIboConfigService(configService);
        iboRequestHandler.setSts(sts);
        iboRequestHandler.setValidateEnvelopeSignature(false);
        iboRequestHandler.setEmptyAttributeValue("NONE");

        testClientHelper = new TestClientHelper(iboRequestHandler);
    }

    // if this is possible with Seal API; maybe just manipulate the Document when using callIBO
    @Test
    public void testNoSignatureWithValidation() throws Exception {
        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory
                .createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        final String audience = audienceStsTest;
        domBuilder.setAudience(audience);
        UserIDCard idCard = createIDCard(null);
        domBuilder.setUserIDCard(idCard);

        setupEnvelopeSignatureValidation();
        final IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());

        assertNotNull(response);
        assertTrue(response.isFault());

        assertTrue("FaultString is not as expected", response.getFaultString().contains("Could not find Liberty signature element"));
    }

    @Test
    public void testInvalidEnvelopeSignature() throws Exception {
        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory
                .createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        final String audience = audienceStsTest;
        domBuilder.setAudience(audience);
        UserIDCard idCard = createIDCard(null);
        domBuilder.setUserIDCard(idCard);
        domBuilder.setSigningVault(getVocesVault(version));


        // Build request and manipulate the signature to make it invalid
        final String soapEnvelope = XmlUtil.node2String(domBuilder.build(), false, false)
                .replaceAll("<ds:SignatureValue>...", "<ds:SignatureValue>abe");
        setupEnvelopeSignatureValidation();
        final Document responseDoc =  testClientHelper.execute(soapEnvelope).document;

        final IDCardToOIOSAMLAssertionResponse response = clientFactory
                .createIDCardToOIOSAMLAssertionResponseModelBuilder().build(responseDoc);

        assertTrue(response.isFault());
        assertTrue("FaultString is not as expected", response.getFaultString()
                .contains("Liberty signature could not be validated"));
    }

    @Test
    public void testRevokedEnvelopeCertificate() {
        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory
                .createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        final String audience = audienceStsTest;
        domBuilder.setAudience(audience);
        UserIDCard idCard = createIDCard(null);
        domBuilder.setUserIDCard(idCard);
        domBuilder.setSigningVault(SealPPCredentialVaults.getRevokedMocesVault(version));

        setupEnvelopeSignatureValidation();
        final IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());

        assertTrue(response.isFault());
        assertTrue(response.getFaultString(), response.getFaultString().contains("Certificate revoked"));
    }

    @Test
    public void testWithoutBST() {
        includeBST = false;
        try {
            final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory
                    .createIDCardToOIOSAMLAssertionRequestDOMBuilder();

            final String audience = audienceStsTest;
            domBuilder.setAudience(audience);
            UserIDCard idCard = createIDCard(null);
            domBuilder.setUserIDCard(idCard);
            domBuilder.setSigningVault(getVocesVault(version));

            final IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());

            assertNotNull(response);
            assertFalse(response.getFaultString(), response.isFault());

            // Decrypt it
            final Element element = EncryptionUtil.decryptAndDetach(response.getEncryptedOIOSAMLAssertionElement(),
                    getAudienceKey(version));
            final OIOSAMLAssertion assertion = new OIOSAMLAssertion(element);

            final UserIDCard assertionUserIDCard = assertion.getUserIDCard();
            assertNull(assertionUserIDCard);

        } finally {
            includeBST = true;
        }
    }

    @Test
    public void testNoMessageID_ValidatingEnvelope() {
        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory
                .createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        domBuilder.setAudience(audienceStsTest);
        domBuilder.setUserIDCard(createIDCard(null));
        domBuilder.setSigningVault(getVocesVault(version));

        final Document reqDoc = domBuilder.build();
        removeMessageID(reqDoc);

        // then a failing request -- since we tampered with the header so the signature gets invalid and we now check
        setupEnvelopeSignatureValidation();
        final IDCardToOIOSAMLAssertionResponse response = callIBO(reqDoc);

        assertTrue(response.isFault());
        assertNotNull(response.getRelatesTo());
        assertTrue("relatesTo should start with '"+ IboRequestHandler.URN_STS_PREFIX+"'",
                response.getRelatesTo().startsWith(IboRequestHandler.URN_STS_PREFIX));
    }

    @Test
    public void testOldIdCard() throws Exception {
        idCardMaxAgeMins = 30;

        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory
                .createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        final String audience = audienceStsTest;
        domBuilder.setAudience(audience);
        final CredentialVault mocesVault = getMocesVault(version);
        final SOSIFactory sosiFactory = new SOSIFactory(mocesVault, System.getProperties());
        final UserIDCard userIDCard = createUnsignedIdCard(mocesVault, sosiFactory);

        // more than 30 mins old
        Date createdDate = new Date(System.currentTimeMillis() - 1*60*60*1000);
        // not expired yet, but still <24 hours after creation
        Date expiredDate = new Date(System.currentTimeMillis() + 1*60*60*1000);

        MockServicesImpl.setCreatedDate(userIDCard, createdDate);
        MockServicesImpl.setExpiredDate(userIDCard, expiredDate);


        UserIDCard idCard = getStsSignedIdCard(sosiFactory, userIDCard);
        domBuilder.setUserIDCard(idCard);
        domBuilder.setSigningVault(getVocesVault(version));

        final IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());
        assertTrue("should be a fault", response.isFault());
        assertEquals(StsFaultCodeValues.WST_INVALID_TIME_RANGE, response.getFaultCode());
        assertTrue(response.getFaultString(),response.getFaultString().contains("IDCard is too old for this audience"));
    }

    private void setupEnvelopeSignatureValidation() {
        iboRequestHandler.setValidateEnvelopeSignature(true);
    }

    @Override
    protected TestClientHelper getTestClient() {
        return testClientHelper;
    }

    @Override
    protected UserIDCard getStsSignedIdCard(SOSIFactory sosiFactory, UserIDCard userIDCard) {
        return getSignedIdCard(this.stsSosiFactory.getSignatureProvider(), sosiFactory, userIDCard);
    }
}
