/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.ibo;

import java.net.MalformedURLException;

import dk.sosi.seal.model.Message;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.testtools.testrequest.MessageGeneratorHelper;
import dk.sosi.testtools.testrequest.STSMessageGenerator;
import dk.sosi.testtools.testrequest.WSTError;
import org.junit.Before;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.sts.server.facade.HttpTestClient;
import dk.sosi.sts.server.facade.MockServicesImpl;
import dk.sosi.sts.server.facade.TestClient;
import org.w3c.dom.Document;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

/**
 *
 */
public class IboIT extends IboTest {

	TestClient testClientHelper;
	private TestClient stsTestClientHelper;

	@Before
	public void initTestClient() throws MalformedURLException {
		testClientHelper = new HttpTestClient("/sts/services/Sosi2OIOSaml");
		stsTestClientHelper = new HttpTestClient("/sts/services/NewSecurityTokenService");
	}

	@Override
	protected TestClient getTestClient() {
		return testClientHelper;
	}

	@Override
	protected UserIDCard getStsSignedIdCard(SOSIFactory sosiFactory, UserIDCard userIDCard) {
		return userIDCard.isValidInTime() ? getRealStsSignedIdCard(sosiFactory, userIDCard)
				: getMockSignedIdCard(sosiFactory, userIDCard); // We can't get STS to sign an invalid id card
	}

	private UserIDCard getRealStsSignedIdCard(SOSIFactory sosiFactory, UserIDCard userIDCard) {
		SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();
		securityTokenRequest.setIDCard(userIDCard);
		Document unsignedRequest = securityTokenRequest.serialize2DOMDocument();
		Document signedRequest = MessageGeneratorHelper.sign(unsignedRequest, sosiFactory.getCredentialVault());

		Document securityTokenResponse = stsTestClientHelper.execute(signedRequest);
		Message responseMessage = checkDeserializeResponse(securityTokenResponse, sosiFactory);
		assertSuccess(securityTokenResponse);
		assertTrue(responseMessage.getIDCard() instanceof UserIDCard);

		return (UserIDCard) responseMessage.getIDCard();
	}

	private UserIDCard getMockSignedIdCard(SOSIFactory sosiFactory, UserIDCard userIDCard) {
		final MockServicesImpl mockServices = new MockServicesImpl(version);
		SOSIFactory stsFactory = mockServices.buildSosiFactory(false);

		return getSignedIdCard(stsFactory.getSignatureProvider(), sosiFactory, userIDCard);
	}

	private Message checkDeserializeResponse(Document response, SOSIFactory factory) {
		Message msg = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(response));
		if (msg.getMessageID() == null || "".equals(msg.getMessageID().trim())) {
			fail("Warning: deserialized response has empty message ID!");
		}
		return msg;
	}

	private void assertSuccess(Document response) {
		WSTError error = new STSMessageGenerator(SignatureUtil.setupCryptoProviderForJVM(), STSMessageGenerator.VERSION.OCES3).getError(response);
		if (error != null) {
			fail(error.getFaultString());
		}
	}
}
