/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.ibo;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.model.constants.WSSETags;
import dk.sosi.seal.model.dombuilders.IDCardToOIOSAMLAssertionRequestDOMBuilder;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.pki.CredentialVaultSignatureProvider;
import dk.sosi.seal.pki.SignatureProvider;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.IboRequestHandler;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.facade.MockServicesImpl;
import dk.sosi.sts.server.facade.TestClient;
import dk.sosi.sts.server.test.*;

import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.security.PrivateKey;
import java.util.Date;
import java.util.Properties;

/**
 *
 */
@RunWith(Parameterized.class)
public abstract class IboTest {
    @Parameter(0)
    public String suiteName;

    @Parameter(1)
    public Version version;

    @Parameter(2)
    public String audienceStsTest;

    @Parameter(3)
    public String cvr;

    @Parameter(4)
    public String cpr;

    OIOSAMLFactory clientFactory = new OIOSAMLFactory();

    final long UNTIL_OFFSET = 60 * 60 * 1000;

    boolean includeBST = true;
    boolean withCertInfoInIdCard = true;
    protected int idCardMaxAgeMins = 24 * 60;

    /* Setting up test parameters */
    @Parameters(name = "{0}")
    public static Iterable<Object[]> data() {
        TestParameterBuilder builder = new TestParameterBuilder();
        TestIdentity oces3Identity = TestIdentityProvider.getIdentityWithTrustedCvrAndOneAuthorization(Version.OCES3);

        builder.newTest("OCES3").add(Version.OCES3, "/ststest3", oces3Identity.getCvr(), oces3Identity.getCpr());

        return builder.build();
    }

    protected static final String getRecipientUrl() {
        return "/ststest/login";
    }

    public static CredentialVault getVocesVault(Version version) {
        return CertificateStore.getValidVocesVault(version);
    }

    public static CredentialVault getFocesVault(Version version) {
        return CertificateStore.getValidFocesVault(version);
    }

    public static PrivateKey getAudienceKey(Version version) {
        return getAudienceVault(version).getSystemCredentialPair().getPrivateKey();
    }

    protected static GenericCredentialVault getAudienceVault(Version version) {
        return CertificateStore.getValidVocesVault(version);
    }

    public static CredentialVault getMocesVault(Version version) {
        return SealPPCredentialVaults.getValidMocesWithCpr(version);
    }

    protected UserIDCard createIDCard(Date expiryDate) {
        final CredentialVault mocesVault = getMocesVault(version);
        final SOSIFactory sosiFactory = new SOSIFactory(mocesVault, System.getProperties());
        final UserIDCard userIDCard = createUnsignedIdCard(mocesVault, sosiFactory);

        if (expiryDate != null) {
            MockServicesImpl.setExpiredDate(userIDCard, expiryDate);
        }

        return getStsSignedIdCard(sosiFactory, userIDCard);
    }

    protected abstract TestClient getTestClient();

    protected abstract UserIDCard getStsSignedIdCard(SOSIFactory sosiFactory, UserIDCard userIDCard);

    protected UserIDCard createUnsignedIdCard(CredentialVault mocesVault, SOSIFactory sosiFactory) {
        final CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, cvr, "Pelles Pølsefabrik");
        final UserInfo userInfo = new UserInfo(cpr, "Hans", "Dampf", null, null, "pølsemager", null);
        final String alternativeIdentifier = withCertInfoInIdCard  // To simulate NewSTS for embedded tests
                ? new CertificateInfo(mocesVault.getSystemCredentialPair().getCertificate()).toString()
                : "test";
        return sosiFactory.createNewUserIDCard("IT-System", userInfo, careProvider, AuthenticationLevel.MOCES_TRUSTED_USER, null, null, null, alternativeIdentifier);
    }

    protected UserIDCard getSignedIdCard(SignatureProvider signatureProvider, SOSIFactory sosiFactory, UserIDCard userIDCard) {
        final SecurityTokenRequest tokenRequest = sosiFactory.createNewSecurityTokenRequest();
        tokenRequest.setIDCard(userIDCard);
        userIDCard.sign(tokenRequest.serialize2DOMDocument(), signatureProvider);
        return userIDCard;
    }

    private SystemIDCard createSystemIdCard() {
        CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "cvr", "navn");
        SystemInfo systemInfo = new SystemInfo(careProvider, "TE-STS");

        final CredentialVault vocesVault = getVocesVault(version);
        final String alternativeIdentifier = new CertificateInfo(vocesVault.getSystemCredentialPair().getCertificate()).toString();
        return TestUtil.createStsSosiFactory(version).createNewSystemIDCard(systemInfo.getITSystemName(), systemInfo.getCareProvider(), AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, alternativeIdentifier);
    }

    protected final IDCardToOIOSAMLAssertionResponse callIBO(Document build) {
        final Document response = sendRequest(build);
        return clientFactory.createIDCardToOIOSAMLAssertionResponseModelBuilder().build(response);
    }

    protected final Document sendRequest(Document build) {
        return getTestClient().execute(XmlUtil.node2String(build, false, false)).document;
    }

    protected void normalRequest(boolean sign) throws Exception {
        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        domBuilder.setAudience(audienceStsTest);
        UserIDCard idCard = createIDCard(null);
        domBuilder.setUserIDCard(idCard);
        CredentialVault clientVault = getVocesVault(version);
        if (sign) {
            domBuilder.setSigningVault(clientVault);
        }

        final IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());

        assertNotNull(response);
        assertFalse(response.getFaultString(), response.isFault());

        // Decrypt it
        Element encryptedOIOSAMLAssertionElement = response.getEncryptedOIOSAMLAssertionElement();
        final Element element = EncryptionUtil.decryptAndDetach(encryptedOIOSAMLAssertionElement, getAudienceKey(version));
        final OIOSAMLAssertion assertion = new OIOSAMLAssertion(element);

        final UserIDCard assertionUserIDCard = assertion.getUserIDCard();
        assertNotNull(assertionUserIDCard);

        assertEquals(idCard.getUserInfo(), assertionUserIDCard.getUserInfo());
        assertEquals(idCard.getAlternativeIdentifier(), assertionUserIDCard.getAlternativeIdentifier());
        assertEquals(idCard.getSystemInfo(), assertionUserIDCard.getSystemInfo());
        assertEquals(idCard.getAuthenticationLevel(), assertionUserIDCard.getAuthenticationLevel());

        assertEquals(audienceStsTest, assertion.getAudienceRestriction());
        // The idcard gets signed 5 minutes in the past
        assertTrue(withinFiveMinutesIsh(assertion.getUserAuthenticationInstant()));

        assertEquals(cpr, assertion.getCpr());
        assertEquals("Dampf", assertion.getSurName());
        assertEquals("Pelles Pølsefabrik", assertion.getOrganizationName());

        assertTrue("NotBefore should be now: " + assertion.getNotBefore().getTime() + " != " + System.currentTimeMillis(), aboutNow(assertion.getNotBefore()));
        final long lifetime = assertion.getNotOnOrAfter().getTime() - System.currentTimeMillis() - UNTIL_OFFSET;
        assertTrue("The lifetime should be right " + lifetime, lifetime < 30000);
        assertTrue("DeliveryNotOnOrAfter should be now", aboutNow(assertion.getDeliveryNotOnOrAfter()));

        // Parameterized validation
        if (Version.OCES3 == version) {
            assertEquals("NONE", assertion.getRidNumberIdentifier()); // RID findes ikke for OCES3
            assertEquals(withCertInfoInIdCard ? "C=DK,O=Den Danske Stat,CN=Den Danske Stat OCES udstedende-CA 1" : "NONE", assertion.getCertificateIssuer());
            assertFalse(assertion.isYouthCertificate()); // 15-18 year old ...
            assertEquals(withCertInfoInIdCard ? "229e2b5e296baf572feb5e9b6d83ee14debcac8a" : "NONE", assertion.getCertificateSerial());
            assertEquals("CVR:" + cvr + "-RID:NONE", assertion.getUID());
            String subjectNameId = withCertInfoInIdCard ? "C=DK,O=Testorganisation nr. 91026150,CN=Thormund Nissen,Serial=UI:DK-E:G:dc7f4a98-d4f9-4d77-84f9-ae83c4e0bf61"
                    : String.format("C=DK,O=%s // CVR: %s,CN=%s %s,Serial=CVR:%s-RID:NONE",
                            idCard.getSystemInfo().getCareProvider().getOrgName(), cvr, idCard.getUserInfo().getGivenName(), idCard.getUserInfo().getSurName(), cvr);
            assertEquals(subjectNameId, assertion.getSubjectNameID());
            assertNotNull(assertion.getDeliveryNotOnOrAfter());
            // From configuration
            assertEquals(getRecipientUrl(), assertion.getRecipient());
        } else {
            throw new Exception("Ukendt version");
        }
    }

    @Test
    @Ignore("SDS-7027")
    public void testRequest() throws Exception {
        normalRequest(true);
    }

    @Test
    @Ignore("SDS-7027")
    public void testRequestWithoutCertificateInfoInIdCard() throws Exception {
        try {
            withCertInfoInIdCard = false;
            normalRequest(true);
        } finally {
            withCertInfoInIdCard = true;
        }
    }

    /**
     * SDS-4600 MitID / OCES3
     *
     * Testcase: udløbet idkort
     *
     */
    @Test
    public void testExpiredIdCard() throws Exception {
        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        final String audience = audienceStsTest;
        domBuilder.setAudience(audience);
        final CredentialVault mocesVault = getMocesVault(version);
        final SOSIFactory sosiFactory = new SOSIFactory(mocesVault, System.getProperties());
        final UserIDCard userIDCard = createUnsignedIdCard(mocesVault, sosiFactory);

        Date expiryDate = new Date(System.currentTimeMillis() - 300000);
        MockServicesImpl.setExpiredDate(userIDCard, expiryDate);

        UserIDCard idCard = getStsSignedIdCard(sosiFactory, userIDCard);
        assertFalse("should not be valid", idCard.isValidInTime());
        domBuilder.setUserIDCard(idCard);
        domBuilder.setSigningVault(getVocesVault(version));

        final IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());
        assertTrue("should be a fault", response.isFault());
        assertEquals(response.getFaultCode() + ":" + response.getFaultString(), StsFaultCodeValues.WST_INVALID_TIME_RANGE, response.getFaultCode());
        assertTrue(response.getFaultCode() + ":" + response.getFaultString(), response.getFaultString().contains("IDCard is created after or expires before current system time"));
    }

    @Test
    @Ignore("SDS-7027")
    public void testRequestWithoutSignature() throws Exception {
        normalRequest(false);
    }

    private boolean withinFiveMinutesIsh(final Date date) {
        return Math.abs(System.currentTimeMillis() - date.getTime()) < 6 * 60 * 1000;
    }

    private boolean aboutNow(final Date date) {
        return System.currentTimeMillis() - date.getTime() < 60000;
    }

    // IMPORTANT: asserts on faultCode, faultString and faultActors are needed

    // don't sign with sts-certificate
    /**
     * SDS-4600 MitID / OCES3
     *
     * Testcase: idkort signeret af en ikke-trustet part
     *
     */
    @Test
    public void testWithUserSignedIdCard() {
        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        final String audience = audienceStsTest;
        domBuilder.setAudience(audience);

        final CredentialVault mocesVault = getMocesVault(version);

        final SOSIFactory sosiFactory = new SOSIFactory(mocesVault, System.getProperties());
        final UserIDCard userIDCard = createUnsignedIdCard(mocesVault, sosiFactory);
        UserIDCard idCard = getSignedIdCard(new CredentialVaultSignatureProvider(mocesVault, new Properties()), sosiFactory, userIDCard);
        domBuilder.setUserIDCard(idCard);
        domBuilder.setSigningVault(getVocesVault(version));

        final IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());

        assertTrue(response.isFault());
        assertTrue(response.getFaultString(), response.getFaultString().contains("The certificate that signed the security token is not trusted"));
    }

    @Test
    public void testMissingConfiguration() {
        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        final String audience = "//nowhere.dk";
        domBuilder.setAudience(audience);
        UserIDCard idCard = createIDCard(null);
        domBuilder.setUserIDCard(idCard);
        domBuilder.setSigningVault(getVocesVault(version));

        final IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());

        assertNotNull(response);
        assertTrue(response.isFault());

        assertTrue(response.getFaultString(), response.getFaultString().contains("No configuration found for audience"));
    }

    protected void removeMessageID(Document doc) {
        final Node soapHeader = doc.getDocumentElement().getChildNodes().item(0); // header then body
        final NodeList headers = soapHeader.getChildNodes();
        for (int i = 0; i < headers.getLength(); i++) {
            final Node header = headers.item(i);
            if (header.getLocalName().equals("MessageID")) {
                soapHeader.removeChild(header);
            }
        }
    }

    // STS-20
    @Test
    public void testNoMessageID() {
        final IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = clientFactory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();

        domBuilder.setAudience(audienceStsTest);
        domBuilder.setUserIDCard(createIDCard(null));
        domBuilder.setSigningVault(getVocesVault(version));

        final Document reqDoc = domBuilder.build();
        removeMessageID(reqDoc);

        // first a successful request
        final IDCardToOIOSAMLAssertionResponse response = callIBO(reqDoc);

        assertFalse(response.getFaultString(), response.isFault());
        assertNotNull(response.getRelatesTo());

        assertTrue("relatesTo should start with '" + IboRequestHandler.URN_STS_PREFIX + "'", response.getRelatesTo().startsWith(IboRequestHandler.URN_STS_PREFIX));

    }

    /**
     * SDS-4600 MitID / OCES3
     *
     * Testcase: idkort med brudt signatur
     *
     */
    @Test
    public void testBrokenIdCard() {
        TestBrokenIDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = new TestBrokenIDCardToOIOSAMLAssertionRequestDOMBuilder();

        CredentialVault clientVault = getMocesVault(version);
        String audience = audienceStsTest;

        UserIDCard idCard = createIDCard(null);
        domBuilder.setUserIDCard(idCard);
        domBuilder.setAudience(audience);
        domBuilder.setSigningVault(clientVault);

        IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());

        assertTrue(response.isFault());
        assertTrue(response.getFaultString(), response.getFaultString().contains("Authentication failed: Signature on IdCard could not be validated"));
    }


    /**
     * SDS-4600 MitID / OCES3
     *
     * Testcase: systemidkort (hvor omveksling fejler)
     *
     */
    @Test
    public void testWithSystemIdCard() {
        // create SystemIdCard
        SystemIDCard systemIDCard = createSystemIdCard();
        final CredentialVault mocesVault = getMocesVault(version);
        final SOSIFactory sosiFactory = new SOSIFactory(mocesVault, System.getProperties());

        // Dette ID-kort skal sættes på domBuilderen ellers er der nogle valideringer der fejler.
        // Det bliver ellers ikke brugt til noget, da denne udgave af builderen bruger system ID-kortet i stedet.
        UserIDCard dummyUserIDCardNotSend = createUnsignedIdCard(mocesVault, sosiFactory);

        TestSystemIDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = new TestSystemIDCardToOIOSAMLAssertionRequestDOMBuilder();
        domBuilder.setAudience(audienceStsTest);
        domBuilder.setSystemIDCard(systemIDCard);
        domBuilder.setUserIDCard(dummyUserIDCardNotSend);

        final IDCardToOIOSAMLAssertionResponse response = callIBO(domBuilder.build());

        assertTrue(response.isFault());
        assertTrue(response.getFaultString(), response.getFaultString().contains("The specified request failed: IDCard in request is not a UserIDCard!"));

    }

    static class TestBrokenIDCardToOIOSAMLAssertionRequestDOMBuilder extends IDCardToOIOSAMLAssertionRequestDOMBuilder {

        @Override
        protected void addExtraHeaders(Document doc, Element header) {
            super.addExtraHeaders(doc, header);
            if(placeIDCardInSOAPHeader) {
                final Element security = doc.createElementNS(NameSpaces.WSSE_SCHEMA, WSSETags.SECURITY_PREFIXED);
                header.appendChild(security);
                addIdCard(doc, security);
            }
        }

        @Override
        protected void addActAsTokens(Document doc, Element actAs) {
            if(placeIDCardInSOAPHeader) {
                addSecurityTokenReferenceToIDCard(doc, actAs);
            } else {
                addIdCard(doc, actAs);
            }
        }

        private void addIdCard(Document doc, Element parent) {
            final Element idCardElement = idCard.serialize2DOMDocument(null, doc);

            idCardElement.getChildNodes().item(0).setTextContent("prins pilfinger");

            parent.appendChild(idCardElement);
        }
    }

    static class TestSystemIDCardToOIOSAMLAssertionRequestDOMBuilder extends IDCardToOIOSAMLAssertionRequestDOMBuilder {

        private SystemIDCard systemIDCard;

        public void setSystemIDCard(SystemIDCard systemIDCard) {
            this.systemIDCard = systemIDCard;
        }

        @Override
        protected void addExtraHeaders(Document doc, Element header) {
            super.addExtraHeaders(doc, header);
            if(placeIDCardInSOAPHeader) {
                final Element security = doc.createElementNS(NameSpaces.WSSE_SCHEMA, WSSETags.SECURITY_PREFIXED);
                header.appendChild(security);
                addIdCard(doc, security);
            }
        }

        @Override
        protected void addActAsTokens(Document doc, Element actAs) {
            if(placeIDCardInSOAPHeader) {
                addSecurityTokenReferenceToIDCard(doc, actAs);
            } else {
                addIdCard(doc, actAs);
            }
        }

        private void addIdCard(Document doc, Element parent) {
            final Element idCardElement = systemIDCard.serialize2DOMDocument(null, doc);
            parent.appendChild(idCardElement);
        }
    }
}
