/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server.facade.jwt;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.fuldmagt.stub.ProcurationTestModel;
import dk.sosi.sts.integrations.blurring.BlurringService;
import dk.sosi.sts.integrations.blurring.BlurringServiceCaller;
import dk.sosi.sts.integrations.blurring.IdCardFactory;
import dk.sosi.sts.integrations.personinformation.Custody;
import dk.sosi.sts.integrations.personinformation.PersonInformationClient;
import dk.sosi.sts.model.BlurredOrganization;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.server.JWT2IdwsRequestHandler;
import dk.sosi.sts.server.StsStatus;
import dk.sosi.sts.server.facade.*;
import dk.sosi.sts.server.service.jwt.OcesKeyCloakJwtStrategy;
import dk.sosi.sts.server.test.StsCredentialVaults;
import dk.sosi.sts.service.ItsConfigService;

import java.io.File;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import static dk.sosi.sts.model.BlurredOrganization.Type.*;
import static java.util.Collections.emptyList;
import static java.util.stream.Stream.of;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class EmbeddedJwtServiceBuilder {

    private final File keyStoreFile = StsCredentialVaults.getJwtTrustStoreFile();

    private final Version version;

    public EmbeddedJwtServiceBuilder(Version version) {
    	this.version = version;
    }

    public TestClientHelper build() throws Exception {
        final MockServicesImpl mockServices = new MockServicesImpl(version);
        SOSIFactory serverSosiFactory = mockServices.buildSosiFactory(true);

        JWT2IdwsRequestHandler impl = new JWT2IdwsRequestHandler();

        ItsConfigService itsConfig = SamlTokenEmbeddedTestHelper.buildItsConfigService(new DefaultSamlTokenTestModel(version));
        impl.setItsConfigService(itsConfig);

        impl.setCrlService(mockServices);
        impl.setSosiFactory(serverSosiFactory);
        impl.setStsStatus(new StsStatus());
        impl.setOcesPidService(SamlTokenEmbeddedTestHelper.initMockOcesPidService());
        impl.setProcurationService(null);

        impl.setFuzzyTime(300000);
        impl.setAllowedDriftInSeconds(0L);
        impl.setTrustStorePassword("Test1234");
        impl.setTrustStoreFile(keyStoreFile);

        final OcesKeyCloakJwtStrategy keyCloakJwtStrategy = new OcesKeyCloakJwtStrategy();
        keyCloakJwtStrategy.setRequiredStsScope("sosi-sts");
        impl.setIssuerStrategies(Collections.singletonMap("http://sts-tester", keyCloakJwtStrategy));
        impl.afterPropertiesSet();
        impl.setProcurationService(new ProcurationMockService());
        impl.setBlurringService(new BlurringService(mock(SOSIFactory.class), mock(IdCardFactory.class),
                                                    mock(BlurringServiceCaller.class)) {
            @Override
            public List<BlurredOrganization> getActiveBlurrings(CprNr patientId, String flowId) {
                return patientId.equals(DefaultSamlTokenTestModel.EXISTING_CPR) ?
                            of(new BlurredOrganization("12345678", Cvr)).collect(Collectors.toList()) :
                       patientId.getValue().equals(ProcurationTestModel.CPR_1) ?
                            of(new BlurredOrganization("87654321", Cvr)).collect(Collectors.toList()) :
                       patientId.getValue().equals(ProcurationTestModel.CPR_4) ?
                               of(new BlurredOrganization("11223344", Cvr)).collect(Collectors.toList()) :
                       emptyList();
            }

            @Override
            public List<BlurredOrganization> getAllActiveOrgBlurrings(String flowId) {
                return of(new BlurredOrganization("650901000016000", Sor),
                          new BlurredOrganization("7042010", Shak)).collect(Collectors.toList());

            }

            @Override
            public String getSalt(String flowId) { return "test_salt"; }
        });

        // PersonInformationClient for testing
        PersonInformationClient personInformationClient = mock(PersonInformationClient.class);
        when(personInformationClient.personCustody("0501792275")).thenReturn(of(
                Custody.fromValues("WARDCUSTODY", "1111111119"),
                Custody.fromValues("PARENTALCUSTODY", "1111111118"))
                    .collect(Collectors.toSet()));
        impl.setPersonInformationClient(personInformationClient);
        return new TestClientHelper(impl);
    }
}
