/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server.facade.jwt;

import java.time.Duration;
import java.time.Instant;
import java.util.Date;

import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.model.CprNr;
import io.jsonwebtoken.Claims;
import io.jsonwebtoken.JwtBuilder;
import io.jsonwebtoken.Jwts;
import io.jsonwebtoken.SignatureAlgorithm;

public class JWTTokenBuilder {
    private final CredentialVault signingVault;
    private final CprNr cpr;
    private final String subjectDN;
    private String scope = "clear sosi-sts";
    private String issuer = "http://sts-tester";
    private Date expiration = new Date(Instant.now().plus(Duration.ofHours(2)).toEpochMilli());
    private String kid = "ssi_foces_2024";

    public JWTTokenBuilder(CredentialVault signingVault, CertificateInfo patientCertInfo, CprNr cpr) {
        this(signingVault, patientCertInfo.getSubjectDN().toString(), cpr);
    }

    public JWTTokenBuilder(CredentialVault signingVault, String subjectDN, CprNr cpr) {
        this.signingVault = signingVault;
        this.subjectDN = subjectDN;
        this.cpr = cpr;
    }

    public JWTTokenBuilder setKid(String kid) {
        this.kid = kid;
        return this;
    }

    public JWTTokenBuilder setExpiration(Date expiration) {
        this.expiration = expiration;
        return this;
    }

    public JWTTokenBuilder setIssuer(String issuer) {
        this.issuer = issuer;
        return this;
    }

    public JWTTokenBuilder setScope(String scope) {
        this.scope = scope;
        return this;
    }

    public String build() {
        final JwtBuilder builder = Jwts.builder();
        builder.signWith(SignatureAlgorithm.RS256, signingVault.getSystemCredentialPair().getPrivateKey());
        builder.setHeaderParam("kid", kid);
        final Claims claims = Jwts.claims();
        claims.setIssuer(issuer);
        if (subjectDN != null) {
            claims.put("certsubdn", subjectDN);
        }
        claims.put("scope", scope);
        if (cpr != null) {
            claims.put("cpr", cpr.getValue());
        }
        claims.setExpiration(expiration);
        builder.setClaims(claims);
        return builder.compact();
    }
}
