/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server.facade.jwt;

import dk.sosi.seal.model.*;
import dk.sosi.sts.fuldmagt.stub.ProcurationTestModel;
import dk.sosi.sts.integrations.blurring.BlurringService;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.server.JWT2IdwsRequestHandler;
import dk.sosi.sts.server.facade.TestClient;
import dk.sosi.sts.server.facade.TestClientHelper;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.w3c.dom.Document;

import java.util.List;
import java.util.stream.Collectors;

import static dk.sosi.seal.model.BlurredOrganisation.OrgType.*;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.*;
import static dk.sosi.seal.model.OnBehalfOfClaim.OnBehalfOfClaimType.*;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.assertSuccess;
import static dk.sosi.sts.server.facade.TestClientHelper.getMessageId;
import static java.util.Collections.emptyList;
import static org.junit.Assert.*;
import static org.mockito.Matchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class JwtTokenServiceEmbeddedTest extends JwtTokenServiceTest {
    private TestClientHelper testClientHelper;

    public JwtTokenServiceEmbeddedTest() {
        super(true);
    }

    @Before
    public void setUp() throws Exception {
        testClientHelper = new EmbeddedJwtServiceBuilder(version).build();
    }

    @Override
    protected TestClient getTestClient() {
        return testClientHelper;
    }

    @Test
    public void testWithAudienceThatDoesNotRequiringBlurring_ResponseDoesNotContainBlurringPrivilege() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr()).build();
        Document requestBuilder = createRequestBuilder(token,
                testModel.getValidWhitelistedVault(),
                testModel.getClearTextAudience(),
                null).build();
        JWTToIdentityTokenResponse response = getResponseModel(getResponseDocument(requestBuilder));
        assertSuccess(response, federation);
        assertNull(response.getIdentityToken().getBlurringInstructions());
    }

    @Test
    public void testWithAudienceThatRequiresBlurringButUserDoesNot_WithoutProcurationClaim_ResponseDoesNotContainBlurring() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                new CprNr("1234567890")).build();
        Document requestBuilder = createRequestBuilder(token,
                testModel.getValidWhitelistedVault(),
                testModel.getBlurringAudience(),
                null).build();
        BlurringInstructions11Impl blurredInstructions = assertResponseAndGetBlurredInstructions(requestBuilder);
        assertEquals(emptyList(), blurredInstructions.getBlurredOrganisations().stream().filter(
                b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList()));
        assertEquals("test_salt", blurredInstructions.getSalt());
    }

    @Test
    public void testWithAudienceThatRequiresBlurringWithoutProcurationClaim_ResponseContainsBlurringForPatient() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr()).build();
        Document requestBuilder = createRequestBuilder(token,
                testModel.getValidWhitelistedVault(),
                testModel.getBlurringAudience(),
                null).build();
        List<BlurredOrganisation> blurring = assertResponseAndGetBlurredOrganisations(requestBuilder).stream().filter(
                b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList());
        assertEquals(1, blurring.size());
        assertEquals(Cvr, blurring.get(0).getOrgType());
        assertEquals("12345678", blurring.get(0).getValue());
        assertEquals(SpecificForPerson, blurring.get(0).getReason());

        List<BlurredOrganisation> blurringDepartment = assertResponseAndGetBlurredOrganisations(requestBuilder).stream().filter(
                b -> b.getReason() == SpecificDepartment).collect(Collectors.toList());
        assertEquals(2, blurringDepartment.size());
        assertEquals(Sor, blurringDepartment.get(0).getOrgType());
        assertEquals(Shak, blurringDepartment.get(1).getOrgType());
    }

    @Test
    public void testWithAudienceThatRequiresBlurring_ResponseContainsBlurringForDepartment() {
        // Arrange
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr()).build();
        Document requestBuilder = createRequestBuilder(token,
                testModel.getValidWhitelistedVault(),
                testModel.getBlurringAudience(),
                null).build();
        // Act
        List<BlurredOrganisation> blurringDepartment = assertResponseAndGetBlurredOrganisations(requestBuilder).stream().filter(
                b -> b.getReason() == SpecificDepartment).collect(Collectors.toList());

        // Assert
        assertEquals(2, blurringDepartment.size());
        assertEquals(Sor, blurringDepartment.get(0).getOrgType());
        assertEquals(Shak, blurringDepartment.get(1).getOrgType());
        assertEquals("650901000016000", blurringDepartment.get(0).getValue());
        assertEquals("7042010", blurringDepartment.get(1).getValue());
    }

    @Test
    public void testWithAudienceThatDoesNotRequireBlurringWithProcurationClaim_ResponseDoesNotContainBlurringPrivilege() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr()).build();
        Document requestBuilder = createRequestBuilder(token,
                testModel.getValidWhitelistedVault(),
                testModel.getClearTextAudience(),
                null)
                .setOnBehalfOfClaim(ProcurationTestModel.CPR_1, Procuration)
                .build();
        JWTToIdentityTokenResponse response = getResponseModel(getResponseDocument(requestBuilder));
        assertSuccess(response, federation);
        assertNull(response.getIdentityToken().getBlurringInstructions());
    }

    @Test
    public void testWithAudienceThatRequiresBlurringWithProcurationClaim_ResponseContainsBlurringForPatients() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr()).build();
        Document requestBuilder =
                createRequestBuilder(token,
                        testModel.getValidWhitelistedVault(),
                        testModel.getBlurringAudience(),
                        null)
                        .setOnBehalfOfClaim(ProcurationTestModel.CPR_1, Procuration)
                        .build();
        List<BlurredOrganisation> blurring = assertResponseAndGetBlurredOrganisations(requestBuilder).stream().filter(
                b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList());
        assertEquals(2, blurring.size());
        assertEquals(Cvr, blurring.get(0).getOrgType());
        assertEquals("12345678", blurring.get(0).getValue());
        assertEquals(SpecificForPerson, blurring.get(0).getReason());
        assertEquals(Cvr, blurring.get(1).getOrgType());
        assertEquals("87654321", blurring.get(1).getValue());
        assertEquals(FromRelatedPerson, blurring.get(1).getReason());
    }

    @Test
    public void testBlurringWithParentalClaim_ResponseContainsBlurring() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr()).build();
        Document requestBuilder =
                createRequestBuilder(token,
                        testModel.getValidWhitelistedVault(),
                        testModel.getBlurringAudience(),
                        null)
                        .setOnBehalfOfClaim(ProcurationTestModel.CPR_1, ParentalCustody)
                        .build();
        List<BlurredOrganisation> blurring = assertResponseAndGetBlurredOrganisations(requestBuilder).stream().filter(
                b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList());
        assertEquals(2, blurring.size());
        assertEquals(Cvr, blurring.get(0).getOrgType());
        assertEquals("12345678", blurring.get(0).getValue());
        assertEquals(SpecificForPerson, blurring.get(0).getReason());
        assertEquals(Cvr, blurring.get(1).getOrgType());
        assertEquals("87654321", blurring.get(1).getValue());
        assertEquals(FromRelatedPerson, blurring.get(1).getReason());
    }

    @Test
    public void testBlurringWithWardCustody_ResponseContainsBlurring() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr()).build();
        Document requestBuilder =
                createRequestBuilder(token,
                        testModel.getValidWhitelistedVault(),
                        testModel.getBlurringAudience(),
                        null)
                        .setOnBehalfOfClaim(ProcurationTestModel.CPR_4, WardCustody)
                        .build();
        List<BlurredOrganisation> blurring = assertResponseAndGetBlurredOrganisations(requestBuilder).stream().filter(
                b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList());
        assertEquals(2, blurring.size());
        assertEquals(Cvr, blurring.get(0).getOrgType());
        assertEquals("12345678", blurring.get(0).getValue());
        assertEquals(SpecificForPerson, blurring.get(0).getReason());
        assertEquals(Cvr, blurring.get(1).getOrgType());
        assertEquals("11223344", blurring.get(1).getValue());
        assertEquals(FromRelatedPerson, blurring.get(1).getReason());
    }

    @Test
    public void testWithAudienceThatRequiresBlurringWithProcurationClaim_OnlyUserRequireBlurring_ResponseContainsBlurringForPatient() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr()).build();
        Document request = createRequestBuilder(token,
                testModel.getValidWhitelistedVault(),
                testModel.getBlurringAudience(),
                null)
                .setOnBehalfOfClaim("0987654321", Procuration)
                .build();
        List<BlurredOrganisation> blurring = assertResponseAndGetBlurredOrganisations(request).stream().filter(
                b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList());
        assertEquals(1, blurring.size());
        assertEquals(Cvr, blurring.get(0).getOrgType());
        assertEquals("12345678", blurring.get(0).getValue());
        assertEquals(SpecificForPerson, blurring.get(0).getReason());
    }

    @Test
    public void testWithAudienceThatRequiresBlurringWithProcurationClaim_OnlyUserRequireBlurring_FlowIdToTraceRequestBetweenComponentsIsCorrect()
            throws Exception {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr()).build();
        Document request = createRequestBuilder(token,
                testModel.getValidWhitelistedVault(),
                testModel.getBlurringAudience(),
                null)
                .setOnBehalfOfClaim("0987654321", Procuration)
                .build();
        String expectedFlowId = getMessageId(request).get();

        ArgumentCaptor<String> getBlurringsFlowId = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> getSaltFlowId = ArgumentCaptor.forClass(String.class);
        BlurringService blurringServiceMock = mock(BlurringService.class);
        when(blurringServiceMock.getActiveBlurrings(any(), getBlurringsFlowId.capture())).thenReturn(emptyList());
        when(blurringServiceMock.getSalt(getSaltFlowId.capture())).thenReturn("dummy");
        TestClientHelper testClientHelper = new EmbeddedJwtServiceBuilder(version).build();
        JWT2IdwsRequestHandler stsService = (JWT2IdwsRequestHandler) testClientHelper.stsService;
        stsService.setBlurringService(blurringServiceMock);
        assertResponseAndGetBlurredOrganisations(testClientHelper, request);
        assertEquals(expectedFlowId, getBlurringsFlowId.getValue());
        assertEquals(expectedFlowId, getSaltFlowId.getValue());
    }

    @Test
    public void testWithAudienceThatRequiresBlurringWithProcurationClaim_OnlyOnBehalfOfRequireBlurring_ResponseContainsBlurringForOnBehalfPatient() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                new CprNr("1234567890")).build();
        Document requestBuilder =
                createRequestBuilder(token,
                        testModel.getValidWhitelistedVault(),
                        testModel.getBlurringAudience(),
                        null)
                        .setOnBehalfOfClaim(ProcurationTestModel.CPR_1, Procuration)
                        .build();
        List<BlurredOrganisation> blurring = assertResponseAndGetBlurredOrganisations(requestBuilder).stream().filter(
                b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList());
        assertEquals(1, blurring.size());
        assertEquals(Cvr, blurring.get(0).getOrgType());
        assertEquals("87654321", blurring.get(0).getValue());
        assertEquals(FromRelatedPerson, blurring.get(0).getReason());
    }

    @Test
    public void testWithAudienceThatRequiresBlurringWithProcurationClaim_NoUserRequireBlurring_ResponseDoesNotContainBlurring() {
        String token = getJwtTokenBuilder(testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                new CprNr("1234567890")).build();
        Document requestBuilder =
                createRequestBuilder(token, testModel.getValidWhitelistedVault(), testModel.getBlurringAudience(), null)
                        .setOnBehalfOfClaim(new CprNr("0987654321").getValue(), Procuration)
                        .build();
        assertEquals(emptyList(), assertResponseAndGetBlurredOrganisations(requestBuilder).stream().filter(
                b -> b.getReason() == SpecificForPerson || b.getReason() == FromRelatedPerson).collect(Collectors.toList()));
    }

    private List<BlurredOrganisation> assertResponseAndGetBlurredOrganisations(Document requestDocument) {
        return assertResponseAndGetBlurredOrganisations(testClientHelper, requestDocument);
    }

    private List<BlurredOrganisation> assertResponseAndGetBlurredOrganisations(TestClient client,
                                                                               Document requestDocument) {
        return assertResponseAndGetBlurredInstructions(client, requestDocument).getBlurredOrganisations();
    }

    private BlurringInstructions11Impl assertResponseAndGetBlurredInstructions(Document requestDocument) {
        return assertResponseAndGetBlurredInstructions(testClientHelper, requestDocument);
    }

    private BlurringInstructions11Impl assertResponseAndGetBlurredInstructions(TestClient client,
                                                                               Document requestDocument) {
        JWTToIdentityTokenResponse response = getResponseModel(getResponseDocument(client, requestDocument));
        assertSuccess(response, federation);
        return (BlurringInstructions11Impl) response.getIdentityToken().getBlurringInstructions();
    }
}