/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server.facade.jwt;

import dk.sosi.seal.model.*;
import dk.sosi.seal.model.dombuilders.JWTToIdentityTokenRequestDOMBuilder;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.facade.TestClient;
import dk.sosi.sts.server.facade.TestClientHelper;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;

import static dk.sosi.seal.model.OnBehalfOfClaim.OnBehalfOfClaimType.*;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.assertFault;
import static org.junit.Assert.*;

public class JwtTokenServiceSubjectRelationsTest extends JwtTokenServiceTest {
    private TestClientHelper testClientHelper;

    public JwtTokenServiceSubjectRelationsTest() {
        super(true);
    }

    @Before
    public void setUp() throws Exception {
        testClientHelper = new EmbeddedJwtServiceBuilder(version).build();
    }

    @Override
    protected TestClient getTestClient() {
        return testClientHelper;
    }

    @Test
    public void testValidRequestWithParentalCustody() {
        // Arrange
        JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(
                testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr());
        String token = tokenBuilder.build();

        String audience = testModel.getClearTextAudience();
        JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
        requestBuilder.setOnBehalfOfClaim("1111111118", ParentalCustody);
        Document requestDocument = requestBuilder.build();

        // Act
        Document responseDoc = getResponseDocument(requestDocument);
        JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
        IdentityToken identityToken = response.getIdentityToken();

        // Assert
        assertNotNull(identityToken);
        SubjectRelations11Impl sr10 = (SubjectRelations11Impl) identityToken.getSubjectRelations();
        assertEquals(1, sr10.getVerifiedRelations().size());
        SubjectRelations11Impl.VerifiedRelation verifiedRelation = sr10.getVerifiedRelations().get(0);
        assertEquals("1111111118", verifiedRelation.getRelatedPersonID());
        assertEquals(SubjectRelations11Impl.RelationType.PARENTAL_CUSTODY_HOLDER, verifiedRelation.getRelationType());
    }

    @Test
    public void testValidRequestWithWardCustody() {
        // Arrange
        JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(
                testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr());
        String token = tokenBuilder.build();

        String audience = testModel.getClearTextAudience();
        JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
        requestBuilder.setOnBehalfOfClaim("1111111119", WardCustody);
        Document requestDocument = requestBuilder.build();

        // Act
        Document responseDoc = getResponseDocument(requestDocument);
        JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
        IdentityToken identityToken = response.getIdentityToken();

        // Assert
        assertNotNull(identityToken);
        SubjectRelations11Impl sr10 = (SubjectRelations11Impl) identityToken.getSubjectRelations();
        assertEquals(1, sr10.getVerifiedRelations().size());
        SubjectRelations11Impl.VerifiedRelation verifiedRelation = sr10.getVerifiedRelations().get(0);
        assertEquals("1111111119", verifiedRelation.getRelatedPersonID());
        assertEquals(SubjectRelations11Impl.RelationType.WARD_CUSTODY_HOLDER, verifiedRelation.getRelationType());
    }

    @Test
    public void testValidRequestWithWardCustodyClaim_certNotWhitelisted_returnsFault() {
        // Arrange
        JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(
                testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr());
        String token = tokenBuilder.build();

        String audience = testModel.getClearTextAudience();
        JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVocesVault(), audience, null);
        requestBuilder.setOnBehalfOfClaim("1111111119", WardCustody);
        Document requestDocument = requestBuilder.build();

        // Act
        Document responseDoc = getResponseDocument(requestDocument);
        JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

        // Assert
        assertFault(response, StsFaultActorValues.ITS, "Ward claim not whitelisted for this certificate SSN and audience");
    }

    @Test
    public void testValidRequestWithWardCustodyClaim_audienceNotWhitelisted_returnsFault() {
        // Arrange
        JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(
                testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr());
        String token = tokenBuilder.build();

        String audience = testModel.getClearTextAudienceNoWardWhitelisting();
        JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
        requestBuilder.setOnBehalfOfClaim("1111111119", WardCustody);
        Document requestDocument = requestBuilder.build();

        // Act
        Document responseDoc = getResponseDocument(requestDocument);
        JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

        // Assert
        assertFault(response, StsFaultActorValues.ITS, "Ward claim not whitelisted for this certificate SSN and audience");
    }


    @Test
    public void testValidRequestWithPartlyWardCustody_returnsFault() {
        // Arrange
        JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(
                testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr());
        String token = tokenBuilder.build();
        String audience = testModel.getClearTextAudience();
        JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
        requestBuilder.setOnBehalfOfClaim("1111111118", PartlyWardCustody);
        Document requestDocument = requestBuilder.build();

        // Act
        Document responseDoc = getResponseDocument(requestDocument);
        JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

        // Assert
        assertFault(response, StsFaultActorValues.ITS, "partlyWardCustody is not allowed as a Subject Relation");
    }

    @Test
    public void testValidRequestWithInvalidParentalCustody_returnsFault() {
        // Arrange
        JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(
                testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr());
        String token = tokenBuilder.build();
        String audience = testModel.getClearTextAudience();
        JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
        requestBuilder.setOnBehalfOfClaim("1111111120", ParentalCustody); // Invalid claim
        Document requestDocument = requestBuilder.build();

        // Act
        Document responseDoc = getResponseDocument(requestDocument);
        JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

        // Assert
        assertFault(response, StsFaultActorValues.ITS, "The person is not in parental custody of the given person");
    }

    @Test
    public void testValidRequestWithInvalidgWardCustody_returnsFault() {
        // Arrange
        JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(
                testModel.getExternalTrustedIdp(),
                testModel.getPatientCertificateInfo(),
                testModel.getPatientCpr());
        String token = tokenBuilder.build();
        String audience = testModel.getClearTextAudience();
        JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
        requestBuilder.setOnBehalfOfClaim("1111111120", WardCustody); // Invalid claim
        Document requestDocument = requestBuilder.build();

        // Act
        Document responseDoc = getResponseDocument(requestDocument);
        JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

        // Assert
        assertFault(response, StsFaultActorValues.ITS, "The person is not in ward custody of the given person");
    }
}