/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.jwt;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.HealthcareSAMLAttributes;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.model.dombuilders.JWTToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.EmptyCredentialVault;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.facade.DefaultSamlTokenTestModel;
import dk.sosi.sts.server.facade.TestClient;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.TestUtil;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.security.cert.X509Certificate;
import java.util.Date;

import static dk.sosi.seal.model.OnBehalfOfClaim.OnBehalfOfClaimType.*;
import static dk.sosi.sts.fuldmagt.stub.ProcurationTestModel.*;
import static dk.sosi.sts.server.facade.IdentityTokenResponseValidation.*;
import static org.junit.Assert.*;

/**
 * Created by Kent on 20-03-2017.
 */
@RunWith(Parameterized.class)
public abstract class JwtTokenServiceTest {
	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	@Parameter(2)
	public String kid_valid;

	@Parameter(3)
	public String kid_revoked;

	protected DefaultSamlTokenTestModel testModel;

	protected Federation federation;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3, "test_foces3_2027", "revoked_foces3_2024");

		return builder.build();
	}

	public JwtTokenServiceTest(boolean isEmbeddedTest) {
		federation = TestUtil.createTestFederation(isEmbeddedTest);
	}

	@Before
	public void setup() {
		testModel = new DefaultSamlTokenTestModel(version);
	}

	public JWTTokenBuilder getJwtTokenBuilder(CredentialVault externalTrustedIdp, CertificateInfo patientCertificateInfo, CprNr patientCpr) {
		return new JWTTokenBuilder(externalTrustedIdp, patientCertificateInfo, patientCpr).setKid(kid_valid);
	}

	public JWTTokenBuilder getJwtTokenBuilder(CredentialVault externalTrustedIdp, String certsubdn, CprNr patientCpr) {
		return new JWTTokenBuilder(externalTrustedIdp, certsubdn, patientCpr).setKid(kid_valid);
	}

	@Test
	public void testTrustedButExpiredMessageSigner() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		final GenericCredentialVault trustedButExpiredConsumer = SealPPCredentialVaults.getExpiredFocesVault(version);
		Document requestDocument = buildRequestDocument(token, trustedButExpiredConsumer, audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Certificate expired or not yet valid");

		assertNull(response.getIdentityToken());
	}

	@Test
	public void testUntrustedMessageSigner() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		final GenericCredentialVault untrusted = SealPPCredentialVaults.getValidMocesNoCpr(version);
		Document requestDocument = buildRequestDocument(token, untrusted, audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Signing certificate not whitelisted for this audience");

		assertNull(response.getIdentityToken());
	}

	@Test
	public void testRevokedTrustedMessageSigner() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		final GenericCredentialVault revoked = SealPPCredentialVaults.getRevokedFocesVault(version);
		Document requestDocument = buildRequestDocument(token, revoked, audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Certificate revoked");

		assertNull(response.getIdentityToken());
	}

	@Test
	public void testInvalidMessageSignature() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());
		Element header = XmlUtil.getFirstChildElementNS(requestDocument.getDocumentElement(), "http://schemas.xmlsoap.org/soap/envelope/", "Header");
		Element security = XmlUtil.getFirstChildElementNS(header, "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd", "Security");
		Element signature = XmlUtil.getFirstChildElementNS(security, DSTags.signature.getNS(), DSTags.signature.getTagName());
		Element signatureValue = XmlUtil.getFirstChildElementNS(signature, "http://www.w3.org/2000/09/xmldsig#", "SignatureValue");

		signatureValue.getFirstChild().setNodeValue(signatureValue.getFirstChild().getNodeValue() + "-break-");

		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	@Test
	public void testExpiredTokenSigner() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(SealPPCredentialVaults.getExpiredFocesVault(version), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.setKid("expired").build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_REQUEST_FAILED, "JWT signer expired");
	}

	@Test
	@Ignore("SDS-7321") // Revoked certificate er udløbet. Skal fixes.
	public void testRevokedTokenSigner() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(SealPPCredentialVaults.getRevokedFocesVault(version), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.setKid(kid_revoked).build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_REQUEST_FAILED, "JWT signer revoked");
	}

	@Test
	public void testKidDoesNotMatchSigningCertificate() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(SealPPCredentialVaults.getValidVocesVault(version), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "JWT signature does not match locally computed signature");
	}

	@Test
	public void testUntrustedTokenSigner() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(SealPPCredentialVaults.getValidVocesVault(version), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.setKid("unknown").build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "Unable to get signing certificate");
	}

	@Test
	public void testJwtExpired() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());

		// expired 10 mins ago
		String token = tokenBuilder.setExpiration(new Date(System.currentTimeMillis() - 600000)).build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "JWT expired");
	}

	@Test
	public void testUnknownJWTIssuer() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());

		// expired 10 mins ago
		String token = tokenBuilder.setIssuer("unknown").build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unknown issuer");
	}

	@Ignore("SDS-6660")
	@Test
	public void testStsMissingInScope() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());

		// expired 10 mins ago
		String token = tokenBuilder.setScope("clear not-sosi-sts").build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Incoming JSON webtoken not scoped for this sts");
	}

	@Ignore("SDS-6660")
	@Test
	public void testAudienceScopeMismatch() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());

		// expired 10 mins ago
		String token = tokenBuilder.setScope("sosi-sts not-clear").build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Incoming JSON webtoken not scoped for audience");
	}

	@Ignore("SDS-6660")
	@Test
	public void testAudienceHasNoScopeConfiguration() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());

		// expired 10 mins ago
		String token = tokenBuilder.setScope("sosi-sts clear").build();

		final String audience = testModel.getCryptAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "JWT scope not configured for audience");
	}

	@Test
	public void testMissingAudience() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		JWTToIdentityTokenRequestDOMBuilder requestBuilder = new JWTToIdentityTokenRequestDOMBuilder() {
			protected void validateBeforeBuild() throws ModelException {
				// not failing on missing audience
			}
		};
		requestBuilder.setCPRNumberClaim(testModel.getPatientCpr().getValue());

		requestBuilder.setJWTToken(token);
		requestBuilder.setSigningVault(testModel.getValidWhitelistedVault());
		requestBuilder.setAudience(null);

		final Document requestDocument = requestBuilder.build();
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Audience is missing");
	}

	/**
	 * This is Testcase #T11 in the specification document
	 */
	@Test
	public void testUnknownAudience() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), "unknown audience", testModel.getPatientCpr().getValue());

		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Audience not configured");
	}

	/**
	 * This is Testcase #T16 in the specification document
	 */
	@Test
	public void testMocesJwtToken() {
		final X509Certificate mocesCert = SealPPCredentialVaults.getValidMocesWithCpr(version).getSystemCredentialPair().getCertificate();
		final CertificateInfo mocesCertInfo = new CertificateInfo(mocesCert);
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), mocesCertInfo, new CprNr("1802602810"));
		String token = tokenBuilder.build();

		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), null);

		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Token not issued to a citizen");
	}

	@Test
	public void testMalformedSubjectDN() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), "nameid-malformed", testModel.getPatientCpr());
		String token = tokenBuilder.build();

		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());

		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Invalid name: nameid");
	}

	@Test
	public void testMissingSubjectDN() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), (String) null, testModel.getPatientCpr());
		String token = tokenBuilder.build();

		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), testModel.getPatientCpr().getValue());

		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToIdentityTokenResponse assertionToIDCardResponse = getResponseModel(responseDoc);
		assertFault(assertionToIDCardResponse, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "JWT claim missing: certsubdn");
	}

	@Ignore("SDS-6660")
	@Test
	public void testMissingCprClaim() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), null);
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, null);

		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR claim missing");
	}

	@Ignore("SDS-6660")
	@Test
	public void testMalformedCprSamlClaim() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), null);
		String token = tokenBuilder.build();

		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "123456789");
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR-number must be 10 digits");
	}

	@Ignore("SDS-6660")
	@Test
	public void testMalformedCprInJWT() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), new CprNr("1234567890") {
			public String getValue() {
				return "123456789";
			}
		});
		String token = tokenBuilder.build();

		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), null);
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR-number must be 10 digits");
	}

	@Ignore("SDS-6660")
	@Test
	public void testValidRequestWithCprInJWT() throws Exception {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, null);
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFalse(response.getFaultString(), response.isFault());
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();

		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		assertEquals(audience, identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute("dk:gov:saml:attribute:SpecVer"));
		assertEquals("3", identityToken.getAssuranceLevel());

		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
	}

	@Ignore("SDS-6660")
	@Test
	public void testValidRequestWithCprInSamlClaim() throws Exception {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), null);
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, testModel.getPatientCpr().getValue());
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFalse(response.getFaultString(), response.isFault());
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();

		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		assertEquals(audience, identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute("dk:gov:saml:attribute:SpecVer"));
		assertEquals("3", identityToken.getAssuranceLevel());

		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
		
		NodeList samlNameIds = identityToken.getDOM().getElementsByTagName("saml:NameID");
		assertNotNull(samlNameIds);
		assertEquals(1, samlNameIds.getLength());
		Node samlNameId = samlNameIds.item(0);
		assertEquals("SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK", samlNameId.getTextContent());
		Node samlNameIdFormat = samlNameId.getAttributes().getNamedItem("Format");
		assertEquals(SAMLValues.NAMEID_FORMAT_X509_SUBJECT_NAME, samlNameIdFormat.getTextContent());
	}

	@Ignore("SDS-6660")
	@Test
	public void testCprInJWTPreferredToSamlClaim() throws Exception {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, "1234567890");
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFalse(response.getFaultString(), response.isFault());
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();

		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		assertEquals(audience, identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute("dk:gov:saml:attribute:SpecVer"));
		assertEquals("3", identityToken.getAssuranceLevel());

		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
	}

	@Ignore("SDS-6660")
	@Test
	public void testWrongCprSamlClaimInRequest() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), null);
		String token = tokenBuilder.build();

		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), testModel.getClearTextAudience(), "1234567890");
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR/PID mismatch");
	}

	@Ignore("SDS-6660")
	@Test
	public void testTrustedCprInJWTNotVerified() throws Exception {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), new CprNr("1234567890"));
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		Document requestDocument = buildRequestDocument(token, testModel.getValidWhitelistedVault(), audience, null);
		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertFalse(response.getFaultString(), response.isFault());
		assertSuccess(response, federation);

		assertNotNull(response.getIdentityToken());
		assertNull(response.getEncryptedIdentityTokenElement());

		final IdentityToken identityToken = response.getIdentityToken();
		assertEquals("1234567890", identityToken.getCpr());
		assertEquals(audience, identityToken.getAudienceRestriction());
		assertTrue(identityToken.getIssuer(), identityToken.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", identityToken.getAttribute("dk:gov:saml:attribute:SpecVer"));
		assertEquals("3", identityToken.getAssuranceLevel());
		assertHolderOfKey(identityToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertValidInTime(identityToken);
		
		NodeList samlNameIds = identityToken.getDOM().getElementsByTagName("saml:NameID");
		assertNotNull(samlNameIds);
		assertEquals(1, samlNameIds.getLength());
		Node samlNameId = samlNameIds.item(0);
		assertEquals("1234567890", samlNameId.getTextContent());
		Node samlNameIdFormat = samlNameId.getAttributes().getNamedItem("Format");
		assertEquals(SAMLValues.NAMEID_FORMAT_PERSISTENT, samlNameIdFormat.getTextContent());
	}

	@Ignore("SDS-6660")
	@Test
	public void testValidRequestWithReadProcurations() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
		requestBuilder.setOnBehalfOfClaim(CPR_1, Procuration);
		Document requestDocument = requestBuilder.build();

		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertSuccess(response, federation);

		final IdentityToken identityToken = response.getIdentityToken();
		assertNotNull(identityToken);
		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		assertTrue(FMK_READ + " missing", identityToken.getPrivileges().getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + CPR_1).contains(FMK_READ));
		assertNull(identityToken.getPrivileges().getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + CPR_2));
		assertEquals(audience, identityToken.getAudienceRestriction());
	}

	@Ignore("SDS-6660")
	@Test
	public void testValidRequestWithWriteProcuration() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
		requestBuilder.setOnBehalfOfClaim(CPR_2, Procuration);
		Document requestDocument = requestBuilder.build();

		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertSuccess(response, federation);

		final IdentityToken identityToken = response.getIdentityToken();
		assertNotNull(identityToken);
		assertEquals(testModel.getPatientCpr().getValue(), identityToken.getCpr());
		final BasicPrivileges privileges = identityToken.getPrivileges();
		assertNull(privileges.getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + CPR_1));
		assertTrue(FMK_WRITE + " missing", identityToken.getPrivileges().getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + CPR_2).contains(FMK_WRITE));
	}

	@Test
	public void testNoPrivilegesIfClaimsAreNotSupplied() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
		// requestBuilder.setProcurationByCprClaim(CPR_1);
		Document requestDocument = requestBuilder.build();

		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);
		assertSuccess(response, federation);

		final IdentityToken identityToken = response.getIdentityToken();
		assertNotNull(identityToken);
		assertNull(identityToken.getPrivileges());
	}

	@Test
	public void testErrorWhenClaimsNotFulfilled() {
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(testModel.getExternalTrustedIdp(), testModel.getPatientCertificateInfo(), testModel.getPatientCpr());
		String token = tokenBuilder.build();

		final String audience = testModel.getClearTextAudience();
		JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
		requestBuilder.setOnBehalfOfClaim("1111111119", Procuration);
		Document requestDocument = requestBuilder.build();

		final Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, "No privileges provided by procuration");
	}

	@Test
	public void testValidRequestWithInvalidSubjectRelationsClaim_returnsFault() {
		// Arrange
		JWTTokenBuilder tokenBuilder = getJwtTokenBuilder(
				testModel.getExternalTrustedIdp(),
				testModel.getPatientCertificateInfo(),
				testModel.getPatientCpr());
		String token = tokenBuilder.build();
		String audience = testModel.getClearTextAudience();
		JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(token, testModel.getValidWhitelistedVault(), audience, null);
		requestBuilder.setOnBehalfOfClaim("1111111120", ParentalCustody); // Wrong claim
		Document requestDocument = requestBuilder.build();

		// Act
		Document responseDoc = getResponseDocument(requestDocument);
		JWTToIdentityTokenResponse response = getResponseModel(responseDoc);

		// Assert
		assertFault(response, StsFaultActorValues.ITS, "The person is not in parental custody of the given person");
	}

	protected final Document buildRequestDocument(String bootstrapToken, CredentialVault signingVault, String appliesTo, String cprClaim) {
		JWTToIdentityTokenRequestDOMBuilder requestBuilder = createRequestBuilder(bootstrapToken, signingVault, appliesTo, cprClaim);
		final Document requestDocument = requestBuilder.build();
		LibertySignatureValidator validator = new LibertySignatureValidator(new EmptyCredentialVault(), requestDocument.getDocumentElement());
		validator.validateSignature();
		return requestDocument;
	}

	public JWTToIdentityTokenRequestDOMBuilder createRequestBuilder(String bootstrapToken,
																	CredentialVault signingVault,
																	String appliesTo,
																	String cprClaim) {
		JWTToIdentityTokenRequestDOMBuilder requestBuilder = new OIOSAMLFactory().createJWTToIdentityTokenRequestDOMBuilder();
		requestBuilder.setCPRNumberClaim(cprClaim);

		requestBuilder.setJWTToken(bootstrapToken);
		requestBuilder.setSigningVault(signingVault);
		requestBuilder.setAudience(appliesTo);
		return requestBuilder;
	}

	public static Document getResponseDocument(TestClient client, Document requestDocument) {
		return client.execute(XmlUtil.node2String(requestDocument, false, false)).document;
	}

	protected final Document getResponseDocument(Document requestDocument) {
		return getTestClient().execute(XmlUtil.node2String(requestDocument, false, false)).document;
	}

	protected final JWTToIdentityTokenResponse getResponseModel(Document responseDoc) {
		return new OIOSAMLFactory().createJWTToIdentityTokenResponseModelBuilder().build(responseDoc);
	}

	protected abstract TestClient getTestClient();

}
