/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.nbo;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.OIOSAMLAssertionToIDCardResponse;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.model.SubjectSerialNumber;
import dk.sosi.sts.server.*;
import dk.sosi.sts.server.facade.*;
import dk.sosi.sts.server.test.StsCredentialVaults;
import dk.sosi.sts.server.test.TestUtil;
import dk.sosi.sts.store.SimulatedMocesStore.SimulatedMoces;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.security.cert.X509Certificate;
import java.util.Date;

import static org.junit.Assert.*;

/**
 * Created by Kent on 24-02-2017.
 */
public class NBOEmbeddedTest extends NboWithHomemadeRequestsTest {
    private static final String NO_ROLE = "urn:dk:healthcare:no-role";

    protected SOSIFactory sosiFactory;
    TestClientHelper testClientHelper;
    private MockServicesImpl mockServices;

    public NBOEmbeddedTest() {
        super(true);
    }

    @Before
    public void setUp() {
    	mockServices = new MockServicesImpl(version);
        sosiFactory = mockServices.buildSosiFactory(true);

        final NboRequestHandler nboRequestHandler = new NboRequestHandler();
        nboRequestHandler.setBlacklistedITSystemName(TestUtil.STS_IT_SYSTEM_NAME);
        nboRequestHandler.setSosiFactory(sosiFactory);
        nboRequestHandler.setCrlService(mockServices);
        nboRequestHandler.setStsStatus(new StsStatus());
        nboRequestHandler.setNboConfig(nboConfig);
        nboRequestHandler.setUserValidationService(mockServices.getUserValidationService());

        testClientHelper = new TestClientHelper(nboRequestHandler);
    }
    
    
    @Override
    boolean isIntegrationTest() {
    	return false;
    }
    
    @Test
    public void testInvalidIdp() throws Exception {
        nboConfig.setTrustedVault(StsCredentialVaults.getExternalTrustedInvalidIdp(version));
        X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
        final NemLoginTokenBuilder nemLoginTokenBuilder = getNemLoginTokenBuilder(cpr, userCertificate, true, StsCredentialVaults.getExternalTrustedIdp(version));
        Element signedAssertion = nemLoginTokenBuilder.expires(new Date(0)).build().getDOM().getDocumentElement();
        Document requestDocument = buildRequestDocument(signedAssertion, actualAutorizationCode, "7170", getClientSystemVault());

        final OIOSAMLAssertionToIDCardResponse assertionToIDCardResponse = callNbo(requestDocument);

        assertTrue(assertionToIDCardResponse.isFault());
        assertEquals(StsFaultActorValues.NBO, assertionToIDCardResponse.getFaultActor());
        assertTrue(assertionToIDCardResponse.getFaultString(),assertionToIDCardResponse.getFaultString().contains("The certificate that signed the security token is not trusted"));
    }

    @Test
    public void testOIOSAML_RequestInvalidXML() {
        final OIOSAMLAssertionToIDCardResponse assertionToIDCardResponse = callNbo(XmlUtil.createEmptyDocument());
        assertTrue(assertionToIDCardResponse.isFault());
        assertEquals(StsFaultActorValues.STS, assertionToIDCardResponse.getFaultActor());
        assertEquals(StsFaultCodeValues.WST_INVALID_REQUEST, assertionToIDCardResponse.getFaultCode());
        assertTrue(assertionToIDCardResponse.getFaultString(), assertionToIDCardResponse.getFaultString().contains("Unable to parse"));
        assertTrue(assertionToIDCardResponse.getFaultString(), assertionToIDCardResponse.getFaultString().contains(StsFaultStringValues.WST_INVALID_REQUEST));
    }
    
    /**
     * In this example the user has two authorizations in the authorization register. The id card request is
     * made for none of the authorizations. 
     * 
     * As no limitations are made in the request, the call must fail because STS will not choose one at random for us.
     */
    @Test
    public void testMultipleAuthorizationsNoneSpecified() throws Exception{
        X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
        Element signedAssertion = getIdpSignedOIOSAMLAssertion("1802609999", userCertificate, true);
        Document requestDocument = buildRequestDocument(signedAssertion, null, null, getClientSystemVault());

        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
        assertTrue("Expected and error as user has multiple authorizations in authorization registry and none specified in request", response.isFault());
        assertEquals(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
        assertTrue(response.getFaultString().startsWith("Authentication failed: multiple authorizations found"));
    }
    
    /**
     * In this example the user has two authorizations in the authorization register. The id card request is
     * made for none of the authorizations. 
     * 
     * A failed limitation have been made with an unknown authorization
     */
    @Test
    public void testMultipleAuthorizationsUnknownAuthorizationCodeSpecified() throws Exception{
        X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
        Element signedAssertion = getIdpSignedOIOSAMLAssertion("1802609999", userCertificate, true);
        Document requestDocument = buildRequestDocument(signedAssertion, "MADEUP", null, getClientSystemVault());

        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
        assertTrue("Expected and error as user has multiple authorizations in authorization registry and none specified in request", response.isFault());
        assertEquals(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
        assertTrue(response.getFaultString().startsWith("Authentication failed: authorization not found"));
    }
    
    /**
     * In this example the user has two authorizations in the authorization register. The id card request is
     * made for none of the authorizations. 
     * 
     * As failed limitation have been made with an unknown role
     */
    @Test
    public void testMultipleAuthorizationsUnknownRoleSpecified() throws Exception{
        X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
        Element signedAssertion = getIdpSignedOIOSAMLAssertion("1802609999", userCertificate, true);
        Document requestDocument = buildRequestDocument(signedAssertion, null, "MADEUP", getClientSystemVault());

        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
        assertTrue("Expected and error as user has multiple authorizations in authorization registry and none specified in request", response.isFault());
        assertEquals(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
        assertTrue(response.getFaultString().startsWith("Authentication failed: multiple authorizations found"));
    }
    
    /**
     * In this example the user has two authorizations in the authorization register and a national role. The id card request is
     * made for no authorizations or national roles.
     * 
     * The role is specified with fixed value 'urn:dk:healthcare:no-role' which will make UserValidationService ignore lookup for authorizations and national roles.
     */
    @Test
    public void testIgnorePrivateAuthorizationsAndNationalRoleWithNoRoleValue() throws Exception{
        X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
        mockServices.getExistingNationalRoleService().put(new SubjectSerialNumber("UI:DK-O:G:8d3fa047-c77e-47e4-bdd2-e91488610ce6"), RoleDefinitionServiceStub.LAEGE_SEKRETAER);
        Element signedAssertion = getIdpSignedOIOSAMLAssertion("1802609999", userCertificate, true);
        Document requestDocument = buildRequestDocument(signedAssertion, null, NO_ROLE, getClientSystemVault());

        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
        assertFalse(response.getFaultString(), response.isFault());

        assertResult(response, "1802609999", null, NO_ROLE, true);
    }
    
    /**
     * In this example the user has two authorizations in the authorization register. The id card request is
     * made for none of the authorizations. 
     * 
     * A failed limitation have been made with an unknown authorization and an unknown role
      */
    @Test
    public void testMultipleAuthorizationsUnknownAuthorizationCodeAndUnknownRoleSpecified() throws Exception{
        X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
        Element signedAssertion = getIdpSignedOIOSAMLAssertion("1802609999", userCertificate, true);
        Document requestDocument = buildRequestDocument(signedAssertion, "MADEUP", "MADEUP", getClientSystemVault());

        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
        assertTrue("Expected and error as user has multiple authorizations in authorization registry and none specified in request", response.isFault());
        assertEquals(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
        assertTrue(response.getFaultString().startsWith("Authentication failed: authorization not found"));
    }
    
    /**
     * In this example the user has two authorizations in the authorization register. The id card request is
     * made for none of the authorizations. 
     * 
     * A failed limitation have been made with an unknown authorization and a known role
      */
    @Test
    public void testMultipleAuthorizationsUnknownAuthorizationCodeAndValidRoleSpecified() throws Exception{
        X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
        Element signedAssertion = getIdpSignedOIOSAMLAssertion("1802609999", userCertificate, true);
        Document requestDocument = buildRequestDocument(signedAssertion, "MADEUP", "1234", getClientSystemVault());

        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
        assertTrue("Expected and error as user has multiple authorizations in authorization registry and none specified in request", response.isFault());
        assertEquals(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
        assertTrue(response.getFaultString().startsWith("Authentication failed: authorization not found"));
    }
    
    /**
     * In this example the user has two authorizations in the authorization register. The id card request is
     * made for one of the authorizations. 
     * 
     * A limitation have been made with an known authorization and an unknown role (could be null)
      */
    @Test
    public void testMultipleAuthorizationsKnownAuthorizationCodeAndUnknownRoleSpecified() throws Exception{
        X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
        Element signedAssertion = getIdpSignedOIOSAMLAssertion("1802609999", userCertificate, true);
        Document requestDocument = buildRequestDocument(signedAssertion, "ABCD", "MADEUP", getClientSystemVault());

        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
        assertFalse(response.getFaultString(), response.isFault());

        assertResult(response, "1802609999", "ABCD", "1234", true);
    }
    
    /**
     * In this example the user has two authorizations in the authorization register. The id card request is
     * made for one of the authorizations. 
     * 
     * A limitation have been made with a known authorization and a known role
      */
    @Test
    public void testMultipleAuthorizationsKnownAuthorizationCodeAndKnownRoleSpecified() throws Exception{
        X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
        Element signedAssertion = getIdpSignedOIOSAMLAssertion("1802609999", userCertificate, true);
        Document requestDocument = buildRequestDocument(signedAssertion, "ABCD", "1234", getClientSystemVault());

        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
        assertFalse(response.getFaultString(), response.isFault());

        assertResult(response, "1802609999", "ABCD", "1234", true);
    }

    @Ignore("SDS-6660")
    @Test
    public void testWhenCprIsTrustedWithSimulatedMoces() throws Exception {
    	SimulatedMoces susanneAndersen = SimulatedMoces.SUSANNE_ANDERSEN;
    	
		X509Certificate userCertificate = susanneAndersen.getPoces();    
        
        NemLoginTokenBuilder assertionBuilder = getNemLoginTokenBuilder(susanneAndersen.getCpr(), userCertificate, true, StsCredentialVaults.getExternalTrustedIdp(version));
        Element signedAssertion = assertionBuilder.setSimulatedMoces(true).build().getDOM().getDocumentElement();
        
        Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());
        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

        assertFalse(response.getFaultString(), response.isFault());
        assertResult(response, susanneAndersen.getCpr(), null, "IGNORED", true);
    }

    @Ignore("SDS-6660")
    @Test
    public void testWhenCprIsUntrustedWithSimulatedMoces() throws Exception {
    	SimulatedMoces susanneAndersen = SimulatedMoces.SUSANNE_ANDERSEN;
    	
		X509Certificate userCertificate = susanneAndersen.getPoces();    
        
    	NemLoginTokenBuilder assertionBuilder = getNemLoginTokenBuilder(susanneAndersen.getCpr(), userCertificate, true, StsCredentialVaults.getExternalTrustedIdpCprNotTrusted(version));
        Element signedAssertion = assertionBuilder.setSimulatedMoces(true).build().getDOM().getDocumentElement();
        
        Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());
        OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

        assertTrue(response.isFault());
        assertEquals(response.getFaultString(), StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
        assertEquals(response.getFaultString(), StsFaultActorValues.CVRRIDCPR, response.getFaultActor());
        assertEquals("Authentication failed: cvrrid-cpr mismatch ["+susanneAndersen.getSubjectSerialNumberValue()+","+susanneAndersen.getCpr()+"]", response.getFaultString());
    }

    @Override
    protected TestClient getTestClient() {
        return testClientHelper;
    }
}
