/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.nbo;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.LibertySignatureValidator;
import dk.sosi.seal.model.OIOSAMLAssertion;
import dk.sosi.seal.model.OIOSAMLAssertionToIDCardResponse;
import dk.sosi.seal.model.SignatureConfiguration.KeyInfoType;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.constants.WSSETags;
import dk.sosi.seal.model.dombuilders.OIOSAMLAssertionToIDCardRequestDOMBuilder;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.EmptyCredentialVault;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfo.CertificateType;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.model.RoleDefinition;
import dk.sosi.sts.server.*;
import dk.sosi.sts.server.facade.NemLoginTokenBuilder;
import dk.sosi.sts.server.facade.RoleDefinitionServiceStub;
import dk.sosi.sts.server.facade.TestClient;
import dk.sosi.sts.server.test.*;
import dk.sosi.sts.store.SimulatedMocesStore.SimulatedMoces;
import org.junit.Assume;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import javax.xml.ws.Holder;
import java.io.IOException;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.Date;
import java.util.Properties;

import static org.junit.Assert.*;

/**
 * User: ksr Date: 22-10-13 Time: 14:58
 */
@RunWith(Parameterized.class)
public abstract class NboWithHomemadeRequestsTest {
	private Federation federation;
	protected NboConfig nboConfig;

	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;
	
	@Parameter(2)
	public String cpr;
	
	@Parameter(3)
	public String actualAutorizationCode;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();
		TestIdentity oces3Identity = TestIdentityProvider.getIdentityWithTrustedCvrAndOneAuthorization(Version.OCES3);

		builder.newTest("OCES3").add(Version.OCES3, oces3Identity.getCpr(), oces3Identity.getAuthorizations().get(0).getCode());

		return builder.build();
	}

	public NboWithHomemadeRequestsTest(boolean isEmbeddedTest) {
		federation = TestUtil.createTestFederation(isEmbeddedTest);
	}

	@Before
	public void initNboConfig() {
		nboConfig = new NboConfig();
		nboConfig.setTrustedVault(StsCredentialVaults.getAllTrustedCertificates(version));
		nboConfig.setIdCardDuration(86400000);
		nboConfig.setFuzzyTime(300000);
		final X509Certificate x509Certificate = StsCredentialVaults.getExternalTrustedIdp(version).getSystemCredentialPair().getCertificate();
		final String ssn = OcesInfoParser.parseCertificate(x509Certificate).getSubjectSerialNumber().getValue();
		nboConfig.setCprTrustCertificates(Arrays.asList(ssn));
	}

	abstract boolean isIntegrationTest();

	@Test
	public void testOIOSAMLRequest_InvalidEnvelopeSignature() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(cpr, userCertificate, true);

		Document requestDocument = buildRequestDocument(signedAssertion, actualAutorizationCode, "7170", getClientSystemVault());
		Element header = XmlUtil.getFirstChildElementNS(requestDocument.getDocumentElement(), "http://schemas.xmlsoap.org/soap/envelope/", "Header");
		Element security = XmlUtil.getFirstChildElementNS(header, "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd", "Security");
		Element signature = XmlUtil.getFirstChildElementNS(security, "http://www.w3.org/2000/09/xmldsig#", "Signature");
		Element signatureValue = XmlUtil.getFirstChildElementNS(signature, "http://www.w3.org/2000/09/xmldsig#", "SignatureValue");

		// Node signatureValue = XmlUtil.findChildElement(header, "Security");
		signatureValue.getFirstChild().setNodeValue(signatureValue.getFirstChild().getNodeValue() + "-break-");

		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertTrue(response.isFault());
		assertEquals(StsFaultCodeValues.WST_INVALID_REQUEST, response.getFaultCode());
		assertEquals(StsFaultActorValues.NBO, response.getFaultActor());
		assertTrue(response.getFaultString(), response.getFaultString().contains("Unable to validate envelope signature"));
	}

	@Test
	public void testOIOSAMLRequest_InvalidEnvelopeCertificate() throws Exception {

		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(cpr, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, actualAutorizationCode, "7170", SealPPCredentialVaults.getExpiredVocesVault(version));
		OIOSAMLAssertionToIDCardResponse assertionToIDCardResponse = callNbo(requestDocument);

		assertTrue(assertionToIDCardResponse.isFault());
		assertEquals(assertionToIDCardResponse.getFaultCode(), StsFaultCodeValues.WST_FAILED_AUTHENTICATION);
		assertEquals(assertionToIDCardResponse.getFaultActor(), StsFaultActorValues.NBO);
		assertTrue(assertionToIDCardResponse.getFaultString().contains("Certificate expired or not yet valid"));
	}

	@Test
	public void testOIOSAML_RequestInvalidTime() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getNemLoginTokenBuilder(cpr, userCertificate, true, StsCredentialVaults.getExternalTrustedIdp(version)).expires(new Date(0)).build().getDOM().getDocumentElement();
		Document requestDocument = buildRequestDocument(signedAssertion, actualAutorizationCode, "7170", getClientSystemVault());

		final OIOSAMLAssertionToIDCardResponse assertionToIDCardResponse = callNbo(requestDocument);

		assertTrue(assertionToIDCardResponse.isFault());
		assertTrue(assertionToIDCardResponse.getFaultString(), assertionToIDCardResponse.getFaultString().contains("token no longer valid"));
		assertEquals(assertionToIDCardResponse.getFaultActor(), StsFaultActorValues.NBO);
	}

	/**
	 * In this example the user has a single authorization in the authorization
	 * register. The id card request is made for that one specifik authorization.
	 */
	@Test
	public void userWithOneSpecificAuthorization() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(cpr, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, actualAutorizationCode, "7170", getClientSystemVault());

		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, cpr, actualAutorizationCode, "7170", true);
	}

	/**
	 * In this example the user has a single authorization in the authorization
	 * register. The id card request is made with only a limitations on the
	 * role/education code. This is only possible because the user has one and only
	 * one authorization with the "Doctor" role.
	 */
	@Test
	public void userWithOneDoctorAuthorization() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(cpr, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, "7170", getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, cpr, actualAutorizationCode, "7170", true);
	}

	/**
	 * In this example the user has a single authorization in the authorization
	 * register. The id card request is made without a limitations on the
	 * authorization that will be used. This is only possible because the user has
	 * one and only one authorization.
	 */
	@Ignore("SDS-6660")
	@Test
	public void userWithOneAuthorization() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(cpr, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());
		new LibertySignatureValidator(new EmptyCredentialVault(), requestDocument.getDocumentElement()).validateSignature();

		String requestString = XmlUtil.node2String(requestDocument, false, false);
		requestDocument = XmlUtil.readXml(new Properties(), requestString, false);
		new LibertySignatureValidator(new EmptyCredentialVault(), requestDocument.getDocumentElement()).validateSignature();

		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, cpr, actualAutorizationCode, "7170", true);
	}

	@Test
	public void testSpecifyWrongAuthorization() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(cpr, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, actualAutorizationCode, "1111", getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
		assertTrue(response.isFault());
		assertTrue(response.getFaultString(), response.getFaultString().contains("authorization not found"));
		assertEquals(StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
		assertEquals(StsFaultActorValues.AUTHORIZATION, response.getFaultActor());
	}

	/**
	 * In this example the user has two authorizations in the authorization
	 * register. The id card request is made for one specifik authorization. If no
	 * limitations where made in the request, the call would fail because STS will
	 * not choose one at random for us.
	 */
	@Test
	public void userWithSeveralAuthorization() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(cpr, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, actualAutorizationCode, "7170", getClientSystemVault());

		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, cpr, actualAutorizationCode, "7170", true);
	}

	/**
	 * In this example the user has no authorization in the authorization register.
	 * The id card request is made without a limitations on the authorization that
	 * will be used. The returned id card does not have an authorization in it, and
	 * it's usage might be limited.
	 */
	@Ignore("SDS-6660")
	@Test
	public void userWithNoAuthorization() throws Exception {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(testIdentity.getCpr(), testIdentity.getCredentialVault().getSystemCredentialPair().getCertificate(), true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());

		NodeList binarySecurityTokenTagsIRequest = TagUtil.getElementsByTagNameNS(requestDocument.getDocumentElement(), WSSETags.binarySecurityToken);
		assertEquals("Forventede ikke BinarySecurityToken i request", 0, binarySecurityTokenTagsIRequest.getLength());

		Holder<Document> responseHolder = new Holder<>();
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument, responseHolder);

		assertFalse(response.getFaultString(), response.isFault());
		assertResult(response, testIdentity.getCpr(), null, "IGNORED", true);

		assertNotNull(responseHolder.value);

		NodeList binarySecurityTokenTagsIResponse = TagUtil.getElementsByTagNameNS(responseHolder.value.getDocumentElement(), WSSETags.binarySecurityToken);
		assertEquals("Forventede ikke BinarySecurityToken i response", 0, binarySecurityTokenTagsIResponse.getLength());

	}

	/**
	 * In this example the user has no authorization in the authorization register.
	 * The id card request is made without a limitations on the authorization that
	 * will be used. The returned id card does not have an authorization in it, and
	 * it's usage might be limited.
	 */
	@Test
	public void userWithNoAuthorizationUsingSecurityTokenReference() throws Exception {
		Assume.assumeTrue(isIntegrationTest());

		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(testIdentity.getCpr(), testIdentity.getCredentialVault().getSystemCredentialPair().getCertificate(), true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault(), KeyInfoType.WSS_REFERENCE, "STS tester");

		NodeList binarySecurityTokenTagsIRequest = TagUtil.getElementsByTagNameNS(requestDocument.getDocumentElement(), WSSETags.binarySecurityToken);
		assertEquals("Forventede at BinarySecurityToken var angivet i request", 1, binarySecurityTokenTagsIRequest.getLength());

		Holder<Document> responseHolder = new Holder<>();
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument, responseHolder);

		assertFalse(response.getFaultString(), response.isFault());
		assertResult(response, testIdentity.getCpr(), null, "IGNORED", true);

		assertNotNull(responseHolder.value);

		NodeList binarySecurityTokenTagsIResponse = TagUtil.getElementsByTagNameNS(responseHolder.value.getDocumentElement(), WSSETags.binarySecurityToken);
		assertEquals("Forventede at BinarySecurityToken var angivet i response", 1, binarySecurityTokenTagsIResponse.getLength());
	}

	@Test
	public void testInputWithMissingRole() throws Exception {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(testIdentity.getCpr(), testIdentity.getCredentialVault().getSystemCredentialPair().getCertificate(), true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, null, getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertTrue(response.isFault());
		assertEquals(StsFaultCodeValues.WST_INVALID_REQUEST, response.getFaultCode());
		assertEquals(StsFaultActorValues.NBO, response.getFaultActor());
		assertTrue(response.getFaultString(), response.getFaultString().contains("EducationCode must be specified"));

	}

	@Ignore("SDS-6660")
	@Test
	public void testAuthorizedUserInputWithMissingRole() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(cpr, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, null, getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertFalse(response.getFaultString(), response.isFault());
		assertResult(response, cpr, actualAutorizationCode, "7170", true);
	}

	@Ignore("SDS-6660")
	@Test
	public void noCertificateInRequest() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(cpr, userCertificate, false);
		Document requestDocument = buildRequestDocument(signedAssertion, actualAutorizationCode, "7170", getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, cpr, actualAutorizationCode, "7170", false);
	}

	@Ignore("SDS-6660")
	@Test
	public void testWrongCprWhenCprIsTrusted() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion("1602761505", userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertFalse(response.getFaultString(), response.isFault());
		assertResult(response, "1602761505", null, "IGNORED", true);
	}

	@Test
	public void testWrongCprWhenCprIsUntrusted() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		NemLoginTokenBuilder assertionBuilder = getNemLoginTokenBuilder("1602761505", userCertificate, true, StsCredentialVaults.getExternalTrustedIdpCprNotTrusted(version));
		Element signedAssertion = assertionBuilder.build().getDOM().getDocumentElement();
		Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertTrue(response.isFault());
		assertEquals(response.getFaultString(), StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
		assertEquals(response.getFaultString(), StsFaultActorValues.UUIDCPR, response.getFaultActor());
		assertTrue(response.getFaultString(), response.getFaultString().contains("-cpr mismatch"));
	}

	@Ignore("SDS-6660")
	@Test
	public void blankCprInRequest() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion("", userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, cpr, actualAutorizationCode, "7170", true);
	}

	@Ignore("SDS-6660")
	@Test
	public void noCprInRequest() throws Exception {
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(null, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);
		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, cpr, actualAutorizationCode, "7170", true);
	}

	@Ignore("SDS-6660")
	@Test
	public void testNoNationalRole() throws Exception {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		X509Certificate userCertificate = testIdentity.getCredentialVault().getSystemCredentialPair().getCertificate();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(null, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, testIdentity.getCpr(), null, "IGNORED", true);
	}

	@Ignore("SDS-6660")
	@Test
	public void testNationalRoleCalculated() throws Exception {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithOneNationalRole(version);
		String role = testIdentity.getNationalRoles().get(0).getPublicName();
		X509Certificate userCertificate = testIdentity.getCredentialVault().getSystemCredentialPair().getCertificate();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(null, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, "IGNORED", getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, testIdentity.getCpr(), null, role, true);
	}

	@Ignore("SDS-6660")
	@Test
	public void testNationalRoleVerified() throws Exception {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithOneNationalRole(version);
		String role = testIdentity.getNationalRoles().get(0).getPublicName();
		X509Certificate userCertificate = testIdentity.getCredentialVault().getSystemCredentialPair().getCertificate();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(null, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, role, getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, testIdentity.getCpr(), null, role, true);
	}

	/*
	Rollen er angivet uden kode.
	 */
	@Ignore("SDS-6660")
	@Test
	public void testNationalRoleVerifiedNoCode() throws Exception {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithOneNationalRole(version);
		RoleDefinition role = testIdentity.getNationalRoles().get(0);
		X509Certificate userCertificate = testIdentity.getCredentialVault().getSystemCredentialPair().getCertificate();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(null, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, role.getPublicNameNoCode(), getClientSystemVault());
		System.out.println(XmlUtil.node2String(requestDocument, false, false));
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertFalse(response.getFaultString(), response.isFault());

		assertResult(response, testIdentity.getCpr(), null, role.getPublicName(), true);
	}

	@Test
	public void testNonExistingNationalRoleDisallowed() throws Exception {
		TestIdentity testIdentity = TestIdentityProvider.getIdentityWithNoAuthorizationsOrNationalRoles(version);
		String nationalRole = RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(null, testIdentity.getCredentialVault().getSystemCredentialPair().getCertificate(), true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, nationalRole, getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertTrue(response.isFault());
		assertEquals(response.getFaultString(), StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
		assertEquals(response.getFaultString(), StsFaultActorValues.AUTHORIZATION, response.getFaultActor());
		assertTrue(response.getFaultString(), response.getFaultString().contains("Authentication failed: user does not have the requested role (" + nationalRole + ")"));

	}

	@Test
	public void testTrustOfCvrIgnoredForNBO() throws Exception {
		X509Certificate userCertificate = getUserWithNoAuthorizationTrustedCvr();
		String nationalRole = RoleDefinitionServiceStub.LAEGE_SEKRETAER.getPublicName();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion(null, userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, null, nationalRole, getClientSystemVault());
		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertTrue(response.isFault());
		assertEquals(response.getFaultString(), StsFaultCodeValues.WST_FAILED_AUTHENTICATION, response.getFaultCode());
		assertEquals(response.getFaultString(), StsFaultActorValues.AUTHORIZATION, response.getFaultActor());
		assertTrue(response.getFaultString(), response.getFaultString().contains("Authentication failed: user does not have the requested role (" + nationalRole + ")"));
	}

	@Test
	public void testWithBlacklistedItSystemName_ReturnsFault() throws Exception{
		X509Certificate userCertificate = getUserWithCprAndSingleAuthorization();
		Element signedAssertion = getIdpSignedOIOSAMLAssertion("1802609999", userCertificate, true);
		Document requestDocument = buildRequestDocument(signedAssertion, "ABCD", "1234", getClientSystemVault(), null, TestUtil.STS_IT_SYSTEM_NAME);

		OIOSAMLAssertionToIDCardResponse response = callNbo(requestDocument);

		assertTrue(response.isFault());
		assertEquals(response.getFaultString(), StsFaultCodeValues.WST_BAD_REQUEST, response.getFaultCode());
		assertEquals(response.getFaultString(), StsFaultActorValues.NBO, response.getFaultActor());
		assertEquals(StsFaultStringValues.WST_BAD_REQUEST + ": Invalid ITSystemName", response.getFaultString());
	}

	protected Document buildRequestDocument(Element signedOIOSAMLAssertion, String code, String role, GenericCredentialVault signingVault) {
		return this.buildRequestDocument(signedOIOSAMLAssertion, code, role, signingVault, null, "STS tester");
	}

	protected Document buildRequestDocument(Element signedOIOSAMLAssertion, String code, String role, GenericCredentialVault signingVault, KeyInfoType keyInfoType, String itSystemName) {
		OIOSAMLAssertionToIDCardRequestDOMBuilder requestBuilder = new OIOSAMLFactory().createOIOSAMLAssertionToIDCardRequestDOMBuilder();
		OIOSAMLAssertion assertion = new OIOSAMLAssertion(signedOIOSAMLAssertion);
		requestBuilder.setOIOSAMLAssertion(assertion);
		requestBuilder.setSigningVault(signingVault, keyInfoType);
		requestBuilder.setITSystemName(itSystemName);
		requestBuilder.setUserGivenName("Test");
		requestBuilder.setUserSurName("Testesen");
		requestBuilder.setUserEducationCode(role);
		requestBuilder.setUserAuthorizationCode(code);
		return requestBuilder.build();
	}

	/**
	 * Retrieve an OIOSamlAssertion, signed by someone trusted by STS (the NemLog-In
	 * IdP) In this test we build it ourselves (exploiting that the test-STS trust
	 * another testIdp besides NemLog-In)
	 */
	protected Element getIdpSignedOIOSAMLAssertion(String cpr, X509Certificate userCertificate, boolean includeCertificate) throws IOException {
		NemLoginTokenBuilder assertionBuilder = getNemLoginTokenBuilder(cpr, userCertificate, includeCertificate, StsCredentialVaults.getExternalTrustedIdp(version));
		return assertionBuilder.build().getDOM().getDocumentElement();
	}

	protected NemLoginTokenBuilder getNemLoginTokenBuilder(String cpr, X509Certificate userCertificate, boolean includeCertificate, CredentialVault trustedIdp) throws IOException {
		OcesInfo parseCertificate = OcesInfoParser.parseCertificate(userCertificate);

		if (parseCertificate.getCertificateType() == CertificateType.POCES) {
			SimulatedMoces simulation = SimulatedMoces.getByCertificate(userCertificate);
			// Create simulated MOCES
			return getNemLoginTokenBuilder(simulation.getCpr(), simulation.getGivenName(), simulation.getSurName(), null, simulation.getCvr(), simulation.getOrgName(), userCertificate, includeCertificate, trustedIdp).setSimulatedMoces(true);

		} else {
			// The identification of the organization. Should match information in the
			// certificate
			return getNemLoginTokenBuilder(cpr, parseCertificate.getCvr(), userCertificate, includeCertificate, trustedIdp);
		}

	}

	protected NemLoginTokenBuilder getNemLoginTokenBuilder(String cpr, String cvr, X509Certificate userCertificate, boolean includeCertificate, CredentialVault trustedIdp) throws IOException {
		return getNemLoginTokenBuilder(cpr, "Test", "Testesen", "test.testesen@nsi.dk", cvr, "Statens Serum Institut", userCertificate, includeCertificate, trustedIdp);
	}

	protected NemLoginTokenBuilder getNemLoginTokenBuilder(String cpr, String givenName, String surName, String email, String cvr, String orgName, X509Certificate userCertificate, boolean includeCertificate, CredentialVault trustedIdp) throws IOException {
		NemLoginTokenBuilder assertionBuilder = new NemLoginTokenBuilder();

		assertionBuilder.setUserCpr(cpr);
		assertionBuilder.setUserName(givenName, surName);
		assertionBuilder.setUserEmail(email);
		assertionBuilder.setCvr(cvr);
		assertionBuilder.setOrgName(orgName);

		// a credential Vault containing the private key for an identity provider
		// trusted by STS (normally NemLog-In IdP)
		assertionBuilder.setSigningVault(trustedIdp);
		assertionBuilder.setIssuer("STS tester/issuer");
		assertionBuilder.setAudienceRestriction("STS tester/audience");
		assertionBuilder.setRecipientURL("STS tester/recipientUrl");

		if (includeCertificate) {
			assertionBuilder.setUserCertificate(userCertificate);
		} else {
			assertionBuilder.setCertificateInfo(new CertificateInfo(userCertificate));
		}
		return assertionBuilder;
	}

	protected final OIOSAMLAssertionToIDCardResponse callNbo(Document requestDocument) {
		return callNbo(requestDocument, null);
	}

	protected final OIOSAMLAssertionToIDCardResponse callNbo(Document requestDocument, Holder<Document> response) {
		final Document responseDoc = sendRequest(requestDocument);
		if (response != null) {
			response.value = responseDoc;
		}
		return parseResponse(responseDoc);
	}

	protected final Document sendRequest(Document requestDocument) {
		return getTestClient().execute(XmlUtil.node2String(requestDocument, false, false)).document;
	}

	protected OIOSAMLAssertionToIDCardResponse parseResponse(Document responseDoc) {
		return new OIOSAMLFactory().createOIOSAMLAssertionToIDCardResponseModelBuilder().build(responseDoc);
	}

	protected void assertResult(OIOSAMLAssertionToIDCardResponse assertionToIDCardResponse, String cprExpected, String authCode, String role, boolean certHashExpected) {
		UserIDCard idCard = (UserIDCard) assertionToIDCardResponse.getIDCard();
		assertNotNull("No user information found", idCard.getUserInfo());
		assertEquals(cprExpected, idCard.getUserInfo().getCPR());
		assertEquals("Incorrect authorization code", authCode, idCard.getUserInfo().getAuthorizationCode());
		assertEquals("Incorrect education code", role, idCard.getUserInfo().getRole());
		if (certHashExpected) {
			assertNotNull("certHash was expected", idCard.getCertHash());
		} else {
			assertNull("No certHash was expected", idCard.getCertHash());
		}

		assertionToIDCardResponse.validateSignature();
		assertionToIDCardResponse.validateSignatureAndTrust(federation);
		final Date now = new Date();

		assertNotNull(assertionToIDCardResponse.getCreated());
		assertNotNull(assertionToIDCardResponse.getExpires());

		assertTrue(assertionToIDCardResponse.getCreated().before(now));
		assertEquals(assertionToIDCardResponse.getExpires().getTime() - assertionToIDCardResponse.getCreated().getTime(), nboConfig.getIdCardDuration());

	}

	protected abstract TestClient getTestClient();

	protected X509Certificate getUserWithCprAndSingleAuthorization() throws Exception {
		return CertificateStore.getValidMoces(version);
	}

	private X509Certificate getUserWithNoAuthorizationTrustedCvr() throws Exception {
		return StsCredentialVaults.getRandomMocesWithCvrTrustedNationalRole(version).getSystemCredentialPair().getCertificate();
	}

	protected GenericCredentialVault getClientSystemVault() {
		return CertificateStore.getValidFocesVault(version);
	}

}
