/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.facade.sbo;

import static dk.sosi.sts.server.facade.OIOWSTrustResponseValidation.assertFault;
import static dk.sosi.sts.server.facade.OIOWSTrustResponseValidation.assertHolderOfKey;
import static dk.sosi.sts.server.facade.OIOWSTrustResponseValidation.assertSuccess;
import static java.time.temporal.ChronoUnit.HOURS;
import static java.time.temporal.ChronoUnit.MINUTES;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.BootstrapOIOSAMLAssertion;
import dk.sosi.seal.model.JWTToOIOSAMLAssertionResponse;
import dk.sosi.seal.model.LibertySignatureValidator;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.OIOBootstrapToken;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.dombuilders.JWTToOIOSAMLAssertionRequestDOMBuilder;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultException;
import dk.sosi.seal.vault.EmptyCredentialVault;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.StsFaultActorValues;
import dk.sosi.sts.server.StsFaultCodeValues;
import dk.sosi.sts.server.facade.DefaultSamlTokenTestModel;
import dk.sosi.sts.server.facade.SamlTokenTestModel;
import dk.sosi.sts.server.facade.TestClient;
import dk.sosi.sts.server.facade.jwt.JWTTokenBuilder;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.StsCredentialVaults;

import dk.sosi.sts.server.test.TestUtil;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.security.PrivateKey;
import java.security.cert.CertificateEncodingException;
import java.time.Instant;
import java.util.Date;

@RunWith(Parameterized.class)
public abstract class SboTokenServiceTest {
	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	@Parameter(2)
	public String kid_valid;

	@Parameter(3)
	public String kid_revoked;

	@Parameter(4)
	public String validSignature;

	protected SamlTokenTestModel testModel;

	protected final Federation federation = TestUtil.createTestFederation();

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3, "test_foces3_2027", "revoked_foces3_2024",
				"RFl5Gh9L2VEke1K8l3KhrEsjPWVa5kwh0pCPonfmzjFVX-jvWs2Omd2e-DQYGv5DuBsphF12NowJcs7JhmNM6PflLYPBgBZdPWcMKSa8wifVAGWqKz9_d9mXaBJyvIBPHiQFs0FVubNycEqPTR3G5-_4O5F16ZS-PHOhkc70QGLuuj9OUduZmiH-ZR-unoDKP0YXa4awIvbkFxpGFS4mYh2CpK9xoWIEO4l7edHOfuysd1zhfewnhmDa7VUEcxA7LU4WrTTzNlVISu_Dk8llhsMfpc2waoWnT3KDAu_mLRrVfIF2ZfD9TkSADzy9JnYXOlF6yd7GLwb4aEzPcb-UaZBxwd0qOwhWoqFHk_LCh1crxgvbSrLh8UuMvrdkh1iw_WcPhiEJooG7wO3pbJaLFPH1rhPQUc7dnA0gO7u-Dj6U4iFGyCmCzS3zzjqXT7gZXvOJ0uicTt9Td_itzJ3ApxA25r24jc_70pj2u45Ssm1AKY84cOJQnCAmcH__eHDU");

		return builder.build();
	}

	@Before
	public void setupTestModel() {
		testModel = new DefaultSamlTokenTestModel(version);
	}

	protected JWTTokenBuilder getJwtTokenBuilder(final CredentialVault trusted, final String certsubdn, final String cpr) {
		return new JWTTokenBuilder(trusted, certsubdn, new CprNr(cpr)).setKid(kid_valid);
	}

	@Ignore("SDS-6660")
	@Test
	public void tXX_validTemplate() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertSuccess(response, federation);
	}

	/**
	 * Given: Valid request hvor JWT: - er gyldig i tid - indeholder et ’certsubdn’
	 * formateret som et X509SubjectName for et POCES certifikat og som dermed
	 * indeholder en PID - er signeret af en trusted part (fx en OIDC AS), som
	 * STS’en er konfigureret med (STS’en bør benytte ’kid’ til at finde rette
	 * udsteders certifikat) - indeholder ”sosi-sts” i dens ’scope’ liste - ikke
	 * indeholder borgerens CPR (som ’cpr’)
	 * 
	 * Og hvor der for WS-Trust requestet gælder at: - signeret af en whitelistet
	 * part - CPR claimet matcher PID’en fra NameID - den requestede audience er
	 * konfigureret i STS - den requestede audience er konfigureret til ikke at
	 * modtage et BST
	 * 
	 * Expected: STS-signeret svar med et krypteret OSA som er: - udstedt til den
	 * requestede audience og krypteret med dennes certifikat - signeret af STS -
	 * har en gyldighedsperiode som konfigureret i STS - indeholder borgerens CPR -
	 * indeholder AssuranceLevel (3) - overholder OIOSAML 2.0 - ikke indeholder et
	 * BST (dvs. ingen DiscoveryEPR attribut)
	 */
	@Ignore("SDS-6660")
	@Test
	public void t01_ValidRequestWithCprClaim() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, null).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final boolean includeBst = false;
		final String audience = testModel.getSBOAudience(includeBst);

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertSuccess(response, federation);

		PrivateKey audienceKey = getAudienceKey(version);
		BootstrapOIOSAMLAssertion bootstrapOIOSAMLAssertion = response.getBootstrapOIOSAMLAssertion(audienceKey);
		assertNotNull(bootstrapOIOSAMLAssertion);

		bootstrapOIOSAMLAssertion.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapOIOSAMLAssertion.validateTimestamp();

		assertEquals(cpr, bootstrapOIOSAMLAssertion.getCpr());
		assertEquals("3", bootstrapOIOSAMLAssertion.getAssuranceLevel());
		assertEquals("DK-SAML-2.0", bootstrapOIOSAMLAssertion.getSpecVersion());

		OIOBootstrapToken bootstrapToken = bootstrapOIOSAMLAssertion.getOIOBootstrapToken();
		assertNull(bootstrapToken);
	}

	/**
	 * Given: Som T1, dog med: - ’cpr’ i JWT og uden CPR claim i requestet
	 * 
	 * Expected: Som T1
	 */
	@Ignore("SDS-6660")
	@Test
	public void t02_ValidRequestWithCprInJwt() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final boolean includeBst = false;
		final String audience = testModel.getSBOAudience(includeBst);

		final Document requestDocument = buildRequestDocument(token, valid, audience, null);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertSuccess(response, federation);

		PrivateKey audienceKey = getAudienceKey(version);
		BootstrapOIOSAMLAssertion bootstrapOIOSAMLAssertion = response.getBootstrapOIOSAMLAssertion(audienceKey);
		assertNotNull(bootstrapOIOSAMLAssertion);

		bootstrapOIOSAMLAssertion.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapOIOSAMLAssertion.validateTimestamp();

		assertEquals(cpr, bootstrapOIOSAMLAssertion.getCpr());
		assertEquals("3", bootstrapOIOSAMLAssertion.getAssuranceLevel());
		assertEquals("DK-SAML-2.0", bootstrapOIOSAMLAssertion.getSpecVersion());

		OIOBootstrapToken bootstrapToken = bootstrapOIOSAMLAssertion.getOIOBootstrapToken();
		assertNull(bootstrapToken);
	}

	@Ignore("SDS-6660")
	@Test
	public void t02x1_TrustedCprInJWTNotVerifiedvalid() throws CertificateEncodingException, CredentialVaultException {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "1234567890";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience(true);

		final Document requestDocument = buildRequestDocument(token, valid, audience, null);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertSuccess(response, federation);

		PrivateKey audienceKey = getAudienceKey(version);
		BootstrapOIOSAMLAssertion bootstrapOIOSAMLAssertion = response.getBootstrapOIOSAMLAssertion(audienceKey);
		assertNotNull(bootstrapOIOSAMLAssertion);

		bootstrapOIOSAMLAssertion.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapOIOSAMLAssertion.validateTimestamp();

		assertEquals(cpr, bootstrapOIOSAMLAssertion.getCpr());
		assertEquals("3", bootstrapOIOSAMLAssertion.getAssuranceLevel());
		assertEquals("DK-SAML-2.0", bootstrapOIOSAMLAssertion.getSpecVersion());
		assertTrue(bootstrapOIOSAMLAssertion.getIssuer(), bootstrapOIOSAMLAssertion.getIssuer().contains("STS"));

		OIOBootstrapToken bootstrapToken = bootstrapOIOSAMLAssertion.getOIOBootstrapToken();
		assertNotNull(bootstrapToken);

		bootstrapToken.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapToken.validateTimestamp();

		assertEquals(cpr, bootstrapToken.getCpr());
		assertEquals("3", bootstrapToken.getAssuranceLevel());
		assertTrue(bootstrapToken.getIssuer(), bootstrapToken.getIssuer().contains("STS"));
		assertEquals(bootstrapOIOSAMLAssertion.getSubjectNameID(), bootstrapToken.getSubjectNameID());

		try {
			assertHolderOfKey(bootstrapToken, valid.getSystemCredentialPair().getCertificate());
		} catch (CertificateEncodingException | CredentialVaultException e) {
			fail(e.getMessage());
		}
	}

	/**
	 * This is Testcase #T2 in the specification document
	 */
	@Ignore("SDS-6660")
	@Test
	public void t02x2_ValidRequestWithCprInJWT() throws Exception {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).setKid(kid_valid).setIssuer("http://sts-tester").setExpiration(Date.from(Instant.now().plus(2, HOURS))).setScope("clear sosi-sts").build();

		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience(true);

		final Document requestDocument = buildRequestDocument(token, valid, audience, null);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFalse(response.getFaultString(), response.isFault());
		assertSuccess(response, federation);
//		STS-signeret svar med et krypteret OSA som er:
//			- udstedt til den requestede audience og krypteret med dennes certifikat
//			- signeret af STS
//			- har en gyldighedsperiode som konfigureret i STS
//			- indeholder borgerens CPR
//			- indeholder AssuranceLevel (3)
//			- overholder OIOSAML 2.0 
//			- ikke indeholder et BST (dvs. ingen DiscoveryEPR attribut)

		BootstrapOIOSAMLAssertion bstOSA = response.getBootstrapOIOSAMLAssertion(getAudienceKey(version));

		assertEquals(cpr, bstOSA.getCpr());
		assertEquals(audience, bstOSA.getAudienceRestriction());
		assertTrue("Fejl med issuer:" + bstOSA.getIssuer(), bstOSA.getIssuer().contains("STS"));
		assertEquals("DK-SAML-2.0", bstOSA.getAttribute("dk:gov:saml:attribute:SpecVer"));
		assertEquals("3", bstOSA.getAssuranceLevel());

		OIOBootstrapToken bst = bstOSA.getOIOBootstrapToken();
		assertNotNull(bst);

		assertHolderOfKey(bst, valid.getSystemCredentialPair().getCertificate());
	}

	@Ignore("SDS-6660")
	@Test
	public void t02x3_CprInJWTPreferredToSamlClaim() throws CertificateEncodingException, CredentialVaultException {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cprJwt = "1234567890";
		final String cprClaim = "0987654321";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cprJwt).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final boolean includeBst = true;
		final String audience = testModel.getSBOAudience(includeBst);

		final Document requestDocument = buildRequestDocument(token, valid, audience, cprClaim);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertSuccess(response, federation);

		PrivateKey audienceKey = getAudienceKey(version);
		BootstrapOIOSAMLAssertion bootstrapOIOSAMLAssertion = response.getBootstrapOIOSAMLAssertion(audienceKey);
		assertNotNull(bootstrapOIOSAMLAssertion);

		bootstrapOIOSAMLAssertion.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapOIOSAMLAssertion.validateTimestamp();

		assertEquals(cprJwt, bootstrapOIOSAMLAssertion.getCpr());
		assertEquals("3", bootstrapOIOSAMLAssertion.getAssuranceLevel());
		assertEquals("DK-SAML-2.0", bootstrapOIOSAMLAssertion.getSpecVersion());

		OIOBootstrapToken bootstrapToken = bootstrapOIOSAMLAssertion.getOIOBootstrapToken();
		assertNotNull(bootstrapToken);

		bootstrapToken.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapToken.validateTimestamp();

		assertEquals(cprJwt, bootstrapToken.getCpr());
		assertEquals("3", bootstrapToken.getAssuranceLevel());
		assertEquals(bootstrapOIOSAMLAssertion.getSubjectNameID(), bootstrapToken.getSubjectNameID());

		assertHolderOfKey(bootstrapToken, valid.getSystemCredentialPair().getCertificate());
	}

	/**
	 * Given: Som T1, men hvor: - den requestede audience er i STS konfigureret til
	 * at modtage et BST
	 * 
	 * Expected: Som T1, bortset fra - OSA indeholder DisoveryEPR attributten med
	 * indlejret BST i Base64 encodet form
	 * 
	 * For BST’en skal der gælde: - udstedt til STS’en selv (dvs. ’audience’ sat til
	 * STS’en) - signeret af STS - har en gyldighedsperiode som konfigureret i STS -
	 * indeholder borgerens CPR - indeholder AssuranceLevel (3) - har samme Subject
	 * NameID som OSA - er af holder-of-key typen, hvor den ønskede audience’s
	 * certifikat (altså webapplikationen som startes op med SBO) indlejres som
	 * holder-of-key certifikat i SubjectConfirmation
	 * 
	 * @throws CredentialVaultException
	 * @throws CertificateEncodingException
	 */
	@Ignore("SDS-6660")
	@Test
	public void t03_ValidRequestWithCprClaimIncludeBst() throws CertificateEncodingException, CredentialVaultException {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, null).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final boolean includeBst = true;
		final String audience = testModel.getSBOAudience(includeBst);

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertSuccess(response, federation);

		PrivateKey audienceKey = getAudienceKey(version);
		BootstrapOIOSAMLAssertion bootstrapOIOSAMLAssertion = response.getBootstrapOIOSAMLAssertion(audienceKey);
		bootstrapOIOSAMLAssertion.validateTimestamp();
		assertNotNull(bootstrapOIOSAMLAssertion);

		bootstrapOIOSAMLAssertion.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapOIOSAMLAssertion.validateTimestamp();

		assertEquals(cpr, bootstrapOIOSAMLAssertion.getCpr());
		assertEquals("3", bootstrapOIOSAMLAssertion.getAssuranceLevel());
		assertEquals("DK-SAML-2.0", bootstrapOIOSAMLAssertion.getSpecVersion());

		OIOBootstrapToken bootstrapToken = bootstrapOIOSAMLAssertion.getOIOBootstrapToken();
		assertNotNull(bootstrapToken);

		bootstrapToken.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapToken.validateTimestamp();

		assertEquals(cpr, bootstrapToken.getCpr());
		assertEquals("3", bootstrapToken.getAssuranceLevel());
		assertEquals(bootstrapOIOSAMLAssertion.getSubjectNameID(), bootstrapToken.getSubjectNameID());

		// assertHolderOfKey(bootstrapToken,
		// testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
		assertHolderOfKey(bootstrapToken, valid.getSystemCredentialPair().getCertificate());
	}

	public OIOBootstrapToken callValidRequestWithCprClaimIncludeBst(final String certsubdn, final String cpr, final String kid) {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();

		final String token = getJwtTokenBuilder(trusted, certsubdn, null).setKid(kid).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final boolean includeBst = true;
		final String audience = testModel.getSBOAudience(includeBst);

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertSuccess(response, federation);

		PrivateKey audienceKey = getAudienceKey(version);
		BootstrapOIOSAMLAssertion bootstrapOIOSAMLAssertion = response.getBootstrapOIOSAMLAssertion(audienceKey);

		bootstrapOIOSAMLAssertion.validateTimestamp();
		assertNotNull(bootstrapOIOSAMLAssertion);

		bootstrapOIOSAMLAssertion.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapOIOSAMLAssertion.validateTimestamp();

		assertEquals(cpr, bootstrapOIOSAMLAssertion.getCpr());
		assertEquals("3", bootstrapOIOSAMLAssertion.getAssuranceLevel());
		assertEquals("DK-SAML-2.0", bootstrapOIOSAMLAssertion.getSpecVersion());

		OIOBootstrapToken bootstrapToken = bootstrapOIOSAMLAssertion.getOIOBootstrapToken();
		assertNotNull(bootstrapToken);

		bootstrapToken.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault(version));
		bootstrapToken.validateTimestamp();

		assertEquals(cpr, bootstrapToken.getCpr());
		assertEquals("3", bootstrapToken.getAssuranceLevel());
		assertEquals(bootstrapOIOSAMLAssertion.getSubjectNameID(), bootstrapToken.getSubjectNameID());

		try {
			assertHolderOfKey(bootstrapToken, valid.getSystemCredentialPair().getCertificate());
		} catch (CertificateEncodingException | CredentialVaultException e) {
			fail("Did not expect exception:" + e.getMessage());
		}
		return bootstrapToken;
	}

	/**
	 * Given: Request indeholdende et (udløbet) JWT udstedt af SDS OIDC AS for at
	 * sikre, at SDS OIDC AS udstedte tokens kan håndteres. (Det benyttes et udløbet
	 * JWT, idet der på runtime ikke er nemt at skaffe et gyldigt JWT) Expected:
	 * Fejlsvar: JWT ikke gyldig længere
	 */
	@Test
	public void t04_JwtExpired() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final Date past = Date.from(Instant.now().minus(10, MINUTES));
		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).setExpiration(past).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "JWT expired");
	}

	/**
	 * Given: Request med ugyldig beskeds-signatur Expected: Fejlsvar: Signatur på
	 * besked kunne ikke valideres
	 */
	@Test
	public void t05_InvalidMessageSignature() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);

		/* Manipulate Request signature */
		Element header = XmlUtil.getFirstChildElementNS(requestDocument.getDocumentElement(), "http://schemas.xmlsoap.org/soap/envelope/", "Header");
		Element security = XmlUtil.getFirstChildElementNS(header, "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd", "Security");
		Element signature = XmlUtil.getFirstChildElementNS(security, DSTags.signature.getNS(), DSTags.signature.getTagName());
		Element signatureValue = XmlUtil.getFirstChildElementNS(signature, "http://www.w3.org/2000/09/xmldsig#", "SignatureValue");
		signatureValue.getFirstChild().setNodeValue(signatureValue.getFirstChild().getNodeValue() + "-break-");
		/* Manipulate done */

		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unable to validate envelope signature");
	}

	/**
	 * Given: Request hvor certifikatet for beskeds-signaturen ikke er whitelistet
	 * Expected: Fejlsvar: Certifikat bag beskeds-signatur ikke whitelistet
	 */
	@Test
	public void t06_UntrustedMessageSigner() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault untrusted = SealPPCredentialVaults.getValidMocesNoCpr(version);
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, untrusted, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Signing certificate not whitelisted for this audience");
	}

	/**
	 * Given: Request hvor certifikatet for beskeds-signaturen er udløbet Expected:
	 * Fejlsvar: Certifikat bag beskeds-signatur udløbet
	 */
	@Test
	public void t07_TrustedButExpiredMessageSigner() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault expired = SealPPCredentialVaults.getExpiredFocesVault(version);
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, expired, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Certificate expired or not yet valid");
	}

	/**
	 * Given: Request hvor certifikatet for beskeds-signaturen er spærret Expected:
	 * Fejlsvar: Certifikat bag beskeds-signatur spærret
	 */
	@Test
	public void t08_RevokedTrustedMessageSigner() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault revoked = SealPPCredentialVaults.getRevokedFocesVault(version);
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, revoked, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_FAILED_AUTHENTICATION, "Certificate revoked");
	}

	/**
	 * Given: Request med JWT uden ’cpr’ og manglende CPR claim Expected: Fejlsvar:
	 * Manglende CPR claim
	 */
	@Ignore("SDS-6660")
	@Test
	public void t09_MissingCprClaim() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = null;

		final String token = getJwtTokenBuilder(trusted, certsubdn, null).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR claim missing");
	}

	/**
	 * Given: Request med ugyldig CPR claim Expected: Fejlsvar: CPR kunne ikke
	 * parses
	 */
	@Ignore("SDS-6660")
	@Test
	public void t10_MalformedCprSamlClaim() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "123456789";

		final String token = getJwtTokenBuilder(trusted, certsubdn, null).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR-number must be 10 digits");
	}

	/**
	 * Given: Request med ugyldig CPR i JWT Expected: Fejlsvar: CPR kunne ikke
	 * parses
	 */
	@Ignore("SDS-6660")
	@Test
	public void t10_MalformedCprInJWT() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "123456789";

		final String token = new JWTTokenBuilder(trusted, certsubdn, new CprNr("xxxxxxxxxx") {
			@Override
			public String getValue() {
				return cpr;
			}
		}).setKid(kid_valid).build();

		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, null);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR-number must be 10 digits");
	}

	/**
	 * Given: Request med CPR claim der ikke matcher PID Expected: Fejlsvar: CPR/PID
	 * mismatch
	 */
	@Ignore("SDS-6660")
	@Test
	public void t11_WrongCprSamlClaimInRequest() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "1234567890";

		final String token = getJwtTokenBuilder(trusted, certsubdn, null).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "CPR/PID mismatch");
	}

	/**
	 * Given: Request hvor certifikatet for beskeds-signaturen er spærret Expected:
	 * Fejlsvar: Certifikat bag beskeds-signatur spærret
	 */
	@Test
	public void t12_MissingAudience() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault revoked = SealPPCredentialVaults.getRevokedFocesVault(version);
		final String audience = null;

		/* Custom Request Builder */
		JWTToOIOSAMLAssertionRequestDOMBuilder requestBuilder = new JWTToOIOSAMLAssertionRequestDOMBuilder() {
			protected void validateBeforeBuild() throws ModelException {
				// not failing on missing audience
			}
		};
		/**/

		final Document requestDocument = buildRequestDocument(requestBuilder, token, revoked, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Audience is missing");
	}

	/**
	 * Given: Request med ikke-konfigureret audience Expected: Fejlsvar: Ønsket
	 * audience er ikke konfigureret
	 */
	@Ignore("SDS-6660")
	@Test
	public void t13_UnknownAudience() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault revoked = SealPPCredentialVaults.getRevokedFocesVault(version);
		final String audience = "unknown audience";

		final Document requestDocument = buildRequestDocument(token, revoked, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Audience not configured");
	}

	/**
	 * Given: Request med JWT med invalid signatur Expected: Fejlsvar: JWT signatur
	 * ugyldigt
	 */
	@Test
	public void t14_InvalidJwtSignature() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String invalidSignature = validSignature.replace('x', 'y');

		String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final String invalidToken = token.replaceAll("^(.*)\\.(.*)$", "$1." + invalidSignature);
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(invalidToken, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "JWT signature does not match locally computed signature");
	}

	/**
	 * Given: Request med JWT signeret af ikke trusted part Expected: Fejlsvar: JWT
	 * signatur ikke trusted
	 */
	@Test
	public void t15_UntrustedTokenSigner() {
		final CredentialVault untrusted = SealPPCredentialVaults.getValidVocesVault(version);
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";
		final String audience = testModel.getSBOAudience();

		final String token = getJwtTokenBuilder(untrusted, certsubdn, cpr).setKid("unknown").build();

		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.STS, StsFaultCodeValues.WST_REQUEST_FAILED, "Unable to get signing certificate");
	}

	/**
	 * Given: Request med JWT signeret af udløbet certifikat Expected: Fejlsvar: JWT
	 * signeret af udløbet certifikat
	 */
	@Test
	public void t16_ExpiredTokenSigner() {
		final CredentialVault expired = SealPPCredentialVaults.getExpiredFocesVault(version);
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";
		final String audience = testModel.getSBOAudience();

		final String token = getJwtTokenBuilder(expired, certsubdn, cpr).setKid("expired").build();

		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_REQUEST_FAILED, "JWT signer expired");
	}

	/**
	 * Given: Request med JWT signeret af spærret certifikat Expected: Fejlsvar: JWT
	 * signeret af spærret certifikat
	 */
	@Ignore("SDS-7321") // Revoked certificate er udløbet. Skal fixes.
	@Test
	public void t17_RevokedTokenSigner() {
		final CredentialVault revoked = SealPPCredentialVaults.getRevokedFocesVault(version);

		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";
		final String audience = testModel.getSBOAudience();

		final String token = getJwtTokenBuilder(revoked, certsubdn, cpr).setKid(kid_revoked).build();

		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_REQUEST_FAILED, "JWT signer revoked");
	}

	/**
	 * Given: Request med JWT med ’certsubdn’ for et MOCES certifikat Expected:
	 * Fejlsvar: JWT er ikke udstedt til en borger"
	 */
	@Test
	public void t18_MocesJwtToken() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String MOCES_certsubdn = "SERIALNUMBER=CVR:30808460-RID:42634739 + CN=TU GENEREL MOCES M CPR gyldig, O=NETS DANID A/S // CVR:30808460, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, MOCES_certsubdn, cpr).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Token not issued to a citizen");
	}

	/**
	 * Given: Request med JWT med ’certsubdn’ som ikke er formateret som
	 * X509SubjectName Expected: Fejlsvar: JWT ’certsubdn’ kan ikke parses som
	 * X509SubjectName
	 */
	@Test
	public void t19_MalformedSubjectDN() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "nameid-malformed";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Invalid name: nameid");
	}

	@Test
	public void t19x1_MissingSubjectDN() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = null;
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "JWT claim missing: certsubdn");
	}

//	/**
//	 * Given:
//	 * 	Som T1, men hvor:
//	 *  - den requestede audience er i STS konfigureret til at modtage et BST
//	 *  
//	 * Expected:
//	 * 	Som T1, bortset fra
//	 *  - OSA indeholder DisoveryEPR attributten med indlejret BST i Base64 encodet form
//	 *  
//	 *  For BST’en skal der gælde:
//	 *  - udstedt til STS’en selv (dvs. ’audience’ sat til STS’en)
//	 *  - signeret af STS
//	 *  - har en gyldighedsperiode som konfigureret i STS
//	 *  - indeholder borgerens CPR
//	 *  - indeholder AssuranceLevel (3)
//	 *  - har samme Subject NameID som OSA
//	 *  - er af holder-of-key typen, hvor den ønskede audience’s certifikat (altså webapplikationen som startes op med SBO) indlejres som holder-of-key certifikat i SubjectConfirmation
//	 * @throws CredentialVaultException 
//	 * @throws CertificateEncodingException 
//	 * @throws IOException 
//	 */
//	@Test
//	public void t20_ValidRequestWithCprClaimIncludeBst() throws CertificateEncodingException, CredentialVaultException, IOException {
//		final CredentialVault trusted = testModel.getExternalTrustedIdp();
//		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
//		final String cpr = "0501792275";
//		
//		final String token = getJwtTokenBuilder(trusted, certsubdn, null).build();
//		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
//		final boolean includeBst = true;
//		final String audience = testModel.getSBOAudience(includeBst);
//	
//		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
//		final Document responseDoc = getResponseDocument(requestDocument);
//		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);
//		
//		assertSuccess(response, federation);
//		
//		PrivateKey audienceKey = getAudienceKey(version);	
//		BootstrapOIOSAMLAssertion bootstrapOIOSAMLAssertion = response.getBootstrapOIOSAMLAssertion(audienceKey);
//		bootstrapOIOSAMLAssertion.validateTimestamp();
//		assertNotNull(bootstrapOIOSAMLAssertion);
//
//		bootstrapOIOSAMLAssertion.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault());
//		bootstrapOIOSAMLAssertion.validateTimestamp();
//		
//		assertEquals(cpr, bootstrapOIOSAMLAssertion.getCpr());
//		assertEquals("3", bootstrapOIOSAMLAssertion.getAssuranceLevel());
//		assertEquals("DK-SAML-2.0", bootstrapOIOSAMLAssertion.getSpecVersion());
//		
//		OIOBootstrapToken bootstrapToken = bootstrapOIOSAMLAssertion.getOIOBootstrapToken();
//		assertNotNull(bootstrapToken);
//		
//		bootstrapToken.validateSignatureAndTrust(StsCredentialVaults.getStsCredentialVault());
//		bootstrapToken.validateTimestamp();
//		
//		assertEquals(cpr, bootstrapToken.getCpr());
//		assertEquals("3", bootstrapToken.getAssuranceLevel());
//		assertEquals(bootstrapOIOSAMLAssertion.getSubjectNameID(), bootstrapToken.getSubjectNameID());
//		
//		assertHolderOfKey(bootstrapToken, testModel.getValidWhitelistedVault().getSystemCredentialPair().getCertificate());
//		
//		BootstrapTokenProcurationLocalTest bootstrapTokenProcurationLocalTest = new BootstrapTokenProcurationLocalTest();
//		bootstrapTokenProcurationLocalTest.initTestModel();
//		bootstrapTokenProcurationLocalTest.setUp();
//		bootstrapTokenProcurationLocalTest.testCallFromSBO(bootstrapToken);
//	}

	@Test
	public void t25x1_UnknownJWTIssuer() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";
		final String audience = testModel.getSBOAudience();

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).setIssuer("unkown").build();

		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Unknown issuer");
	}

	@Ignore("SDS-6660")
	@Test
	public void t30x1_StsMissingInScope() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).setScope("clear not-sosi-sts").build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Incoming JSON webtoken not scoped for this sts");
	}

	@Ignore("SDS-6660")
	@Test
	public void t30x2_AudienceScopeMismatch() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).setScope("sosi-sts not-clear").build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getSBOAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "Incoming JSON webtoken not scoped for audience");
	}

	@Ignore("SDS-6660")
	@Test
	public void t30x3_AudienceHasNoScopeConfiguration() {
		final CredentialVault trusted = testModel.getExternalTrustedIdp();
		final String certsubdn = "SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK";
		final String cpr = "0501792275";

		final String token = getJwtTokenBuilder(trusted, certsubdn, cpr).setScope("sosi-sts clear").build();
		final GenericCredentialVault valid = testModel.getValidWhitelistedVault();
		final String audience = testModel.getCryptAudience();

		final Document requestDocument = buildRequestDocument(token, valid, audience, cpr);
		final Document responseDoc = getResponseDocument(requestDocument);
		final JWTToOIOSAMLAssertionResponse response = getResponseModel(responseDoc);

		assertFault(response, StsFaultActorValues.ITS, StsFaultCodeValues.WST_INVALID_REQUEST, "JWT scope not configured for audience");
	}

	public static PrivateKey getAudienceKey(Version version) {
		return getAudienceVault(version).getSystemCredentialPair().getPrivateKey();
	}

	protected static GenericCredentialVault getAudienceVault(Version version) {
		return CertificateStore.getValidVocesVault(version);
	}

	protected final Document buildRequestDocument(String jwtToken, CredentialVault signingVault, String appliesTo, String cprClaim) {
		JWTToOIOSAMLAssertionRequestDOMBuilder requestBuilder = createRequestBuilder(jwtToken, signingVault, appliesTo, cprClaim);
		return buildRequestDocument(requestBuilder);
	}

	protected final Document buildRequestDocument(JWTToOIOSAMLAssertionRequestDOMBuilder requestBuilderRef, String jwtToken, CredentialVault signingVault, String appliesTo, String cprClaim) {
		JWTToOIOSAMLAssertionRequestDOMBuilder requestBuilder = initRequestBuilder(requestBuilderRef, jwtToken, signingVault, appliesTo, cprClaim);
		return buildRequestDocument(requestBuilder);
	}

	protected final Document buildRequestDocument(JWTToOIOSAMLAssertionRequestDOMBuilder requestBuilder) {
		final Document requestDocument = requestBuilder.build();
		LibertySignatureValidator validator = new LibertySignatureValidator(new EmptyCredentialVault(), requestDocument.getDocumentElement());
		validator.validateSignature();
		return requestDocument;
	}

	private JWTToOIOSAMLAssertionRequestDOMBuilder createRequestBuilder(String jwtToken, CredentialVault signingVault, String appliesTo, String cprClaim) {
		JWTToOIOSAMLAssertionRequestDOMBuilder requestBuilder = new OIOSAMLFactory().createJWTToOIOSAMLAssertionRequestDOMBuilder();
		return initRequestBuilder(requestBuilder, jwtToken, signingVault, appliesTo, cprClaim);
	}

	private JWTToOIOSAMLAssertionRequestDOMBuilder initRequestBuilder(JWTToOIOSAMLAssertionRequestDOMBuilder requestBuilder, String jwtToken, CredentialVault signingVault, String appliesTo, String cprClaim) {
		/**/
		requestBuilder.setCPRNumberClaim(cprClaim);
		requestBuilder.setSigningVault(signingVault);
		requestBuilder.setAudience(appliesTo);
		/**/
		requestBuilder.setJWTToken(jwtToken);
		return requestBuilder;
	}

	protected final Document getResponseDocument(Document requestDocument) {
		return getTestClient().execute(XmlUtil.node2String(requestDocument, false, false)).document;
	}

	protected final JWTToOIOSAMLAssertionResponse getResponseModel(Document responseDoc) {
		return new OIOSAMLFactory().createJWTToOIOSAMLAssertionResponseModelBuilder().build(responseDoc);
	}

	protected abstract TestClient getTestClient();

}
