package dk.sosi.sts.server.service;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import dk.sosi.seal.SecurityProviderUtil;
import dk.sosi.seal.pki.CRLCache;
import dk.sosi.seal.pki.CertificateResolver;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.pki.InMemoryCRLCache;
import dk.sosi.seal.pki.OCESCertificateResolver;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.seal.pki.impl.intermediate.HashMapIntermediateCertificateCache;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.SealTestHelper;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.StsCredentialVaults;
import dk.sosi.sts.server.test.TestUtil;

import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;

import java.io.IOException;
import java.io.InputStream;
import java.security.Provider;
import java.security.Security;
import java.security.cert.CRLException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509CRL;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

/**
 *
 * @author moa@arosii.dk
 */
@RunWith(Parameterized.class)
public class CRLCertificateStatusCheckerCopyRunner {
	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	private static int INT5MIN = 300;

	private final CRLCache cache = new InMemoryCRLCache();
	private final int defaultTTL = 5 * 60 * 60 * 1000;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3);

		return builder.build();
	}

	public static class CRLCertificateStatusCheckerCopyBC extends CRLCertificateStatusCheckerCopy {

		public CRLCertificateStatusCheckerCopyBC(CRLCache cache, int interval, boolean strict, int ttl, CertificateResolver certificateResolver) {
			super(cache, interval, strict, ttl, certificateResolver);
		}

		@Override
		protected Provider getCertificateFactoryProvider() {
			// use the BouncyCastle provider if installed
			Provider result = Security.getProvider(BouncyCastleProvider.PROVIDER_NAME);
			assertNotNull(result);
			return result;
		}
	}

	CRLCache cache() {
		cache.clear();
		return cache;
	}

	CertificateResolver resolver() {
		return new OCESCertificateResolver(new HashMapIntermediateCertificateCache());
	}

	CRLCertificateStatusCheckerCopy checker(boolean useBounceCastle, boolean strict) {
		CRLCertificateStatusCheckerCopy result;
		if (useBounceCastle) {
			result = new CRLCertificateStatusCheckerCopyBC(cache(), CRLCertificateStatusCheckerCopy.ALWAYS, strict, defaultTTL, resolver());
		} else {
			result = new CRLCertificateStatusCheckerCopy(cache(), CRLCertificateStatusCheckerCopy.ALWAYS, strict, defaultTTL, resolver());
		}
		result.setConnectTimeout(2 * 60000);
		result.setReadTimeout(2 * 60000);
		return result;
	}

	CRLCertificateStatusCheckerCopy checker(boolean strict) {
		return checker(true, strict);
	}

	@Before
	public void setUp() throws Exception {
		TestUtil.disableLogging();
	}

	@After
	public void tearDown() throws Exception {
		TestUtil.revertLogging();
	}

	@Test
	public void testIllegalArguments() {
		try {
			new CRLCertificateStatusCheckerCopyBC(null, CRLCertificateStatusCheckerCopy.ALWAYS, false, defaultTTL, resolver());
			fail("IllegalArgumentException expected");
		} catch (IllegalArgumentException e) {
			// expected
		}

		try {
			new CRLCertificateStatusCheckerCopyBC(cache(), -100, false, defaultTTL, resolver());
			fail("IllegalArgumentException expected");
		} catch (IllegalArgumentException e) {
			// expected
		}

		try {
			new CRLCertificateStatusCheckerCopyBC(cache(), CRLCertificateStatusCheckerCopy.ALWAYS, false, -defaultTTL, null);
			fail("IllegalArgumentException expected");
		} catch (IllegalArgumentException e) {
			// expected
		}

	}

	@Test
	public void testIsNullRevoked() throws Exception {
		CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), CRLCertificateStatusCheckerCopy.NEVER, false, defaultTTL, resolver());
		try {
			statusChecker.getRevocationStatus(null);
			fail("IllegalArgumentException expected");
		} catch (IllegalArgumentException e) {
			// OK
		}
	}

	@Test
	public void testCertificateWithoutCRLRef() throws Exception {
		X509Certificate cert = SealTestHelper.issueCertificate();

		CRLCertificateStatusCheckerCopy statusChecker = checker(false);
		statusChecker.getRevocationStatus(cert);

		SecurityProviderUtil.addBCAsProvider();
		statusChecker = checker(true);
		try {
			statusChecker.getRevocationStatus(cert);
			fail();
		} catch (IllegalStateException e) {
			// OK
		}
		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testRevokedCertificate() throws Exception {
		SecurityProviderUtil.addBCAsProvider();
		CRLCertificateStatusCheckerCopy statusChecker = checker(true);

		X509Certificate cert = getRevokedMoces();
		assertFalse(statusChecker.getRevocationStatus(cert).isValid());
		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testValidCertificate() throws Exception {
		SecurityProviderUtil.addBCAsProvider();
		CRLCertificateStatusCheckerCopy statusChecker = checker(true);

		X509Certificate cert = getValidMoces();
		assertTrue(statusChecker.getRevocationStatus(cert).isValid());
		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testValidCertificateButUrlUnreachable_Strict() throws Exception {
		final CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), INT5MIN, true, defaultTTL, resolver()) {
			@Override
			protected CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				throw new IOException("Artificial Timeout.");
			}
		};

		X509Certificate cert = getValidMoces();
		try {
			assertFalse(statusChecker.getRevocationStatus(cert).isValid());
			fail("Strict case should throw exception in case of unreachable CRL URL");
		} catch (IllegalStateException t) {
			// OK
		}
	}

	@Test
	public void testValidCertificateButUrlUnreachableNonStrict() throws Exception {
		final CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), INT5MIN, false, defaultTTL, resolver()) {
			@Override
			protected CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				throw new IOException("Artificial Timeout.");
			}
		};

		X509Certificate cert = getValidMoces();
		CertificateStatus revoked = statusChecker.getRevocationStatus(cert);
		assertEquals(revoked.getTimestamp(), CRLCertificateStatusCheckerCopy.INVALID_TIMESTAMP);
		assertTrue(revoked.isValid());
	}

	@Test
	public void testValidCertificateCheckIntervalSemantic() throws Exception {
		SecurityProviderUtil.addBCAsProvider();
		final AtomicInteger i = new AtomicInteger(0);
		final CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), INT5MIN, true, defaultTTL, resolver()) {
			@Override
			protected CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				i.incrementAndGet();
				return super.load(url, crlInfo);
			}
		};
		statusChecker.setConnectTimeout(2 * 60000);
		statusChecker.setReadTimeout(2 * 60000);

		X509Certificate cert = getValidMoces();
		assertTrue(statusChecker.getRevocationStatus(cert).isValid());

		assertTrue(statusChecker.getRevocationStatus(cert).isValid());

		assertEquals(i.get(), 1);
		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testValidCertificateCheckAlways() throws Exception {
		SecurityProviderUtil.addBCAsProvider();

		final AtomicInteger i = new AtomicInteger(0);
		final CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), CRLCertificateStatusCheckerCopy.ALWAYS, true, defaultTTL, resolver()) {
			@Override
			protected CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				i.incrementAndGet();
				return super.load(url, crlInfo);
			}
		};
		statusChecker.setConnectTimeout(2 * 60000);
		statusChecker.setReadTimeout(2 * 60000);

		X509Certificate cert = getValidMoces();
		assertTrue(statusChecker.getRevocationStatus(cert).isValid());

		assertTrue(statusChecker.getRevocationStatus(cert).isValid());

		assertEquals(i.get(), 2);

		SecurityProviderUtil.removeBCAsProvider();
	}

	/**
	 * The certificate, tu-voces-spxrret, is revoked!
	 */
	@Test
	public void testRevokedOces2Certificate() throws Exception {
		SecurityProviderUtil.addBCAsProvider();
		Date now = new Date();
		CredentialVault oces2CredentialVault = SealPPCredentialVaults.getRevokedFocesVault(version);
		X509Certificate certificate = oces2CredentialVault.getSystemCredentialPair().getCertificate();

		assertNotNull(certificate);

		final CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), INT5MIN, true, defaultTTL, resolver()) {
			@Override
			protected CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				assertEquals("http://crl.systemtest8.trust2408.com/systemtest8.crl", url);
				return super.load(url, crlInfo);
			}
		};

		statusChecker.setConnectTimeout(2 * 60000);
		statusChecker.setReadTimeout(2 * 60000);
		CertificateStatus certificateStatus = statusChecker.getRevocationStatus(certificate);
		assertFalse(certificateStatus.isValid());
		assertTrue(certificateStatus.getTimestamp().before(new Date(now.getTime() + INT5MIN * 1000)));
		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testTTL() throws Exception {
		SecurityProviderUtil.addBCAsProvider();

		CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), CRLCertificateStatusCheckerCopy.ALWAYS, true, Integer.MIN_VALUE, resolver());
		statusChecker.setConnectTimeout(2 * 60000);
		statusChecker.setReadTimeout(2 * 60000);

		X509Certificate cert = getRevokedMoces();
		try {
			statusChecker.getRevocationStatus(cert);
			fail("CRL should not be valid");
		} catch (IllegalStateException e) {
			// OK
		}

		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testNoTTL() throws Exception {
		final AtomicBoolean loaded = new AtomicBoolean(false);

		// serve the crl, which has next update in the past (2009) for any certificate
		CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), CRLCertificateStatusCheckerCopy.ALWAYS, true, CRLCertificateStatusCheckerCopy.NEVER, resolver()) {
			@Override
			protected CRLCache.CRLInfo load(String url, CRLCache.CRLInfo crlInfo) throws IOException {
				InputStream inStream = Thread.currentThread().getContextClassLoader().getResourceAsStream("ocesNextUpdateInPast.crl");

				X509CRL crl;
				try {
					CertificateFactory cf = CertificateFactory.getInstance("X.509");
					crl = (X509CRL) cf.generateCRL(inStream);
					assertNotNull(crl);
					assertTrue(loaded.compareAndSet(false, true));
				} catch (CRLException e) {
					throw new RuntimeException(e);
				} catch (CertificateException e) {
					throw new RuntimeException(e);
				} finally {
					inStream.close();
				}
				return new UncheckedCRLInfo(crl, new Date().getTime());
			}

			@Override
			protected boolean verify(X509CRL crl, X509Certificate cert) {
				return true;
			}
		};
		statusChecker.setConnectTimeout(2 * 60000);
		statusChecker.setReadTimeout(2 * 60000);

		// doesn't matter which certificate we use
		X509Certificate cert = getRevokedMoces();
		assertTrue(statusChecker.getRevocationStatus(cert).isValid());
		assertTrue(loaded.get());
	}

	@Test
	public void testVerificationFailed() throws Exception {
		SecurityProviderUtil.addBCAsProvider();
		X509Certificate certificate = getRevokedMoces();

		// serve a totally unrelated crl for the certificate
		final CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), INT5MIN, true, defaultTTL, resolver()) {
			@Override
			protected CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				return super.load("http://crl.systemtest10.trust2408.com/systemtest10.crl", crlInfo);
			}
		};
		statusChecker.setConnectTimeout(2 * 60000);
		statusChecker.setReadTimeout(2 * 60000);

		try {
			statusChecker.getRevocationStatus(certificate);
			fail("CRL should not be valid");
		} catch (IllegalStateException e) {
			// OK
		}
		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testPartitioned() throws Exception {
		final AtomicBoolean loaded = new AtomicBoolean(false);

		// serve the partitioned crl from any certificate
		CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), CRLCertificateStatusCheckerCopy.ALWAYS, true, defaultTTL, resolver()) {
			@Override
			protected CRLCache.CRLInfo load(String url, CRLCache.CRLInfo crlInfo) throws IOException {
				InputStream inStream = Thread.currentThread().getContextClassLoader().getResourceAsStream("partcrl.crl");

				X509CRL crl;
				try {
					CertificateFactory cf = CertificateFactory.getInstance("X.509");
					crl = (X509CRL) cf.generateCRL(inStream);
					assertNotNull(crl);
					assertTrue(loaded.compareAndSet(false, true));
				} catch (CRLException e) {
					throw new RuntimeException(e);
				} catch (CertificateException e) {
					throw new RuntimeException(e);
				} finally {
					inStream.close();
				}
				return new UncheckedCRLInfo(crl, new Date().getTime());
			}

			@Override
			protected boolean verify(X509CRL crl, X509Certificate cert) {
				return true;
			}
		};
		statusChecker.setConnectTimeout(2 * 60000);
		statusChecker.setReadTimeout(2 * 60000);

		// doesn't matter which certificate we use (might by when the order of crl
		// checks is taken into account)
		X509Certificate cert = getRevokedMoces();
		try {
			statusChecker.getRevocationStatus(cert);
			fail("CRL should not be valid");
		} catch (IllegalStateException e) {
			// OK
		}
		assertTrue(loaded.get());
	}

	@Test
	public void testChecksNotDoneTwice() throws Exception {
		SecurityProviderUtil.addBCAsProvider();
		final AtomicInteger counter = new AtomicInteger(0);
		CRLCache crlCache = new InMemoryCRLCache();
		CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(crlCache, CRLCertificateStatusCheckerCopy.ALWAYS, true, defaultTTL, resolver()) {
			@Override
			protected boolean verify(X509CRL crl, X509Certificate cert) {
				counter.incrementAndGet();
				return super.verify(crl, cert);
			}
		};
		statusChecker.setConnectTimeout(2 * 60000);
		statusChecker.setReadTimeout(2 * 60000);

		X509Certificate cert = getRevokedMoces();
		assertEquals("not invoked yet", 0, counter.get());
		int lastInvocationCount = 0;
		long lastTimestamp = -1L;
		for (int i = 0; i < 50; i++) {
			assertFalse(statusChecker.getRevocationStatus(cert).isValid());
			assertEquals(1, crlCache.entries().size());
			long lastModified = crlCache.entries().iterator().next().getValue().getLastModified();
			// lastModified for the crl might not be consistent
			// we make sure, that if we invoke several times, it is because the timestamp
			// has changed
			if (counter.get() > lastInvocationCount) {
				assertTrue(lastModified != lastTimestamp);
			}
			lastInvocationCount = counter.get();
			lastTimestamp = lastModified;
		}
		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testUpdateAll() {
		final String KEY = "hey";
		CRLCache.CRLInfo crlInfo = new CRLCache.CRLInfo(null, System.currentTimeMillis());

		final CRLCertificateStatusCheckerCopy statusChecker = new CRLCertificateStatusCheckerCopyBC(cache(), INT5MIN, true, defaultTTL, resolver()) {
			@Override
			protected CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				assertEquals(KEY, url);
				return crlInfo;
			}
		};
		statusChecker.setConnectTimeout(2 * 60000);
		statusChecker.setReadTimeout(2 * 60000);

		cache.update(KEY, crlInfo);

		statusChecker.updateAll();

	}

	// -- test download

	private static final String CRL_URL = "http://crl.systemtest8.trust2408.com/systemtest8.crl";

	@Test
	public void testDownload() throws Exception {
		SecurityProviderUtil.addBCAsProvider();

		final CRLCertificateStatusCheckerCopy checker = checker(true);
		checker.setConnectTimeout(2 * 60000);
		checker.setReadTimeout(2 * 60000);
		final CRLCache.CRLInfo crlInfo = checker.downloadCRL(CRL_URL, null);

		assertNotNull(crlInfo);
		assertTrue(System.currentTimeMillis() - crlInfo.getCreated() < 2 * 60000L); // 2 minute to download should be enough
		assertNotNull(crlInfo.getCrl());

		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testInvalidURL() throws Exception {
		SecurityProviderUtil.addBCAsProvider();
		try {
			checker(true, false).downloadCRL("http://www.google.com", null);
			fail("Invalid url");
		} catch (PKIException e) {
			// OK
		}
		SecurityProviderUtil.removeBCAsProvider();
	}

	private static void clearJDKInternalCRLCache() {
		try {
			CertificateFactory.getInstance("X.509").generateCRL(null);
		} catch (CRLException e) {
			// OK: see implementation ... (X509Factory.java)
		} catch (CertificateException e) {
			throw new RuntimeException(e);
		}
	}

	/*
	 * Test that the download/creation of a new CRL depends on the lastModified
	 * flag.
	 */
	@Test
	public void testNoDownloading() throws Exception {
		SecurityProviderUtil.addBCAsProvider();
		final CRLCache.CRLInfo crlInfo1 = checker(false).downloadCRL(CRL_URL, null);
		clearJDKInternalCRLCache();
		final CRLCache.CRLInfo crlInfo2 = checker(false).downloadCRL(CRL_URL, null);

		assertTrue(crlInfo1.getCrl().equals(crlInfo2.getCrl()));
		assertFalse(crlInfo1.getCrl() == crlInfo2.getCrl());

		clearJDKInternalCRLCache();

		final CRLCache.CRLInfo crlInfo3 = checker(false).downloadCRL(CRL_URL, crlInfo2);
		assertTrue(crlInfo2.getCrl().equals(crlInfo3.getCrl()));
		assertSame(crlInfo2.getCrl(), crlInfo3.getCrl());
		SecurityProviderUtil.removeBCAsProvider();
	}

	@Test
	public void testInvalidConnectTimeout() {
		final CRLCertificateStatusCheckerCopy checker = checker(false);
		try {
			checker.setConnectTimeout(-1500);
			fail();
		} catch (IllegalArgumentException e) {
			// OK
		}
		try {
			checker.setConnectTimeout(0);
			fail();
		} catch (IllegalArgumentException e) {
			// OK
		}
	}

	@Test
	public void testInvalidReadTimeout() {
		final CRLCertificateStatusCheckerCopy checker = checker(false);
		try {
			checker.setReadTimeout(-1500);
			fail();
		} catch (IllegalArgumentException e) {
			// OK
		}
		try {
			checker.setReadTimeout(0);
			fail();
		} catch (IllegalArgumentException e) {
			// OK
		}
	}

	@Test
	public void testStsCertificateNotRevoked() {
		X509Certificate certificate = StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate();
		CertificateStatusChecker statusChecker = new CRLCertificateStatusCheckerCopyRunner.CRLCertificateStatusCheckerCopyBC(new InMemoryCRLCache(), 3600, false, 300, new OCESCertificateResolver(new HashMapIntermediateCertificateCache()));
		final CertificateStatus revocationStatus = statusChecker.getRevocationStatus(certificate);
		if (!revocationStatus.isValid()) {
			fail("STS certificate is revoked < " + revocationStatus.getTimestamp());
		}
	}

	private X509Certificate getValidMoces() throws Exception {
		return CertificateStore.getValidMoces(version);
	}

	private X509Certificate getRevokedMoces() throws Exception {
		return SealPPCredentialVaults.getRevokedMocesVault(version).getSystemCredentialPair().getCertificate();
	}
}
