package dk.sosi.sts.server.service;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import dk.sosi.seal.SecurityProviderUtil;
import dk.sosi.seal.pki.CRLCache;
import dk.sosi.seal.pki.CertificateResolver;
import dk.sosi.seal.pki.InMemoryCRLCache;
import dk.sosi.seal.pki.OCESCertificateResolver;
import dk.sosi.seal.pki.impl.intermediate.HashMapIntermediateCertificateCache;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.service.CRLCertificateStatusCheckerCopyRunner.CRLCertificateStatusCheckerCopyBC;
import dk.sosi.sts.server.test.TestUtil;

import org.apache.log4j.Logger;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.management.ManagementFactory;
import java.net.URL;
import java.net.URLConnection;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.security.cert.X509Certificate;
import java.util.concurrent.TimeUnit;

/**
 * Integration test cases for CRL code. This tests the very large 20 MB CRL
 * file, so we don't do that in the unit tests.
 * 
 * @author moa@arosii.dk
 */
@RunWith(Parameterized.class)
public class CRLHttpIT {

	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	private static int INT5MIN = 30000;
	private final int defaultTTL = 5 * 60 * 60 * 1000;

	private final CRLCache cache = new InMemoryCRLCache();

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3);

		return builder.build();
	}

	@Before
	public void setUp() throws Exception {
		TestUtil.disableLogging();
		SecurityProviderUtil.addBCAsProvider();
	}

	@After
	public void tearDown() throws Exception {
		SecurityProviderUtil.removeBCAsProvider();
		TestUtil.revertLogging();
	}

	/**
	 * Test that the Bouncy Castle CRL implementation uses less than 128 MB for the
	 * 20 MB OCES2 CRL.
	 */
	@Test
	public void testDownloadLargeCRL() throws Exception {
		File crlFile = largeCRLDownload();
		try {
			System.gc();
			long start = System.nanoTime();
			long heapUsed = largeCRLMemoryUsage(true, crlFile);
			long time = System.nanoTime() - start;

			Logger.getLogger(CRLHttpIT.class).debug("CRL memory usage BC:  " + heapUsed + " B/" + (heapUsed / 1024) + " KB/" + (heapUsed / (1024 * 1024)) + " MB (" + TimeUnit.NANOSECONDS.toMillis(time) + " ms)");
			assertTrue("Bouncy Castle CRL heap used not smaller than 224 MB (" + heapUsed + ")", heapUsed < 224L * 1024 * 1024);
		} finally {
			crlFile.delete();
		}
	}

	/**
	 * Manual test for comparing Sun default CRL impl. with BouncyCastle CRL impl.
	 * <p>
	 * Note this requires a large heap size (at least 512 mb).
	 * </p>
	 */
	public void manual_testCompareLargeCRL() throws Exception {
		File crlFile = largeCRLDownload();
		try {
			System.gc();
			long start = System.nanoTime();
			long heapUsedDefault = largeCRLMemoryUsage(false, crlFile);
			long timeDefault = System.nanoTime() - start;

			System.gc();
			start = System.nanoTime();
			long heapUsedBouncyCastle = largeCRLMemoryUsage(true, crlFile);
			long timeBC = System.nanoTime() - start;

			Logger.getLogger(CRLHttpIT.class).debug("CRL memory usage SUN: " + heapUsedDefault + " B/" + (heapUsedDefault / 1024) + " KB/" + (heapUsedDefault / (1024 * 1024)) + " MB (" + TimeUnit.NANOSECONDS.toMillis(timeDefault) + " ms)");
			Logger.getLogger(CRLHttpIT.class).debug("CRL memory usage BC:  " + heapUsedBouncyCastle + " B/" + (heapUsedBouncyCastle / 1024) + " KB/" + (heapUsedBouncyCastle / (1024 * 1024)) + " MB (" + TimeUnit.NANOSECONDS.toMillis(timeBC) + " ms)");
			double factor = ((double) heapUsedDefault) / heapUsedBouncyCastle;
			double tfactor = ((double) timeDefault) / timeBC;
			assertTrue("factor smaller than 4: " + factor, factor > 4);
			assertTrue("time factor smaller than 4: " + tfactor, tfactor > 4);
		} finally {
			crlFile.delete();
		}
	}

	private static final String LARGE_CRL_URL = "http://ca1.cti-gov.dk/oces/issuing/1/crl/issuing.crl";

	private File largeCRLDownload() throws IOException {
		Logger.getLogger(CRLHttpIT.class).debug("Downloading large CRL from " + LARGE_CRL_URL + " ..");
		long start = System.nanoTime();
		long size;
		File tmpFile = File.createTempFile(CRLHttpIT.class.getSimpleName(), ".crl");
		tmpFile.deleteOnExit();
		FileOutputStream fos = new FileOutputStream(tmpFile);
		try {
			URL url = new URL(LARGE_CRL_URL);
			URLConnection conn = url.openConnection();
			conn.setConnectTimeout(2 * 60000);
			conn.setReadTimeout(10 * 60000);
			conn.connect();
			InputStream is = conn.getInputStream();
			try {
				ReadableByteChannel rbc = Channels.newChannel(is);
				try {
					size = fos.getChannel().transferFrom(rbc, 0, Long.MAX_VALUE);
				} finally {
					rbc.close();
				}
			} finally {
				is.close();
			}
		} finally {
			fos.close();
		}
		Logger.getLogger(CRLHttpIT.class).debug("Downloaded large CRL (" + (size / (1024 * 1024)) + " MB) in " + TimeUnit.NANOSECONDS.toMillis(System.nanoTime() - start) + " ms");
		return tmpFile;
	}

	private long largeCRLMemoryUsage(boolean useBounceCastle, File crlFile) throws Exception {
		CRLCertificateStatusCheckerCopy checker;
		if (useBounceCastle) {
			checker = new CRLCertificateStatusCheckerCopyBC(cache(), INT5MIN, true, defaultTTL, resolver());
		} else {
			checker = new CRLCertificateStatusCheckerCopy(cache(), INT5MIN, true, defaultTTL, resolver());
		}

		checker.setConnectTimeout(2 * 60000);
		checker.setReadTimeout(10 * 60000);

		long count = 10;
		long heapUsedSum = 0;
		for (int i = 0; i < count; i++) {
			System.gc();
			long heapUsedBefore = ManagementFactory.getMemoryMXBean().getHeapMemoryUsage().getUsed();
			final CRLCache.CRLInfo crlInfo = checker.downloadCRL(crlFile.toURI().toURL().toString(), null);
			assertNotNull(crlInfo);
			assertNotNull(crlInfo.getCrl());

			X509Certificate validCert = CertificateStore.getValidMoces(version);
			assertTrue(checker.getRevocationStatus(validCert).isValid());
			X509Certificate revokedCert = CertificateStore.getRevokedMoces(version);
			assertFalse(checker.getRevocationStatus(revokedCert).isValid());

			heapUsedSum += ManagementFactory.getMemoryMXBean().getHeapMemoryUsage().getUsed() - heapUsedBefore;
		}
		return heapUsedSum / count;
	}

	private CertificateResolver resolver() {
		return new OCESCertificateResolver(new HashMapIntermediateCertificateCache());
	}

	private CRLCache cache() {
		cache.clear();
		return cache;
	}

	@Test
	public void testFile() throws Exception {
		final CRLCertificateStatusCheckerCopy checker;
		checker = new CRLCertificateStatusCheckerCopyBC(cache(), INT5MIN, true, defaultTTL, resolver());

		final X509Certificate certificate = CertificateStore.getValidMoces(version);
		final Runnable runnable = new Runnable() {

			@Override
			public void run() {
				synchronized (this) {
					try {
						this.wait();
					} catch (InterruptedException e) {
						throw new RuntimeException(e);
					}
				}
				try {
					Thread.sleep((long) (Math.random() * 100));
				} catch (InterruptedException e) {
					throw new RuntimeException(e);
				}
				final boolean valid = checker.getRevocationStatus(certificate).isValid();
				assertTrue(valid);
			}
		};

		final int tCount = 3;
		final Thread[] threads = new Thread[tCount];

		for (int i = 0; i < tCount; i++) {
			threads[i] = new Thread(runnable);
			threads[i].start();
		}

		synchronized (runnable) {
			runnable.notify();
		}

		// This should be tuned to the download speed
		Thread.sleep(200);

		synchronized (runnable) {
			runnable.notifyAll();
		}

		for (int i = 0; i < tCount; i++) {
			threads[i].join();
		}

	}
}
