/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.security.cert.X509Certificate;
import java.sql.Connection;
import java.sql.Statement;

import org.bouncycastle.asn1.x509.Extension;
import org.hsqldb.jdbc.JDBCDataSource;
import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.mockito.Mockito;

import dk.sosi.seal.pki.CRLCertificateStatusChecker;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.server.test.StsCredentialVaults;

/**
 * 
 * @author Jacob Qvortrup
 */
public class CraOcesCrlServiceDbTest {

	private static final String[] DDL_CREATE = new String[] { "CREATE TABLE crl (id BIGINT NOT NULL, url VARCHAR(2000) NOT NULL, lastmodified BIGINT NOT NULL, nextupdate BIGINT NOT NULL, PRIMARY KEY (id))", "CREATE TABLE revoked ( id BIGINT NOT NULL, crlid BIGINT NOT NULL, serialnumber VARCHAR(48), added BIGINT NOT NULL, since BIGINT NOT NULL, PRIMARY KEY (id))", };

	private static final String[] DDL_DROP = new String[] { "DROP TABLE crl", "DROP TABLE revoked", };

	private CraOcesCrlServiceDbImpl craOcesCrlServiceImpl;

	private JDBCDataSource dataSource;

	private long now;

	@Before
	public void createCraOcesCrlServiceImpl() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		Connection conn = dataSource.getConnection();
		Statement statement = conn.createStatement();
		for (String sql : DDL_CREATE) {
			statement.execute(sql);
		}
		conn.close();

		craOcesCrlServiceImpl = new CraOcesCrlServiceDbImpl();
		craOcesCrlServiceImpl.setDataSource(dataSource);
		now = System.currentTimeMillis();
	}

	@After
	public void cleanDatabase() throws Exception {
		Connection conn = dataSource.getConnection();
		Statement statement = conn.createStatement();
		for (String sql : DDL_DROP) {
			statement.execute(sql);
		}
		conn.close();
	}

	@Test
	public void nullCertificate() {
		X509Certificate certificate = null;
		try {
			craOcesCrlServiceImpl.getRevocationStatus(certificate);
			fail("Expected PKIException");
		} catch (PKIException e) {
			assertEquals("Certificate must not be null", e.getMessage());
		}
	}

	@Test
	public void certificateWithoutUrl() {
		X509Certificate certificate = Mockito.mock(X509Certificate.class);
		Mockito.when(certificate.getExtensionValue(Extension.cRLDistributionPoints.getId())).thenReturn(null);
		try {
			craOcesCrlServiceImpl.getRevocationStatus(certificate);
			fail("Expected PKIException");
		} catch (PKIException e) {
			assertEquals("Certificate must have a CRL url", e.getMessage());
		}
	}

	@Test
	public void certificateWithoutSerialNumber() {
		X509Certificate base = StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate();
		X509Certificate certificate = Mockito.mock(X509Certificate.class);
		Mockito.when(certificate.getExtensionValue(Extension.cRLDistributionPoints.getId())).thenReturn(base.getExtensionValue(Extension.cRLDistributionPoints.getId()));
		Mockito.when(certificate.getSerialNumber()).thenReturn(null);
		try {
			craOcesCrlServiceImpl.getRevocationStatus(certificate);
			fail("Expected PKIException");
		} catch (PKIException e) {
			assertEquals("Certificate must have a serial number", e.getMessage());
		}
	}

	@Test
	public void unknownCertificate() {
		X509Certificate certificate = StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate();
		try {
			craOcesCrlServiceImpl.getRevocationStatus(certificate);
			fail("Expected PKIException");
		} catch (PKIException e) {
			assertEquals("Certificate has an unknown CRL url: " + CraOcesCrlServiceDbImpl.getCRLUrlFromCertificate(certificate), e.getMessage());
		}
	}

	@Test
	public void revokedCertificate() throws Exception {
		X509Certificate certificate = StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate();
		dataSource.getConnection().createStatement().execute("INSERT INTO crl (id, url, lastmodified, nextupdate) VALUES (1, '" + CraOcesCrlServiceDbImpl.getCRLUrlFromCertificate(certificate) + "', " + now + ", " + (now + 5000) + ")");
		dataSource.getConnection().createStatement().execute("INSERT INTO revoked (id, crlid, serialnumber, added, since) VALUES (1, 1, '" + certificate.getSerialNumber().toString() + "', 1, 1)");
		CertificateStatus status = craOcesCrlServiceImpl.getRevocationStatus(certificate);
		assertFalse(status.isValid());
		assertEquals(now, status.getTimestamp().getTime());
	}

	@Test
	public void revokedIssuer() throws Exception {
		X509Certificate certificate = StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate();
		dataSource.getConnection().createStatement().execute("INSERT INTO crl (id, url, lastmodified, nextupdate) VALUES (1, '" + CraOcesCrlServiceDbImpl.getCRLUrlFromCertificate(certificate) + "', " + now + ", " + (now + 5000) + ")");
		dataSource.getConnection().createStatement().execute("INSERT INTO revoked (id, crlid, serialnumber, added, since) VALUES (1, 1, null, 1, 1)");
		CertificateStatus status = craOcesCrlServiceImpl.getRevocationStatus(certificate);
		assertFalse(status.isValid());
		assertEquals(now, status.getTimestamp().getTime());
	}

	@Test
	public void outdatedData() throws Exception {
		X509Certificate certificate = StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate();
		dataSource.getConnection().createStatement().execute("INSERT INTO crl (id, url, lastmodified, nextupdate) VALUES (1, '" + CraOcesCrlServiceDbImpl.getCRLUrlFromCertificate(certificate) + "', " + (now - 10000) + ", " + (now - 5000) + ")");
		try {
			craOcesCrlServiceImpl.getRevocationStatus(certificate);
			fail("Expected PKIException");
		} catch (PKIException e) {
			assertEquals("Certificate Revocation Authority not up to date", e.getMessage());
		}
	}

	@Test
	public void validNow() throws Exception {
		X509Certificate certificate = StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate();
		dataSource.getConnection().createStatement().execute("INSERT INTO crl (id, url, lastmodified, nextupdate) VALUES (1, '" + CraOcesCrlServiceDbImpl.getCRLUrlFromCertificate(certificate) + "', " + now + ", " + (now + 5000) + ")");
		CertificateStatus status = craOcesCrlServiceImpl.getRevocationStatus(certificate);
		assertTrue(status.isValid());
		assertEquals(now, status.getTimestamp().getTime());
	}

	@Test
	public void validTTL() throws Exception {
		X509Certificate certificate = StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate();
		long past = System.currentTimeMillis() - 5000;
		dataSource.getConnection().createStatement().execute("INSERT INTO crl (id, url, lastmodified, nextupdate) VALUES (1, '" + CraOcesCrlServiceDbImpl.getCRLUrlFromCertificate(certificate) + "', " + past + ", " + (past + 5000) + ")");
		craOcesCrlServiceImpl.setTtl(1);
		CertificateStatus status = craOcesCrlServiceImpl.getRevocationStatus(certificate);
		assertTrue(status.isValid());
		assertEquals(past, status.getTimestamp().getTime());
	}

	@Test
	public void validTTLDisabled() throws Exception {
		X509Certificate certificate = StsCredentialVaults.getStsCredentialVault(Version.OCES3).getSystemCredentialPair().getCertificate();
		long past = System.currentTimeMillis() - 5000;
		dataSource.getConnection().createStatement().execute("INSERT INTO crl (id, url, lastmodified, nextupdate) VALUES (1, '" + CraOcesCrlServiceDbImpl.getCRLUrlFromCertificate(certificate) + "', " + past + ", " + (past + 5000) + ")");
		craOcesCrlServiceImpl.setTtl(CRLCertificateStatusChecker.NEVER);
		CertificateStatus status = craOcesCrlServiceImpl.getRevocationStatus(certificate);
		assertTrue(status.isValid());
		assertEquals(past, status.getTimestamp().getTime());
	}
}
