/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import dk.sosi.seal.pki.CertificateCache;
import dk.sosi.seal.pki.CertificateResolver;
import dk.sosi.seal.pki.OCESCertificateResolver;
import dk.sosi.seal.pki.SealInternalApi;
import dk.sosi.seal.pki.impl.intermediate.IntermediateCertificateStoreAdapter;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.server.CertificateStore;
import dk.sosi.sts.server.StsHashMapCertificateCacheFactory;
import dk.sosi.sts.server.test.SealPPCredentialVaults;
import dk.sosi.sts.server.test.TestUtil;
import dk.sosi.sts.service.OcesCvrRidService;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;

import java.security.cert.X509Certificate;
import java.util.Collections;

/**
 *
 */
@RunWith(Parameterized.class)
public class CvrRidEndpointTest {
	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	final CertificateCache intermediateCertificateCache = StsHashMapCertificateCacheFactory.create();
	final OCESCertificateResolver ocesCertificateResolver = new OCESCertificateResolver(new IntermediateCertificateStoreAdapter(intermediateCertificateCache));

	final CertificateResolver resolver = ocesCertificateResolver;
	final StsOcesCvrRidService ocesCvrRidService = new StsOcesCvrRidService();

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3);

		return builder.build();
	}

	@Before
	public void setUp() throws Exception {
		TestUtil.disableLogging();
		ocesCvrRidService.setWsEndpoints(Collections.<String, OcesCvrRidService>emptyMap());
		ocesCvrRidService.setCertificateResolver(resolver);
	}

	@After
	public void tearDown() {
		TestUtil.revertLogging();
	}

	private void assertRootFound(CredentialVault credentialVault) throws StsException {
		X509Certificate certificate = credentialVault.getSystemCredentialPair().getCertificate();
		final X509Certificate rootIssuer = ocesCvrRidService.getRootIssuingCertificate(certificate);
		assertTrue("A root certificate could not be found!", SealInternalApi.getOces3RootPPCertificates().contains(rootIssuer));
	}

	@Test
	public void testAllValidKnown() throws StsException {
		assertRootFound(SealPPCredentialVaults.getValidMocesNoCpr(version));
		assertRootFound(SealPPCredentialVaults.getValidMocesWithCpr(version));

		assertRootFound(SealPPCredentialVaults.getValidFocesVault(version));
		assertRootFound(SealPPCredentialVaults.getRevokedFocesVault(version));

		assertRootFound(SealPPCredentialVaults.getValidVocesVault(version));
		assertRootFound(SealPPCredentialVaults.getRevokedVocesVault(version));
	}

	@Test
	public void testAllThatFailsWithOces3() throws StsException {
		assertRootFound(SealPPCredentialVaults.getRevokedMocesVault(version));
		assertRootFound(SealPPCredentialVaults.getExpiredMocesVault(version));
		assertRootFound(SealPPCredentialVaults.getExpiredFocesVault(version));
		assertRootFound(SealPPCredentialVaults.getExpiredVocesVault(version));
	}

	@Test
	public void testfindService() throws Exception {

		// World!
		X509Certificate rootCertProd = SealInternalApi.getOces2RootPRCertificate();
		X509Certificate rootCertPP = SealInternalApi.getOces2RootPPCertificate();

		assertCertName("C=DK,O=TRUST2408,CN=TRUST2408 OCES Primary CA", rootCertProd);
		assertCertName("C=DK,O=TRUST2408,CN=TRUST2408 Systemtest VII Primary CA", rootCertPP);

		// All root certificates are self-issued
		assertIssuing(rootCertProd, rootCertProd);
		assertIssuing(rootCertPP, rootCertPP);

		assertFalse(SealInternalApi.isIntermediateCertificate(rootCertProd));
		assertFalse(SealInternalApi.isIntermediateCertificate(rootCertPP));
	}

	@Test
	public void testFindCertificateChain() throws Exception {
		final X509Certificate certificate = CertificateStore.getValidMoces(version);
		assertFalse(SealInternalApi.isIntermediateCertificate(certificate));

		X509Certificate intermediate = ocesCvrRidService.getIssuingCertificate(certificate);
		assertTrue(SealInternalApi.isIntermediateCertificate(intermediate));

		X509Certificate intermediateParent = ocesCvrRidService.getIssuingCertificate(intermediate);
		assertTrue(SealInternalApi.getOces3RootPPCertificates().contains(intermediateParent));

		final X509Certificate rootIssuer = ocesCvrRidService.getRootIssuingCertificate(certificate);
		assertTrue(SealInternalApi.getOces3RootPPCertificates().contains(rootIssuer));
		assertEquals("CN=Den Danske Stat OCES udstedende-CA 1,OU=Test - cti,O=Den Danske Stat,C=DK", ocesCvrRidService.getKey(intermediate));
	}

	private void assertCertName(String name, X509Certificate certificate) {
		assertEquals(name, ocesCvrRidService.getKey(certificate));
		assertEquals(name, certificate.getSubjectDN().getName());
	}

	private void assertIssuing(X509Certificate parent, X509Certificate child) {
		assertEquals(parent, ocesCvrRidService.getIssuingCertificate(child));
		assertTrue(SealInternalApi.isIssuerOf(child, parent));
	}
}
