/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfoParser;

import org.hsqldb.jdbc.JDBCDataSource;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;

import java.security.cert.X509Certificate;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;

@RunWith(Parameterized.class)
public class StsAclServiceTest {

	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public X509Certificate testCert;

	private JDBCDataSource dataSource;
	private String SYSTEM_NAME = "orgName";
	private static OcesInfo ocesInfo;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(CredentialVaultTestUtil.getVoces3CredentialVault().getSystemCredentialPair().getCertificate());

		return builder.build();
	}

	@Before
	public void setUp() throws Exception {
		dataSource = new JDBCDataSource();
		dataSource.setDatabase("jdbc:hsqldb:mem:memoryDB");
		dataSource.setUser("sa");
		dataSource.setPassword("");
		Connection conn = dataSource.getConnection();
		Statement statement = conn.createStatement();
		statement.execute("DROP TABLE whitelist IF EXISTS");
		statement.execute("CREATE TABLE whitelist (ssn VARCHAR(100), created DATE)");
		statement.execute("DROP TABLE blacklist IF EXISTS");
		statement.execute("CREATE TABLE blacklist (ssn VARCHAR(100), created DATE)");
		conn.close();

		ocesInfo = OcesInfoParser.parseCertificate(testCert);
	}

	@After
	public void tearDown() throws Exception {
		Connection conn = dataSource.getConnection();
		conn.createStatement().execute("DROP TABLE whitelist");
		conn.createStatement().execute("DROP TABLE blacklist");
		conn.close();
	}

	@Test
	public void testEmptyWhiteList() {
		StsAclExtendedService service = new StsAclExtendedService(dataSource);
		assertFalse(service.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testEmptyBlackList() {
		StsAclExtendedService service = new StsAclExtendedService(dataSource);
		assertFalse(service.isBlacklisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testBlacklisted() throws SQLException {
		Connection conn = dataSource.getConnection();
		PreparedStatement pstmt = conn.prepareStatement("INSERT INTO blacklist VALUES (?, ?)");
		pstmt.setString(1, OcesInfoParser.parseCertificate(testCert).getSubjectSerialNumber().getValue());
		pstmt.setDate(2, null);
		pstmt.execute();
		pstmt.close();
		conn.close();

		StsAclExtendedService service = new StsAclExtendedService(dataSource);
		assertTrue(service.isBlacklisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testWhitelisted() throws SQLException {
		Connection conn = dataSource.getConnection();
		PreparedStatement pstmt = conn.prepareStatement("INSERT INTO whitelist VALUES (?, ?)");
		pstmt.setString(1, "CVR:" + OcesInfoParser.parseCertificate(testCert).getCvr() + "-SYSTEM:" + SYSTEM_NAME);
		pstmt.setDate(2, null);
		pstmt.execute();
		pstmt.close();
		conn.close();

		StsAclExtendedService service = new StsAclExtendedService(dataSource);
		assertTrue(service.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testNoWhitelisting() throws Exception {
		StsAclExtendedService service = new StsAclService(dataSource);

		assertTrue(service.isWhitelisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testEmptyBlackListWithNoWhitelisting() {
		StsAclExtendedService service = new StsAclService(dataSource);
		assertFalse(service.isBlacklisted(ocesInfo, SYSTEM_NAME));
	}

	@Test
	public void testBlacklistedWithNoWhitelisting() throws SQLException {
		Connection conn = dataSource.getConnection();
		PreparedStatement pstmt = conn.prepareStatement("INSERT INTO blacklist VALUES (?, ?)");
		pstmt.setString(1, OcesInfoParser.parseCertificate(testCert).getSubjectSerialNumber().getValue());
		pstmt.setDate(2, null);
		pstmt.execute();
		pstmt.close();
		conn.close();

		StsAclExtendedService service = new StsAclService(dataSource);
		assertTrue(service.isBlacklisted(ocesInfo, SYSTEM_NAME));
	}
}
