/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import dk.sosi.seal.pki.CRLCache;
import dk.sosi.seal.pki.CertificateResolver;
import dk.sosi.seal.pki.InMemoryCRLCache;
import dk.sosi.seal.pki.OCESCertificateResolver;
import dk.sosi.seal.pki.impl.intermediate.IntermediateCertificateStoreAdapter;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.its.DocumentResponse;
import dk.sosi.sts.server.IdentityTokenRequestHandler;
import dk.sosi.sts.server.StsHashMapCertificateCacheFactory;
import dk.sosi.sts.server.test.SealPPCredentialVaults;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.mockito.Mockito;

import java.io.IOException;
import java.lang.reflect.Method;
import java.security.cert.X509Certificate;
import java.util.AbstractMap;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;

@RunWith(Parameterized.class)
public class StsOcesCrlServiceImplTest {
	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public Version version;

	// private URL crl_url;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();

		builder.newTest("OCES3").add(Version.OCES3);

		return builder.build();
	}

	@Before
	public void setUp() throws Exception {
		// OCES TEST CRL:
		// crl_url = new URL("http://test.crl.oces.certifikat.dk/oces.crl");
	}

	// TODO: this test is redundant, but some tests for the implementation should be
	// created.
	@Test
	public void testMissingEveryConfigConfiguration() {

		try {
			OCESCertificateResolver resolver = new OCESCertificateResolver(new IntermediateCertificateStoreAdapter(StsHashMapCertificateCacheFactory.create()));
			final StsOcesCrlServiceImpl service = new StsOcesCrlServiceImpl();
			service.setCertificateResolver(resolver);
			service.afterPropertiesSet();
		} catch (Exception e) {
			fail("Should default to in-memory if property for cache not set.");
		}
	}

	public void manual_testCrl() throws IOException {
		final OCESCertificateResolver resolver = new OCESCertificateResolver(new IntermediateCertificateStoreAdapter(StsHashMapCertificateCacheFactory.create()));
		final StsOcesCrlServiceImpl stsOcesCrlService = new StsOcesCrlServiceImpl() {
			{
				setCrlCache(new InMemoryCRLCache());
				setCertificateResolver(resolver);
				try {
					afterPropertiesSet();
				} catch (Exception e) {
					fail("Exception:" + e);
				}
			}

			@Override
			public CRLCache.CRLInfo load(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				final CRLCache.CRLInfo load = super.load(url, crlInfo);
				System.out.println("URL: " + url);
				System.out.println(load.getCrl());
				return load;
			}

		};

//		stsOcesCrlService.load("http://test.crl.oces.certifikat.dk/oces.crl", null);
		// stsOcesCrlService.load("http://crl.systemtest17.trust2408.com/systemtest17.crl",
		// null);
		final X509Certificate certificate = SealPPCredentialVaults.getValidVocesVault(version).getSystemCredentialPair().getCertificate();
		System.out.println(certificate);
		stsOcesCrlService.getRevocationStatus(certificate);

	}

	// Remove this one?
	@Test
	public void testAnnop() throws Exception {
		final IdentityTokenRequestHandler identityTokenRequestHandler = new IdentityTokenRequestHandler(null, null, null, null, null);

		final Method isFault = identityTokenRequestHandler.getClass().getDeclaredMethod("isFault", DocumentResponse.class);
		System.out.println(Arrays.toString(isFault.getDeclaredAnnotations()));

	}

	@Test
	public void testRefresh() throws Exception {

		final StsOcesCrlServiceImpl stsOcesCrlService = new StsOcesCrlServiceImpl();

		// Mocks
		final StsOcesCrlServiceImpl.StsCertificateStatusChecker checker = Mockito.mock(StsOcesCrlServiceImpl.StsCertificateStatusChecker.class);
		final CRLCache crlCache = Mockito.mock(CRLCache.class);

		stsOcesCrlService.setCertificateStatusChecker(checker);
		stsOcesCrlService.setCrlCache(crlCache);

		// Cache setup
		final HashSet<Map.Entry<String, CRLCache.CRLInfo>> cacheContents = new HashSet<Map.Entry<String, CRLCache.CRLInfo>>();
		final CRLCache.CRLInfo crlInfo = new CRLCache.CRLInfo(null, -1);
		cacheContents.add(new AbstractMap.SimpleEntry<String, CRLCache.CRLInfo>("URL", crlInfo));
		Mockito.when(crlCache.entries()).thenReturn(cacheContents);

		// Download of crl
		final CRLCache.CRLInfo result = new CRLCache.CRLInfo(null, System.currentTimeMillis());
		Mockito.when(checker.load("URL", crlInfo)).thenReturn(result);

		stsOcesCrlService.refresh();

		Mockito.verify(crlCache).update("URL", result);
	}

	@Test
	public void testRefresh_Failed() throws Exception {

		final StsOcesCrlServiceImpl stsOcesCrlService = new StsOcesCrlServiceImpl();

		// Mocks
		final StsOcesCrlServiceImpl.StsCertificateStatusChecker checker = Mockito.mock(StsOcesCrlServiceImpl.StsCertificateStatusChecker.class);
		final CRLCache crlCache = Mockito.mock(CRLCache.class);

		stsOcesCrlService.setCertificateStatusChecker(checker);
		stsOcesCrlService.setCrlCache(crlCache);

		// cache setup
		final HashSet<Map.Entry<String, CRLCache.CRLInfo>> cacheContents = new HashSet<Map.Entry<String, CRLCache.CRLInfo>>();
		final CRLCache.CRLInfo crlInfo = new CRLCache.CRLInfo(null, -1);
		cacheContents.add(new AbstractMap.SimpleEntry<String, CRLCache.CRLInfo>("URL", crlInfo));
		Mockito.when(crlCache.entries()).thenReturn(cacheContents);

		// load problem
		final CRLCache.CRLInfo result = new CRLCache.CRLInfo(null, System.currentTimeMillis());
		Mockito.when(checker.load("URL", crlInfo)).thenThrow(new RuntimeException("problem"));

		stsOcesCrlService.refresh();

		Mockito.verify(crlCache, Mockito.never()).update("URL", result);
	}

	@Test
	public void testLoad() throws Exception {

		final StsOcesCrlServiceImpl stsOcesCrlService = new StsOcesCrlServiceImpl();

		// Mocks
		final CRLCache cache = Mockito.mock(CRLCache.class);
		final CertificateResolver certificateResolver = Mockito.mock(CertificateResolver.class);
		final StsOcesCrlServiceImpl.StsCertificateStatusChecker checker = stsOcesCrlService.new StsCertificateStatusChecker(cache, -1, true, 0, certificateResolver) {
			@Override
			public CRLCache.CRLInfo fetchURL(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				return null;
			}
		};

		stsOcesCrlService.setCertificateStatusChecker(checker);

		stsOcesCrlService.load("URL", null);

		assertTrue(System.currentTimeMillis() - stsOcesCrlService.getLastSuccessful() < 1000);
		assertEquals(-1, stsOcesCrlService.getLastFailed());
	}

	@Test
	public void testLoad_Failed() throws Exception {

		final StsOcesCrlServiceImpl stsOcesCrlService = new StsOcesCrlServiceImpl();

		// Mocks
		final CRLCache cache = Mockito.mock(CRLCache.class);
		final CertificateResolver certificateResolver = Mockito.mock(CertificateResolver.class);
		final StsOcesCrlServiceImpl.StsCertificateStatusChecker checker = stsOcesCrlService.new StsCertificateStatusChecker(cache, -1, true, 0, certificateResolver) {
			@Override
			public CRLCache.CRLInfo fetchURL(final String url, final CRLCache.CRLInfo crlInfo) throws IOException {
				throw new RuntimeException("problem");
			}
		};

		stsOcesCrlService.setCertificateStatusChecker(checker);

		try {
			stsOcesCrlService.load("URL", null);
			fail("Exception should propagate.");
		} catch (Throwable t) {

		}

		assertTrue(System.currentTimeMillis() - stsOcesCrlService.getLastFailed() < 1000);
		assertEquals(-1, stsOcesCrlService.getLastSuccessful());
	}

	@Test
	public void testInitialSetup() throws Exception {

		final StsOcesCrlServiceImpl stsOcesCrlService = new StsOcesCrlServiceImpl();

		// Mocks
		final CRLCache crlCache = Mockito.mock(CRLCache.class);
		final CertificateResolver resolver = Mockito.mock(CertificateResolver.class);

		stsOcesCrlService.setCrlCache(crlCache);
		stsOcesCrlService.setCertificateResolver(resolver);
		stsOcesCrlService.setInterval(3600);
		stsOcesCrlService.setCrl_ttl(0);
		stsOcesCrlService.setReadTimeout(10000);
		stsOcesCrlService.setConnectTimeout(10000);

		stsOcesCrlService.afterPropertiesSet();

		assertNotNull(stsOcesCrlService.getCertificateStatusChecker());

	}
}
