/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service;

import java.security.Principal;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import org.junit.Ignore;
import org.mockito.Mockito;

import dk.sosi.seal.pki.CertificateResolver;
import dk.sosi.sts.StsException;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.service.OcesCvrRidService;
import junit.framework.TestCase;

public class StsOcesCvrRidServiceTest extends TestCase {

	CertificateResolver resolver;
	Map<String, OcesCvrRidService> map;
	StsOcesCvrRidService stsOcesCvrRidService;

	@Override
	public void setUp() throws Exception {
		resolver = Mockito.mock(CertificateResolver.class);
		map = new HashMap<String, OcesCvrRidService>();

		stsOcesCvrRidService = new StsOcesCvrRidService();
		stsOcesCvrRidService.setWsEndpoints(map);
		stsOcesCvrRidService.setCertificateResolver(resolver);
	}

	private ArrayList<X509Certificate> setupChain(String... sdns) {
		X509Certificate prev = null;
		final ArrayList<X509Certificate> x509Certificates = new ArrayList<X509Certificate>();
		for (String sdn : sdns) {
			final X509Certificate cert = Mockito.mock(X509Certificate.class);
			final Principal principal = Mockito.mock(Principal.class);
			Mockito.when(cert.getSubjectDN()).thenReturn(principal);
			if (prev == null) {
				Mockito.when(principal.getName()).thenReturn(sdn);
			} else {
				Mockito.when(principal.getName()).thenReturn("SERIALNUMBER="+sdn);
			}

			x509Certificates.add(cert);

			if (prev == null) {
				Mockito.when(resolver.getIssuingCertificate(cert)).thenReturn(cert);
			} else {
				Mockito.when(resolver.getIssuingCertificate(cert)).thenReturn(prev);
			}
			prev = cert;

		}
		return x509Certificates;
	}

	private void setupService(OcesCvrRidService mock, OcesInfo ocesInfo, CprNr cprNr) throws StsException {
		Mockito.when(mock.findRelatedCpr(ocesInfo)).thenReturn(cprNr);

	}

	public void testRouting() throws Exception {
		final ArrayList<X509Certificate> chain = setupChain("OCES1", "CVR:19343634-RID:1165813849809");

		final OcesInfo ocesInfo = OcesInfoParser.parseCertificate(chain.get(1));

		final CprNr cprNr = new CprNr("1111111119");

		final OcesCvrRidService mock = Mockito.mock(OcesCvrRidService.class);
		map.put("OCES1", mock);
		setupService(mock, ocesInfo, cprNr);

		final CprNr relatedCpr = stsOcesCvrRidService.findRelatedCpr(ocesInfo);

		assertEquals(cprNr, relatedCpr);
	}

	public void testRoutingDeep() throws Exception {
		final ArrayList<X509Certificate> chain = setupChain("OCES1",
															"CVR:19343634-RID:1165811849809",
															"CVR:19343634-RID:2165813149809",
															"CVR:19343634-RID:3165813841809",
															"CVR:19343634-RID:4161813849809",
															"CVR:19343634-RID:5165811849809",
															"CVR:19343634-RID:6165813849801",
															"CVR:19343634-RID:7165813849809"
														   );
		final OcesInfo ocesInfo = OcesInfoParser.parseCertificate(chain.get(7));

		final CprNr cprNr = new CprNr("1111111119");

		final OcesCvrRidService mock = Mockito.mock(OcesCvrRidService.class);
		map.put("OCES1", mock);
		setupService(mock, ocesInfo, cprNr);

		final CprNr relatedCpr = stsOcesCvrRidService.findRelatedCpr(ocesInfo);

		assertEquals(cprNr, relatedCpr);


	}

	
	public void _testRoutingTooDeep() throws Exception {
		final ArrayList<X509Certificate> chain = setupChain("OCES1",
															"CVR:19343634-RID:1165811849809",
															"CVR:19343634-RID:2165813149809",
															"CVR:19343634-RID:3165813841809",
															"CVR:19343634-RID:4161813849809",
															"CVR:19343634-RID:5165811849809",
															"CVR:19343634-RID:6165813849801",
															"CVR:19343634-RID:7165813841809",
															"CVR:19343634-RID:8161813849809",
															"CVR:19343634-RID:9165811849809",
															"CVR:19343634-RID:1000000000000",
															"CVR:19343634-RID:1100000000000",
															"CVR:19343634-RID:1200000000000"
														   );
		final OcesInfo ocesInfo = OcesInfoParser.parseCertificate(chain.get(12));
    assertEquals("CVR:19343634-RID:1200000000000", ocesInfo.getSubjectSerialNumber().getValue());

		final CprNr cprNr = new CprNr("1111111119");

		final OcesCvrRidService mock = Mockito.mock(OcesCvrRidService.class);
		map.put("OCES1", mock);
		setupService(mock, ocesInfo, cprNr);

		try {
			stsOcesCvrRidService.findRelatedCpr(ocesInfo);
			fail("expected exception for depth > 10");
		} catch (StsException e) {

		}
	}

	// TODO: test with setup with multiple root certificates/more than one service
}
