/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.service.procuration;

import dk.sosi.sts.ServiceInfoCollector;
import dk.sosi.sts.StsException;
import dk.sosi.sts.fuldmagt.stub.ProcurationTestModel;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.model.PrivilegeGroup;
import dk.sosi.sts.service.ProcurationService;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.junit.Assume;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;

import static org.junit.Assert.*;

public class AggregatedProcurationServiceIT {
    private static final String mockEndpoint1 = "http://localhost:8082/fuldmagt-stub/MockService1";
    private static final String mockEndpoint2 = "http://localhost:8082/fuldmagt-stub/MockService2";
    private ProcurationService service;

    @BeforeClass
    public static void init() {
        // Testen kører kun, hvis de lokale mock services er tændt
        try (CloseableHttpClient httpClient = HttpClients.createDefault()) {
            HttpGet request = new HttpGet(mockEndpoint1);
            CloseableHttpResponse response = httpClient.execute(request);
            Assume.assumeNotNull(response);

        } catch (IOException e) {
            Assume.assumeNoException(e);
        }
    }

    @Before
    public void initService() throws MalformedURLException {
        ProcurationWebservice service1 = new ProcurationWebservice(new ServiceInfoCollector(0));
        service1.setEntityId("https://saml.sts.nspop.dk");
        service1.setEndpoint(new URL(mockEndpoint1));
        service1.init();

        ProcurationWebservice service2= new ProcurationWebservice(new ServiceInfoCollector(0));
        service2.setEntityId("https://saml.sts.nspop.dk");
        service2.setEndpoint(new URL(mockEndpoint2));
        service2.init();

        this.service = new AggregatedProcurationService(new ServiceInfoCollector(0), service1, service2);
    }

    @Test
    public void testReadOnly() throws StsException {
        PrivilegeGroup procurationPrivileges = service.getProcurationPrivileges(new CprNr(ProcurationTestModel.CPR_1), new CprNr(ProcurationTestModel.CPR_3));
        List<String> privileges = procurationPrivileges.getPrivileges();
        assertFalse(privileges.isEmpty());
        assertTrue(procurationPrivileges.getConstraints().isEmpty());
        assertTrue(privileges.contains(ProcurationTestModel.FMK_READ)); // From MockService1
        assertTrue(privileges.contains(ProcurationTestModel.PRIVILEGE_1)); // From MockService2
    }

    @Test
    public void testNoProcurations() throws StsException {
        PrivilegeGroup procurationPrivileges = service.getProcurationPrivileges(new CprNr(ProcurationTestModel.CPR_3), new CprNr(ProcurationTestModel.CPR_1));
        assertEquals(0, procurationPrivileges.getPrivileges().size());
        assertEquals(0, procurationPrivileges.getConstraints().size());
    }

    @Test
    public void testWrite() throws StsException {
        PrivilegeGroup procurationPrivileges = service.getProcurationPrivileges(new CprNr(ProcurationTestModel.CPR_2), new CprNr(ProcurationTestModel.CPR_3));
        List<String> privileges = procurationPrivileges.getPrivileges();
        assertFalse(privileges.isEmpty());
        assertTrue(procurationPrivileges.getConstraints().isEmpty());
        assertTrue(privileges.contains(ProcurationTestModel.FMK_WRITE)); // From MockService1
        assertFalse(privileges.contains(ProcurationTestModel.FMK_READ)); // From MockService1
        assertTrue(privileges.contains(ProcurationTestModel.PRIVILEGE_2)); // From MockService2
    }
}