/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.test;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.security.KeyStoreException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateFactory;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.Enumeration;

import dk.sosi.sts.ServiceInfoCollector;
import dk.sosi.sts.server.StsHashMapCertificateCacheFactory;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.DERIA5String;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.x509.CRLDistPoint;
import org.bouncycastle.asn1.x509.DistributionPoint;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;

import dk.sosi.seal.pki.CRLCertificateStatusChecker;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.pki.InMemoryCRLCache;
import dk.sosi.seal.pki.IntermediateCertificateCache;
import dk.sosi.seal.pki.OCESCertificateResolver;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.seal.pki.SealInternalApi;
import dk.sosi.seal.pki.impl.intermediate.IntermediateCertificateStoreAdapter;
import dk.sosi.seal.pki.internal.remote.HttpCertificateLoader;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.sts.certificate.OcesInfo;
import dk.sosi.sts.certificate.OcesInfoParser;
import dk.sosi.sts.model.CprNr;
import dk.sosi.sts.ocescpr.cvrrid.service.impl.WsOcesCvrRidService;

/**
 * User: ksr
 * Date: 16-01-15
 * Time: 11:00
 */
public class CertificateDescriptionTool {
	public static final String SERVICE_ENDPOINT_PP = "https://ws-erhverv.pp.certifikat.dk/rid_serviceprovider_server/services/HandleSundhedsportalWSPort";

	public static void main(String[] args) throws Exception {
		// SDS-6660
		// Kan denne slettes helt?
	}


	public static X509Certificate loadCertificateBase64(String base64String) throws FileNotFoundException, CertificateException {
		InputStream fileInputStream = new ByteArrayInputStream(XmlUtil.fromBase64(base64String));
		return loadCertificate(fileInputStream);
	}

	public static X509Certificate loadCertificateJks(File x509file) throws FileNotFoundException, CertificateException {
		InputStream fileInputStream = new FileInputStream(x509file);
		return loadCertificate(fileInputStream);
	}

	private static X509Certificate loadCertificate(InputStream inputStream) throws CertificateException {
		CertificateFactory cf = CertificateFactory.getInstance("X.509");
		return (X509Certificate) cf.generateCertificate(inputStream);
	}

	public static String describeCertificate(CredentialVault vault) throws KeyStoreException, IOException {
		StringBuilder builder = new StringBuilder();
		Enumeration<String> aliases = vault.getKeyStore().aliases();
		while (aliases.hasMoreElements()) {
			String alias = aliases.nextElement();
			builder.append(alias).append('\n');
			X509Certificate cert = (X509Certificate) vault.getKeyStore().getCertificate(alias);
			builder.append(describeCertificate(cert)).append('\n');
		}
		return builder.toString();
	}

	public static String describeCertificate(X509Certificate cert) throws IOException {
		Appendable writer = new StringBuilder();
		writer.append("============================\n");
		appendDescription(cert, writer);
		writer.append("============================\n");
		return writer.toString();
	}

	private static void appendDescription(X509Certificate cert, Appendable appendable) throws IOException {
		appendable.append(cert.getSubjectDN().getName()).append('\n');
		appendable.append("issuer=").append(cert.getIssuerX500Principal().getName()).append('\n');
		try {
			cert.checkValidity(new Date());
			appendable.append("valid=true (until " + cert.getNotAfter() + ")");
		} catch (CertificateExpiredException e) {
			appendable.append("valid=expired at " + cert.getNotAfter());
		} catch (CertificateNotYetValidException e) {
			appendable.append("valid=not until " + cert.getNotBefore());
		}
		CRLCertificateStatusChecker statusChecker = new CRLCertificateStatusChecker(new InMemoryCRLCache(), 3600, true, 1860, new OCESCertificateResolver(new IntermediateCertificateStoreAdapter(StsHashMapCertificateCacheFactory.create())));
		appendable.append(", revoked=");
		try {
			CertificateStatus status = statusChecker.getRevocationStatus(cert);
			appendable.append(""+!status.isValid());
		} catch (IllegalStateException e) {
			appendable.append(e.getMessage());
		}
		if (SealInternalApi.isOces2Certificate(cert) && new DistinguishedName(cert.getSubjectDN().getName()).getSubjectSerialNumber() != null) {
			OcesInfo ocesInfo = OcesInfoParser.parseCertificate(cert);
			OcesInfo.CertificateType certificateType = ocesInfo.getCertificateType();
			appendable.append(", ").append(certificateType.toString());
			if (certificateType == OcesInfo.CertificateType.MOCES) {
				String cpr = getCpr(ocesInfo);
				appendable.append("(cvrrid->cpr = ").append(cpr).append(')');
			}
		}
		appendable.append('\n');
		if (SealInternalApi.isEndCertificate(cert)) {
			URI intermediateURI = SealInternalApi.getIntermediateCertificateURL(cert);
			appendable.append("crlUrl=" + getCRLUrlFromCertificate(cert)).append('\n');
			appendable.append("intermediateUri=" + intermediateURI.toString()).append('\n');
			appendable.append("intermediateCertificate:-----------------------------------\n");
			try {
				byte[] rawCertificate = new HttpCertificateLoader().loadCertificate(intermediateURI.toString());
				X509Certificate intermediate = CertificateParser.asCertificate(rawCertificate);
				appendDescription(intermediate, appendable);
			} catch (PKIException e) {
				appendable.append(e.getMessage()).append('\n');
			}
		} else if (SealInternalApi.isIntermediateCertificate(cert)) {
			appendable.append("rootCertificate:-----------------------------------\n");
			X509Certificate root = new OCESCertificateResolver((IntermediateCertificateCache)null).getIssuingCertificate(cert);
			appendDescription(root, appendable);
		}
	}

	private static String getCpr(OcesInfo ocesInfo)  {
		try {
			CprNr relatedCpr = initService().findRelatedCpr(ocesInfo);
			return relatedCpr != null ? relatedCpr.getValue() : null;
		} catch (Exception e) {
			return e.getMessage();
		}
	}

	private static WsOcesCvrRidService initService() throws MalformedURLException {
		WsOcesCvrRidService service  = new WsOcesCvrRidService(new ServiceInfoCollector(0));
		service.setEndpoint(new URL(SERVICE_ENDPOINT_PP));
		service.setTrustStore("src/main/config/testtdc.keystore");
		service.setTrustStorePassword("gnyffe");
		service.setKeyStore("src/main/config/nsp-test-rid.jks");
		service.setKeyStorePassword("Test1234");
		service.init();
		return service;
	}


	private static String getCRLUrlFromCertificate(X509Certificate cert) {
		byte[] extensionValue = cert.getExtensionValue(org.bouncycastle.asn1.x509.X509Extension.cRLDistributionPoints.getId());
		if(extensionValue == null) {
			return null;
		} else {
			ASN1Sequence seq;
			try {
				DEROctetString oct = (DEROctetString)(new ASN1InputStream(new ByteArrayInputStream(extensionValue))).readObject();
				seq = (ASN1Sequence)(new ASN1InputStream(oct.getOctets())).readObject();
			} catch (IOException var13) {
				return null;
			}

			CRLDistPoint distPoint = CRLDistPoint.getInstance(seq);

			for(DistributionPoint distributionPoint : distPoint.getDistributionPoints()) {
				if(distributionPoint.getDistributionPoint().getType() == 0) {
					GeneralNames generalNames = (GeneralNames) distributionPoint.getDistributionPoint().getName();
					for(GeneralName name  : generalNames.getNames()) {
						if(name.getTagNo() == 6) {
							return DERIA5String.getInstance(name.getName().toASN1Primitive()).getString();
						}
					}
				}
			}

			return null;
		}
	}


}
