package dk.sosi.sts.server.test;

import dk.nsp.test.idp.NspSoapClient;
import dk.nsp.test.idp.builder.Builders;
import dk.nsp.test.idp.builder.CitizenBuilder;
import dk.nsp.test.idp.model.CitizenIdentity;
import dk.nsp.test.idp.model.Identity;
import dk.nsp.test.idp.vault.RemoteCredentialVault;
import org.apache.commons.io.IOUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPException;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URI;
import java.util.Optional;
import java.util.function.Consumer;

import static dk.sosi.sts.XmlUtil.toXml;

public class NspSoap {

    public final static String minlogAudience = "https://minlog";

    public static CitizenIdentity citizenIdentity(String cpr) {
        return citizenIdentity(cpr, minlogAudience);
    }

    public static CitizenIdentity citizenIdentity(String cpr, String audience) {
        return citizenIdentity(Builders.citizen().identifiedBy(cpr), audience);
    }

    public static CitizenIdentity citizenIdentityOnBehalfOf(String cpr, String onBehalfOf) {
        return citizenIdentity(Builders.citizen()
                                       .identifiedBy(cpr)
                                       .actingOnBehalfOf(onBehalfOf),
                               minlogAudience);
    }

    public static CitizenIdentity citizenIdentityOnBehalfOf(String cpr, String onBehalfOf, String audience) {
        return citizenIdentity(Builders.citizen()
                                       .identifiedBy(cpr)
                                       .actingOnBehalfOf(onBehalfOf),
                               audience);
    }

    public static CitizenIdentity citizenIdentity(CitizenBuilder citizen, String audience) {
        return Builders.citizenIdentity()
                .representing(citizen)
                .usedBy(Builders.serviceConsumer()
                                .named("Service Consumer Test")
                                .identifiedBy(new RemoteCredentialVault(
                                        URI.create("https://www.nspop.dk/download/attachments/190481050" +
                                                           "/NSP_Test_Service_Consumer_sds.p12"),
                                        "Test1234".toCharArray(),
                                        true)))
                .verifiedBy(Builders.identityProvider()
                                    .named("https://idp.test.nspop.dk")
                                    .withAudience("https://bootstrap.sts.nspop.dk/")
                                    .identifiedBy(new RemoteCredentialVault(
                                            URI.create(
                                                    "https://www.nspop.dk/download/attachments/190481050" +
                                                            "/NSP_Test_Identity_Provider_sds.p12"),
                                            "Test1234".toCharArray(),
                                            true)))
                .using(Builders.securityTokenService()
                               .at(URI.create("http://test1.ekstern-test.nspop.dk:8080/sts/services/Bst2Idws")))
                .consumedBy(Builders.serviceProvider().withAudience(audience))
                .build();
    }

    public static <T> T fromNspSoapResponse(NspSoapResponseData response, Class<T> c)
            throws SOAPException, IOException, JAXBException {
        final Document message = MessageFactory
                .newInstance()
                .createMessage(null, new ByteArrayInputStream(response.content.getBytes()))
                .getSOAPBody()
                .extractContentAsDocument();
        return JAXBContext.newInstance(c)
                .createUnmarshaller()
                .unmarshal(message, c)
                .getValue();
    }

    public static class NspSoapResponseData {
        public final String content, statusMessage;
        public final int status;
        public final boolean isFault;

        public NspSoapResponseData(String content, int status, String statusMessage, boolean isFault) {
            this.content = content;
            this.status = status;
            this.statusMessage = statusMessage;
            this.isFault = isFault;
        }

        public static NspSoapResponseData from(NspSoapClient.NspSoapResponse res) {
            try {
                return new NspSoapResponseData(IOUtils.toString(res.getResponse()),
                                               res.getStatus(),
                                               res.getStatusMessage(),
                                               res.isFault());
            } catch (IOException e) { throw new RuntimeException(e); }
        }
    }

    public static <T> NspSoapResponseData executeNspRequest(URI uri, String soapAction, Identity id, T req,
                                                            Element... hs)
            throws JAXBException, ParserConfigurationException, IOException, SAXException {
        return executeNspRequest(uri, soapAction, Optional.of(id), toXml(req).getDocumentElement(), u -> {}, hs);
    }

    public static <T> NspSoapResponseData executeNspRequest(URI uri, String soapAction, Identity id, JAXBElement<T> req,
                                                            Element... hs)
            throws JAXBException, ParserConfigurationException, IOException, SAXException {
        return executeNspRequest(uri, soapAction, Optional.of(id), toXml(req).getDocumentElement(), u -> {}, hs);
    }

    public static <T> NspSoapResponseData executeNspRequest(URI uri, String soapAction, Optional<Identity> id,
                                                            JAXBElement<T> req, Element... hs)
            throws JAXBException, ParserConfigurationException, IOException, SAXException {
        return executeNspRequest(uri, soapAction, id, toXml(req).getDocumentElement(), u -> {}, hs);
    }

    public static <T> NspSoapResponseData executeNspRequest(URI uri, String soapAction, Identity id, T req,
                                                            Consumer<String> logger, Element... hs)
            throws JAXBException, ParserConfigurationException, IOException, SAXException {
        return executeNspRequest(uri, soapAction, Optional.of(id), toXml(req).getDocumentElement(), logger, hs);
    }

    public static <T> NspSoapResponseData executeNspRequest(URI uri, String soapAction, Identity id, JAXBElement<T> req,
                                                            Consumer<String> logger, Element... hs)
            throws JAXBException, ParserConfigurationException, IOException, SAXException {
        return executeNspRequest(uri, soapAction, Optional.of(id), toXml(req).getDocumentElement(), logger, hs);
    }

    public static <T> NspSoapResponseData executeNspRequest(URI uri, String soapAction, Optional<Identity> id,
                                                            Consumer<String> logger, JAXBElement<T> req, Element... hs)
            throws JAXBException, ParserConfigurationException, IOException, SAXException {
        return executeNspRequest(uri, soapAction, id, toXml(req).getDocumentElement(), logger, hs);
    }

    public static NspSoapResponseData executeNspRequest(URI uri, String soapAction, Optional<Identity> identity,
                                                        Element req, Consumer<String> logger, Element... hs) {
        try (NspSoapClient c = new NspSoapClient(logger)) {
            NspSoapClient.NspSoapService soapReq = c.request(uri, soapAction);
            NspSoapClient.NspSoapRequest as = identity.map(id -> {
                try { return soapReq.as(id); } catch (IOException e) { throw new RuntimeException(e); }
            }).orElse(soapReq.withoutSecurity());
            try (NspSoapClient.NspSoapResponse res = as.execute(req, hs)) { return NspSoapResponseData.from(res); }
        } catch (Exception e) { throw new RuntimeException(e); }
    }
}