/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.test;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.pki.CertificateStatusChecker;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.server.StsHashMapCertificateCacheFactory;
import dk.sosi.sts.server.facade.HttpTestClient;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

public class TestUtil {
    public static final String STS_IT_SYSTEM_NAME = "http://system.nspop.dk/sts";
	
	public static SOSIFactory createStsSosiFactory(Version version) {
		final CredentialVault stsCredentialVault = StsCredentialVaults.getStsCredentialVault(version);
		return createSOSIFactory(stsCredentialVault, false, null);
	}

	public static SOSIFactory createSOSIFactory(CredentialVault vault, boolean validate, CertificateStatusChecker csc) {
		Properties props = new Properties(System.getProperties());
		props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_ISSUER, "TESTSTS");
		props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(validate));
        Federation federation = createTestFederation(csc);
        return new SOSIFactory(federation, vault, props);
	}

    public static Federation createTestFederation() {
        return createTestFederation(null);
    }

    public static Federation createTestFederation(boolean isEmbeddedTest) {
        return createTestFederation(null, isEmbeddedTest);
    }

    public static Federation createTestFederation(CertificateStatusChecker csc) {
        return createTestFederation(csc, false);
    }

    public static Federation createTestFederation(CertificateStatusChecker csc, boolean isEmbeddedTest) {
        // Use MockFederation for local tests, since real sts certificate is only used on test environments
        return (HttpTestClient.testEnvironmentHost.contains("localhost") || isEmbeddedTest) ? createTestFederationWithSsnMatcher(csc)
                : createSosiTestFederation(csc);
    }

    private static Federation createTestFederationWithSsnMatcher(CertificateStatusChecker csc) {
        return csc == null ? new SOSITestFederation(new Properties(), "UI:DK-O:G:8d3fa047-c77e-47e4-bdd2-e91488610ce6")
                : new SOSITestFederation(new Properties(), csc, "UI:DK-O:G:8d3fa047-c77e-47e4-bdd2-e91488610ce6");
    }

    private static SOSITestFederation createSosiTestFederation(CertificateStatusChecker csc) {
        SOSITestFederation federation;
        if (csc == null) {
            federation = new SOSITestFederation(new Properties(),
                    StsHashMapCertificateCacheFactory.create());
        } else {
            federation = new SOSITestFederation(new Properties(),
                    StsHashMapCertificateCacheFactory.create(),
                    csc);
        }
        return federation;
    }

    public static GenericCredentialVault getVaultFromResource(String path, String password) {
        if (path.endsWith("pkcs12") || path.endsWith("p12")) {
            GenericCredentialVault vault = getEmptyVault(password);
            byte[] pkcs12 = readResource(path);
            vault.setSystemCredentialPair(new ByteArrayInputStream(pkcs12), password);
            return vault;
        } else {
            final Properties properties = SignatureUtil.setupCryptoProviderForJVM();
            return new ClasspathCredentialVault(properties,path, password);
        }
    }

    public static GenericCredentialVault getEmptyVault(String password) {
        return new GenericCredentialVault(SignatureUtil.setupCryptoProviderForJVM(), password);
    }


    private static byte[] readResource(String resourceName) {
        try {
            InputStream in = Thread.currentThread().getContextClassLoader().getResourceAsStream(resourceName);
            if (in == null) {
                throw new NullPointerException("Unknown resource " + resourceName);
            }
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            byte[] buffer = new byte[8192];
            int n;
            while ((n = in.read(buffer, 0, buffer.length)) != -1) {
                baos.write(buffer, 0, n);
            }
            return baos.toByteArray();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }


	public static void disableLogging() {
		LogManager.getRootLogger().setLevel(Level.FATAL);
	}

  public static void revertLogging() {
		BasicConfigurator.resetConfiguration();
	}

}
