/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sosi.sts.server.test;

import dk.sosi.sts.certificate.OcesInfo.Version;
import dk.sosi.sts.roles.authorizationdefinitions.CachingAuthorizationDefinitionsService;
import dk.sosi.sts.roles.nationalroles.MockNationalRolesService;
import dk.sosi.sts.roles.roledefinitions.CachingRoleDefinitionsService;
import dk.sosi.sts.server.UserValidationService;
import dk.sosi.sts.server.facade.AuthorizationDefinitionServiceStub;
import dk.sosi.sts.server.facade.RoleDefinitionServiceStub;
import dk.sosi.sts.service.*;
import org.apache.log4j.LogManager;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Optional;

import static org.mockito.Mockito.mock;
import static java.util.Optional.*;

public final class UserValidationServiceTestUtil {
	static class UserValidationServiceTestCache <D> {
		private final D cacheService;
		private final D stubImpl;
		
		public <CX extends D, SX extends D> UserValidationServiceTestCache(Class<CX> cache, Class<SX> stub) {
			this(null, cache, stub);
		}
		
		public <CX extends D, SX extends D> UserValidationServiceTestCache(Version version, Class<CX> cache, Class<SX> stub) {
			try {
				if(version != null) {
					stubImpl = stub.getConstructor(Version.class).newInstance(version);
				} else {
					stubImpl = stub.newInstance();
				}
				
			} catch (Throwable t) {
				LogManager.getLogger(UserValidationServiceTestCache.class).fatal("Failed to create stub:"+stub.getName());
				t.printStackTrace();
				throw new RuntimeException(t);
			}
			
			CX cacheInstance = null;
			
			Class<?>[] interfaces = stub.getInterfaces();
			for (Class<?> stubInterface : interfaces) {
				if (stub.getSimpleName().startsWith(stubInterface.getSimpleName())) {
					try {
						cacheInstance = cache.getConstructor(stub.getInterfaces()).newInstance(stubImpl);
						break;
					} catch (Throwable t) {
						LogManager.getLogger(UserValidationServiceTestCache.class).warn("Failed to create cache:" + cache.getName()+" using "+stubInterface.getName()+" in constructor");
					}
				}
			}
			
			if(cacheInstance == null) {
				LogManager.getLogger(UserValidationServiceTestCache.class).error("Failed to create cache:" + cache.getName());
			}
					
			cacheService = cacheInstance;
		}
		
		public D getRoleDefinitionService(Version version) {	
			return cacheService != null ? cacheService : stubImpl;
		}
		
	}
	
	private static final Map<Version,UserValidationServiceTestCache<RoleDefinitionService>> roleDefinitionServiceTestCaches = new LinkedHashMap<>();
	static {
		roleDefinitionServiceTestCaches.put(Version.OCES2, new UserValidationServiceTestCache<>(Version.OCES2, CachingRoleDefinitionsService.class, RoleDefinitionServiceStub.class));
		roleDefinitionServiceTestCaches.put(Version.OCES3, new UserValidationServiceTestCache<>(Version.OCES3, CachingRoleDefinitionsService.class, RoleDefinitionServiceStub.class));
	}
	private static final UserValidationServiceTestCache<AuthorizationDefinitionService> authorizationDefinitionServiceTestCache = new UserValidationServiceTestCache<>(CachingAuthorizationDefinitionsService.class, AuthorizationDefinitionServiceStub.class);
	
	private static UserValidationService createUserValidationService(
			Version version,
			Optional<OcesCvrRidService> cvrRidService,
			Optional<AuthorizationService> authorizationService,
			Optional<RoleDefinitionService> roleDefinitionService, 
			Optional<NationalRoleService> rolesService,
			Optional<AuthorizationDefinitionService> authorizationDefinitionService,
			Optional<OcesUUID2CPRService> uuid2CPRService) {
		return new UserValidationService(
				cvrRidService.orElse(mock(OcesCvrRidService.class)),
				authorizationService.orElse(mock(AuthorizationService.class)),
				roleDefinitionService.orElse(roleDefinitionServiceTestCaches.get(version).getRoleDefinitionService(version)), 
				rolesService.orElse(new MockNationalRolesService()),
				authorizationDefinitionService.orElse(authorizationDefinitionServiceTestCache.getRoleDefinitionService(version)),
				uuid2CPRService.orElse(mock(OcesUUID2CPRService.class)));
	}
	public static UserValidationService createUserValidationService(Version version, OcesCvrRidService cvrRidService, OcesUUID2CPRService uuid2CPRService) {
		return createUserValidationService(
				version,
				Optional.ofNullable(cvrRidService),
				Optional.empty(),
				Optional.empty(),
				Optional.empty(),
				Optional.empty(),
				Optional.ofNullable(uuid2CPRService));
	}
	public static UserValidationService createUserValidationService(Version version, OcesUUID2CPRService uuidService) {
		return createUserValidationService(
				version,
				Optional.empty(),
				Optional.empty(),
				Optional.empty(),
				Optional.empty(),
				Optional.empty(),
				Optional.ofNullable(uuidService));
	}
	
	public static UserValidationService createUserValidationService(Version version, OcesCvrRidService cvrRidService, AuthorizationService authorizationService, OcesUUID2CPRService uuid2CPRService) {
		return createUserValidationService(
				version,
				Optional.ofNullable(cvrRidService),
				Optional.ofNullable(authorizationService),
				Optional.empty(),
				Optional.empty(),
				Optional.empty(),
				Optional.ofNullable(uuid2CPRService));
	}
	
	public static UserValidationService createUserValidationService(Version version, AuthorizationService authorizationService) {
		return createUserValidationService(
				version,
				Optional.empty(),
				Optional.ofNullable(authorizationService),
				Optional.empty(),
				Optional.empty(),
				Optional.empty(),
				Optional.empty());
	}

	public static UserValidationService createUserValidationService(Version version, AuthorizationService authorizationService, NationalRoleService rolesService) {
		return createUserValidationService(
				version,
				Optional.empty(),
				Optional.ofNullable(authorizationService),
				Optional.empty(),
				Optional.ofNullable(rolesService),
				Optional.empty(),
				Optional.empty());
	}
	
	public static UserValidationService createUserValidationService(Version version, AuthorizationService authorizationService, NationalRoleService rolesService, RoleDefinitionService roleDefinitionService) {
		return createUserValidationService(
				version,
				empty(),
				ofNullable(authorizationService),
				of(roleDefinitionService),
				ofNullable(rolesService),
				empty(),
				empty());
	}
	
	public static UserValidationService createUserValidationService(Version version, OcesCvrRidService cvrRidService, AuthorizationService authorizationService, NationalRoleService nationalRoleService, OcesUUID2CPRService uuidService) {
		return createUserValidationService(
				version,
				Optional.ofNullable(cvrRidService),
				Optional.ofNullable(authorizationService),
				Optional.empty(),
				Optional.ofNullable(nationalRoleService),
				Optional.empty(),
				Optional.ofNullable(uuidService));
	}

    public static UserValidationService createUserValidationService(Version version, NationalRoleService nationalRoleService, OcesUUID2CPRService uuidService) {
        return createUserValidationService(
                version,
                Optional.empty(),
                Optional.empty(),
                Optional.empty(),
                Optional.ofNullable(nationalRoleService),
                Optional.empty(),
                Optional.ofNullable(uuidService));
    }
}
