/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.util;

import java.io.IOException;
import java.util.Properties;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.SimpleHttpConnectionManager;
import org.apache.commons.httpclient.methods.GetMethod;

public class STSHttpClient {
	
	public static final String CONNECTION_MANAGER_ALWAYS_CLOSE 	= "connMan.alwaysClose";
	public static final String PARAM_CONNECTION_TIMEOUT 		= "connMan.param.connTimeout";
	public static final String URI 								= "method.URI";
	public static final String FOLLOW_REDIRECTS					= "method.followRedirects";
	
	private static final ThreadLocal<Properties> DEFAULT_PROPERTIES = new ThreadLocal<Properties>() {
		protected Properties initialValue() {
			Properties props = new Properties();
			props.put(CONNECTION_MANAGER_ALWAYS_CLOSE, "true");
			props.put(PARAM_CONNECTION_TIMEOUT, "4000");
			props.put(FOLLOW_REDIRECTS, "true");
			return props;
		};
	};
	
	public static class Response {
		private String uri;
		private int statusCode;
		private String message;
		
		public String getURI() {
			return uri;
		}
		
		public int getStatusCode() {
			return statusCode;
		}
		
		public String getMessage() {
			return message;
		}
		
		@Override
		public String toString() {
			return "URI="+uri+", statusCode="+statusCode+", message="+message;
		}
	}
	
	public static Response executeGet(String url) throws HttpException, IOException {
		Properties properties = new Properties();
		properties.put(URI, url);
		return executeGet(properties);
	}
	
	public static Response executeGet(Properties customProperties) throws HttpException, IOException {
		Properties config = new Properties(DEFAULT_PROPERTIES.get());
		config.putAll(customProperties);
		
		Boolean alwaysClose = Boolean.valueOf(config.getProperty(CONNECTION_MANAGER_ALWAYS_CLOSE));
		SimpleHttpConnectionManager httpConnectionManager = new SimpleHttpConnectionManager(alwaysClose);
		
		Integer connectionTimeout = Integer.valueOf(config.getProperty(PARAM_CONNECTION_TIMEOUT));
		httpConnectionManager.getParams().setConnectionTimeout(connectionTimeout);
		
		HttpClient httpClient = new HttpClient(httpConnectionManager);	
		
		HttpMethod method = null;
		try {
			String uri = config.getProperty(URI);
			method = new GetMethod(uri);
			
			Boolean followRedirects = Boolean.valueOf(config.getProperty(FOLLOW_REDIRECTS));
			method.setFollowRedirects(followRedirects);
			
			Response response = new Response();
			response.uri = uri;
			response.statusCode = httpClient.executeMethod(method);
			response.message = method.getResponseBodyAsString();
			
			return response;
		} finally {
			if(method != null) {
				method.releaseConnection();
			}
		}
	}
}