/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.sts.server.war;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.sts.server.*;
import junit.framework.TestCase;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.FileSystemXmlApplicationContext;
import org.springframework.core.io.Resource;
import org.springframework.jdbc.datasource.embedded.EmbeddedDatabaseBuilder;
import org.springframework.mock.jndi.SimpleNamingContextBuilder;

import javax.naming.NamingException;
import javax.sql.DataSource;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Arrays;

/**
 *
 */
public class SpringTester extends TestCase{

	private static SimpleNamingContextBuilder builder;
	
	@SuppressWarnings("resource")
	private static DataSource createDataSource(String sqlFolder, String dbname) throws IOException {
		final Resource sqlRoot = new FileSystemXmlApplicationContext().getResource(sqlFolder);
		File[] sqlFiles = sqlRoot.getFile().listFiles(new FilenameFilter() {		
			@Override
			public boolean accept(File dir, String name) {
				return name.endsWith(".sql");
			}
		});		
		Arrays.sort(sqlFiles);
		EmbeddedDatabaseBuilder embeddedDatabaseBuilder = new EmbeddedDatabaseBuilder().setName(dbname);
		
		for (File sqlFile : sqlFiles) {
			embeddedDatabaseBuilder.addScript("file:"+sqlFile.getAbsolutePath());
		}
		
		return embeddedDatabaseBuilder.build();		
	}
	
	public static final SimpleNamingContextBuilder getNamingContextBuilderWithDataSources() throws IOException, NamingException{
		if(builder == null){
			builder = SimpleNamingContextBuilder.emptyActivatedContextBuilder();
			builder.bind("java:jboss/datasources/STS", createDataSource("etc/sql/hsql/sts-db", "STS"));
			builder.bind("java:jboss/datasources/CRASTS", createDataSource("etc/sql/hsql/cra-db", "CRASTS"));
		}
		return builder;
	}
	
	private static final ThreadLocal<ApplicationContext> APPLICATION_CTX = new ThreadLocal<ApplicationContext>() {
		@Override
		protected ApplicationContext initialValue() {
			try {
				return new FileSystemXmlApplicationContext("src/test/resources/etc/embedded/config.xml");
			} catch (RuntimeException t) {
				t.printStackTrace();
				throw t;
			}
		}
	};
	
	public static ApplicationContext getApplicationContext() {
		System.setProperty("catalina.home", "target/log/");
		System.setProperty("jboss.server.log.dir", "target/log/");
		System.setProperty("jboss.server.config.dir", "etc/examples");
		return APPLICATION_CTX.get();
	}

	public void testSpring() throws Exception {
		getNamingContextBuilderWithDataSources().activate();
		
		ApplicationContext appCtx = getApplicationContext();

		final SOSIFactory sosiFactory = appCtx.getBean("sosiFactory", SOSIFactory.class);

		sosiFactory.getCredentialVault();

		check(NboRequestHandler.class, appCtx);
		check(StsRequestHandler.class, appCtx, "stsRequestHandler","stsRequestHandler2");
		//check(IdentityTokenRequestHandler.class, appCtx);
		check(IboRequestHandler.class, appCtx);
		check(SboRequestHandler.class, appCtx);
		check(Bootstrap2IdwsRequestHandler.class, appCtx);
		check(JWT2IdwsRequestHandler.class, appCtx);
	}

	public<S1,S2, T extends AbstractStsRequestHandler<S1, S2, ?>> void check(Class<T> t, final ApplicationContext appCtx) {
		this.check(t, appCtx, new String[] {});
	}
	
	public<S1,S2, T extends AbstractStsRequestHandler<S1, S2, ?>> void check(Class<T> t, final ApplicationContext appCtx, final String... names) {
		if(names != null && names.length > 0) {
			for (String name : names) {
				final T bean = name != null ? appCtx.getBean(name, t) : appCtx.getBean(t);
				assertNotNull(bean.getSosiFactory());		
			}
		} else {
			assertNotNull(appCtx.getBean(t).getSosiFactory());		
		}
	}

}
